/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Bumper.hh"
#include "SensorObject.hh"
#include "RendererRelay.hh"
#include "serialization/DeserializationInterface.hh"


using namespace peekabot;
using namespace peekabot::renderer;


/**
 * \page bumper Bumpers (DEPRECATED)
 *
 * \attention The sensor visualization framework was deprecated in 0.6, and
 * support will be removed in 0.7.
 *
 * The type identifier for bumper sensors is \b \c bumper.
 *
 * Bumpers visualize collisions by changing the color of all child objects.
 *
 * \section rangefinder_data Input data
 *
 * Bumpers take a single boolean as input, where \c true denotes a collision.
 *
 * \section rangefinder_params Parameters
 *
 * The parameter's type is shown in parantheses after the parameter name.
 *
 * \par active_color (color)
 * The color to use to indicate collision.
 * \par
 * <em>Default:</em> (1,0,0)
 *
 * \section rangefinder_examples Examples
 *
 * Below is an example of how to define a bumper in your robot 
 * definition file:
 *
 * \code
 * <sensor type="bumper">
 *   <name>bumper1</name>
 *
 *   <transform>
 *     <translate system="world">0.12 0 0</translate>
 *   </transform>
 *
 *   <children>
 *     <cube>
 *       <scale>0.02 0.1 0.05</scale>
 *     </cube>
 *   </children>
 *
 *   <params>
 *     <color name="active_color" r="0.8" g="0" b="0"/>
 *   </params>
 * </sensor>
 * \endcode
 *
 * In the client, feed the sensor data by doing:
 *
 * \code
 * bool bumper_collision = ...;
 * peekabot::SensorData s;
 * scan.write(bumper_collision);
 * proxy.set_sensor_data(handle_to_your_sensor, s);
 * \endcode
 */


static const RGBColor DEFAULT_COLOR(1,0,0);


Bumper::Bumper() throw()
    : m_prev_state(false)
{
}


void Bumper::validate_params()
{
    check_param_type<RGBColor>("active_color");
}


void Bumper::interpret_sensor_data(
    SensorObject *sensor, DeserializationInterface &sensor_data) 
    throw(std::exception)
{
    bool bumper_state;
    sensor_data >> bumper_state;

    if( bumper_state == m_prev_state )
        return;
    
    if( bumper_state == true )
    {
        enable_bumper(sensor);
    }
    else
    {
        disable_bumper(sensor);
        m_overriden.clear();
    }

    m_prev_state = bumper_state;
}

void Bumper::enable_bumper(SceneObject *obj) throw()
{
    m_overriden.insert(std::make_pair(obj, obj->get_color()));

    obj->set_color(
        get_param<RGBColor>("active_color", DEFAULT_COLOR), false);
    
    const SceneObject::ChildContainer &children = obj->get_children();
    for( SceneObject::ChildContainer::const_iterator it = children.begin(); 
         it != children.end(); ++it )
    {
        enable_bumper(*it);
    }
}

void Bumper::disable_bumper(SceneObject *obj) throw()
{
    OverrideMap::iterator it = m_overriden.find(obj);

    if( it != m_overriden.end() )
    {
        it->first->set_color(it->second, false);
    }
    
    const SceneObject::ChildContainer &children = obj->get_children();
    for( SceneObject::ChildContainer::const_iterator it = children.begin(); 
         it != children.end(); ++it )
    {
        disable_bumper(*it);
    }
}

