/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/bind.hpp>
#include <cstring>
#include <limits>

#include "GeometricPrimitives.hh"
#include "ObjectVisitor.hh"
#include "ScopedHandler.hh"
#include "ScopedMap.hh"

using namespace peekabot;

// ----------------- SphereObject implementation -------------------

HandlerInformer SphereObject::ms_handler_informer(
    "sphere",
    &SphereObject::start_handler);

SphereObject::SphereObject() throw()
    : SceneObject("sphere")
{
}

SphereObject::SphereObject(ScopedHandler* handler)
    : SceneObject("sphere", handler),
      ScalableObject(handler)
{
}

SphereObject::~SphereObject() throw()
{

}

void SphereObject::accept(ObjectVisitor* visitor) throw()
{
    visitor->visit(this);
}

void SphereObject::start_handler(
    const std::string & name,
    ScopedHandler::AttributeMap &attributes,
    ScopedHandler *handler) throw()
{
    // Creating a scene object by passing a ScopedHandler will cause it to
    // enter a new scope with all registered tag start handlers plus any
    // specific handlers for SceneObject properties which are registered
    // by the parent constructor.
    SceneObject* tmp = new SphereObject(handler);
    
    // Set the new object as the current object
    ScopedMap & variables = handler->get_variables();
    variables.push_variable("current_object", tmp);
}

// ----------------- CircleObject implementation -------------------

HandlerInformer CircleObject::ms_handler_informer(
    "circle",
    &CircleObject::start_handler);

CircleObject::CircleObject() throw()
    : SceneObject("circle")
{
}

CircleObject::CircleObject(ScopedHandler* handler)
    : SceneObject("circle", handler),
      ScalableObject(handler),
      LineBased(handler)
{
}

CircleObject::~CircleObject() throw()
{
}

void CircleObject::accept(ObjectVisitor* visitor) throw()
{
    visitor->visit(this);
}

void CircleObject::start_handler(
    const std::string & name,
    ScopedHandler::AttributeMap &attributes,
    ScopedHandler *handler) throw()
{
    // Creating a scene object by passing a ScopedHandler will cause it to
    // enter a new scope with all registered tag start handlers plus any
    // specific handlers for SceneObject properties which are registered
    // by the parent constructor.
    SceneObject* tmp = new CircleObject(handler);
    
    // Set the new object as the current object
    ScopedMap & variables = handler->get_variables();
    variables.push_variable("current_object", tmp);
}


// ----------------- CubeObject implementation -------------------

HandlerInformer CubeObject::ms_handler_informer(
    "cube",
    &CubeObject::start_handler);


CubeObject::CubeObject() throw()
    : SceneObject("cube")
{
}


CubeObject::CubeObject(ScopedHandler* handler)
    : SceneObject("cube", handler),
      ScalableObject(handler)
{
}

CubeObject::~CubeObject() throw()
{
}

void CubeObject::accept(ObjectVisitor* visitor) throw()
{
    visitor->visit(this);
}

void CubeObject::start_handler(
    const std::string & name,
    ScopedHandler::AttributeMap &attributes,
    ScopedHandler *handler) throw()
{
    // Creating a scene object by passing a ScopedHandler will cause it to
    // enter a new scope with all registered tag start handlers plus any
    // specific handlers for SceneObject properties which are registered
    // by the parent constructor.
    SceneObject* tmp = new CubeObject(handler);
    
    // Set the new object as the current object
    ScopedMap & variables = handler->get_variables();
    variables.push_variable("current_object", tmp);
}

// ----------------- CylinderObject implementation -------------------


HandlerInformer CylinderObject::ms_handler_informer(
    "cylinder",
    &CylinderObject::start_handler);

CylinderObject::CylinderObject() throw()
    : SceneObject("cylinder")
{
}

CylinderObject::CylinderObject(ScopedHandler* handler)
    : SceneObject("cylinder", handler),
      ScalableObject(handler)
{
}


CylinderObject::~CylinderObject() throw()
{
}

void CylinderObject::accept(ObjectVisitor* visitor) throw()
{
    visitor->visit(this);
}


void CylinderObject::start_handler(
    const std::string & name,
    ScopedHandler::AttributeMap &attributes,
    ScopedHandler *handler) throw()
{
    // Creating a scene object by passing a ScopedHandler will cause it to
    // enter a new scope with all registered tag start handlers plus any
    // specific handlers for SceneObject properties which are registered
    // by the parent constructor.
    SceneObject* tmp = new CylinderObject(handler);
    
    // Set the new object as the current object
    ScopedMap & variables = handler->get_variables();
    variables.push_variable("current_object", tmp);
}



// ----------------- PolygonObject implementation -------------------


HandlerInformer PolygonObject::ms_handler_informer(
    "polygon", 
    &PolygonObject::start_handler);

PolygonObject::PolygonObject() throw()
    : SceneObject("polygon"),
      VertexObject(0, std::numeric_limits<size_t>::max())
{
}


PolygonObject::PolygonObject(const Vertices &vertices) throw()
    : SceneObject("polygon"),
      VertexObject(0, std::numeric_limits<size_t>::max())
{
    set_vertices(vertices);
}

PolygonObject::PolygonObject(ScopedHandler* handler)
    : SceneObject("polygon", handler),
      ScalableObject(handler),
      VertexObject(0, std::numeric_limits<size_t>::max(), handler)
{
}

PolygonObject::~PolygonObject() throw()
{
}

void PolygonObject::accept(ObjectVisitor* visitor) throw()
{
    visitor->visit(this);
}

void PolygonObject::start_handler(
    const std::string & name,
    ScopedHandler::AttributeMap &attributes,
    ScopedHandler *handler) throw()
{
    // Creating a scene object by passing a ScopedHandler will cause it to
    // enter a new scope with all registered tag start handlers plus any
    // specific handlers for SceneObject properties which are registered
    // by the parent constructor.
    SceneObject* tmp = new PolygonObject(handler);
    
    // Set the new object as the current object
    ScopedMap & variables = handler->get_variables();
    variables.push_variable("current_object", tmp);
}
