/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GEOMETRIC_PRIMITIVES_HH
#define __PEEKABOT_GEOMETRIC_PRIMITIVES_HH


#include "SceneObject.hh"
#include "ScalableObject.hh"
#include "VertexObject.hh"
#include "LineBased.hh"
#include "Vector.hh"
#include "HandlerInformer.hh"


namespace peekabot
{
    class ScopedHandler;

    /**
     * \internal
     *
     * \brief Class representing the cube geometric primitive.
     *
     * A primitive object inheriting the colorable and scalable properties.
     *
     * \sa ColorableObject, ScalableObject
     */
    class CubeObject : public SceneObject, public ScalableObject
    {
    public:
        CubeObject() throw();

        /** \brief Construct object from XML data.
         */
        CubeObject(ScopedHandler *handler);
        virtual ~CubeObject() throw();

        /** \brief Method for accepting \c ObjectVisitor objects.
         *
         * This method performs an automatic type cast by calling the \c visit() method of the visitor, allowing it to perform type-dependent task.
         */
        virtual void accept(ObjectVisitor* visitor) throw();

        /** \brief Element start handler for XML parsing.
         */
        static void start_handler(
            const std::string & name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();
    private:
        static HandlerInformer ms_handler_informer;
    };




    /**
     * \internal
     *
     * \brief Class representing the cylinder geometric primitive.
     *
     * A primitive object inheriting the colorable and scalable properties.
     *
     * \sa ColorableObject, ScalableObject, TessellatedObject
     */
    class CylinderObject : public SceneObject, public ScalableObject
    {
    public:
        CylinderObject() throw();

        /** \brief Construct object from XML data.
         */
        CylinderObject(ScopedHandler *handler);

        virtual ~CylinderObject() throw();

        /** \brief Method for accepting \c ObjectVisitor objects.
         *
         * This method performs an automatic type cast by calling the \c visit() method of the visitor, allowing it to perform type-dependent task.
         */
        virtual void accept(ObjectVisitor* visitor) throw();

        /** \brief Element start handler for XML parsing.
         */
        static void start_handler(
            const std::string & name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();
    private:
        static HandlerInformer ms_handler_informer;
    };



    /**
     * \internal
     *
     * \brief Class representing the sphere geometric primitive.
     *
     * A primitive object inheriting the colorable and scalable properties.
     *
     * \sa ColorableObject, ScalableObject, TessellatedObject
     */
    class SphereObject : public SceneObject, public ScalableObject
    {
    public:
        SphereObject() throw();

        /** \brief Construct object from XML data.
         */
        SphereObject(ScopedHandler *handler);

        virtual ~SphereObject() throw();

        /** \brief Method for accepting \c ObjectVisitor objects.
         *
         * This method performs an automatic type cast by calling the \c visit() method of the visitor, allowing it to perform type-dependent task.
         */
        virtual void accept(ObjectVisitor* visitor) throw();

        /** \brief Element start handler for XML parsing.
         */
        static void start_handler(
            const std::string & name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();
    private:
        static HandlerInformer ms_handler_informer;
    };



    /**
     * \internal
     *
     * \brief Class representing the circle geometric primitive.
     *
     * A primitive object inheriting the colorable and scalable properties.
     *
     * \sa ColorableObject, ScalableObject, TessellatedObject
     */
    class CircleObject : public SceneObject, 
                         public ScalableObject, 
                         public LineBased
    {
    public:
        CircleObject() throw();

        /** \brief Construct object from XML data.
         */
        CircleObject(ScopedHandler *handler);

        virtual ~CircleObject() throw();

        /** \brief Method for accepting \c ObjectVisitor objects.
         *
         * This method performs an automatic type cast by calling the \c visit() method of the visitor, allowing it to perform type-dependent task.
         */
        virtual void accept(ObjectVisitor* visitor) throw();

        /** \brief Element start handler for XML parsing.
         */
        static void start_handler(
            const std::string & name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

    private:
        real_unit m_radius;

        static HandlerInformer ms_handler_informer;
    };




    /**
     * \internal
     *
     * \brief Class representing the polyline geometric primitive.
     *
     * A primitive object inheriting the colorable and scalable properties. It is specified by a list of vertices, where the order in the list is the same as the order in which the line is drawn.
     *
     * \sa ColorableObject, ScalableObject
     */
    class PolyLineObject : public SceneObject,
                           public ScalableObject,  
                           public VertexObject,
                           public LineBased
    {
    public:
        /** \brief Creates a new \c PolyLineObject from a list of coordinate vectors.
         */
        PolyLineObject(const std::list<Vector4ru> &vertex_list);


        /** \brief Construct object from XML data.
         */
        PolyLineObject(ScopedHandler *handler);

        virtual ~PolyLineObject() throw();

        /** \brief Method for accepting \c ObjectVisitor objects.
         *
         * This method performs an automatic type cast by calling the \c visit() method of the visitor, allowing it to perform type-dependent task.
         */
        virtual void accept(ObjectVisitor* visitor) throw();

        /** \brief Element start handler for XML parsing.
         */
        static void start_handler(
            const std::string & name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();
    private:
        static HandlerInformer ms_handler_informer;
    };





    class PolygonObject : public SceneObject, 
                          public ScalableObject, 
                          public VertexObject
    {
    public:
        /** \brief Creates a polygon with four corners as specified.
         */
        PolygonObject() throw();

        PolygonObject(const Vertices &vertices) throw();


        /** \brief Construct object from XML data.
         */
        PolygonObject(ScopedHandler *handler);

        virtual ~PolygonObject() throw();

        /** \brief Method for accepting \c ObjectVisitor objects.
         *
         * This method performs an automatic type cast by calling the \c visit() method of the visitor, allowing it to perform type-dependent task.
         */
        virtual void accept(ObjectVisitor* visitor) throw();

        /** \brief Element start handler for XML parsing.
         */
        static void start_handler(
            const std::string & name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

    private:
        static HandlerInformer ms_handler_informer;
    };


}

#endif // __PEEKABOT_GEOMETRIC_PRIMITIVES_HH
