/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GEOM_GEOMETRY_TOOLBOX_HH
#define __PEEKABOT_GEOM_GEOMETRY_TOOLBOX_HH


#include <cmath>
#include <cassert>


namespace peekabot
{
    namespace geom
    {
        template<class FwdIt, class VertexType>
        void calculate_optimal_bounding_sphere(
            FwdIt begin, FwdIt end, 
            VertexType &midp, float &r,
            float x_scale_factor = 1.0f,
            float y_scale_factor = 1.0f,
            float z_scale_factor = 1.0f) throw()
        {
            float max_sq_dist = 0;
            midp = VertexType(0,0,0);

            size_t count = 0;
            
            for( FwdIt it = begin; it < end; )
            {
                ++count;

                const VertexType &v1 = *it;
                FwdIt next_it = ++it;

                for( FwdIt jt = next_it; jt < end; ++jt )
                {
                    VertexType diff = v1 - (*jt);
                    diff(0) *= x_scale_factor;
                    diff(1) *= y_scale_factor;
                    diff(2) *= z_scale_factor;
                    float sq_dist = diff * diff;

                    if( sq_dist > max_sq_dist )
                    {
                        max_sq_dist = sq_dist;
                        midp = *jt;
                        midp(0) *= x_scale_factor;
                        midp(1) *= y_scale_factor;
                        midp(2) *= z_scale_factor;
                        midp += diff/2;
                    }
                }
            }

            // Special case for when there's only one vertex
            if( count == 1 )
                midp = *begin;

            r = sqrtf(max_sq_dist)/2;
        }


        template<class FwdIt, class VertexType>
        void calculate_approximate_bounding_sphere(
            FwdIt begin, FwdIt end, 
            VertexType &midp, float &r,
            float x_scale_factor = 1.0f,
            float y_scale_factor = 1.0f,
            float z_scale_factor = 1.0f) throw()
        {
            // Special case: no coordinates
            if( begin == end )
            {
                midp = VertexType(0,0,0);
                r = 0;
                return;
            }

            VertexType min[3], max[3];

            for( int i = 0; i < 3; ++i )
                min[i] = max[i] = *begin;

            FwdIt it = begin;
            for( ++it; it < end; ++it )
            {
                for( int i = 0; i < 3; ++i )
                {
                    if( (*it)(i) < min[i](i) )
                        min[i] = *it;
                    else if( (*it)(i) > max[i](i) )
                        max[i] = *it;
                }
            }


            midp = VertexType(0,0,0);
            r = -1;

            for( int i = 0; i < 3; ++i )
            {
                VertexType diff = max[i] - min[i];
                diff(0) *= x_scale_factor;
                diff(1) *= y_scale_factor;
                diff(2) *= z_scale_factor;
                float d = sqrtf(diff * diff);

                if( d > 2*r )
                {
                    r = d/2;
                    midp = min[i];
                    midp(0) *= x_scale_factor;
                    midp(1) *= y_scale_factor;
                    midp(2) *= z_scale_factor;
                    midp += diff/2;
                }
            }

            assert( r >= 0 );

            for( FwdIt it = begin; it < end; ++it )
            {
                // Transform vertex to properly scaled coordinates
                VertexType v((*it)(0)*x_scale_factor,
                             (*it)(1)*y_scale_factor,
                             (*it)(2)*z_scale_factor);
                             
                VertexType diff = v - midp;
                float sq_d = diff * diff;

                if( sq_d > r*r+0.0000000001 )
                {
                    float d = sqrtf(sq_d);
                    r = (d + r)/2;
                    midp += (d-r)/d * diff;
                }
            }
        }
    }
}


#endif // __PEEKABOT_GEOM_GEOMETRY_TOOLBOX_HH
