/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GRID_OBJECT_HH
#define __PEEKABOT_GRID_OBJECT_HH


#include "SceneObject.hh"
#include "ScalableObject.hh"
#include "LineBased.hh"
#include "XMLHandler.hh"
#include "HandlerInformer.hh"


namespace peekabot
{
    class ScopedHandler;

    /**
     * \internal
     *
     * \brief A grid object.
     *
     * Enables showing radial and cartesian grids.
     */
    class GridObject : public SceneObject,
                       public ScalableObject, 
                       public LineBased
    {
    public:
        GridObject() throw();

        GridObject(ScopedHandler *handler) throw();
        
        virtual void accept(ObjectVisitor* visitor) throw();
        
        GridType get_grid_type() const throw();

        void set_grid_type(GridType grid_type) throw();

        unsigned int get_segment_count() const throw();

        void set_segment_count(uint32_t segment_count) throw();

        float get_central_angle() const throw();
        
        void set_central_angle(float central_angle) throw();

        /// \name Signals
        /// @{

        typedef boost::signal<
            void (GridObject *grid, uint32_t segments)> OnSegmentCountSetSignal;
        
        OnSegmentCountSetSignal m_on_segment_count_set;

        typedef boost::signal<
            void (GridObject *grid, GridType type)> OnGridTypeSetSignal;
        
        OnGridTypeSetSignal m_on_grid_type_set;

        typedef boost::signal<
            void (GridObject *grid, float angle)> OnCentralAngleSetSignal;
        
        OnCentralAngleSetSignal m_on_central_angle_set;

        /// @}

    private:
        /** 
         * \brief Element start handler for XML parsing.
         */
        static void start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        void type_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        void type_cdata_handler(
            const std::string &cdata,
            ScopedHandler *handler) throw();
        
        void segments_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        void segments_cdata_handler(
            const std::string &cdata,
            ScopedHandler *handler) throw();

        void central_angle_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        void radians_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        void degrees_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        void central_angle_cdata_handler(
            const std::string &cdata,
            ScopedHandler *handler,
            bool value_in_degrees) throw();

    private:
        static HandlerInformer ms_handler_informer;

        GridType m_grid_type;

        unsigned int m_segment_count;

        float m_central_angle;
    };
}

#endif // __PEEKABOT_GRID_OBJECT_HH
