/*
 * Copyright Staffan Gimåker 2006-2009.
 * Copyright Anders Boberg 2007-2008.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <boost/format.hpp>
#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/ref.hpp>
#include <limits>
#include <iostream>

#include "Joint.hh"
#include "ObjectVisitor.hh"
#include "ScopedHandler.hh"


using namespace peekabot;

//
// The limit outside which matrices must differ to treated as different
//
static const real_unit LIMIT = 0.000001;


Joint::Joint(const std::string &default_name) throw()
    : SceneObject(default_name),
      m_dof_xform(Matrix4ru::identity()),
      m_min(-std::numeric_limits<real_unit>::max()),
      m_max(std::numeric_limits<real_unit>::max()),
      m_val(0),
      m_offset(0)
{
    m_pre_dof_mtop = get_transformation();
}


Joint::Joint(const std::string &default_name, 
         ScopedHandler *handler) throw()
    : SceneObject(default_name, handler),
      m_dof_xform(Matrix4ru::identity()),
      m_min(-std::numeric_limits<real_unit>::max()),
      m_max(std::numeric_limits<real_unit>::max()),
      m_val(0),
      m_offset(0)

{
    m_pre_dof_mtop = get_transformation();
}


Joint::~Joint() throw()
{
}


real_unit Joint::get_min_value() const throw()
{
    return m_min;
}


void Joint::set_min_value(real_unit min) throw(std::domain_error)
{
    if( min > get_max_value() )
        throw std::domain_error(
            "Specified minimum value is greater than the joint's maximum "
            "value (name='" + get_name() + "')");

    m_min = min;

    if( get_value() < m_min )
        set_value(m_min);

    m_on_dof_domain_set(this, min, get_max_value());
}


real_unit Joint::get_max_value() const throw()
{
    return m_max;
}


void Joint::set_max_value(real_unit max) throw(std::domain_error)
{
    if( max < get_min_value() )
        throw std::domain_error(
            "Specified maximum value is less than the joint's minimum "
            "value (name='" + get_name() + "'");

    m_max = max;

    if( get_value() > m_max )
        set_value(m_max);

    m_on_dof_domain_set(this, get_min_value(), max);
}


real_unit Joint::get_value() const throw()
{
    return m_val;
}

real_unit Joint::get_user_value() const throw()
{
    return m_user_val;
}

Matrix4ru Joint::get_full_transform(real_unit value) const throw()
{
    Matrix4ru dof_xform = calculate_dof_transform(value - m_offset);
    return m_pre_dof_mtop * dof_xform;
}

void Joint::set_value(real_unit value) throw(std::domain_error)
{
    if( value < get_min_value() || value > get_max_value() )
    {
        std::string msg = (
            boost::format(
                "The given joint value, %1%, is outside the "
                "allowed range [%2%, %3%] (name='%4%')")
            % value % get_min_value() % get_max_value() % get_name()).str();

        throw std::domain_error(msg);
    }

    m_val = value;
    m_dof_xform = calculate_dof_transform(value - m_offset);
    set_transformation(m_pre_dof_mtop * m_dof_xform);

    m_on_dof_value_set(this, value);
}


void Joint::set_value_offset(float offset) throw()
{
    m_offset = offset;
    set_value( m_val );
}


float Joint::get_value_offset() const throw()
{
    return m_offset;
}


void Joint::check_pre_transform() const throw()
{
    const Matrix4ru &mtop = get_transformation();
    
    if( !mtop.equals(m_pre_dof_mtop * m_dof_xform, LIMIT) )
    {
        // The object has been moved/rotated
        //
        // Find the object's mtop as it would have been
        // given that no DOF xform has been applied:
        m_pre_dof_mtop = mtop * m_dof_xform.inverse();
    }
}

void Joint::set_user_value(real_unit value) throw(std::domain_error)
{
    if( value < get_min_value() || value > get_max_value() )
    {
        std::string msg = (
            boost::format(
                "The given joint value, %1%, is outside "
                "the allowed range [%2%, %3%]")
            % value % get_min_value() % get_max_value()).str();

        throw std::domain_error(msg);
    }

    m_user_val = value;

    m_on_dof_user_value_set(this, value);
}

