/*
 * Copyright Staffan Gimåker 2006-2009.
 * Copyright Anders Boberg 2007.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PEEKABOT_JOINT_HH_INCLUDED
#define PEEKABOT_JOINT_HH_INCLUDED


#include <stdexcept>

#include "Types.hh"
#include "Vector.hh"
#include "Matrix.hh"
#include "SceneObject.hh"
#include "XMLHandler.hh"


namespace peekabot
{
    class ScopedHandler;


    /**
     * \internal
     *
     * \brief An interface for providing a robot with a degree of freedom, in a
     * way that's convinient to work with for the the users of both the server
     * and proxy library.
     *
     * \sa JointDOF, TrackDOF
     */
    class Joint : public SceneObject
    {
    public:
        /**
         * \brief Construct a new DOF object.
         *
         * \post <tt>get_value() == 0</tt>, <tt>get_min_value() == 0</tt>, 
         * <tt>get_max_value() == 0</tt>
         */
        Joint(const std::string &default_name) throw();

        Joint(const std::string &default_name, 
            ScopedHandler *handler) throw();

        virtual ~Joint() throw();

        /**
         * \brief Get the lower bound of allowed values (inclusive).
         */
        real_unit get_min_value() const throw();

        /**
         * \brief Set the lower bound of allowed values (inclusive).
         *
         * If the DOF's current value is less than the new lower bound,
         * the value will automatically be set to the lowest allowed value.
         *
         * \pre <tt>min <= get_max_value()</tt>
         * \post <tt>get_min_value() <= get_max_value()</tt>
         *
         * \exception std::domain_error Cast if \f$min>get\_max\_value()\f$.
         */
        void set_min_value(real_unit min) throw(std::domain_error);

        /**
         * \brief Get the upper bound of allowed values (inclusive).
         */
        real_unit get_max_value() const throw();

        /**
         * \brief Set the upper bound of allowed values (inclusive).
         *
         * If the DOF's current value is greater than the new upper bound,
         * the value will automatically be set to the largest allowed value.
         *
         * \pre <tt>max >= get_min_value()</tt>
         * \post <tt>get_min_value() <= get_max_value()</tt>
         *
         * \exception std::domain_error Cast if \f$max<get\_min\_value()\f$.
         */
        void set_max_value(real_unit max) throw(std::domain_error);

        /**
         * \brief Get the DOF's current value.
         *
         * \invariant Let \f$x\f$ be the return value, then 
         * \f$get\_min\_value() \leq x \leq get\_max\_value()\f$.
         */
        real_unit get_value() const throw();

        void set_value_offset(float offset) throw();

        float get_value_offset() const throw();

        /**
         * \brief Get the DOF's current user value.
         *
         * \invariant Let \f$x\f$ be the return value, then 
         * \f$get\_min\_value() \leq x \leq get\_max\_value()\f$.
         */
        real_unit get_user_value() const throw();

        /**
         * \brief Set the value of the DOF, actual value semantics
         * are subclass-specific.
         *
         * \exception std::domain_error Cast if the specified value lies outside
         * the \f$[get\_min\_value(),get\_max\_value()]\f$ range.
         */
        void set_value(real_unit value) throw(std::domain_error);

        /**
         * \brief Set the user value of the DOF.
         *
         * \exception std::domain_error Cast if the specified value lies outside
         * the \f$[get\_min\_value(),get\_max\_value()]\f$ range.
         */
        void set_user_value(real_unit value) throw(std::domain_error);

       /**
         * \brief Subclass-specific method for calculating how a given DOF value
         * should affect the transformation matrix.
         *
         * This method will call check_pre_mtop() to update its pre-dof
         * transformation matrix.
         */
        virtual Matrix4ru calculate_dof_transform(real_unit value)
            const throw() = 0;

        Matrix4ru get_full_transform(real_unit value)
            const throw();

        /// \name Signals
        /// @{

        typedef boost::signal<void (Joint *dof, real_unit value)> OnDOFValueSetSignal;

        OnDOFValueSetSignal m_on_dof_value_set;

        typedef boost::signal<void (Joint *dof, real_unit value)> OnDOFUserValueSetSignal;

        OnDOFUserValueSetSignal m_on_dof_user_value_set;

        typedef boost::signal<void (Joint *dof, real_unit min, real_unit max)> OnDOFDomainSetSignal;

        OnDOFDomainSetSignal m_on_dof_domain_set;

        /// @}


    protected:
        /**
         * \brief This method is used to check if the object's transformation
         * matrix has changed due to non-dof-related translation/rotation.
         */
        void check_pre_transform() const throw();


        /**
         * \brief The DOF's mtop prior to applying the DOF-transformation 
         * governed by the DOF's value.
         */
        mutable Matrix4ru m_pre_dof_mtop;


        /** 
         * \brief The DOF transform for the current DOF-value.
         *
         * We need to store the previous transform in order
         * to correctly calculate it the next time the DOF is set, or 
         * it's transformation changed.
         */
        Matrix4ru m_dof_xform;


    private:
        /**
         * \brief The lower bound (inclusive) on allowed values.
         *
         * The initial value is zero.
         */
        real_unit m_min;

        /**
         * \brief The upper bound (inclusive) on allowed values.
         *
         * The initial value is zero.
         */
        real_unit m_max;

        /**
         * \brief The DOF's current value.
         *
         * The initial value is zero.
         */
        real_unit m_val; 

        float m_offset;
        
        /**
         * \brief The DOF's user value, i.e. the value manipulated
         * by the GUI user.
         *
         * The initial value is zero.
         */
        real_unit m_user_val;
    };


}


#endif // PEEKABOT_JOINT_HH_INCLUDED
