/*
 * Copyright Staffan Gimåker 2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <fstream>

#include "LexicalMap.hh"


using namespace peekabot;


const std::string &LexicalMap::get_raw(const std::string &key) const
{
    OptionMap::const_iterator it = m_options.find(key);
    if( it == m_options.end() )
        throw std::runtime_error("Key '" + key + "' not found");
    return it->second;
}


void LexicalMap::set_raw(const std::string &key, const std::string &val)
{
    m_options[key] = val;
}


void LexicalMap::unset(const std::string &key)
{
    m_options.erase(key);
}


void LexicalMap::clear()
{
    m_options.clear();
}


bool LexicalMap::has(const std::string &key) const
{
    return m_options.find(key) != m_options.end();
}


void LexicalMap::load_ini(std::istream &is)
{
    std::string curr_section = "default";

    while( !is.eof() )
    {
        std::string str;
        std::getline(is, str);

        std::vector<std::string> components;
        boost::split( components, str, boost::is_any_of("#;") );

        std::string &line = components[0];
        boost::trim(line);


        if( line.length() == 0  )
            continue;

        if( line[0] == '[' )
        {
            // Section directive
            if( line[line.length()-1] != ']' )
                throw std::runtime_error(
                    "Malformed section directive: " + line);

            std::string section_name(line, 1, line.length()-2);
            boost::trim(section_name);

            if( section_name.empty() )
                throw std::runtime_error(
                    "Zero-length section names are not allowed: " + line);

            curr_section = section_name;
        }
        else
        {
            // Option directive
            std::vector<std::string> components;
            boost::split( components, line, boost::is_any_of("=") );

            if( components.size() != 2 )
                throw std::runtime_error(
                    "Malformed option directive: " + line);

            std::string &lhs = components[0];
            std::string &rhs = components[1];

            boost::trim(lhs);
            boost::trim(rhs);

            /*if( curr_section.empty() )
                throw std::runtime_error(
                    "Option directive encountered without "
                    "a previous section directive");*/

            set_raw(curr_section + "." + lhs, rhs);
        }
    }
}


void LexicalMap::load_ini(const std::string &filename)
{
    std::ifstream ifs(filename.c_str());
    if( !ifs )
        throw std::runtime_error(
            "Failed to open config file '" + filename + "' for reading");
    load_ini(ifs);
}


void LexicalMap::save_ini(std::ostream &os) const
{
    std::string prev_section = "default";

    for( OptionMap::const_iterator it = m_options.begin();
         it != m_options.end(); ++it )
    {
        std::string section, option;

        // Split the key into a <section-name, option-name> pair
        std::vector<std::string> components;
        boost::split( components, it->first, boost::is_any_of(".") );

        if( components.size() == 1 )
        {
            section = "default";
            option = components[0];
        }
        else
        {
            section = components[0];
            for( size_t i = 1; i < components.size(); ++i )
                option += components[i];
        }

        if( section != prev_section )
        {
            os << "\n[" << section << "]\n\n";
            prev_section = section;
        }

        os << option << " = " << it->second << "\n";
    }

    os.flush();
}


void LexicalMap::save_ini(const std::string &filename) const
{
    std::ofstream ofs(filename.c_str());
    if( !ofs )
        throw std::runtime_error(
            "Failed to open config file '" + filename + "' for writing");
    save_ini(ofs);
}
