/*
 * Copyright Staffan Gimåker 2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PEEKABOT_LEXICAL_MAP_HH_INCLUDED
#define PEEKABOT_LEXICAL_MAP_HH_INCLUDED


#include <map>
#include <string>
#include <vector>
#include <set>
#include <list>
#include <stdexcept>
#include <istream>
#include <ostream>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>


namespace peekabot
{
    namespace detail
    {
        template<class S, class T>
        inline void lexical_assign(S &lhs, const T &rhs)
        {
            lhs = boost::lexical_cast<S>(rhs);
        }

        template<>
        inline void lexical_assign(bool &lhs, const std::string &rhs)
        {
            if( rhs == "true" || rhs == "on" ||
                rhs == "enable" || rhs == "1" )
                lhs = true;
            else if( rhs == "false" || rhs == "off" ||
                     rhs == "disable" || rhs == "0" )
                lhs = false;
            else
                throw boost::bad_lexical_cast();
        }

        template<class T>
        inline void lexical_assign(std::list<T> &lhs, const std::string &rhs)
        {
            lhs.clear();

            std::vector<std::string> components;
            boost::split( components, rhs, boost::is_any_of(",") );

            for( size_t i = 0; i < components.size(); ++i )
            {
                boost::trim(components[i]);
                if( !components[i].empty() )
                {
                    T tmp;
                    lexical_assign(tmp, components[i]);
                    lhs.push_back(tmp);
                }
            }
        }

        template<class T>
        inline void lexical_assign(std::string &lhs, const std::list<T> &rhs)
        {
            lhs = "";
            for( typename std::list<T>::const_iterator it = rhs.begin();
                 it != rhs.end(); ++it )
            {
                std::string tmp;
                lexical_assign(tmp, *it);
                if( it != rhs.begin() )
                    lhs += ", ";
                lhs += tmp;
            }
        }
    }

    class LexicalMap
    {
    public:
        const std::string &get_raw(const std::string &key) const;

        template<class T>
        T get(const std::string &key) const
        {
            OptionMap::const_iterator it = m_options.find(key);
            if( it == m_options.end() )
                throw std::runtime_error("Key '" + key + "' not found");

            try
            {
                T ret;
                detail::lexical_assign(ret, it->second);
                return ret;
            }
            catch(...)
            {
                throw std::runtime_error(
                    "Malformed option '" + it->second +
                    "' for key '" + key + "'");
            }
        }

        void set_raw(const std::string &key, const std::string &val);

        template<class T>
        void set(const std::string &key, const T &val)
        {
            std::string raw;
            detail::lexical_assign(raw, val);
            set_raw(key, raw);
        }

        void unset(const std::string &key);

        void clear();

        bool has(const std::string &key) const;

        template<class T>
        bool has_type(const std::string &key) const
        {
            if( !has(key) )
            {
                throw std::runtime_error("Key '" + key + "' not found");
            }
            else
            {
                try
                {
                    T tmp;
                    detail::lexical_assign(tmp, get_raw(key));
                    return true;
                }
                catch(...)
                {
                    return false;
                }
            }
        }

        void load_ini(std::istream &is);

        void load_ini(const std::string &filename);

        void save_ini(std::ostream &os) const;

        void save_ini(const std::string &filename) const;

    protected:
        typedef std::map<std::string, std::string> OptionMap;

        OptionMap m_options;
    };
}


#endif // PEEKABOT_LEXICAL_MAP_HH_INCLUDED
