/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string/case_conv.hpp>

#include "LineBased.hh"
#include "ScopedHandler.hh"


using namespace peekabot;


LineBased::LineBased() throw()
    : m_line_width(1),
      m_line_style(LINE_STYLE_SOLID),
      m_stipple_factor(1)
{
}

LineBased::LineBased(ScopedHandler *handler)
    : m_line_width(1),
      m_line_style(LINE_STYLE_SOLID),
      m_stipple_factor(1)
{
    // Register start tag handlers

    handler->add_start_handler(
        "line_width", 
        boost::bind(
            &LineBased::line_width_start_handler, this, _1, _2, _3));

    handler->add_start_handler(
        "line_style",
        boost::bind(
            &LineBased::line_style_start_handler, this, _1, _2, _3));
}

float LineBased::get_line_width() const throw()
{
    return m_line_width;
}

void LineBased::set_line_width(float line_width) throw()
{
    m_line_width = line_width;
    m_on_line_params_set(this, m_line_width, m_line_style, m_stipple_factor);
}

std::pair<LineStyle, int> LineBased::get_line_style() const throw()
{
    return std::pair<LineStyle, int>(m_line_style, m_stipple_factor);
}

void LineBased::set_line_style(LineStyle style, int factor) throw()
{
    m_line_style = style;
    m_stipple_factor = factor;
    m_on_line_params_set(this, m_line_width, m_line_style, m_stipple_factor);
}


//
// --------------- XML handler methods --------------------
//


void LineBased::line_width_start_handler(
    const std::string &name,
    ScopedHandler::AttributeMap &attributes,
    ScopedHandler *handler) throw()
{
    ScopedHandler::TagScope scope;
    
    scope.cdata_functor = boost::bind(
        &LineBased::line_width_cdata_handler, this, _1, _2);
    
    handler->enter_scope(scope);    
}

void LineBased::line_width_cdata_handler(
    const std::string &cdata,
    ScopedHandler *handler)
{
    set_line_width(boost::lexical_cast<float>(cdata));
}

void LineBased::line_style_start_handler(
    const std::string &name,
    ScopedHandler::AttributeMap &attributes,
    ScopedHandler *handler)
{
    ScopedHandler::TagScope scope;

    int factor = boost::lexical_cast<int>(attributes["factor"]);
    
    scope.cdata_functor = boost::bind(
        &LineBased::line_style_cdata_handler, this, _1, _2, factor);
    
    handler->enter_scope(scope);
}

void LineBased::line_style_cdata_handler(
    const std::string &cdata,
    ScopedHandler *handler,
    int factor)
{
    std::string style = cdata;
    boost::to_lower(style);

    if( style == "solid" )
        set_line_style(LINE_STYLE_SOLID, factor);
    else if( style == "dashed" )
        set_line_style(LINE_STYLE_DASHED, factor);
    else if( style == "dotted" )
        set_line_style(LINE_STYLE_DOTTED, factor);
    else if( style == "dash-dot" )
        set_line_style(LINE_STYLE_DASH_DOT, factor);
    else if( style == "dash-dot-dot" )
        set_line_style(LINE_STYLE_DASH_DOT_DOT, factor);
    else
        throw std::runtime_error("Unknown line style");
}
