/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_LINE_BASED_HH
#define __PEEKABOT_LINE_BASED_HH


#include <boost/thread_safe_signal.hpp>

#include "Types.hh"
#include "XMLHandler.hh"


namespace peekabot
{

    class ScopedHandler;

    /**
     * \internal
     *
     * \brief An abstraction of objects in the scene that are scalable.
     *
     * A non-instantiable class that adds to scene objects the scale 
     * property and methods for manipulating it. Classes that need 
     * this property inherit it from this class.
     */
    class LineBased
    {
    public:
        LineBased() throw();

        /**
         * \brief Construct a line-based object using an XML document handler.
         */
        LineBased(ScopedHandler *handler);


        float get_line_width() const throw();

        void set_line_width(float line_width) throw();


        std::pair<LineStyle, int> get_line_style() const throw();

        void set_line_style(LineStyle style, int factor) throw();


        /// \name Signals
        /// @{

        boost::signal<
            void (LineBased *object, float line_width, LineStyle line_pattern, int pattern_factor)> m_on_line_params_set;

        /// @}


    private:
        /// \name XML handler methods
        /// @{

        /**
         * \brief <line_width> start handler for XML parsing.
         */
        void line_width_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        /**
         *\brief <line_width> cdata handler for XML parsing.
         */
        void line_width_cdata_handler(
            const std::string &cdata,
            ScopedHandler *handler);

        /**
         * \brief <line_style> start handler for XML parsing.
         */
        void line_style_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler);

        /**
         *\brief <line_style> cdata handler for XML parsing.
         */
        void line_style_cdata_handler(
            const std::string &cdata,
            ScopedHandler *handler,
            int factor);

        /// @}

    private:
        float m_line_width;

        LineStyle m_line_style;

        int m_stipple_factor;
    };

}


#endif // __PEEKABOT_LINE_BASED_HH
