/*
 * Copyright Staffan Gimåker 2006-2009.
 * Copyright Anders Boberg 2007.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Types.hh"
#include "ServerActuator.hh"
#include "renderer/Renderer.hh"
#include "gui/UserInterface.hh"
#include "atl/ActionTransporter.hh"
#include "atl/ActionRouter.hh"
#include "actions/LoadScene.hh"

#include "Log.hh"
#include "PrintfFormatter.hh"
#include "Configuration.hh"
#include "ModelObject.hh"
#include "Subscriber.hh"
#include "FsToolbox.hh"
#include "Init.hh"

#include <cstdlib>
#include <unistd.h>
#include <iostream>
#include <cmath>
#include <boost/format.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/program_options.hpp>
#include <FL/Fl_File_Chooser.H>

using namespace peekabot;
using namespace gui;
using namespace renderer;
using namespace atl;
namespace po = boost::program_options;



static const int DEFAULT_PORT = 5050;


void init_logs() throw();
void setup_config() throw();
void init_scene(const std::vector<std::string> &scenes) throw();


class Console : public Subscriber
{
    bool m_is_active;

    /**
     * \brief The formatter used to format the log messages.
     */
    const PrintfFormatter m_formatter;
    
public:
    Console() throw() : m_is_active(false), m_formatter("%m\n") {}

    /**
     * \brief Writes the message to file immediately (i.e. after writing it,
     * it flushes the buffers).
     */
    virtual void handle_message(const Message &message) throw()
    {
        std::cout << m_formatter.format(message);
        std::cout.flush();
    }

    void activate() throw()
    {
        if( !m_is_active )
        {
            TheMessageHub::instance().subscribe(INFO_MESSAGE, this);
            TheMessageHub::instance().subscribe(WARNING_MESSAGE, this);
            TheMessageHub::instance().subscribe(ERROR_MESSAGE, this);
            m_is_active = true;
        }
    }

    void deactivate() throw()
    {
        if( m_is_active )
        {
            TheMessageHub::instance().unsubscribe(INFO_MESSAGE, this);
            TheMessageHub::instance().unsubscribe(WARNING_MESSAGE, this);
            TheMessageHub::instance().unsubscribe(ERROR_MESSAGE, this);
            m_is_active = false;
        }
    }
};


int main(int argc, char** argv)
{
    Console console;
    console.activate();

    fs::create_dir_if_inexistant(fs::get_resource_path());
    fs::create_dir_if_inexistant(fs::get_resource_path() / "data");
    fs::create_dir_if_inexistant(fs::get_resource_path() / "snapshots");
    fs::create_dir_if_inexistant(fs::get_temp_path() / "peekabot");

    //
    // Load configuration and setup default options
    //
    setup_config();

    init_logs();
    console.deactivate();


    std::vector<std::string> scene_files;
    int port = TheConfiguration::instance().get<int>(
        "network.listen_port", DEFAULT_PORT);

    po::variables_map vm;
    po::options_description desc("Allowed options");
    desc.add_options()
        ("help,h", "Show this help message and exit")
        ("version,v", "Show the peekabot version and exit")
        ("load-scene,s", po::value<std::vector<std::string> >(&scene_files), 
         "Load the specified scene file at startup (can be used multiple times)")
        ("port,p", po::value<int>(&port), 
         "Listening port. Overrides the port set in the configuration file");

    po::positional_options_description p;
    p.add("input", 1);
    p.add("output", 2);

    try
    {
        po::store(po::command_line_parser(argc, argv).
                  options(desc).positional(p).run(), vm);
        po::notify(vm);
    }
    catch(po::error &e)
    {
        std::cerr << "Error parsing command line arguments: " 
                  << e.what() << std::endl;
        return -1;
    }

    if( vm.count("help") > 0 )
    {
        std::cout << desc << std::endl;
        return 0;
    }

    if( vm.count("version") > 0 )
    {
        std::cout << "peekabot " << version::PEEKABOT_VERSION_STRING 
                  << std::endl << std::endl
                  << "Copyright (C) 2006, 2007, 2008 Staffan Gimåker, " 
                  << "Anders Boberg" << std::endl << "http://www.peekabot.org/"
                  << std::endl << std::endl
                  << "peekabot comes with ABSOLUTELY NO WARRANTY. peekabot "
                  << "is free software, and\nyou may use, modify and "
                  << "redistribute it under the terms of the GNU\n"
                  << "General Public License version 2 or later." << std::endl;
        return 0;
    }


    //
    // Log various settings and data that might be useful in dismissing 
    // bug reports from users ;)
    //
    TheMessageHub::instance().publish(
        LOG_MESSAGE, "Resource directory: " + 
        fs::get_resource_path().string());

    TheMessageHub::instance().publish(
        LOG_MESSAGE, "Pkgdatadir: " + 
        fs::get_pkgdata_path().string());

    TheMessageHub::instance().publish(
        LOG_MESSAGE, "Share path: " +
        TheConfiguration::instance().get<fs::path>(
            "resources.share_path").string());

    peekabot::init();


    peekabot::gui::TheUserInterface::instance().set_fps(
        TheConfiguration::instance().get<int>("gui.fps", 20));
    peekabot::gui::TheUserInterface::instance().set_gui_fps(
        TheConfiguration::instance().get<int>("gui.gui_fps", 25));

    TheServerActuator::instance().start();
    init_scene(scene_files);
    TheActionTransporter::instance().start(port);

    peekabot::gui::TheUserInterface::instance().run();

    TheActionTransporter::instance().stop();
    TheServerActuator::instance().stop();


    peekabot::cleanup();

    return 0;
}


void init_logs() throw()
{
    static PrintfFormatter formatter("[%t]: %m (%y, %s)\n");

    static Log log(
        (fs::get_resource_path() / "peekabot.log").string(), formatter);

    TheMessageHub::instance().subscribe(LOG_MESSAGE, &log);
    TheMessageHub::instance().subscribe(INFO_MESSAGE, &log);
    TheMessageHub::instance().subscribe(WARNING_MESSAGE, &log);
    TheMessageHub::instance().subscribe(ERROR_MESSAGE, &log);

    if( TheConfiguration::instance().get<bool>(
            "debug.debug_log", false) )
    {
        static Log dbg_log(
            (fs::get_resource_path() / fs::make_path("debug.log")).string(), formatter);
        TheMessageHub::instance().subscribe(DEBUG_MESSAGE, &dbg_log);
    }
}


bool is_share_path(const boost::filesystem::path &path)
{
    return (
        boost::filesystem::exists(path) &&
        boost::filesystem::is_directory(path) &&
        boost::filesystem::exists(path / "schemas") &&
        boost::filesystem::is_directory(path / "schemas") &&
        boost::filesystem::exists(path / "images") &&
        boost::filesystem::is_directory(path / "images") &&
        boost::filesystem::exists(path / "fonts") &&
        boost::filesystem::is_directory(path / "fonts") );
}


boost::filesystem::path get_share_path()
{
    Fl::scheme("plastic");

    fl_message(
        "peekabot couldn't determine where its shared resources (icons, "
        "XML schemas, etc.) are installed.\nYou will have to manually "
        "select in directory where they can be found."
        "\n\n"
        " - On Windows, the directory where the shared resources are "
        "installed is usually INSTALL_DIR/share/peekabot.\n"
        " - On other operating systems, it's typically '/usr/share/peekabot' "
        "or similar."
        "\n\n"
        "This setting will be saved in your configuration file.");

    bool cancelled = false;
    std::string dir;

    while( true )
    {
        Fl_File_Chooser w(
            dir.empty() ? 0:dir.c_str(), 0, Fl_File_Chooser::DIRECTORY, 
            "Choose shared resource directory");
        w.preview(false);
        w.show();

        while( w.shown() )
            Fl::wait();

        cancelled = (w.value() == 0);
        if( w.value() != 0 )
            dir = w.value();

        if( cancelled || is_share_path(dir) )
            break;
        else
            fl_message(
                "The directory you specified doesn't seem to contain the "
                "shared peekabot resources - try again.");
    }

    if( !cancelled && is_share_path(dir) )
    {
        return boost::filesystem::path(dir);
    }

    throw std::runtime_error(
        "Couldn't find the location where the peekabot "
        "shared resources were installed");
}



std::list<boost::filesystem::path> get_data_paths()
{
    std::list<boost::filesystem::path> ret;

    TheConfiguration::instance().get_non_generated(
        "resources.data_paths", ret);

    ret.push_front(fs::get_resource_path() / "models"); // deprecated
    ret.push_front(fs::get_resource_path() / "scenes"); // deprecated

    ret.push_front(fs::get_pkgdata_path() / "data");
    ret.push_front(fs::get_resource_path() / "data");

    ret.push_back(fs::get_pkgdata_path() / "scenes"); // deprecated
    ret.push_back(fs::get_pkgdata_path() / "models"); // deprecated

    return ret;
}




void setup_config() throw()
{
    //
    // Load config file
    //
    boost::filesystem::path conf_path =
        fs::get_resource_path() / fs::make_path("peekabot.conf");

    if( boost::filesystem::exists(conf_path) )
        TheConfiguration::instance().load(conf_path.string());
    else
    {
        TheMessageHub::instance().publish(
            INFO_MESSAGE,
            "No configuration file found - "
            "creating ~/.peekabot/peekabot.conf");

        std::ofstream tmp(conf_path.string().c_str());
    }

    //
    // Setup default values
    //
    Plist &defaults = TheConfiguration::instance().get_defaults();
    defaults.set("resources.share_path", fs::get_pkgdata_path());

    //
    // Determine share-path
    //

    // If the currently configured share_path isn't valid, we need the user
    // to input the share_path and save it to in configuration file.
    if( !is_share_path(
            TheConfiguration::instance().get<boost::filesystem::path>(
                "resources.share_path")) )
    {
        boost::filesystem::path share_path = get_share_path();

        // Store share_path in configuration file
        TheConfiguration::instance().get_persistent().set(
            "resources.share_path", share_path);

        // ... and save it!
        TheConfiguration::instance().save(
            (fs::get_resource_path() / "peekabot.conf").string());
    }

    //
    // Set up special/generated options
    //
    Plist &gen = TheConfiguration::instance().get_generated();

    gen.add_property_adapter<std::list<boost::filesystem::path> >(
        "data_paths", &get_data_paths);
}


void init_scene(const std::vector<std::string> &scenes) throw()
{
    fs::path path = TheConfiguration::instance().get<fs::path>(
        "default_scene", fs::make_path("default_scene.xml"));

    atl::TheActionRouter::instance().post_action(
        new LoadScene(ROOT_OBJECT_ID, path.string(), AUTO_ENUMERATE_ON_CONFLICT),
        SERVER_SOURCE_ID, SERVER_SOURCE_ID);

    for( std::size_t i = 0; i < scenes.size(); ++i )
    {
        atl::TheActionRouter::instance().post_action(
            new LoadScene(ROOT_OBJECT_ID, scenes[i], AUTO_ENUMERATE_ON_CONFLICT),
            SERVER_SOURCE_ID,  SERVER_SOURCE_ID);
    }
}
