/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_MODEL_OBJECT_HH
#define __PEEKABOT_MODEL_OBJECT_HH


#include <string>
#include <list>

#include "SceneObject.hh"
#include "ScalableObject.hh"
#include "ObjectVisitor.hh"
#include "HandlerInformer.hh"
#include "FsToolbox.hh"


namespace peekabot
{
    class ScopedHandler;

    /**
     * \internal
     *
     * \brief Class representing externally modeled objects in the scene.
     */
    class ModelObject : public SceneObject, public ScalableObject
    {
    public:        
        /** \brief Creates a new \c ModelObject from the given model file.
         *
         * \param model_file_name The file name of the model file to import.
         */
        ModelObject(
            const std::string &model_filename,
            const std::list<fs::path> &extra_dirs = std::list<fs::path>()) 
            throw();

        /** \brief Construct object from XML data.
         */
        ModelObject(
            ScopedHandler *handler,
            const std::list<fs::path> &extra_dirs = std::list<fs::path>());

        /** \brief Adds a texture to a model object.
         *
         * \param texture_file_name The file name of the texture file to be loaded.
         */
        //void set_texture(std::string texture_file_name) {}
        
        /** \brief Method for accepting \c ObjectVisitor objects.
         *
         * This method performs an automatic type cast by calling the \c visit() method of the visitor, allowing it to perform type-dependent task.
         */
        virtual void accept(ObjectVisitor* visitor) throw();

        const std::string &get_model_filename() const throw();

        const std::string &get_expanded_model_filename() const throw();

        void set_filename(
            const std::string &filename,
            const std::list<fs::path> &extra_dirs = std::list<fs::path>());

    private:
        /// \name XML handler methods
        /// @{

        /** \brief Element start handler for XML parsing.
         */
        static void start_handler(
            const std::string & name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        /** \brief File start handler for XML parsing.
         */
        void file_start_handler(
            const std::string & name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        /** \brief File CDATA handler.
         */
        void file_cdata_handler(
            const std::string &cdata,
            ScopedHandler *handler);

        /// @}

    private:
        std::string m_filename;

        std::string m_expanded_filename;

        static HandlerInformer ms_handler_informer;
    };




}

#endif // __PEEKABOT_MODEL_OBJECT_HH
