/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_OCCUPANCY_GRID_2D_HH
#define __PEEKABOT_OCCUPANCY_GRID_2D_HH

#include <vector>
#include "Vector.hh"
#include "SceneObject.hh"
#include "ScalableObject.hh"
#include "ObjectVisitor.hh"
#include "HandlerInformer.hh"


namespace peekabot
{
    class ScopedHandler;

    class OccupancyGrid2D : public SceneObject, 
                            public ScalableObject
    {
    public:        
        /** \brief Creates a new \c OccupancyGrid2D from the given model file.
         *
         * \param model_file_name The file name of the model file to import.
         */
        OccupancyGrid2D(float cell_size, 
                        const RGBColor &free_color,
                        const RGBColor &occupied_color) throw();

        /** \brief Construct object from XML data.
         */
        OccupancyGrid2D(ScopedHandler *handler);

        /** \brief Method for accepting \c ObjectVisitor objects.
         *
         * This method performs an automatic type cast by calling the \c visit()
         *  method of the visitor, allowing it to perform type-dependent task.
         */
        virtual void accept(ObjectVisitor *visitor) throw();

        float get_cell_size() const throw();

        const RGBColor &get_free_color() const throw();

        const RGBColor &get_occupied_color() const throw();

        // belief = -1 <=> clear cell
        void set_cell(const Vector2f &x, float belief) throw();

        void set_cells(const std::vector<std::pair<Vector2f, float> > &cells) throw();

        /// \name Signals
        /// @{

        typedef boost::signal<void (OccupancyGrid2D *object, const std::vector<std::pair<Vector2f, float> > &cells)> OnCellsSet;

        OnCellsSet m_on_cells_set;

        /// @}

    private:
        /// \name XML handler methods
        /// @{

        /** \brief Element start handler for XML parsing.
         */
        static void start_handler(
            const std::string & name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        /// @}

    private:
        //static HandlerInformer ms_handler_informer;

        float m_cell_size;

        RGBColor m_free_color;

        RGBColor m_occupied_color;
    };

}

#endif // __PEEKABOT_OCCUPANCY_GRID_2D_HH
