/*
 * Copyright Staffan Gimåker 2009.
 *
 * ---
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */

#ifndef PEEKABOT_PLIST_HH_INCLUDED
#define PEEKABOT_PLIST_HH_INCLUDED


#include <map>
#include <string>
#include <cassert>
#include <stdexcept>
#include <boost/any.hpp>
#include <boost/shared_ptr.hpp>


namespace peekabot
{
    class Prop
    {
    public:
        virtual ~Prop() {}

        virtual Prop *clone() const = 0;

        virtual boost::any get() const = 0;

        virtual void set(const boost::any &val) = 0;
    };


    namespace detail
    {
        template<class K>
        class PlistImpl
        {
        public:
            typedef K KeyType;

            PlistImpl()
            {
            }

            PlistImpl(const PlistImpl &other)
                : m_prototype(other.m_prototype)
            {
                for( typename PropMap::const_iterator it = other.m_props.begin();
                     it != other.m_props.end(); ++it )
                {
                    m_props.insert(std::make_pair(it->first, it->second->clone()));
                }
            }

            ~PlistImpl()
            {
                for( typename PropMap::iterator it = m_props.begin();
                     it != m_props.end(); ++it )
                {
                    delete it->second;
                }

                m_props.clear();
            }

            Prop *get_prop(const KeyType &key)
            {
                typename PropMap::iterator it = m_props.find(key);
                if( it == m_props.end() )
                    return 0;
                else
                    return it->second;
            }

            const Prop *get_prop(const KeyType &key) const
            {
                typename PropMap::const_iterator it = m_props.find(key);
                if( it == m_props.end() )
                    return 0;
                else
                    return it->second;
            }

            Prop *get_prop_recursive(const KeyType &key)
            {
                typename PropMap::iterator it = m_props.find(key);
                if( it == m_props.end() )
                {
                    if( m_prototype )
                        return m_prototype->get_prop_recursive(key);
                    else
                        return 0;
                }
                else
                    return it->second;
            }

            const Prop *get_prop_recursive(const KeyType &key) const
            {
                typename PropMap::const_iterator it = m_props.find(key);
                if( it == m_props.end() )
                {
                    if( m_prototype )
                        return m_prototype->get_prop_recursive(key);
                    else
                        return 0;
                }
                else
                    return it->second;
            }

            void set_prop(const KeyType &key, Prop *p)
            {
                m_props[key] = p;
            }

            void unset_prop(const KeyType &key)
            {
                m_props.erase(key);
            }

            boost::shared_ptr<PlistImpl<KeyType> > get_prototype()
            {
                return m_prototype;
            }

            const boost::shared_ptr<PlistImpl<KeyType> > get_prototype() const
            {
                return m_prototype;
            }

            void set_prototype(boost::shared_ptr<PlistImpl<KeyType> > prototype)
            {
                m_prototype = prototype;
            }

            bool has(const KeyType &key) const
            {
                return defines(key) || (m_prototype && m_prototype->has(key));
            }

            inline bool defines(const KeyType &key) const
            {
                return get_prop(key) != 0;
            }

        private:
            mutable boost::shared_ptr<PlistImpl<KeyType> > m_prototype;

            typedef std::map<KeyType, Prop *> PropMap;

            PropMap m_props;
        };


        class IntrinsicProp : public Prop
        {
        public:
            virtual Prop *clone() const
            {
                Prop *p = new IntrinsicProp;
                p->set(m_val);
                return p;
            }

            virtual boost::any get() const
            {
                return m_val;
            }

            virtual void set(const boost::any &val)
            {
                m_val = val;
            }

        private:
            boost::any m_val;
        };


        template<class T>
        class AdapterProp : public Prop
        {
        public:
            AdapterProp(
                boost::function<T (void)> getter,
                boost::function<void (const T &)> setter)
                : m_getter(getter), m_setter(setter) {}

            virtual Prop *clone() const
            {
                return new AdapterProp<T>(m_getter, m_setter);
            }

            virtual boost::any get() const
            {
                return boost::any(m_getter());
            }

            virtual void set(const boost::any &val)
            {
                if( m_setter )
                    m_setter(boost::any_cast<T>(val));
                else
                    throw std::runtime_error("Property is read only");
            }

        private:
            boost::function<T (void)> m_getter;
            boost::function<void (const T &)> m_setter;
        };
    }


    /**
     * \internal
     *
     * \brief A basic property-list class.
     */
    class Plist
    {
    public:
        typedef std::string KeyType;

        class PropNotFound : public std::runtime_error
        {
        public:
            PropNotFound(const std::string &msg) : std::runtime_error(msg) {}
        };

        class InvalidPropType : public std::runtime_error
        {
        public:
            InvalidPropType(const std::string &msg) : std::runtime_error(msg) {}
        };

        //typedef boost::function<
        //    Prop *(const std::string &key)> PropNotFoundHook;

        Plist()
            : m_impl(new detail::PlistImpl<std::string>())
        {
        }

        // Deep-copy ctor
        Plist(const Plist &other)
            : m_impl(new detail::PlistImpl<std::string>(*other.m_impl))
        {
        }

        // Deep-copy assignment
        Plist &operator=(const Plist &other)
        {
            m_impl = boost::shared_ptr<detail::PlistImpl<std::string> >(
                new detail::PlistImpl<std::string>(*other.m_impl));

            return *this;
        }


        template<class T>
        T get(const std::string &key) const
        {
            Prop *p = m_impl->get_prop_recursive(key);

            //if( !p )
            //    p = m_prop_not_found_hook(key);

            if( !p )
                throw PropNotFound("Property '" + key + "' not found");

            try
            {
                return boost::any_cast<T>(p->get());
            }
            catch(boost::bad_any_cast &)
            {
                throw InvalidPropType(
                    "Property '" + key + "' does not have the requested type");
            }
        }

        template<class T>
        void set(const std::string &key, const T &val)
        {
            Prop *p = m_impl->get_prop(key);

            if( !p )
            {
                //p = m_prop_not_found_hook(key);

                p = new detail::IntrinsicProp();
                add_property(key, p);
            }

            assert( p );

            p->set(boost::any(val));
        }

        void unset(const std::string &key)
        {
            if( !defines(key) )
                throw std::runtime_error(
                    "Cannot unset key '" + key +
                    "' that's not defined in plist");
            else
                m_impl->unset_prop(key);
        }

        inline bool has(const std::string &key) const
        {
            return m_impl->has(key);
        }

        template<class T>
        inline bool has_type(const std::string &key) const
        {
            Prop *p = m_impl->get_prop_recursive(key);

            if( !p )
                return false;
            else
            {
                try
                {
                    boost::any_cast<T>(p->get());
                    return true;
                }
                catch(...)
                {
                    return false;
                }
            }
        }

        inline bool defines(const std::string &key) const
        {
            return m_impl->defines(key);
        }

        inline bool has_prototype() const
        {
            return m_impl->get_prototype();
        }

        Plist prototype()
        {
            if( !has_prototype() )
                throw std::runtime_error("Plist has no prototype");
            else
                return Plist(m_impl->get_prototype());
        }

        const Plist prototype() const
        {
            if( !has_prototype() )
                throw std::runtime_error("Plist has no prototype");
            else
                return Plist(m_impl->get_prototype());
        }

        void set_prototype(Plist &prototype)
        {
            m_impl->set_prototype(prototype.m_impl);
        }

        // Enables prop-adapters!
        void add_property(const std::string &key, Prop *p)
        {
            m_impl->set_prop(key, p);
        }


        template<class T>
        void add_property_adapter(
            const std::string &key,
            boost::function<T (void)> getter,
            boost::function<void (const T &)> setter =
            boost::function<void (const T &)>())
        {
            Prop *p = new detail::AdapterProp<T>(getter, setter);
            try
            {
                add_property(key, p);
            }
            catch(...)
            {
                delete p;
                throw;
            }
        }

        /// \name Meta-programming hooks
        /// @{

        /*void set_prop_not_found_hook(PropNotFoundHook hook)
        {
            m_impl->m_prop_not_found_hook = hook;
        }*/

        /// @}

    private:
        Plist(boost::shared_ptr<detail::PlistImpl<std::string> > impl)
            : m_impl(impl) {}

    private:
        boost::shared_ptr<detail::PlistImpl<std::string> > m_impl;
    };
}


#endif // PEEKABOT_PLIST_HH_INCLUDED
