/*
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <cassert>
#include "PseudonymCatalog.hh"
#include "SceneObject.hh"
#include "atl/ActionTransporter.hh"


using namespace peekabot;



PseudonymCatalog::PseudonymCatalog(SceneObject *root) throw()
{
    on_child_attached(root);

    // Remove old pseudonyms for source whenever a new client with the same ID
    // is connected... a bit of hack, but it works for now.
    TheActionTransporter::instance().m_on_client_connected.connect(
        boost::bind(&PseudonymCatalog::on_client_connected, this, _1));
}


PseudonymCatalog::~PseudonymCatalog() 
{
    /*TheActionTransporter::instance().m_on_client_connected.disconnect(
        boost::bind(&PseudonymCatalog::on_client_connected, this, _1));*/
}


ObjectID PseudonymCatalog::translate_pseudonym(
    SourceID source, ObjectID id) const throw()
{
    SourceMap::const_iterator it = m_source_map.find(source);

    if( it == m_source_map.end() )
        return 0; // No pseudonyms registered for source

    const PseudoToReal &p_to_r = it->second.m_pseudo_to_real;
    PseudoToReal::const_iterator jt = p_to_r.find(id);

    if( jt == p_to_r.end() )
        return 0; // Pseudonym not found

    return jt->second;
}


void PseudonymCatalog::add_pseudonym(
    SourceID source, ObjectID id, ObjectID real_id) throw(std::runtime_error)
{
    PseudoBimap &pbi = m_source_map[source];

    // Add pseudo -> real lookup, and check if pseudonym is already registered
    if( !pbi.m_pseudo_to_real.insert(std::make_pair(id, real_id)).second )
        throw std::runtime_error(
            "Failed to add pseudonym - pseudonym already exists");

    // Add real -> pseudo lookup
    pbi.m_real_to_pseudo.insert(std::make_pair(real_id, id));
}


void PseudonymCatalog::remove_pseudonym(
    SourceID source, ObjectID id) throw(std::runtime_error)
{
    SourceMap::iterator pbi_it = m_source_map.find(source);
    if( pbi_it == m_source_map.end() )
        throw std::runtime_error( // No pseudonyms registered for source
            "Failed to remove pseudonym - no such pseudonym");

    PseudoBimap &pbi = pbi_it->second;

    // Find the real ID corresponding to the pseudonym
    ObjectID real_id = translate_pseudonym(source, id);
    if( real_id == 0 )
        throw std::runtime_error( // No such pseudonym
            "Failed to remove pseudonym - no such pseudonym");

    // Remove pseudo -> real lookup
    pbi.m_pseudo_to_real.erase(id);
    // Remove real -> pseudo lookups
    pbi.m_real_to_pseudo.erase(real_id);
}


void PseudonymCatalog::remove_pseudonyms_for_object(ObjectID real_id) throw()
{
    for( SourceMap::iterator it = m_source_map.begin(); 
         it != m_source_map.end(); ++it )
    {
        PseudoBimap &pbi = it->second;
        RealToPseudo::iterator jt = pbi.m_real_to_pseudo.lower_bound(real_id);
        RealToPseudo::iterator ub = pbi.m_real_to_pseudo.upper_bound(real_id);

        // Iterate over and remove all pseudonyms for real_id
        for( ; jt != ub; ++jt )
        {
            ObjectID pseudo_id = jt->second;
            // Remove pseduo -> real lookup
            pbi.m_pseudo_to_real.erase(pseudo_id);
        }
    }
}


void PseudonymCatalog::remove_pseudonyms_for_client(SourceID source) throw()
{
    m_source_map.erase(source);
}




void PseudonymCatalog::on_detached(SceneObject *object) throw()
{
    object->m_on_child_attached.disconnect(
        boost::bind(&PseudonymCatalog::on_child_attached, this, _1));

    object->m_on_detached.disconnect(
        boost::bind(&PseudonymCatalog::on_detached, this, _1));
    
    SceneObject::ChildContainer &children = object->get_children();
    for( SceneObject::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        on_detached(*it);
    }
}


void PseudonymCatalog::on_child_attached(SceneObject *object) throw()
{
    object->m_on_child_attached.connect(
        boost::bind(&PseudonymCatalog::on_child_attached, this, _1));

    object->m_on_detached.connect(
        boost::bind(&PseudonymCatalog::on_detached, this, _1));

    if( m_tracked.insert(object).second )
    {
        object->m_on_delete.connect(
            boost::bind(&PseudonymCatalog::on_deleted, this, _1));
    }

    SceneObject::ChildContainer &children = object->get_children();
    for( SceneObject::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        on_child_attached(*it);
    }
}


void PseudonymCatalog::on_deleted(SceneObject *object) throw()
{
    remove_pseudonyms_for_object(object->get_object_id());

    object->m_on_delete.disconnect(
        boost::bind(&PseudonymCatalog::on_deleted, this, _1));

    m_tracked.erase(object);
}


void PseudonymCatalog::on_client_connected(atl::ClientConnection *conn) throw()
{
    remove_pseudonyms_for_client(conn->get_id());
}
