/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <cmath>
#include <stdexcept>
#include <limits>
#include <boost/scoped_array.hpp>

#include "Rangefinder.hh"
#include "renderer/Renderer.hh"
#include "RendererRelay.hh"
#include "SensorObject.hh"
#include "GroupObject.hh"
#include "IDTranslationTable.hh"
#include "serialization/DeserializationInterface.hh"


using namespace peekabot;
using namespace peekabot::renderer;


/**
 * \page rangefinder Rangefinders (DEPRECATED)
 *
 * \attention The sensor visualization framework was deprecated in 0.6, and
 * support will be removed in 0.7.
 *
 * 2D-rangefinder sensors are natively supported in peekabot. Its type 
 * identifier is \b \c rangefinder.
 *
 * How the range data is visualized is configurable through a \e mode 
 * parameter. Making it suitable for a wide range of sensor devices, with
 * laser rangefinders and sonars being two examples.
 *
 * Measurements that lie outside the range \f$(lower\_bound,upper\_bound)\f$ are
 * rejected as "invalid" and put in a separate group and marked in a different
 * color.
 *
 * \section rangefinder_data Input data
 *
 * Rangefinders accept an array of floats or doubles from clients. Mixing 
 * different data types in the same reading is not allowed.
 *
 * Sending datatypes other than float or double, or mixing them, results in 
 * undefined behaviour.
 *
 * \section rangefinder_params Parameters
 *
 * The parameter's type is shown in parantheses after the parameter name.
 *
 * \par mode (string)
 * The visualization mode to use. Must be \c dots or \c discs. See 
 * \ref rangefinder_modes for more documentation on the available modes.
 * \par
 * <em>Default:</em> dots
 *
 * \par samples (int)
 * The number of data points per scan. Sending a different number of data 
 * samples results in undefined behaviour.
 * \par
 * <em>Default:</em> 361
 *
 * \par start_angle (float)
 * The angle of the first data point, in degrees. 
 * \par
 * <em>Default:</em> -90
 *
 * \par end_angle (float)
 * The angle of the last data point, in degrees. 
 * \par
 * <em>Default:</em> 90
 *
 * \par angle_uncertainty (float)
 * The angular uncertainty, in degrees. E.g. if your particular device is an 
 * optical rangefinder with a beam opening of 1 degree, you should put 0.5 here.
 * \par
 * <em>Default:</em> 0
 * 
 * \par dist_uncertainty_prop (float)
 * Proportional distance uncertainty. 
 * \par
 * The total uncertainty \f$\sigma(d)\f$ in distance for a measurement 
 * \f$d\f$ is calculated according to:
 * \par
 * \f$\sigma(d)=d \cdot c_{dist\_uncertainty\_prop} + 
 * c_{dist\_uncertainty\_const}\f$
 * \par
 * <em>Default:</em> 0
 *
 * \par dist_uncertainty_const (float)
 * Constant distance uncertainty, in meters. 
 * \par
 * <em>Default:</em> 0
 *
 * \par lower_bound (float)
 * The exclusive lower bound on distance for valid measurements, in meters. 
 * \par
 * <em>Default:</em> \c std::numeric_limits<float>::min()
 *
 * \par upper_bound (float)
 * The exclusive upper bound on distance for valid measurements, in meters.
 * \par
 * <em>Default:</em> \c std::numeric_limits<float>::max()
 *
 * \par offset (float)
 * The distance each measurement is offset from the sensor's
 * origin. E.g. if you have a sonar ring, each sonar and measurement will be
 * offset a (typically positive) distance from the sensor object's origin. 
 * Given in meters.
 * \par
 * <em>Default:</em> 0
 *
 * \par valid_color (color)
 * The default color to use for valid readings.
 * \par
 * <em>Default:</em> blue (0,0,1)
 *
 * \par invalid_color (color)
 * The default color to use for invalid readings.
 * \par
 * <em>Default:</em> red (1,0,0)
 *
 * \subsection rangefinder_params_templates Parameter templates
 *
 * None currently available.
 *
 * \section rangefinder_modes Visualization modes
 *
 * \par dots
 * Each range sample is visualized as a dot with distance and angular 
 * uncertainties indicated by lines. This is typically used for laser scanners.
 * \par
 * Visualization of uncertainties can be toggled in the GUI, or turned off
 * altogether by setting all uncertainty parameters to zero.
 *
 * \par discs 
 * Each range sample is visualized as a black semi-transparent line from the 
 * measurement's origin (the sensor's origin plus an distance offset, 
 * specified by the \e offset param) with a disc at the end of that line 
 * signifying the angular uncertainty/beam width. The distance uncertainty 
 * is shown as a line perpendicular to the disc.
 * \par
 * It's good for visualizing low sample-density data and rangings with high
 * angular uncertainty, such as sonars.
 * \par
 * Note that this mode is rather resource intensive when many samples are used.
 *
 * \section rangefinder_examples Examples
 *
 * Below is an example of how to define a laser rangefinder in your robot 
 * definition file:
 *
 * \code
 * <sensor type="rangefinder">
 *   <name>scanner</name>
 *
 *   <transform>
 *     <translate system="world">0.34 0 0.58</translate>
 *   </transform>
 *
 *   <children>
 *     <model>
 *       <file>sick_lms.pbmf</file>
 *     </model>
 *   </children>
 *
 *   <params>
 *     <string name="mode">dots</string>
 *     <int name="samples">181</int>
 *     <float name="angle_uncertainty">0.5</float>
 *     <float name="dist_uncertainty_prop">0.002</float>
 *     <float name="dist_uncertainty_const">0.01</float>
 *     <float name="lower_bound">0.05</float>
 *     <float name="upper_bound">8</float>
 *     <color name="invalid_color" r="1" g="0" b="1"/>
 *   </params>
 * </sensor>
 * \endcode
 *
 * In the client, feed the sensor data by doing:
 *
 * \code
 * float dists[361];
 * ...
 * peekabot::SensorProxy p;
 * ...
 * p.write(dists, 361); // buffer data to be sent
 * p.update(); // send and update
 * \endcode
 */


Rangefinder::Rangefinder() throw()
    : m_impl(0)
{
}


void Rangefinder::validate_params()
{
    check_param_type<int>("samples");
    check_param_type<float>("start_angle");
    check_param_type<float>("end_angle");
    check_param_type<float>("dist_uncertainty_prop");
    check_param_type<float>("dist_uncertainty_const");
    check_param_type<float>("angle_uncertainty");
    check_param_type<float>("lower_bound");
    check_param_type<float>("upper_bound");
    check_param_type<RGBColor>("valid_color");
    check_param_type<RGBColor>("invalid_color");
    check_param_type<std::string>("mode");

    if( has_param("mode") )
    {
        const std::string &mode = get_param<std::string>("mode");
        
        if( mode != "dots" && 
            mode != "polygon" &&
            mode != "discs" &&
            mode != "cones" )
            throw std::runtime_error("Invalid mode");
    }
}


void Rangefinder::set_param_defaults(
    const std::string &template_name) throw()
{
    // TODO
}


void Rangefinder::interpret_sensor_data(
    SensorObject *sensor, DeserializationInterface &sensor_data) 
    throw(std::exception)
{
    const std::string mode = get_param<std::string>("mode", "dots");
    int n = get_param<int>("samples", 361);

    boost::scoped_array<float> d(new float[n]);

    try
    {
        // Read sensor data
        for( int i = 0; i < n; ++i )
            sensor_data >> d[i];


        if( mode != m_curr_mode || m_impl == 0 )
        {
            m_curr_mode = mode;
            
            if( mode == "dots" )
                m_impl = new DotMode();
            else if( mode == "discs" )
                m_impl = new DiscMode();
            else
                throw std::runtime_error(
                    "Unimplemented rangefinder mode " + mode);
        }

        m_impl->update_params(*this);
        m_impl->visualize(d.get(), sensor, *this);
    }
    catch(...)
    {
        throw;
    }
}


// ----------------- ModeImpl implementation ----------------


void Rangefinder::ModeImpl::update_params(const Sensor &parent)
{
    m_samples = parent.get_param<int>(
        "samples", 361);
    m_start_angle = parent.get_param<float>(
        "start_angle", -90) / 180 * M_PI;
    m_end_angle = parent.get_param<float>(
        "end_angle", 90) / 180 * M_PI;
    m_dist_uncertainty_prop = parent.get_param<float>(
        "dist_uncertainty_prop", 0);
    m_dist_uncertainty_const = parent.get_param<float>(
        "dist_uncertainty_const", 0);
    m_angle_uncertainty = parent.get_param<float>(
        "angle_uncertainty", 0) / 180 * M_PI;
    m_lower_bound = parent.get_param<float>(
        "lower_bound", std::numeric_limits<real_unit>::min());
    m_upper_bound = parent.get_param<float>(
        "upper_bound", std::numeric_limits<real_unit>::max());
    m_offset = parent.get_param<float>(
        "offset", 0);

    m_step = (m_end_angle-m_start_angle)/(m_samples-1);
}


// ----------------- DotMode implementation --------------------


Rangefinder::DotMode::DotMode()
    : m_valid_measurements(static_cast<NodeID>(-1)),
      m_invalid_measurements(static_cast<NodeID>(-1)),
      m_valid_angle_errs(static_cast<NodeID>(-1)),
      m_invalid_angle_errs(static_cast<NodeID>(-1)),
      m_valid_dist_errs(static_cast<NodeID>(-1)),
      m_invalid_dist_errs(static_cast<NodeID>(-1))
{
}

Rangefinder::DotMode::~DotMode()
{
    Renderer &renderer = TheRenderer::instance();

    if( m_valid_measurements != static_cast<NodeID>(-1) )
        renderer.remove_primitive(m_valid_measurements);
    if( m_invalid_measurements != static_cast<NodeID>(-1) )
        renderer.remove_primitive(m_invalid_measurements);
    if( m_valid_angle_errs != static_cast<NodeID>(-1) )
        renderer.remove_primitive(m_valid_angle_errs);
    if( m_invalid_angle_errs != static_cast<NodeID>(-1) )
        renderer.remove_primitive(m_invalid_angle_errs);
    if( m_valid_angle_errs != static_cast<NodeID>(-1) )
        renderer.remove_primitive(m_valid_angle_errs);
    if( m_invalid_dist_errs != static_cast<NodeID>(-1) )
        renderer.remove_primitive(m_invalid_dist_errs);
}

void Rangefinder::DotMode::visualize(
    const float *dists, SensorObject *sensor, const Rangefinder &rf)
{
    const float *d = dists;
    const int n = m_samples;

    // Create groups...
    SceneObject *valid;
    if( sensor->has_child("valid") )
        valid = sensor->get_child("valid");
    else
    {
        valid = new GroupObject;
        valid->set_name("valid");
        valid->set_color(
            rf.get_param<RGBColor>("valid_color", RGBColor(0,0,1)), false);
        sensor->attach(valid);
    }
    
    SceneObject *invalid;
    if( sensor->has_child("invalid") )
        invalid = sensor->get_child("invalid");
    else
    {
        invalid = new GroupObject;
        invalid->set_name("invalid");
        invalid->set_color(
            rf.get_param<RGBColor>("invalid_color", RGBColor(1,0,0)), false);
        sensor->attach(invalid);
    }
    
    
    show_distances(d, n, m_start_angle, m_step, 
                   m_lower_bound, m_upper_bound, 
                   true, m_valid_measurements, valid);
    
    show_distances(d, n, m_start_angle, m_step, 
                   m_lower_bound, m_upper_bound, 
                   false, m_invalid_measurements, invalid);
    
    if( m_dist_uncertainty_prop != 0 || m_dist_uncertainty_const != 0 )
    {
        show_distance_uncertainties(
            d, n, m_start_angle, m_step, 
            m_dist_uncertainty_prop, m_dist_uncertainty_const,
            m_lower_bound, m_upper_bound, 
            true, m_valid_dist_errs, valid);
        
        show_distance_uncertainties(
            d, n, m_start_angle, m_step, 
            m_dist_uncertainty_prop, m_dist_uncertainty_const,
            m_lower_bound, m_upper_bound, 
            false, m_invalid_dist_errs, invalid);
    }
    
    if( m_angle_uncertainty != 0 )
    {
        show_angle_uncertainties(
            d, n, m_start_angle, m_step, 
            m_angle_uncertainty, 
            m_lower_bound, m_upper_bound, 
            true, m_valid_angle_errs, valid);
        
        show_angle_uncertainties(
            d, n, m_start_angle, m_step, 
            m_angle_uncertainty, 
            m_lower_bound, m_upper_bound, 
            false, m_invalid_angle_errs, invalid);
    }
}


void Rangefinder::DotMode::show_distances(
    const real_unit d[], uint32_t n, 
    real_unit a, real_unit m_step, 
    real_unit m_lower_bound, real_unit m_upper_bound, 
    bool plot_valid, PrimitiveID &id,
    SceneObject *parent) throw()
{
    renderer::Renderer &renderer = TheRenderer::instance();

    // Get/create groups
    SceneObject *measurements;
    if( parent->has_child("measurements") )
        measurements = parent->get_child("measurements");
    else
    {
        measurements = new GroupObject;
        measurements->set_name("measurements");
        measurements->set_color(parent->get_color(), false);
        parent->attach(measurements);
    }

    std::vector<Vector3ru> points;

    // Plot distance measurements
    for( uint32_t i = 0; i < n; ++i )
    {
        bool is_valid = d[i] > m_lower_bound && d[i] < m_upper_bound;

        if( (plot_valid && is_valid) || (!plot_valid && !is_valid) )
            points.push_back(Vector3ru(d[i]*cosf(a), d[i]*sinf(a), 0));

        a += m_step;
    }


    // Create/clear entitites
    if( id == static_cast<NodeID>(-1) )
    {
        NodeID node_id = TheIDTranslationTable::instance().get_node_id(
            measurements->get_object_id());
        id = renderer.add_point_blob(
            node_id, false, points);
    }
    else
    {
        renderer.set_primitive_vertices(id, points);
    }
}


void Rangefinder::DotMode::show_distance_uncertainties(
    const real_unit d[], uint32_t n, 
    real_unit a, real_unit m_step,
    real_unit m_dist_uncertainty_prop, real_unit m_dist_uncertainty_const,
    real_unit m_lower_bound, real_unit m_upper_bound, 
    bool plot_valid, PrimitiveID &id,
    SceneObject *parent) throw()
{
    renderer::Renderer &renderer = TheRenderer::instance();

    SceneObject *uncertainties;
    if( parent->has_child("distance_uncertainties") )
        uncertainties = parent->get_child("distance_uncertainties");
    else
    {
        uncertainties = new GroupObject;
        uncertainties->set_name("distance_uncertainties");
        uncertainties->set_color(RGBColor(0,0,0), false);
        parent->attach(uncertainties);
    }


    std::vector<Vector3ru> verts;


    for( uint32_t i = 0; i < n; ++i )
    {
        real_unit u = d[i]*m_dist_uncertainty_prop + m_dist_uncertainty_const;
        bool is_valid = d[i] > m_lower_bound && d[i] < m_upper_bound;

        if( plot_valid && is_valid )
        {
            verts.push_back( Vector3ru((d[i]-u)*cosf(a), (d[i]-u)*sinf(a), 0) );
            verts.push_back( Vector3ru((d[i]+u)*cosf(a), (d[i]+u)*sinf(a), 0) );
        }
        else if( !plot_valid && !is_valid )
        {
            verts.push_back( Vector3ru((d[i]-u)*cosf(a), (d[i]-u)*sinf(a), 0) );
            verts.push_back( Vector3ru((d[i]+u)*cosf(a), (d[i]+u)*sinf(a), 0) );
        }

        a += m_step;
    }



    if( id == static_cast<NodeID>(-1) )
    {
        NodeID node_id = TheIDTranslationTable::instance().get_node_id(
            uncertainties->get_object_id());
        id = renderer.add_line_blob(node_id, false, verts);
    }
    else
    {
        renderer.set_primitive_vertices(id, verts);
    }
}

void Rangefinder::DotMode::show_angle_uncertainties(
    const real_unit d[], uint32_t n, 
    real_unit a, real_unit m_step,
    real_unit m_angle_uncertainty, 
    real_unit m_lower_bound, real_unit m_upper_bound, 
    bool plot_valid, PrimitiveID &id,
    SceneObject *parent) throw()
{
    renderer::Renderer &renderer = TheRenderer::instance();

    SceneObject *uncertainties;
    if( parent->has_child("angle_uncertainties") )
        uncertainties = parent->get_child("angle_uncertainties");
    else
    {
        uncertainties = new GroupObject;
        uncertainties->set_name("angle_uncertainties");
        uncertainties->set_color(RGBColor(0,0,0), false);
        parent->attach(uncertainties);
    }


    std::vector<Vector3ru> verts;

    for( uint32_t i = 0; i < n; ++i )
    {
        bool is_valid = d[i] > m_lower_bound && d[i] < m_upper_bound;

        if( plot_valid && is_valid )
        {
            verts.push_back( Vector3ru(d[i]*cosf(a-m_angle_uncertainty), 
                                       d[i]*sinf(a-m_angle_uncertainty), 0) );
            verts.push_back( Vector3ru(d[i]*cosf(a+m_angle_uncertainty), 
                                       d[i]*sinf(a+m_angle_uncertainty), 0) );
        }
        else if( !plot_valid && !is_valid )
        {
            verts.push_back( Vector3ru(d[i]*cosf(a-m_angle_uncertainty), 
                                       d[i]*sinf(a-m_angle_uncertainty), 0) );
            verts.push_back( Vector3ru(d[i]*cosf(a+m_angle_uncertainty), 
                                       d[i]*sinf(a+m_angle_uncertainty), 0) );
        }

        a += m_step;
    }


    if( id == static_cast<NodeID>(-1) )
    {
        NodeID node_id = TheIDTranslationTable::instance().get_node_id(
            uncertainties->get_object_id());
        id = renderer.add_line_blob(node_id, false, verts);
    }
    else
    {
        renderer.set_primitive_vertices(id, verts);
    }
}



// ------------------- DiscMode implementation --------------------------


Rangefinder::DiscMode::DiscMode()
{
}

Rangefinder::DiscMode::~DiscMode()
{
}

void Rangefinder::DiscMode::visualize(
    const float *dists, SensorObject *sensor, const Rangefinder &rf)
{
    for( size_t i = 0; i < (size_t)m_samples; ++i )
    {
        if( m_data_points.size() <= i )
        {
            m_data_points.push_back(
                boost::shared_ptr<DataPoint>(new DataPoint(sensor, rf)));
        }

        m_data_points[i]->set(dists[i], m_step*i + m_start_angle, rf);
    }
}


// -------------------- DataPoint implementation ----------------------


Rangefinder::DiscMode::DataPoint::DataPoint(
    SensorObject *sensor, const Rangefinder &rf)
    : m_is_created(false),
      m_reject(false) 
{
    if( sensor->has_child("valid") )
        m_valid = sensor->get_child("valid");
    else
    {
        m_valid = new GroupObject;
        m_valid->set_name("valid");
        m_valid->set_color(
            rf.get_param<RGBColor>("valid_color", RGBColor(0,0,1)), false);
        sensor->attach(m_valid);
    }
    
    if( sensor->has_child("invalid") )
        m_invalid = sensor->get_child("invalid");
    else
    {
        m_invalid = new GroupObject;
        m_invalid->set_name("invalid");
        m_invalid->set_color(
            rf.get_param<RGBColor>("invalid_color", RGBColor(1,0,0)), false);
        sensor->attach(m_invalid);
    }
}

Rangefinder::DiscMode::DataPoint::~DataPoint()
{
    if( is_created() )
    {
        Renderer &renderer = TheRenderer::instance();

        renderer.remove_node(m_disc_node);
    }
}

void Rangefinder::DiscMode::DataPoint::set(
    float dist, float theta, const Rangefinder &rf)
{
    Renderer &renderer = TheRenderer::instance();

    float angle = rf.get_param<float>(
        "angle_uncertainty", 30) / 180 * M_PI * 2;
    float dist_uncertainty_prop = rf.get_param<float>(
        "dist_uncertainty_prop", 0);
    float dist_uncertainty_const = rf.get_param<float>(
        "dist_uncertainty_const", 0);
    float lower_bound = rf.get_param<float>(
        "lower_bound", std::numeric_limits<real_unit>::min());
    float upper_bound = rf.get_param<float>(
        "upper_bound", std::numeric_limits<real_unit>::max());
    float offset = rf.get_param<float>(
        "offset", 0);

    bool reject = dist <= lower_bound || dist >= upper_bound;

    if( !is_created() )
        create();

    SceneObject *parent = reject ? m_invalid : m_valid;
 
    if( reject != m_reject )
    {
        m_reject = reject;

        renderer.rearrange_node(
            m_disc_node,
            TheIDTranslationTable::instance().get_node_id(
                parent->get_object_id()));
    }

    // Update color of the disc node to match its parent node
    renderer.set_node_color(m_disc_node, parent->get_color());

    // Move the disc to the distance measured
    renderer.set_node_transformation(
        m_disc_node,
        Matrix4ru::rotate_y(M_PI/2) *
        Matrix4ru::rotate_x(-theta) *
        Matrix4ru::translate(0, 0, dist+offset));

    // Set the size of the disc according to angle and
    // distance
    renderer.set_node_scale(
        m_disc_node, dist*sinf(angle/2), dist*sinf(angle/2), 1);

    std::vector<Vector3r> verts(2);

    // Set the line from the disc to the sensor.
    verts[0] = Vector3r(
        0,0, -(dist_uncertainty_prop*dist+dist_uncertainty_const));
    verts[1] = Vector3r(
        0,0, -dist);

    renderer.set_primitive_vertices(m_line_id, verts);

    // Set the uncertainty indicator
    verts[0] = Vector3r(
        0,0, -(dist_uncertainty_prop*dist + dist_uncertainty_const));
    verts[1] = Vector3r(
        0,0, dist_uncertainty_prop*dist + dist_uncertainty_const);

    renderer.set_primitive_vertices(m_uncertainty_id, verts);

    renderer.set_primitive_alpha(
        m_line_id, parent->get_accumulated_opacity() * 0.3);
    renderer.set_primitive_color(m_line_id, RGBColor(0,0,0));
}

void Rangefinder::DiscMode::DataPoint::create()
{
    Renderer &renderer = TheRenderer::instance();

    m_is_created = true;
    m_reject = false;

    NodeID data_node = TheIDTranslationTable::instance().get_node_id(
        m_valid->get_object_id());

    // The disc is created as a node so that it can
    // be scaled
    m_disc_node = renderer.add_node(data_node);
    renderer.add_disc(m_disc_node, true);

    // Create primitive for uncertainty indicator
    m_uncertainty_id = renderer.add_line_mesh(m_disc_node, false);

    // Create primitive for line marker
    m_line_id = renderer.add_line_mesh(m_disc_node, false);

    renderer.set_node_color(m_disc_node, m_valid->get_color());
}

bool Rangefinder::DiscMode::DataPoint::is_created() const throw()
{
    return m_is_created;
}
