/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_RANGEFINDER_HH
#define __PEEKABOT_RANGEFINDER_HH


#include <stdexcept>
#include <boost/shared_ptr.hpp>

#include "Sensor.hh"
#include "SensorObject.hh"
#include "renderer/IDTypes.hh"


namespace peekabot
{
   
    /**
     * \internal
     *
     * \brief A sensor data interpreter class for laser rangefinders.
     *
     * \sa For user documentation, see \ref laser_rangefinder.
     */
    class Rangefinder : public Sensor
    {
        template<class T>
        class ObjectProxy
        {
            ObjectProxy(SceneObject *parent, const std::string &name)//, color, opacity, visibility);
                : m_parent(parent), m_name(name)
            {
            }

            ~ObjectProxy()
            {
                if( m_obj )
                {
                    m_obj->detach();
                    delete m_obj;
                }
            }

            T *get()
            {
                if( !m_obj )
                {
                    if( m_parent->has_child(m_name) )
                    {
                        m_obj = m_parent->get_child(m_name);
                        return m_obj;
                    }
                    else
                    {
                        m_obj = new T;
                        m_obj->set_name(m_name);
                        m_parent->attach(m_obj, AUTO_ENUMERATE_ON_CONFLICT);
                    }
                }

                return m_obj;
            }

            T *operator->()
            {
                return get();
            }

            T *operator*()
            {
                return get();
            }

          private:
            std::string m_name;
            SceneObject *m_parent;
            T *m_obj;
        };

    public:
        Rangefinder() throw();

        virtual void interpret_sensor_data(
            SensorObject *sensor,
            DeserializationInterface &sensor_data) throw(std::exception);

        virtual void set_param_defaults(
            const std::string &template_name) throw();

        virtual void validate_params();

    private:
        class ModeImpl
        {
        public:
            virtual ~ModeImpl() {}

            virtual void visualize(const float *dists, SensorObject *sensor, const Rangefinder &rf) = 0;

            void update_params(const Sensor &parent);

        protected:
            int   m_samples;
            float m_start_angle;
            float m_end_angle;
            float m_dist_uncertainty_prop;
            float m_dist_uncertainty_const;
            float m_angle_uncertainty;
            float m_lower_bound;
            float m_upper_bound;
            float m_offset;
            float m_step;
        };

        class DotMode : public ModeImpl
        {
        public:
            DotMode();

            virtual ~DotMode();
       
            virtual void visualize(const float *dists, SensorObject *sensor, const Rangefinder &rf);

        private:
            void show_distances(
                const real_unit d[], uint32_t n, 
                real_unit a, real_unit step, 
                real_unit lower_bound, real_unit upper_bound, 
                bool plot_valid, PrimitiveID &id,
                SceneObject *parent) throw();
            
            void show_distance_uncertainties(
                const real_unit d[], uint32_t n, 
                real_unit a, real_unit step, 
                real_unit dist_uncertainty_prop, real_unit dist_uncertainty_const,
                real_unit lower_bound, real_unit upper_bound, 
                bool plot_valid, PrimitiveID &id,
                SceneObject *parent) throw();
            
            void show_angle_uncertainties(
                const real_unit d[], uint32_t n, 
                real_unit a, real_unit step,
                real_unit angle_uncertainty,
                real_unit lower_bound, real_unit upper_bound, 
                bool plot_valid, PrimitiveID &id,
                SceneObject *parent) throw();

        private:
            PrimitiveID m_valid_measurements;
            PrimitiveID m_invalid_measurements;
            PrimitiveID m_valid_angle_errs;
            PrimitiveID m_invalid_angle_errs;
            PrimitiveID m_valid_dist_errs;
            PrimitiveID m_invalid_dist_errs;
        };


        class DiscMode : public ModeImpl
        {
        public:
            DiscMode();
            
            virtual ~DiscMode();
           
            virtual void visualize(const float *dists, SensorObject *sensor, const Rangefinder &rf);

        private:
            class DataPoint
            {
            public:
                DataPoint(SensorObject *sensor, const Rangefinder &rf);

                ~DataPoint();

                void set(float dist, float theta, const Rangefinder &rf);

            private:
                void create();

                bool is_created() const throw();

            private:
                bool m_is_created;
                bool m_reject;

                NodeID m_disc_node;
                PrimitiveID m_uncertainty_id;
                PrimitiveID m_line_id;

                SceneObject *m_valid;
                SceneObject *m_invalid;
            };

            std::vector<boost::shared_ptr<DataPoint> > m_data_points;
        };

        ModeImpl *m_impl;

        std::string m_curr_mode;
    };

}


#endif // __PEEKABOT_RANGEFINDER_HH
