/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cassert>
#include <cmath>
#include <boost/bind.hpp>

#include "IDTranslationTable.hh"
#include "MessageHub.hh"
#include "RendererRelay.hh"
#include "renderer/Renderer.hh"

#include "CameraObject.hh"
#include "SceneObject.hh"
#include "GroupObject.hh"
#include "ScalableObject.hh"
#include "GeometricPrimitives.hh"
#include "Hinge.hh"
#include "Slider.hh"
#include "PointSetObject.hh"
#include "ModelObject.hh"
#include "SensorObject.hh"
#include "GridObject.hh"
#include "PointCloud.hh"
#include "LineCloud.hh"
#include "OccupancyGrid2D.hh"
#include "Label.hh"

#include "SceneTreePtr.hh"
#include "SceneTree.hh"


using namespace peekabot;
using peekabot::renderer::TheRenderer;
using peekabot::renderer::Renderer;



//
// ------------------- RendererRelay implementation -----------------------
//

RendererRelay::RendererRelay() throw()
    : m_impl(TheIDTranslationTable::instance(),
             TheRenderer::instance())
{
    m_impl.init(TheSceneTree::acquire()->get_root());
}

RendererRelay::RendererRelay(SceneObject *tree_root) throw()
    : m_impl(TheIDTranslationTable::instance(),
             TheRenderer::instance())
{
    m_impl.init(tree_root);
}

NodeID RendererRelay::get_node_id(ObjectID object_id) const throw()
{
    return m_impl.get_node_id(object_id);
}

PrimitiveID RendererRelay::get_primitive_id(ObjectID object_id) const throw()
{
    return m_impl.get_primitive_id(object_id);
}



//
// -------------------- Impl implementation ------------------------
//

RendererRelay::Impl::Impl(IDTranslationTable &id_lut,
                          Renderer &renderer) throw()
    : m_id_lut(id_lut),
      m_renderer(renderer)
{
}

void RendererRelay::Impl::init(SceneObject *tree_root) throw()
{
    // Add ID translation entry for the root node
    m_id_lut.add_node_translation_entry(
        tree_root->get_object_id(), m_renderer.get_root_node_id());

    on_child_attached(tree_root);
}

NodeID RendererRelay::Impl::get_node_id(ObjectID object_id) const throw()
{
    NodeID id = m_id_lut.get_node_id(object_id);

    if( id == static_cast<NodeID>(-1) )
        TheMessageHub::instance().publish(
            WARNING_MESSAGE, 
            "%s%:d: Invalid object ID provided to "
            "RendererRelay::Impl::get_node_id()",
            __FILE__, __LINE__);

    return id;
}

PrimitiveID RendererRelay::Impl::get_primitive_id(ObjectID object_id) const throw()
{
    return m_id_lut.get_primitive_id(object_id);
}


void RendererRelay::Impl::relay_shared_node_attributes(
    SceneObject *object, NodeID node_id) const throw()
{
    // Relay all common object attributes: mtop, color, opacity, id, layer.
    m_renderer.set_node_transformation(
        node_id, object->get_transformation());

    m_renderer.set_node_opacity(
        node_id, object->get_opacity());

    m_renderer.set_node_color(
        node_id, object->get_color());

    m_renderer.set_node_layer(
        node_id, object->get_layer());

    m_renderer.set_node_visibility(
        node_id, object->is_hidden());

    m_renderer.set_node_name(
        node_id, (uint32_t)object->get_object_id());

    // ... and scale if applicable
    ScalableObject *scalable = dynamic_cast<ScalableObject *>(object);
    if( scalable )
    {
        m_renderer.set_node_scale(
            node_id, 
            scalable->get_x_scale(),
            scalable->get_y_scale(),
            scalable->get_z_scale());
    }
}


void RendererRelay::Impl::relay_shared_primitive_attributes(
    SceneObject *object) const throw()
{
    // ... and vertices if applicable
    VertexObject *vertex = dynamic_cast<VertexObject *>(object);
    if( vertex )
    {
        m_renderer.set_primitive_vertices(
            m_id_lut.get_primitive_id(object->get_object_id()), 
            vertex->get_vertices());
    }
    
    LineBased *line_based = dynamic_cast<LineBased *>(object);
    if( line_based )
    {
        std::pair<LineStyle, int> style;
        style = line_based->get_line_style();

        m_renderer.set_primitive_line_style(
            m_id_lut.get_primitive_id(object->get_object_id()), 
            style.first, style.second, line_based->get_line_width());
    }
}


void RendererRelay::Impl::on_child_attached(SceneObject *object)
{
    NodeID node_id =  m_renderer.get_root_node_id();
    // Unless it's a root node, add a corresponding node in the renderer
    if( object->get_parent() )
    {
        node_id = m_renderer.add_node(
            m_id_lut.get_node_id(object->get_parent()->get_object_id()));

        m_id_lut.add_node_translation_entry(object->get_object_id(), node_id);
    }

    // Create primitivies & connect to type-specific signals
    object->accept(this);


    Connections &connections = m_connection_map[object];

    connections.push_back(
        object->m_on_child_attached.connect(
            boost::bind(&RendererRelay::Impl::on_child_attached, this, _1)));
    
    connections.push_back(
        object->m_on_detached.connect(
            boost::bind(&RendererRelay::Impl::on_detached, this, _1, true)));

    /*object->m_on_object_id_set.connect(
      boost::bind(&RendererRelay::on_object_id_set, this, _1, _2, _3));*/


    connections.push_back(
        object->m_on_transformation_set.connect(
            boost::bind(&Renderer::set_node_transformation, &m_renderer, 
                        node_id, _2, PARENT_COORDINATES)));

    connections.push_back(
        object->m_on_opacity_set.connect(
            boost::bind(&Renderer::set_node_opacity, &m_renderer, node_id, _2)));

    connections.push_back(
        object->m_on_hidden_set.connect(
            boost::bind(&Renderer::set_node_visibility, &m_renderer, node_id, _2)));

    connections.push_back(    
        object->m_on_layer_set.connect(
            boost::bind(&Renderer::set_node_layer, &m_renderer, node_id, _2)));

    connections.push_back(    
        object->m_on_color_set.connect(
            boost::bind(&Renderer::set_node_color, &m_renderer, node_id, _2)));


    ScalableObject *scalable = dynamic_cast<ScalableObject *>(object);
    if( scalable )
    {
        connections.push_back(
            scalable->m_on_scale_set.connect(
                boost::bind(&Renderer::set_node_scale, &m_renderer, 
                            node_id, _2, _3, _4)));
    }

    VertexObject *vertex = dynamic_cast<VertexObject *>(object);
    if( vertex )
    {
        connections.push_back(
            vertex->m_on_vertices_set.connect(
                boost::bind(&Renderer::set_primitive_vertices, &m_renderer, 
                            m_id_lut.get_primitive_id(object->get_object_id()), _2)));
    }

    LineBased *line_based = dynamic_cast<LineBased *>(object);
    if( line_based )
    {
        connections.push_back(
            line_based->m_on_line_params_set.connect(
                boost::bind(&Renderer::set_primitive_line_style, &m_renderer, 
                            m_id_lut.get_primitive_id(object->get_object_id()),
                            _3, _4, _2)));
    }


    // Relay common node attributes (mtop, color, etc.)
    relay_shared_node_attributes(object, node_id);

    // relay common primitive attributes
    relay_shared_primitive_attributes(object);



    SceneObject::ChildContainer &children = object->get_children();
    for( SceneObject::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        on_child_attached(*it);
    }
}



void RendererRelay::Impl::on_detached(SceneObject *object, bool subtree_root)
{
    if( subtree_root )
    {
        //
        // Remove rendering node - but only if object is the root of the
        // subtree being removed
        //
        m_renderer.remove_node(
            m_id_lut.get_node_id(object->get_object_id()));
    }

    //
    // Remove ObjectID -> NodeID translation entries
    //
    m_id_lut.remove_node_translation_entries(object);
    m_id_lut.remove_primitive_translation_entries(object);

    //
    // Disconnect all slots for object
    //
    ConnectionMap::iterator conn_it = m_connection_map.find(object);
    assert( conn_it != m_connection_map.end() );
    Connections &connections = conn_it->second;

    for( Connections::iterator it = connections.begin();
         it != connections.end(); ++it )
    {
        it->disconnect();
    }
    
    m_connection_map.erase(conn_it);


    SceneObject::ChildContainer &children = object->get_children();
    for( SceneObject::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        on_detached(*it, false);
    }
}


void RendererRelay::Impl::set_grid_vertices(PrimitiveID id, GridObject *object)
{
    std::vector<Vector3f> verts;
    unsigned int segments = object->get_segment_count();

    if( object->get_grid_type() == REGULAR_GRID )
    {
        float foo = segments/2.0f;
        
        for( unsigned int i = 0; i < segments+1; ++i )
        {
            verts.push_back(Vector3f(-foo+i, -foo, 0));
            verts.push_back(Vector3f(-foo+i,  foo, 0));
        }
        for( unsigned int i = 0; i < segments+1; ++i )
        {
            verts.push_back(Vector3f(-foo, -foo+i, 0));
            verts.push_back(Vector3f( foo, -foo+i, 0));
        }
    }
    else if( object->get_grid_type() == RADIAL_GRID )
    {
        float theta = object->get_central_angle();
        float offset = -theta/2;

        for( unsigned int i = 0; i < segments; ++i )
        {
            unsigned int r = i+1;
            unsigned int circle_segments = static_cast<unsigned int>(
                ceilf((64+10*i) * theta/M_PI/2));

            for( unsigned int j = 0; j < circle_segments; ++j )
            {
                verts.push_back(Vector3f(r*cosf(theta*j/circle_segments+offset), 
                                         r*sinf(theta*j/circle_segments+offset), 0));
                verts.push_back(Vector3f(r*cosf(theta*(j+1)/circle_segments+offset), 
                                         r*sinf(theta*(j+1)/circle_segments+offset), 0)); 
            }
        }
    }
    else if( object->get_grid_type() == ANGULAR_GRID )
    {
        float theta = object->get_central_angle();
        float offset = -theta/2;

        unsigned int lines;
        if( fabsf(theta-2*M_PI) < 0.000001 )
            lines = segments;
        else
            lines = segments+1;

        for( unsigned int i = 0; i < lines; ++i )
        {
            verts.push_back(Vector3f(0,0,0));
            verts.push_back(Vector3f(cosf(theta*i/segments+offset), 
                                     sinf(theta*i/segments+offset), 0));
        }
    }
    else
        assert( false );

    m_renderer.set_primitive_vertices(id, verts);
}



void RendererRelay::Impl::visit(CameraObject *object)
{
    NodeID node_id = m_id_lut.get_node_id(object->get_object_id());

    //
    // Create a corresponding camera primitive in the renderer
    //
    PrimitiveID id = m_renderer.add_camera(node_id, false);
    m_id_lut.add_primitive_translation_entry(
        object->get_object_id(), id);

    //
    // Relay camera specific attributes.
    //
    m_renderer.set_camera_fov(
        id, static_cast<real32>(object->get_fov()));

    m_renderer.set_camera_view_mode(
        id, object->is_orthographic());

    m_renderer.set_camera_near_plane(
        id, static_cast<real32>(object->get_near_plane()));

    m_renderer.set_camera_far_plane(
        id, static_cast<real32>(object->get_far_plane()));

    m_renderer.set_camera_zoom_distance(
        id, object->get_zoom_distance());



    //
    // Connect the renderer to get updated when camera-specific 
    // attributes are changed.
    //
    Connections &connections = m_connection_map[object];

    connections.push_back(
        object->m_on_fov_set.connect(
            boost::bind(&Renderer::set_camera_fov, &m_renderer, id, _2)));

    connections.push_back(
        object->m_on_view_mode_set.connect(
            boost::bind(&Renderer::set_camera_view_mode, &m_renderer, id, _2)));

    connections.push_back(
        object->m_on_near_plane_set.connect(
            boost::bind(&Renderer::set_camera_near_plane, &m_renderer, id, _2)));

    connections.push_back(
        object->m_on_far_plane_set.connect(
            boost::bind(&Renderer::set_camera_far_plane, &m_renderer, id, _2)));

    connections.push_back(
        object->m_on_zoom_distance_set.connect(
            boost::bind(&Renderer::set_camera_zoom_distance, &m_renderer, id, _2)));
}

void RendererRelay::Impl::visit(ModelObject *object)
{
    m_renderer.add_model(
        m_id_lut.get_node_id(object->get_object_id()), 
        true, object->get_expanded_model_filename());
}

void RendererRelay::Impl::visit(SphereObject *object)
{
    m_renderer.add_sphere(
        m_id_lut.get_node_id(object->get_object_id()), true);
}

void RendererRelay::Impl::visit(GroupObject *object)
{
    // Groups have no graphical representation...
}

void RendererRelay::Impl::visit(CubeObject *object)
{
    m_renderer.add_cube(
      m_id_lut.get_node_id(object->get_object_id()), true);
}

void RendererRelay::Impl::visit(PolygonObject *object)
{
    PrimitiveID id = m_renderer.add_polygon(
        m_id_lut.get_node_id(object->get_object_id()), 
        false, object->get_vertices());

    m_id_lut.add_primitive_translation_entry(
        object->get_object_id(), id);
}

void RendererRelay::Impl::visit(CylinderObject *object)
{
    m_renderer.add_cylinder(
        m_id_lut.get_node_id(object->get_object_id()), true);
}

void RendererRelay::Impl::visit(CircleObject *object)
{
    PrimitiveID id = m_renderer.add_circle(
        m_id_lut.get_node_id(object->get_object_id()), false);

    m_id_lut.add_primitive_translation_entry(
        object->get_object_id(), id);
}

void RendererRelay::Impl::visit(PolyLineObject *object)
{
    // Not yet implemented
}

void RendererRelay::Impl::visit(SensorObject *object)
{
    // Sensor object have no graphical representation of their own...
}

void RendererRelay::Impl::visit(PointSetObject *object)
{
    // TODO
}

void RendererRelay::Impl::visit(Hinge *object)
{
    // DOFs have no graphical representation of their own...
}

void RendererRelay::Impl::visit(Slider *object)
{
    // DOFs have no graphical representation of their own...
}

void RendererRelay::Impl::visit(GridObject *object)
{
    PrimitiveID id = m_renderer.add_line_blob(
        m_id_lut.get_node_id(object->get_object_id()), false, 
        std::vector<Vector3f>());

    m_id_lut.add_primitive_translation_entry(
        object->get_object_id(), id);

    //
    // Connect the renderer to get updated when grid-specific 
    // attributes are changed.
    //
    Connections &connections = m_connection_map[object];

    connections.push_back(
        object->m_on_segment_count_set.connect(
            boost::bind(&RendererRelay::Impl::set_grid_vertices, this, id, _1)));

    connections.push_back(
        object->m_on_grid_type_set.connect(
            boost::bind(&RendererRelay::Impl::set_grid_vertices, this, id, _1)));

    connections.push_back(
        object->m_on_central_angle_set.connect(
            boost::bind(&RendererRelay::Impl::set_grid_vertices, this, id, _1)));

    set_grid_vertices(id, object);
}

void RendererRelay::Impl::visit(PointCloud *object)
{
    PrimitiveID id = m_renderer.add_point_blob(
        m_id_lut.get_node_id(object->get_object_id()), false, 
        object->get_vertices());

    m_id_lut.add_primitive_translation_entry(
        object->get_object_id(), id);
}

void RendererRelay::Impl::visit(LineCloud *object)
{
    PrimitiveID id = m_renderer.add_line_blob(
        m_id_lut.get_node_id(object->get_object_id()), false, 
        object->get_vertices());

    m_id_lut.add_primitive_translation_entry(
        object->get_object_id(), id);
}

void RendererRelay::Impl::visit(OccupancyGrid2D *object)
{
    PrimitiveID id = m_renderer.add_occupancy_grid_2d(
        m_id_lut.get_node_id(object->get_object_id()), false, 
        object->get_cell_size(),
        object->get_free_color(),
        object->get_occupied_color());

    m_id_lut.add_primitive_translation_entry(
        object->get_object_id(), id);

    //
    // Connect the renderer to get updated when occupancy grid-specific 
    // attributes are changed.
    //
    Connections &connections = m_connection_map[object];

    connections.push_back(
        object->m_on_cells_set.connect(
            boost::bind(&Renderer::set_occupancy_grid_2d_cells, &m_renderer, id, _2)));
}

void RendererRelay::Impl::visit(Label *object)
{
    PrimitiveID id = m_renderer.add_label(
      m_id_lut.get_node_id(object->get_object_id()), false);

    m_id_lut.add_primitive_translation_entry(
        object->get_object_id(), id);

    m_renderer.set_label_alignment(id, object->get_alignment());
    m_renderer.set_label_text(id, object->get_text());

    //
    // Connect the renderer to get updated when occupancy lavel-specific 
    // attributes are changed.
    //
    Connections &connections = m_connection_map[object];

    connections.push_back(
        object->m_on_text_set.connect(
            boost::bind(&Renderer::set_label_text, &m_renderer, id, _2)));

    connections.push_back(
        object->m_on_alignment_set.connect(
            boost::bind(&Renderer::set_label_alignment, &m_renderer, id, _2)));
}
