/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_RENDERER_RELAY_HH
#define __PEEKABOT_RENDERER_RELAY_HH


#include <map>
#include <vector>
#include <boost/utility.hpp>
#include <boost/thread_safe_signal.hpp>

#include "Types.hh"
#include "ObjectVisitor.hh"
#include "renderer/IDTypes.hh"
#include "Singleton.hh"
#include "SceneObject.hh"


namespace peekabot
{
    namespace renderer
    {
        class Renderer;
    }

    class IDTranslationTable;
  

    /**
     * \internal
     *
     * \brief Traps graphical changes to the scene and forwards the information
     * to the renderer.
     */
    class RendererRelay : public boost::noncopyable
    {
    public:
        RendererRelay() throw();

        RendererRelay(SceneObject *tree_root) throw();

        NodeID get_node_id(ObjectID object_id) const throw();

        PrimitiveID get_primitive_id(ObjectID object_id) const throw();

    private:
        class Impl : public boost::noncopyable, public ObjectVisitor
        {
        public:
            Impl(IDTranslationTable &id_lut, renderer::Renderer &renderer) throw();

            void init(SceneObject *tree_root) throw();

            NodeID get_node_id(ObjectID object_id) const throw();

            PrimitiveID get_primitive_id(ObjectID object_id) const throw();


            /// \name Visitor methods
            /// @{

            virtual void visit(CameraObject *object);

            virtual void visit(ModelObject *object);

            virtual void visit(SphereObject *object);

            virtual void visit(GroupObject *object);

            virtual void visit(CubeObject *object);

            virtual void visit(PolygonObject *object);

            virtual void visit(CylinderObject *object);

            virtual void visit(CircleObject *object);

            virtual void visit(PolyLineObject *object);

            virtual void visit(SensorObject *object);

            virtual void visit(PointSetObject *object);

            virtual void visit(Hinge *object);

            virtual void visit(Slider *object);

            virtual void visit(GridObject *object);

            virtual void visit(PointCloud *object);

            virtual void visit(LineCloud *object);

            virtual void visit(OccupancyGrid2D *object);

            virtual void visit(Label *object);

            /// @}

        private:
            void relay_shared_node_attributes(
                SceneObject *object, NodeID node_id) const throw();

            void relay_shared_primitive_attributes(
                SceneObject *object) const throw();

            void on_child_attached(SceneObject *object);

            void on_detached(SceneObject *object, bool subtree_root);

            void set_grid_vertices(PrimitiveID id, GridObject *object);

        private:
            IDTranslationTable &m_id_lut;

            renderer::Renderer &m_renderer;

            typedef std::vector<boost::signalslib::connection> Connections;

            typedef std::map<SceneObject *, Connections> ConnectionMap;

            ConnectionMap m_connection_map;
        };

        Impl m_impl;
    };

    typedef Singleton<RendererRelay> TheRendererRelay;


}


#endif // __PEEKABOT_RENDERER_RELAY_HH
