/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_SCALABLE_OBJECT_HH
#define __PEEKABOT_SCALABLE_OBJECT_HH


#include <boost/thread_safe_signal.hpp>

#include "Types.hh"
#include "XMLHandler.hh"


namespace peekabot
{

    class ScopedHandler;

    /**
     * \internal
     *
     * \brief An abstraction of objects in the scene that are scalable.
     * 
     * A non-instantiable class that adds to scene objects the scale 
     * property and methods for manipulating it. Classes that need 
     * this property inherit it from this class.
     */
    class ScalableObject
    {
    public:
        ScalableObject() throw();

        /** \brief Construct a scalable object using an XML document handler.
         */
        ScalableObject(ScopedHandler *handler);

        /** \brief Returns the scaling factor along the x axis.
         */
        real_unit get_x_scale() const throw();

        /** \brief Returns the scaling factor along the y axis.
         */
        real_unit get_y_scale() const throw();

        /** \brief Returns the scaling factor along the z axis.
         */
        real_unit get_z_scale() const throw();


        void set_scale(real_unit x_scale, real_unit y_scale, real_unit z_scale) throw();

        /** \brief Set the scaling along the x axis.
         * \exception AccessDenied The object is read-only, from the context it
         *            was accesed from.
         * \param x_scale The scaling factor.
         */
        void set_x_scale(real_unit x_scale) throw();


        /** \brief Set the scaling along the y axis.
         * \exception AccessDenied The object is read-only, from the context it
         *            was accesed from.
         * \param y_scale The scaling factor.
         */
        void set_y_scale(real_unit y_scale) throw();

        /** \brief Set the scaling along the z axis.
         * \exception AccessDenied The object is read-only, from the context it
         *            was accesed from.
         * \param z_scale The scaling factor.
         */
        void set_z_scale(real_unit z_scale) throw();


        /// \name Signals
        /// @{

        //boost::signal<void (ScalableObject *object)> m_on_set_scale;

        typedef boost::signal<void (ScalableObject *object, float x_scale, float y_scale, float z_scale)> OnScaleSetSignal;

        OnScaleSetSignal m_on_scale_set;

        /// @}


    private:
        /// \name XML handler methods
        /// @{

        /** \brief <scale> start handler for XML parsing.
         */
        void scale_start_handler(
            const std::string & name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        /** \brief <scale> cdata handler for XML parsing
         */
        void scale_cdata_handler(
            const std::string &cdata,
            ScopedHandler *handler) throw();

        /// @}

    private:
        /** \brief The scaling factor along the x axis.
         */
        real_unit m_x_scale;
        /** \brief The scaling factor along the y axis.
         */
        real_unit m_y_scale;
        /** \brief The scaling factor along the z axis.
         */
        real_unit m_z_scale;
    };

}


#endif // __PEEKABOT_SCALABLE_OBJECT_HH
