/*
 * Copyright Staffan Gimåker 2006-2009.
 * Copyright Anders Boberg 2006-2007.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * \brief \c SceneObject member function declarations.
 */

#include <stack>
#include <iostream>
#include <sstream>
#include <stdexcept>
#include <boost/bind.hpp>
#include <boost/function.hpp>
#include <boost/format.hpp>
#include <boost/foreach.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <boost/algorithm/string/trim.hpp>

#include "SceneObject.hh"
#include "SceneFileLoader.hh"
#include "IDAllocator.hh"


using namespace peekabot;


namespace
{
    void end_handler_wrapper(
        const std::string &, ScopedHandler *, 
        boost::function<void (void)> f)
    {
        f();
    }
}



// Do one normalization for every 10 transforms (25 breaks it, 15 almost breaks it..)
static int TRANSFORMS_PER_NORMALIZATION = 10;



namespace
{
    void do_nothing(const std::string &, ScopedHandler *) {}

    /*
     * XML handler that ignores all contained cdata and elements.
     */
    void external_handler(
        const std::string &name,
        ScopedHandler::AttributeMap &attributes,
        ScopedHandler *handler) throw()
    {
        ScopedHandler::TagScope scope;
        scope.cdata_functor = &do_nothing;
        scope.unhandled_functor = &external_handler;

        handler->enter_scope(scope);
    }
}




// Reserve IDs in the range [0,FIRST_UNRESERVED_ID)
IDAllocator<ObjectID> *SceneObject::ms_id_allocator(
    new DefaultIDAllocator<ObjectID>(0, FIRST_UNRESERVED_ID-1));



SceneObject::SceneObject(const std::string &default_name)
    : m_is_hidden(false),
      m_parent(0),
      m_layer(0),
      m_opacity(Opacity(false, 1)),
      m_name(default_name),
      m_id(ms_id_allocator->allocate()),
      m_transforms_since_normalization(0),
      m_color(0, 0.5, 0),
      m_mtow_is_valid(false)
{
    m_mtop.load_identity();
}

SceneObject::~SceneObject() 
{
    m_on_delete(this);

    // TODO: Who removes objects from the parent's child list?
    while( !m_children.empty() )
    {
        delete *m_children.begin();
        m_children.erase(m_children.begin());
    }

    // Release ID...
    ms_id_allocator->release(m_id);
}


bool SceneObject::mtow_is_valid() const throw()
{
    return m_mtow_is_valid;
}



const Matrix4ru & SceneObject::get_mtow() throw()
{
    if( !m_mtow_is_valid )
        validate_mtow_cache();

    return m_mtow;
}


const Matrix4ru &SceneObject::get_parent_mtow() throw()
{
    if( m_parent != 0 )
        return m_parent->get_mtow();
    else
        return get_mtow();
}


void SceneObject::validate_mtow_cache() throw()
{
    std::stack<SceneObject *> invalid;

    SceneObject *obj = this;
    while( obj && !obj->m_mtow_is_valid )
    {
        invalid.push(obj);
        obj = obj->m_parent;
    }
    
    while( !invalid.empty() )
    {
        SceneObject *obj = invalid.top();
        invalid.pop();
        
        if( obj->m_parent == 0 )
        {
            // Special case: the root node.
            obj->m_mtow = obj->m_mtop;
        }
        else
        {
            // The parent _always_ has a valid mtow, since we just 
            // calculated it.
            obj->m_mtow = obj->m_parent->m_mtow * obj->m_mtop;
        }
        
        obj->m_mtow_is_valid = true;
    }
}


void SceneObject::invalidate_mtow_cache()
{
    // In the default implementation we simply set m_mtow_is_valid to false
    // for this object and recursively down the tree.

    m_mtow_is_valid = false;
    m_on_mtow_changed(this);

    for( ChildContainer::iterator it = m_children.begin();
        it != m_children.end(); it++ )
    {
        (*it)->invalidate_mtow_cache();
    }
}


void SceneObject::set_object_id(ObjectID id)
{
    ObjectID old = m_id;
    m_on_set_object_id(this);
    m_id = id;
    m_on_object_id_set(this, old);
}


ObjectID SceneObject::get_object_id() const throw()
{
    return m_id;
}

uint32_t SceneObject::get_layer() const throw()
{
    return m_layer;
}

bool SceneObject::set_layer(uint32_t layer)
{
    //Then check that the layer number is valid
    if(layer >= NUMBER_OF_LAYERS)
        return true;

    m_layer = layer;
    m_on_layer_set(this, layer);

    return false;
}

const std::string& SceneObject::get_name() const throw()
{
    return m_name;
}

std::string SceneObject::get_name_string() const throw()
{
    return (get_meta_name().length() == 0) ? 
        (get_name()):(get_name() + ": " + get_meta_name());
}


void SceneObject::set_name(const std::string& name) throw(std::exception)
{
    // In order for the new name to be valid, we must make sure there are
    // no siblings of this object who have the same name.
    if(m_parent != 0)
    {
        ChildContainer::iterator child_it = m_parent->m_children.begin();
        
        for( ; child_it != m_parent->m_children.end(); ++child_it )
        {
            if( (*child_it)->get_name() == name && 
                (*child_it) != this )
                throw std::runtime_error(
                    "Failed to change name - name uniqueness constraint violated");
        }
    }

    m_name = name;
    m_on_name_set(this, name);
}

const std::string& SceneObject::get_meta_name() const throw()
{
    return m_meta_name;
}

void SceneObject::set_meta_name(const std::string &meta_name) 
{
    m_meta_name = meta_name;
    m_on_meta_name_set(this, meta_name);
}

const Opacity &SceneObject::get_opacity() const throw()
{
    return m_opacity;
}

void SceneObject::set_opacity(const Opacity& opacity)
{
    m_opacity = opacity;
    m_on_opacity_set(this, opacity);

    // Notify the relevant parts (i.e. the subtrees that doesn't have a 
    // node of absolute opacity at its root) of the subtree that the 
    // accumulated opacity has changed
    std::stack<SceneObject *> s;
    s.push( this );

    while( !s.empty() )
    {
        SceneObject *obj = s.top();
        s.pop();

        obj->m_on_accum_opacity_changed(obj);

        for( ChildContainer::iterator it = obj->m_children.begin();
             it != obj->m_children.end(); it++ )
        {
            if( !(*it)->get_opacity().is_absolute )
                s.push(*it);
        }
    }
}

void SceneObject::set_hidden(bool hidden)
{
    m_is_hidden = hidden;
    m_on_hidden_set(this, hidden);
}


bool SceneObject::is_hidden() const throw()
{
    return m_is_hidden;
}

real_unit SceneObject::get_accumulated_opacity() const throw()
{
    std::stack<const SceneObject*> path;
    real_unit accumulated_opacity;

    // Trace a path up to the first object with absolute opacity.
    const SceneObject *current = this;
    path.push(current);
    while(!(current->get_opacity().is_absolute) && current->m_parent != 0)
    {
        path.push(current->m_parent);
        current = current->m_parent;
    }
    
    // Start the accumulated opacity to the value of the absolute one.
    current = path.top();
    path.pop();
    accumulated_opacity = current->m_opacity.opacity;

    // Start accumulating...
    while(!path.empty())
    {
        current = path.top();
        path.pop();
        
        accumulated_opacity *= current->m_opacity.opacity;
    }

    return accumulated_opacity;
}


bool SceneObject::attach(
    SceneObject *subtree, NameConflictPolicy conflict_policy) 
    throw(std::runtime_error)
{
    bool ret = false;

    // Check for and/or resolve name conflicts...
    if( has_child(subtree->get_name()) )
    {
        if( conflict_policy == AUTO_ENUMERATE_ON_CONFLICT )
        {
            // Do a binary search to find a suitable suffixed name, we assume
            // that names are continously distributed... so it might fail, if 
            // it does we resort to using a linear search
            std::string name;
            uint32_t lb = 1;
            uint32_t ub = std::numeric_limits<uint32_t>::max();

            while( ub-lb > 1 )
            {
                uint32_t n = lb + (ub-lb)/2;
                name = (boost::format("%1%%2%") % subtree->get_name() % n).str();

                if( has_child(name) )
                    lb = n+1;
                else
                    ub = n;
            }

            if( ub-lb == 1 )
            {
                name = (boost::format("%1%%2%") % subtree->get_name() % lb).str();
                if( has_child(name) )
                    name = (boost::format("%1%%2%") % subtree->get_name() % ub).str();
            }

            // Did the binary search fail? If so, resort to a linear search
            if( has_child(name) )
            {
                uint32_t n = 0;
                do
                {
                    name = (boost::format("%1%%2%") % subtree->get_name() % ++n).str();
                }
                while( has_child(name) );
            }

            assert( !has_child(name) );

            subtree->set_name(name);
            ret = true;
        }
        else if( conflict_policy == REPLACE_ON_CONFLICT )
        {
            SceneObject *conflicting = get_child(subtree->get_name());
            conflicting->detach();
            delete conflicting;
        }
        else
            throw std::runtime_error(
                "Failed to attach subtree due to a name uniqueness violation");
    }

    subtree->m_parent = this;
    m_children.insert(subtree);

    m_on_child_attached(subtree);

    return ret;
}


void SceneObject::detach()
{
    if( get_parent() == 0 )
        throw std::logic_error("Can't detach parentless nodes");

    m_parent->m_children.erase(this);
    m_parent = 0;
    
    m_on_detached(this);

    //
    // Notify the object AND any descendents that the subtree they
    // belong to was detached
    //
    std::stack<SceneObject *> temp_stack;
    temp_stack.push(this);
    while( !temp_stack.empty() )
    {
        SceneObject *obj = temp_stack.top();
        temp_stack.pop();

        obj->m_on_subtree_detached(obj);

        BOOST_FOREACH( SceneObject *s, obj->m_children )
        {
            temp_stack.push(s);
        }
    }
}


bool SceneObject::insert(SceneObject* object)
{
    if( m_children.find(object) == m_children.end() )
        return true;

    m_children.insert(object);
    object->m_parent = this;

    return false;
}

bool SceneObject::is_descendant(SceneObject *object) const throw()
{
    // An object is a descendant of this object if it is a child of
    // this object, or if it is a descendant of one of this object's
    // children.
    for(ChildContainer::const_iterator it = m_children.begin();
        it != m_children.end(); ++it)
    {
        if(*it == object)
            return true;
        else 
            return (*it)->is_descendant(object);
    }
    return false;
}

const Matrix4ru & SceneObject::get_transformation() const throw()
{
    return m_mtop;
}


void SceneObject::set_transformation(const Matrix4ru &mtop) throw()
{
    m_mtop = mtop;
    // apply subclass specific changes
    //m_mtop *= post_set_transformation();
    invalidate_mtow_cache();

    // TODO: remove this...? set_transformation should get already valid matrices
    if( ++m_transforms_since_normalization >= TRANSFORMS_PER_NORMALIZATION )
        orthonormalize_transform();

    m_on_transformation_set(this, m_mtop);
}


void SceneObject::apply_transformation(const Matrix4ru &transform) throw()
{
    m_mtop *= transform;
    invalidate_mtow_cache();

    if( ++m_transforms_since_normalization >= TRANSFORMS_PER_NORMALIZATION )
        orthonormalize_transform();

    m_on_transformation_set(this, m_mtop);
}


void SceneObject::orthonormalize_transform() throw()
{
    //
    // Apply Gram-Schmidt's orthonormalization procedure 
    // to get an orthonormal basis (assuming the upper left 3x3 
    // matrix is a basis for R^3):
    //
    m_mtop(0).normalize();
    m_mtop(1).normalize();
    // The cross product of two normalized vectors is a normalized vector, thus
    // we don't need to normalize the third basis vector
    m_mtop(2) = Vector4ru(Vector3ru(m_mtop(0)).cross_product(m_mtop(1)), 0);
    m_mtop(1) = Vector4ru(Vector3ru(m_mtop(2)).cross_product(m_mtop(0)), 0);

    m_transforms_since_normalization = 0;

    invalidate_mtow_cache();
}


SceneObject *SceneObject::get_parent() throw()
{
    return m_parent;
}


SceneObject::ChildContainer &SceneObject::get_children() throw()
{
    return m_children;
}


const SceneObject::ChildContainer &SceneObject::get_children() const throw()
{
    return m_children;
}


SceneObject *SceneObject::get_child(const std::string &name) 
    throw(ChildNotFound)
{
    for(ChildContainer::iterator it = m_children.begin();
        it != m_children.end(); ++it)
        if((*it)->get_name() == name)
            return *it;

    throw ChildNotFound(name);
}


const SceneObject *SceneObject::get_child(const std::string &name) 
    const throw(ChildNotFound)
{
    for(ChildContainer::const_iterator it = m_children.begin();
        it != m_children.end(); ++it)
        if((*it)->get_name() == name)
            return *it;

    throw ChildNotFound(name);
}


uint32_t SceneObject::get_child_count() const throw()
{
    return m_children.size();
}

bool SceneObject::has_child(const std::string &name) const throw()
{
    for(ChildContainer::const_iterator it = m_children.begin();
        it != m_children.end(); ++it)
        if((*it)->get_name() == name)
            return true;
    return false;
}

bool SceneObject::has_descendent(const std::string &rel_path) const throw()
{
    std::vector<std::string> components;
    boost::split( components, rel_path, boost::is_any_of(".") );

    return _get_descendent(components.begin(), components.end()) != 0;
}


SceneObject *SceneObject::get_descendent(const std::string &rel_path) 
    throw(std::runtime_error)
{
    std::vector<std::string> components;
    boost::split( components, rel_path, boost::is_any_of(".") );

    SceneObject *ret = _get_descendent(
        components.begin(), components.end());
    
    if( ret == 0 )
        throw std::runtime_error("Descendent '" + rel_path + "' not found");

    return ret;
}


const SceneObject *SceneObject::get_descendent(const std::string &rel_path) 
    const throw(std::runtime_error)
{
    std::vector<std::string> components;
    boost::split( components, rel_path, boost::is_any_of(".") );

    const SceneObject *ret = _get_descendent(
        components.begin(), components.end());
    
    if( ret == 0 )
        throw std::runtime_error("Descendent '" + rel_path + "' not found");

    return ret;
}


void SceneObject::clear_children() throw()
{
    m_children.clear();
}


const Vector4ru &SceneObject::get_position() const throw()
{
    return m_mtop(3);
}


void SceneObject::set_position(const Vector3ru &pos) throw()
{
    Matrix4ru t = m_mtop;
    t(3) = Vector4ru(pos, 1);
    set_transformation(t);
}


const Vector4ru &SceneObject::get_world_position() throw()
{
    return get_mtow()(3);
}


void SceneObject::set_world_position(const Vector3ru &pos) throw()
{
    if( get_parent() )
    {
        const Matrix4ru &m = get_parent()->get_mtow();
        set_position(m.inverse() * Vector4ru(pos, 1));
    }
    else
    {
        // No parent found, e.g. we might just say the parent's
        // transformation is the identity:
        set_position(pos);
    }
}

void SceneObject::set_world_position(
    real_unit x, real_unit y, real_unit z) throw()
{
    set_world_position(Vector3ru(x,y,z));
}


void SceneObject::rotate_x(real_unit angle) throw()
{
    apply_transformation(Matrix4ru::rotate_x(angle));
}


void SceneObject::rotate_y(real_unit angle) throw()
{
    apply_transformation(Matrix4ru::rotate_y(angle));
}


void SceneObject::rotate_z(real_unit angle) throw()
{
    apply_transformation(Matrix4ru::rotate_z(angle));
}


void SceneObject::set_color(const RGBColor &color, bool recursive)
{
    m_color = color;
    m_on_color_set(this, color);

    if( recursive )
    {
        for( ChildContainer::iterator it = m_children.begin();
             it != m_children.end(); it++ )
        {
            (*it)->set_color(color, true);
        }
    }
}


RGBColor SceneObject::get_color() const throw()
{
    return m_color;
}


SceneObject::SceneObject(
    const std::string &default_name, ScopedHandler *handler)
    : m_is_hidden(false),
      m_parent(0),
      m_layer(0),
      m_opacity(Opacity(false, 1)),
      m_name(default_name),
      m_id(ms_id_allocator->allocate()),
      m_transforms_since_normalization(0),
      m_color(0, 0.5, 0),
      m_mtow_is_valid(false)
{
    m_mtop.load_identity();
    
    // If there is a layer or color variable set, change the color and/or
    // layer of this object accordingly.
    ScopedMap & variables = handler->get_variables();
       
    if(variables.exists<uint32_t>("layer"))
        m_layer = variables.get_variable<uint32_t>("layer") - 1;

    // Save the original name of the object for later (overriden by
    // the name_cdata_handler, when a name tag is present).
    variables.get_variable<std::map<SceneObject *, std::string> >(
        "orig_names")[this] = m_name;

    // Attach the object to its parent.
    SceneObject* parent = variables.get_variable<SceneObject*>("current_object");
    parent->attach(this, AUTO_ENUMERATE_ON_CONFLICT);

    ScopedHandler::TagScope scope;

    // Specify the scene object end handler for the new scope.
    scope.end_functors.push_back(
        boost::bind(&SceneObject::end_handler, this, _1, _2));

    // If the object's color is being overriden, set up a function
    // that will set the appropriate color when the end tag is encountered -
    // we can't set it right away since ctors higher up in the hierarchy
    // might change m_color (if they have special defaults, like models and 
    // cameras)
    if(variables.exists<RGBColor>("color"))
    {
        const RGBColor &color = variables.get_variable<RGBColor>("color");

        boost::function<void (void)> f = boost::bind(
            &SceneObject::set_color, this, color, false);

        scope.end_functors.push_back(
            boost::bind(&end_handler_wrapper, _1, _2, f));
    }


    handler->enter_scope(scope);


    // Register the name tag start handler
    handler->add_start_handler(
        "name",        
        boost::bind(&SceneObject::name_start_handler, this, _1, _2, _3));

    // Register the meta name tag start handler
    handler->add_start_handler(
        "meta_name",        
        boost::bind(&SceneObject::meta_name_start_handler, this, _1, _2, _3));

    // Register the opacity tag start handler
    handler->add_start_handler(
        "opacity",        
        boost::bind(&SceneObject::opacity_start_handler, this, _1, _2, _3));

    // Register the hidden tag start handler
    handler->add_start_handler(
        "hidden",        
        boost::bind(&SceneObject::hidden_start_handler, this, _1, _2, _3));

    // Register the transform tag start handler
    handler->add_start_handler(
        "transform",        
        boost::bind(&SceneObject::transform_start_handler, this, _1, _2, _3));

    // Register the children tag start handler
    handler->add_start_handler(
        "children",        
        boost::bind(&SceneObject::children_start_handler, this, _1, _2, _3));

    // Register handler for ignoring all cdata/elements wrapped in an
    // <external>-element.
    handler->add_start_handler("external", &external_handler);
}

void SceneObject::end_handler(const std::string &name,
                              ScopedHandler *handler) 
    throw(std::runtime_error)
{
    // Pop the current_object variable since we're exiting an object element
    ScopedMap & variables = handler->get_variables();
    variables.pop_variable<SceneObject*>("current_object");

    ++variables.get_variable<unsigned int>("object_counter");
}

void SceneObject::name_start_handler(
    const std::string &name,
    ScopedHandler::AttributeMap &attributes,
    ScopedHandler *handler) throw()
{
    ScopedHandler::TagScope name_scope;
    name_scope.cdata_functor = boost::bind(&SceneObject::name_cdata_handler,
                                           this, _1, _2);

    // Enter a new scope where only cdata is valid, and is handled by
    // the name_cdata_handler() method.
    handler->enter_scope(name_scope);
}

void SceneObject::name_cdata_handler(const std::string &cdata,
                                     ScopedHandler *handler)
    throw(std::runtime_error)
{
    std::string name = cdata;
    boost::trim(name);

    try
    {
        set_name(name);
        // Override the default original name
        handler->get_variables().get_variable<
            std::map<SceneObject *, std::string> >("orig_names")[this] = name;
    }
    catch(std::exception &e)
    {
        std::string error = "Name conflict during parsing: ";
        error += name;
        throw(std::runtime_error(error));
    }
}

void SceneObject::meta_name_start_handler(
    const std::string &name,
    ScopedHandler::AttributeMap &attributes,
    ScopedHandler *handler) throw()
{
    ScopedHandler::TagScope meta_scope;
    meta_scope.cdata_functor = boost::bind(
        &SceneObject::meta_name_cdata_handler, this, _1, _2);
    
    // Enter a new scope where only cdata is valid, and is handled by
    // the meta_cdata_handler() method.
    handler->enter_scope(meta_scope);
}

void SceneObject::meta_name_cdata_handler(const std::string &cdata,
                                          ScopedHandler *handler)
    throw()
{
    set_meta_name(cdata);
}

void SceneObject::opacity_start_handler(
    const std::string &name,
    ScopedHandler::AttributeMap &attributes,
    ScopedHandler *handler) throw()
{
    ScopedHandler::TagScope opacity_scope;
    
    bool absolute = false;

    // Check if the opacity is absolute
    absolute = attributes["absolute"] == "true";

    opacity_scope.cdata_functor = boost::bind(
        &SceneObject::opacity_cdata_handler, this, absolute, _1, _2);
    
    // Enter a new scope where only cdata is valid, and is handled by
    // the meta_cdata_handler() method.
    handler->enter_scope(opacity_scope);
}

void SceneObject::opacity_cdata_handler(bool absolute,
                                        const std::string &cdata,
                                        ScopedHandler *handler)
    throw()
{
    set_opacity(Opacity(absolute, atof(cdata.c_str())));
}

void SceneObject::hidden_start_handler(
    const std::string &name,
    ScopedHandler::AttributeMap &attributes,
    ScopedHandler *handler) throw()
{

    m_is_hidden = true;

    // Enter a new scope to make sure the scope stack
    // won't be f***ed up (empty tags have end tags,
    // even if they're not explicit, so the top
    // end handler will get called).
    handler->enter_scope();
}

void SceneObject::transform_start_handler(
    const std::string &name,
    XMLHandler::AttributeMap &attributes,
    ScopedHandler *handler) throw()
{
    ScopedHandler::TagScope scope;

    // Add transformation tag handlers to the scope.
    scope.start_functors["rotate"] = boost::bind(
        &SceneObject::rotate_start_handler, this, _1, _2, _3);

    scope.start_functors["translate"] = boost::bind(
        &SceneObject::translate_start_handler, this, _1, _2, _3);

    scope.start_functors["matrix"] = boost::bind(
        &SceneObject::matrix_start_handler, this, _1, _2, _3);

    handler->enter_scope(scope);
}


void SceneObject::rotate_start_handler(
    const std::string &name,
    XMLHandler::AttributeMap &attributes,
    ScopedHandler *handler) throw()
{
    using namespace peekabot::transformation_target;

    if(name == "rotate")
    {
        ScopedHandler::TagScope rotation_scope;
      
        rotation_scope.start_functors["axis"] = boost::bind(
            &SceneObject::rotate_start_handler, this, _1, _2, _3);
        
        rotation_scope.start_functors["pivot"] = boost::bind(
            &SceneObject::rotate_start_handler, this, _1, _2, _3);
        
        rotation_scope.start_functors["radians"] = boost::bind(
            &SceneObject::rotate_start_handler, this, _1, _2, _3);    
        
        rotation_scope.start_functors["degrees"] = boost::bind(
            &SceneObject::rotate_start_handler, this, _1, _2, _3);

        rotation_scope.end_functors.push_back(
            boost::bind(&SceneObject::rotate_end_handler, this, _1, _2));
        
        handler->enter_scope(rotation_scope);
    }
    else if(name == "axis")
    {
        // Check which coordinate system we're using
        CoordinateSystem system;
        if(attributes["system"] == "world")
            system = WORLD_COORDINATES;
        else if(attributes["system"] == "parent")
            system = PARENT_COORDINATES;
        else
            system = LOCAL_COORDINATES;

        ScopedHandler::TagScope axis_scope;

        axis_scope.cdata_functor = boost::bind(
            &SceneObject::rotate_cdata_handler, this, 1, // 1 -> tag = axis
            system, _1, _2); 

        handler->enter_scope(axis_scope);
    }
    else if(name == "pivot")
    {
        // Check which coordinate system we're using
        CoordinateSystem system;
        if(attributes["system"] == "world")
            system = WORLD_COORDINATES;
        else if(attributes["system"] == "parent")
            system = PARENT_COORDINATES;
        else
            system = LOCAL_COORDINATES;

        ScopedHandler::TagScope pivot_scope;

        pivot_scope.cdata_functor = boost::bind(
            &SceneObject::rotate_cdata_handler, this, 2, // 2 -> tag = pivot
            system, _1, _2); 

        handler->enter_scope(pivot_scope);
    }
    else if(name == "radians")
    {
        ScopedHandler::TagScope radians_scope;

        radians_scope.cdata_functor = boost::bind(
            &SceneObject::rotate_cdata_handler, this, 3, // 3 -> tag = radians
            LOCAL_COORDINATES, _1, _2); // bogus system

        handler->enter_scope(radians_scope);
    }
    else if(name == "degrees")
    {
        ScopedHandler::TagScope degrees_scope;

        degrees_scope.cdata_functor = boost::bind(
            &SceneObject::rotate_cdata_handler, this, 4, // 4 -> tag = degrees 
            LOCAL_COORDINATES, _1, _2); // bogus system

        handler->enter_scope(degrees_scope);
    }
}

void SceneObject::rotate_cdata_handler(
    uint32_t tag,
    transformation_target::CoordinateSystem system,
    const std::string &cdata,
    ScopedHandler *handler) throw()
{
    using namespace peekabot::transformation_target;
    ScopedMap &variables = handler->get_variables();

    Matrix4ru parent_mtow = get_parent_mtow();

    std::stringstream s(cdata);

    if(tag == 1 || tag == 2) // axis or pivot tag
    {
        Vector4ru vec;
        s >> vec(0) >> vec(1) >> vec(2);

        if(tag == 1) // If we're transforming an axis, vec(3) needs to be 0
            vec(3) = 0;
        if(tag == 2) // For a point coordinate, it needs to be 1.
            vec(3) = 1;

        // If we're using parent or local coordinates we need to transform
        // the vector.
        if(system == PARENT_COORDINATES)
        {
            vec = parent_mtow * vec;
        }
        else if(system == LOCAL_COORDINATES)
        {
            vec = parent_mtow * m_mtop * vec;
        }

        // Always zero for axes, also zero when we translate to pivot!
        vec(3) = 0;

        if(tag == 1)
            variables.push_variable("rotation_axis", vec);
        else
            variables.push_variable("rotation_pivot", vec);
    }
    else // radians or degrees tag
    {
        real_unit rad;
        s >> rad;
        
        if(tag == 4) // If it's in degrees, convert into radians
            rad *= 2.0 * 3.141592654 / 360.0;

        variables.push_variable("rotation_rad", rad);
    }
}

void SceneObject::rotate_end_handler(const std::string &name,
                                     ScopedHandler *handler) throw()
{
    // Fetch rotation parameters
    ScopedMap &variables = handler->get_variables();
    Vector4ru axis;
    if(variables.exists<Vector4ru>("rotation_axis"))
    {
        axis = variables.get_variable<Vector4ru>("rotation_axis");
        variables.pop_variable<Vector4ru>("rotation_axis");
    }
    else
        axis = Vector4ru(1, 0, 0, 0);

    Vector4ru pivot;
    if(variables.exists<Vector4ru>("rotation_pivot"))
    {
        pivot = variables.get_variable<Vector4ru>("rotation_pivot");
        variables.pop_variable<Vector4ru>("rotation_pivot");
    }
    else
        pivot = Vector4ru(0, 0, 0, 0);

    real_unit rad =  variables.get_variable<real_unit>("rotation_rad");
    variables.pop_variable<real_unit>("rotation_rad");

    // Get the parent and local mtow matrices
    Matrix4ru parent_m = get_parent_mtow();
    Matrix4ru my_m = get_mtow();

    // Transpose the object to the pivot
    my_m(3) -= pivot;

    // Rotate
    my_m *= Matrix4ru::rotate(axis, rad);
    
    // Move back
    my_m(3) += pivot;

    // Translate back to local...
    my_m *= parent_m.inverse();
        
    set_transformation(my_m);
}


void SceneObject::translate_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw()
{
    using namespace peekabot::transformation_target;

    // Check which coordinate system we're using
    CoordinateSystem system;
    if(attributes["system"] == "world")
        system = WORLD_COORDINATES;
    else if(attributes["system"] == "parent")
        system = PARENT_COORDINATES;
    else
        system = LOCAL_COORDINATES;

    ScopedHandler::TagScope scope;
    
    scope.cdata_functor = boost::bind(
        &SceneObject::translate_cdata_handler, this, system, _1, _2); 
    
    handler->enter_scope(scope);    
}


void SceneObject::translate_cdata_handler(
    transformation_target::CoordinateSystem system,
    const std::string &cdata,
    ScopedHandler *handler) throw()
{
    using namespace peekabot::transformation_target;

    // Read coordinate information
    std::stringstream s(cdata);
    real_unit x, y, z;
    s >> x >> y >> z;

    Matrix4ru m = m_mtop;

    if(system == PARENT_COORDINATES)
    {
        m *= Matrix4ru::translate(x, y, z);
    }
    else if(system == WORLD_COORDINATES)
    {  
        Matrix4ru parent_mtow = get_parent_mtow();
        Matrix4ru mtow = get_mtow();
        
        m = (parent_mtow.inverse() * 
             Matrix4ru::translate(x, y, z))* mtow;
    }

    else if(system == LOCAL_COORDINATES)
    {
        m(3) += (x * m(0) +
                 y * m(1) +
                 z * m(2));
    }

    set_transformation(m);
}

void SceneObject::matrix_start_handler(
    const std::string &name,
    XMLHandler::AttributeMap &attributes,
    ScopedHandler *handler) throw()
{
    using namespace peekabot::transformation_target;

    // Check which coordinate system we're using
    CoordinateSystem system;
    if(attributes["system"] == "world")
        system = WORLD_COORDINATES;
    else if(attributes["system"] == "parent")
        system = PARENT_COORDINATES;
    else
        system = LOCAL_COORDINATES;

    ScopedHandler::TagScope scope;
    
    scope.cdata_functor = boost::bind(
        &SceneObject::matrix_cdata_handler, this, system, _1, _2); 
    
    handler->enter_scope(scope);        
}


void SceneObject::matrix_cdata_handler(
    transformation_target::CoordinateSystem system,
    const std::string &cdata,
    ScopedHandler *handler) throw()
{
    using namespace peekabot::transformation_target;

    std::stringstream s(cdata);
    Matrix4ru m;



    // Read our matrix.
    s >> m(0, 0) >> m(0, 1) >> m(0, 2) >> m(0, 3)
      >> m(1, 0) >> m(1, 1) >> m(1, 2) >> m(1, 3)
      >> m(2, 0) >> m(2, 1) >> m(2, 2) >> m(2, 3)
      >> m(3, 0) >> m(3, 1) >> m(3, 2) >> m(3, 3);

    // Normalize to make sure it's good.
    m(0).normalize();
    m(1).normalize();
    m(2).normalize();

    if(system == WORLD_COORDINATES)
    {
        // Here we need to apply the inverse parent to translate into local
        Matrix4ru parent_mtow = get_parent_mtow();
        m *= parent_mtow.inverse();
    }

    // If the system was PARENT_COORDINATES, we're safe.

    set_transformation(m);
}

void SceneObject::children_start_handler(
    const std::string &name,
    XMLHandler::AttributeMap &attributes,
    ScopedHandler *handler) throw()
{
    // children scope - allow all global start handlers
    ScopedHandler::TagScope children_scope;
    children_scope.start_functors = SceneFileLoader::get_start_handlers();
    handler->enter_scope(children_scope);
}


SceneObject *SceneObject::_get_descendent(
    const std::vector<std::string>::const_iterator first,
    const std::vector<std::string>::const_iterator last) throw()
{
    if( first == last )
        return 0; // Path is empty <=> no such descendent!

    for( ChildContainer::iterator it = m_children.begin();
         it != m_children.end(); ++it )
    {
        if( (*it)->get_name() == *first )
        {
            if( std::distance(first, last) == 1 )
            {
                return *it;
            }
            else
            {
                return (*it)->_get_descendent(first+1,last);
            }
        }
    }

    return 0;
}


const SceneObject *SceneObject::_get_descendent(
    const std::vector<std::string>::const_iterator first,
    const std::vector<std::string>::const_iterator last) const throw()
{
    if( first == last )
        return 0; // Path is empty <=> no such descendent!

    for( ChildContainer::const_iterator it = m_children.begin();
         it != m_children.end(); ++it )
    {
        if( (*it)->get_name() == *first )
        {
            if( std::distance(first, last) == 1 )
            {
                return *it;
            }
            else
            {
                return (*it)->_get_descendent(first+1,last);
            }
        }
    }

    return 0;
}
