/*
 * Copyright Staffan Gimåker 2006-2009.
 * Copyright Anders Boberg 2006-2007.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PEEKABOT_SCENE_OBJECT_HH_INCLUDED
#define PEEKABOT_SCENE_OBJECT_HH_INCLUDED


#include <utility>
#include <string>
#include <exception>
#include <stdexcept>
#include <set>
#include <boost/thread_safe_signal.hpp>

#include "Types.hh"
#include "Matrix.hh"
#include "Vector.hh"
#include "XMLHandler.hh"


namespace peekabot
{
    class SceneObject;
    class ObjectVisitor;
    template<class IDType> class IDAllocator;

    class ScopedHandler;

    /**
     * \internal
     *
     * \brief The base object class.
     *
     * The SceneObject class and its derivates constitute the core of peekabot's
     * data repository.
     * The scene, or world if you wish, in peekabot is comprised of a tree of 
     * SceneObject-derivates that are manipulated by the user and clients.
     *
     * All objects share a set of common attributes such as a transformation, 
     * color, layer, etc.
     *
     * \section object_naming Naming and paths
     *
     * In order for clients to have the ability to manipulate already existing 
     * objects all objects are named. A \e path to an object consists of its name
     * appended to its parent's path, separated by a dot. For example 
     * <em>root.robot.scanner</em> is a path to the scanner object attached to 
     * robot, which is in turned attached to the scene root.
     *
     * Given the explaination of paths above the two restrictions that apply to 
     * names seem natural:
     * \li They cannot contain dots.
     * \li All children of an object must have unique names.
     *
     * To ease the burden on the end user, name conflicts can often be solved 
     * automatically by peekabot, see NameConflictPolicy.
     *
     * \sa set_name(), get_name(), NameConflictPolicy
     */
    class SceneObject
    {
        friend class SceneTree;

    public:
        typedef std::set<SceneObject *> ChildContainer;

        class ChildNotFound : public std::runtime_error
        {
        public:            
            ChildNotFound(const std::string &name)
                : std::runtime_error("Child '" + name + "' not found") {}
        };



        SceneObject(const std::string &default_name);

        /** \brief Create a \c SceneObject to be initialized by a \c XMLHandler.
         */
        SceneObject(const std::string &default_name, ScopedHandler *handler);

        virtual ~SceneObject();

        /**
         * \brief Retrieves the model space to world space matrix.
         *
         * If \c mtow is \c true, the cached value stored in \c m_mtow
         * will promptly be returned. Otherwise, the cache will be 
         * recalculated and updated.
         *
         * \pre None.
         * \post The model space to world space matrix has been returned.
         *       \c m_mtow_is_valid is \c true.
         * \return The model space to world space matrix.
         * \note Calling this method may be of high cost!
         * \note The method might alter internal state of the object.
         *
         * \todo Ponder the possibility of exceptions, and how to deal with 
         * them.
         */
        const Matrix4ru &get_mtow() throw();

        /**
         * \brief Get the parent space to world space transformation matrix.
         *
         * If the object has no parent, its mtow (= mtop) will be returned.
         */
        const Matrix4ru &get_parent_mtow() throw();

        // Throws if the parent is read only
        // returns true if the operation failed - failure occurs when the 
        // operation is called on the root or the default camera, which cannot
        // be removed.
        void detach();

        /** \brief Returns true if \c object is a descendant of this object.
         */
        bool is_descendant(SceneObject *object) const throw();

        /**
         * \brief Returns the parent of an object.
         *
         * If the object is protected it will return a protected instance of 
         * the object's parent. If it is not, a protected instance with the
         * server as the accessing context will be returned. This behaviour is
         * specified by \c SceneObjectGuard::get_parent().
         *
         * \return The object's parent, or a zero pointer if it has no parent.
         * 
         * \note No class other than \c SceneObjectGuard is allowed to
         *       override this method.
         */
        SceneObject *get_parent() throw();

        /**
         * \brief Get the object's children.
         *
         * If the object is protected it will return a list of protected instance of 
         * the object's children. If it is not, a list of protected instance with 
         * the server as the accessing context will be returned. This behaviour 
         * is specified and implemented in \c SceneObjectGuard::get_children().
         *
         * \return A list of (guarded) references to all the object's children. 
         *         The list can be empty.
         */
        ChildContainer &get_children() throw();

        const ChildContainer &get_children() const throw();

        SceneObject *get_child(const std::string &name) 
            throw(ChildNotFound);

        const SceneObject *get_child(const std::string &name) 
            const throw(ChildNotFound);

        bool has_child(const std::string &name) const throw();

        /**
         * \brief Check whether the given (relative) path is a descendent.
         *
         * \param rel_path The path of the descendent to return, relative the 
         * object.
         * \return \c true if \a rel_path is a descendent of the object, 
         * \c false otherwise.
         */
        bool has_descendent(const std::string &rel_path) const throw();

        /**
         * \brief Return the descendent described by the given (relative) path.
         *
         * \param rel_path The path of the descendent to return, relative the 
         * object.
         * \return The descendent object.
         * \throw std::runtime_error Thrown if \a rel_path points to a 
         * non-existant object.
         */
        SceneObject *get_descendent(const std::string &rel_path) 
            throw(std::runtime_error);

        /**
         * \copydoc SceneObject::get_descendent(const std::string &)
         */
        const SceneObject *get_descendent(const std::string &rel_path) 
            const throw(std::runtime_error);

        /** \brief Returns the number of children of this object.
         */
        uint32_t get_child_count() const throw();

        /** \brief Clears the child set of this object.
         *
         * \c WARNING: This is not the proper way to remove objects from the scene!
         * Instead use detach() on the child. This method exists to enable the user
         * to clear the child set silently without setting off any triggers and
         * without deleting the objects, which may be desirable under some
         * circumstances.
         */
         void clear_children() throw();

        /* ... */

        /**
         * \brief Get the object's \c ObjectID.
         *
         * \return The object's \c ObjectID, if it has not been assigned an ID 
         *         the value returned is undefined.
         */
        ObjectID get_object_id() const throw();

        // !!!!!! !!!!! TODO SHOULDN'T BE PUBLIC!!!!!!!! !!!!!
        void set_object_id(ObjectID id);
        // !!!!!! !!!!! TODO SHOULDN'T BE PUBLIC!!!!!!!! !!!!!


        /**
         * \brief Attach a child object to the object.
         *
         * In order for this operation to succeed one condition must be met:
         * No child of the object has the same name as the object to be added.
         *
         * \return \c true is returned if auto-enumeration was enabled and the
         * name was changed to resolve a name conflict.
         * 
         * \exception std::runtime_error Cast if adding the subtree would've 
         * caused a name uniqueness violation.
         */
        bool attach(SceneObject *subtree, 
                    NameConflictPolicy conflict_policy = FAIL_ON_CONFLICT) 
            throw(std::runtime_error);

        /**
         * \brief Get which layer the object is (shown) in.
         *
         * If no layer has been assigned to the object, the layer it is in
         * is undefined, but consistent among all objects. The default layer 
         * is guaranteed to be a valid layer.
         *
         * \return The layer in which the object is in.
         */
        uint32_t get_layer() const throw();

        /**
         * \brief Set the layer the object is (shown) in.
         *
         * Not all values for \c layer are valid. Negative numbers
         * are never allowed and other number might be rejected. In case
         * of failure to comply, the operation will return \c true.
         *
         * \param layer The layer in which to move the object into.
         * \return \c false if the operation succeeded, \c true if it failed.
         */
        bool set_layer(uint32_t layer);

        /**
         * \brief Get the object's name.
         *
         * \note The reference returned isn't guaranteed to have a long 
         *       lifetime - it may change or be deleted after the calling
         *       method has exited.
         * \note An object's name is always unique among its parent's children.
         *
         * \return A reference to the object's meta name.
         * \sa default_name()
         */
        const std::string &get_name() const throw();

        /**
         * \brief Get a concatenated string consisting of the objects name
         * followed by its meta name, if it has one.
         *
         * This method returns a string on the format
         * <tt>name: meta name</tt>. 
         *
         * \return A copy of the concatenated string.
         */
        std::string get_name_string() const throw();


        /**
         * \brief Set the object's name.
         *
         * The name of an object must be unique among it's parent's children. 
         * Thus, if the \c name parameter fails to fullfill this goal, the
         * operation will fail.
         *
         * \pre The \c name parameter is unique among the object's parent's 
         *      children.
         * \post The name of the object is changed to the value supplied.
         *
         * \return \c false if the operation succeeded, \c true if it failed.
         * \param name The name to assign the object.
         */
        void set_name(const std::string &name) throw(std::exception);

        /**
         * \brief Get the object's meta name.
         *
         * The default meta name for an object is the empty string, which
         * will be returned if no meta name has been assigned the object.
         *
         * \note The reference returned isn't guaranteed to have a long 
         *       lifetime - it may change or be deleted after the calling
         *       method has exited.
         *
         * \return A reference to the object's meta name.
         */
        const std::string &get_meta_name() const throw();

        /**
         * \brief Set the object's meta name.
         *
         * \param meta_name The meta name to assign to the object.
         */
        void set_meta_name(const std::string &meta_name);

        /**
         * \brief Get the object's opacity value.
         *
         * The default opacity value for a node is <tt><false, 1.0></tt>. If the
         * opacity value has not been altered, this value will be returned.
         *
         * \note Note that the value returned by this method is not necessarily
         *       the same as it's \e accumulated opacity (i.e. the opacity used
         * to draw the object).
         *
         * \return The object's opacity.
         * \sa \c get_accumulated_opacity()
         */
        const Opacity &get_opacity() const throw();
            
        /**
         * \brief Set the object's opacity.
         *
         * For well defined semantics of the \c Opacity type, see its 
         * documentation.
         *
         * \param opacity The opacity to assign to the object.
         */
        void set_opacity(const Opacity &opacity);


        /** 
         * \brief Sets the visibility of an object.
         *
         * Sets whether this object is hidden or visible.
         *
         * \param hide \c true to set object as hidden, \c false to set as
         * shown.
         */
        void set_hidden(bool hidden);

        /**
         * \brief Returns whether this object is hidden or not.
         *
         * \return \c true if the object is hidden, \c false if it is shown.
         */
        bool is_hidden() const throw();

        /**
         * \brief Get the object's accumulated opacity value.
         *
         * See \c Opacity for a discussion of how opacity propagates down the 
         * object hierarchy.
         *
         * \return The object's accumulated opacity.
         */
        real_unit get_accumulated_opacity() const throw();


        /** \brief \brief Accept \c ObjectVisitor objects.
         *
         * Concrete derived classes implement this method by simply calling
         * \c ObjectVisitor::visit(this), thus allowing for a convenient way to
         * do automatic type casting.
         *
         */

        virtual void accept(ObjectVisitor *visitor) throw() = 0;

        /** \brief Returns a constant reference to the transformation matrix of
         * this \c SceneObject.
         */
        const Matrix4ru &get_transformation() const throw();

        /** \brief Sets the transformation matrix of this \c SceneObject.
         *
         * \param Constant reference to the transformation matrix to be copied
         * for this object.
         */
        void set_transformation(const Matrix4ru &mtop) throw();

        /** \brief Applies a transformation through multiplication with a
         *  transformation matrix.
         *
         *  \param transform The transformation to apply.
         */
        void apply_transformation(const Matrix4ru &transform) throw();

        /** \brief Convinience functions implemented in terms of get-, set- and 
         * apply_transformation.
         */
        const Vector4ru &get_position() const throw();

        /** \brief Sets position by entering a positional vector into the
         *  transformation matrix of this object.
         *
         * \param pos The vector to be inserted into the position column of the
         * the transformation matrix.
         */
        void set_position(const Vector3ru &pos) throw();

        /** \brief Sets position by entering coordinates into transformation
         *  matrix.
         *
         * \param x The value to be inserted into the x row of the positional
         * column of the transformation matrix. 
         * \param y The value to be inserted into the y row of the positional
         * column of the transformation matrix. 
         * \param z The value to be inserted into the z row of the positional
         * column of the transformation matrix. 
         */
        void set_position(real_unit x, real_unit y, real_unit z);

        /** \brief Returns the positional vector of this object in world
         * coordinates.
         */
        const Vector4ru &get_world_position() throw();


        /** \brief Sets the position of this object to a positional vector given
         * in world coordinates.
         *
         * \param pos The vector to be transformed into the model coordinate
         * system and entered into the position column of the transformation
         * matrix.
         */
        void set_world_position(const Vector3ru &pos) throw();

        /** \brief Sets the position of this object to a poisition given in
         * world coordinates.
         *
         * \param x The value to be transformed into the model coordinate
         * system and inserted into the x row of the positional column of the
         * transformation matrix. 
         * \param y The value to be transformed into the model coordinate
         * system and inserted into the y row of the positional column of the
         * transformation matrix. 
         * \param z The value to be transformed into the model coordinate
         * system and inserted into the z row of the positional column of the
         * transformation matrix. 
         */
        void set_world_position(real_unit x, real_unit y, real_unit z) throw();

        /** \brief Rotates the object around a given axis.
         *
         * \param axis The axis vector of rotation.
         * \param angle The angle of rotation in radians.
         */
        void rotate(const Vector4ru &axis, real_unit angle);

        /** \brief Rotates the object around a given axis.
         *
         * \param x The x coordinate of the axis vector of rotation.
         * \param y The y coordinate of the axis vector of rotation.
         * \param z The z coordinate of the axis vector of rotation.
         * \param angle The angle of rotation in radians.
         */
        void rotate(real_unit x, real_unit y, real_unit z, real_unit angle) throw();

        /** \brief Rotates the object around its x axis.
         *
         * \param angle The angle of rotation in radians.
         */
        void rotate_x(real_unit angle) throw();

        /** \brief Rotates the object around its y axis.
         *
         * \param angle The angle of rotation in radians.
         */
        void rotate_y(real_unit angle) throw();

        /** \brief Rotates the object around its z axis.
         *
         * \param angle The angle of rotation in radians.
         */
        void rotate_z(real_unit angle) throw();


        /** \brief Sets the color of this object to the specified RGB value.
         *
         * \param color The RGB color value.
         * \param recursive If true the color will also be set on the 
         * descendents of the object.
         */
        void set_color(const RGBColor &color, bool recursive = false);
        
        /** \brief Gets the current color value of this object.
         */
        RGBColor get_color() const throw();


        /// \name Signals
        /// @{

        typedef boost::signal<void (SceneObject *child)> OnChildAttachedSignal;
        
        OnChildAttachedSignal m_on_child_attached;

        typedef boost::signal<void (SceneObject *object)> OnDetachedSignal;

        OnDetachedSignal m_on_detached;

        typedef boost::signal<void (SceneObject *object)> OnSubtreeDetachedSignal;

        OnSubtreeDetachedSignal m_on_subtree_detached;

        typedef boost::signal<void (SceneObject *object, const Matrix4ru &mtop)> OnTransformationSetSignal;
        
        OnTransformationSetSignal m_on_transformation_set;

        typedef boost::signal<void (SceneObject *object, const std::string &name)> OnNameSetSignal;

        OnNameSetSignal m_on_name_set;

        typedef boost::signal<void (SceneObject *object, uint32_t layer)> OnLayerSetSignal;

        OnLayerSetSignal m_on_layer_set;

        typedef boost::signal<void (SceneObject *object, const std::string &meta_name)> OnMetaNameSetSignal;

        OnMetaNameSetSignal m_on_meta_name_set;
        
        typedef boost::signal<void (SceneObject *object, const Opacity &opacity)> OnOpacitySetSignal;

        OnOpacitySetSignal m_on_opacity_set;

        typedef boost::signal<void (SceneObject *object, const RGBColor &color)> OnColorSetSignal;

        OnColorSetSignal m_on_color_set;

        typedef boost::signal<void (SceneObject *object, bool hidden)> OnHiddenSetSignal;

        OnHiddenSetSignal m_on_hidden_set;

        typedef boost::signal<void (SceneObject *object)> OnSetObjectIDSignal;

        OnSetObjectIDSignal m_on_set_object_id;

        typedef boost::signal<void (SceneObject *object, ObjectID old)> OnObjectIDSetSignal;

        OnObjectIDSetSignal m_on_object_id_set;

        typedef boost::signal<void (SceneObject *object)> OnAccumOpacityChangedSignal;

        OnAccumOpacityChangedSignal m_on_accum_opacity_changed;

        typedef boost::signal<void (SceneObject *object)> OnMTOWChangedSignal;

        OnMTOWChangedSignal m_on_mtow_changed;

        typedef boost::signal<void (SceneObject *object)> OnDeleteSignal;
        
        OnDeleteSignal m_on_delete;

        /// @}



    private:

        bool insert(SceneObject* object);

        void validate_mtow_cache() throw();


    protected:
        /**
         * \brief Returns \c true if a valid model space to world space matrix
         *        is cached.
         *
         * This meaning of this function is equivalent to that of the 
         * private member \c m_mtow_is_valid.
         *
         * \pre None.
         * \post If the method returns \c true, an imminent call to \c mtow()
         *       is guaranteed to be very fast. This guarantee expires
         *       as soon as the object's, or an ancestor's, 
         *       model-to-parent transformation is modified.
         */
        bool mtow_is_valid() const throw();

        void orthonormalize_transform() throw();

        /**
         * \brief Overridable method responsible for invalidating the 
         *        model-to-world matrix cache.
         *
         * An implementation of this method is responsible to invalidate
         * the cached matrix for the object and call \c invalidate_mtow_cahce()
         * on all the object's children. This \e must be done to insure
         * correct behaviour.
         * That does not, however, rule out the possibility to recalculate
         * the cached model-to-world matrix immediately when it is 
         * invalidated, marking it valid again after its been updated.
         *
         * <em>Default implementation:</em> The default implementation 
         * simply sets \c m_mtow_is_valid to \c false and calls 
         * \c invalidate_mtow_cache() on all the object's children.
         *
         * Called by:
         * - \c apply_transformation(const Mat4 &)
         * - \c set_transformation(const Mat4 &)
         *
         * \pre None.
         * \post For each object in object's subtree, including itself, 
         *       \c mtow returns false if the object's \c m_mtow
         *       is invalid.
         * \sa \c apply_transformation(const Mat4 &), \c set_transformation(const Mat4 &)
         */
        void invalidate_mtow_cache();



    protected:

        /// \name XML handler methods
        /// @{

        /** \brief Element end handler for XML parsing.
         */        
        void end_handler(const std::string &name,
                         ScopedHandler *handler) 
            throw(std::runtime_error);

        /** \brief Name start handler for XML parsing.
         */
        void name_start_handler(
            const std::string & name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        /** \brief Name cdata handler for XML parsing.
         */
        void name_cdata_handler(const std::string &cdata,
                                ScopedHandler *handler)
            throw(std::runtime_error);

        /** \brief Meta name start handler for XML parsing.
         */
        void meta_name_start_handler(
            const std::string & name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        /** \brief Meta name cdata handler for XML parsing.
         */
        void meta_name_cdata_handler(const std::string &cdata,
                                     ScopedHandler *handler) throw();

        /** \brief Opacity start handler for XML parsing
         */
        void opacity_start_handler(
            const std::string & name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        /** \brief Opacity cdata handler for XML parsing
         */
        void opacity_cdata_handler(bool absolute,
                                   const std::string &cdata,
                                   ScopedHandler *handler)
            throw();

        /** \brief Hidden start handler for XML parsing
         */
        void hidden_start_handler(
            const std::string & name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        /** \brief XML handler for the transform tag.
         */
        void transform_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        /** \brief XML handler for the transform end tag.
         */        
        void transform_end_handler(const std::string &name,
                         ScopedHandler *handler) throw();

        /** \brief XML handler for the transformation/rotate tag.
         */
        void rotate_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        /** \brief rotate cdata handler for XML parsing
         */
        void rotate_cdata_handler(
            uint32_t tag,
            transformation_target::CoordinateSystem system,
            const std::string &cdata,
            ScopedHandler *handler) throw();
        
        /** \brief XML handler for the end tag of rotate
         */        
        void rotate_end_handler(const std::string &name,
                         ScopedHandler *handler) throw();


        /** \brief XML handler for the transformation/translate tag.
         */
        void translate_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        /** \brief translate cdata handler for XML parsing
         */
        void translate_cdata_handler(
            transformation_target::CoordinateSystem system,
            const std::string &cdata,
            ScopedHandler *handler) throw();

        /** \brief XML handler for the transformation/matrix tag.
         */
        void matrix_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        /** \brief matrix cdata handler for XML parsing
         */
        void matrix_cdata_handler(
            transformation_target::CoordinateSystem system,
            const std::string &cdata,
            ScopedHandler *handler) throw();

        /** \brief XML handler for the <children> tag.
         */
        void children_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler) throw();

        /// @}

        SceneObject *_get_descendent(
            const std::vector<std::string>::const_iterator first,
            const std::vector<std::string>::const_iterator last) throw();

        const SceneObject *_get_descendent(
            const std::vector<std::string>::const_iterator first,
            const std::vector<std::string>::const_iterator last) const throw();


        //
        // ----------------------- Members ---------------------------
        //

    private:
        /** \brief \c true if this object is hidden, \c false if it is visible.
         */
        bool m_is_hidden;

        /** \brief \c A pointer to the parent of this object in the scene tree.
         */
        SceneObject *m_parent;

        /** \brief \c The visibility layer that this object resides in.
         */
        uint32_t m_layer;

        /** \brief \c The opacity value of this object.
         */
        Opacity m_opacity;

        /** \brief The name of the object.
         */
        std::string m_name;
            
        /** \brief The meta name of the object.
         */
        std::string m_meta_name;

        /** \brief The unique ObjectID number of the object.
         */
        ObjectID m_id;

        int m_transforms_since_normalization;

        /** \brief The RGB color value of the object.
         */
        RGBColor m_color;

        static IDAllocator<ObjectID> *ms_id_allocator;

        /** \brief A list of pointers to this object's children.
         */
        ChildContainer m_children;

        /**
         * \brief Model space to parent space transformation matrix.
         * \todo Add reference to page discussing the different coordinate 
         *       systems referred to through the documentation.
         */
        Matrix4ru m_mtop;

        /**
         * \brief Model space to world space transformation matrix.
         *
         * \c m_mtow is guaranteed to be valid only when \c m_mtow_is_valid
         * is true.
         * \c m_mtow is typically invalidated when the object, or an 
         * ancestor to it, is transformed.
         *
         * \todo Add reference to page discussing the different coordinate 
         *       systems referred to through the documentation.
         */
        Matrix4ru m_mtow;

        /**
         * \brief Signifies whether or not \c m_mtow is valid and can be 
         *        read.
         * \sa \c m_mtow
         */
        bool m_mtow_is_valid;
    };


}

#endif // PEEKABOT_SCENE_OBJECT_HH_INCLUDED
