/*
 * Copyright Staffan Gimåker 2006-2009.
 * Copyright Anders Boberg 2006.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>

#include "SceneTree.hh"
#include "SceneObject.hh"
#include "CameraObject.hh"
#include "GroupObject.hh"
#include "GeometricPrimitives.hh"
#include "IDAllocator.hh"
#include "renderer/Renderer.hh"

using namespace peekabot;
using peekabot::renderer::TheRenderer;



// --------------------- TranslationTable implementation -----------------------

SceneTree::TranslationTable::TranslationTable(SceneObject *root) throw()
{
    on_child_attached(root);
}


void SceneTree::TranslationTable::on_child_attached(SceneObject *object) throw()
{
    add_rule(object);

    object->m_on_child_attached.connect(
        boost::bind(&SceneTree::TranslationTable::on_child_attached, this, _1));

    object->m_on_detached.connect(
        boost::bind(&SceneTree::TranslationTable::on_detached, this, _1));

    object->m_on_set_object_id.connect(
        boost::bind(&SceneTree::TranslationTable::remove_rule, this, _1));

    object->m_on_object_id_set.connect(
        boost::bind(&SceneTree::TranslationTable::add_rule, this, _1));

    SceneObject::ChildContainer &children = object->get_children();
    for( SceneObject::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        on_child_attached(*it);
    }
}


void SceneTree::TranslationTable::on_detached(SceneObject *object) throw()
{
    remove_rule(object);

    object->m_on_child_attached.disconnect(
        boost::bind(&SceneTree::TranslationTable::on_child_attached, this, _1));

    object->m_on_detached.disconnect(
        boost::bind(&SceneTree::TranslationTable::on_detached, this, _1));

    object->m_on_set_object_id.disconnect(
        boost::bind(&SceneTree::TranslationTable::remove_rule, this, _1));

    object->m_on_object_id_set.disconnect(
        boost::bind(&SceneTree::TranslationTable::add_rule, this, _1));

    SceneObject::ChildContainer &children = object->get_children();
    for( SceneObject::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        on_detached(*it);
    }
}


bool SceneTree::TranslationTable::add_rule(SceneObject *object) 
    throw()
{
    TranslationMap::iterator it = m_id_map.find(object->get_object_id());

    if( it != m_id_map.end() )
        return true;

    m_id_map.insert(std::make_pair(object->get_object_id(), object));

    return false;
}


bool SceneTree::TranslationTable::remove_rule(SceneObject *object) 
    throw()
{
    TranslationMap::iterator it = m_id_map.find(object->get_object_id());

    if( it == m_id_map.end() )
        return true;

    m_id_map.erase(it);

    return false;
}


SceneObject *SceneTree::TranslationTable::translate(
    ObjectID id) const throw()
{
    TranslationMap::const_iterator it = m_id_map.find(id);

    if( it != m_id_map.end() )
        return it->second;
    else
        return 0;
}



// ------------------------- SceneTree implementation -----------------------



SceneTree::SceneTree()
    : m_root(new GroupObject()),
      m_translation_table(m_root)
{
    SceneTreeMutex::scoped_lock lock(m_scene_mutex);

    // TODO: fix this! ?
    
    m_root->set_object_id(ROOT_OBJECT_ID);
    m_root->set_name("root");
    m_root->set_meta_name("Root");
    //add_rule(ROOT_OBJECT_ID, m_root);

    CameraObject *default_cam = new CameraObject();
    default_cam->set_object_id(DEFAULT_CAMERA_OBJECT_ID);
    default_cam->set_name("default_camera");
    default_cam->set_meta_name("Default camera");
    m_root->attach(default_cam);
    //add_rule(DEFAULT_CAMERA_OBJECT_ID, default_cam);
}

SceneTree::~SceneTree() throw()
{
    SceneTreeMutex::scoped_lock lock(m_scene_mutex);

    if( m_root )
        delete m_root;
}


SceneObject *SceneTree::get_root() throw()
{
    return m_root;
}

SceneObject *SceneTree::get_object(ObjectID id) throw()
{
    return m_translation_table.translate(id);
}


SceneObject *SceneTree::get_object(
    const std::string &path, SceneObject *path_parent) throw()
{
    std::vector<std::string> components;
    boost::split( components, path, boost::is_any_of(".") );

    if( components.empty() )
        return 0;

    SceneObject *obj = 0;
    size_t idx = 0;

    if( path_parent != 0 )
    {
        obj = path_parent;
    }
    else
    {
        obj = m_root;
        if( components[0] != "root" )
            return 0;
        ++idx;
    }

    try
    {
        for( ; idx < components.size(); ++idx )
        {
            assert( obj != 0 );
            obj = obj->get_child(components[idx]);
        }
    }
    catch(...)
    {
        return 0;
    }

    return obj;
}
