/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_SENSOR_HH
#define __PEEKABOT_SENSOR_HH


#include <string>
#include <stdexcept>
#include <boost/any.hpp>

#include "serialization/DeserializationInterface.hh"
#include "ScopedHandler.hh"
#include "MessageHub.hh"


namespace peekabot
{
    class SensorObject;


    /**
     * \internal
     *
     * \brief Represents a sensor data interpreter for a specific type of 
     * sensor.
     *
     * Derivatives of this class are tasked with interpreting the data 
     * coming from clients and transforming that into its proper graphical 
     * representation.
     *
     * Each SensorObject aggregates a Sensor object. It handles the
     * intepretation and transformation of the SensorObjects received data.
     *
     * <em>Important:</em> 
     * Each Sensor-subclass must be registered with the SensorFactory.
     *
     * \sa SensorObject, SensorFactory
     */
    class Sensor
    {
    public:
        /**
         * \brief Stores sensor configuration parameters.
         *
         * E.g. uncertainty values, number of samples, etc.
         */
        typedef std::map<std::string, boost::any> Params;


        virtual ~Sensor() {}

        /**
         * \brief This method should be implemented to transform an incoming
         * stream of bytes (the sensor data) to its proper graphical 
         * representation in peekabot.
         *
         * \remarks This shouldn't be declared no-throw in derived classes - 
         * reading from DeserializationBuffer can always cause an exception to
         * be thrown.
         */
        virtual void interpret_sensor_data(
            SensorObject *sensor,
            DeserializationInterface &sensor_data) = 0;

        /**
         * \brief If the sensor should have parameter templates, override this 
         * to provide them.
         *
         * This method is called (or should be called, depending on 
         * who's reading) immediately after the sensor has been 
         * constructed, before any parameters are parsed. Use it to give 
         * parameters default values based on a template.
         *
         * Note that this method is only called *if* a template is specified, 
         * and only if that template name is not an empty string.
         *
         * Template-invariant default values shouldn't be set here. Defaults 
         * not associated with a template should be handled by using the version
         * of get_param() that takes a second default parameter argument.
         */
        virtual void set_param_defaults(
            const std::string &template_name) throw() {}

        /**
         * \brief This method is called after parsing and setting the sensor 
         * parameters defined in the definition file. 
         *
         * It's intended for sensor-specific error checking - types, required
         * parameters, range checking, etc.
         */
        virtual void validate_params() {}

        void params_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler);

        /**
         * \brief Return the map of parmeters 
         */
        const Params &get_params() const throw()
        {
            return m_params;
        }

        /**
         * \brief Returns true if the given parameter is defined and has the
         * given type. Returns false otherwise.
         */
        template<class T>
        bool has_param(const std::string &name) const throw()
        {
            Params::const_iterator it = m_params.find(name);

            if( it == m_params.end() )
                return false;

            try
            {
                boost::any_cast<const T &>(it->second);
                return true;
            }
            catch(boost::bad_any_cast)
            {
                return false;
            }
        }

        /**
         * \brief Returns true if the given parameter is defined, but doesn't 
         * check its type.
         */
        bool has_param(const std::string &name) const throw()
        {
            return m_params.find(name) != m_params.end();
        }

        /**
         * \brief Get the parameter with the given name.
         *
         * \exception std::runtime_error Cast if the stored parameter type is 
         * different from the requested type, or the parameter is undefined.
         */
        template<class T>
        const T &get_param(const std::string &name) 
            const throw(std::runtime_error)
        {
            Params::const_iterator it = m_params.find(name);

            if( it == m_params.end() )
                throw std::runtime_error("No such parameter");

            try
            {
                return boost::any_cast<const T &>(it->second);
            }
            catch(boost::bad_any_cast)
            {
                throw std::runtime_error(
                    "The parameters type is different from the requested type");
            }
        }

        /**
         * \brief Get the parameter with the given name. If undefined, return 
         * the given default value.
         *
         * \exception std::runtime_error Cast if the stored parameter type is 
         * different from the requested type.
         */
        template<class T>
        T get_param(const std::string &name, const T &default_value) 
            const throw(std::runtime_error)
        {
            if( has_param(name) )
                return get_param<T>(name);
            else
                return default_value;
        }

    protected:
        /**
         * \brief Check the existance and type of a required parameter.
         *
         * Will throw, with an indication the cause of failure, if the
         * parameter is undefined or of another.
         */
        template<class T>
        void check_required_param(const std::string &name) 
            throw(std::runtime_error)
        {
            if( !has_param(name) )
                throw std::runtime_error(
                    "Required parameter '" + name + "' not present");

            if( !has_param<T>(name) )
                throw std::runtime_error(
                    "Illegal type for parameter '" + name + "'");
        }

        /**
         * \brief Check the type of a parameter, if it's defined.
         *
         * Will throw, with an indication the cause of failure, if the
         * parameter is of another type than specified by the template
         * argument.
         *
         * If the parameter isn't defined, the method is a no-op.
         */
        template<class T>
        void check_param_type(const std::string &name) 
            throw(std::runtime_error)
        {
            if( has_param(name) && !has_param<T>(name) )
                throw std::runtime_error(
                    "Illegal type for parameter '" + name + "'");
        }

        /**
         * \brief Sets a sensor parameter.
         *
         * If a parameter of the same name already exists it will be 
         * overwritten by the new value (and type). A warning will, however,
         * be generated in that case.
         */
        template<class T>
        void set_param(const std::string &name, const T &value)
        {
            T copy = value;
            m_params[name] = copy;
        }

    private:
        void bool_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler);

        void bool_cdata_handler(
            const std::string &cdata, ScopedHandler *handler,
            const std::string &param_name);


        void int_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler);

        void int_cdata_handler(
            const std::string &cdata, ScopedHandler *handler,
            const std::string &param_name);


        void float_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler);

        void float_cdata_handler(
            const std::string &cdata, ScopedHandler *handler,
            const std::string &param_name);

            
        void string_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler);

        void string_cdata_handler(
            const std::string &cdata, ScopedHandler *handler,
            const std::string &param_name);


        void color_start_handler(
            const std::string &name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler);


    private:
        Params m_params;
    };

}


#endif // __PEEKABOT_SENSOR_HH
