/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/bind.hpp>

#include "SensorObject.hh"
#include "ObjectVisitor.hh"
#include "ScopedHandler.hh"
#include "SensorFactory.hh"


using namespace peekabot;


HandlerInformer SensorObject::ms_handler_informer(
    "sensor", &SensorObject::start_handler);


SensorObject::SensorObject(const std::string &sensor_type)
    : SceneObject("sensor"),
      m_sensor_type(sensor_type),
      m_data_interpreter(
          boost::shared_ptr<Sensor>(
              TheSensorFactory::instance().instantiate_interpreter(
                  sensor_type)))
{
}

SensorObject::SensorObject(
    ScopedHandler *handler, 
    const std::string &sensor_type, 
    const std::string &params_template)
    : SceneObject("sensor", handler),
      m_sensor_type(sensor_type),
      m_data_interpreter(
          boost::shared_ptr<Sensor>(
              TheSensorFactory::instance().instantiate_interpreter(
                  sensor_type)))
{
    handler->add_start_handler(
        "params",
        boost::bind(&Sensor::params_start_handler, 
                    m_data_interpreter.get(), _1, _2, _3));

    // Set up default parameter values based on the template
    if( params_template != "" )
        m_data_interpreter->set_param_defaults(params_template);
}

SensorObject::~SensorObject() throw()
{
}

void SensorObject::accept(ObjectVisitor* visitor) throw()
{
    visitor->visit(this);
}

const std::string &SensorObject::get_sensor_type() const throw()
{
    return m_sensor_type;
}

void SensorObject::set_sensor_data(DeserializationInterface &sensor_data)
{
    m_data_interpreter->interpret_sensor_data(this, sensor_data);
}


//
// ----- XML handler methods ------
//


void SensorObject::start_handler(
    const std::string &name,
    XMLHandler::AttributeMap &attributes,
    ScopedHandler *handler)
{
    TheMessageHub::instance().publish(
        WARNING_MESSAGE,
        "<sensor> is deprecated and will be removed in peekabot 0.7, "
        "see the release notes for more information.");

    XMLHandler::AttributeMap::const_iterator template_it = 
        attributes.find("template");

    std::string params_template = "";
    if( template_it != attributes.end() )
        params_template = template_it->second;

    // Creating a scene object by passing a ScopedHandler will cause it to
    // enter a new scope with all registered tag start handlers plus any
    // specific handlers for SceneObject properties which are registered
    // by the parent constructor.
    SensorObject *tmp = new SensorObject(
        handler, attributes["type"], params_template);

    // Set the new object as the current object
    ScopedMap &variables = handler->get_variables();
    variables.push_variable("current_object", (SceneObject *)tmp);

    handler->get_current_scope().end_functors.push_back(
        boost::bind(&SensorObject::end_handler, tmp, _1, _2));
}


void SensorObject::end_handler(
    const std::string &name, ScopedHandler* handler)
{
    TheMessageHub::instance().publish(
        DEBUG_MESSAGE,
        "Running post-parsing sensor parameter validity checks");

    m_data_interpreter->validate_params();

    TheMessageHub::instance().publish(
        DEBUG_MESSAGE,
        "Sensor parameter validity checks passed");
}
