/*
 * Copyright Staffan Gimåker 2006-2007, 2009.
 * Copyright Anders Boberg 2006-2007.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <iostream>
#include <exception>
#include <algorithm>
#include <boost/function.hpp>
#include <boost/bind.hpp>
#include "ServerActuator.hh"
#include "IDAllocator.hh"
#include "PseudonymCatalog.hh"
#include "SceneTree.hh"
#include "SceneTreePtr.hh"
#include "ServerExecutionContext.hh"
#include "atl/ActionRouter.hh"
#include "MessageHub.hh"

using namespace peekabot;


//
// How many actions that are processed by each call
// to process_actions().
// Affects the action throughput/responsivity ratio (high batch size increases
// throughput).
//
static const size_t DEFAULT_ACTION_BATCH_SIZE = 20;


ServerActuator::ServerActuator() throw()
    : m_thread(0)
{
    m_action_batch_size = TheConfiguration::instance().get<int>(
        "general.action_batch_size", DEFAULT_ACTION_BATCH_SIZE);
}


ServerActuator::~ServerActuator() throw()
{
    delete m_pseudonyms;
}


void ServerActuator::run() throw()
{
    while( !m_stop_signal )
    {
        {
            boost::recursive_mutex::scoped_lock lock(m_mutex);

            // Wait for new data (or that we're being shut down)
            if( m_queue.empty() )
                m_queue_cond.wait(lock);
        }
        
        bool do_redraw;
        process_actions(do_redraw);

    }
}


bool ServerActuator::process_actions(bool &do_redraw) throw()
{
    bool processed_gui_actions = false;

    // Get ready for some action(s)!!
    size_t n_actions;
    {
        boost::recursive_mutex::scoped_lock lock(m_mutex);

        n_actions = std::min((int)m_queue.size(), m_action_batch_size);

        do_redraw = m_queue.size() - n_actions == 0;
    }
        

    if( n_actions > 0 )
    {
        ServerExecutionContext context(*m_pseudonyms);
        
        for( size_t i = 0; i < n_actions; ++i )
        {
            Action *action = 0;
            SourceID source;

            {
                boost::recursive_mutex::scoped_lock lock(m_mutex);

                action = m_queue.front().first;
                source = m_queue.front().second;
                m_queue.pop();

                if( !action )
                    continue;
            }

            context.set_accessor(source);

            try
            {
                action->execute(&context);

                if( source == GUI_SOURCE_ID )
                    processed_gui_actions = true;
            }
            catch(std::exception &e)
            {
                TheMessageHub::instance().publish(
                    DEBUG_MESSAGE, 
                    source,
                    "Action failed to execute, actuator caught exception: " + 
                    std::string(e.what()) );

                TheMessageHub::instance().publish(
                    ERROR_MESSAGE, 
                    source,
                    std::string(e.what()) );
            }
            catch(...)
            {
                TheMessageHub::instance().publish(
                    ERROR_MESSAGE, 
                    source,
                    "Actuator caught unexpected exception of "
                    "non-std::exception type");
            }

            delete action;
        }
    }
        
    return processed_gui_actions;
}


bool ServerActuator::start() throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    if( is_running() )
        return true;

    m_stop_signal = false;

    SceneTreePtr scene_ptr = TheSceneTree::acquire();

    m_pseudonyms = new PseudonymCatalog(scene_ptr->get_root());

    TheActionRouter::instance().set_action_handler(
        SERVER_SOURCE_ID,
        boost::bind(&ServerActuator::handle_inbound_action, this, _1, _2) );

    boost::function0<void> f = boost::bind(&ServerActuator::run, this);
    m_thread = new boost::thread(f);

    return false;
}

bool ServerActuator::stop() throw()
{
    if( !is_running() )
        return true;

    // Signal the actuator to halt
    m_stop_signal = true;

    m_queue_cond.notify_all();

    // Wait for it to finish what it's doing...
    m_thread->join();

    {
        boost::recursive_mutex::scoped_lock lock(m_mutex);

        // Clean up...
        delete m_thread;
        m_thread = 0;
    }

    return false;
}

bool ServerActuator::is_running() throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    return m_thread != 0;
}


void ServerActuator::handle_inbound_action(
    Action *action, SourceID source) throw()
{
    {
        boost::recursive_mutex::scoped_lock lock(m_mutex);

        m_queue.push(std::make_pair(action, source));
    }

    m_queue_cond.notify_all();
}
