/*
 * Copyright Staffan Gimåker 2006-2009.
 * Copyright Anders Boberg 2007.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <stdexcept>

#include "ServerExecutionContext.hh"
#include "SceneTree.hh"
#include "SceneTreePtr.hh"
#include "PseudonymCatalog.hh"
#include "atl/ActionRouter.hh"
#include "SceneFileLoader.hh"

using namespace peekabot;


ServerExecutionContext::ServerExecutionContext(PseudonymCatalog &pseudonyms) throw()
    : m_pseudonyms(pseudonyms),
      m_scene_ptr(TheSceneTree::acquire())
{
}


void ServerExecutionContext::set_accessor(const SourceID &accessor) throw()
{
    m_accessor = accessor;
}


SceneObject *ServerExecutionContext::get_root() throw()
{
    return m_scene_ptr->get_root();
}



SceneObject *ServerExecutionContext::get_object(ObjectID id) throw()
{
    ObjectID real_id;

    // If the accessor is anyone but the server, they talk in pseudonym IDs
    // that first needs to be translated.
    if( m_accessor >= 1000 )
    {
        // Find real ID
        real_id = m_pseudonyms.translate_pseudonym(m_accessor, id);
    }
    else
        real_id = id;

    // Since we own a lock on the scene tree it's safe to return an ObjectPtr,
    // which will be valid for the duration of the action's execute method
    // provided it itself does not delete it.
    return m_scene_ptr->get_object(real_id);
}



SceneObject *ServerExecutionContext::get_object(
    const std::string &path,
    SceneObject *path_parent) throw()
{
    return m_scene_ptr->get_object(path, path_parent);
}



void ServerExecutionContext::register_pseudonym(
    ObjectID real_id, ObjectID pseudonym_id) throw(std::exception)
{
    // Find object...
    SceneObject *obj = m_scene_ptr->get_object(real_id);
    if( !obj )
        throw std::runtime_error("Failed to register pseudonym, object not found");

    m_pseudonyms.add_pseudonym(m_accessor, pseudonym_id, obj->get_object_id());
}


void ServerExecutionContext::deregister_pseudonym(
    ObjectID pseudonym_id) throw(std::exception)
{
    m_pseudonyms.remove_pseudonym(m_accessor, pseudonym_id);
}


void ServerExecutionContext::post_action(
    Action *action, const SourceID &dest) throw()
{
    TheActionRouter::instance().post_action(
        action, SERVER_SOURCE_ID, dest);
}


void ServerExecutionContext::post_response(Action *action) throw()
{
    TheActionRouter::instance().post_action(
        action, SERVER_SOURCE_ID, m_accessor);
}


std::vector<SceneObject *> ServerExecutionContext::load_scene(
    const std::string &filename)
{
    SceneFileLoader loader;
    return loader.load_scene(filename);
}

