/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_SERVER_EXECUTION_CONTEXT_HH
#define __PEEKABOT_SERVER_EXECUTION_CONTEXT_HH


#include <string>
#include <list>
#include <vector>
#include <exception>
#include "Types.hh"
#include "SceneObject.hh"
#include "SceneTreePtr.hh"

// TODO: Removing pseudonym (double hashing and triggers?)


namespace peekabot
{

    class Action;
    class PseudonymCatalog;
    class ServerActuator;
    class SensorObject;
    class RobotObject;

    class ServerExecutionContext
    {
        friend class ServerActuator;

        SourceID m_accessor;

        PseudonymCatalog &m_pseudonyms;

        // Provides automatic locking of the scene tree the for the context's lifetime
        SceneTreePtr m_scene_ptr;


    private:

        void set_accessor(const SourceID &accessor) throw();

        ServerExecutionContext(PseudonymCatalog &pseudonyms) throw();

        virtual ~ServerExecutionContext() {}


    public:
        class UniquenessViolation : public std::exception
        {
            std::string m_name;

        public:            
            UniquenessViolation(const std::string &name) throw() 
                : m_name(name) {}

            virtual ~UniquenessViolation() throw() {}

            virtual const char *what() const throw()
            {
                return "Failed to add object, due to the name uniquness constraint.";
            }

            const std::string &get_name() const throw()
            {
                return m_name;
            }
        };




        SceneObject *get_root() throw();

        SceneObject *get_object(ObjectID id) throw();

        template<class T>
        T *get_object(ObjectID id) throw()
        {
            return dynamic_cast<T *>(get_object(id));
        }

        SceneObject *get_object(
            const std::string &path, 
            SceneObject *path_parent = 0) throw();

        void register_pseudonym(ObjectID real_id, ObjectID pseudonym_id) throw(std::exception);

        void deregister_pseudonym(ObjectID pseudonym_id) throw(std::exception);

        void post_action(Action *action, const SourceID &dest) throw();

        void post_response(Action *action) throw();


        std::vector<SceneObject *> load_scene(const std::string &filename);
        
        /** \brief Tries to apply the ALIAS_ON_CONFLICT policy.
         *
         * This method will check if parameter \c parent has a child
         * with the name \c name, which is also of type \c T. If so,
         * it considers the ALIAS_ON_CONFLICT policy applicable
         * and registers the pseudonym as an alias for the child if
         * the \c register_pseudo argument is \c true.
         *
         * \return \c true if the ALIAS_ON_CONFLICT policy was applied
         * \c false if not.
         */
        template<class T> 
        bool aliasing_policy_applied(SceneObject* parent,
                                     const std::string &name,
                                     ObjectID pseudonym_id,
                                     bool register_pseudo)
        {
            if(parent->has_child(name))
            {
                SceneObject *object = parent->get_child(name);
                if(dynamic_cast<T*>(object) != 0)
                {
                    if(register_pseudo)
                        register_pseudonym(object->get_object_id(),
                                           pseudonym_id);
                    return true;
                }
            }
            
            return false;
        }

    };

}


#endif // __PEEKABOT_SERVER_EXECUTION_CONTEXT_HH
