/*
 * Copyright Staffan Gimåker 2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sstream>
#include <fstream>
#include <limits>
#include <ctime>
#include <boost/filesystem/operations.hpp>
#include <boost/random.hpp>
#include <boost/cstdint.hpp>
#include <boost/algorithm/string/replace.hpp>

#include "UploadCache.hh"
#include "FsToolbox.hh"


using namespace peekabot;


UploadCache::UploadCache()
{
    m_rng.seed(std::time(0));

    if( boost::filesystem::exists(upload_directory()) )
        boost::filesystem::remove_all(upload_directory());

    boost::filesystem::create_directory(upload_directory());
}


UploadCache::~UploadCache()
{
    clear();
    boost::filesystem::remove_all(upload_directory());
}


void UploadCache::add(const std::string &key, const void *data, std::size_t n)
{
    if( !has_key(key) )
        m_files[key] = generate_filename(key);

    boost::filesystem::path path = upload_directory() / m_files[key];
    std::ofstream ofs(path.string().c_str(), std::ios::binary);

    if( !ofs )
        throw std::runtime_error(
            "Failed to open file '" + path.string() + "' for writing");

    if( n > 0 )
        ofs.write((const char *)data, n);

    ofs.close();
}


void UploadCache::remove(const std::string &key)
{
    if( !has_key(key) )
        throw std::runtime_error("No such file");

    boost::filesystem::path p = upload_directory() / m_files[key];
    if( boost::filesystem::exists(p) )
        boost::filesystem::remove(p);

    m_files.erase(key);
}


bool UploadCache::has_key(const std::string &key) const
{
    KeyFilenameMap::const_iterator it = m_files.find(key);
    return it != m_files.end();
}


void UploadCache::clear()
{
    while( !m_files.empty() )
    {
        std::string k = m_files.begin()->first;
        remove(k);
    }
}


boost::filesystem::path UploadCache::expand_path(
    const std::string &key) const
{
    KeyFilenameMap::const_iterator it = m_files.find(key);
    if( it == m_files.end() )
        return boost::filesystem::path(key);
    else
        return upload_directory() / it->second;
}


boost::filesystem::path UploadCache::upload_directory()
{
    return boost::filesystem::path(
        fs::get_temp_path() / "peekabot" / "uploads");
}


std::string UploadCache::generate_filename(const std::string &key) const
{
    boost::uniform_int<boost::uint32_t> dist(
        0, std::numeric_limits<boost::uint32_t>::max());
    boost::variate_generator<boost::mt19937 &,
        boost::uniform_int<boost::uint32_t> > die(m_rng, dist);

    std::string key_ = key;
    boost::replace_all(key_, "/", "-"); // for "normal" paths
    boost::replace_all(key_, "\\", "-"); // for Windows paths

    while( true )
    {
        std::stringstream ss;
        ss << die() << key_;
        if( !boost::filesystem::exists(upload_directory() / ss.str()) )
            return ss.str();
    }
}
