/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/bind.hpp>
#include <stdexcept>
#include <sstream>

#include "VertexObject.hh"
#include "ScopedHandler.hh"
#include "ScopedMap.hh"


using namespace peekabot;



VertexObject::VertexObject(
    size_t min_vertices, size_t max_vertices, size_t multiple_of) throw()
    : m_min_vertices(min_vertices),
      m_max_vertices(max_vertices),
      m_multiple_of(multiple_of)
{
    //
    // Add a few dummy vertices to satisfy the vertex count constraints
    //
    for( size_t i = 0; i < min_vertices; ++i )
        m_vertices.push_back(Vector3ru(0,0,0));
}


VertexObject::VertexObject(
    size_t min_vertices, size_t max_vertices, 
    ScopedHandler *handler, size_t multiple_of)
    : m_min_vertices(min_vertices),
      m_max_vertices(max_vertices),
      m_multiple_of(multiple_of)
{
    //
    // Add a few dummy vertices to satisfy the vertex count constraints
    //
    for( size_t i = 0; i < min_vertices; ++i )
        m_vertices.push_back(Vector3ru(0,0,0));


    boost::function<void (const std::string&, ScopedHandler*)> fn =
        boost::bind(&VertexObject::vertices_end_handler, this, _1, _2);

    // Register the vertices tag start handler
    handler->add_start_handler(
        "vertices",
        boost::bind(&VertexObject::vertices_start_handler, 
                    this, fn, _1, _2, _3));
}


void VertexObject::set_vertices(const Vector3ru *vertices, size_t n) throw(std::exception)
{
    if( n < m_min_vertices || n > m_max_vertices || n % m_multiple_of != 0 )
        throw std::length_error("Invalid number of vertices");
    Vertices v;
    for( size_t i = 0; i < n; ++i )
        v.push_back(vertices[i]);

    set_vertices(v);
}


void VertexObject::set_vertices(const Vertices &vertices) throw(std::exception)
{
    if( vertices.size() < m_min_vertices || 
        vertices.size() > m_max_vertices || 
        vertices.size() % m_multiple_of != 0 )
        throw std::length_error("Invalid number of vertices");

    m_vertices = vertices;
    m_on_vertices_set(this, m_vertices);
}


void VertexObject::add_vertices(const Vertices &vertices) throw(std::exception)
{
    size_t n = vertices.size() + get_vertex_count();

    if( n < m_min_vertices || 
        n > m_max_vertices || 
        n % m_multiple_of != 0 )
        throw std::length_error("Invalid number of vertices");

    std::copy(vertices.begin(), vertices.end(), std::back_inserter(m_vertices));
    // TODO: should use vertices added hook instead
    m_on_vertices_set(this, m_vertices);
}


const VertexObject::Vertices &VertexObject::get_vertices() const throw()
{
    return m_vertices;
}


size_t VertexObject::get_vertex_count() const throw()
{
    return m_vertices.size();
}


// ---------------------- XML handler functions ----------------------


void VertexObject::vertices_start_handler(
    boost::function<void (const std::string&, ScopedHandler*)> end_handler,
    const std::string tag_name,
    XMLHandler::AttributeMap & attributes,
    ScopedHandler* handler)
{
    ScopedHandler::TagScope vertices_scope;
    
    vertices_scope.start_functors["vertex"] = boost::bind(
        &VertexObject::vertex_start_handler, _1, _2, _3);

    vertices_scope.end_functors.push_back(end_handler);

    handler->enter_scope(vertices_scope);
}


void VertexObject::vertices_end_handler(const std::string &name,
                                        ScopedHandler *handler)
    throw(std::runtime_error, std::length_error)
{
    ScopedMap & variables = handler->get_variables();

    Vertices vertices;
    while( variables.exists<Vector3ru>("vertex") )
    {
        vertices.push_back(variables.get_variable<Vector3ru>("vertex"));
        variables.pop_variable<Vector3ru>("vertex");
    }

    set_vertices(vertices);
}


void VertexObject::vertex_cdata_handler(
    const std::string &cdata,
    ScopedHandler *handler)
{
    ScopedMap & variables = handler->get_variables();

    // Read coordinate information
    std::stringstream s(cdata);
    Vector3ru vec;
    s >> vec(0) >> vec(1) >> vec(2);

    variables.push_variable("vertex", vec);
}


void VertexObject::vertex_start_handler(
    const std::string & tag_name,
    XMLHandler::AttributeMap &attributes,
    ScopedHandler *handler)
{
    ScopedHandler::TagScope vertex_scope;
    
    vertex_scope.cdata_functor = boost::bind(
        &vertex_cdata_handler, _1, _2); 
    
    handler->enter_scope(vertex_scope);    
}

