/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_VERTEX_OBJECT_HH
#define __PEEKABOT_VERTEX_OBJECT_HH


#include <exception>
#include <vector>
#include <boost/thread_safe_signal.hpp>

#include "Vector.hh"
#include "Types.hh"
#include "XMLHandler.hh"


namespace peekabot
{

    class ScopedHandler;

    class VertexObject
    {
    public:
        typedef std::vector<Vector3ru> Vertices;

        VertexObject(size_t min_vertices, 
                     size_t max_vertices, 
                     size_t multiple_of = 1) throw();


        /**
         * \brief Construct a vertex object using an XML document handler.
         */
        VertexObject(size_t min_vertices, 
                     size_t max_vertices, 
                     ScopedHandler *handler, 
                     size_t multiple_of = 1);


        void set_vertices(const Vector3ru *vertices, size_t n) throw(std::exception);

        void set_vertices(const Vertices &vertices) throw(std::exception);

        void add_vertices(const Vertices &vertices) throw(std::exception);

        const Vertices &get_vertices() const throw();

        size_t get_vertex_count() const throw();


        /// \name Signals
        /// @{

        boost::signal<void (VertexObject *object, const VertexObject::Vertices &vertices)> m_on_vertices_set;

        /// @}


    private:
        /// \name XML handler methods
        /// @{

        void vertices_start_handler(
            boost::function<void (const std::string&, ScopedHandler*)> end_handler,
            const std::string tag_name,
            XMLHandler::AttributeMap & attributes,
            ScopedHandler* handler);

        void vertices_end_handler(
            const std::string &name, ScopedHandler *handler) 
            throw(std::runtime_error, std::length_error);


        static void vertex_cdata_handler(
            const std::string &cdata,
            ScopedHandler *handler);
        

        static void vertex_start_handler(
            const std::string & tag_name,
            XMLHandler::AttributeMap &attributes,
            ScopedHandler *handler);

        /// @}

    private:
        const size_t m_min_vertices;
        const size_t m_max_vertices;
        const size_t m_multiple_of;

        Vertices m_vertices;
    };


}

#endif // __PEEKABOT_VERTEX_OBJECT_HH
