/*
 * Copyright Anders Boberg 2006-2007.
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */

#include "MoveObjectRelative.hh"

#ifdef __PEEKABOT_SERVER
#  include "../ServerExecutionContext.hh"
#  include "../SceneObject.hh"
#  include "../CameraObject.hh"
#endif


using namespace peekabot;
using namespace peekabot::transformation_target;


MoveObjectRelative::MoveObjectRelative() throw()
{
}


MoveObjectRelative::MoveObjectRelative(
    ObjectID target, real_unit x, real_unit y, real_unit z,
    CoordinateSystem system, ObjectID camera) throw()
    : m_target(target), m_x(x), m_y(y), m_z(z),
      m_system(system), m_camera(camera)
    
{
}


Action *MoveObjectRelative::clone() const
{
    return new MoveObjectRelative(m_target, m_x, m_y, m_z, m_system);
}


void MoveObjectRelative::execute(
    ServerExecutionContext *context) 
    const throw(std::exception)
{
#ifdef __PEEKABOT_SERVER
    CoordinateSystem system = m_system;

    SceneObject *ptr = context->get_object(m_target);
    if( !ptr )
        throw std::runtime_error(
            "Target object does not exist.");

    Matrix4ru trans_matrix = ptr->get_transformation();

    // We currently support only world coordinates or local coordinates.
    if(system == PARENT_COORDINATES)
    {
        trans_matrix *= Matrix4ru::translate(m_x, m_y, m_z);
    }
    else if(system == WORLD_COORDINATES)
    {
        SceneObject *parent_ptr = ptr->get_parent();
        if(parent_ptr)
        {
            Matrix4ru parent_mtow = parent_ptr->get_mtow();
            Matrix4ru mtow = ptr->get_mtow();

            trans_matrix = (parent_mtow.inverse() * 
                            Matrix4ru::translate(m_x, m_y, m_z))* mtow;
        }
        // If the object has no parent, it's the root, so we might as well
        // transform it using world coordinates.
        else
            trans_matrix *= Matrix4ru::translate(m_x, m_y, m_z);
    }

    else if(system == LOCAL_COORDINATES)
    {
        trans_matrix(3) += (m_x * trans_matrix(0) +
                            m_y * trans_matrix(1) +
                            m_z * trans_matrix(2));
    }
    else if(system == CAMERA_COORDINATES)
    {
        // First, fetch the mtow matrix of the camera...
        CameraObject *camera_ptr = dynamic_cast<CameraObject *>(
            context->get_object(m_camera));
        if(!camera_ptr)
            throw new std::runtime_error(
                "The specified camera does not exist.");
        Matrix4ru camera_mtow = camera_ptr->get_mtow();
        camera_mtow(3) = Vector4ru(0, 0, 0, 0); // We only want axes

        // Then, in order to do a mtow->mtop translation, we need
        // the mtow of the target object's parent...
        Matrix4ru parent_mtow;
        SceneObject *parent_ptr = ptr->get_parent();
        if(parent_ptr)
            parent_mtow = parent_ptr->get_mtow();
        // If the object has no parent, it's the root, so we might as well
        // transform it using world coordinates.
        else
            parent_mtow = ptr->get_transformation();
        //parent_mtow(3) = Vector4ru(0, 0, 0, 0); // We only want axes

        // Next, we need the target object's mtow to apply the
        // transformation to.        
        Matrix4ru mtow = ptr->get_mtow();

        // Apply transformation
        mtow(3) += (m_x * -camera_mtow(1) +
                    m_y * camera_mtow(0) +
                    m_z * camera_mtow(2));

        trans_matrix = (parent_mtow.inverse() * mtow);
    }
    
   
    ptr->set_transformation(trans_matrix);
#endif
}


void MoveObjectRelative::save(SerializationInterface &ar) const
{
    ar << m_target 
       << m_x << m_y << m_z
       << m_system
       << m_camera;
}

void MoveObjectRelative::load(DeserializationInterface &ar)
{
    ar >> m_target 
       >> m_x >> m_y >> m_z
       >> m_system
       >> m_camera;
}
