/*
 * Copyright Anders Boberg 2006-2007.
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */

#include <stdexcept>

#include "RotateObject.hh"

#ifdef __PEEKABOT_SERVER
#  include "../ServerExecutionContext.hh"
#  include "../SceneObject.hh"
#  include "../CameraObject.hh"
#endif


using namespace peekabot;
using namespace peekabot::transformation_target;


RotateObject::RotateObject() throw()
{
}


RotateObject::RotateObject(
    ObjectID target, real_unit x, real_unit y, real_unit z,
    CoordinateSystem axis_system, const Vector4ru &center_point,
    CoordinateSystem pivot_system) throw()
    : m_target(target), m_x(x), m_y(y), m_z(z),
      m_axis_system(axis_system), m_center_point(center_point),
      m_pivot_system(pivot_system)
    
{
}


Action *RotateObject::clone() const
{
    return new RotateObject(m_target, m_x, m_y, m_z, m_axis_system,
                            m_center_point, m_pivot_system);
}


void RotateObject::execute(
    ServerExecutionContext *context) 
    const throw(std::exception)
{
#ifdef __PEEKABOT_SERVER
    SceneObject *ptr = context->get_object(m_target);
    if( !ptr )
        throw std::runtime_error("Object not found.");

    SceneObject *parent_ptr = ptr->get_parent();
    Matrix4ru parent_mtow;
    if(parent_ptr)
        parent_mtow = parent_ptr->get_mtow();
    else
        parent_mtow.load_identity();


    Matrix4ru trans_matrix = ptr->get_mtow();

    Matrix4ru rotational_matrix;
    
    if(m_axis_system == WORLD_COORDINATES)
    {
        rotational_matrix = Matrix4ru::identity();
    }
    else if(m_axis_system == LOCAL_COORDINATES)
        rotational_matrix = ptr->get_mtow();
    else // PARENT_COORDINATES
    {
        rotational_matrix = parent_mtow;
    }

    // Transform pivot coordinates into world
    Vector4ru tmp = m_center_point;
    if(m_pivot_system == LOCAL_COORDINATES)
        tmp = trans_matrix * m_center_point;
    else if(m_pivot_system == PARENT_COORDINATES)
        tmp = parent_mtow * m_center_point;

    // If WORLD_COORDINATES we don't need to do anything

    tmp(3) = 0;
    trans_matrix(3) -= tmp;
    trans_matrix *= Matrix4ru::rotate(rotational_matrix(0), m_x);
    trans_matrix *= Matrix4ru::rotate(rotational_matrix(1), m_y);
    trans_matrix *= Matrix4ru::rotate(rotational_matrix(2), m_z);
    trans_matrix(3) += tmp;

    trans_matrix *= parent_mtow.inverse();

    ptr->set_transformation(trans_matrix);
#endif
}


void RotateObject::save(SerializationInterface &ar) const
{
    ar << m_target
       << m_x << m_y << m_z
       << m_axis_system
       << m_center_point
       << m_pivot_system;
}

void RotateObject::load(DeserializationInterface &ar)
{
    ar >> m_target
       >> m_x >> m_y >> m_z
       >> m_axis_system
       >> m_center_point
       >> m_pivot_system;
}
