/*
 * Copyright Staffan Gimåker 2007-2009.
 *
 * ---
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */

#include <boost/math/fpclassify.hpp>

#include "SetOrientation.hh"

#ifdef __PEEKABOT_SERVER
#  include "../ServerExecutionContext.hh"
#  include "../SceneObject.hh"
#endif


using namespace peekabot;


SetOrientation::SetOrientation(
    ObjectID target,
    const Vector3ru &orientation,
    CoordinateSystem coord_sys) throw(std::logic_error)
    : m_target(target),
      m_orientation(orientation),
      m_coord_sys(coord_sys)
{
    if( m_orientation == Vector3ru(0,0,0) )
        throw std::logic_error("Orientation must be a non-zero vector");

    m_orientation.normalize();
}


SetOrientation::SetOrientation() throw()
{
}


SetOrientation::SetOrientation(const SetOrientation &action) throw()
    : m_target(action.m_target),
      m_orientation(action.m_orientation),
      m_coord_sys(action.m_coord_sys)
{
}


SetOrientation::~SetOrientation() throw()
{
}


Action *SetOrientation::clone() const
{
    return new SetOrientation(*this);
}


void SetOrientation::execute(
    ServerExecutionContext *context)
    const throw(std::exception)
{
#ifdef __PEEKABOT_SERVER
    SceneObject *ptr = context->get_object(m_target);

    if( !ptr )
        throw std::runtime_error(
            "Failed to set orientation, target object not found");

    // Check for infinity and NaN
    for( int i = 0; i < 3; i++ )
        if( !boost::math::isfinite(m_orientation(i)) )
            throw std::logic_error(
                "Failed to set orientation: orientation "
                "vector cannot contain infinity or NaN");

    Vector3ru orient;

    switch( m_coord_sys )
    {
        case LOCAL_COORDINATES:
            orient = ptr->get_transformation() * Vector4ru(m_orientation, 0);
            break;

        case PARENT_COORDINATES:
            orient = m_orientation;
            break;

        case WORLD_COORDINATES:
            orient = ptr->get_parent_mtow().inverse() *
                Vector4ru(m_orientation, 0);
            break;

        default:
            throw std::runtime_error(
                "Failed to set oreintation, unsupported coordinate system");
            break;
    }


    orient.normalize(); // Muy importante! The equality test will fail otherwise

    Matrix4ru m = ptr->get_transformation();
    Vector3ru old_orient = m(0);

    if( old_orient.equals(orient, 0.001) )
        return; // It's already correctly oriented

    Vector3ru axis = orient.cross_product(m(0));

    Vector4ru pos = m(3);
    m(3) = Vector4ru(0,0,0,1);
    real_unit theta = acosf(
        (old_orient*orient) / (old_orient.norm()*orient.norm()) );
    m = Matrix4ru::rotate(Vector4ru(axis, 0), theta) * m;
    m(3) = pos;

    ptr->set_transformation(m);
#endif
}


void SetOrientation::save(SerializationInterface &ar) const
{
    ar << m_target << m_orientation << m_coord_sys;
}

void SetOrientation::load(DeserializationInterface &ar)
{
    ar >> m_target >> m_orientation >> m_coord_sys;
}
