/*
 * Copyright Staffan Gimåker 2007-2009.
 *
 * ---
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */

#include <boost/math/fpclassify.hpp>

#include "SetPose.hh"

#ifdef __PEEKABOT_SERVER
#  include "../ServerExecutionContext.hh"
#  include "../SceneObject.hh"
#endif


using namespace peekabot;


SetPose::SetPose(
    ObjectID target,
    const Vector3ru &pos,
    real_unit yaw,
    real_unit roll,
    real_unit pitch,
    CoordinateSystem coord_sys) throw()
    : m_target(target),
      m_pos(pos),
      m_yaw(yaw),
      m_roll(roll),
      m_pitch(pitch),
      m_coord_sys(coord_sys)
{
}


SetPose::SetPose() throw()
{
}


SetPose::SetPose(const SetPose &action) throw()
    : m_target(action.m_target),
      m_pos(action.m_pos),
      m_yaw(action.m_yaw),
      m_roll(action.m_roll),
      m_pitch(action.m_pitch),
      m_coord_sys(action.m_coord_sys)
{
}


SetPose::~SetPose() throw()
{
}


Action *SetPose::clone() const
{
    return new SetPose(*this);
}


void SetPose::execute(
    ServerExecutionContext *context)
    const throw(std::exception)
{
#ifdef __PEEKABOT_SERVER
    SceneObject *ptr = context->get_object(m_target);

    if( !ptr )
        throw std::runtime_error(
            "Failed to set pose: target object not found");

    // Check for infinity and NaN
    for( int i = 0; i < 3; i++ )
        if( !boost::math::isfinite(m_pos(i)) )
            throw std::logic_error(
                "Failed to set pose: position cannot contain infinity or NaN");

    if( !boost::math::isfinite(m_yaw) ||
        !boost::math::isfinite(m_roll) ||
        !boost::math::isfinite(m_pitch) )
        throw std::logic_error(
            "Failed to set pose: angles contain infinity or NaN");


    Matrix4r pose;

    switch( m_coord_sys )
    {
        case LOCAL_COORDINATES:
            pose = (Matrix4ru::translate(m_pos) *
                    Matrix4ru::rotate_z(m_yaw) *
                    Matrix4ru::rotate_y(m_roll) *
                    Matrix4ru::rotate_x(m_pitch));
            break;

        case PARENT_COORDINATES:
            pose = (Matrix4ru::translate(m_pos) *
                    Matrix4ru::rotate_z(m_yaw) *
                    Matrix4ru::rotate_y(m_roll) *
                    Matrix4ru::rotate_x(m_pitch));
            break;

        case WORLD_COORDINATES:
            // Rotate, then translate and at last convert to parent coordinates
            pose = (ptr->get_parent_mtow().inverse() *
                    Matrix4ru::translate(m_pos) *
                    Matrix4ru::rotate_z(m_yaw) *
                    Matrix4ru::rotate_y(m_roll) *
                    Matrix4ru::rotate_x(m_pitch));
            break;

        default:
            throw std::runtime_error(
                "Failed to set pose: unsupported coordinate system");
            break;
    }

    ptr->set_transformation(pose);
#endif
}


void SetPose::save(SerializationInterface &ar) const
{
    ar << m_target << m_pos
       << m_yaw << m_roll << m_pitch
       << m_coord_sys;
}

void SetPose::load(DeserializationInterface &ar)
{
    ar >> m_target >> m_pos
       >> m_yaw >> m_roll >> m_pitch
       >> m_coord_sys;
}
