/*
 * Copyright Anders Boberg 2007.
 * Copyright Staffan Gimåker 2008-2009.
 *
 * ---
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */

#include "ViewportMove.hh"

#ifdef __PEEKABOT_SERVER
#  include "../ServerExecutionContext.hh"
#  include "../SceneObject.hh"
#  include "../CameraObject.hh"
#endif


using namespace peekabot;
using namespace peekabot::transformation_target;


ViewportMove::ViewportMove() throw()
{
}


ViewportMove::ViewportMove(
    ObjectID target, real_unit x, real_unit y,
    Axis axes, CoordinateSystem system, ObjectID camera) throw()
    : m_target(target), m_dx(x), m_dy(y), m_axes(axes),
      m_system(system), m_camera(camera)
    
{
}


Action *ViewportMove::clone() const
{
    return new ViewportMove(m_target, m_dx, m_dy, m_axes, m_system, m_camera);
}


void ViewportMove::execute(
    ServerExecutionContext *context) 
    const throw(std::exception)
{
#ifdef __PEEKABOT_SERVER

    SceneObject *ptr = context->get_object(m_target);
    if( !ptr )
        throw std::runtime_error(
            "Target object does not exist.");

    // First, fetch the mtow matrix of the camera...
    CameraObject *camera_ptr = dynamic_cast<CameraObject *>(
        context->get_object(m_camera));
    if(!camera_ptr)
        throw new std::runtime_error(
            "The specified camera does not exist.");
    Matrix4ru camera_mtow = camera_ptr->get_mtow();
    Matrix4ru cam_inverse = camera_mtow.inverse();
    //real_unit f = camera_ptr->get_near_plane();
    //real_unit zoom = camera_ptr->get_zoom_distance();

    //camera_mtow(3) = Vector4ru(0, 0, 0, 0); // We only want axes
    
    // Then, in order to do a mtow->mtop translation, we need
    // the mtow of the target object's parent...
    Matrix4ru parent_mtow;
    SceneObject *parent_ptr = ptr->get_parent();
    if(parent_ptr)
        parent_mtow = parent_ptr->get_mtow();
    // If the object has no parent, it's the root, so we might as well
    // transform it using world coordinates.
    else
        parent_mtow = ptr->get_transformation();
    //parent_mtow(3) = Vector4ru(0, 0, 0, 0); // We only want axes
    
    // Next, we need the target object's mtow to apply the
    // transformation to.        
    Matrix4ru mtow = ptr->get_mtow();

    Matrix4ru trans_matrix = ptr->get_transformation();

    // We currently support only world coordinates or local coordinates.
    if(m_system == PARENT_COORDINATES)
    {
        Vector4ru trans(0, -m_dx, -m_dy , 1);
        parent_mtow(3) = Vector4ru(0, 0, 0, 1);
        camera_mtow(3) = Vector4ru(0, 0, 0, 1);        
        trans = parent_mtow.inverse() * camera_mtow * trans;
        trans_matrix(0, 3) += ((m_axes & X_AXIS) > 0) * trans(0);
        trans_matrix(1, 3) += ((m_axes & Y_AXIS) > 0) * trans(1);
        trans_matrix(2, 3) += ((m_axes & Z_AXIS) > 0) * trans(2);
    }
    else if(m_system == WORLD_COORDINATES)
    {
        Vector4ru trans(0, -m_dx, -m_dy , 1);
        Matrix4ru par_mtow_inv = parent_mtow.inverse();
        //par_mtow_inv(3) = Vector4ru(0, 0, 0, 1);
        camera_mtow(3) = Vector4ru(0, 0, 0, 1);
        trans = camera_mtow * trans;
        trans_matrix = (par_mtow_inv * 
                        Matrix4ru::translate(((m_axes & X_AXIS) > 0) * trans(0),
                                             ((m_axes & Y_AXIS) > 0) * trans(1),
                                             ((m_axes & Z_AXIS) > 0) * trans(2))
                        * mtow);
    }

    else if(m_system == LOCAL_COORDINATES)
    {
        Vector4ru trans(0, -m_dx, -m_dy , 1);
        mtow(3) = Vector4ru(0, 0, 0, 1);
        camera_mtow(3) = Vector4ru(0, 0, 0, 1);
        trans = mtow.inverse() * camera_mtow * trans;
        trans_matrix(3) += ((m_axes & X_AXIS) > 0) * trans(0) * trans_matrix(0);
        trans_matrix(3) += ((m_axes & Y_AXIS) > 0) * trans(1) * trans_matrix(1);
        trans_matrix(3) += ((m_axes & Z_AXIS) > 0) * trans(2) * trans_matrix(2);

        //trans_matrix = (parent_mtow.inverse() * mtow);        
    }
    else if(m_system == CAMERA_COORDINATES)
    {
        // Apply transformation
        mtow(3) += (((m_axes & Y_AXIS) > 0) * m_dx * -camera_mtow(1) +
                    ((m_axes & Z_AXIS) > 0) * m_dy * -camera_mtow(2));

        trans_matrix = (parent_mtow.inverse() * mtow);
    }
    
   
    ptr->set_transformation(trans_matrix);
#endif
}


void ViewportMove::save(SerializationInterface &ar) const
{
    ar << m_target << m_dx << m_dy
       << m_axes << m_system << m_camera;
}

void ViewportMove::load(DeserializationInterface &ar)
{
    ar >> m_target >> m_dx >> m_dy
       >> m_axes >> m_system >> m_camera;
}
