/*
 * Copyright Anders Boberg 2007.
 * Copyright Staffan Gimåker 2008-2009.
 *
 * ---
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */

#include <stdexcept>
#include <cmath>
#include "ViewportRotate.hh"
#include "../serialization/Containers.hh"

#ifdef __PEEKABOT_SERVER
#  include "../ServerExecutionContext.hh"
#  include "../SceneObject.hh"
#  include "../CameraObject.hh"
#endif


#define SIGN(a) ((a < 0) ? -1 : 1)

using namespace peekabot;
using namespace peekabot::transformation_target;


ViewportRotate::ViewportRotate() throw()
{
}


ViewportRotate::ViewportRotate(
    ObjectSet targets, real_unit dx, real_unit dy, Axis axes, 
    CoordinateSystem axis_system, RotationCenterpoint pivot,
    ObjectID camera) throw()
    : m_targets(targets),
      m_dx(dx),
      m_dy(dy),
      m_camera(camera),
      m_axes(axes),
      m_axis_system(axis_system),
      m_pivot(pivot)

{
}

Action *ViewportRotate::clone() const
{
    return new ViewportRotate(m_targets, m_dx, m_dy, m_axes, m_axis_system,
                              m_pivot, m_camera);
}


void ViewportRotate::execute(
    ServerExecutionContext *context) 
    const throw(std::exception)
{
#ifdef __PEEKABOT_SERVER

    // First, calculate the average centerpoint if used
    Vector4ru average_center(0, 0, 0, 0);
    if(m_pivot == AVERAGE_CENTER)
    {
        for(ObjectSet::iterator it = m_targets.begin();
            it != m_targets.end(); ++it)
        {
            SceneObject *ptr = context->get_object(*it);            
            average_center += ptr->get_mtow()(3);
        }
        if(m_targets.size() > 0)
            average_center /= m_targets.size();
    }
    
    // Next, apply transformation to all objects in the set.
    for(ObjectSet::iterator it = m_targets.begin();
        it != m_targets.end(); ++it)
    {
        // First fetch all relevant transformation matrices.
        // We want to transform them into world coordinates
        // so that they are all handled uniformly.

        SceneObject *ptr = context->get_object(*it);
        if( !ptr )
            throw std::runtime_error("ViewportRotate: Object not found.");

        SceneObject *cam_ptr = context->get_object(m_camera);
        if( !ptr )
            throw std::runtime_error("ViewportRotate: Camera object not found.");

        Matrix4ru cam_mtow = cam_ptr->get_mtow();
        Matrix4ru trans_matrix = ptr->get_mtow();
   
        SceneObject *parent_ptr = ptr->get_parent();
        Matrix4ru parent_mtow;
        if(parent_ptr)
            parent_mtow = parent_ptr->get_mtow();
        else
            parent_mtow.load_identity();

        Matrix4ru rotational_matrix;
        real_unit x_amount, y_amount, z_amount;

        if(m_axis_system == WORLD_COORDINATES)
        {
            rotational_matrix = Matrix4ru::identity();
        }
        else if(m_axis_system == LOCAL_COORDINATES)
            rotational_matrix = ptr->get_mtow();
        else if(m_axis_system == PARENT_COORDINATES)
        {
            rotational_matrix = parent_mtow;
        }
        else
            rotational_matrix = cam_mtow;

        // The amount of rotation is scaled by the amount the cursor
        // has moved in the othogonal direction of the rotational axis.

        float yscalar, zscalar;
        yscalar = cam_mtow(1) * rotational_matrix(0);
        zscalar = cam_mtow(2) * rotational_matrix(0);
        x_amount = ((m_dx * SIGN(yscalar) * (1 - fabs(yscalar)) +
                     m_dy * SIGN(zscalar) * (1 - fabs(zscalar))) *
                    (m_axes & X_AXIS));
        yscalar = cam_mtow(1) * rotational_matrix(1);
        zscalar = cam_mtow(2) * rotational_matrix(1);
        y_amount = ((m_dx * SIGN(yscalar) * (1 - fabs(yscalar)) +
                     m_dy * SIGN(yscalar) * (1 - fabs(zscalar))) *
                    (m_axes & Y_AXIS));
        yscalar = cam_mtow(1) * rotational_matrix(2);
        zscalar = cam_mtow(2) * rotational_matrix(2);
        z_amount = ((m_dx * SIGN(yscalar) * (1 - fabs(yscalar)) +
                     m_dy * SIGN(yscalar) * (1 - fabs(zscalar))) *
                    (m_axes & Z_AXIS));
        /*
        else
        {
            rotational_matrix = cam_mtow;
            
            x_amount = 0;
            y_amount = m_dy * (m_axes & Y_AXIS);
            z_amount = -m_dx * (m_axes & Z_AXIS);
        }
        
        if(m_axis_system != CAMERA_COORDINATES)
        {
            x_amount = ((-m_dx * cam_mtow(2) * rotational_matrix(0) +
                         m_dy * cam_mtow(1) * rotational_matrix(0)) *
                        (m_axes & X_AXIS));
            y_amount = ((-m_dx * cam_mtow(2) * rotational_matrix(1) +
                         m_dy * cam_mtow(1) * rotational_matrix(1)) *
                        (m_axes & Y_AXIS));
            z_amount = ((-m_dx * cam_mtow(2) * rotational_matrix(2) +
                         m_dy * cam_mtow(1) * rotational_matrix(2)) *
                        (m_axes & Z_AXIS));
        }
        */

        // Transform pivot coordinates into world
        Vector4ru tmp(0, 0, 0, 0);
        if(m_pivot == LOCAL_CENTER)
            tmp = trans_matrix(3);
        else if(m_pivot == PARENT_CENTER)
            tmp = parent_mtow(3);
        else if(m_pivot == AVERAGE_CENTER)
            tmp = average_center;
        
        // If WORLD_COORDINATES we don't need to do anything

        tmp(3) = 0;
        trans_matrix(3) -= tmp;
        trans_matrix *= Matrix4ru::rotate(rotational_matrix(0), x_amount);
        trans_matrix *= Matrix4ru::rotate(rotational_matrix(1), y_amount);
        trans_matrix *= Matrix4ru::rotate(rotational_matrix(2), z_amount);
        trans_matrix(3) += tmp;

        trans_matrix *= parent_mtow.inverse();

        ptr->set_transformation(trans_matrix);
    }
#endif
}


void ViewportRotate::save(SerializationInterface &ar) const
{
    ar << m_targets << m_dx << m_dy 
       << m_camera << m_axes << m_axis_system
       << m_center_point << m_pivot;
}

void ViewportRotate::load(DeserializationInterface &ar)
{
    ar >> m_targets >> m_dx >> m_dy 
       >> m_camera >> m_axes >> m_axis_system
       >> m_center_point >> m_pivot;
}
