/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "ActionRouter.hh"
#include "../Action.hh"

using namespace peekabot;
using namespace peekabot::atl;



bool ActionRouter::set_action_handler(
    SourceID target, ActionHandler handler) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    if( m_handlers.find( target ) != m_handlers.end() )
        return true;
    
    m_handlers.insert( std::make_pair(target, handler) );

    return false;
}



bool ActionRouter::unset_action_handler(SourceID target) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    HandlerMap::iterator it = m_handlers.find(target);
  
    if( it == m_handlers.end() )
        return true;

    m_handlers.erase(it);

    return false;
}


void ActionRouter::post_action(
    Action *action, SourceID src, SourceID dest) throw()
{
    HandlerMap::const_iterator it;

    {
        boost::recursive_mutex::scoped_lock lock(m_mutex);
        it = m_handlers.find(dest);
    }
  
    if( it != m_handlers.end() )
    {
        (it->second)(action, src);
        notify(ON_ACTION_ROUTED);
    }
    else
    {
        notify(ON_ACTION_DROPPED);
    }
}


bool ActionRouter::register_observer(
    RouterEvent event, RouterObserver *observer) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    ObserverSet &trigger_set = m_observers[event];

    // Check to see that the observer is not already registered
    bool ret_val = trigger_set.insert(observer).second;

    return !ret_val;
}


bool ActionRouter::deregister_observer(
    RouterEvent event, RouterObserver *observer) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    ObserverMap::iterator map_it = m_observers.find(event);

    // Check to see that there is any observers registered for this event
    if(map_it == m_observers.end() || map_it->second.empty())
        return true;

    // Find and erase the observer we're looking for.
    ObserverSet::iterator set_it = map_it->second.begin();
    for(; set_it != map_it->second.end(); ++set_it)
    {
        if(*set_it == observer)
        {
            map_it->second.erase(set_it);
            return false;
        }        
    }

    return true;
}

        
void ActionRouter::notify(RouterEvent event) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    ObserverMap::iterator it = m_observers.find(event);

    if( it != m_observers.end() )
    {
        for( ObserverSet::iterator obs = it->second.begin();
             obs != it->second.end(); ++obs )
        {
            (*obs)->on_notify(event, this);
        }
    }
}
