/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_ATL_ACTION_ROUTER_HH
#define __PEEKABOT_ATL_ACTION_ROUTER_HH

// NOTE to self: When an action is denied, a failure response must be formed to 
// the client! Otherwise the client may get stuck querying the result/status of 
// the dropped action.
//
// !!! THIS IS YET TO BE IMPLEMENTED!

#include <boost/function.hpp>
#include <boost/thread/recursive_mutex.hpp>
#include <map>
#include <set>

#include "../Types.hh"
#include "../Singleton.hh"

namespace peekabot
{
//
// Forward declarations
//
class Action;

namespace atl
{
    //
    // Forward declarations
    //
    class ActionRouter;


    /// \todo Document this, in detail.
    enum RouterEvent
    {
        ON_ACTION_DROPPED,
        ON_ACTION_ROUTED
    };

    /// \todo Document this, in detail.
    class RouterObserver
    {
    public:
        virtual ~RouterObserver() {}

        virtual void on_notify(RouterEvent event, ActionRouter *router) throw() {}
    };


    /**
     * \brief Handles routing \c Actions from the poster to the 
     * designated target.
     *
     * The router also employed a configurable \c ActionFilter that filters out
     * unwanted \c Actions.  Any \c Action that has a target with no sink set 
     * will be dropped, too.
     *
     * The \c RouterObservers can be used to, for example, to send error messages
     * to the sender of a dropped \c Action.
     */
    class ActionRouter
    {
    public:
        typedef boost::function2<void, Action *, SourceID> ActionHandler;
      
        /**
         * \brief Set the function to handle actions designated for \c target.
         *
         * \return Returns \c false on success, \c true if an action handler for
         * the given \e target is already installed.
         */
        bool set_action_handler(SourceID target, ActionHandler handler) throw();

        /**
         * \brief Unset the handler for actions designated for \c target.
         *
         * \return \c false on success, \c true if no handler for the given
         * \e target was found.
         */        
        bool unset_action_handler(SourceID target) throw();
       
        /**
         * \brief Post an action, which will automagically arrive at it's 
         * specified designation.
         *
         * Does not, as it is right now at least, take ownership of the action
         * passed to it.
         */
        void post_action(Action *action, SourceID src, SourceID dest) throw();
        // 1. run through filter - drop?
        // 2. if no, find sink
        // 3. no sink found => drop, otherwise: call the sink -- sink(action)

        /// \todo Document this, in detail.
        bool register_observer(RouterEvent event, RouterObserver *observer) throw();

        /// \todo Document this, in detail.
        bool deregister_observer(RouterEvent event, RouterObserver *observer) throw();
        
    private:
        void notify(RouterEvent event) throw();

    private:
        typedef std::map<SourceID, ActionHandler> HandlerMap;

        /**
         * \brief Stores the mapping of which handler is used for which
         * source ID.
         */        
        HandlerMap m_handlers;

        typedef std::set<RouterObserver *> ObserverSet;
        typedef std::map<RouterEvent, ObserverSet> ObserverMap;

        /**
         * \brief Stores the mapping from event to registered observers.
         */
        ObserverMap m_observers;

        /**
         * \brief Used to synchronize access to \c m_observers and 
         * \c m_handlers.
         */
        mutable boost::recursive_mutex m_mutex;
    };


    typedef Singleton<ActionRouter> TheActionRouter;


}

using peekabot::atl::TheActionRouter;

}



#endif // __PEEKABOT_ATL_ACTION_ROUTER_HH
