/*
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <cstdio>
#include <boost/bind.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>

#include "ActionTransporter.hh"
#include "ClientConnection.hh"

#ifdef HAVE_SOCKETPAIR
#  ifdef HAVE_UNISTD_H
#    include <unistd.h>
#  else
#    error "Don't know which file to include to get socketpair!"
#  endif
#endif



using namespace peekabot;
using namespace peekabot::atl;




//
// --------- Constants ----------
//
static const int BACKLOG = 10;
static const int SELECT_TIMEOUT = 50*1000; // 50 ms (in microsecs)



ActionTransporter::ActionTransporter() throw()
    : m_client_id_allocator(new DefaultIDAllocator<SourceID>(0,999)),
      m_select_thread(0),
      m_stop_signal(false),
      m_is_running(false),
      m_listen_port(0),
      m_listen_fd(0),
      m_wake_wrt_fd(0),
      m_wake_rd_fd(0)
{
    FD_ZERO(&m_fds);
    FD_ZERO(&m_busy_fds);
}


ActionTransporter::~ActionTransporter() throw()
{
    if( is_running() )
        stop();

    assert( m_clients.empty() );
}


bool ActionTransporter::register_observer(
    ConnectionEvent event, ConnectionObserver *observer) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_observers_mutex);

    ObserverSet &observer_set = m_observers[event];

    // Check to see that the observer is not already registered
    bool ret_val = observer_set.insert(observer).second;

    return !ret_val;
}


bool ActionTransporter::deregister_observer(
    ConnectionEvent event, ConnectionObserver *observer) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_observers_mutex);

    ObserverMap::iterator map_it = m_observers.find(event);

    // Check to see that there is any observers registered for this event
    if(map_it == m_observers.end() || map_it->second.empty())
        return true;

    // Find and erase the observer we're looking for.
    ObserverSet::iterator set_it = map_it->second.begin();
    for(; set_it != map_it->second.end(); ++set_it)
    {
        if(*set_it == observer)
        {
            map_it->second.erase(set_it);
            return false;
        }        
    }

    return true;
}




bool ActionTransporter::stop() throw()
{
    if( !is_running() )
        return true;

    // keep anyone else from calling start/stop while we're at it
    boost::recursive_mutex::scoped_lock lock(m_stop_signal_mutex);

    // stop listening for new connections
    stop_listening();

    {
        boost::recursive_mutex::scoped_lock lock(m_clients_mutex);

        // close all connections
        for( ConnectionMap::iterator it = m_clients.begin();
             it != m_clients.end(); ++it )
            it->second->close();
    }

    // signal the select thread to halt (i.e. run())
    m_stop_signal = true;
    
    // wake the select thread, so that it can be stopped right away
    // (not strictly necessary - the select call will time out sooner
    // or later anywhow)
    wake();

    // wake for it [the select thread] to finish
    m_select_thread->join();
    delete m_select_thread;
    m_select_thread = 0;

    assert( m_closed.empty() );
    assert( m_clients.empty() );
    
    // close notification fds
    tear_down_notification_fds();

    {
        boost::recursive_mutex::scoped_lock lock(m_resource_mutex);
        m_is_running = false;
    }
    {
        boost::recursive_mutex::scoped_lock lock(m_observers_mutex);
        
        m_observers.clear();
    }

    m_stop_signal = false;

    return false;
}


bool ActionTransporter::is_running() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_resource_mutex);

    return m_is_running;
}



        /**
         * \brief Retrieve a list of all connections.
         */
std::list<ClientConnection *> ActionTransporter::get_connections() throw()
{
    // TODO
    return std::list<ClientConnection *>();
}


        /**
         * \brief For a client ID, get the associated \c ClientConnection.
         *
         * \return Returns \c null if a client with the given ID is not 
         * connected.
         */
ClientConnection *ActionTransporter::get_connection(SourceID client_id) throw()
{
    // TODO
    return 0;
}


        /**
         * \brief Get the number of currently connected clients.
         */
int ActionTransporter::get_connection_count() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_clients_mutex);
    
    return m_clients.size();
}

    
bool ActionTransporter::start(unsigned int port) throw()
{
    if( is_running() )
        return true;

    if( start_listening(port) )
        return true;

    // lock the stop signal mutex to make sure no one else calls stop()
    // while we're starting up
    boost::recursive_mutex::scoped_lock stop_lock(m_stop_signal_mutex);

    // synchronize resource access and prevent anyone from doing anything
    // else, basically.
    boost::recursive_mutex::scoped_lock lock(m_resource_mutex);

    // 1. Create a notification pipe/socket-pair (e.g. connect to self)
    if( set_up_notification_fds() )
    {
        stop_listening();
        return true;
    }

    // 2. Create (and start) the select thread
    m_select_thread = new boost::thread(
        boost::bind(&ActionTransporter::run, this));

    m_is_running = true;
    
    return false;
}



void ActionTransporter::run() throw()
{
    boost::posix_time::ptime last_tick_time(
        boost::posix_time::microsec_clock::local_time());

    fd_set readfds;
    fd_set writefds;

    while( !m_stop_signal )
    {
        {
            boost::recursive_mutex::scoped_lock lock(m_resource_mutex);

            handle_closed_connections();

            // set up fd set
            readfds = m_fds;
            writefds = m_busy_fds;
            // add the listener socket
            FD_SET(m_listen_fd, &readfds);
            // add the wake fd
#ifdef HAVE_SOCKETPAIR
            FD_SET(m_wake_rd_fd, &readfds);
#endif
        }

        timeval tv;
        tv.tv_sec = 0;
        tv.tv_usec = SELECT_TIMEOUT;
        int n_ready = select(get_max_fd()+1, &readfds, &writefds, 0, &tv);

        // generate tick events
        generate_tick_events(last_tick_time);

        // TODO: check for failure (n_ready == -1) // A: INVALID_SOCKET!
        if( n_ready < 0 )
            continue; // Timed out, no sockets ready..
       
        handle_new_connections(&readfds);

        {
            boost::recursive_mutex::scoped_lock lock(m_resource_mutex);

            // Make sure we don't get a hit on the listen socket
            FD_CLR(m_listen_fd, &readfds);
            // If there's data on the wake fd, read it (one byte per wake request) 
            // and clear it
            if( FD_ISSET(m_wake_rd_fd, &readfds) )
            {
                // TODO: check for errors
                uint8_t x[128];
                while( ::recv(m_wake_rd_fd, (sockets::BufType)x, 128, 0) > 0 );
                FD_CLR(m_wake_rd_fd, &readfds);
            }
        }
        {
            boost::recursive_mutex::scoped_lock lock(m_clients_mutex);

            // The m_clients data structure is guaranteed to remain valid, even
            // if a connection closes itself - implying that it should be updated -
            // until after all callbacks have been dealt with.
            // This is done by recording such changes and executing them later - 
            // see below!
            for( ConnectionMap::iterator it = m_clients.begin();
                 it != m_clients.end(); ++it )
            {
                if( !it->second->is_closed() )
                {
                    // Note: it would make sense to store fds in the transporter,
                    // since we now have to do a lock/unlock for each client - 
                    // potentially costly!
                    sockets::SocketType fd = it->second->get_socket_fd();
                    
                    if( FD_ISSET(fd, &readfds) )
                        it->second->on_data_available();
                    
                    if( FD_ISSET(fd, &writefds) )
                        it->second->on_writable();
                }
            }
        }
    }

    handle_closed_connections();
}



bool ActionTransporter::set_up_notification_fds() throw()
{
#ifdef HAVE_SOCKETPAIR
    boost::recursive_mutex::scoped_lock lock(m_resource_mutex);

    // This works only on posix systems!!!
    int fd[2];

    /*if( pipe(fd) != 0 )
      return true;*/

    if( socketpair(AF_UNIX, SOCK_STREAM, 0, fd) != 0 )
        return true;

    m_wake_rd_fd  = fd[0];
    m_wake_wrt_fd = fd[1]; // note: it really should be blocking.

    // shut the unused end of each socket down
    shutdown(m_wake_rd_fd, SHUT_WR);
    shutdown(m_wake_wrt_fd, SHUT_RD);

    sockets::set_nonblocking(m_wake_rd_fd);
    sockets::set_nonblocking(m_wake_wrt_fd);

    m_fd_set.insert(m_wake_rd_fd);
    m_fd_set.insert(m_wake_wrt_fd);
#endif // HAVE_SOCKETPAIR
    
    return false;
}


void ActionTransporter::tear_down_notification_fds() throw()
{
#ifdef HAVE_SOCKETPAIR
    boost::recursive_mutex::scoped_lock lock(m_resource_mutex);

    close(m_wake_rd_fd);
    close(m_wake_wrt_fd);

    m_fd_set.erase(m_wake_rd_fd);
    m_fd_set.erase(m_wake_wrt_fd);
#endif // HAVE_SOCKETPAIR
}


void ActionTransporter::ActionTransporter::wake() const throw()
{
#ifdef HAVE_SOCKETPAIR
    //boost::recursive_mutex::scoped_lock lock(m_resource_mutex);

    //if( is_running() )
    {
        uint8_t x = 0;
        //assert( ::send(m_wake_wrt_fd, &x, 1, 0) == 1 );
        ::send(m_wake_wrt_fd, (sockets::BufType)&x, 1, 0);
    }
#endif // HAVE_SOCKETPAIR
}


bool ActionTransporter::start_listening(unsigned int port) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_resource_mutex);


    m_listen_fd = socket(PF_INET, SOCK_STREAM, 0);
    
    // Make sure the listen socket doesn't block.
    sockets::set_nonblocking(m_listen_fd);

    if( m_listen_fd == INVALID_SOCKET )
    {
        perror("socket");
        return true;
    }

    int yes = 1;
    // lose the pesky "address already in use" error message
    if( setsockopt(m_listen_fd, SOL_SOCKET, SO_REUSEADDR, 
                   (sockets::BufType)&yes, sizeof(int)) == SOCKET_ERROR ) {
        perror("setsockopt");
        return true;
    }

    // bind
    sockaddr_in myaddr;
    myaddr.sin_family = AF_INET;
    myaddr.sin_addr.s_addr = INADDR_ANY;
    myaddr.sin_port = htons(port);
    memset(&(myaddr.sin_zero), '\0', 8);

    if( bind(m_listen_fd, (sockaddr *)&myaddr, sizeof(myaddr)) == SOCKET_ERROR )
    {
        perror("bind"); // ?
        return true;
    }

    if( listen(m_listen_fd, BACKLOG) == SOCKET_ERROR )
    {
        perror("listen"); // ?
        return true;
    }

    m_listen_port = port;

    m_fd_set.insert(m_listen_fd);

    return false;
}


void ActionTransporter::stop_listening() throw()
{
    boost::recursive_mutex::scoped_lock lock(m_resource_mutex);

    sockets::close_socket(m_listen_fd);

    m_fd_set.erase(m_listen_fd);

    m_listen_port = 0;
    m_listen_fd = 0;
}



void ActionTransporter::notify(
    ConnectionEvent event, ClientConnection *conn) throw()
{
    {
        boost::recursive_mutex::scoped_lock lock(m_observers_mutex);

        // Generate observer notifications
        ObserverMap::iterator it = m_observers.find(event);
        
        if( it != m_observers.end() )
        {
            for( ObserverSet::iterator obs = it->second.begin();
                 obs != it->second.end(); ++obs )
            {
                (*obs)->on_notify(event, conn);
            }
        }
    }

    {
        boost::recursive_mutex::scoped_lock lock(m_resource_mutex);

        // Record disconnected clients -- we'll use
        // the list of recorded clients to update our internal
        // state at the end after each iteration of the loop in run().
        //
        // Also add/remove busy/idle clients from the set of fds that are 
        // active sending data.
        if( event == ON_CLIENT_DISCONNECTED )
        {
            m_closed.push_back(conn);
            m_on_client_disconnected(conn);
        }
        else if( event == ON_IDLE_SENDING )
            FD_CLR(conn->get_socket_fd(), &m_busy_fds);
        else if( event == ON_BUSY_SENDING )
            FD_SET(conn->get_socket_fd(), &m_busy_fds);
    }
}



void ActionTransporter::handle_new_connections(fd_set *readfds) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_resource_mutex);

    // A new connection?
    if( FD_ISSET(m_listen_fd, readfds) )
    {
        // Eureka!
        sockaddr_in remote_addr;
        
        socklen_t addr_len = sizeof(sockaddr_in);
        int new_fd = accept(m_listen_fd, (sockaddr *)&remote_addr, &addr_len);
        
        if( new_fd != INVALID_SOCKET )
        {
            boost::recursive_mutex::scoped_lock lock(m_clients_mutex);

            // Okay, accept succeeded.           
            SourceID client_id = m_client_id_allocator->allocate();
            ClientConnection *conn = new ClientConnection(
                this, new_fd, client_id);
          
            // Update internal state:
            FD_SET(new_fd, &m_fds);
            m_clients.insert(std::make_pair(client_id, conn));
            m_fd_set.insert(new_fd);
                       
            notify(ON_CLIENT_CONNECTED, conn);
            m_on_client_connected(conn);
        }
        // else: TODO report error
    }
}



void ActionTransporter::handle_closed_connections() throw()
{
    boost::recursive_mutex::scoped_lock lock(m_resource_mutex);
    
    // Handle closed connections
    for( ClosedList::iterator it = m_closed.begin();
         it != m_closed.end(); ++it )
    {
        boost::recursive_mutex::scoped_lock lock(m_clients_mutex);

        ClientConnection *conn = *it;
        sockets::SocketType fd = conn->get_socket_fd();
               
        FD_CLR(fd, &m_fds);
        FD_CLR(fd, &m_busy_fds);
        m_client_id_allocator->release(conn->get_id());
        m_fd_set.erase(fd);
        m_clients.erase(conn->get_id());
        
        delete conn;
    }
    
    m_closed.clear();
}


void ActionTransporter::generate_tick_events(
    boost::posix_time::ptime &last_tick_time) throw()
{
    boost::posix_time::time_duration dt = 
        boost::posix_time::microsec_clock::local_time() - last_tick_time;

    if( dt.total_milliseconds() > 1000 )
    {
        boost::recursive_mutex::scoped_lock lock(m_clients_mutex);
        
        for( ConnectionMap::iterator it = m_clients.begin();
             it != m_clients.end(); ++it )
            it->second->on_tick(dt);
        
        last_tick_time = boost::posix_time::microsec_clock::local_time();
    }
}



sockets::SocketType ActionTransporter::get_max_fd() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_resource_mutex);
    //assert( *m_fd_set.rbegin() >= *m_fd_set.begin() );

    return !m_fd_set.empty() ? *m_fd_set.rbegin() : 0;
}

