/*
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_ATL_ACTION_TRANSPORTER_HH
#define __PEEKABOT_ATL_ACTION_TRANSPORTER_HH


#include <set>
#include <map>
#include <list>
#include <boost/thread.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/thread_safe_signal.hpp>

#include "../Types.hh"
#include "../IDAllocator.hh"
#include "../Singleton.hh"
#include "ClientConnection.hh"


namespace peekabot
{
namespace atl
{



    /**
     * \brief An interface for being notified of \c ClientConnection events.
     */
    class ConnectionObserver
    {
    public:
        virtual ~ConnectionObserver() {}

        /**
         * \brief Called when the observer is triggered by an event its been 
         * registered with.
         *
         * \param event The stimuli that caused the observer to fire.
         * \param conn The connection that caused the stimuli.
         */
        virtual void on_notify(
            ConnectionEvent event, 
            ClientConnection *conn) throw() = 0;
    };




    /**
     * \brief The class that manages and oversees the operations of a set of 
     * \c ClientConnections. It listens for and handles new connections and 
     * performs some of the grunt, low-level socket operations.
     */
    class ActionTransporter
    {
        friend class ClientConnection;

    public:
        ActionTransporter() throw();

        virtual ~ActionTransporter() throw();

        /**
         * \brief Register an observer for a specific event. The observer is 
         * triggered by any connection owned by the transporter.
         */
        bool register_observer(ConnectionEvent event, ConnectionObserver *observer) throw();

        /**
         * \brief Deregister an observer previously registered with the
         * \c ActionTransporter.
         */
        bool deregister_observer(ConnectionEvent event, ConnectionObserver *observer) throw();


        /*bool kill_connection(SourceID::IDType client_id) throw()
          {
          find conn from in m_clients (id -> conn)
          conn->close();
          }*/


        // can be called multiple times to listen on several ports

        /**
         * \brief Start listening on the specified port. The server can be 
         * instructed to listen to several ports by calling this method repeatedly.
         *
         * Calling this method will start the action transporter if it has not 
         * previously been started.
         *
         * \param port The port on which to listen (\e note: port numbers below 1024
         * generally require root permission on unices).
         * \return Returns \c true in case of error.
         */
        //bool listen_on(unsigned int port) throw();
        // bind
        // if( running )
        //   wake()
        // else
        //   start()

        /**
         * \brief Returns the ports that are used to listen for new connections.
         */
        //std::set<unsigned int> get_listen_ports() const throw();

        unsigned int get_port() const throw();


        /**
         * \brief Start the action transporter.
         *
         * No connections can be established before the action transporter has 
         * been started.
         *
         * \post <tt>is_running() = true</tt>
         *
         * \return \c true if the transporter is already running, \c false 
         * otherwise.
         */
        bool start(unsigned int port) throw();


        /**
         * \brief Stop the action transporter.
         *
         * Stopping the transporter will disconnect all connected clients. No new 
         * connections can be made until the action transporter is once again 
         * started.
         *
         * \return \c true if the transporter is not running or in case of 
         * error, \c false otherwise.
         */
        bool stop() throw();

        /**
         * \brief Check whether the action transporter is running or not.
         */
        bool is_running() const throw();



        /**
         * \brief Retrieve a list of all connections.
         */
        std::list<ClientConnection *> get_connections() throw();

        /**
         * \brief For a client ID, get the associated \c ClientConnection.
         *
         * \return Returns \c null if a client with the given ID is not 
         * connected.
         */
        ClientConnection *get_connection(SourceID client_id) throw();

        /**
         * \brief Get the number of currently connected clients.
         */
        int get_connection_count() const throw();


        typedef boost::signal<void (ClientConnection *)> OnClientConnectedSignal;

        OnClientConnectedSignal m_on_client_connected;

        typedef boost::signal<void (ClientConnection *)> OnClientDisconnectedSignal;
        
        OnClientDisconnectedSignal m_on_client_disconnected;

   
    private:
        /**
         * \brief The method doing all the work - i.e. select on the socket fd set 
         * and notify the connections appropriately.
         *
         * Runs in it's own thread!
         */
        void run() throw();
        // create notification pipe/socket (e.g. connect to self)
        // if socket pair, call shutdown() to disallow futher sends/recvs as appropriate
        // while(!m_stop_signal)
        // {
        //   select
        //   accept, handle new connections
        //   do callbacks
        //   generate notifications
        // }


        /**
         * \brief Wake the select-thread. E.g., force it to take any connection
         * state changes into account (e.g. a connection that just started 
         * writing).
         */
        void wake() const throw();


        bool set_up_notification_fds() throw();

        void tear_down_notification_fds() throw();




        bool start_listening(unsigned int port) throw();

        void stop_listening() throw();


        /**
         * \brief Called by owned connections to propagate state information
         * from the to the transporter.
         *
         * This method notifies the appropriate observers and also performs
         * some internal state handling - e.g. update it's set of socket file
         * descriptors which it uses in it's \c select() calls.
         *
         * \sa \c run(), \c ConnectionEvent, \c ClientConnection
         */
        void notify(ConnectionEvent event, ClientConnection *conn) throw();

        void handle_new_connections(fd_set *readfds) throw();

        void handle_closed_connections() throw();

        void generate_tick_events(
            boost::posix_time::ptime &last_tick_time) throw();

        /**
         * \brief Return the largest of all file descriptors, including the 
         * listening socket file descriptor.
         */
        sockets::SocketType get_max_fd() const throw();


    private:
        /**
         * \brief An allocator for allocating unique client IDs to every
         * client connection.
         */
        IDAllocator<SourceID> *m_client_id_allocator;

        mutable boost::recursive_mutex m_stop_signal_mutex;

        mutable boost::recursive_mutex m_resource_mutex;

        mutable boost::recursive_mutex m_observers_mutex;

        mutable boost::recursive_mutex m_clients_mutex;

        boost::thread *m_select_thread;

        /**
         * \brief Governs thread halting.
         *
         * Set to \c true by \c stop() when the transporter should halt.
         * Reset to its initial value of \c false when the transported is 
         * halted.
         *
         * The variable is monitored by \c run(), which resets this cue and
         * returns once it detects that it's been set to \c true.
         */
        volatile bool m_stop_signal;

        bool m_is_running;


        unsigned int m_listen_port;

        sockets::SocketType m_listen_fd;

        typedef std::map<SourceID, ClientConnection *> ConnectionMap;
        ConnectionMap m_clients;


        typedef std::list<ClientConnection *> ClosedList;

        /**
         * \brief This list stores all connections that have been closed,
         * but have not yet had that state change propagated into the other
         * data structures (e.g. \c m_fds, \c m_clients).
         *
         * Values are added from \c notify, called by \c ClientConnections, in 
         * response from callbacks generated by \c run().
         * They are then handled and removed by the end of each iteration 
         * of the select loop in \c run().
         */
        ClosedList m_closed;


        typedef std::set<ConnectionObserver *> ObserverSet;
        typedef std::map<ConnectionEvent, ObserverSet> ObserverMap;
        ObserverMap m_observers;

        std::set<sockets::SocketType> m_fd_set;


        /**
         * \brief The set of all file descriptors checked for a ready-to-write
         * state in \c run().
         */
        fd_set m_busy_fds;

        /**
         * \brief The set of all file descriptors of all the transporter's 
         * connections. Used to check for data-availablity in \c run().
         */
        fd_set m_fds;

        /**
         * \brief The write end of file descriptor pair used to wake
         * the select thread (i.e. \c run()).
         *
         * You should \e not assume that this is a two way communication
         * channel - this file descriptor must be used for 
         * <em>writing only</em>.
         */
        sockets::SocketType m_wake_wrt_fd;

        /**
         * \brief The read end of file descriptor pair used to wake
         * the select thread (i.e. \c run()).
         *
         * See  \c m_wake_wrt_fd, restrictions anologous to those described 
         * there apply to this file descriptor.
         */
        sockets::SocketType m_wake_rd_fd;
    };



}

    typedef Singleton<atl::ActionTransporter> TheActionTransporter;

}


#endif // __PEEKABOT_ATL_ACTION_TRANSPORTER_HH

