/*
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <boost/bind.hpp>
#include <boost/scoped_array.hpp>

#include "ActionRouter.hh"
#include "ConnectionStates.hh"
#include "ActionTransporter.hh"
#include "../serialization/SerializationInterface.hh"
#include "../serialization/DeserializationInterface.hh"
#include "../serialization/MemAdapters.hh"
#include "../serialization/ChunkedBufferAdapter.hh"
#include "../MessageHub.hh"

extern "C" {
#include <liblzf/lzf.h>
}


using namespace peekabot;
using namespace peekabot::atl;



        // is_big_endian is propagated from UnauthenticatedState
AuthenticatedState::AuthenticatedState(
    ClientConnection *connection, bool is_big_endian)
    : ConnectionState(connection),
      m_read_state(RS_HEADER_1),
      m_recv_buf(8*1024),
      m_is_big_endian(is_big_endian)
{
}


void AuthenticatedState::on_data_available() throw()
{
    // just read it!!
    size_t n;
    char ibuf[16*1024];

    do
    {
        n = get_connection()->recv(ibuf, 16*1024);
        m_recv_buf.write(ibuf, n);
    }
    // Is there more to read?
    while( get_connection()->peek(ibuf, 1) == 1 );

    std::size_t tmp;
    do
    {
        tmp = m_recv_buf.get_size();
        handle_data();
    }
    while( tmp != m_recv_buf.get_size() );
}


void AuthenticatedState::handle_data()
{
    if( m_read_state == RS_HEADER_1 && m_recv_buf.get_size() >= 5 )
    {
        ChunkedBufferAdapter adapter(m_recv_buf);
        DeserializationInterface depickle(adapter, m_is_big_endian);

        depickle >> m_body_size >> m_ctrl_byte;
        if( m_ctrl_byte == 1 )
            m_read_state = RS_HEADER_2;
        else
            m_read_state = RS_BODY;
    }
    else if( m_read_state == RS_HEADER_2 && m_recv_buf.get_size() >= 4 )
    {
        ChunkedBufferAdapter adapter(m_recv_buf);
        DeserializationInterface depickle(adapter, m_is_big_endian);
        depickle >> m_uncomp_size;

        m_read_state = RS_BODY;
    }
    else if( m_read_state == RS_BODY && m_recv_buf.get_size() >= m_body_size )
    {
        consume_action();

        m_read_state = RS_HEADER_1;
    }
}


void AuthenticatedState::consume_action()
{
    size_t recv_buf_size = m_recv_buf.get_size();
    Action *action = 0;
    try
    {
        ChunkedBufferAdapter adapter(m_recv_buf);
        DeserializationInterface depickle(adapter, m_is_big_endian);

        if( m_ctrl_byte == 1 )
        {
            boost::scoped_array<uint8_t> uncomp(
                new uint8_t[m_uncomp_size]);
            boost::scoped_array<uint8_t> comp(
                new uint8_t[m_body_size]);

            depickle.load_binary(comp.get(), m_body_size);

            if( lzf_decompress(
                    comp.get(), m_body_size, uncomp.get(), m_uncomp_size)
                != m_uncomp_size )
                throw std::runtime_error("Failed to decompress action");

            MemDeserializationBuffer buf(uncomp.get(), m_uncomp_size);
            DeserializationInterface ar(
                buf, PEEKABOT_BYTE_ORDER == PEEKABOT_BIG_ENDIAN);

            ar >> action;
        }
        else
        {
            depickle >> action;
        }
    }
    catch(std::exception &e)
    {
        TheMessageHub::instance().publish(
            ERROR_MESSAGE,
            "Failed to deserialize action, exception caught: %s",
            e.what());
    }
    catch(...)
    {
        TheMessageHub::instance().publish(
            ERROR_MESSAGE,
            "Failed to deserialize action, unknown exception caught");
    }

    if( action )
        TheActionRouter::instance().post_action(
            action, get_connection()->get_id(), SERVER_SOURCE_ID);
    else
        delete action;

    // The number of bytes consumed by deserialize_state should be
    // the same as published in the ation's header. If it's not,
    // we print a warning and consume the unconsumed bytes.
    if( m_recv_buf.get_size() != recv_buf_size-m_body_size )
    {
        size_t n_read = recv_buf_size - m_recv_buf.get_size();

        TheMessageHub::instance().publish(
            WARNING_MESSAGE,
            "The number of bytes consumed during action "
            "deserialization  doesn't match the number advertised "
            "in the action's header");

        if( n_read < m_body_size )
        {
            m_recv_buf.discard(m_body_size-n_read);
            TheMessageHub::instance().publish(
                WARNING_MESSAGE,
                "Deserializaton consumed %d bytes less than advertised",
                m_body_size-n_read);
        }
        else
            TheMessageHub::instance().publish(
                WARNING_MESSAGE,
                "Deserialization consumed %d bytes more than advertised", 
                n_read-m_body_size);
    }
}


void AuthenticatedState::on_activating_state() throw()
{
    TheActionRouter::instance().set_action_handler(
        get_connection()->get_id(), 
        boost::bind(
            &AuthenticatedState::handle_outbound_action, this, _1, _2) );
}


void AuthenticatedState::on_state_active() throw()
{
    notify_connection(ON_CLIENT_AUTHENTICATED);
}


void AuthenticatedState::on_state_inactive() throw()
{
    TheActionRouter::instance().unset_action_handler(
        get_connection()->get_id() );
}


void AuthenticatedState::handle_outbound_action(
    Action *action, SourceID src) throw()
{
    get_connection()->send_action(action);

    // We're done with it, delete it.
    delete action;
}


bool AuthenticatedState::is_big_endian() const throw()
{
    return m_is_big_endian; 
}

