/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_ATL_CLIENT_CONNECTION_HH
#define __PEEKABOT_ATL_CLIENT_CONNECTION_HH


#include <string>
#include <boost/thread/recursive_mutex.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>

#include "../Types.hh"
#include "../ChunkedBuffer.hh"
#include "../Sockets.hh"
#include "../Action.hh"


namespace peekabot
{
namespace atl
{

    // Note: In case on or more events are generated the order in
    // which they fire is usually not specified (e.g. \c ON_CLIENT_DISCONNECTED
    // and \c ON_CLIENT_LOCALLY_DISCONNECTED may fire in any order).
    // The two pairs \c ON_SENDING_ACTION, \c ON_ACTION_SENT and
    // \c ON_RECEIVING_ACTION, \c ON_ACTION_RECEIVED is an expection and do have
    // a well-defined, and obvious, order.
    enum ConnectionEvent
    {
        ON_CLIENT_CONNECTED,
        ON_CLIENT_AUTHENTICATED,
        ON_CLIENT_DISCONNECTED,

        //ON_SENDING_ACTION,
        //ON_ACTION_SENT,
        //ON_RECEIVING_ACTION,
        //ON_ACTION_RECEIVED,
        ON_IDLE_SENDING,
        ON_BUSY_SENDING
        //ON_IDLE_RECEIVING,
        //ON_BUSY_RECEIVING
    };



    //
    // Forward declarations
    //
    class ClientConnection;
    class ActionTransporter;



    /**
     * \brief A client connection's state information and accompanying 
     * procedures encapsulated in an object. Plugs in to \c ClientConnection.
     *
     * The current \c ConnectionState of a \c ClientConnection can be varied
     * to transparently change its behaviour.
     *
     * \e Note: This is an application of the \e State pattern.
     */
    class ConnectionState
    {
        /**
         * \brief The connection to which the state belongs.
         *
         * \invariant <tt>m_conn != 0</tt>
         */
        ClientConnection *m_conn;

    protected:
        /**
         * \brief Get the connection to which the state belongs.
         *
         * \invariant The return value is always a valid, non-null pointer.
         *
         * \return A pointer to the connection.
         */
        ClientConnection *get_connection() throw();

        /**
         * \brief Get the connection to which the state belongs.
         *
         * \invariant The return value is always a valid, non-null pointer.
         *
         * \return A pointer to the connection.
         */
        const ClientConnection *get_connection() const throw();

        /**
         * \brief Notify the state's connection and any of its associated
         * observers that the specified event has occurred.
         *
         * \param event The type of event that has occurred.
         */
        void notify_connection(ConnectionEvent event) throw();

        /**
         * \brief Set the state's connection's proxy version.
         *
         * \param major The client proxy's major version.
         * \param minor The client proxy's minor version.
         * \param revision The client proxy's version revision.
         *
         * \sa \c ClientConnection::set_proxy_version(uint8, uint8, uint8)
         */
        void set_proxy_version(uint8_t major, uint8_t minor, uint8_t rev) throw();

        /**
         * \brief Set the state's connection's protocol version.
         *
         * \param version The client's protocol version.
         */
        void set_protocol_version(uint32_t version) throw();

    public:
        /**
         * \brief Create a new \c ConnectionState for the given 
         * \c ClientConnection.
         *
         * \param connection The connection which to associate the state with.
         * Note that the state's associated connection cannot be changed after 
         * its construction.
         */
        ConnectionState(ClientConnection *connection) : m_conn(connection) {}

        /**
         * \brief An empty but virtual destructor, needed for derivatives.
         */
        virtual ~ConnectionState() {}

        /**
         * \brief Called when there's data available for reading.
         */
        virtual void on_data_available() throw() = 0;

        /**
         * \brief Called approximately once per second.
         *
         * Can be used to, for example, close unauthenticated connections or 
         * calculate interesting, time-dependent statistics.
         *
         * The default implementation does nothing.
         */
        virtual void on_tick(boost::posix_time::time_duration &dt) throw() {}

        /**
         * \brief Called by the state's \c ClientConnection, when the buffer 
         * of outbound data has been exhausted, prompting for potential new 
         * data to send.
         *
         * The default implementation does nothing.
         */
        virtual void on_send_buffers_exhausted() throw() {}

        /**
         * \brief An event callback called prior to activating the state.
         *
         * This method is triggered by 
         * \c ClientConnection::set_state(ConnectionState *). A call to
         * \c on_state_active() on the same object follows.
         *
         * The default implementation does nothing.
         */
        virtual void on_activating_state() throw() {}

        /**
         * \brief Called by the owning \c ClientConnection when the state
         * is made active by a call to 
         * \c ClientConnection::set_state(ConnectionState *).
         *
         * The default implementation does nothing.
         */
        virtual void on_state_active() throw() {}

        /**
         * \brief Called by the directing \c ActionTransporter when the state
         * just before the state is made inactive by a call to 
         * \c ActionTransporter::set_state(ConnectionState *), assuming that
         * that method is called with another state.
         *
         * The default implementation does nothing.
         */
        virtual void on_state_inactive() throw() {}
    };





    /**
     * \brief Provides "low-level" interface for sending and receiving data to 
     * a remote client.
     *
     * The class provides low-level statistics and send buffering. It utilizes
     * \e State objects as an abstraction of the connections state (and it's 
     * handler).
     */
    class ClientConnection
    {
        friend class ActionTransporter;
        friend class ConnectionState;

    public:
        /**
         * \brief Denotes all valid reasons that can be or are given
         * for the client being disconnected. It can be retrieved after
         * a connection has been closed to, for example, generating meaningful
         * information messages.
         *
         * \sa \c get_disconnect_reason(), \c close(DisconnectReason)
         */
        enum DisconnectReason
        {
            /**
             * \brief The connection was closed locally with no special reason
             * given.
             */
            LOCALLY_DISCONNECTED,
            /**
             * \brief The connection was terminated by the client.
             */
            REMOTELY_DISCONNECTED,
            /**
             * \brief The client timed out while authenticating, and was 
             * subsequently dropped.
             */
            AUTHENTICATION_TIMED_OUT,
            /**
             * \brief Erroneous authentication data was received, yielding an
             * unsuccessful authentication leading to the connection's 
             * termination.
             */
            AUTHENTICATION_FAILED,
            /**
             * \brief The server's and remote proxy's network protocols are
             * incompatible.
             */
            INCOMPATIBLE_PROTOCOL,
            /**
             * \brief The server's and the remote proxy's versions are 
             * incompatible.
             */
            INCOMPATIBLE_PROXY
        };


        /**
         * \brief Create a new \c ClientConnection and associate
         * it with the given transporter, file descriptor and client ID.
         *
         * \param transporter The owning \c ActionTransporter.
         * \param socket_fd The client connection's socket file descriptor.
         * \param client_id The client's unique identity to the peekabot server.
         */
        ClientConnection(
            ActionTransporter *transporter, 
            sockets::SocketType socket_fd,
            SourceID client_id) throw();

        /**
         * \brief Takes care of freeing resources and closing its socket 
         * if not already closed.
         */
        virtual ~ClientConnection() throw();

        /**
         * \brief Returns true if the socket to the client has been closed.
         *
         * \return \c true if the connection is closed, \c false otherwise.
         */
        bool is_closed() const throw();

        /**
         * \brief Terminate the connection to the client.
         *
         * \param reason The reason for closing the connection.
         *
         * \sa DisconnectReason
         */
        void close(DisconnectReason reason = LOCALLY_DISCONNECTED) throw();

        /**
         * \brief Get the set reason for the connection to the client being
         * lost.
         *
         * \pre The connection is closed, but has been open.
         * \return The reason given to the disconnect the connection at its 
         * time of termination.
         */
        DisconnectReason get_disconnect_reason() const throw();

        /**
         * \brief Returns the amount of unsent data, pending to be sent.
         *
         * \return The number of unsent bytes in all send buffers.
         */
        size_t get_outbound_count() const throw();

        /**
         * \brief Returns the uptime of the connection.
         *
         * \return The duration for which the connection has been up.
         */
        boost::posix_time::time_duration get_uptime() const throw();

        /**
         * \brief Returns the send and recveive throughput.
         *
         * \return A pair of throughput per second (bytes/second), 
         * <sent/s, recevied/s>.
         */
        std::pair<float, float> get_throughput() const throw();


        /**
         * \brief Return the total number of bytes this connection has sent.
         */
        uint64_t get_bytes_sent() const throw();

        /**
         * \brief Return the total number of bytes this connection has received.
         */
        uint64_t get_bytes_recv() const throw();


        /**
         * \brief Get the client's address (IP).
         *
         * \pre \c is_closed() is \c false.
         *
         * \return The remote client's IP.
         */        
        std::string get_remote_address() const throw();



        /**
         * \brief Request some data to be sent. The call is non-blocking and
         * the data will be sent as soon as possible.
         *
         * There are no guarantees that the data will \e ever be transmitted -
         * if the connection is closed before it can be sent it will be lost.
         * Furthermore, there are no guarnatees on how long it will be before
         * the data can be sent.
         *
         * Internally, the data is buffered (see \c m_send_buf, 
         * \c m_intermediary_buf) and transmitted when the \c on_writable() 
         * event callback is called (by the connection's owning transporter).
         *
         * \c send() will take care of changing the send state of the socket
         * to \c busy. \c on_writable() takes care of resetting the send state
         * to \c idle when no more data is availble for writing.
         *
         * \param buf A pointer to the data to be sent.
         * \param bytes The number of bytes to send from the given buffer.
         */
        void send(const void *buf, size_t bytes) throw();

        /**
         * \brief Consume and send the contents of the entire passed buffer.
         *
         * \post <tt>buf.is_empty() = true</tt>
         *
         * \param buf The data to send.
         */
        void send(ChunkedBuffer &buf) throw();

        /**
         * \brief Serialize and send the specified action.
         *
         * Takes care of everything that is needed, including forming and
         * sending the action header.
         *
         * Does \e not take ownership of the given action.
         *
         * \param action The action to be sent.
         */
        void send_action(Action *action) throw();

        /**
         * \brief Recv some data from the socket. The call is non-blocking.
         *
         * \param buf The buffer which to store received data in. The given 
         * buffer should have storage enough to accommodate \c max_bytes 
         * bytes.
         * \param max_bytes The maximum number of bytes to receive.
         * \return The number of bytes received into the given buffer.
         */
        size_t recv(void *buf, size_t max_bytes) throw();

        /**
         * \brief Recv some data without removing it from the socket data queue.
         * This call is non-blocking.
         *
         * \invariant <tt>peek(b1, n) == peek(b2, n), memcmp(b1,b2) == 0</tt>
         *
         * \param buf The buffer which to store read data in. The given 
         * buffer should have storage enough to accommodate \c max_bytes 
         * bytes.
         * \param max_bytes The maximum number of bytes to read.
         * \return The number of bytes read into the given buffer.
         */
        size_t peek(void *buf, size_t max_bytes) throw();

    
        /**
         * \brief Set the state of the connection.
         *
         * Note that the \c ClientConnection takes ownership of the given
         * state object.
         *
         * \param state The new connection state. Ownership of the state 
         * is transmitted from the caller to the connection!
         *
         * \sa \c ConnectionState, \c AuthenticatedState, 
         * \c UnauthenticatedState
         */
        void set_state(ConnectionState *state) throw();

        /**
         * \brief Return the client's source ID.
         */
        SourceID get_id() const throw();

        
        /**
         * \brief Get the client connection's protocol version.
         */
        uint32_t get_protocol_version() const throw();

        /**
         * \brief Return the client's proxy's version as a string.
         *
         * The returned string has the form <tt>major.minor.revision</tt>.
         */
        std::string get_proxy_version() const throw();

        /**
         * \brief Return the client's proxy's major version.
         */
        uint8_t get_proxy_version_major() const throw();

        /**
         * \brief Return the client's proxy's minor version.
         */        
        uint8_t get_proxy_version_minor() const throw();

        /**
         * \brief Return the client's proxy's version revision.
         */
        uint8_t get_proxy_version_revision() const throw();


        /**
         * \brief Return the connection's owning and directing
         * \c ActionTransporter.
         */
        ActionTransporter *get_transporter() throw();

        
        /**
         * \brief Return the connection's owning and directing
         * \c ActionTransporter.
         */
        const ActionTransporter *get_transporter() const throw();

        void toggle_low_latency_mode(bool enable) throw();

        bool low_latency_mode_enabled() const throw();


    private:
        /**
         * \brief An event handler called by the owning \c ActionTransporter
         * approximately every second.
         *
         * It is used to generate \e tick events for the current conenction 
         * state and calculate throughput statistics.
         *
         * \param dt The time interval since the method was last called.
         */
        void on_tick(boost::posix_time::time_duration &dt) throw();

        /**
         * \brief Event callback called when there's data available for recveivng
         *        on the socket.
         *
         * Used to recv actions and handshaking data.
         */
        void on_data_available() throw();


        sockets::SocketType get_socket_fd() const throw();


        /**
         * \brief Toggle the socket as active writing data.
         *
         * Causes it to receive \e on_writable events from its owning
         * \c ActionTransporter.
         */
        void set_active_sending() throw();

        /**
         * \brief Toggle the socket as idle, not writing data.
         *
         * Causes it not to receive \e on_writable events from its owning
         * \c ActionTransporter.
         */
        void set_idle_sending() throw();


        /**
         * \brief Called when the socket is in active state (see 
         * \c set_send_active()) and the socket is ready for writing.
         *
         * This method, in turn, raises the same event on the active 
         * connection state.
         *
         * \sa \c ActionTransporter
         */
        void on_writable() throw();


        //ConnectionState *use_state() throw();




    private:
        /**
         * \brief Notify the directing \c ActionTransporter of connection events.
         *
         * This is used to update a variety of internal information in the 
         * \c ActionTransporter as well as to generate notifications to registered
         * observers.
         */
        void notify(ConnectionEvent event) throw();


        void set_protocol_version(uint32_t version) throw();

        void set_proxy_version(
            uint8_t major, uint8_t minor, uint8_t revision) throw();


        /**
         * \brief Called by the connection's \e state or internally when
         * it's been detected that the socket has been closed.
         *
         * Takes care of generating notifications to observers and
         * updating it's internal state to reflect that the connection
         * has been closed.
         *
         * \param remotely_closed \c true should be sent if the connection
         * was closed by the client, \c false if it was closed locally.
         */
        void connection_closed(bool remotely_closed) throw();


    private:
        /**
         * \brief A small, uncomplicated intermediary send buffer, with no 
         * wrap-around.
         */
        struct IntermediaryBuffer
        {
            IntermediaryBuffer(int capacity) 
                : m_capacity(capacity), m_rptr(0), m_wptr(0)
            {
                m_buf = new uint8_t[m_capacity];
            }

            ~IntermediaryBuffer()
            {
                delete[] m_buf;
            }

            uint8_t *m_buf;

            const size_t m_capacity;

            /**
             * \brief The index of the recv head into the buffer.
             *
             * \invariant m_rptr <= m_capacity
             */
            size_t m_rptr;

            size_t m_wptr;

            bool is_empty() const throw()
            {
                return m_wptr-m_rptr == 0;
            }

            size_t get_bytes_available() const throw()
            {
                return m_wptr-m_rptr;
            }

            size_t get_capacity_available() const throw()
            {
                return m_capacity-m_wptr;
            }

            void clear() throw()
            {
                m_rptr = m_wptr = 0;
            }

            const void *get_read_ptr() const throw()
            {
                return m_buf+m_rptr;
            }
        };


        /**
         * \brief An intermediary buffer that serves as the source for writing
         * actual data to the socket. It's feed from \c m_send_buf whenever it's
         * empty.
         */
        IntermediaryBuffer m_intermediary_buf;


        /**
         * \brief A buffer to hold the outgoing data.
         *
         * Data is transfered from this buffer to an intermediary buffer,
         * \c m_intermediary_buf, for sending.
         */
        ChunkedBuffer m_send_buf;


        /**
         * \brief Governs access to all variables not synchronized by any other
         * mutex.
         *
         * \sa \c m_stat_mutex
         */
        mutable boost::recursive_mutex m_mutex;

        /**
         * \brief Governs access to all statistics variables.
         */
        mutable boost::recursive_mutex m_stat_mutex;



        /**
         * \brief The connection's socket, used to actually communicate with
         * the remote proxy at the client-side.
         */
        const sockets::SocketType m_socket_fd;

        /**
         * \brief A pointer to the \c ActionTransporter that owns and
         * directs this particular \c ClientConnection.
         */
        ActionTransporter *m_transporter;
    
        /**
         * \brief The connection's current state, encapsulated in an object.
         *
         * Requests to this objet is made to handle state specific details,
         * such as handling incoming data.
         *
         * This allows for different states to be handled easily, without 
         * large if-blocks to handle states and using separated code.
         * Code separation in itself is a huge structural benefit.
         */
        ConnectionState *m_state;

        /**
         * \brief The client's source ID.
         */
        const SourceID m_id;

        /**
         * \brief Keeps track of whether the connection has been closed
         * or is in an open, active state.
         *
         * The life-time of the connection outlives their period of being open
         * by a small time, thus we need this variable to track if the 
         * connection has been closed but not yet deleted - see the
         * \c ActionTransporter class for information on when the connection
         * objects are deleted (when they've been closed).
         */
        bool m_is_closed;

        /**
         * \brief If the connection has been \e closed, this contains
         * the disconnect reason, otherwise it is undefined.
         *
         * Access is governed by the \c m_mutex mutex.
         *
         * \sa close(DisconnectReason)
         */
        DisconnectReason m_disconnect_reason;


        //
        // ----- Statistics -----
        //

        /**
         * \brief The number of bytes/s sent.
         *
         * Access is governed by the \c m_stat_mutex mutex.
         */
        float m_sent_throughput;

        /**
         * \brief The number of bytes/s received.
         *
         * Access is governed by the \c m_stat_mutex mutex.
         */
        float m_recv_throughput;

        /**
         * \brief The total number of bytes sent.
         *
         * Access is governed by the \c m_stat_mutex mutex.
         */
        uint64_t m_bytes_sent;

        /**
         * \brief The total number of bytes received.
         *
         * Access is governed by the \c m_stat_mutex mutex.
         */
        uint64_t m_bytes_recv;

        /**
         * \brief The number of bytes sent since the last tick-event, 
         * used to calculate throughput.
         *
         * Access is governed by the \c m_stat_mutex mutex.
         *
         * \sa get_throughput(), \c m_stat_mutex mutex
         */
        unsigned int m_sent_since_last_tick;

        /**
         * \brief The number of bytes received since the last tick-event, 
         * used to calculate throughput.
         *
         * Access is governed by the \c m_stat_mutex mutex.
         *
         * \sa get_throughput(), \c m_stat_mutex mutex
         */
        unsigned int m_recv_since_last_tick;


        /**
         * \brief Stores the time of the connection's establishment.
         *
         * Used to determine the connection's uptime.
         */
        const boost::posix_time::ptime m_up_since;


        //
        // --------- Client/proxy information ----------
        //


        /**
         * \brief The client proxy's protocol version.
         */
        uint32_t m_protocol_version;

        /**
         * \brief The major version of the client's peekabot proxy.
         */
        uint8_t m_proxy_version_major;

        /**
         * \brief The minor version of the client's peekabot proxy.
         */
        uint8_t m_proxy_version_minor;

        /**
         * \brief The revision of the client's peekabot proxy.
         */
        uint8_t m_proxy_version_revision;

        /**
         * \brief Stores address (IP) of the remote end of the connection,
         * i.e. the client.
         *
         * The value is undefined if the remote address could not be determined.
         *
         * Set in the constructor, and manipulated \e only there. Can thus be 
         * safely accessed without the need for synchronization.
         */
        std::string m_remote_address;

        bool m_low_latency_mode;
    };


}
}


#endif // __PEEKABOT_ATL_CLIENT_CONNECTION_HH
