/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_ATL_CONNECTION_STATES_HH
#define __PEEKABOT_ATL_CONNECTION_STATES_HH


#include <queue>
#include <boost/function.hpp>
#include <boost/thread/recursive_mutex.hpp>

#include "../Types.hh"
#include "../Protocol.hh"
#include "../Action.hh"
#include "../ChunkedBuffer.hh"
#include "ClientConnection.hh"


namespace peekabot
{

class Action;

namespace atl
{


    /**
     * \brief The state handler for \e unauthenticated connections.
     *
     * Handles authentication and other handshaking matters and then
     * changes the state to \c AuthenticatedState, or closes the connection
     * if the authentication failed.
     *
     * <b>Handshake procedure:</b>
     * \li Server/proxy sends: A byte describing byte order (uint8), 0=little endian, 1=big endian.
     * \li Server/proxy sends: A unique, static ID: 0x7065656B (uint32).
     * \li Server/proxy sends: The protocol version (uint32).
     * \li Server/proxy sends: peekabot version (uint32_t), coded as major<<16 | minor<<8 | revision.
     * The 8 most significant bits are ignored and must be zero.
     * \li Proxy sends: Enable/disable low latency mode (uint8).
     * \li Server/proxy sends: Authentication result (uint9_t) 0=ok, non-zero means connection refused. 
     * 1=no more connections allowed, 2=incompatible peekabot version, 3=incompatible protocol version, 
     * 4=unexpected data received
     *
     * There's also a time out on unauthenticated connections, typically 10 seconds.
     */
    class UnauthenticatedState : public ConnectionState
    {
    protected:
        virtual void part_one_read() throw();

        virtual void part_two_read() throw();


        bool m_is_big_endian;
        
        uint32_t m_unique_id;

        uint32_t m_protocol_version;

        uint32_t m_peekabot_version;

        uint8_t m_enable_low_latency_mode;

        uint8_t m_auth_res_local, m_auth_res_remote;

        bool m_close_if_outbound_empty;


    protected:
        typedef boost::function0<void> ReadCallback;

        void read(void *dest, size_t n) throw();

        void read(void *dest, size_t n, ReadCallback callback) throw();

    private:
        static void null_callback() throw() {}

        struct ReadRecord
        {
            void         *m_dest;
            size_t        m_size;
            size_t        m_n_read;
            ReadCallback  m_callback;
        };

        std::queue<ReadRecord> m_read_queue;

    public:
        UnauthenticatedState(ClientConnection *connection) throw();

        virtual ~UnauthenticatedState() throw();

        /**
         * \brief Receive and deal appropriately with incoming handshake data.
         */
        virtual void on_data_available() throw();

        /**
         * \brief Handle time-outs.
         */
        virtual void on_tick(boost::posix_time::time_duration &dt) throw();

        /**
         * \brief The state's become active, i.e. we're need to get 
         * authenticated. It's a good idea to send that authentication data now!
         */
        virtual void on_state_active() throw();
    };







    /**
     * \brief The state handler for \e authenticated connections.
     *
     * It's duties include serializing and sending outbound actions
     * and receiving and deserializing incoming actions. Received actions
     * are then forwarde to the \c ActionRouter.
     *
     * <b>Communication protocol:</b>
     *
     * The protocol once authenticated is very simple as handles only sending
     * and receiving \c Actions. Sending and receiving data is completely 
     * decoupled.
     *
     * The data transmitted for each \c Action consists, in the given order, of:
     * \li An action header
     * \li The action body
     *
     * Note, however, that this protocol may chage depending on the protocol
     * version and extraneous data sent, such as checksums.
     */
    class AuthenticatedState : public ConnectionState
    {
    public:
        // is_big_endian is propagated from UnauthenticatedState
        AuthenticatedState(
            ClientConnection *connection, bool is_big_endian);
        // 1. instruct TheActionRouter to route this client's
        //    actions to the handle_outbound_action() method of this class.


        //virtual void on_tick() throw();

        /**
         * \brief When the connection has finished sending the previous action,
         * instruct it to serialize and send another one, if available.
         */
        //virtual void on_send_buffers_exhausted() throw();
            /*{
            if( queue not empty )
                send_action(queue.top);
        }*/

        /**
         * \brief Handle incoming data.
         */
        virtual void on_data_available() throw();
            /*{
            //state_handler->on_data_available(this);

            if( state = unauth )
            {
                send(ababa);
                ...
                }
            if( state = header )
            {
                recv until m_recv_buf.size() == sizeof(header)
                construct header from recv_buf
                state = body
                }
            if( state = body )
            {
                recv until m_recv_buf.size() == header.body_size
                construct action
                state = header
                }
                }*/

        virtual void on_activating_state() throw();

        /**
         * \brief Fire notifications and set up routing of the client's
         * outgoing actions to this state's \c handle_outbound_action() method.
         */
        virtual void on_state_active() throw();
        // 0. TheActionTransporter->set_action_handler(handle_outbound_action)
        // 1. conn->notify(ON_CLIENT_AUTHENTICATED)


        /**
         * \brief Stop routing actions designated for this client's
         * connection to this state object.
         */
        virtual void on_state_inactive() throw();
        // 1. TheActionTransporter->unset_action_handler(for this conn)

    private:
        /**
         * \brief Handles actions passed by the router.
         */
        void handle_outbound_action(Action *action, SourceID src) throw();
        // 1. m_queue.insert(action)
        // 2. connecton->transporter->wake()
        // 3. if queue was empty, connection->send_action(queue.pop())

        /**
         * \brief Returns true if the remote client is sending data
         * in little endian byte order.
         */
        bool is_big_endian() const throw();

        void handle_data();

        void consume_action();


        /**
         * \brief Send an action.
         */
        //void send_action(Action *action) throw();
        // 0. notify(SENDING_ACTION)
        // 1. serialize
        // 2. construct & send header
        // 3. send body
        // 4. ???


    private:
        enum ReadState
        {
            RS_HEADER_1,
            RS_HEADER_2,
            RS_BODY
        };

        ReadState m_read_state;

        /**
         * \brief A mutex for synchronizing access to \c m_queue.
         */
        //boost::recursive_mutex m_queue_mutex;

        /**
         * \brief Stores actions to be sent.
         * 
         * The actions are taken off one by one from this queue when the 
         * previous action has been sent.
         */
        //std::queue<Action *> m_queue;

        /**
         * \brief Used to read the header portion of an incoming, serialized
         * \c Action.
         */
        uint32_t m_body_size;

        uint8_t m_ctrl_byte;

        uint32_t m_uncomp_size;

        /**
         * \brief A buffer that stores body data of incoming actions.
         *
         * Once a complete action has been read, the buffer is used to 
         * deserialize and reconstruct the transmitted action.
         */
        ChunkedBuffer m_recv_buf;

        /**
         * \brief Set to true if the remote client is sending data in little
         * endian byte order.
         */
        bool m_is_big_endian;
    };


}
}

#endif // __PEEKABOT_ATL_CONNECTION_STATES_HH
