/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <boost/test/unit_test.hpp>

#include "../ActionRouter.hh"
#include "../../Action.hh"

using namespace peekabot;
using namespace peekabot::atl;


typedef uint16_t ActionID;


// A small helper class to make testing easier
class TestAction : public Action
{
public:
    TestAction() throw() {}
    
    virtual ~TestAction() throw() {}
    
    virtual Action *clone() const
    {
        // we don't need this
        return 0;
    }
    
    virtual void translate_ids(const ObjectIDTranslator *translator)
    {
        // we don't need this
    }

private:
    virtual void save(SerializationInterface &) const {}

    virtual void load(DeserializationInterface &) {}
};



template<int T>
struct Handler
{
    static bool m_fired;

    static void handle(
        Action *action, 
        SourceID source) throw()
    {
        m_fired = true;
    }

    static void reset() throw()
    {
        m_fired = false;
    }

    static bool has_fired() throw()
    {
        bool tmp = m_fired;
        reset();
        return tmp;
    }
};

template<int T> bool Handler<T>::m_fired = false;





BOOST_AUTO_TEST_SUITE( ActionRouterTests );


BOOST_AUTO_TEST_CASE( set_action_handler )
{
    ActionRouter router;

    BOOST_CHECK( 
        router.set_action_handler(
            SourceID(SourceID::CORE_ID, SourceID::SERVER), 
            Handler<0>::handle) == false );

    BOOST_CHECK( 
        router.set_action_handler(
            SourceID(SourceID::CORE_ID, SourceID::SERVER), 
            Handler<0>::handle) == true );

    BOOST_CHECK( 
        router.set_action_handler(
            SourceID(SourceID::GUI_ID, SourceID::SERVER), 
            Handler<0>::handle) == false );

    BOOST_CHECK( 
        router.set_action_handler(
            SourceID(SourceID::GUI_ID, SourceID::SERVER), 
            Handler<0>::handle) == true );

    BOOST_CHECK( 
        router.set_action_handler(
            SourceID(1, SourceID::CLIENT), 
            Handler<0>::handle) == false );

    BOOST_CHECK( 
        router.set_action_handler(
            SourceID(2, SourceID::CLIENT), 
            Handler<0>::handle) == false );
}


BOOST_AUTO_TEST_CASE( unset_action_handler )
{
    ActionRouter router;

    BOOST_CHECK( 
        router.set_action_handler(
            SourceID(SourceID::CORE_ID, SourceID::SERVER), 
            Handler<0>::handle) == false );

    // there's no handler registered for (client,5)
    BOOST_CHECK( 
        router.unset_action_handler(
            SourceID(5, SourceID::CLIENT)) == true );

    BOOST_CHECK( 
        router.unset_action_handler(
            SourceID(SourceID::CORE_ID, SourceID::SERVER)) == false );

    // the handler for this source was already removed
    BOOST_CHECK( 
        router.unset_action_handler(
            SourceID(SourceID::CORE_ID, SourceID::SERVER)) == true );
}


BOOST_AUTO_TEST_CASE( post_action )
{
    ActionRouter router;

    Handler<0>::reset();
    Handler<1>::reset();
    Handler<2>::reset();
    Handler<3>::reset();

    BOOST_CHECK( 
        router.set_action_handler(
            SourceID(SourceID::CORE_ID, SourceID::SERVER), 
            Handler<0>::handle) == false );

    BOOST_CHECK( 
        router.set_action_handler(
            SourceID(SourceID::GUI_ID, SourceID::SERVER), 
            Handler<1>::handle) == false );

    BOOST_CHECK( 
        router.set_action_handler(
            SourceID(100, SourceID::CLIENT), 
            Handler<2>::handle) == false );

    BOOST_CHECK( 
        router.set_action_handler(
            SourceID(101, SourceID::CLIENT), 
            Handler<3>::handle) == false );

    TestAction action;

    router.post_action(
        &action, 
        SourceID(SourceID::CORE_ID, SourceID::SERVER),
        SourceID(5, SourceID::CLIENT));

    BOOST_CHECK( Handler<0>::has_fired() == false );
    BOOST_CHECK( Handler<1>::has_fired() == false );
    BOOST_CHECK( Handler<2>::has_fired() == false );
    BOOST_CHECK( Handler<3>::has_fired() == false );

    router.post_action(
        &action,
        SourceID(SourceID::CORE_ID, SourceID::SERVER),
        SourceID(100, SourceID::CLIENT));

    BOOST_CHECK( Handler<0>::has_fired() == false );
    BOOST_CHECK( Handler<1>::has_fired() == false );
    BOOST_CHECK( Handler<2>::has_fired() == true );
    BOOST_CHECK( Handler<3>::has_fired() == false );

    router.post_action(
        &action,
        SourceID(5, SourceID::CLIENT),
        SourceID(SourceID::CORE_ID, SourceID::SERVER));

    BOOST_CHECK( Handler<0>::has_fired() == true );
    BOOST_CHECK( Handler<1>::has_fired() == false );
    BOOST_CHECK( Handler<2>::has_fired() == false );
    BOOST_CHECK( Handler<3>::has_fired() == false );

    // unset action handler and test the same ation again
    BOOST_CHECK(
        router.unset_action_handler(
            SourceID(SourceID::CORE_ID, SourceID::SERVER)) == false );

    router.post_action(
        &action,
        SourceID(5, SourceID::CLIENT),
        SourceID(SourceID::CORE_ID, SourceID::SERVER));

    BOOST_CHECK( Handler<0>::has_fired() == false );
    BOOST_CHECK( Handler<1>::has_fired() == false );
    BOOST_CHECK( Handler<2>::has_fired() == false );
    BOOST_CHECK( Handler<3>::has_fired() == false );
}



BOOST_AUTO_TEST_CASE( observers )
{
    class TestObserver : public RouterObserver
    {
    public:
        virtual void on_notify(RouterEvent event, ActionRouter *router) throw()
        {
            m_fired = true;
        }

        bool m_fired;
    };



    ActionRouter router;
    TestObserver obs;

    BOOST_CHECK( router.register_observer(ON_ACTION_ROUTED, &obs) == false );

    // we shouldn't be able to register the same observer twice for 
    // the same event
    BOOST_CHECK( router.register_observer(ON_ACTION_ROUTED, &obs) == true );

    BOOST_REQUIRE( 
        router.set_action_handler(
            SourceID(SourceID::CORE_ID, SourceID::SERVER), 
            Handler<0>::handle) == false );

    TestAction action;

    router.post_action(
        &action,
        SourceID(5, SourceID::CLIENT),
        SourceID(SourceID::CORE_ID, SourceID::SERVER));
    BOOST_CHECK( obs.m_fired == true );
    obs.m_fired = false;

    // try to remove non-existant observer
    BOOST_CHECK( router.deregister_observer(ON_ACTION_DROPPED, &obs) == true );
    BOOST_CHECK( router.deregister_observer(ON_ACTION_DROPPED, 0) == true );
}


BOOST_AUTO_TEST_SUITE_END();
