/*
 * Copyright Staffan Gimåker 2007-2008.
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */

#ifndef __PEEKABOT_CLIENT_OPERATION_RESULT_HH
#define __PEEKABOT_CLIENT_OPERATION_RESULT_HH


#include <stdexcept>
#include <boost/thread/mutex.hpp>
#include "any_dl.hpp" // GCC/Boost.Any workaround

#include "OperationStatus.hh"


namespace peekabot
{
    namespace client
    {
        /**
         * \internal
         *
         * \brief Provides internal storage for and interaction with results.
         *
         *
         * All methods lock the OperationResult but functionality inherited 
         * from OperationStatus can still be used simultaneously.
         *
         * \sa Result, OperationStatus
         */
        class OperationResult : public OperationStatus
        {
        public:
            OperationResult() throw();

            virtual ~OperationResult() throw();

            /**
             * \brief Set the result data contained in the object.
             *
             * This method is called when the result is received from the 
             * server, by the ActionResultContainer that carries it.
             *
             * \pre In order to ensure correct behaviour, 
             * Status::set_outcome(OperationOutcome, const std::string &)
             * must not have been called prior to calling this method.
             *
             * \param result The result of the operation.
             */
            void set_result(const boost::any &result) throw();

            /**
             * \brief Return a copy of the stored result.
             *
             * \pre get_outcome() == OPERATION_SUCCEEDED, otherwise an
             * exception will be thrown.
             *
             * \throw std::logic_error Thrown if 1) the operation failed or 2)
             * the outcome operation is still pending.
             */
            boost::any get_result() const throw(std::logic_error);


        private:
            /**
             * \brief Governs \e all (read and write) access to \e any member.
             */
            mutable boost::mutex m_result_mutex;

            /**
             * \brief The result data.
             */
            boost::any m_result;

            /**
             * \brief Set to \c true when the result data member (m_result) has
             * been set. Used for error checking purposes only.
             */
            bool m_result_is_set;
        };
    }
}


#endif // __PEEKABOT_CLIENT_OPERATION_RESULT_HH
