/*
 * Copyright Staffan Gimåker 2008-2009.
 *
 * ---
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */

#ifndef PEEKABOT_CLIENT_LABEL_PROXY_HH_INCLUDED
#define PEEKABOT_CLIENT_LABEL_PROXY_HH_INCLUDED


#include "ScalableProxy.hh"


namespace peekabot
{
    namespace client
    {
        /**
         * \brief Base class for LabelProxy.
         */
        class LabelProxyBase : public ScalableProxyBase
        {
        public:
            LabelProxyBase();

            LabelProxyBase(const LabelProxyBase &p);

            /**
             * \brief Set the label's text.
             */
            DelayedDispatch set_text(const std::string &text);

            /**
             * \brief Change the label's text alignment.
             *
             * The default is left-aligned text.
             */
            DelayedDispatch set_alignment(TextAlignment alignment);
        };


        /**
         * \brief Proxy class used to add and manipulate orientable text labels.
         *
         * Text labels provide a convinient way to add simple annotations to
         * a visualization - marking rooms in a map according to name, for 
         * example.
         */
        class LabelProxy : public LabelProxyBase
        {
        public:
            LabelProxy();

            LabelProxy(const LabelProxyBase &p);

            LabelProxy &operator=(const LabelProxy &p) throw();

            LabelProxy &operator=(const LabelProxyBase &p) throw();

            /**
             * \copydoc ScalableProxy::assign(const ObjectProxyBase&)
             */
            DelayedDispatch assign(const ObjectProxyBase &p);

            /**
             * \copydoc ObjectProxy::assign(PeekabotClient&,const std::string&)
             */
            DelayedDispatch assign(
                PeekabotClient &client,
                const std::string &path);

            /**
             * \copydoc 
             * ObjectProxy::assign(const ObjectProxyBase&,const std::string&)
             */
            DelayedDispatch assign(
                const ObjectProxyBase &parent, 
                const std::string &rel_path);

            /**
             * \brief Add a label to the scene.
             *
             * Adds a label to the scene at the given path and assigns 
             * the proxy to point to it. If an object already exists at the 
             * given path the outcome is governed by \a conflict_policy.
             *
             * \param path The path where the object will be added, including 
             * the name of the object.
             * \param filename The (server-side) path to the label file.
             * \param conflict_policy Determines how name conflicts are handled.
             */
            DelayedDispatch add(
                PeekabotClient &client,
                const std::string &path,
                NameConflictPolicy conflict_policy = AUTO_ENUMERATE_ON_CONFLICT);

            /**
             * \brief Add a label to the scene.
             *
             * Adds a label with the given name under the the object referred 
             * to by \a parent and assigns the proxy to point to it. If an 
             * object already exists at the given path the outcome is governed 
             * by \a conflict_policy.
             *
             * \param parent The parent under which the object will be added.
             * \param name The name to assign to the created object.
             * \param conflict_policy Determines how name conflicts are handled.
             */
            DelayedDispatch add(
                const ObjectProxyBase &parent, 
                const std::string &name,
                NameConflictPolicy conflict_policy = AUTO_ENUMERATE_ON_CONFLICT);
        };
    }
}


#endif // PEEKABOT_CLIENT_LABEL_PROXY_HH_INCLUDED
