/*
 * Copyright Staffan Gimåker 2008-2009.
 *
 * ---
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */

#include <boost/thread/mutex.hpp>

#include "ObjectProxy.hh"
#include "../PeekabotClient.hh"
#include "../../IDAllocator.hh"

#include "../../actions/RegisterPseudonym.hh"
#include "../../actions/DeregisterPseudonym.hh"
#include "../../actions/SetTransformation.hh"
#include "../../actions/SetPosition.hh"
#include "../../actions/SetOrientation.hh"
#include "../../actions/SetPose.hh"
#include "../../actions/SetOpacity.hh"
#include "../../actions/SetVisibility.hh"
#include "../../actions/SetLayer.hh"
#include "../../actions/SetColor.hh"
#include "../../actions/SetName.hh"
#include "../../actions/SetMetaName.hh"
#include "../../actions/RemoveObject.hh"
#include "../../actions/ClearChildren.hh"
#include "../../actions/LoadScene.hh"
#include "../../actions/GetTransformation.hh"
#include "../../actions/GetPosition.hh"
#include "../../actions/GetOrientation.hh"
#include "../../actions/GetChildren.hh"
#include "../../actions/SetRotation.hh"
#include "../../actions/RearrangeObject.hh"


using namespace peekabot;
using namespace peekabot::client;


namespace
{
    /*
     * A simple locking-decorator for another IDAllocator class, making it 
     * thread-safe.
     */
    template<
        class IDType, 
        template<class> class IDAllocatorImpl = DefaultIDAllocator>
    class ThreadSafeIDAllocator : public IDAllocator<IDType>
    {
    public:
        virtual IDType allocate()
        {
            boost::mutex::scoped_lock lock(m_mutex);
            return m_backend.allocate();
        }

        virtual void release(IDType id)
        {
            boost::mutex::scoped_lock lock(m_mutex);
            m_backend.release(id);
        }

    private:
        IDAllocatorImpl<IDType> m_backend;
        boost::mutex m_mutex;
    };

    ThreadSafeIDAllocator<ObjectID> pseudonym_allocator;
}

//
// ------------------- ObjectProxyBase implementation ---------------------
//

ObjectProxyBase::ObjectProxyBase()
{
}


ObjectProxyBase::ObjectProxyBase(const ObjectProxyBase &p)
    : PeekabotProxyBase(p),
      m_pseudonym(p.get_pseudonym())
{
}


ObjectProxyBase::~ObjectProxyBase()
{
    // Reset pseudonym (will release it if no one else is using it)
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    if( m_pseudonym )
    {
        // Pseudonym previously assigned, release it?
        if( m_pseudonym.unique() )
        {
            pseudonym_allocator.release(*m_pseudonym);
            dispatch_action(new DeregisterPseudonym(*m_pseudonym), 0);
        }

        m_pseudonym.reset();
    }
}


bool ObjectProxyBase::operator==(const ObjectProxyBase &p) const throw()
{
    return unchecked_get_client_impl() == p.unchecked_get_client_impl() &&
        get_object_id() == p.get_object_id();
}


bool ObjectProxyBase::operator!=(const ObjectProxyBase &p) const throw()
{
    return !(p == *this);
}


ObjectID ObjectProxyBase::get_object_id() const throw(std::logic_error)
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);
    if( !is_assigned() )
        throw std::logic_error("The object proxy is unassigned, "
                               "i.e. not yet tied to a remote object");
    assert( m_pseudonym );
    return *m_pseudonym;
}


ObjectID ObjectProxyBase::get_object_id(const ObjectProxyBase &p) 
    throw(std::logic_error)
{
    return p.get_object_id();
}


void ObjectProxyBase::unchecked_assign(
    boost::shared_ptr<ClientImpl> client, boost::shared_ptr<ObjectID> id)
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    // NOTE: Order matters here - the pseudonym must be released before we
    // change the client (since it might be a different client!)
    if( m_pseudonym )
    {
        assert( client );

        // Pseudonym previously assigned, release it?
        boost::shared_ptr<ObjectID> tmp = m_pseudonym;
        m_pseudonym.reset();

        if( tmp.unique() )
        {
            pseudonym_allocator.release(*tmp);
            dispatch_action(new DeregisterPseudonym(*tmp), 0);
        }
    }

    set_client_impl(client);
    m_pseudonym = id;
}


void ObjectProxyBase::unchecked_assign(const ObjectProxyBase &other)
{
    boost::shared_ptr<ClientImpl> client;
    boost::shared_ptr<ObjectID> pid;

    {
        boost::recursive_mutex::scoped_lock lock(other.m_mutex);
        client = other.unchecked_get_client_impl();
        pid = other.get_pseudonym();
    }

    unchecked_assign(client, pid);
}


boost::shared_ptr<ObjectID> ObjectProxyBase::get_pseudonym() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);
    return m_pseudonym;
}


boost::shared_ptr<ObjectID> ObjectProxyBase::get_pseudonym(
    const ObjectProxyBase &p) throw()
{
    return p.get_pseudonym();
}


boost::shared_ptr<ObjectID> ObjectProxyBase::allocate_pseudonym()
{
    boost::shared_ptr<ObjectID> ret(new ObjectID);
    *ret = pseudonym_allocator.allocate();
    return ret;
}


//


DelayedDispatch ObjectProxyBase::set_transformation(
    const Matrix4f &mat,
    CoordinateSystem coord_sys)
{
    return DelayedDispatch(
        get_client_impl(),
        new SetTransformation(get_object_id(), mat, coord_sys));
}


DelayedDispatch ObjectProxyBase::set_position(
    float x, float y, float z,
    CoordinateSystem coord_sys)
{
    return DelayedDispatch(
        get_client_impl(),
        new SetPosition(get_object_id(), Vector3f(x,y,z), coord_sys));
}


DelayedDispatch ObjectProxyBase::set_orientation(
    float vx, float vy, float vz,
    CoordinateSystem coord_sys)
{
    return DelayedDispatch(
        get_client_impl(),
        new SetOrientation(get_object_id(), Vector3f(vx,vy,vz), coord_sys));
}


DelayedDispatch ObjectProxyBase::set_rotation(
    float yaw, float roll, float pitch,
    CoordinateSystem coord_sys)
{
    return DelayedDispatch(
        get_client_impl(),
        new SetRotation(get_object_id(), yaw, roll, pitch, coord_sys));
}


DelayedDispatch ObjectProxyBase::set_pose(
    float x, float y, float z,
    float yaw, float roll, float pitch,
    CoordinateSystem coord_sys)
{
    return DelayedDispatch(
        get_client_impl(),
        new SetPose(get_object_id(),
                    Vector3f(x, y, z), yaw, roll, pitch, coord_sys));
}


DelayedDispatch ObjectProxyBase::set_opacity(float opacity, bool absolute)
{
    return DelayedDispatch(
        get_client_impl(),
        new SetOpacity(get_object_id(), Opacity(absolute, opacity)));
}


DelayedDispatch ObjectProxyBase::set_visibility(bool visible)
{
    return DelayedDispatch(
        get_client_impl(),
        new SetVisibility(get_object_id(), visible));
}


DelayedDispatch ObjectProxyBase::set_layer(
    unsigned int layer, bool recursive)
{
    if( layer < 1 || layer > 16 )
        throw std::range_error("Invalid layer number");

    return DelayedDispatch(
        get_client_impl(),
        new SetLayer(get_object_id(), layer-1, recursive));
}


DelayedDispatch ObjectProxyBase::set_color(
    float r, float g, float b, bool recursive)
{
    return DelayedDispatch(
        get_client_impl(),
        new SetColor(get_object_id(), RGBColor(r,g,b), recursive));
}


DelayedDispatch ObjectProxyBase::set_name(const std::string &name)
{
    return DelayedDispatch(
        get_client_impl(),
        new SetName(get_object_id(), name));
}


DelayedDispatch ObjectProxyBase::set_meta_name(const std::string &meta_name)
{
    return DelayedDispatch(
        get_client_impl(),
        new SetMetaName(get_object_id(), meta_name));
}


DelayedDispatch ObjectProxyBase::remove()
{
    return DelayedDispatch(
        get_client_impl(),
        new RemoveObject(get_object_id()));
}


DelayedDispatch ObjectProxyBase::clear()
{
    return DelayedDispatch(
        get_client_impl(),
        new ClearChildren(get_object_id()));
}


DelayedDispatch ObjectProxyBase::rearrange(
    const ObjectProxyBase &new_parent,
    bool retain_world_pose,
    NameConflictPolicy conflict_policy) const
{
    return DelayedDispatch(
        get_client_impl(),
        new RearrangeObject(
            PathIdentifier(get_object_id()),
            PathIdentifier(get_object_id(new_parent)),
            retain_world_pose, conflict_policy));
}


DelayedDispatch ObjectProxyBase::load_scene(
    const std::string &filename,
    NameConflictPolicy conflict_policy)
{
    return DelayedDispatch(
        get_client_impl(),
        new LoadScene(get_object_id(), filename, conflict_policy));
}


Result<Matrix4f> ObjectProxyBase::get_transformation(
    CoordinateSystem coord_sys) const
{
    uint32_t request_id = allocate_request_id();

    return Result<Matrix4f>(
        dispatch_get_action(
            new GetTransformation(
                request_id, get_object_id(), coord_sys), request_id) );
}


Result<Vector3f> ObjectProxyBase::get_position(
    CoordinateSystem coord_sys) const
{
    uint32_t request_id = allocate_request_id();

    return Result<Vector3f>(
        dispatch_get_action(
            new GetPosition(
                request_id, get_object_id(), coord_sys), request_id) );
}


Result<Vector3f> ObjectProxyBase::get_orientation(
    CoordinateSystem coord_sys) const
{
    uint32_t request_id = allocate_request_id();

    return Result<Vector3f>(
        dispatch_get_action(
            new GetOrientation(
                request_id, get_object_id(), coord_sys), request_id) );
}


Result<std::vector<std::string> > ObjectProxyBase::get_children() const
{
    uint32_t request_id = allocate_request_id();

    return Result<std::vector<std::string> >(
        dispatch_get_action(
            new GetChildren(request_id, get_object_id()), request_id) );
}


//
// ------------------- ObjectProxy implementation ---------------------
//


ObjectProxy::ObjectProxy()
{
}

ObjectProxy::ObjectProxy(const ObjectProxyBase &p)
    : ObjectProxyBase(p)
{
}


ObjectProxy &ObjectProxy::operator=(const ObjectProxy &p) throw()
{
    return *this = (ObjectProxyBase &)p;
}


ObjectProxy &ObjectProxy::operator=(const ObjectProxyBase &p) throw()
{
    unchecked_assign(p);
    return *this;
}

DelayedDispatch ObjectProxy::assign(
    PeekabotClient &client, const std::string &path) throw()
{
    unchecked_assign(get_client_impl(client), allocate_pseudonym());

    return DelayedDispatch(
        get_client_impl(),
        new RegisterPseudonym(path, get_object_id()));
}

DelayedDispatch ObjectProxy::assign(
    const ObjectProxyBase &parent,
    const std::string &rel_path) throw()
{
    unchecked_assign(get_client_impl(parent), allocate_pseudonym());

    return DelayedDispatch(
        get_client_impl(),
        new RegisterPseudonym(
            get_object_id(parent), rel_path, get_object_id()));
}
