/*
 * Copyright Staffan Gimåker 2008-2009.
 *
 * ---
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */

#include "SensorProxy.hh"
#include "../PeekabotClient.hh"
#include "../../ChunkedBuffer.hh"
#include "../../serialization/SerializationInterface.hh"
#include "../../serialization/ChunkedBufferAdapter.hh"

#include "../../actions/AssignSensor.hh"
#include "../../actions/SetSensorData.hh"


using namespace peekabot;
using namespace peekabot::client;


//
// ------------------ SensorProxyBase implementation -------------------
//


SensorProxyBase::SensorProxyBase()
{
}


SensorProxyBase::SensorProxyBase(const SensorProxyBase &p)
    : ObjectProxyBase(p)
{
}

//

DelayedDispatch SensorProxyBase::update(const SensorData &reading)
{
    Action *a = 0;
    try
    {
        a = new SetSensorData(get_object_id(), *reading.m_buf);
        return DelayedDispatch(get_client_impl(), a);
    }
    catch(...)
    {
        if( a )
            delete a;
        throw;
    }
}


//
// ------------------ SensorData implementation ----------------------
//


SensorProxyBase::SensorData::SensorData()
    : m_buf(new ChunkedBuffer(8*1024))
{
}

SensorProxyBase::SensorData::~SensorData()
{
    if( m_buf )
    {
        delete m_buf;
        m_buf = 0;
    }
}

void SensorProxyBase::SensorData::write(float x) throw()
{
    ChunkedBufferAdapter adapter(*m_buf);
    SerializationInterface ar(adapter);
    ar << x;
}

void SensorProxyBase::SensorData::write(float x[], size_t n) throw()
{
    ChunkedBufferAdapter adapter(*m_buf);
    SerializationInterface ar(adapter);

    for( size_t i = 0; i < n; ++i )
        ar << x[i];
}

void SensorProxyBase::SensorData::write(double x) throw()
{
    // Note: doubles are mangled into floats
    write((float)x);
}

void SensorProxyBase::SensorData::write(double x[], size_t n) throw()
{
    ChunkedBufferAdapter adapter(*m_buf);
    SerializationInterface ar(adapter);

    // Note: doubles are mangled into floats
    for( size_t i = 0; i < n; ++i )
        write(x[i]);
}

void SensorProxyBase::SensorData::write(bool x) throw()
{
    ChunkedBufferAdapter adapter(*m_buf);
    SerializationInterface ar(adapter);
    ar << x;
}

void SensorProxyBase::SensorData::write(bool x[], size_t n) throw()
{
    ChunkedBufferAdapter adapter(*m_buf);
    SerializationInterface ar(adapter);

    for( size_t i = 0; i < n; ++i )
        ar << x[i];
}

void SensorProxyBase::SensorData::clear() throw()
{
    m_buf->clear();
}


//
// ------------------ SensorProxy implementation -------------------
//


SensorProxy::SensorProxy()
{
}


SensorProxy::SensorProxy(const SensorProxyBase &p)
    : SensorProxyBase(p)
{
}


SensorProxy &SensorProxy::operator=(const SensorProxy &p) throw()
{
    return *this = (SensorProxyBase &)p;
}


SensorProxy &SensorProxy::operator=(const SensorProxyBase &p) throw()
{
    unchecked_assign(unchecked_get_client_impl(p), get_pseudonym(p));
    return *this;
}


DelayedDispatch SensorProxy::assign(const ObjectProxyBase &p)
{
    unchecked_assign(get_client_impl(p), allocate_pseudonym());

    return DelayedDispatch(
        get_client_impl(),
        new AssignSensor(PathIdentifier(get_object_id(p)), get_object_id()));
}


DelayedDispatch SensorProxy::assign(
    PeekabotClient &client,
    const std::string &path)
{
    unchecked_assign(get_client_impl(client), allocate_pseudonym());

    return DelayedDispatch(
        get_client_impl(),
        new AssignSensor(PathIdentifier(path), get_object_id()));
}


DelayedDispatch SensorProxy::assign(
    const ObjectProxyBase &parent, 
    const std::string &rel_path)
{
    unchecked_assign(get_client_impl(parent), allocate_pseudonym());

    return DelayedDispatch(
        get_client_impl(), new AssignSensor(
            PathIdentifier(get_object_id(parent), rel_path), get_object_id()));
}
