/*
 * Copyright Anders Boberg 2006-2007.
 * Copyright Staffan Gimåker 2007-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
//////////////////////   BrowserFrame methods /////////////////////////////
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////

#include <FL/Fl.H>

#include "BrowserFrame.hh"
#include "BrowserTrigger.hh"
#include "Workspace.hh"
#include "UserInterfaceUtilities.hh"
#include "../SceneTreePtr.hh"

using namespace peekabot;
using namespace gui;

BrowserFrame::BrowserFrame(int x, int y, int w, int h, char* l) :
    Frame(x, y, w, h, l)
{
    init(x, y, w, h);
}

void BrowserFrame::init(int x, int y, int w, int h)
{
    m_title_box = new Fl_Box(x+31, y, w-31, 25, "Tree Browser");
    m_title_box->box(FL_PLASTIC_UP_BOX);
    m_title_box->color(FL_LIGHT3);
    m_title_box->labeltype(FL_ENGRAVED_LABEL);
    m_title_box->labelcolor((Fl_Color)43);
    add(m_title_box);

    m_browser = new TreeBrowser(x + 5, y + 30, w - 10, h - 35);
    add(m_browser);

    type(0); // No scrollbars.

    // Lock the scene tree to make sure nothing is modified while we
    // register triggers.
    {
        SceneTreePtr ptr = TheSceneTree::acquire();
        
        m_trigger = new BrowserTrigger(m_browser);
        
        // This causes the trigger to traverse the scene tree,
        // searching for cameras and registering itself for
        // every node.
        m_trigger->on_child_attached(ptr->get_root());
    }
}

BrowserFrame::BrowserFrame(Frame* frame) :
    Frame(frame->x(), frame->y(), frame->w(), frame->h())
{
    // Give the new frame the same size as the one it replaces
    init(frame->x(), frame->y(), frame->w(), frame->h());

    // Register the frame with the currently active workspace
    TheUIUtilities::instance().active_workspace()->register_browser_frame(this);
    
    // Since the old frame must have been selected to use the content selection
    // menu, we want the new frame to be selected as well.
    TheUIUtilities::instance().active_workspace()->select_frame(this);

    // Remove the old frame safely.
    frame->deregister();
    frame->hide();
    Fl::delete_widget(frame);

}

BrowserFrame::~BrowserFrame()
{
    SceneTreePtr ptr = TheSceneTree::acquire();

    m_trigger->deregister_recursive(ptr->get_root());

    delete m_trigger;
}

void BrowserFrame::deregister()
{
    Workspace* ws = (Workspace*)parent();

    ws->deregister_browser_frame(this);
}

void BrowserFrame::resize(int x, int y, int w, int h)
{
    Fl_Scroll::resize(x, y, w, h);
    m_title_box->resize(x+31, y, (w-31 > 110) ? w - 31 : 110 , 25);
    m_browser->resize(x + 5, y + 30, w - 10, h - 35);;
}

Frame::FrameType BrowserFrame::frame_type() const
{
    return Frame::BROWSER_FRAME;
}

bool BrowserFrame::resizable_width(int width)
{
    return 50;
}

bool BrowserFrame::resizable_height(int height)
{
    return 50;
}


Frame* BrowserFrame::clone(int x, int y, int w, int h)
{
    return new BrowserFrame(x, y, w, h);
}

void BrowserFrame::select_object(ObjectID object)
{
    m_browser->set_selected(object, true);    
}

void BrowserFrame::deselect_object(ObjectID object)
{
    m_browser->set_selected(object, false);
}

void BrowserFrame::clear_selection()
{
    m_browser->select_all(false);
}

TreeBrowser* BrowserFrame::get_tree_browser()
{
    return m_browser;
}

