/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_BROWSERFRAME_HH
#define __PEEKABOT_GUI_BROWSERFRAME_HH

#include <set>
#include <string>
#include <FL/Fl_Image.H>
#include <FL/Fl_Box.H>
#include "Frame.hh"
#include "TreeBrowser.hh"
#include "../Types.hh"

namespace peekabot
{
    class SceneObject;
    
    namespace gui
    {
        class Workspace;
        class BrowserTrigger;
        

        /** \brief Represents scene browser frames.
         *
         * The \c BrowserFrame object has a \c TreeBrowser object, which
         * it uses to present the user with a browsable graphical
         * representation of the scene tree.
         * To generate this tree, it uses a \c BrowserVisitor, which
         * traverses the entire tree, building \c NodeWidget objects,
         * which are graphical representations of \c SceneObject objects,
         * and \c TreeBrowser::TreeNode objects which contain the
         * NodeWidget objects.
         *
         * The \c BrowserVisitor will also register \c BrowserTriggers for
         * the \c TreeNode objects, to make sure they are updated when the
         * scene tree is. 
         *
         * The \c BrowserFrame also inherits the basic frame content
         * selection menu and border drawing functionality from the \c Frame base class.
         *
         */
        class BrowserFrame : public Frame
        {
            friend class BrowserVisitor;
            friend class BrowserTrigger;
        public:
            /** \brief Constructs a new \c BrowserFrame with the specified size.
             *
             * It will call the \c Frame base class constructor to initialize
             * the frame and draw the frame content selection menu.
             *
             * \param x The x-coordinate of the top left corner of this frame.
             * \param y The y-coordinate of the top left corner of this frame.
             * \param w The width of this frame.
             * \param h The height of this frame.
             * \param l The widget label (for FLTK compatibility)
             */
            BrowserFrame(int x, int y, int w, int h, char* l = 0);


            /** \brief Creates a new browser frame.
             *
             * The new frame will be created to fit the space occupied by \c
             * frame. The constructor will also take care of registering the
             * frame with the active workspace, and of removing the old frame.
             *
             * \param frame The \c Frame to replace.
             */
            BrowserFrame(Frame* frame);

            virtual ~BrowserFrame();

            /** \brief Selects or deselects nodes in the \c TreeBrowser.
             *
             * The purpose of these methods is to synchronize selection in
             * viewports with selection in the scene browser.
             *
             */
            void select_object(ObjectID object);
            void deselect_object(ObjectID object);
            void clear_selection();
   
            /** \brief Deregisters the \c Frame with its \c Workspace.
             */
            virtual void deregister();

            /** \brief Returns true if the frame can be resized to width \c width.
             *
             * Currently, the minimum width is 50 for browser frames.
             */
            virtual bool resizable_width(int width);
            /** \brief Returns true if the frame can be resized to height \c height.
             * 
             * Currently, the minimum height is 50 for browser frames.
             */
            virtual bool resizable_height(int height);


            /** \brief Creates a clone of the frame with the specified screen
             * dimensions.
             */
            virtual Frame* clone(int x, int y, int w, int h);


            virtual void resize(int x, int y, int w, int h);
            
            /** \brief Returns the \c TreeBrowser widget associated
             * with this frame.
             */
            TreeBrowser* get_tree_browser();
    
        protected:
            /** \brief Initializes a new browser frame.
             *
             * Uses a \c BrowserVisitor to generate the tree browser.
             *
             * \param x The x-coordinate of the top left corner of this frame.
             * \param y The y-coordinate of the top left corner of this frame.
             * \param w The width of this frame.
             * \param h The height of this frame.
             */
            virtual void init(int x, int y, int w, int h);

            /** \brief Get the type of the frame.
             *
             * \return A \c FrameType of value BROWSER_FRAME
             */
            virtual FrameType frame_type() const;

        private:
            /** \brief A pointer to the \c TreeBrowser in this frame.
             */
            TreeBrowser *m_browser;

            /** \brief The frame title box "Tree browser"
             */
            Fl_Box* m_title_box;

            /** \brief The trigger that this browser frame registers with objects.
             */
            BrowserTrigger *m_trigger;
        };

    }
}
#endif //__PEEKABOT_GUI_BROWSERFRAME_HH
