/*
 * Copyright Anders Boberg 2006-2007.
 * Copyright Staffan Gimåker 2007-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdexcept>
#include <iostream>

#include "BrowserTrigger.hh"
#include "BrowserFrame.hh"
#include "UserInterfaceUtilities.hh"
#include "UserInterface.hh"
#include "IconManager.hh"
#include "../MessageHub.hh"
#include "UpdateActions.hh"
#include "../SceneObject.hh"

using namespace peekabot;
using namespace gui;

BrowserTrigger::BrowserTrigger(TreeBrowser *browser)
    : m_browser(browser)
{
    
}

void BrowserTrigger::on_detached(SceneObject *object) throw()
{
    class RemoveTreeNode : public UpdateAction
    {
    public:
        RemoveTreeNode(TreeBrowser *browser, ObjectID id)
            : m_browser(browser),
              m_id(id)
        {
        }

        virtual void execute()
        {
            m_browser->remove_node(m_id);
        }

        virtual bool replaces()
        {
            return false;
        }
    private:
        TreeBrowser *m_browser;
        ObjectID m_id;
    };

    // Remove the object from the tree browser (and all children...)
    TheUserInterface::instance().post_action(
        m_browser, new RemoveTreeNode(m_browser, object->get_object_id()));

    // Recursively deregister this trigger from the object and all of
    // its children.
    deregister_recursive(object);   
}

void BrowserTrigger::deregister_recursive(SceneObject *object) throw()
{
    object->m_on_child_attached.disconnect(
        boost::bind(&BrowserTrigger::on_child_attached, this, _1));

    object->m_on_detached.disconnect(
        boost::bind(&BrowserTrigger::on_detached, this, _1));

    object->m_on_name_set.disconnect(
        boost::bind(&BrowserTrigger::on_name_set, this, _1));

    object->m_on_meta_name_set.disconnect(
        boost::bind(&BrowserTrigger::on_name_set, this, _1));


    SceneObject::ChildContainer &children = object->get_children();
    for( SceneObject::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        deregister_recursive((*it));
    }
}

void BrowserTrigger::on_child_attached(SceneObject *object) throw()
{
    class InsertTreeNode : public UpdateAction
    {
    public:
        InsertTreeNode(TreeBrowser *browser, ObjectID parent_id,
                   ObjectID child_id, const std::string &l, Fl_Image *img)
            : m_browser(browser),
              m_parent_id(parent_id),
              m_child_id(child_id),
              m_label(l),
              m_image(img)
        {
        }

        virtual void execute()
        {
            if(m_child_id == ROOT_OBJECT_ID)
                m_browser->add_root(ROOT_OBJECT_ID, m_label.c_str(), m_image);
            else
                m_browser->insert_node(m_child_id, m_parent_id,
                                       m_label.c_str(), m_image);
        }

        virtual bool replaces()
        {
            return false;
        }

    private:
        TreeBrowser *m_browser;
        ObjectID m_parent_id;
        ObjectID m_child_id;
        std::string m_label;
        Fl_Image *m_image;
    };

    // Make sure this trigger propagates in the scene tree...
    object->m_on_child_attached.connect(
        boost::bind(&BrowserTrigger::on_child_attached, this, _1));

    object->m_on_detached.connect(
        boost::bind(&BrowserTrigger::on_detached, this, _1));

    object->m_on_name_set.connect(
        boost::bind(&BrowserTrigger::on_name_set, this, _1));

    object->m_on_meta_name_set.connect(
        boost::bind(&BrowserTrigger::on_name_set, this, _1));
 
    // This will cause the object to call the appropriate visit method
    // setting our m_image member.
    object->accept(this);
    
    ObjectID parent_id, child_id;
    if(object->get_parent() == 0)
        parent_id = 0;
    else
        parent_id = object->get_parent()->get_object_id();

    child_id = object->get_object_id();

    // Add the object from the tree browser
    TheUserInterface::instance().post_action(
        m_browser, new InsertTreeNode(m_browser, parent_id, child_id,
                                      object->get_name_string().c_str(),
                                      m_image));

    // Recursively register and add all children.
    SceneObject::ChildContainer &children = object->get_children();
    for( SceneObject::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        on_child_attached((*it));
    }
}

void BrowserTrigger::on_name_set(SceneObject *object) throw()
{
    class SetTreeNodeLabel : public UpdateAction
    {
    public:
        SetTreeNodeLabel(TreeBrowser *browser, ObjectID id, const std::string &l)
            : m_browser(browser),
              m_id(id),
              m_label(l)
        {
        }

        virtual void execute()
        {
            m_browser->set_label(m_id, m_label.c_str());
        }

        virtual bool replaces()
        {
            return false;
        }

    private:
        TreeBrowser *m_browser;
        ObjectID m_id;
        std::string m_label;
    };

    // Add the object from the tree browser
    TheUserInterface::instance().post_action(
        m_browser, new SetTreeNodeLabel(m_browser, object->get_object_id(),
                                        object->get_name_string().c_str()));
}





void BrowserTrigger::visit(CameraObject* camera_object)
{
    m_image = TheIconManager::instance().get_icon("camera_object2.png");
    
}



void BrowserTrigger::visit(ModelObject* model_object)
{
    m_image = TheIconManager::instance().get_icon("teapot.png");
    
}



void BrowserTrigger::visit(SphereObject* sphere_object)
{
    m_image = TheIconManager::instance().get_icon("sphere_object.png");
    
}




void BrowserTrigger::visit(GroupObject* group_object)
{
    m_image = TheIconManager::instance().get_icon("group_object.png");
    
}




void BrowserTrigger::visit(CubeObject* cube_object)
{
    m_image = TheIconManager::instance().get_icon("cube_object.png");
    
}



void BrowserTrigger::visit(PolygonObject* polygon_object)
{
    m_image = TheIconManager::instance().get_icon("polygon_object.png");
    
}




void BrowserTrigger::visit(CylinderObject* cylinder_object)
{
    m_image = TheIconManager::instance().get_icon("cylinder_object.png");
    
}




void BrowserTrigger::visit(CircleObject* circle_object)
{
    m_image = TheIconManager::instance().get_icon("circle_object.png");
    
}



void BrowserTrigger::visit(PolyLineObject* poly_line_object)
{
    m_image = TheIconManager::instance().get_icon("poly_line_object.png");
    
}



void BrowserTrigger::visit(SensorObject* sensor_object)
{
    m_image = TheIconManager::instance().get_icon("sensor_object.png");
    
}




void BrowserTrigger::visit(PointSetObject* object)
{
    // TODO!!!
}


void BrowserTrigger::visit(Hinge* object)
{
    m_image = TheIconManager::instance().get_icon("dof.png");
}

void BrowserTrigger::visit(Slider* object)
{
    m_image = TheIconManager::instance().get_icon("dof.png");
}

void BrowserTrigger::visit(GridObject* object)
{
    m_image = TheIconManager::instance().get_icon("grid.png");
}

void BrowserTrigger::visit(PointCloud* object)
{
    m_image = TheIconManager::instance().get_icon("point_cloud.png");
}

void BrowserTrigger::visit(LineCloud* object)
{
    m_image = TheIconManager::instance().get_icon("line_cloud.png");
}

void BrowserTrigger::visit(OccupancyGrid2D* object)
{
    m_image = TheIconManager::instance().get_icon("occupancy_grid_2d.png");
}

void BrowserTrigger::visit(Label* object)
{
    m_image = TheIconManager::instance().get_icon("label.png");
}

