/*
 * Copyright Anders Boberg 2006-2007.
 * Copyright Staffan Gimåker 2007-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PEEKABOT_GUI_BROWSER_TRIGGER_HH_INCLUDED
#define PEEKABOT_GUI_BROWSER_TRIGGER_HH_INCLUDED

#include <FL/Fl_Image.H>

#include "../Types.hh"
#include "../ObjectVisitor.hh"

namespace peekabot
{
    namespace gui
    {
        class BrowserFrame;
        class TreeBrowser;

        /** \brief Trigger for updating \c BrowserFrame objects when
         * objects are modified in the scene.
         *
         * This trigger is registered when a \c TreeNode is created in a
         * \c BrowserFrame. It will contain a pointer to the \c
         * TreeBrowser that registered it.
         *
         * It is registered for the ON_CHILD_ADDED, ON_REMOVAL,
         * ON_REARRANGE, ON_NAME_SET and ON_META_NAME_SET,  stimuli types.
         *
         * When triggered, it will obtain a mutex for the GUI update queue
         * and then post an action corresponding to the type of event that
         * triggered it.
         *
         */

        class BrowserTrigger : public ObjectVisitor
        {

        public:

            /** \brief This constructor simply sets the pointer to the \c
             * TreeBrowser that registered the trigger.
             */
            BrowserTrigger(TreeBrowser *tree);

            ~BrowserTrigger() throw()
            {
            }

            void on_detached(SceneObject *object) throw();

            void deregister_recursive(SceneObject *object) throw();
            void on_child_attached(SceneObject *object) throw();
            void on_name_set(SceneObject *object) throw();

            /** \brief Creates \c TreeNode with a label and icon based on a CameraObject.
             * \param camera_object A pointer to the \c CameraObject being visited.
             */
            virtual void visit(CameraObject* camera_object);



            /** \brief Creates \c TreeNode with a label and icon based on a ModelObject.
             * \param model_object A pointer to the \c ModelObject being visited.
             */
            virtual void visit(ModelObject* model_object);



            /** \brief Creates \c TreeNode with a label and icon based on a SphereObject.
             * \param sphere_object A pointer to the \c SphereObject being visited.
             */
            virtual void visit(SphereObject* sphere_object);



            /** \brief Creates \c TreeNode with a label and icon based on a GroupObject.
             * \param group_object A pointer to the \c GroupObject being visited.
             */
            virtual void visit(GroupObject* group_object);



            /** \brief Creates \c TreeNode with a label and icon based on a CubeObject.
             * \param cube_object A pointer to the \c CubeObject being visited.
             */
            virtual void visit(CubeObject* cube_object);


            /** \brief Creates \c TreeNode with a label and icon based on a PolygonObject.
             * \param polygon_object A pointer to the \c PolygonObject being visited.
             */
            virtual void visit(PolygonObject* polygon_object);


            /** \brief Creates \c TreeNode with a label and icon based on a CylinderObject.
             * \param cylinder_object A pointer to the \c CylinderObject being visited.
             */
            virtual void visit(CylinderObject* cylinder_object);



            /** \brief Creates \c TreeNode with a label and icon based on a CircleObject.
             * \param circle_object A pointer to the \c CircleObject being visited.
             */
            virtual void visit(CircleObject* circle_object);


            /** \brief Creates \c TreeNode with a label and icon based on a PolyLineObject.
             * \param line_object A pointer to the \c LineObject being visited.
             */
            virtual void visit(PolyLineObject* poly_line_object);

            /** \brief Creates \c TreeNode with a label and icon based on a SensorObject.
             * \param sensor_object A pointer to the \c SensorObject being visited.
             */
            virtual void visit(SensorObject* sensor_object);

            /** \brief Creates \c TreeNode with a label and icon based on a PointSetObject.
             * \param object A pointer to the \c PointSetObject being visited.
             */
            virtual void visit(PointSetObject* object);

            /** \brief Creates \c TreeNode with a label and icon based on a Hinge.
             * \param object A pointer to the \c Hinge being visited.
             */
            virtual void visit(Hinge* object);

            /** \brief Creates \c TreeNode with a label and icon based on a Slider.
             * \param object A pointer to the \c Slider being visited.
             */
            virtual void visit(Slider* object);

            virtual void visit(GridObject* object);

            virtual void visit(PointCloud* object);

            virtual void visit(LineCloud* object);

            virtual void visit(OccupancyGrid2D* object);

            virtual void visit(Label* object);

        private:
            /** \brief The \c TreeBrowser that registered this trigger.
             */
            TreeBrowser *m_browser;



            Fl_Image *m_image;
        };

    }

}

#endif // PEEKABOT_GUI_BROWSER_TRIGGER_HH_INCLUDED
