/*
 * Copyright Anders Boberg 2006-2007.
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <FL/Fl_Box.H>

#include "CameraGroup.hh"
#include "IconManager.hh"
#include "UserInterfaceUtilities.hh"
#include "UserInterface.hh"
#include "CameraPropertyTrigger.hh"
#include "../SceneTreePtr.hh"
#include "../CameraObject.hh"
#include "../actions/SetOrthographic.hh"
#include "../actions/SetFOV.hh"
#include "../actions/SetZoomDistance.hh"

using namespace peekabot;
using namespace gui;
using namespace boost::posix_time;

CameraGroup::CameraGroup(int x, int y) :
    PropertyGroup(x, y, 230, 120, "Camera"),
    m_last_action_time(microsec_clock::local_time())
{
    hide();
    begin();
    
    // This is the box upon which the fill mode and view mode buttons are
    // placed
    Fl_Box* mode_box = new Fl_Box(x+10, y+25, 210, 80);
    mode_box->box(FL_PLASTIC_UP_FRAME);
    mode_box->color((Fl_Color)55);
       
    m_orthographic_button = new Fl_Button(x+145, y+62, 32, 32);
    m_orthographic_button->type(FL_TOGGLE_BUTTON);
    m_orthographic_button->image(TheIconManager::instance().get_icon(
                                  "../images/ortho_mode.png"));
    m_orthographic_button->callback(camera_mode_callback, this);
    m_orthographic_button->tooltip("Orthographic");
     
    m_zoom_roller = new Fl_Roller(x+20, y+45, 190, 15, "Zoom");
    m_zoom_roller->type(1);
    m_zoom_roller->minimum(0);
    m_zoom_roller->maximum(10000);
    m_zoom_roller->step(0.05);
    m_zoom_roller->value(1);
    m_zoom_roller->align(FL_ALIGN_TOP);
    m_zoom_roller->callback(&CameraGroup::zoom_callback);

    
    m_fov_input = new Fl_Value_Input(x+60, y+65, 55, 25, "FOV");
    m_fov_input->maximum(180);
    m_fov_input->step(0.1);
    m_fov_input->value(90);
    m_fov_input->when(FL_WHEN_CHANGED);
    m_fov_input->callback(camera_mode_callback, this);
    

    end();
    set_open(false);
    show();

    m_trigger = new CameraPropertyTrigger(
        false, m_orthographic_button,
        m_fov_input, m_zoom_roller);

    set_properties(TheUIUtilities::instance().get_selection());
}

CameraGroup::~CameraGroup()
{
    delete m_trigger;
}

void CameraGroup::set_properties(const std::set<ObjectID> &objects)
{
    if(!objects.empty())
    {
        activate();

        // Lock the scene tree and the selection set.
        SceneTreePtr ptr = TheSceneTree::acquire();
        
        Lock selection_lock(TheUIUtilities::instance().get_selection_mutex());
        
        // Deregister old triggers and register new ones...
        deregister_triggers(objects, ptr);
        register_triggers(objects, ptr);
        
        m_objects = objects;
    }
    else
        deactivate();
}

void CameraGroup::camera_mode_callback(Fl_Widget* w, void* data)
{
    CameraGroup* group = static_cast<CameraGroup*>(data);

    // Lock selection set
    Lock selection_lock(TheUIUtilities::instance().get_selection_mutex());
    std::set<ObjectID> objects = TheUIUtilities::instance().get_selection();

    real_unit fov = group->m_fov_input->value();
    
    bool orthographic;

    if(group->m_orthographic_button->value() == 1)
        orthographic = true;
    else
        orthographic = false;

    for(std::set<ObjectID>::iterator it = objects.begin();
        it != objects.end(); ++it)
    {
        TheUIUtilities::instance().post_action(
            new SetFOV(*it, fov));
        TheUIUtilities::instance().post_action(
            new SetOrthographic(*it, orthographic));
    }
}

void CameraGroup::zoom_callback(Fl_Widget* w, void* data)
{
    Fl_Roller *roller = static_cast<Fl_Roller*>(w);
    // Lock selection set
    Lock selection_lock(TheUIUtilities::instance().get_selection_mutex());
    std::set<ObjectID> objects = TheUIUtilities::instance().get_selection();

    real_unit zoom = roller->value();

    for(std::set<ObjectID>::iterator it = objects.begin();
        it != objects.end(); ++it)
        TheUIUtilities::instance().post_action(
            new SetZoomDistance(*it, zoom, false));
}

void CameraGroup::deregister_triggers(const std::set<ObjectID> &objects,
                                          SceneTreePtr &ptr)
{
    std::set<ObjectID> tmp;
    std::insert_iterator<std::set<ObjectID> > result(tmp, tmp.begin());

    std::insert_iterator<std::set<ObjectID> > end =
        std::set_difference(m_objects.begin(), m_objects.end(),
                            objects.begin(), objects.end(), result);

    for(std::set<ObjectID>::iterator it = tmp.begin();
        it != tmp.end(); ++it)
    {
        CameraObject *object = dynamic_cast<CameraObject *>(
            ptr->get_object(*it));
        if(object)
            m_trigger->attach_to(object);
    }
}

void CameraGroup::register_triggers(const std::set<ObjectID> &objects,
                                        SceneTreePtr &ptr)
{
    std::set<ObjectID> tmp;
    std::insert_iterator<std::set<ObjectID> > result(tmp, tmp.begin());

    std::insert_iterator<std::set<ObjectID> > end =
        std::set_difference(objects.begin(), objects.end(),
                            m_objects.begin(), m_objects.end(), result);

    for(std::set<ObjectID>::iterator it = tmp.begin();
        it != tmp.end(); ++it)
    {
        CameraObject *object = dynamic_cast<CameraObject *>(
            ptr->get_object(*it));
        if(object)
            m_trigger->attach_to(object);
    }
}

void CameraGroup::clear_updates()
{
    // Scrap any incoming updates for this group's widgets
    TheUserInterface::instance().remove_updates(m_fov_input);
    TheUserInterface::instance().remove_updates(m_orthographic_button);
    TheUserInterface::instance().remove_updates(m_zoom_roller);
}
