/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_CAMERA_GROUP_HH
#define __PEEKABOT_GUI_CAMERA_GROUP_HH


#include "PropertyGroup.hh"
#include "../SceneTreePtr.hh"

#include <set>
#include <FL/Fl_Value_Input.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Roller.H>

#include <boost/date_time/posix_time/posix_time.hpp>

namespace peekabot
{

    namespace gui
    {

        class CameraPropertyTrigger;

        /** \brief A group for setting camera properties.
         *
         * This group contains buttons for switching the fill and view mode of a camera, a roller for zooming (which is implemented as scaling the positional vector of the camera from its origo) and an input field for setting the field of view (FOV).
         *
         */
        class CameraGroup : public PropertyGroup
        {
        public:
            /** \brief Creates a new \c CameraGroup object
             *
             * Allocates and adds the sub widgets and fills them with properties by  *
             * calling \c set_properties() on the current selection.
             */
            CameraGroup(int x, int y);

            /** \brief Deletes the trigger allocated for this group.
             */
            virtual ~CameraGroup();

            /** \brief Virtual method for filling this group with information from a camera object.
             * 
             * It creates an \c Fl_Value_Input widget for FOV and registers the \c set_fov_callback() function with it. It creates an \c Fl_Roller widget and registers the \c zoom_callback() function with it. It creates two groups of \c Fl_Radio_Button objects each, and registers the \c set_fill_mode_callback() and \c set_view_mode_callback() functions respectively. It also registers a PropertyTrigger with the object corresponding to this group.
             *
             * This method is called by the \c PropertyVisitor as a new \c SceneObject is visited, or by a \c PropertyTrigger as some modification affecting this group of properties is made.
             *
             */
            virtual void set_properties(const std::set<ObjectID> &objects);


            void register_triggers(const std::set<ObjectID> &objects,
                                   SceneTreePtr &ptr);

            void deregister_triggers(const std::set<ObjectID> &objects,
                                     SceneTreePtr &ptr);

            /** \brief Callback for setting the fill mode, view mode and FOV of
             * a camera.
             *
             * This is called whenever the user changes the value of the "FOV" \c Fl_Value_Input widget or clicks one of the fill or view mode buttons.
             *
             * Will generate a \c SetCameraMode action  based on the value entered
             * by the user.
             *
             * \param w A pointer to the \c Fl_Value_Input control triggering the callback.
             * \param data A pointer to the \c CameraGroup of the widget.
             */
            static void camera_mode_callback(Fl_Widget* w, void* data);



            /** \brief Callback for zooming the camera in and out.
             *
             * This is called whenever the user drags the "Zoom" \c Fl_Roller widget.
             *
             * Generates a \c ZoomCamera action.
             *
             * \param w A pointer to the \c Fl_Roller control triggering the callback.
             * \param data A pointer to void, as required by FLTK (not used).
             */
            static void zoom_callback(Fl_Widget* w, void* data);
 
            /** \brief Removes all pending updates for this group.
             */
            void clear_updates();

        private:
            CameraPropertyTrigger* m_trigger;

            std::set<ObjectID> objects;

            Fl_Button *m_orthographic_button;
    
            Fl_Roller *m_zoom_roller;
    
            Fl_Value_Input *m_fov_input;

            boost::posix_time::ptime m_last_action_time;
        };

    }
}

#endif // __PEEKABOT_GUI_CAMERA_GROUP_HH
