/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/ref.hpp>
#include <FL/Fl_Value_Input.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Roller.H>

#include "CameraPropertyTrigger.hh"
#include "UserInterface.hh"
#include "UpdateActions.hh"
#include "../CameraObject.hh"
#include "../SceneTreePtr.hh"
#include "UserInterfaceUtilities.hh"

using namespace peekabot;
using namespace gui;


CameraPropertyTrigger::~CameraPropertyTrigger()
{
    SceneTreePtr ptr = TheSceneTree::acquire();
    while( !m_objects.empty() )
        detach_from(*m_objects.begin());
}


void CameraPropertyTrigger::on_fov_set(CameraObject *object)
{
    // Set the relative and accumulated opacity sliders to the average of
    // all objects in the activators set.

    real_unit fov = 0;
    int count = m_objects.size();
    for(SubjectSet::const_iterator it = m_objects.begin();
        it != m_objects.end(); ++it)
    {
        // We need to make sure that the object is a camera
        CameraObject* cam = dynamic_cast<CameraObject *>(*it);
        if(cam != 0)
        {
            fov += cam->get_fov();
        }
    }
    fov /= count;
    
    // Update the fov field    
    TheUserInterface::instance().post_action(
        m_fov_field, new UpdateValuator(m_fov_field, fov));    
}

void CameraPropertyTrigger::on_view_mode_set(CameraObject *object)
{
    // Set the relative and accumulated opacity sliders to the average of
    // all objects in the activators set.

    bool is_orthographic = object->is_orthographic();

     // Update the orthographic button.
    TheUserInterface::instance().post_action(
        m_orthographic_button, 
        new UpdateButton(m_orthographic_button, (is_orthographic) ? 1 : 0));
}

void CameraPropertyTrigger::on_zoom_distance_set(CameraObject *object)
{
    TheUserInterface::instance().post_action(
        m_zoom_roller, new UpdateValuator(m_zoom_roller,
                                          object->get_zoom_distance()));
}


void CameraPropertyTrigger::attach_to(SceneObject *object)
{
    // Add object to the set of registered objects
    if( m_objects.insert(object).second )
    {
        CameraObject *cam = dynamic_cast<CameraObject *>(object);
        assert( cam );

        // Register triggers
        cam->m_on_view_mode_set.connect(
            boost::bind(&CameraPropertyTrigger::on_view_mode_set, this, _1));
        
        if( !m_view_frame )
        {
            cam->m_on_fov_set.connect(
                boost::bind(&CameraPropertyTrigger::on_fov_set, this, _1));
        
            cam->m_on_zoom_distance_set.connect(
                boost::bind(&CameraPropertyTrigger::on_zoom_distance_set, this, _1));
        }
        
        object->m_on_subtree_detached.connect(
            boost::bind(&CameraPropertyTrigger::detach_from, this, _1));
        
        if( !m_view_frame )
        {
            on_fov_set(cam);
            on_zoom_distance_set(cam);
        }        

        on_view_mode_set(cam);
    }
}

void CameraPropertyTrigger::detach_from(SceneObject *object)
{
    // Erase object from set of registered objects
    if( m_objects.erase(object) > 0 )
    {
        CameraObject *cam = dynamic_cast<CameraObject *>(object);
        assert( cam );

        // Deregister triggers
        cam->m_on_view_mode_set.disconnect(
            boost::bind(&CameraPropertyTrigger::on_view_mode_set, this, _1));
        

        if( !m_view_frame )
        {
            cam->m_on_fov_set.disconnect(
                boost::bind(&CameraPropertyTrigger::on_fov_set, this, _1));
            
            cam->m_on_zoom_distance_set.disconnect(
                boost::bind(&CameraPropertyTrigger::on_zoom_distance_set, this, _1));
        }
        
        object->m_on_subtree_detached.disconnect(
            boost::bind(&CameraPropertyTrigger::detach_from, this, _1));
    }
}
