/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_CAMERA_PROPERTY_TRIGGER_HH
#define __PEEKABOT_GUI_CAMERA_PROPERTY_TRIGGER_HH

#include <cstring>
#include <set>
#include "../CameraObject.hh"

class Fl_Value_Input;
class Fl_Button;
class Fl_Roller;
namespace peekabot
{

namespace gui
{
    /** \brief Trigger for updating the cameraProperty fields in a\c CameraGroup.
     *
     * This trigger is registered when a \c CameraGroup is created in a \c
     * PropertyFrame. It will contain pointers to the visibility fields that it
     * is registered for.
     *
     * It is registered for the ON_FOV_SET and ON_VIEW_MODE_SET
     * triggers.
     *
     * When triggered it generates update events and posts them on the GUI update
     * queue via \c UserInterface::post_event().
     *
     */

class CameraPropertyTrigger
{

public:
    /** \brief This constructor simply sets the pointers to the opacity and
     * relative opacity sliders, the hidden checkbox and the layer radio group.
     */
    CameraPropertyTrigger(bool is_viewframe, Fl_Button *orthographic_button,
                          Fl_Value_Input *fov = 0,
                          Fl_Roller *zoom_roller = 0)
        : m_view_frame(is_viewframe), m_fov_field(fov),
          m_zoom_roller(zoom_roller),
          m_orthographic_button(orthographic_button) {}

    virtual ~CameraPropertyTrigger();

    void on_fov_set(CameraObject *object);
    void on_view_mode_set(CameraObject *object);
    void on_zoom_distance_set(CameraObject *object);

    void attach_to(SceneObject *object);
    void detach_from(SceneObject *object);

private:

    /** \brief Boolean specifying whether the trigger is registered for a
     * \c ViewFrame or not.
     */
    bool m_view_frame;

    /** \brief Pointer to the FOV field of the \c CameraGroup.
     */
    Fl_Value_Input* m_fov_field;

    Fl_Roller *m_zoom_roller;

    /** \brief A pointer to the orthogonal button.
     */
    Fl_Button* m_orthographic_button;

    typedef std::set<SceneObject *> SubjectSet;
    
    SubjectSet m_objects;
};


}
}

#endif //__PEEKABOT_GUI_CAMERA_PROPERTY_TRIGGER_HH
