/*
 * Copyright Anders Boberg 2006-2007.
 * Copyright Staffan Gimåker 2006-2007, 2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "CameraTrigger.hh"
#include "../SceneObject.hh"
#include "../CameraObject.hh"
#include "UserInterfaceUtilities.hh"
#include "UserInterface.hh"
#include <FL/Fl.H>
#include "../SceneTreePtr.hh"
#include "UpdateActions.hh"
#include "ViewFrame.hh"

using namespace peekabot;
using namespace gui;


CameraTrigger::CameraTrigger(ViewFrame* view_frame)
    : m_view_frame(view_frame)
{
    SceneTreePtr ptr = TheSceneTree::acquire();
    on_child_attached(ptr->get_root());
}


CameraTrigger::~CameraTrigger()
{
    SceneTreePtr ptr = TheSceneTree::acquire();
    deregister_recursive(ptr->get_root());
}

void CameraTrigger::deregister_recursive(SceneObject *object) throw()
{
    object->m_on_child_attached.disconnect(
        boost::bind(&CameraTrigger::on_child_attached, this, _1));

    object->m_on_detached.disconnect(
        boost::bind(&CameraTrigger::on_detached, this, _1));

    // Check if the trigger was a CameraObject...
    CameraObject *camera_ptr = dynamic_cast<CameraObject *>(object);
    if(camera_ptr)
    {
        object->m_on_name_set.disconnect(
            boost::bind(&CameraTrigger::on_name_set, this, _1));

        object->m_on_meta_name_set.disconnect(
            boost::bind(&CameraTrigger::on_name_set, this, _1));
    }

    SceneObject::ChildContainer &children = object->get_children();
    for( SceneObject::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        deregister_recursive(*it);
    }
}


void CameraTrigger::on_detached(SceneObject *object) throw()
{
    object->m_on_child_attached.disconnect(
        boost::bind(&CameraTrigger::on_child_attached, this, _1));

    object->m_on_detached.disconnect(
        boost::bind(&CameraTrigger::on_detached, this, _1));

    // Check if the trigger was a CameraObject...
    CameraObject *camera_ptr = dynamic_cast<CameraObject *>(object);
    if(camera_ptr)
    {
        TheUserInterface::instance().post_action(
            m_view_frame, new RemoveCamera(m_view_frame, camera_ptr->get_object_id()));

        object->m_on_name_set.disconnect(
            boost::bind(&CameraTrigger::on_name_set, this, _1));

        object->m_on_meta_name_set.disconnect(
            boost::bind(&CameraTrigger::on_name_set, this, _1));
    }

    SceneObject::ChildContainer &children = object->get_children();

    for( SceneObject::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        on_detached((*it));
    }
}


void CameraTrigger::on_child_attached(SceneObject *object) throw()
{
    // Make sure this trigger propagates in the scene tree...
    object->m_on_child_attached.connect(
        boost::bind(&CameraTrigger::on_child_attached, this, _1));

    object->m_on_detached.connect(
        boost::bind(&CameraTrigger::on_detached, this, _1));

    // Check if the trigger was a CameraObject...
    CameraObject *camera_ptr = dynamic_cast<CameraObject *>(object);
    if(camera_ptr)
    {
        // The name change triggers are registered only for camera objects...
        object->m_on_name_set.connect(
            boost::bind(&CameraTrigger::on_name_set, this, _1));

        object->m_on_meta_name_set.connect(
            boost::bind(&CameraTrigger::on_name_set, this, _1));

        // Add the camera to the camera selection menu.
        TheUserInterface::instance().post_action(
            m_view_frame, new UpdateCamera(m_view_frame, camera_ptr->get_object_id(),
                                           camera_ptr->get_name_string()));
    }

    SceneObject::ChildContainer &children = object->get_children();

    for( SceneObject::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        on_child_attached((*it));
    }
}

void CameraTrigger::on_name_set(SceneObject *object) throw()
{
    // Update the camera selection menu with the new name.
    TheUserInterface::instance().post_action(
        m_view_frame, new UpdateCamera(m_view_frame, object->get_object_id(),
                                       object->get_name_string()));
}

