/*
 * Copyright Anders Boberg 2006-2007.
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ColorGroup.hh"
#include "ColorTrigger.hh"
#include "UserInterfaceUtilities.hh"
#include "UserInterface.hh"
#include "../SceneObject.hh"
#include "../SceneTreePtr.hh"
#include "../actions/SetColor.hh"

using namespace peekabot;
using namespace peekabot::gui;
using namespace boost::posix_time;

ColorGroup::ColorGroup(int x, int y) :
    PropertyGroup(x, y, 230, 150, "Color"),
    m_last_action_time(microsec_clock::local_time())
{
    hide();
    begin();

    m_color_chooser = new Fl_Color_Chooser(x + 15, y + 25, 200, 95);
    m_color_chooser->callback(color_callback, this);

    m_apply_recursively = new Fl_Check_Button(x + 20, y + 125, 15, 15,
                                              "Set recursively");

    end();
    set_open(false);
    show();

    m_trigger = new ColorTrigger(m_color_chooser);

    set_properties(TheUIUtilities::instance().get_selection());
}

ColorGroup::~ColorGroup()
{
    delete m_trigger;
}


void ColorGroup::set_properties(const std::set<ObjectID> &objects)
{
    if(!objects.empty())
    {
        activate();

        // Lock the scene tree and the selection set.
        SceneTreePtr ptr = TheSceneTree::acquire();
             
        //Deregister old triggers and register new ones...
        deregister_triggers(objects, ptr);
        register_triggers(objects, ptr);
        
        m_objects = objects;
    }
    else
        deactivate();
}


void ColorGroup::deregister_triggers(const std::set<ObjectID> &objects,
                                          SceneTreePtr &ptr)
{
    // The set difference m_objects - objects gives all objects that
    // were previously selected, but are no longer.
    std::set<ObjectID> tmp;
    std::insert_iterator<std::set<ObjectID> > result(tmp, tmp.begin());

    std::insert_iterator<std::set<ObjectID> > end =
        std::set_difference(m_objects.begin(), m_objects.end(),
                            objects.begin(), objects.end(), result);

    // Deregister triggers for all objects that have been deselected.
    for(std::set<ObjectID>::iterator it = tmp.begin();
        it != tmp.end(); ++it)
    {
        SceneObject *object = ptr->get_object(*it);
        if(object)
            /*{
            object->m_on_color_set.disconnect(
                boost::bind(&ColorTrigger::on_color_set, m_trigger, _1));

            m_trigger->on_deregistered(object);
            }*/
            m_trigger->detach_from(object, true);
    }
}

void ColorGroup::register_triggers(const std::set<ObjectID> &objects,
                                        SceneTreePtr &ptr)
{
    // The set difference objects - m_objects gives all objects that
    // weren't previously selected, but are now.
    std::set<ObjectID> tmp;
    std::insert_iterator<std::set<ObjectID> > result(tmp, tmp.begin());

    std::insert_iterator<std::set<ObjectID> > end =
        std::set_difference(objects.begin(), objects.end(),
                            m_objects.begin(), m_objects.end(), result);

    // Register triggers for all newly selected objects.
    for(std::set<ObjectID>::iterator it = tmp.begin();
        it != tmp.end(); ++it)
    {
        SceneObject *object = ptr->get_object(*it);
        if(object)
            /*{
            object->m_on_color_set.connect(
                boost::bind(&ColorTrigger::on_color_set, m_trigger, _1));

            m_trigger->on_registered(object);
            }*/
            m_trigger->attach_to(object);
    }
}

void ColorGroup::color_callback(Fl_Widget* w, void* data)
{
    ColorGroup* group = static_cast<ColorGroup*>(data);
    Fl_Color_Chooser *cc = static_cast<Fl_Color_Chooser*>(w);
    
    // Lock selection set
    Lock selection_lock(TheUIUtilities::instance().get_selection_mutex());
    std::set<ObjectID> objects = TheUIUtilities::instance().get_selection();

    
    // Check to see that at least one action interval has passed
    // since our last visit here, else bail out.
    ptime current_time = microsec_clock::local_time();
    if(current_time < group->m_last_action_time +
       TheUserInterface::instance().get_action_interval())
        return;
    group->m_last_action_time = current_time;

    RGBColor color(cc->r(), cc->g(), cc->b()); 
    
    // Get the status of the set recursively checkbox
    bool recursive = group->m_apply_recursively->value() == 1;

    for(std::set<ObjectID>::iterator it = objects.begin();
        it != objects.end(); ++it)
        TheUIUtilities::instance().post_action(
            new SetColor(*it, color, recursive));
}

void ColorGroup::clear_updates()
{
    // Scrap any incoming updates for this group's widgets
    TheUserInterface::instance().remove_updates(m_color_chooser);
}
