/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_COLOR_GROUP_HH
#define __PEEKABOT_GUI_COLOR_GROUP_HH

#include <set>
#include "PropertyVisitor.hh"
#include "PropertyGroup.hh"
#include "../SceneTreePtr.hh"

#include <FL/Fl_Color_Chooser.H>
#include <FL/Fl_Check_Button.H>
#include <boost/date_time/posix_time/posix_time.hpp>


namespace peekabot
{

namespace gui
{
    class ColorTrigger;

    /** \brief A group of properties and controls for setting color properties.
     *
     * This group contains a \c Fl_Color_Chooser widget for selecting RGB color
     * values and uses the \c ColorTrigger class to fetch data for and update
     * its controls.
     *
     */
class ColorGroup : public PropertyGroup
{
public:
    /** \brief Creates a new \c ColorGroup object.
     *
     * Allocates and adds the sub widgets and fills them with properties by  *
     * calling \c set_properties() on the current selection.
     *
     * It also registers the \c ColorTrigger to make sure the controls are
     * set to the correct values.
     */
    ColorGroup(int x, int y);

    ~ColorGroup();

    /** \brief Virtual method for filling this group with color information from
     * an object.
     *
     * This method registers a \c ColorTrigger for all objects in the selection
     * in order to fetch property data. It also deregisters triggers for objects
     * that have been deselected.
     *
     * This method is called by the \c PropertyVisitor upon creation, and
     * whenever the selection changes.
     *
     */
    virtual void set_properties(const std::set<ObjectID> &objects);

    /** \brief Callback for applying color changes.
     *
     * This is called whenever the user releases the mouse button in the color area, or hits enter in one of the value input fields.
     *
     * Will generate a SetColorAction based on the value entered by the user.
     *
     * \param w A pointer to the \c Fl_Button control triggering the callback.
     * \param data A pointer to void, as required by FLTK (not used).
     */
    static void color_callback(Fl_Widget* w, void* data);


    void register_triggers(const std::set<ObjectID> &objects,
                           SceneTreePtr &ptr);

    void deregister_triggers(const std::set<ObjectID> &objects,
                             SceneTreePtr &ptr);

    /** \brief Removes all pending updates for this group.
     */
    void clear_updates();

private:
    Fl_Color_Chooser *m_color_chooser;

    Fl_Check_Button *m_apply_recursively;

    /** \brief A pointer to the trigger this group registers with selected
     * objects.
     */
    ColorTrigger* m_trigger;

    std::set<ObjectID> m_objects;

    boost::posix_time::ptime m_last_action_time;
};

}
}
#endif //__PEEKABOT_GUI_COLOR_GROUP_HH
