/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_COLOR_TRIGGER_HH
#define __PEEKABOT_GUI_COLOR_TRIGGER_HH

#include <set>

class Fl_Color_Chooser;

namespace peekabot
{

class SceneObject;

namespace gui
{


    /** \brief Trigger for updating the color chooser in a\c ColorGroup.
     *
     * This trigger is registered when a \c ColorGroup is created in a \c
     * PropertyFrame. It will contain a pointer to the \c Fl_Color_Chooser
     * that it is registered for.
     *
     * It is registered for the ON_COLOR_SET stimulus.
     *
     * When triggered it generates update events and posts them on the GUI
     * update queue via \c UserInterface::post_event().
     *
     */

class ColorTrigger
{

public:
    /** \brief This constructor simply sets the pointers to color and meta color
     * fields.
     */
    ColorTrigger(Fl_Color_Chooser* color_chooser);

    ~ColorTrigger();


    /** \brief Updates the color or meta color fields of a \c ColorGroup.
     *
     * This method generates an \c UpdateEvent for the affected input
     * widget and the corresponding data, which it fetches from the
     * activator set, and posts it on the GUI update
     * queue.
     *
     * \param stimuli The type of stimuli that caused this trigger.
     * \param object The \c SceneObject that this trigger was registered with.
     *
     */
    void on_color_set(SceneObject *object);

    void attach_to(SceneObject *object);
    void detach_from(SceneObject *object, bool update_widget);

private:
    /** \brief Pointers to the color fields that the trigger is registered for.
     */
    Fl_Color_Chooser* m_color_chooser;

    /** \brief A set storing all objects the trigger is registered with.
     */
    std::set<SceneObject *> m_objects;
};


}
}

#endif //__PEEKABOT_GUI_COLOR_TRIGGER_HH
