/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
//////////////////////   ConsoleFrame methods   ///////////////////////////
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////

#include <FL/Fl.H>
#include <FL/Fl_Box.H>

#include "ConsoleFrame.hh"
#include "Workspace.hh"
#include "UserInterfaceUtilities.hh"

using namespace peekabot;
using namespace gui;

ConsoleFrame::ConsoleFrame(int x, int y, int w, int h, char* l) :
    Frame(x, y, w, h, l)
{
    init(x, y, w, h);
}

void ConsoleFrame::init(int x, int y, int w, int h)
{
    Fl_Box* o = new Fl_Box(Frame::x()+31, Frame::y(), 120, 25, "Console");
    o->box(FL_PLASTIC_UP_BOX);
    o->color(FL_LIGHT3);
    o->labeltype(FL_ENGRAVED_LABEL);
    o->labelcolor((Fl_Color)43);

    add(o);

    m_display_buffer = new Fl_Text_Display (ConsoleFrame::x() + 5,
                                            ConsoleFrame::y()+25,
                                            ConsoleFrame::w() - 10,
                                            ConsoleFrame::h()-25);
    m_display_buffer->buffer(TheUIUtilities::instance().get_text_buffer());
    add(m_display_buffer);

    type(0); // No scroll bars!
    
}

ConsoleFrame::ConsoleFrame(Frame* frame) :
    Frame(frame->x(), frame->y(), frame->w(), frame->h())
{
    // Give the new frame the same size as the one it replaces
    init(frame->x(), frame->y(), frame->w(), frame->h());

    // Register the frame with the currently active workspace
    TheUIUtilities::instance().active_workspace()->register_console_frame(this);
    
    // Since the old frame must have been selected to use the content selection
    // menu, we want the new frame to be selected as well.
    TheUIUtilities::instance().active_workspace()->select_frame(this);

    // Remove the old frame safely.
    frame->deregister();
    frame->hide();
    Fl::delete_widget(frame);

    redraw();   
}

void ConsoleFrame::deregister()
{
    Workspace* ws = (Workspace*)parent();

    ws->deregister_console_frame(this);
}

Frame::FrameType ConsoleFrame::frame_type() const
{
    return Frame::CONSOLE_FRAME;
}

void ConsoleFrame::resize(int xx, int yy, int ww, int hh)
{
    if(hh < 60)
    {
        hh = 60;
        yy = y();
    }
    Fl_Scroll::resize(xx, yy, ww, hh);
    m_display_buffer->resize(xx + 5, yy + 30, ww - 10, hh - 35);    
    m_display_buffer->redraw();
}

bool ConsoleFrame::resizable_width(int width)
{
    return 50;
}

bool ConsoleFrame::resizable_height(int height)
{
    return 60;
}

void ConsoleFrame::update()
{
    int disp_lines = m_display_buffer->h() / (m_display_buffer->textsize()+ 5);
    int lines = m_display_buffer->buffer()->count_lines(
        0, m_display_buffer->buffer()->length());
    m_display_buffer->scroll(lines - disp_lines + 1, 0);
    m_display_buffer->redraw();
}


Frame* ConsoleFrame::clone(int x, int y, int w, int h)
{
    return new ConsoleFrame(x, y, w, h);
}

void ConsoleFrame::draw()
{
    Lock lock(TheUIUtilities::instance().get_console_mutex());

    Frame::draw();
}
