/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_CONSOLE_FRAME_HH
#define __PEEKABOT_GUI_CONSOLE_FRAME_HH

#include <list>
#include <string>
#include <FL/Fl_Text_Buffer.H>
#include <FL/Fl_Text_Display.H>
#include "Frame.hh"

namespace peekabot
{



namespace gui
{

    class Workspace;

    /** \brief Represents console frames.
     *
     * The console frame contains a Fl_Text_Display object, which is where messages to the console are displayed. The Fl_Text_Display object in turn contains a Fl_Text_Buffer object, which is the data structure storing the actual text. Currently, all ConsoleFrame objects share the same Fl_Text_Buffer, although this may be changed in later revisions.
     * 
     * The console frame also has a tool button for saving the contents of the Fl_Text_Buffer to file. Clicking it saves the buffer to a file with a name in the format <tt>robolook_console_yymmdd_hhmmss.txt</tt>. If such a file already exists, it will be overwritten.
     *
     * The ConsoleFrame also inherits the basic frame content selection menu and border drawing functionality from the Frame base class.
     *
     */

class ConsoleFrame : public Frame
{
public:
    /** \brief Constructs a new ConsoleFrame with the specified location and size.
     * Will call the Frame base class constructor to initialize the frame and get the frame content selection menu.
     *
     * \param x The x-coordinate of the top left corner of this frame.
     * \param y The y-coordinate of the top left corner of this frame.
     * \param w The width of this frame.
     * \param h The height of this frame.
     * \param l The widget label (for FLTK compatibility)
     */
    ConsoleFrame(int x, int y, int w, int h, char* l = 0);



    /** \brief Creates a new console frame.
     *
     * The new frame will be created to fit the space occupied by \c frame. The constructor will also take care of registering the frame with the active workspace, and of removing the old frame.
     *
     * \param frame The \c Frame to replace.
     */
    ConsoleFrame(Frame* frame);


    /** \brief Callback for the "Save to file" button. Saves the content of the Fl_Text_Buffer to a file.
     *
     * This callback will generate a file name in the format <tt>robolook_console_yymmdd_hhmmss.txt</tt>, and then call Fl_Text_Buffer::savefile(). If such a file already exists, it will be overwritten.
     *
     * This callback is registered by the ConsoleFrame constructor, and is called whenever the user clicks the "Save to file" button in a console frame.
     *
     * \param w A pointer to the Fl_Text_Display widget in this frame.
     * \param data A pointer to the Fl_Text_Buffer containing the text to be saved.
     *
     */
    static void save_callback(Fl_Widget* w, void* data);

    /** \brief Updates the console frame.
     *
     * This method calls Fl_Text_Display::redraw() on its display buffer to update it.
     *
     * This method is called during a UserInterfaceUtils::send_to_console() call.
     */
    void update();

    /** \brief Deregisters the \c Frame with its \c Workspace.
     */
    virtual void deregister();

    /** \brief Returns true if the frame can be resized to width \c width.
     *
     * Currently, the minimum width is 50 for console frames.
     */
    virtual bool resizable_width(int width);
    /** \brief Returns true if the frame can be resized to height \c height.
     * 
     * Currently, the minimum height is 60 for console frames.
     */
    virtual bool resizable_height(int height);

    /** \brief Creates a clone of the frame with the specified screen
     * dimensions.
     */
    virtual Frame* clone(int x, int y, int w, int h);

    /** \brief Reimplements the parent \c draw() method to ensure thread-safe
     * operation by locking the console mutex.
     */
    virtual void draw();

protected:

    /** \brief Constructs a new ConsoleFrame with the specified location and size.
     * This method will initialize a Fl_Text_Display object for displaying console text, and an Fl_Button for saving the file (for which it registers the save_callback() method.
     *
     * The constructor will also automatically fetch a pointer to the common Fl_Text_Buffer object from UserInterfaceUtilities for use with the Fl_Text_Display object. 
     *
     * \param x The x-coordinate of the top left corner of this frame.
     * \param y The y-coordinate of the top left corner of this frame.
     * \param w The width of this frame.
     * \param h The height of this frame.
     */
    void init(int x, int y, int w, int h);

    /** \brief Get the type of the frame.
     *
     * \return A \c FrameType of value CONSOLE_FRAME
     */
    virtual FrameType frame_type() const;

    /** \brief Overrides the \c Fl_Scroll::resize() method to resize the
     * console display window.
     */
    virtual void resize(int xx, int yy, int ww, int hh);


private:
    /** \brief A pointer to the \c Fl_Text_Display widget containing the display buffer for this frame.
     */
    Fl_Text_Display* m_display_buffer;
};





}
}
#endif // __PEEKABOT_GUI_CONSOLE_FRAME_HH
