/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <FL/Fl_Value_Input.H>

#include "CoordinateTrigger.hh"
#include "UserInterface.hh"
#include "UpdateActions.hh"
#include "../SceneObject.hh"
#include "../Vector.hh"
#include "../SceneTreePtr.hh"


using namespace peekabot;
using peekabot::gui::CoordinateTrigger;

CoordinateTrigger::CoordinateTrigger(
    Fl_Value_Input* x_field, 
    Fl_Value_Input* y_field,
    Fl_Value_Input* z_field)
    : m_x_field(x_field), m_y_field(y_field), m_z_field(z_field)
{
}


CoordinateTrigger::~CoordinateTrigger()
{
    SceneTreePtr ptr = TheSceneTree::acquire();
    while( !m_objects.empty() )
        detach_from(*m_objects.begin());
}


void CoordinateTrigger::on_transformation(SceneObject *object)
{
    peekabot::Vector3ru position;

    switch(m_system)
    {
        case peekabot::transformation_target::LOCAL_COORDINATES:
            return; // No use sending coordinates - they will be 0 anyway.
        case peekabot::transformation_target::PARENT_COORDINATES:
            position = object->get_transformation()(3);
            break;
        case peekabot::transformation_target::WORLD_COORDINATES:
            position = calculate_mean();
            break;
        default:
            break;
    }


    TheUserInterface::instance().post_action(
        m_x_field, new UpdateValuator(m_x_field, position(0)));

    TheUserInterface::instance().post_action(
        m_y_field, new UpdateValuator(m_y_field, position(1)));

    TheUserInterface::instance().post_action(
        m_z_field, new UpdateValuator(m_z_field, position(2)));
}


void CoordinateTrigger::attach_to(SceneObject *object)
{
    // Add object to the set of registered objects
    if( m_objects.insert(object).second )
    {
        // Register triggers
        object->m_on_transformation_set.connect(
            boost::bind(&CoordinateTrigger::on_transformation, this, _1));
        
        object->m_on_mtow_changed.connect(
            boost::bind(&CoordinateTrigger::on_transformation, this, _1));
        
        object->m_on_subtree_detached.connect(
            boost::bind(&CoordinateTrigger::detach_from, this, _1));
        
        on_transformation(object);
    }
}

void CoordinateTrigger::detach_from(SceneObject *object)
{
    // Erase object from set of registered objects
    if( m_objects.erase(object) > 0 )
    {
        // Deregister triggers
        object->m_on_transformation_set.disconnect(
            boost::bind(&CoordinateTrigger::on_transformation, this, _1));
        
        object->m_on_mtow_changed.disconnect(
            boost::bind(&CoordinateTrigger::on_transformation, this, _1));
        
        object->m_on_subtree_detached.disconnect(
            boost::bind(&CoordinateTrigger::detach_from, this, _1));

        //on_transformation(object);
    }
}

peekabot::Vector3ru CoordinateTrigger::calculate_mean()
{
    // Calculate average position for objects
    Vector4r average_position(0, 0, 0, 0);

    int count = m_objects.size();
    for(std::set<SceneObject *>::const_iterator it = m_objects.begin();
        it != m_objects.end(); ++it)
    {
        average_position += (*it)->get_mtow()(3);
    }
    average_position /= count;

    return average_position;
}



//    Fl_Value_Input* m_x_field, *m_y_field, *m_z_field;
