/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_COORDINATE_TRIGGER_HH
#define __PEEKABOT_GUI_COORDINATE_TRIGGER_HH

#include <set>

#include "../Vector.hh"
#include "../Types.hh"

class Fl_Value_Input;

namespace peekabot
{

class SceneObject;

namespace gui
{


    /** \brief Trigger for updating the coordinate fields in a\c MoveRotateGroup.
     *
     * This trigger is registered when a \c MoveRotateGroup is created in a \c
     * PropertyFrame. It will contain pointers to the coordinate fields that it
     * is registered for.
     *
     * It is registered for the ON_TRANSFORMATION_SET and ON_MTOW_CHANGED
     * stimuli.
     *
     * It keeps a list of objects for which any stimuli has been registered to
     * cause triggers, and assumes both stimuli have been registered for each
     * object.
     *
     * When triggered it generates update events and posts them on the GUI update
     * queue via \c UserInterface::post_event().
     *
     */

class CoordinateTrigger
{

public:
    /** \brief This constructor simply sets the pointers to the X- Y- and
     * Z-coordinate fields.
     */
    CoordinateTrigger(Fl_Value_Input* x_field, 
                      Fl_Value_Input* y_field,
                      Fl_Value_Input* z_field);

    ~CoordinateTrigger();

    /** \brief Specifies which system the coordinates are to be specified in.
     */
    inline void set_coordinate_system(
        transformation_target::CoordinateSystem system)
    {
        m_system = system;
        if(m_objects.empty() == false)
            on_transformation(*m_objects.begin());
    }


    /** \brief Method for handling transformation events.
     *
     * Calls the \c calculate_mean() method to get the averaged positional
     * vector for all objects it is registered for, and then generates
     * \c UpdateAction objects  for the corresponding object data
     * and property fields and post it on the GUI update queue.
     */
    void on_transformation(SceneObject *object);

    void attach_to(SceneObject *object);

    void detach_from(SceneObject *object);


private:
    /** \brief Calculates the mean coordinate values for a set of objects.
     *
     * This method will first determine whether the trigger was registered for
     * any other scene objects, indicating a multiple selection. In that case,
     * it will fetch the coordinate values of those other objects and calculate
     * the mean value.
     *
     * This method uses the \c ObjectTrigger::get_activator_set() method to
     * Gain access to the other objects that the trigger is registered for,
     * and then proceeds to calculate the mean positional vector.
     */
    Vector3ru calculate_mean();

    /** \brief Pointers to the coordinate fields that the trigger is registered
     * for.
     */
    Fl_Value_Input* m_x_field, *m_y_field, *m_z_field;

    /** \brief Stores which coordinate system the coordinate fields are using.
     */
    transformation_target::CoordinateSystem m_system;

    /** \brief A set storing all objects the trigger is registered with.
     */
    std::set<SceneObject *> m_objects;
};


}
}

#endif //__PEEKABOT_GUI_COORDINATE_TRIGGER_HH
