/**
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/bind.hpp>
#include <iostream>

#include "Decorator.hh"
#include "../IDTranslationTable.hh"
#include "UserInterfaceUtilities.hh"
#include "../renderer/Renderer.hh"
#include "Manipulator.hh"


using namespace peekabot;
using namespace peekabot::gui;
using peekabot::renderer::TheRenderer;


Decorator::Decorator(UserInterfaceUtilities &uiu)
    : m_uiu(uiu)
{
    // Connect slots...
    // ... for select/deselect
    m_uiu.on_object_selected().connect(
        boost::bind(&Decorator::selection_changed, this, _1, true));
    m_uiu.on_object_deselected().connect(
        boost::bind(&Decorator::selection_changed, this, _1, false));
    // ... tool changed
    m_uiu.get_manipulator()->get_mode_signal().connect(
        boost::bind(&Decorator::mode_changed, this, _1));
    // ... coordinate system changed
    m_uiu.get_manipulator()->get_system_signal().connect(
        boost::bind(&Decorator::coordinate_system_changed, this, _1));
    // ... pivot changed
    m_uiu.get_manipulator()->get_pivot_signal().connect(
        boost::bind(&Decorator::pivot_changed, this, _1));
    // ... axes changed
    m_uiu.get_manipulator()->get_axis_signal().connect(
        boost::bind(&Decorator::axes_changed, this, _1));
}

Decorator::~Decorator()
{
    // Disconnect slots...
    m_uiu.on_object_selected().disconnect(
        boost::bind(&Decorator::selection_changed, this, _1, true));
    m_uiu.on_object_deselected().disconnect(
        boost::bind(&Decorator::selection_changed, this, _1, false));
    
    if(m_uiu.get_manipulator() != 0)
    {
        m_uiu.get_manipulator()->get_mode_signal().disconnect(
            boost::bind(&Decorator::mode_changed, this, _1));
        m_uiu.get_manipulator()->get_system_signal().disconnect(
            boost::bind(&Decorator::coordinate_system_changed, this, _1));
        m_uiu.get_manipulator()->get_pivot_signal().disconnect(
            boost::bind(&Decorator::pivot_changed, this, _1));
        m_uiu.get_manipulator()->get_axis_signal().disconnect(
            boost::bind(&Decorator::axes_changed, this, _1));
    }
}

void Decorator::selection_changed(ObjectID id, bool selected) throw()
{
    NodeID node_id = TheIDTranslationTable::instance().get_node_id(id);

    if( selected )
        TheRenderer::instance().select_node(node_id);
    else
        TheRenderer::instance().deselect_node(node_id);
}

void Decorator::mode_changed(ManipulatorMode mode) throw()
{
    //std::cout << "mode=" << mode << std::endl;

    if( mode == MOVE_MODE )
    {
        TheRenderer::instance().set_transformation_guides_mode(
            renderer::MOVE_TRANSFORMATION_GUIDES);
        TheRenderer::instance().set_transformation_guides_coord_sys(
            m_uiu.get_manipulator()->get_system());
        TheRenderer::instance().set_transformation_guides_axes(
            m_uiu.get_manipulator()->get_axes());
    }
    else if( mode == ROTATE_MODE )
    {
        TheRenderer::instance().set_transformation_guides_mode(
            renderer::ROTATE_TRANSFORMATION_GUIDES);
        TheRenderer::instance().set_transformation_guides_coord_sys(
            m_uiu.get_manipulator()->get_system());
        TheRenderer::instance().set_transformation_guides_axes(
            m_uiu.get_manipulator()->get_axes());
        TheRenderer::instance().set_transformation_guides_pivot(
            m_uiu.get_manipulator()->get_pivot());
    }
    else
    {
        TheRenderer::instance().set_transformation_guides_mode(
            renderer::NO_TRANSFORMATION_GUIDES);
    }
}

void Decorator::coordinate_system_changed(CoordinateSystem coord_sys) throw()
{
    //std::cout << "coord_sys=" << coord_sys << std::endl;

    TheRenderer::instance().set_transformation_guides_coord_sys(coord_sys);
}


void Decorator::pivot_changed(RotationCenterpoint pivot) throw()
{
    TheRenderer::instance().set_transformation_guides_pivot(pivot);
}


void Decorator::axes_changed(Axis axes) throw()
{
    TheRenderer::instance().set_transformation_guides_axes(axes);
}

