/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_FRAME_HH
#define __PEEKABOT_GUI_FRAME_HH

#include <FL/Fl_Menu_Button.H>
#include <FL/Fl_Scroll.H>
#include <list>


namespace peekabot
{

namespace gui
{

    /** \brief Class representing a frame in the GUI.
     *
     * Provides some basic functionality common to all frames, such as the frame content selection menu displayed in the top left corner (created by calling \c create_menu() ), and the drawing of the extra-thick frame border of selected frames.
     *
     */
class Frame : public Fl_Scroll
{
public:
    /** \brief Enumeration describing the type of frame.
     */
    enum FrameType {PROPERTY_FRAME, VIEW_FRAME, BROWSER_FRAME, CONSOLE_FRAME};


    /** \brief Frame content selection menu callback. Changes the frame type according to the users choice in the menu.
     *
     * This method will hide the current frame and create a new \c Frame object over it. It will mark the current frame for deletion by calling the \c Fl::delete_widget() function. 
     *
     * This function is triggered by the user selecting an option in the frame content selection menu.
     *
     * \param w A pointer to the Fl_Menu_Button representing the frame content selection menu.
     * \param data A pointer to void, as required by FLTK (not used)
     *
     */
    static void menu_callback(Fl_Widget* w, void* data);




    /** \brief FLTK \c handle() method for selecting frames.
     *
     * This method is called by FLTK when an event occurs. If the event is a \c FL_PUSHED event, it means the user has clicked inside this frame. In that case, we set this frame as selected (by calling \c Workspace::select_frame()).
     *
     * We also pass all events on to \c Fl_Scroll::handle() to make sure any child widgets get the event as usual.
     *
     * \param msg The FLTK event passed to this method.
     *
     * \return If msg was \c FL_PUSHED, we return 1, else we return whatever Fl_Scroll::handle() returns.
     *
     */
    virtual int handle(int msg);

    /** \brief Set the \ m_selected property.
     *
     * This method sets the m_selected property, which indicates whether the
     * frame border should be drawn thick or not.
     *
     * \param selected \c true if the frame is selected, \c false if not. 
     */
    void selected(bool selected);

    /** \brief Get the selection status of this frame.
     *
     * Returns \c true if the frame is selected, \c false if it is not.
     */
    bool is_selected() const;

    /** \brief Get the type of the frame.
     *
     * This method is abstract and must be implemented by derived classes.
     *
     * \return The \c FrameType of this frame.
     */
    virtual FrameType frame_type() const = 0;

    /** \brief Deregisters the \c Frame with its \c Workspace.
     */
    virtual void deregister() = 0;

    /** \brief Returns true if the frame can be resized to width \c width.
     */
    virtual bool resizable_width(int width) = 0;
    /** \brief Returns true if the frame can be resized to height \c height.
     */
    virtual bool resizable_height(int height) = 0;

    /** \brief Creates a clone of the frame with the specified screen
     * dimensions.
     */
    virtual Frame* clone(int x, int y, int w, int h) = 0;

protected:
    /** \brief Creates the frame content selection menu.
     *
     * This constructor creates the FLTK \c Fl_Menu_Button object that implements the frame content selection menu, and registers \c menu_callback() as the callback function for the menu.
     *
     *
     * This constructor should be called in the constructor of the derived classes.
     *
     * \param x The x-coordinate of the top left corner of this frame.
     * \param y The y-coordinate of the top left corner of this frame.
     * \param w The width of this frame.
     * \param h The height of this frame.
     * \param l The widget label (for FLTK compatibility)
     */
    Frame(int x, int y, int w, int h, char* l = 0);

    /** \brief Draws an extra thick frame border if the frame is selected.
     *
     * This method uses FLTK drawing mode functions to draw an extra thick frame border during drawing mode.
     *
     * This method should only be called by FLTK, during a redraw, or from the \c draw() methods of derived classes.
     */
    virtual void draw();

private:
    /** \brief Indicates whether this is the selected frame or not. \c true if selected \c false if not.
     */
    bool m_selected;
};

}
}
#endif // __PEEKABOT_GUI_FRAME_HH
