/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <FL/Fl_Widget.H>
#include "GUIUpdater.hh"
#include "UpdateActions.hh"

using peekabot::gui::GUIUpdater;

GUIUpdater::~GUIUpdater()
{
    // Delete all actions.
    for(UpdateMap::iterator it = m_action_queue.begin();
        it != m_action_queue.end(); ++it)
    {
        delete it->second;
    }
    
}

void GUIUpdater::post_action(Fl_Widget* widget, UpdateAction* action)
{
    // Lock the 'queue'
    boost::recursive_mutex::scoped_lock lock(m_queue_mutex);

    // First, see if the update should replace any currently pending updates.
    if(action->replaces())
    {      
        // Delete the old update(s)
        std::pair<UpdateMap::iterator, UpdateMap::iterator> range =
            m_action_queue.equal_range(widget);

        for(UpdateMap::iterator it = range.first; it != range.second; ++it)
            delete it->second;
        
        // Erase previous entries
        m_action_queue.erase(widget);
    }
       
    // Insert new update
    m_action_queue.insert(std::make_pair(widget, action));
}



void GUIUpdater::update_gui()
{
    UpdateMap tmp_queue;

    // Lock the 'queue', copy it, clear it and unlock it.
    {
    boost::recursive_mutex::scoped_lock lock(m_queue_mutex);

    tmp_queue = m_action_queue;

    m_action_queue.clear();
    }

    // Process actions from the copy
    for(UpdateMap::iterator it = tmp_queue.begin();
        it != tmp_queue.end(); ++it)
    {
        // Execute the action
        it->second->execute();

        // Force a widget redraw
        if(it->first != 0)
            it->first->redraw();

        // Delete it when we're done with it
        delete it->second;
    }
}

void GUIUpdater::remove_updates(Fl_Widget* widget)
{
    // Lock the 'queue'
    boost::recursive_mutex::scoped_lock lock(m_queue_mutex);

    // Delete the update(s)
    std::pair<UpdateMap::iterator, UpdateMap::iterator> range =
        m_action_queue.equal_range(widget);
    
    for(UpdateMap::iterator it = range.first; it != range.second; ++it)
        delete it->second;
    
    // Erase previous entries
    m_action_queue.erase(widget);
}

bool GUIUpdater::has_actions()
{   
    return !m_action_queue.empty();
}
