/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_GUI_UPDATER_HH
#define __PEEKABOT_GUI_GUI_UPDATER_HH

#include <map>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/thread/recursive_mutex.hpp>


#include "../Types.hh"

class Fl_Widget;

namespace peekabot
{
namespace gui
{
    class UpdateAction;

    
    /** \brief The \cGUIUpdater class is responsible for maintaining a
     * queue of actions to update GUI widgets.
     *
     * These actions are posted by triggers when scene objects containing
     * data that is displayed in a GUI widget is modified. When this
     * queue contains elements, and enough time has passed since the last
     * GUI update, the server actuator thread will wake the main thread
     * which will then proceed to process the actions and update the
     * GUI elements.
     *
     * The 'queue' used by the GUI updater is not actually a queue, but
     * a map. This is because the order in wich widgets are updated is
     * irrelevant, since all updates are absolute and do not depend on
     * other widgets. Further, storing the updates in a map ensures that
     * only the most recent update to a widget is actually carried out -
     * updates that would have been overridden in the same update are
     * skipped.
     *
     * Finally, having a map container enables us to easily look up
     * updates for widgets that have been deleted since the update was
     * posted and remove them.
     *
     * It uses scoped locks and mutexes to ensure thread consistency
     * between the server actuator and the main thread.
     *
     */
    class GUIUpdater
    {
    public:
        /** \brief Deletes all the memory allocated for
         * the actions currently in the queue.
         */
        ~GUIUpdater();

        /** \brief Posts an update action on the queue.
         *
         * This method is intended to be called exclusively from the
         * server actuator thread. It acquires a lock on the queue
         * mutex, making sure the queue is not simultaneously polled
         * by the main thread.
         *
         * The first argument is the widget that the action affects.
         * If this widget is deleted, the \c remove_widget() method
         * should be called with a pointer to the widget passed to it
         * to make sure any updates are deleted along with it.
         *
         */
        void post_action(Fl_Widget* widget, UpdateAction* action);

        /** \brief Process all pending GUI updates.
         *
         * This method is intended to be called from the main thread.
         * It acquires a lock on the queue mutex and then copies the
         * queue before processing it. The mutex is unlocked before
         * processing starts, so that the server actuator thread can
         * simultaneously post actions to the queue while the copy is
         * being processed.
         */
        void update_gui();

        /** \brief Returns \c true if there are pending updates, \c false
         * if not.
         *
         * This method doesn't actually lock the mutex, as that would
         * possibly be unnecessarily time-consuming and this is not a
         * critical operation. The thread will be locked when the actual
         * processing takes place instead.
         */
        bool has_actions();

        /** \brief Removes any updates for the specified widget.
         *
         * This method is called when a widget gets deleted. Makes sure
         * the updater doesn't try to dereference a widget pointer that
         * has been deallocated.
         */
        void remove_updates(Fl_Widget* widget);
        
    private:
        typedef std::multimap<Fl_Widget*, UpdateAction*> UpdateMap;
        
        /** \brief Container implementing the 'queue'.
         */
        UpdateMap m_action_queue;

        /** \brief Mutex ensuring the thread safety of the queue.
         */
        boost::recursive_mutex m_queue_mutex;
    };


}
}

#endif //__PEEKABOT_GUI_GUI_UPDATER_HH
