/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_MAIN_MENU_HH
#define __PEEKABOT_GUI_MAIN_MENU_HH

#include <string>
#include <FL/Fl_Menu_Bar.H>
#include <FL/fl_ask.H>
//#include <FL/Fl_Image.H>

namespace peekabot
{

    namespace gui
    {


        /** \brief Class representing the main menu bar.
         *
         *  This class inherits most of its functionality from Fl_Menu_Bar, but
         *  adds callbacks for menu options as member functions.
         *
         */
        class MainMenu : public Fl_Menu_Bar
        {
        public:
            /** \brief Creates the main menu and registers the callback functions.
             *
             *  Calls base class constructor to create a menu bar that is 20 pixels in height, and covers the width of the main window. Calls the \c Fl_Menu_::copy() method  to build the menu options and register the callbacks.
             *
             */
            MainMenu();

            /** \brief Standard FLTK-style widget constructor.
             */
            MainMenu(int x, int y, int w, int h, char* l = 0);


            /** \brief FLTK callback function for the "Quit" menu option.
             *
             * This function starts by asking the user to confirm quitting by calling \c fl_ask(). If the user clicks "Yes", it will create a \c QuitAction, and then call \c UserInterFace::post_action to post it on the inbound action queue.
             *
             * \pre The program is running.
             *
             * \post The \c QuitAction has been posted.
             *
             * \param w \c Fl_Widget pointer required by FLTK (not used)
             * \param data void pointer required by FLTK (not used)
             * 
             */
            static void quit_callback(Fl_Widget* w, void* data);





            /** \brief FLTK callback function for the "About" menu option.
             *
             * This function calls the \c fl_message() function to display a brief text message telling the user which version of RoboLook he is running.
             *
             * \pre The main application window is the active window.
             *
             * \post The "About" dialog is displayed.
             *
             * \param w \c Fl_Widget pointer required by FLTK (not used)
             * \param data void pointer required by FLTK (not used)
             * 
             */
            static void about_callback(Fl_Widget* w, void* data);




            /** \brief FLTK callback function for the "Help" menu option.
             *
             * Creates a new \c Fl_Help_Dialog object and loads the RoboLook help files into it. 
             *
             * \pre The main application window is the active window.
             *
             * \post The "Help" dialog is displayed.
             *
             * \param w \c Fl_Widget pointer required by FLTK (not used)
             * \param data void pointer required by FLTK (not used)
             * 
             */
            static void help_callback(Fl_Widget* w, void* data);


            static void save_snapshot_callback(Fl_Widget* w, void* data);

            static void load_callback(Fl_Widget *w, void* data);
        };



    }
}
#endif //__PEEKABOT_GUI_MAIN_MENU_HH
