/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "MainWindow.hh"
#include "MainMenu.hh"
#include "Workspace.hh"
#include "UserInterfaceUtilities.hh"
#include "UserInterface.hh"
#include "ViewFrame.hh"
#include "BrowserFrame.hh"
#include "PropertyFrame.hh"
#include "ConsoleFrame.hh"
#include "ToolButton.hh"
#include "IconManager.hh"
#include "../renderer/Renderer.hh"
#include "../SceneTreePtr.hh"
#include "../actions/AddCamera.hh"
#include "../Types.hh"
#include "../MessageHub.hh"
#include "../RendererRelay.hh"
#include "GUITypes.hh"
#include "Manipulator.hh"
#include "CommandRepository.hh"
#include "StatusBar.hh"
#include "../Configuration.hh"

#include <FL/fl_ask.H>
#include <FL/Fl_Gl_Window.H>
#include <FL/Fl_Input.H>
#include <FL/Fl.H>
#include <FL/Fl_Menu_Window.H>
#include <FL/Fl_Pack.H>
#include <string>
#include <cstdlib>

#include <boost/bind.hpp>

using namespace peekabot;
using namespace gui;
using peekabot::renderer::TheRenderer;

namespace peekabot
{
    namespace gui
    {
        char move_tooltip[128];
        char rotate_tooltip[128];

        char add_camera_tooltip[128];
    }
}

MainWindow::MainWindow() :
    Fl_Overlay_Window(100, 100, Fl::w()-200, Fl::h()-200, "peekabot")
{
    end();
}

void MainWindow::init()
{
    // Set the Fl scheme
    Fl::scheme("plastic");


    begin();

    // Generate main menu
    m_main_menu = new MainMenu(0,0,w(), 25);


    ///////////////////////////////////////////////////////////
    ///////////////// Tool bar ////////////////////////////////
    ///////////////////////////////////////////////////////////

    Fl_Group *tool_bar = new Fl_Group(0, 25, w(), 50);
    tool_bar->box(FL_UP_BOX);
    tool_bar->resizable(0);
    ToolButton *move_objects = new ToolButton(
        10, 30, 40, 40, 0,
        TheUIUtilities::instance().get_command_repository()->get_command(CMD_TOGGLE_MOVE_MODE));
    move_objects->type(FL_TOGGLE_BUTTON);
    sprintf(move_tooltip, "%s: Move objects",
            TheConfiguration::instance().get_option<std::string>(
                "shortcuts.move_mode", "m").c_str());
    move_objects->tooltip(move_tooltip);
    move_objects->image(TheIconManager::instance().get_icon("move.png"));
    // Connect to the mode signal of the manipulator! This will toggle the
    // button when the setting changes.
    TheUIUtilities::instance().get_manipulator()->get_mode_signal().connect(
        boost::bind(&ToolButton::set_if_equal, move_objects, MOVE_MODE, _1));
    
    ToolButton *rotate_objects = new ToolButton(
        55, 30, 40, 40, 0,
        TheUIUtilities::instance().get_command_repository()->get_command(CMD_TOGGLE_ROTATE_MODE));
    rotate_objects->type(FL_TOGGLE_BUTTON);
    sprintf(rotate_tooltip, "%s: Rotate objects",
            TheConfiguration::instance().get_option<std::string>(
                "shortcuts.rotate_mode", "r").c_str());
    rotate_objects->tooltip(rotate_tooltip);
    rotate_objects->image(TheIconManager::instance().get_icon("rotate.png"));
    // Connect to the mode signal of the manipulator! This will toggle the
    // button when the setting changes.
    TheUIUtilities::instance().get_manipulator()->get_mode_signal().connect(
        boost::bind(&ToolButton::set_if_equal, rotate_objects, ROTATE_MODE, _1));

    // Separator
    Fl_Box * sep1 = new Fl_Box(101, 32, 3, 36);
    sep1->box(FL_DOWN_BOX);  


    ToolButton *add_camera = new ToolButton(
        110, 30, 40, 40, 0, &MainWindow::add_camera_callback);
    add_camera->tooltip("Create a new camera");
    add_camera->image(TheIconManager::instance().get_icon("add_camera.png"));
       

    tool_bar->end();

    ///////////////////////////////////////////////////////////
    ///////////////// Status bar///////////////////////////////
    ///////////////////////////////////////////////////////////
    
    new StatusBar(0, h()-26, w(), 26);   
    

    // Generate dummy FL_GL_Window to hold the shared context
    class GLDummy : public Fl_Gl_Window
    {
    public:
        GLDummy()
            : Fl_Gl_Window(1,1,1,1)
        {
            mode(FL_DOUBLE | FL_RGB | FL_DEPTH  |
                 FL_STENCIL | FL_MULTISAMPLE);
        }

        virtual ~GLDummy()
        {
        }
        
        virtual void draw()
        {
            // We need this to display the an error message in case we can't
            // initialize the renderer.
            // If we simply call fl_alert(), this draw() method will be called
            // again, it will fail again, ... Infinite recursion in other words.
            // Set this flag to high to prevent further initialization attempts.
            static bool dont_init = false;

            if( (!TheUserInterface::instance().get_gl_context() || !valid()) &&
                !dont_init )
            {
                TheUserInterface::instance().set_gl_context(context());

                try 
                {
                    TheRenderer::instance().initialize();
                    // If we could init the renderer, start forwarding changes
                    // to it
                    TheRendererRelay::instance();
                }
                catch(std::exception &e)
                {
                    TheMessageHub::instance().publish(
                        ERROR_MESSAGE,
                        "Failed to initialize renderer, shutting down... "
                        "Error: %s", e.what());

                    dont_init = true;
                    fl_alert("Fatal error! Could not initialize renderer:\n%s",
                             e.what());
                    exit(-1);
                }
            }
        }
    };

    GLDummy *dummy = new GLDummy();

    dummy->redraw();
    //dummy->make_current();


    // Generate workspaces
    m_workspaces = new Fl_Tabs(0, 75, w(), h()-101);
    m_workspaces->end();
    resizable(m_workspaces);
    end();

    // Load the workspace layout (latest, or default)
    load_layout();

}






void MainWindow::load_layout(std::string filename) 
{
    /* Temporary solution - we create a workspace consisting of 4 view frames
     * a browser frame, a property frame and a console frame.
     */
    
    Workspace* ws = new Workspace(m_workspaces->x(), m_workspaces->y()+15,
                                  m_workspaces->w(), m_workspaces->h()-15,
                                  "Workspace 1");

    ws->register_view_frame(new ViewFrame(ws->x(), ws->y(),
                                          ws->w()*4/5, ws->h()));

    BrowserFrame *browser = new BrowserFrame(ws->x()+ws->w()*4/5, ws->y(),
                                             ws->w()-ws->w()*4/5, ws->h()*3/5);
    ws->register_browser_frame(browser);
    ws->register_property_frame(new PropertyFrame(ws->x()+ws->w()*4/5,
                                                  ws->y()+browser->h(),
                                                  ws->w()-ws->w()*4/5, 
                                                  ws->h()-browser->h()));

    TheUIUtilities::instance().activate_workspace(ws);
    m_workspaces->add(ws);

    // TODO: remember to change this if more than one workspace is
    // supported.
    m_workspaces->resizable(ws);
    /*
    Frameset* frameset_test = new Frameset(m_workspaces->x(),
                                           m_workspaces->y()+15,
                                           m_workspaces->w(),
                                           m_workspaces->h()-15,
                                           "Workspace 2");

    frameset_test->init(new PropertyFrame(ws->x()+ws->w()*4/5,
                                                ws->y()+ws->h()*2/5,
                                                ws->w()*1/5, ws->h()*3/5));
    m_workspaces->add(frameset_test);
    */
    m_workspaces->clear_visible_focus();
}

int MainWindow::handle(int msg)
{
    return Fl_Double_Window::handle(msg);
}


void MainWindow::add_camera_callback()
{
    TheUIUtilities::instance().post_action(
        new AddCamera(PathIdentifier("root.camera"), 
                      AUTO_ENUMERATE_ON_CONFLICT, false, -1));
}

void MainWindow::draw_overlay()
{
    
    // Iterate through the overlay function map
    // and call all overlay drawers.
    for(OverlayMap::const_iterator it = m_overlays.begin(); it != m_overlays.end(); ++it)
    {
        it->second();
    }
}

