/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_MAIN_WINDOW_HH
#define __PEEKABOT_GUI_MAIN_WINDOW_HH

#include <string>
#include <map>
#include <list>
#include <FL/Fl.H>
#include <FL/Fl_Overlay_Window.H>
#include <FL/Fl_Tabs.H>
#include <FL/Fl_Widget.H>
#include <boost/function.hpp>


namespace peekabot
{

namespace gui
{

    class ToolButton;
    class MainMenu;

/** \brief Class representing the main application window.
 *
 * The main objective of this class is to draw the main window, the toolbar and the main menu. These are all created during construction.
 *
 * The Fl_Double_Window base class handles most of the functionality, such as double-buffering and downward propagation of events.
 *
 */
class MainWindow : public Fl_Overlay_Window
{
public:
    typedef boost::function<void (void)> OverlayFunction;

    /** \brief Creates the main application window.
     *
     * This constructor initializes the main window to its default size by calling the base class constructor.
     */
    MainWindow();

    /** \brief Initializes the \c MainWindow.
     *
     * Creates the main menu, a default toolbar and a set of default workspace tabs.
     *
     */
    void init();

    /** \brief Loads a session layout file.
     *
     * Loads the layout file specified by the \c filename parameter.
     * If this parameter is blank (default), then this method will attempt
     * to load the file "latest.wsp", or failing that, "default.wsp". If
     * none of these files exist or are valid, it will create a single
     * workspace contain a single ViewFrame.
     *
     */
    void load_layout(std::string filename = std::string());

    virtual int handle(int msg);

    static void add_camera_callback();
    
    void register_overlay(Fl_Widget *w, const OverlayFunction &func)
    {
        m_overlays[w] = func;
    }
    
    void deregister_overlay(Fl_Widget *w)
    {
        m_overlays.erase(w);
    }

protected:
    virtual void draw_overlay();
private:
    typedef std::map<Fl_Widget*, OverlayFunction> OverlayMap;


    /** \brief A pointer to the main menu widget.
     */
    MainMenu* m_main_menu;
    /** \brief The list of \c ToolButton objects that make up the toolbar.
     */
    std::list<ToolButton*> m_toolbar;
    /** \brief A pointer to the \c Fl_Tabs widget that contains the workspaces.
     */
    Fl_Tabs* m_workspaces;
    
    OverlayMap m_overlays;
};


}
}
#endif //__PEEKABOT_GUI_MAIN_WINDOW_HH
