/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_MOVE_ROTATE_GROUP_HH
#define __PEEKABOT_GUI_MOVE_ROTATE_GROUP_HH

#include <list>
#include <FL/Fl_Value_Input.H>
#include <FL/Fl_Round_Button.H>
#include <FL/Fl_Check_Button.H>

#include "../Vector.hh"
#include "PropertyGroup.hh"
#include "../SceneTreePtr.hh"

namespace peekabot
{

    class SceneObject;

namespace gui
{
    class MoveInput;
    class RotateInput;
    class TransformationInput;
    class CoordinateTrigger;
    class PropertyTrigger;

    /** \brief A group of properties and controls for moving and rotating
     * objects.
     *
     * This group is divided into a "Move" tab and a "Rotate" tab.
     *
     * - The "Move" tab contains \c Fl_Value_Input widgets for inputting
     *  movement coordinates, as well as radio buttons for selecting which
     *  coordinate system to view the coordinates in. It also features a check
     *  box for viewing absolute or relative coordinates.
     * - The "Rotate" tab contains \c Fl_Value_Input widgets for inputting
     *  degrees for rotational transformation. It also features radio button
     *  groups for selecting the rotational center and axis.
     *
     * Based on user input, the group will generate transformation actions and
     * have \c UserInterfaceUtilities forward them to the inbound actions queue.
     *
     * This group uses the \c CoordinateTrigger to fill and update its
     * coordinate fields with data from the selected objects.
     *
     */



class MoveRotateGroup : public PropertyGroup
{
    friend class MoveInput;
    friend class RotateInput;
public:
    /** \brief Creates a new \c MoveRotateGroup object.
     *
     * Allocates and adds the sub widgets and makes sure they are filled with
     * properties by registering a \c CoordinateTrigger for them.
     */
    MoveRotateGroup(int x, int y);

    /** \brief Deletes the trigger allocated for this group.
     */
    virtual ~MoveRotateGroup();

    /** \brief Virtual method for preparing this property group for displaying
     * the properties of the selected objects.
     *
     * For multiple selections, this method disallows all transformations except
     * movement in relative world coordinates, and rotation around local center.
     *
     * It also registers a \c CoordinateTrigger with the currently selected
     * objects to fill the coordinate input fields with values. It also
     * it also deregisters triggers for objects that have been deselected.
     *
     *
     * This method is called by the \c PropertyVisitor upon creation, and
     * whenever the selection changes.
     *
     */
    virtual void set_properties(const std::set<ObjectID> &objects);

    /** \brief Removes all pending updates for this group.
     */
    void clear_updates();

    /** \brief Registers a \c CoordinateTrigger with all objects in the set
     * passed to the method.
     */
    void register_triggers(const std::set<ObjectID> &objects,
                           SceneTreePtr &ptr);

    /** \brief Deregisters the \c CoordinateTrigger from all objects in the set
     * passed to the method.
     */
    void deregister_triggers(const std::set<ObjectID> &objects,
                             SceneTreePtr &ptr);

private:
    /** \brief Callback for the "Relative coordinates" checkbox.
     *
     * This will either set each coordinate input box to 0,
     * or fill them with coordinate information, depending on whether
     * the button is checked or not.
     *
     * \param w A pointer to the \c Fl_Value_Input control triggering the
     * callback.
     * \param data A pointer to void, as required by FLTK (not used).
     */
    static void relative_coordinates_callback(Fl_Widget* w, void* data);

    static void coordinate_choice_callback(Fl_Widget* w, void* data);

    static void rotational_system_callback(Fl_Widget* w, void* data);

    /** \brief A list of the \c SceneObject objects this group is displaying the
     * properties of.
     * \deprecated
     */
    std::set<ObjectID> m_objects;

    /** \brief The averaged position of a selection of multiple objects.
     * \deprecated{This is now handled by \c CoordinateTrigger.}
     */
    Vector4ru m_average_position;
    Vector4ru m_local_position;
    Vector4ru m_parent_position;

    /** \brief Pointers to the \c Fl_Value_Input widget for changing this
     * coordinate.
     */
    MoveInput *m_x_coordinate, *m_y_coordinate, *m_z_coordinate;

    /** \brief Pointers to the \c Fl_Round_Button widget for selecting this
     * coordinate system.
     */
    Fl_Round_Button *m_world_coordinates, *m_parent_coordinates,
        *m_local_coordinates;

    /** \brief Pointers to the \c Fl_Check_Button widget for selecting whether
     * to use relative or absolute coordinates.
     */
    Fl_Check_Button *m_relative_coordinates;

    /** \brief Pointers to the \c Fl_Value_Input widget for rotating about this
     * axis.
     */
    RotateInput *m_x_axis, *m_y_axis, *m_z_axis;

    /** \brief Pointers to the \c Fl_Round_Button widget for selecting this
     * rotational center.
     */
    Fl_Round_Button *m_world_center, *m_parent_center, *m_local_center;

    /** \brief Pointers to the \c Fl_Round_Button widget for selecting this rotational axis.
     */
    Fl_Round_Button *m_world_axis, *m_local_axis, *m_parent_axis;

    /** \brief The coordinate trigger this group registered with the selected
     * objects.
     */
    CoordinateTrigger* m_trigger;

};

}
}
#endif //__PEEKABOT_GUI_MOVE_ROTATE_GROUP_HH
