/*
 * Copyright Anders Boberg 2006-2007.
 * Copyright Staffan Gimåker 2007-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "NameGroup.hh"
#include "NameTrigger.hh"
#include "UserInterfaceUtilities.hh"
#include "UserInterface.hh"
#include "../SceneObject.hh"
#include "../SceneTreePtr.hh"
#include "../actions/SetName.hh"
#include "../actions/SetMetaName.hh"

using namespace peekabot;
using namespace peekabot::gui;

NameGroup::NameGroup(int x, int y) :
        PropertyGroup(x, y, 230, 135, "Name")
{
    hide();
    begin();

    m_name_input = new Fl_Input(x + 85, y + 35 , 125, 25, "Name");
    m_meta_name_input = new Fl_Input(x + 85, y + 70, 125, 25, "Meta name");
    
    m_apply_button = new Fl_Button(x + 85, y + 105, 50, 25, "Apply");
    m_apply_button->callback(name_callback, this);

    end();
    set_open(false);
    show();

    m_trigger = new NameTrigger(m_name_input, m_meta_name_input);

    set_properties(TheUIUtilities::instance().get_selection());
}

NameGroup::~NameGroup()
{
    delete m_trigger;
}


void NameGroup::set_properties(const std::set<ObjectID> &objects)
{
    if(!objects.empty())
    {
        activate();

        // Lock the scene tree while we register our trigger
        SceneTreePtr ptr = TheSceneTree::acquire();
       
        // Deregister old triggers and register new ones...
        deregister_triggers(objects, ptr);
        register_triggers(objects, ptr);
        
        m_objects = objects; 
    }
    else
        deactivate();
}

void NameGroup::name_callback(Fl_Widget* w, void* data)
{
    NameGroup* group = static_cast<NameGroup*>(data);
    

    // Lock selection set
    Lock selection_lock(TheUIUtilities::instance().get_selection_mutex());
    std::set<ObjectID> objects = TheUIUtilities::instance().get_selection();

    std::string name = group->m_name_input->value();
    std::string meta_name = group->m_meta_name_input->value();

    TheUIUtilities::instance().post_action(
        new SetName(*objects.begin(), name));
    TheUIUtilities::instance().post_action(
        new SetMetaName(*objects.begin(), meta_name));
}


void NameGroup::deregister_triggers(const std::set<ObjectID> &objects,
                                          SceneTreePtr &ptr)
{
    std::set<ObjectID> tmp;
    std::insert_iterator<std::set<ObjectID> > result(tmp, tmp.begin());

    std::insert_iterator<std::set<ObjectID> > end =
        std::set_difference(m_objects.begin(), m_objects.end(),
                            objects.begin(), objects.end(), result);

    for(std::set<ObjectID>::iterator it = tmp.begin();
        it != tmp.end(); ++it)
    {
        SceneObject *object = ptr->get_object(*it);
        if(object)
            m_trigger->detach_from(object);
    }
}

void NameGroup::register_triggers(const std::set<ObjectID> &objects,
                                        SceneTreePtr &ptr)
{
    std::set<ObjectID> tmp;
    std::insert_iterator<std::set<ObjectID> > result(tmp, tmp.begin());

    std::insert_iterator<std::set<ObjectID> > end =
        std::set_difference(objects.begin(), objects.end(),
                            m_objects.begin(), m_objects.end(), result);

    for(std::set<ObjectID>::iterator it = tmp.begin();
        it != tmp.end(); ++it)
    {
        SceneObject *object = ptr->get_object(*it);
        if(object)
            m_trigger->attach_to(object);
    }
}

void NameGroup::clear_updates()
{
    // Scrap any incoming updates for this group's widgets
    TheUserInterface::instance().remove_updates(m_name_input);
    TheUserInterface::instance().remove_updates(m_meta_name_input);
}
