/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_NAME_GROUP_HH
#define __PEEKABOT_GUI_NAME_GROUP_HH

#include "../Types.hh"
#include "../SceneTreePtr.hh"
#include "PropertyVisitor.hh"
#include "PropertyGroup.hh"
#include <set>
#include <FL/Fl_Button.H>
#include <FL/Fl_Input.H>

namespace peekabot
{

namespace gui
{
    class NameTrigger;


    /** \brief A group of properties and controls for setting name and meta name properties.
     *
     * This group contains two \c Fl_Input widgets for entering the name and meta name properties of object, and a \c Fl_Button for applying changes.
     *
     * It creates and registers a \c NameTrigger with the selected object to
     * fill and update the name and meta name fields.
     *
     */
class NameGroup : public PropertyGroup
{
public:
    /** \brief Creates a new \c NameGroup object.
     *
     * Allocates and adds the sub widgets and fills them with properties by
     * registering the \c NameTrigger for this group.
     */
    NameGroup(int x, int y);

    /** \brief Deletes the trigger allocated for this group.
     */
    ~NameGroup();

    /** \brief Virtual method for filling this group with name information from
     * an object.
     *
     * This method registers a NameTrigger with the currently selected object
     * in order to fill the data fields of the group. It also deregisters
     * triggers for any object that has been deselected.
     *
     * This method is called by the \c PropertyVisitor upon creation, and
     * whenever the selection changes.
     *
     */
    virtual void set_properties(const std::set<ObjectID> &objects);

    /** \brief Callback for applying name changes.
     *
     * This is called whenever the user clicks the "Apply" button.
     *
     * Will generate a NameChangeAction based on the values entered by the user.
     *
     * \param w A pointer to the \c Fl_Button control triggering the callback.
     * \param data A pointer to the \c NameGroup that the button belongs to.
     */
    static void name_callback(Fl_Widget* w, void* data);

    void register_triggers(const std::set<ObjectID> &objects,
                           SceneTreePtr &ptr);

    void deregister_triggers(const std::set<ObjectID> &objects,
                             SceneTreePtr &ptr);

    /** \brief Removes all pending updates for this group.
     */
    void clear_updates();

private:
    /** \brief A pointer to the Fl_Input widget for entering the object name.
     */
    Fl_Input* m_name_input;

    /** \brief A pointer to the Fl_Input widget for entering the object meta name.
     */
    Fl_Input* m_meta_name_input;

    /** \brief The button for applying changes made in the name fields.
     */
    Fl_Button* m_apply_button;

    /** \brief A pointer to the trigger this group registers with selected
     * objects.
     */
    NameTrigger* m_trigger;

    std::set<ObjectID> m_objects;
};

}
}
#endif //__PEEKABOT_GUI_NAME_GROUP_HH
