/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_NAME_TRIGGER_HH
#define __PEEKABOT_GUI_NAME_TRIGGER_HH

#include <set>

class Fl_Input;

namespace peekabot
{

class SceneObject;

namespace gui
{


    /** \brief Trigger for updating the name fields in a\c NameGroup.
     *
     * This trigger is registered when a \c NameGroup is created in a \c
     * PropertyFrame. It will contain pointers to the name fields that it
     * is registered for.
     *
     * It is registered for the ON_NAME_SET and ON_META_NAME_SET stimuli.
     *
     * When triggered it generates update events and posts them on the GUI update
     * queue via \c UserInterface::post_event().
     *
     */

class NameTrigger
{
public:
    /** \brief This constructor simply sets the pointers to name and meta name
     * fields.
     */
    NameTrigger(Fl_Input* name_field, Fl_Input* meta_name_field);

    ~NameTrigger();
    
    /** \brief Updates the name field of a \c NameGroup.
     *
     * This method will generate an \c UpdateEvent for the corresponding object
     * name and input widget and post it on the GUI update queue.
     */
    void on_name_set(SceneObject *object);

    /** \brief Updates the meta name field of a \c NameGroup.
     *
     * This method will generate an \c UpdateEvent for the corresponding object
     * meta name and input widget and post it on the GUI update queue.
     */
    void on_meta_name_set(SceneObject *object);

    void attach_to(SceneObject *object);
    void detach_from(SceneObject *object);

private:
    /** \brief Pointers to the name fields that the trigger is registered for.
     */
    Fl_Input* m_name_field, *m_meta_name_field;

    /** \brief A set storing all objects the trigger is registered with.
     */
    std::set<SceneObject *> m_objects;
};


}
}

#endif //__PEEKABOT_GUI_NAME_TRIGGER_HH
