/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
//////////////////   PropertyFrame methods   //////////////////////////////
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <FL/Fl.H>
#include "PropertyFrame.hh"
#include "UserInterfaceUtilities.hh"
#include "Workspace.hh"

#include "MoveRotateGroup.hh"
#include "CameraGroup.hh"
#include "ColorGroup.hh"
#include "NameGroup.hh"
#include "VisibilityGroup.hh"
#include "ScaleGroup.hh"
#include "DOFGroup.hh"

using namespace peekabot;
using namespace gui;


PropertyFrame::PropertyFrame(int x, int y, int w, int h, char* l) :
    Frame(x, y, w, h, l)
{
    init(x, y, w, h);
    //resizable(0);
    select_objects(TheUIUtilities::instance().get_selection());
}

void PropertyFrame::init(int x, int y, int w, int h)
{    
    Fl_Box* o = new Fl_Box(Frame::x()+31, Frame::y(), 210, 25, "Property inspector");
    o->box(FL_PLASTIC_UP_BOX);
    o->color(FL_LIGHT3);
    o->labeltype(FL_ENGRAVED_LABEL);
    o->labelcolor((Fl_Color)43);
    add(o);

    m_group_pack = new Fl_Pack(x+5, y+30, 230, h-35);
    add(m_group_pack);
   
    m_move_rotate_group = 0;
    m_camera_group = 0;
    m_color_group = 0;
    m_name_group = 0;
    m_scale_group = 0;
    m_visibility_group = 0;
    m_dof_group = 0;

    m_property_builder = new PropertyVisitor(this);
}

PropertyFrame::~PropertyFrame()
{
    delete m_property_builder;
}

PropertyFrame::PropertyFrame(Frame* frame) :
    Frame(frame->x(), frame->y(), frame->w(), frame->h())
{
    // Give the new frame the same size as the one it replaces
    init(frame->x(), frame->y(), frame->w(), frame->h());

    // Register the frame with the currently active workspace
    TheUIUtilities::instance().active_workspace()->register_property_frame(this);
    
    // Since the old frame must have been selected to use the content selection
    // menu, we want the new frame to be selected as well.
    TheUIUtilities::instance().active_workspace()->select_frame(this);

    // Remove the old frame safely.
    frame->deregister();
    frame->hide();
    Fl::delete_widget(frame);

    select_objects(TheUIUtilities::instance().get_selection());
    redraw();   
}

void PropertyFrame::select_objects(const std::set<ObjectID> &objects)
{
    m_property_builder->visit_multiple(objects);
    redraw();
}

void PropertyFrame::deregister()
{
    Workspace* ws = (Workspace*)parent();
    ws->deregister_property_frame(this);

    if(m_move_rotate_group != 0)
        m_move_rotate_group->clear_updates();
    if(m_camera_group != 0)
        m_camera_group->clear_updates();
    if(m_color_group != 0)
        m_color_group->clear_updates();
    if(m_name_group != 0)
        m_name_group->clear_updates();
    if(m_scale_group != 0)
        m_scale_group->clear_updates();
    if(m_visibility_group != 0)
        m_visibility_group->clear_updates();
    /*if(m_dof_group != 0)
      m_dof_group->clear_updates();*/
}

Frame::FrameType PropertyFrame::frame_type() const
{
    return Frame::PROPERTY_FRAME;
}


bool PropertyFrame::resizable_width(int width)
{
    return 50;
}

bool PropertyFrame::resizable_height(int height)
{
    return 50;
}

Frame* PropertyFrame::clone(int x, int y, int w, int h)
{
    return new PropertyFrame(x, y, w, h);
}


void PropertyFrame::add_name_group()
{
    if(m_name_group == 0)
    {
        m_name_group = new NameGroup(m_group_pack->x(),
                                         m_group_pack->y());
        m_group_pack->add(m_name_group);
    }
    else
    {
        m_name_group->set_properties(
            TheUIUtilities::instance().get_selection());
        m_name_group->show();
    }
}

void PropertyFrame::remove_name_group()
{
    if(m_name_group != 0)
        m_name_group->hide();
}



void PropertyFrame::add_move_rotate_group()
{
    if(m_move_rotate_group == 0)
    {
        m_move_rotate_group = new MoveRotateGroup(m_group_pack->x(),
                                                  m_group_pack->y());
        m_group_pack->add(m_move_rotate_group);
    }
    else
    {
        m_move_rotate_group->set_properties(
            TheUIUtilities::instance().get_selection());
        m_move_rotate_group->show();
    }
}

void PropertyFrame::remove_move_rotate_group()
{
    if(m_move_rotate_group != 0)
        m_move_rotate_group->hide();
}



void PropertyFrame::add_camera_group()
{
    if(m_camera_group == 0)
    {
        m_camera_group = new CameraGroup(m_group_pack->x(),
                                         m_group_pack->y());
        m_group_pack->add(m_camera_group);
    }
    else
    {
        m_camera_group->set_properties(
            TheUIUtilities::instance().get_selection());
        m_camera_group->show();
    }
}

void PropertyFrame::remove_camera_group()
{
    if(m_camera_group != 0)
        m_camera_group->hide();
}

void PropertyFrame::add_visibility_group()
{
    if(m_visibility_group == 0)
    {
        m_visibility_group = new VisibilityGroup(m_group_pack->x(),
                                                 m_group_pack->y());
        m_group_pack->add(m_visibility_group);
    }
    else
    {
        m_visibility_group->set_properties(
            TheUIUtilities::instance().get_selection());
        m_visibility_group->show();
    }
}

void PropertyFrame::remove_visibility_group()
{
    if(m_visibility_group != 0)
        m_visibility_group->hide();
}


void PropertyFrame::add_color_group()
{
    if(m_color_group == 0)
    {
        m_color_group = new ColorGroup(m_group_pack->x(),
                                                 m_group_pack->y());
        m_group_pack->add(m_color_group);
    }
    else
    {
        m_color_group->set_properties(
            TheUIUtilities::instance().get_selection());
        m_color_group->show();
    }
}

void PropertyFrame::remove_color_group()
{
    if(m_color_group != 0)
        m_color_group->hide();
}

void PropertyFrame::add_scale_group()
{
    if(m_scale_group == 0)
    {
        m_scale_group = new ScaleGroup(m_group_pack->x(),
                                                 m_group_pack->y());
        m_group_pack->add(m_scale_group);
    }
    else
    {
        m_scale_group->set_properties(
            TheUIUtilities::instance().get_selection());
        m_scale_group->show();
    }
}

void PropertyFrame::remove_scale_group()
{
    if(m_scale_group != 0)
        m_scale_group->hide();
}

void PropertyFrame::add_dof_group()
{
    /*if(m_dof_group == 0)
    {
        m_dof_group = new DOFGroup(m_group_pack->x(),
                                   m_group_pack->y());
        m_group_pack->add(m_dof_group);
    }
    else
    {
        m_dof_group->set_properties(
            TheUIUtilities::instance().get_selection());
        m_dof_group->show();
    }*/
}

void PropertyFrame::remove_dof_group()
{
    /*if(m_dof_group != 0)
      m_dof_group->hide();*/
}
