/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_PROPERTY_FRAME_HH
#define __PEEKABOT_GUI_PROPERTY_FRAME_HH

#include <set>
#include <FL/Fl_Menu_Button.H>
#include <FL/Fl_Pack.H>
#include <FL/Fl_Box.H>
#include "Frame.hh"
#include "../Types.hh"

namespace peekabot
{

    class SceneObject;


namespace gui
{
    class UserInterfaceUtilities;
    class PropertyVisitor;
    class Workspace;    


    class MoveRotateGroup;
    class CameraGroup;
    class ColorGroup;
    class NameGroup;
    class ScaleGroup;
    class VisibilityGroup;
    class DOFGroup;

    /** \brief Represents property inspector frames.
     *
     * This subclass of \c Frame presents the user with controls for examining and changing the properties of objects in the scene. 
     *
     * It uses a \c PropertyVisitor to extract a context sensitive selection of properties depending on the type of the selected object(s).
     *
     * It registers \c PropertyTrigger objects to keep these properties updated.
     *
     * It inherits the frame content selection menu from \c Frame.
     * 
     * \sa \ref triggers
     */



class PropertyFrame : public Frame
{
public:
    /** \brief Creates a new property inspector frame.
     *
     * This method calls \c init() to initialize the frame.
     *
     * It will also call the constructor in \c Frame, in order to get the frame content selection menu.
     *
     * \param x The x-coordinate of the top left corner of this frame.
     * \param y The y-coordinate of the top left corner of this frame.
     * \param w The width of this frame.
     * \param h The height of this frame.
     * \param l The widget label (for FLTK compatibility)
     */
    PropertyFrame(int x, int y, int w, int h, char* l = 0);


    /** \brief Creates a new property inspector frame.
     *
     * The new frame will be created to fit the space occupied by \c frame. The constructor will also take care of registering the frame with the active workspace, and of removing the old frame.
     *
     * \param frame The \c Frame to replace.
     */
    PropertyFrame(Frame* frame);

    /** \brief Deletes the \c PropertyVisitor allocated for the frame.
     */
    ~PropertyFrame();
    
    /** \brief Generates property groups for the currently selected object(s).
     *
     * This method deletes the current \c PropertyGroup objects in this frame and creates new ones by creating a \c PropertyVisitor and calling PropertyVisitor::visit_multiple() on \c objects.
     *
     * This will generate \c PropertyGroup objects and add them to this frame. It will also register \c PropertyTrigger objects (see \ref triggers) for these properties.
     *
     * This function is typically called during construction of a new \c PropertyFrame, and by \c UserInterfaceUtilities::select() whenever the current selection changes.
     *
     * \param objects The list of currently selected objects.
     */
    void select_objects(const std::set<ObjectID> &objects);

    /** \brief Deregisters the \c Frame with its \c Workspace.
     */
    virtual void deregister();

    /** \brief Returns true if the frame can be resized to width \c width.
     *
     * Currently, the minimum width is 50 for property frames.
     */
    virtual bool resizable_width(int width);
    /** \brief Returns true if the frame can be resized to height \c height.
     * 
     * Currently, the minimum height is 50 for property frames.
     */
    virtual bool resizable_height(int height);

    /** \brief Creates a clone of the frame with the specified screen
     * dimensions.
     */
    virtual Frame* clone(int x, int y, int w, int h);

    void add_move_rotate_group();
    void remove_move_rotate_group();

    void add_camera_group();
    void remove_camera_group();

    void add_color_group();
    void remove_color_group();

    void add_name_group();
    void remove_name_group();

    void add_scale_group();
    void remove_scale_group();

    void add_visibility_group();
    void remove_visibility_group();

    void add_dof_group();
    void remove_dof_group();

protected:

    /** \brief Initialize a new property frame.
     *
     * The property inspector will show the properties for the currently selected object(s) by calling \c select_objects() on \c UserInterfaceUtilities::get_scelection().
     *
     *
     * \param x The x-coordinate of the top left corner of this frame.
     * \param y The y-coordinate of the top left corner of this frame.
     * \param w The width of this frame.
     * \param h The height of this frame.
     */
    void init(int x, int y, int w, int h);


    /** \brief Get the type of the frame.
     *
     * \return A \c FrameType of value PROPERTY_FRAME
     */
    virtual FrameType frame_type() const;

    PropertyVisitor *m_property_builder;

    Fl_Pack* m_group_pack;

    MoveRotateGroup* m_move_rotate_group;

    CameraGroup* m_camera_group;

    ColorGroup* m_color_group;

    NameGroup* m_name_group;

    ScaleGroup* m_scale_group;

    VisibilityGroup* m_visibility_group;

    DOFGroup* m_dof_group;

    //MaterialGroup* m_material_group;

};

}
}
#endif // __PEEKABOT_GUI_PROPERTY_FRAME_HH
