/*
 * Copyright Anders Boberg 2006-2007.
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "PropertyVisitor.hh"
#include "PropertyFrame.hh"
#include "../SceneObject.hh"
//#include "../GeometricPrimitives.hh"
#include "../GroupObject.hh"
#include "../Types.hh"
#include "../SceneTreePtr.hh"

#include <algorithm>

using namespace peekabot;
using namespace gui;

/* Remember property types:

        MOVE_ROTATE = 1;
        CAMERA = 2;
        COLOR = 4;
        NAME = 8;
        SCALE = 16;
        TESSELLATION = 32;
        VISIBILITY = 64;
        MATERIAL = 128;
*/


PropertyVisitor::PropertyVisitor(PropertyFrame* frame) :
    m_frame(frame), m_properties(0) {}

void PropertyVisitor::visit_multiple(const std::set<ObjectID> &objects)
{
    int32_t previous = m_properties;

    if(objects.size() > 0)
    {
    // Start with the union of all properties
    m_properties = (MOVE_ROTATE | CAMERA | COLOR | NAME | SCALE | TESSELLATION |
                    VISIBILITY | MATERIAL | DOF);
    }
    else
        m_properties = 0;

    SceneTreePtr ptr = TheSceneTree::acquire(); // Lock the scene tree

    // Visit all objects in the selection, and-ing together the applicable 
    // properties with m_properties, creating the intersection of properties.
    std::set<ObjectID>::iterator it;
    for(it = objects.begin(); it != objects.end(); ++it)
    {
        SceneObject *object = ptr->get_object(*it);
        // When multiple objects are deleted, the selection set 
 	  	// temporarily contains invalid IDs. Watch out for those. 
        if( object )
            object->accept(this);
    }

    // If the selection was multiple, we disallow the NAME property, but allow the SCALE property(NOT!).
    if (objects.size() > 1)
        m_properties = m_properties & (MOVE_ROTATE | COLOR | CAMERA | TESSELLATION | VISIBILITY);// | SCALE;

    // Create the property groups...

    // Do we have the NAME property?
    if(m_properties & NAME)
        m_frame->add_name_group();
    else if(previous & NAME)
        m_frame->remove_name_group();


    // ... the MOVE_ROTATE property?
    if(m_properties & MOVE_ROTATE)
        m_frame->add_move_rotate_group();
    else if (previous & MOVE_ROTATE) // The previous group should be removed
        m_frame->remove_move_rotate_group();
    
    
    if(m_properties & CAMERA)
        m_frame->add_camera_group();
    else if(previous & CAMERA)
        m_frame->remove_camera_group();


    if(m_properties & VISIBILITY)
        m_frame->add_visibility_group();
    else if(previous & VISIBILITY)
        m_frame->remove_visibility_group();

    
    if(m_properties & COLOR)
        m_frame->add_color_group();
    else if(previous & COLOR)
        m_frame->remove_color_group();
    


    if(m_properties & SCALE)
        m_frame->add_scale_group(); 
    else if(previous & SCALE)
        m_frame->remove_scale_group();

    if(m_properties & DOF)
        m_frame->add_dof_group(); 
    else if(previous & DOF)
        m_frame->remove_dof_group();


    /*
      if(m_properties & MATERIAL)
      m_frame->add_material_group();
      else if(previous & MATERIAL)
      m_frame->remove_material_group();*/
}

void PropertyVisitor::visit(CameraObject* camera_object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR | CAMERA);
}

void PropertyVisitor::visit(ModelObject* model_object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR | SCALE);
}

void PropertyVisitor::visit(SphereObject* sphere_object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR | SCALE | TESSELLATION);
}

void PropertyVisitor::visit(GroupObject* group_object)
{
    // For the root object, only the VISIBILITY property is applicable.
    /*  if(group_object->get_parent() == 0)
        m_properties = m_properties & (VISIBILITY );
        else*/
        m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR);
}


void PropertyVisitor::visit(CubeObject* cube_object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR | SCALE);
}

void PropertyVisitor::visit(PolygonObject* polygon_object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR | SCALE);
}

void PropertyVisitor::visit(CylinderObject* cylinder_object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR | SCALE | TESSELLATION);
}

void PropertyVisitor::visit(CircleObject* circle_object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR | SCALE | TESSELLATION);
}

void PropertyVisitor::visit(PolyLineObject* poly_line_object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR | SCALE);
}

void PropertyVisitor::visit(SensorObject* sensor_object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR);
}


void PropertyVisitor::visit(PointSetObject* object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR);
}


void PropertyVisitor::visit(Hinge* object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR | DOF );
}


void PropertyVisitor::visit(Slider* object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR | DOF);
}

void PropertyVisitor::visit(GridObject* object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR | SCALE | TESSELLATION);
}

void PropertyVisitor::visit(PointCloud* object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR | SCALE);
}

void PropertyVisitor::visit(LineCloud* object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR | SCALE);
}

void PropertyVisitor::visit(OccupancyGrid2D* object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR | SCALE);
}

void PropertyVisitor::visit(Label* object)
{
    m_properties = m_properties & (MOVE_ROTATE | NAME | VISIBILITY | COLOR | SCALE);
}



