/*
 * Copyright Anders Boberg 2006-2007.
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PEEKABOT_GUI_PROPERTY_VISITOR_HH_INCLUDED
#define PEEKABOT_GUI_PROPERTY_VISITOR_HH_INCLUDED

#include <set>

#include "../ObjectVisitor.hh"
#include "../Types.hh"

namespace peekabot
{

    class SceneObject;

namespace gui
{
    /** \brief Enum describing the types of properties available to an object
     *
     * The values are powers of 2, so that the properties can be manipulated
     * by simple bit operations.
     */
    enum PropertyTypes
    {
        MOVE_ROTATE = 1,
        CAMERA = 2,
        COLOR = 4,
        NAME = 8,
        SCALE = 16,
        TESSELLATION = 32,
        VISIBILITY = 64,
        MATERIAL = 128,
        DOF = 256,
        ROBOT = 512
    };    

    class PropertyFrame;

    /** \brief Visitor for extracting subclass-specific properties from \c SceneObject-derived classes.
     *
     * This visitor is used by the \c PropertyFrame class as it fills its frame with \c PropertGroups applicable to the currently selected object(s).
     *
     */
class PropertyVisitor : public ObjectVisitor
{
public:
     /** \brief Creates a new \c PropertyVisitor for a \c PropertyFrame.
     *
     * This constructor simply sets the m_frame property.
     */
    PropertyVisitor(PropertyFrame* frame);



    /** \brief Generates groups for multiple selections.
     *
     * Will generate the intersection of properties available to the set of
     * selected objects.
     *
     */
    void visit_multiple(const std::set<ObjectID> &objects);




    /** \brief Creates \c PropertyGroup objects specific to CameraObject.
     *
     * Generates a \c MoveRotateGroup, a \c NameGroup, a \c VisibilityGroup and a \c CameraGroup and calls their respective \c set_properties() methods for the object being visited.
     *
     * \param camera_object A pointer to the \c CameraObject being visited.
     */
    virtual void visit(CameraObject* camera_object);



    /** \brief Creates \c PropertyGroup objects specific to ModelObject.
     *
     * Generates a \c MoveRotateGroup, a \c NameGroup, a \c ScaleGroup and a \c VisibilityGroup and calls their respective \c set_properties() methods for the object being visited.
     *
     * \param model_object A pointer to the \c ModelObject being visited.
     */
    virtual void visit(ModelObject* model_object);



    /** \brief Creates \c PropertyGroup objects specific to SphereObject.
     *
     * Generates a \c MoveRotateGroup, a \c NameGroup, a \c ScaleGroup, a \c ColorGroup, a \c TessellationGroup and a \c VisibilityGroup and calls their respective \c set_properties() methods for the object being visited.
     *
     * \param sphere_object A pointer to the \c SphereObject being visited.
     */
    virtual void visit(SphereObject* sphere_object);



    /** \brief Creates \c PropertyGroup objects specific to GroupObject.
     *
     * Generates a \c MoveRotateGroup, a \c NameGroup and a \c VisibilityGroup and calls their respective \c set_properties() methods for the object being visited.
     *
     * \param group_object A pointer to the \c GroupObject being visited.
     */
    virtual void visit(GroupObject* group_object);



    /** \brief Creates \c PropertyGroup objects specific to CubeObject.
     *
     * Generates a \c MoveRotateGroup, a \c NameGroup, a \c ScaleGroup, a \c ColorGroup and a \c VisibilityGroup and calls their respective \c set_properties() methods for the object being visited.
     *
     * \param cube_object A pointer to the \c CubeObject being visited.
     */
    virtual void visit(CubeObject* cube_object);


    /** \brief Creates \c PropertyGroup objects specific to PolygonObject.
     *
     * Generates a \c MoveRotateGroup, a \c NameGroup, a \c ColorGroup and a \c VisibilityGroup and calls their respective \c set_properties() methods for the object being visited.
     *
     * \param polygon_object A pointer to the \c PolygonObject being visited.
     */
    virtual void visit(PolygonObject* polygon_object);



    /** \brief Creates \c PropertyGroup objects specific to CylinderObject.
     *
     * Generates a \c MoveRotateGroup, a \c NameGroup, a \c ScaleGroup, a \c ColorGroup, a \c TessellationGroup and a \c VisibilityGroup and calls their respective \c set_properties() methods for the object being visited.
     *
     * \param cylinder_object A pointer to the \c CylinderObject being visited.
     */
    virtual void visit(CylinderObject* cylinder_object);



    /** \brief Creates \c PropertyGroup objects specific to CircleObject.
     *
     * Generates a \c MoveRotateGroup, a \c NameGroup, a \c ScaleGroup, a \c ColorGroup, a \c TessellationGroup and a \c VisibilityGroup and calls their respective \c set_properties() methods for the object being visited.
     *
     * \param circle_object A pointer to the \c CircleObject being visited.
     */
    virtual void visit(CircleObject* circle_object);



    /** \brief Creates \c PropertyGroup objects specific to PolyLineObject.
     *
     * Generates a \c MoveRotateGroup, a \c NameGroup, a \c ScaleGroup, a \c ColorGroup and a \c VisibilityGroup and calls their respective \c set_properties() methods for the object being visited.
     *
     * \param line_object A pointer to the \c PolyLineObject being visited.
     */
    virtual void visit(PolyLineObject* poly_line_object);



    /** \brief Creates \c PropertyGroup objects specific to SensorObject.
     *
     * Generates a \c MoveRotateGroup, a \c NameGroup and a \c VisibilityGroup and calls their respective \c set_properties() methods for the object being visited.
     *
     * \param sensor_object A pointer to the \c SensorObject being visited.
     */
    virtual void visit(SensorObject* sensor_object);

    virtual void visit(PointSetObject* object);

    virtual void visit(Hinge* object);

    virtual void visit(Slider* object);

    virtual void visit(GridObject* object);

    virtual void visit(PointCloud* object);

    virtual void visit(LineCloud* object);

    virtual void visit(OccupancyGrid2D* object);

    virtual void visit(Label* object);

private:
    /** \brief The \c PropertyFrame this \c PropertyVisitor is generating \c PropertyGroups for.
     */
    PropertyFrame* m_frame;

    /** \brief Bit field describing the applicable properties for the current
     *  selection.
     */
    int m_properties;
};


}
}

#endif // PEEKABOT_GUI_PROPERTY_VISITOR_HH_INCLUDED
