/*
 * Copyright Anders Boberg 2007.
 * Copyright Staffan Gimåker 2007-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <FL/Fl_Check_Button.H>
#include <FL/Fl_Button.H>

#include "ScaleGroup.hh"
#include "ScaleTrigger.hh"
#include "../ScalableObject.hh"
#include "../SceneTreePtr.hh"
#include "UserInterface.hh"
#include "UserInterfaceUtilities.hh"
#include "TransformationInput.hh"


using namespace peekabot;
using namespace gui;

ScaleGroup::ScaleGroup(int x, int y) :
    PropertyGroup(x, y, 230, 75, "Scale")
{
    hide();
    begin();
    m_x_scale = new ScaleInput(x + 25, y + 25, 50, 20, "X");
    m_x_scale->minimum(0.001);
    m_x_scale->maximum(1000000);
    m_x_scale->step(0.001);
    m_x_scale->value(1);
    m_x_scale->set_uniform(true);

    m_y_scale = new ScaleInput(x + 95, y + 25, 50, 20, "Y");
    m_y_scale->minimum(0.001);
    m_y_scale->maximum(1000000);
    m_y_scale->step(0.001);
    m_y_scale->value(1);
    m_y_scale->set_uniform(true);

    m_z_scale = new ScaleInput(x + 165, y + 25, 50, 20, "Z");
    m_z_scale->minimum(0.001);
    m_z_scale->maximum(1000000);
    m_z_scale->step(0.001);
    m_z_scale->value(1);
    m_z_scale->set_uniform(true);

    m_x_scale->set_xyz_links(m_x_scale, m_y_scale, m_z_scale);
    m_y_scale->set_xyz_links(m_x_scale, m_y_scale, m_z_scale);
    m_z_scale->set_xyz_links(m_x_scale, m_y_scale, m_z_scale);

    m_uniform_button = new Fl_Check_Button(x + 90, y + 50, 20, 20, "Uniform");
    m_uniform_button->value(1);
    m_uniform_button->callback(&uniform_callback, this);

    end();
    set_open(false);
    show();

    m_trigger = new ScaleTrigger(m_x_scale, m_y_scale, m_z_scale);

    set_properties(TheUIUtilities::instance().get_selection());
}

ScaleGroup::~ScaleGroup()
{
    delete m_trigger;
}


void ScaleGroup::uniform_callback(Fl_Widget* w, void* data)
{
    Fl_Check_Button *uniform_button = static_cast<Fl_Check_Button*>(w);
    ScaleGroup* group = static_cast<ScaleGroup*>(data);

    bool uniform = uniform_button->value() == 1;

    group->m_x_scale->set_uniform(uniform);
    group->m_y_scale->set_uniform(uniform);
    group->m_z_scale->set_uniform(uniform);
}

void ScaleGroup::set_properties(const std::set<ObjectID> & objects)
{
    if(!objects.empty())
    {
        activate();

        // Lock the scene tree to make sure nothing is modified while we build.
        SceneTreePtr ptr = TheSceneTree::acquire();
        
        // Deregister triggers
        deregister_triggers(objects, ptr);

        // Register new ones
        register_triggers(objects, ptr); 

        m_objects = objects;  
    }
    else
        deactivate();
}

void ScaleGroup::register_triggers(const std::set<ObjectID> &objects,
                                   SceneTreePtr &ptr)
{
    // The set difference objects - m_objects gives all objects that
    // weren't previously selected, but are now.
    std::set<ObjectID> tmp;
    std::insert_iterator<std::set<ObjectID> > result(tmp, tmp.begin());

    std::insert_iterator<std::set<ObjectID> > end =
        std::set_difference(objects.begin(), objects.end(),
                            m_objects.begin(), m_objects.end(), result);

    // Register triggers for all newly selected objects.
    for(std::set<ObjectID>::iterator it = tmp.begin();
        it != tmp.end(); ++it)
    {
        SceneObject *object = ptr->get_object(*it);
        if(object)
            m_trigger->attach_to(object);
    }
}

void ScaleGroup::deregister_triggers(const std::set<ObjectID> &objects,
                                     SceneTreePtr &ptr)
{
    // The set difference m_objects - objects gives all objects that
    // were previously selected, but are no longer.
    std::set<ObjectID> tmp;
    std::insert_iterator<std::set<ObjectID> > result(tmp, tmp.begin());

    std::insert_iterator<std::set<ObjectID> > end =
        std::set_difference(m_objects.begin(), m_objects.end(),
                            objects.begin(), objects.end(), result);

    // Deregister triggers for all objects that have been deselected.
    for(std::set<ObjectID>::iterator it = tmp.begin();
        it != tmp.end(); ++it)
    {
        SceneObject *object = ptr->get_object(*it);
        if(object)
            m_trigger->detach_from(object);
    }
}

void ScaleGroup::clear_updates()
{
    // Scrap any incoming updates for this group's widgets
    TheUserInterface::instance().remove_updates(m_x_scale);
    TheUserInterface::instance().remove_updates(m_y_scale);
    TheUserInterface::instance().remove_updates(m_z_scale);

}

