/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_SCALEGROUP_HH
#define __PEEKABOT_GUI_SCALEGROUP_HH


#include "PropertyGroup.hh"
#include "../SceneTreePtr.hh"

class Fl_Check_Button;

namespace peekabot
{

namespace gui
{
    
    class ScaleTrigger;
    class ScaleInput;

    /** \brief A group controls for scaling objects uniformly or along specific axes.
     *
     * This group contains three \c Fl_Value_Input widgets for entering the scale
     * factor along each axis. It also includes a checkbox for uniform scaling.
     *
     * The scale values for single-selected objects are fetched and updated by
     * the \c ScaleTrigger class.
     *
     */
class ScaleGroup : public PropertyGroup
{   
public:
    /** \brief Creates a new \c ScaleGroup object.
     *
     * Allocates and adds the sub widgets and fills them with properties by
     * creating and registering the \c ScaleTrigger.
     */
    ScaleGroup(int x, int y);

    /** \brief Deletes the trigger allocated for this group.
     */
    virtual ~ScaleGroup();

    /** \brief Virtual method for filling this group with scale information from an object.
     *
     * This method registers a ScaleTrigger with the currently selected
     * object in order to fill the input fields with scale information. When
     * the selection changes it deregisters triggers for any object that has
     * been deselected.
     *
     * This method is called by the \c PropertyVisitor upon creation, and
     * whenever the selection changes.
     *
     */
    virtual void set_properties(const std::set<ObjectID> &objects);

    /** \brief Callback for scaling controls
     *
     * This is called whenever the user manipulates one of the \c Fl_Value_Input controls for scaling.
     *
     * Will generate a SetScale action based on the values entered by the user.
     *
     * \param w A pointer to the \c Fl_Value_Input control triggering the callback.
     * \param data A pointer to void, as required by FLTK (not used).
     */
    static void scale_callback(Fl_Widget* w, void* data);

    /** \brief Registers a \c ScaleTrigger with all objects in the set
     * passed to the method.
     */
    void register_triggers(const std::set<ObjectID> &objects,
                           SceneTreePtr &ptr);
    /** \brief Deregisters the \c ScaleTrigger from all objects in the set
     * passed to the method.
     */
    void deregister_triggers(const std::set<ObjectID> &objects,
                             SceneTreePtr &ptr);


    /** \brief Removes all pending updates for this group.
     */
    void clear_updates();

    /** \brief Sets the uniform/non-uniform property of the \c ScaleInput widgets
     *
     * Called when the user checks/unchecks the "uniform" check button.
     */
    static void uniform_callback(Fl_Widget* w, void* data);

private:
    ScaleInput *m_x_scale, *m_y_scale, *m_z_scale;

    Fl_Check_Button* m_uniform_button;

    ScaleTrigger *m_trigger;
};

}
}
#endif //__PEEKABOT_GUI_SCALEGROUP_HH
