/*
 * Copyright Anders Boberg 2007.
 * Copyright Staffan Gimåker 2007, 2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <FL/Fl_Output.H>
#include <FL/Fl.H>
#include <FL/fl_draw.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Choice.H>
#include <boost/bind.hpp>
#include <boost/format.hpp>
#include <string>
#include <cstdlib>
#include <iostream>

#include "StatusBar.hh"
#include "UserInterface.hh"
#include "UserInterfaceUtilities.hh"
#include "Manipulator.hh"
#include "ToolButton.hh"
#include "CommandRepository.hh"
#include "../Configuration.hh"
#include "../PrintfFormatter.hh"
#include "UpdateActions.hh"

using namespace peekabot;
using namespace gui;

#define D_FRAME FL_FREE_BOXTYPE

namespace peekabot
{
    namespace gui
    {      
        char x_axis_tooltip[128];
        char y_axis_tooltip[128];
        char z_axis_tooltip[128];

        char rot_coord_syst_tooltip[128];
        char mov_coord_syst_tooltip[128];
    }
}

StatusBar::StatusBar(int x, int y, int w, int h, char *l)
    : Fl_Group(x, y, w, h, l),
      m_select_count(0)
{
    end();

    // Register our custom box type.
    Fl::set_boxtype(D_FRAME, &StatusBar::frame_draw, 0, 0, 0, 0);

    box(FL_FLAT_BOX);

    sprintf(x_axis_tooltip, "%s: Enable/disable transformation along x-axis",
            TheConfiguration::instance().get_option<std::string>(
                "shortcuts.x_axis", "x").c_str());
    sprintf(y_axis_tooltip, "%s: Enable/disable transformation along y-axis",
            TheConfiguration::instance().get_option<std::string>(
                "shortcuts.y_axis", "y").c_str());  
    sprintf(z_axis_tooltip, "%s: Enable/disable transformation along z-axis",
            TheConfiguration::instance().get_option<std::string>(
                "shortcuts.z_axis", "z").c_str());
    sprintf(mov_coord_syst_tooltip, "Coordinate system of the translatational axes (%s cycles)",
            TheConfiguration::instance().get_option<std::string>(
                "shortcuts.move_mode", "r").c_str());
    sprintf(rot_coord_syst_tooltip, "Coordinate system of the rotational axis (%s cycles)",
            TheConfiguration::instance().get_option<std::string>(
                "shortcuts.rotate_mode", "r").c_str());
   
    
    m_default_group = init_default_group();
    m_move_group = init_move_group();
    m_rotate_group = init_rotate_group();

    Manipulator *manipulator = TheUIUtilities::instance().get_manipulator();
    set_mode(manipulator->get_mode());
    manipulator->get_mode_signal().connect(
        boost::bind(&StatusBar::set_mode, this, _1));

    update_count(0, 0);

    show();

    TheMessageHub::instance().subscribe(INFO_MESSAGE, this);
    TheMessageHub::instance().subscribe(WARNING_MESSAGE, this);
    TheMessageHub::instance().subscribe(ERROR_MESSAGE, this);

    TheUIUtilities::instance().on_object_selected().connect(
        boost::bind(&StatusBar::update_count, this, 1, _1));
    TheUIUtilities::instance().on_object_deselected().connect(
        boost::bind(&StatusBar::update_count, this, -1, _1));
    //box(FL_THIN_UP_BOX);
}

StatusBar::~StatusBar()
{
    TheMessageHub::instance().unsubscribe(INFO_MESSAGE, this);
    TheMessageHub::instance().unsubscribe(WARNING_MESSAGE, this);
    TheMessageHub::instance().unsubscribe(ERROR_MESSAGE, this);
    /*
      Manipulator *manip = TheUIUtilities::instance().get_manipulator();
      manip->get_axis_signal().disconnect(0);
    */
}

Fl_Group *StatusBar::init_default_group()
{
    Fl_Group *group = new Fl_Group(x(), y(), w(), h());

    m_default_output = new Fl_Output(x()+1, y()+1, (int)(w() * 0.7)-2, h()-2);
    m_default_output->box(D_FRAME);
    m_default_output->color(FL_BACKGROUND_COLOR);
    m_default_output->clear_visible_focus();

    m_default_select = new Fl_Output(x() + (int)(w() * 0.7)+1, y()+1, (int)(w() * 0.3)-2, h()-2);
    m_default_select->box(D_FRAME);
    m_default_select->color(FL_BACKGROUND_COLOR);
    m_default_select->clear_visible_focus();

    group->end();
    group->resizable(m_default_output);
    
    return group;
}

Fl_Group *StatusBar::init_move_group()
{
    Fl_Group *group = new Fl_Group(x(), y(), w(), h());

    m_move_output = new Fl_Output(x()+1, y()+1, (int)((w()-134) * 0.7)-2, h()-2);
    m_move_output->box(D_FRAME);
    m_move_output->color(FL_BACKGROUND_COLOR);
    m_move_output->clear_visible_focus();

    m_move_select = new Fl_Output(x() + (int)((w()-134) * 0.7)+1, y()+1, (int)((w()-134) * 0.3)-2, h()-2);
    m_move_select->box(D_FRAME);
    m_move_select->color(FL_BACKGROUND_COLOR);
    m_move_select->clear_visible_focus();    

    CommandRepository *commands = TheUIUtilities::instance().get_command_repository();
    Manipulator *manip = TheUIUtilities::instance().get_manipulator();

    // Axis buttons
    ToolButton *x_axis = new ToolButton(w()-50, group->y()+4, 15, 18, "X",
                                        commands->get_command(CMD_TOGGLE_X_AXIS));
    x_axis->type(FL_TOGGLE_BUTTON);
    x_axis->selection_color(FL_RED);
    x_axis->tooltip(x_axis_tooltip);
    // Initialize to the correct axis setting for the manipulator.
    x_axis->value(manip->get_axes() & transformation_target::X_AXIS);
    // Connect to the mode signal of the manipulator! This will toggle the
    // button when the setting changes.
    manip->get_axis_signal().connect(boost::bind(&ToolButton::set_if_logical_and, x_axis,
                                                 transformation_target::X_AXIS, _1));
    // Connect to the system signal in order to disable this axis when in camera mode.
    manip->get_system_signal().connect(boost::bind(&ToolButton::activate_if_logical_and, x_axis,
                                                   (transformation_target::LOCAL_COORDINATES | 
                                                    transformation_target::PARENT_COORDINATES |
                                                    transformation_target::WORLD_COORDINATES), _1));
    m_axis_subscribers.insert(x_axis);

    ToolButton *y_axis = new ToolButton(w()-35, group->y()+4, 15, 18, "Y",
                                        commands->get_command(CMD_TOGGLE_Y_AXIS));
    y_axis->type(FL_TOGGLE_BUTTON);
    y_axis->selection_color(FL_GREEN);
    y_axis->tooltip(y_axis_tooltip);
    // Initialize to the correct axis setting for the manipulator.
    y_axis->value(manip->get_axes() & transformation_target::Y_AXIS);
    // Connect to the mode signal of the manipulator! This will toggle the
    // button when the setting changes.
    manip->get_axis_signal().connect(boost::bind(&ToolButton::set_if_logical_and, y_axis,
                                                 transformation_target::Y_AXIS, _1));
    m_axis_subscribers.insert(y_axis);

    ToolButton *z_axis = new ToolButton(w()-20, group->y()+4, 15, 18, "Z",
                                        commands->get_command(CMD_TOGGLE_Z_AXIS));
    z_axis->type(FL_TOGGLE_BUTTON);
    z_axis->tooltip(z_axis_tooltip);
    z_axis->selection_color(FL_BLUE);
    // Initialize to the correct axis setting for the manipulator.
    z_axis->value(manip->get_axes() & transformation_target::Z_AXIS);
    // Connect to the mode signal of the manipulator! This will toggle the
    // button when the setting changes.
    manip->get_axis_signal().connect(boost::bind(&ToolButton::set_if_logical_and, z_axis,
                                                 transformation_target::Z_AXIS, _1));
    m_axis_subscribers.insert(z_axis);

    //Separator.
    Fl_Box * sep2 = new Fl_Box(w()-55, group->y()+3, 3, 20);
    sep2->box(D_FRAME);
    
    //Coordinate system menu button
    Fl_Choice *coord_menu = new Fl_Choice(w()-131, group->y()+4, 74, 18);
    coord_menu->clear_visible_focus();
    coord_menu->tooltip(mov_coord_syst_tooltip);
    coord_menu->add("Camera",0, &StatusBar::menu_callback,
                    &commands->get_command(CMD_SET_CAMERA_COORDINATES));
    coord_menu->add("World",0, &StatusBar::menu_callback,
                    &commands->get_command(CMD_SET_WORLD_COORDINATES));
    coord_menu->add("Local",0, &StatusBar::menu_callback,
                    &commands->get_command(CMD_SET_LOCAL_COORDINATES));
    coord_menu->add("Parent",0, &StatusBar::menu_callback,
                    &commands->get_command(CMD_SET_PARENT_COORDINATES));
    // Connect choice button to system signal and set it to the current system.
    manip->get_system_signal().connect(boost::bind(&StatusBar::set_system_item,
                                                   coord_menu, _1));
    set_system_item(coord_menu, manip->get_system());

    group->end();
    group->resizable(m_move_output);

    return group;
}

Fl_Group *StatusBar::init_rotate_group()
{
    Fl_Group *group = new Fl_Group(x(), y(), w(), h());

    m_rotate_output = new Fl_Output(x()+1, y()+1, (int)((w()-215) * 0.7)-2, h()-2);
    m_rotate_output->box(D_FRAME);
    m_rotate_output->color(FL_BACKGROUND_COLOR);
    m_rotate_output->clear_visible_focus();  
    
    m_rotate_select = new Fl_Output(x() + (int)((w()-215) * 0.7)+1, y()+1, (int)((w()-215) * 0.3)-2, h()-2);
    m_rotate_select->box(D_FRAME);
    m_rotate_select->color(FL_BACKGROUND_COLOR);
    m_rotate_select->clear_visible_focus();  

    CommandRepository *commands = TheUIUtilities::instance().get_command_repository();
    Manipulator *manip = TheUIUtilities::instance().get_manipulator();

    // Axis buttons
    ToolButton *x_axis = new ToolButton(w()-50, group->y()+4, 15, 18, "X",
                                        commands->get_command(CMD_SET_X_AXIS));
    x_axis->type(FL_TOGGLE_BUTTON);
    x_axis->selection_color(FL_RED);
    x_axis->tooltip(x_axis_tooltip);
    // Initialize to the correct axis setting for the manipulator.
    x_axis->value(manip->get_axes() & transformation_target::X_AXIS);
    // Connect to the mode signal of the manipulator! This will toggle the
    // button when the setting changes.
    manip->get_axis_signal().connect(boost::bind(&ToolButton::set_if_logical_and, x_axis,
                                                 transformation_target::X_AXIS, _1));
    m_axis_subscribers.insert(x_axis);
    
    ToolButton *y_axis = new ToolButton(w()-35, group->y()+4, 15, 18, "Y",
                                        commands->get_command(CMD_SET_Y_AXIS));
    y_axis->type(FL_TOGGLE_BUTTON);
    y_axis->selection_color(FL_GREEN);
    y_axis->tooltip(y_axis_tooltip);
    // Initialize to the correct axis setting for the manipulator.
    y_axis->value(manip->get_axes() & transformation_target::Y_AXIS);
    // Connect to the mode signal of the manipulator! This will toggle the
    // button when the setting changes.
    manip->get_axis_signal().connect(boost::bind(&ToolButton::set_if_logical_and, y_axis,
                                                 transformation_target::Y_AXIS, _1));
    m_axis_subscribers.insert(y_axis);
    
    ToolButton *z_axis = new ToolButton(w()-20, group->y()+4, 15, 18, "Z",
                                        commands->get_command(CMD_SET_Z_AXIS));
    z_axis->type(FL_TOGGLE_BUTTON);
    z_axis->tooltip(z_axis_tooltip);
    z_axis->selection_color(FL_BLUE);
    // Initialize to the correct axis setting for the manipulator.
    z_axis->value(manip->get_axes() & transformation_target::Z_AXIS);
    // Connect to the mode signal of the manipulator! This will toggle the
    // button when the setting changes.
    manip->get_axis_signal().connect(boost::bind(&ToolButton::set_if_logical_and, z_axis,
                                                 transformation_target::Z_AXIS, _1));    
    m_axis_subscribers.insert(z_axis);

    //Separator.
    Fl_Box * sep2 = new Fl_Box(w()-55, group->y()+3, 3, 20);
    sep2->box(D_FRAME);

    //Coordinate system menu button
    Fl_Choice *coord_menu = new Fl_Choice(w()-131, group->y()+4, 74, 18);
    coord_menu->clear_visible_focus();
    coord_menu->tooltip(rot_coord_syst_tooltip);
    coord_menu->add("Camera",0, &StatusBar::menu_callback,
                    &commands->get_command(CMD_SET_CAMERA_COORDINATES));
    coord_menu->add("World",0, &StatusBar::menu_callback,
                    &commands->get_command(CMD_SET_WORLD_COORDINATES));
    coord_menu->add("Local",0, &StatusBar::menu_callback,
                    &commands->get_command(CMD_SET_LOCAL_COORDINATES));
    coord_menu->add("Parent",0, &StatusBar::menu_callback,
                    &commands->get_command(CMD_SET_PARENT_COORDINATES));
    // Connect choice button to system signal and set it to the current system.
    manip->get_system_signal().connect(boost::bind(&StatusBar::set_system_item,
                                                   coord_menu, _1));
    set_system_item(coord_menu, manip->get_system());

    //Pivot menu button
    Fl_Choice *pivot_menu = new Fl_Choice(w()-210, group->y()+4, 74, 18);
    pivot_menu->clear_visible_focus();
    pivot_menu->tooltip("Rotational pivot");
    pivot_menu->add("Local",0, &StatusBar::menu_callback,
                    &commands->get_command(CMD_SET_LOCAL_PIVOT));
    pivot_menu->add("Parent",0, &StatusBar::menu_callback,
                    &commands->get_command(CMD_SET_PARENT_PIVOT));
    pivot_menu->add("World",0, &StatusBar::menu_callback,
                    &commands->get_command(CMD_SET_WORLD_PIVOT));
    pivot_menu->add("Average",0, &StatusBar::menu_callback,
                    &commands->get_command(CMD_SET_AVERAGE_PIVOT));
    // Connect choice button to system signal and set it to the current system.
    manip->get_pivot_signal().connect(boost::bind(&StatusBar::set_pivot_item,
                                                  pivot_menu, _1));
    set_pivot_item(pivot_menu, manip->get_pivot());

    group->end();
    group->resizable(m_rotate_output);

    return group;
}

void StatusBar::set_mode(ManipulatorMode mode)
{
    if(mode == NAVIGATE_MODE)
    {
        m_default_group->show();
        m_move_group->hide();
        m_rotate_group->hide();
    }
    else if(mode == MOVE_MODE)
    {
        m_default_group->hide();
        m_move_group->show();
        m_rotate_group->hide();
    }
    else if(mode == ROTATE_MODE)
    {
        m_default_group->hide();
        m_move_group->hide();
        m_rotate_group->show();
    }
}

void StatusBar::handle_message(const Message &m) throw()
{
    class UpdateTextOutput : public UpdateAction
    {
    public:
        UpdateTextOutput(Fl_Output *output, const std::string &str)
            : m_output(output),
              m_string(str)
        {
        }

        virtual void execute()
        {
            m_output->value(m_string.c_str());
            m_output->redraw();
        }

        virtual bool replaces()
        {
            return true;
        }

    private:
        Fl_Output *m_output;
        std::string m_string;
    };
    
    if( m.get_type() == INFO_MESSAGE ||
        (m.has_source() && m.get_source() == GUI_SOURCE_ID) )
    {
        // Format the string.
        static PrintfFormatter formatter("%m");
        std::string string = formatter.format(m);
        
        // Post the printing action on the update queue.
        TheUserInterface::instance().post_action(
            m_default_output, new UpdateTextOutput(m_default_output, string));
        
        // Post the printing action on the update queue.
        TheUserInterface::instance().post_action(
            m_move_output, new UpdateTextOutput(m_move_output, string));
        
        // Post the printing action on the update queue.
        TheUserInterface::instance().post_action(
            m_rotate_output, new UpdateTextOutput(m_rotate_output, string));
    }
}

void StatusBar::menu_callback(Fl_Widget *w, void *data)
{
    (*static_cast<CommandRepository::NullaryCommand*>(data))();
}

void StatusBar::set_system_item(Fl_Choice *c,
                                transformation_target::CoordinateSystem system)
{
    int choice = 0;
    switch(system)
    {
        case CAMERA_COORDINATES:
            choice = 0;
            break;
        case WORLD_COORDINATES:
            choice = 1;
            break;
        case LOCAL_COORDINATES:
            choice = 2;
            break;
        case PARENT_COORDINATES:
            choice = 3;
            break;
    }
    c->value(choice);
}


void StatusBar::set_pivot_item(Fl_Choice *c,
                                transformation_target::RotationCenterpoint pivot)
{
    int choice = 0;
    switch(pivot)
    {
        case LOCAL_CENTER:
            choice = 0;
            break;
        case PARENT_CENTER:
            choice = 1;
            break;
        case WORLD_CENTER:
            choice = 2;
            break;
        case AVERAGE_CENTER:
            choice = 3;
            break;
    }
    c->value(choice);
}


void StatusBar::update_count(int change, ObjectID id)
{
    // The sign of change reflects whether an object was
    // selected or deselected.
    m_select_count += change;

    // Special case for singular
    if(m_select_count == 1)
    {
        m_default_select->value("1 object selected");
        m_move_select->value("1 object selected");
        m_rotate_select->value("1 object selected");
    }
    else
    {
        m_default_select->value((boost::format("%1% objects selected")
                                % m_select_count).str().c_str());
        m_move_select->value((boost::format("%1% objects selected")
                                % m_select_count).str().c_str());
        m_rotate_select->value((boost::format("%1% objects selected")
                                % m_select_count).str().c_str());
    }

    m_default_group->redraw();
    m_move_group->redraw();
    m_rotate_group->redraw();

}

void StatusBar::frame_draw(int x, int y, int w, int h, Fl_Color c)
{
    fl_color(c);
    fl_rectf(x, y, w, h);
    fl_color(fl_darker(c));
    fl_line(x, y+h-1, x, y, x+w-1, y);
    fl_color(fl_lighter(c));
    fl_line(x, y+h-1, x+w-1, y+h-1, x+w-1, y);
}
