/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_STATUS_BAR_HH
#define __PEEKABOT_GUI_STATUS_BAR_HH

#include <FL/Fl_Group.H>
#include <set>

#include "../Subscriber.hh"
#include "GUITypes.hh"
#include "Types.hh"

class Fl_Output;

namespace peekabot
{
    namespace gui
    {
        /**
         * \brief Class representing the status bar widget.
         *
         * The status bar is shown in the bottom of the application
         * window. It contains three separate groups of information
         * and settings; one for navigation mode, one for the move
         * tool and one for the rotate tool. Only one of these is
         * visible at any given time, reflecting the current mode.
         *
         * \c StatusBar interacts with the \c Manipulator class
         * through the coordinate system and rotation pivot menu
         * buttons and the axis buttons.
         *
         * It also interacts with \c MessageHub in order to display
         * the latest user console message, and \c
         * UserInterfaceUtilities in order to provide information on
         * the number of objects selected.
         */
        class StatusBar : public Fl_Group, public Subscriber
        {
        public:
            /**
             * \brief Standard FLTK constructor.
             *
             * Sets the dimensions and look of the status bar. It
             * initializes tooltips used by the buttons, and
             * initializes the three property groups by calling
             * the methods \c init_default_group(), \c init_move_group()
             * and \c init_rotate_group().
             *
             * It connects to the mode signal of the \c Manipulator
             * object, in order to display the settings for the current
             * mode.
             *
             * Finally, it subscribes to \c INFO_MESSAGE messages
             * from the \c MessageHub and connects to the selection
             * signals of the \c UserInterfaceUtilities.
             */
            StatusBar(int x, int y, int w, int h, char *l = 0);

            /**
             * \brief Cancels the \c INFO_MESSAGE subscription.
             */
            ~StatusBar();

            /**
             * \brief Sets which group of settings is displayed on
             * the status bar.
             *
             * This method is connected to the mode signal of the
             * \c Manipulator object at construction.
             */
            void set_mode(ManipulatorMode mode);     
            

            /**
             * \brief Prints a message on the status bar.
             *
             * This method is called by the \c MessageHub when a
             * new message of type \c INFO_MESSAGE is sent.
             */
            virtual void handle_message(const Message &m) throw();

        private:

            /** 
             * \brief Callback for menu buttons. Assumes \c data
             * points to a nullary functor object and calls it.
             */
            static void menu_callback(Fl_Widget *w, void *data);

            /**
             * \brief Sets the system shown on the coordinate system
             * menu button.
             *
             * This method is connected to the system signal of the
             * \c Manipulator object.
             */
            static void set_system_item(
                Fl_Choice* c, transformation_target::CoordinateSystem system);

            /**
             * \brief Sets the pivot shown on the rotation pivot
             * menu button.
             *
             * This method is connected to the pivot signal of the
             * \c Manipulator object.
             */
            static void set_pivot_item(
                Fl_Choice *c,
                transformation_target::RotationCenterpoint pivot);

            /**
             * \brief Draws a thin, square down box.
             *
             * This is used as the boxtype for the text outputs on the
             * status bar, since the plastic ones look real ugly, and
             * the ordinary ones aren't accessible when the scheme
             * has been set.
             */
            static void frame_draw(int x, int y, int w, int h, Fl_Color c);

            /**
             * \brief Updates the selected objects count on the status
             * bar.
             *
             * This method is connected to the on_select and on_deselect
             * signals of the \c UserInterfaceUtilities object.
             */
            void update_count(int change, ObjectID id);

            /**
             * \brief Initializes the default group (i.e. the one shown
             * when in navigation mode).
             *
             * This method creates the two text outputs shown when in
             * default mode.
             */
            Fl_Group *init_default_group();

            /**
             * \brief Initializes the move group (i.e. the one shown
             * when the move tool is active).
             *
             * This method creates the two text outputs shown when in
             * move mode, as well as the coordinate system menu button
             * and the axis selection buttons.
             */
            Fl_Group *init_move_group();

            /**
             * \brief Initializes the rotate group (i.e. the one shown
             * when the rotate tool is active).
             *
             * This method creates the two text outputs shown when in
             * rotate mode, as well as the coordinate system and pivot
             * menu buttons and the axis selection buttons.
             */
            Fl_Group *init_rotate_group();

            Fl_Group *m_default_group;
            Fl_Output *m_default_output;
            Fl_Output *m_default_select;

            Fl_Group *m_move_group;
            Fl_Output *m_move_output;
            Fl_Output *m_move_select;

            Fl_Group *m_rotate_group;
            Fl_Output *m_rotate_output;
            Fl_Output *m_rotate_select;

            typedef std::set<Fl_Widget*> AxisSubscribers;
            AxisSubscribers m_axis_subscribers;
            typedef std::set<Fl_Widget*> SystemSubscribers;
            SystemSubscribers m_system_subscribers;

            int m_select_count;
        };        
    }
}

#endif//__PEEKABOT_GUI_STATUS_BAR_HH 
