/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_TOOL_BUTTON_HH
#define __PEEKABOT_TOOL_BUTTON_HH

#include <FL/Fl_Button.H>
#include <FL/Fl_Image.H>

#include <boost/function.hpp>

namespace peekabot
{

namespace gui
{


/** \brief Base class for Toolbar buttons.
 *
 * Implements a handle() method for taking care of mouse over-type events
 * and overrides the resize() method to make sure the button can't be resized
 * 
 */
class ToolButton : public Fl_Button
{
    static void do_nothing() {}

public:
    typedef boost::function<void (void)> Callback;

    /** \brief Default FLTK widget constructor.
     */
    ToolButton(int x, int y, int w, int h, const char* title = 0,
               const Callback &_callback = &ToolButton::do_nothing)
        : Fl_Button(x, y, w, h, title),
          m_callback(_callback)
    {
        callback(&fltk_callback);
        clear_visible_focus();
    }

    /** \brief FLTK event handling method. Acts on \c Fl::FL_ENTER and
     * \c Fl::FL_LEAVE events and changes the icon if the mouse is over the
     * button area.
     *
     *
     * \pre \c m_icon has been initialized, and either \c m_mouse_over has been
     * initialized, or \c m_mouse_over_enabled is false.
     *
     * \post
     * - If \c msg is \c Fl::FL_LEAVE or \c m_mouse_over_enabled is false, the
     * image of the base \c Fl_Button will be set to \c m_icon.
     * - If msg is \c Fl::FL_ENTER and \c m_mouse_over_enabled is true, the
     * image of the base \c Fl_Button will be set to \c m_mouse_over.
     *
     * \param msg The ID of the message sent by FLTK.
     *
     * \return If \c msg is either \c Fl::FL_ENTER or \c Fl::FL_LEAVE, the
     * return value will be 1, otherwise it will be 0.
     *
     * \todo At the moment, this just calls the Fl_Button::handle() method.
    */
    virtual int handle(int msg) { return Fl_Button::handle(msg);}

    void set_if_equal(int a, int b)
    {
        if(a == b)
            value(1);
        else
            value(0);
    }
    
    void set_if_logical_and(int a, int b)
    {
        if(a & b)
            value(1);
        else
            value(0);
    }

    void show_if_logical_and(int a, int b)
    {
        if(a & b)
            show();
        else
            hide();
    }
    
    void activate_if_logical_and(int a, int b)
    {
        if(a & b)
            activate();
        else
            deactivate();
    }

protected:
    Callback m_callback;

    static void fltk_callback(Fl_Widget *me, void *d)
    {
        static_cast<ToolButton*>(me)->m_callback();
    }

    /** \brief True if the button is highlighted on FL_MOVE events.     */
    bool m_mouse_over_enabled;
    /** \brief The icon displayed when the mouse is outside the button area.*/
    Fl_Image* m_icon;
    /** \brief The icon displayed when the mouse is inside the button area.*/
    Fl_Image* m_mouse_over;

    /** \brief Reimplement the resize method change only positiong, since we don't want these buttons
     * to change size.
     */
    /*virtual void resize(int x, int y, int w, int h)
    {
        position(x, y);
        }*/
};

}
}
#endif //__PEEKABOT_TOOL_BUTTON_HH
