/*
 * Copyright Anders Boberg 2006-2007.
 * Copyright Staffan Gimåker 2006-2007, 2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <set>

#include "TransformationInput.hh"
#include "MoveRotateGroup.hh"
#include "../actions/MoveObjectRelative.hh"
#include "../actions/RotateObject.hh"
#include "../actions/SetScale.hh"
#include "UserInterfaceUtilities.hh"
#include "UserInterface.hh"
#include "../SceneTreePtr.hh"

using namespace peekabot;
using namespace gui;
using namespace boost::posix_time;
using namespace transformation_target;

TransformationInput::TransformationInput(int x, int y, int w, int h, char* l):
    Fl_Value_Input(x, y, w, h, l),
    m_last_action_time(microsec_clock::local_time())   
{    
    // Ugly hacky conversion, but it workz
    callback((void (*)(Fl_Widget*, void*))enter_value_callback, (void*)0);
    when(FL_WHEN_ENTER_KEY);
}

/* In the handle method, we let the base Fl_Value_Input handle method process
 * events just like normal, and then perform the extra task of sending updated
 * values to the actuator. The action sending frequency is capped.
 */ 
int TransformationInput::handle(int msg)
{
    if(msg == FL_PUSH)
        m_last_value = value(); 
    int ret_val = Fl_Value_Input::handle(msg); // Process events
    real_unit new_value = value(); // The value after processing any new user input

    // Calculate how much the value changed due to the user dragging.
    real_unit delta = new_value - m_last_value;

    ptime current_time = microsec_clock::local_time();
            
    switch(msg)
    {
        case FL_DRAG:
            // Check to see that at least one action interval has passed
            // since our last visit here, else bail out.
            if(current_time < m_last_action_time +
               TheUserInterface::instance().get_action_interval())
                return 1;
            m_last_action_time = current_time;

            m_last_value = new_value;
            
            // Send our transformation to the actuator.
            on_drag(delta);
            break;
        case FL_RELEASE:
            // Always send on release, to guarantee the values shown in
            // the field are current.
            m_last_value = new_value;

            on_release(delta);            
    }

    return ret_val;
}

void TransformationInput::clear()
{
    m_last_value = 0;
    value(0);
}

void TransformationInput::enter_value_callback(TransformationInput *me, void* data)
{
    // The value entered into the field
    real_unit new_value = me->value();

    real_unit delta = new_value - me->m_last_value;

    me->m_last_value = new_value;
    
    // This counts the same as releasing the mouse.
    me->on_release(delta);
}

void TransformationInput::set_group(PropertyGroup* group)
{
    m_group = group;
}

MoveInput::MoveInput(int x, int y, int w, int h, char* l) :
    TransformationInput(x, y, w, h, l)
{
}

void MoveInput::set_coordinate(Coordinate coordinate)

{
    m_coordinate = coordinate;
}

void MoveInput::on_release(real_unit delta)
{
    // The group should always be MoveRotateGroup for this type of input.
    MoveRotateGroup* group = static_cast<MoveRotateGroup*>(m_group);

    if(group->m_local_coordinates->value() == 1)
    {
        // If we're displaying local coordinates, reset upon release.
        // Whichever coordinates we moved to will be the new origin
        // anyway.
        value(0);   
    }

    on_drag(delta);
}

void MoveInput::on_drag(real_unit delta)
{
    // The group should always be MoveRotateGroup for this type of input.
    MoveRotateGroup* group = static_cast<MoveRotateGroup*>(m_group);

    real_unit x = 0, y = 0, z = 0;
    if (m_coordinate == X_COORDINATE)
        x = delta;
    else if (m_coordinate == Y_COORDINATE)
        y = delta;
    else if(m_coordinate == Z_COORDINATE)
        z = delta;
    else
        throw;

    CoordinateSystem system;

    if(group->m_world_coordinates->value() == 1)
        system = WORLD_COORDINATES;
    else if(group->m_parent_coordinates->value() == 1)
        system = PARENT_COORDINATES;
    else
    {
        system = LOCAL_COORDINATES;

    }

    // This is where we generate actions for all objects in the selection.

    // Lock the scene tree and the selection set.
    SceneTreePtr ptr = TheSceneTree::acquire();
    Lock selection_lock(TheUIUtilities::instance().get_selection_mutex());

    std::set<ObjectID> objects = TheUIUtilities::instance().get_selection();
    
    for(std::set<ObjectID>::iterator it = objects.begin();
        it != objects.end(); ++it)
    {
        if(!ptr->get_object(*it)) // If the object's invalid, deselect it
            TheUIUtilities::instance().deselect_object(*it);
        else // Else send out action...
            TheUIUtilities::instance().post_action(
                new MoveObjectRelative(*it, x, y, z, system));
    }
}   

RotateInput::RotateInput(int x, int y, int w, int h, char* l) :
    TransformationInput(x, y, w, h, l)
{

}

void RotateInput::set_axis(Axis axis)
{
    m_axis = axis;
}

void RotateInput::on_drag(float delta)
{
    // The group should always be MoveRotateGroup for this type of input.
    MoveRotateGroup* group = dynamic_cast<MoveRotateGroup*>(m_group);

    if(group == 0)
        throw;

    real_unit x = 0, y = 0, z = 0;
    if (m_axis == X_AXIS)
        x = delta;
    else if (m_axis == Y_AXIS)
        y = delta;
    else if(m_axis == Z_AXIS)
        z = delta;
    else
        throw;

    CoordinateSystem axis_system, pivot_system;

    if(group->m_world_axis->value() == 1)
        axis_system = WORLD_COORDINATES;
    else if(group->m_parent_axis->value() == 1)
        axis_system = PARENT_COORDINATES;
    else
        axis_system = LOCAL_COORDINATES;

    // This is where we generate actions for all objects in the selection.

    // Lock the scene tree and the selection set.
    SceneTreePtr ptr = TheSceneTree::acquire();
    Lock selection_lock(TheUIUtilities::instance().get_selection_mutex());

    // Get the center of rotation...
    Vector4ru center_point;
    if(group->m_parent_center->value() == 1)
    {
        center_point = Vector4ru(0, 0, 0, 1);
        pivot_system = PARENT_COORDINATES;
    }
    else if(group->m_world_center->value() == 1)
    {
        center_point = Vector4ru(0, 0, 0, 1);
        pivot_system = WORLD_COORDINATES;
    }
    else
    {
        // Check which coordinate system the pivot is in...
        if(group->m_world_coordinates->value() == 1)
            pivot_system = WORLD_COORDINATES;
        else if(group->m_parent_coordinates->value() == 1)
            pivot_system = PARENT_COORDINATES;
        else
            pivot_system = LOCAL_COORDINATES;
        center_point(0) = group->m_x_coordinate->value();
        center_point(1) = group->m_y_coordinate->value();
        center_point(2) = group->m_z_coordinate->value();
        center_point(3) = 1;
    }

    // Iterate through the selection set and send actions for all objects
    std::set<ObjectID> objects =
    TheUIUtilities::instance().get_selection();
    for(std::set<ObjectID>::iterator it = objects.begin();
        it != objects.end(); ++it)
    {
        // If the object's invalid, deselect it
        if(!ptr->get_object(*it)) 
            TheUIUtilities::instance().deselect_object(*it);
        else // Else send out action...
            TheUIUtilities::instance().post_action(
                new RotateObject(*it, x*M_PI/180, y*M_PI/180, z*M_PI/180,
                                 axis_system, center_point, pivot_system));
    }
    /*    TheUIUtilities::instance().post_action(
          new PanCamera(m_camera, delta_x, delta_y)); */
}   

void RotateInput::on_release(float delta)
{
    on_drag(delta);
}




ScaleInput::ScaleInput(int x, int y, int w, int h, char* l) :
    TransformationInput(x, y, w, h, l)
{
}

void ScaleInput::set_xyz_links(ScaleInput *x, ScaleInput *y, ScaleInput *z)
{
    m_links[0] = x;
    m_links[1] = y;
    m_links[2] = z;
}

void ScaleInput::set_uniform(bool uniform)
{
    m_uniform = uniform;
}


void ScaleInput::on_drag(real_unit delta)
{
    double old = value() - delta;

    double factor, x_factor, y_factor, z_factor;

    if(old != 0)
        factor = value() / old;
    else
        factor = 0;

    // If the uniform-box is checked, we update all value fields.
    if(m_uniform)
    {
        x_factor = y_factor = z_factor = factor;
    }
    else
    {
        x_factor = y_factor = z_factor = 1;
        if(m_links[0] == this)
            x_factor = factor;
        else if(m_links[1] == this)
            y_factor = factor;
        else
            z_factor = factor;
    }

    
    // Iterate through the selection set and send actions for all objects
    std::set<ObjectID> objects =
    TheUIUtilities::instance().get_selection();
    for(std::set<ObjectID>::iterator it = objects.begin();
        it != objects.end(); ++it)
    {   
        if(factor > 0)
            TheUIUtilities::instance().post_action(
                new SetScale(*it, x_factor, y_factor, z_factor, true));
        else
            TheUIUtilities::instance().post_action(
                new SetScale(*it, m_links[0]->value(), m_links[1]->value(),
                             m_links[2]->value())); 
    }
}

void ScaleInput::on_release(real_unit delta)
{
    on_drag(delta);
}
