/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//
// TreeBrowser.cc
//  
// Made by Anders Boberg
// Login   <anders@robot-lab2>
// 
// Started on  Mon Jul 23 15:52:09 2007 Anders Boberg
// Started on  Mon Jul 23 15:52:09 2007 Anders Boberg
//

#include <cassert>
#include <boost/bind.hpp>
#include <limits>

#include <FL/fl_draw.H>
#include <FL/Fl.H>

#include "../actions/RearrangeObject.hh"
#include "TreeBrowser.hh"
#include "UserInterfaceUtilities.hh"
#include "UserInterface.hh"

static const int MAX_SHORT = std::numeric_limits<short>::max();
static const int MIN_SHORT = std::numeric_limits<short>::min();


using namespace peekabot;
using namespace gui;

TreeBrowser::TreeBrowser(int x, int y, int w, int h, const char *l) throw()
    : Fl_Group(x, y, w, h, l),
      m_root_node(0),
      m_indent(10),
      m_dnd(false),
      m_highlighted_widget(0)
{
    end();


    m_vertical_bar = new Fl_Scrollbar(x + w - 16, y, 16, h - 16);
    m_vertical_bar->linesize(ROW_HEIGHT);
    m_vertical_bar->value(0, 10, 0, h);
    m_vertical_bar->callback(scrollbar_callback);
    m_horizontal_bar = new Fl_Scrollbar(x, y + h - 16, w - 16, 16); 
    m_horizontal_bar->linesize(ROW_HEIGHT);
    m_horizontal_bar->type(FL_HORIZONTAL);
    m_horizontal_bar->value(0, 10, 0, w);
    m_horizontal_bar->callback(scrollbar_callback);
    add(m_vertical_bar);
    add(m_horizontal_bar);
    m_vertical_bar->show();
    m_horizontal_bar->show();

    box(FL_PLASTIC_DOWN_BOX);
    resizable(0);
    color(FL_WHITE);
}

TreeBrowser::~TreeBrowser() throw()
{
    if(m_root_node != 0)
        delete m_root_node;
}

void TreeBrowser::add_root(IDType id, const char *l, Fl_Image *img)
    throw(std::runtime_error)
{
    if(m_root_node != 0)
        throw(std::runtime_error("TreeBrowser::set_root(): "
                                 "root must be set only once!"));
          
    m_root_node = new TreeNode(this, id, l, img, false);
    m_root_node->set_expanded(true);
    m_node_map[id] = m_root_node;
    redraw();
}

bool TreeBrowser::insert_node(IDType id, IDType parent,
                              const char *l, Fl_Image *img)
{
    assert(m_root_node != 0);

    if(m_node_map.find(id) != m_node_map.end() ||
       m_node_map.find(parent) == m_node_map.end())
        return false;
  
    TreeNode *node = new TreeNode(this, id, l, img);
    TreeNode *p_node = m_node_map[parent];
    m_node_map[id] = node;
    p_node->add_child(node);
    node->set_parent(p_node);

    if(TheUIUtilities::instance().is_selected(id))
        node->set_selected(true);
    redraw();
    return true;
}

bool TreeBrowser::remove_node(IDType id)
{
    NodeMap::iterator it = m_node_map.find(id);
    if(it == m_node_map.end())
        return false;

    // Release memory
    delete it->second;

    redraw();
    return true;
}

void TreeBrowser::resize(int x, int y, int w, int h)
{
    m_vertical_bar->resize(x + w - 16, y, 16, h - 16);
    m_horizontal_bar->resize(x, y + h - 16, w - 16, 16);
    Fl_Group::resize(x, y, w, h);
}

void TreeBrowser::draw() throw()
{
    fl_draw_box(box(), x(), y(), w(), h(), color());
    int height = h();
    if(m_root_node == 0)
        return;
    else
    {
        fl_push_clip(x()+2, y()+2, w()-22, h()-22);
        height = m_root_node->draw_recursive(x() + 12 - m_horizontal_bar->value(),
                                             y() + 5 - m_vertical_bar->value());
        
        m_root_node->draw_buttons(x() + 10, y()+5);
        fl_pop_clip();
    }
    int bar_height = (height - h() + 35 < 0 ? 0 :
                      height - h() + 35);
    m_vertical_bar->value(m_vertical_bar->value(), 10, 0, bar_height);
    draw_child(*m_vertical_bar);
    draw_child(*m_horizontal_bar);
    fl_draw_box(FL_FLAT_BOX, x()+w()-16, y()+h()-16, 16, 16, FL_BACKGROUND_COLOR);
}

void TreeBrowser::set_selected(IDType id, bool select)
{
    assert(m_root_node != 0);
    NodeMap::iterator it = m_node_map.find(id);
    if(it == m_node_map.end())
        return;

    it->second->set_selected(select);
    redraw();    
}

void TreeBrowser::set_label(IDType id, const char *label)
{
    assert(m_root_node != 0);
    NodeMap::iterator it = m_node_map.find(id);
    if(it == m_node_map.end())
        return;

    it->second->get_widget()->copy_label(label);
    redraw();
}

void TreeBrowser::set_dnd(NodeWidget *nw)
{
    Fl::belowmouse(this);
    m_dnd = true;
    m_dnd_widget = nw;
}

void TreeBrowser::stop_dnd()
{
    m_dnd = false;
    m_dnd_widget = 0;
}

void TreeBrowser::select_all(bool select) throw()
{
    for(NodeMap::iterator it = m_node_map.begin();
        it != m_node_map.end(); ++it)
        it->second->set_selected(select);

    redraw();
}

int TreeBrowser::handle(int msg)
{
    switch(msg)
    {
        case FL_ENTER:
            return 1;
        case FL_RELEASE:
            if(m_dnd && Fl::event_button() == FL_LEFT_MOUSE)
            {
                m_dnd_widget->handle(msg);
                m_dnd = false;
                m_dnd_widget = 0;
                return 1;
            }
            else if(Fl::belowmouse() == m_vertical_bar)
                return m_vertical_bar->handle(FL_RELEASE);
            else if(Fl::belowmouse() == m_horizontal_bar)
                return m_horizontal_bar->handle(FL_RELEASE);

            return Fl_Group::handle(msg);
        case FL_LEAVE:
            if(Fl::belowmouse() != 0 && Fl::belowmouse() != this)
                return Fl::belowmouse()->handle(msg);
            else
                return Fl_Group::handle(msg);
        case FL_MOVE:
            /*if(Fl::event_inside(m_vertical_bar))
            {
                Fl::belowmouse(m_vertical_bar);
                return m_vertical_bar->handle(FL_ENTER);
            }            
            else if(Fl::event_inside(m_horizontal_bar))
            {
                Fl::belowmouse(m_horizontal_bar);
                    return m_horizontal_bar->handle(FL_ENTER);
                    }*/
            if(m_highlighted_widget &&
               !Fl::event_inside(m_highlighted_widget))
            {
                m_highlighted_widget->handle(FL_LEAVE);
                m_highlighted_widget = 0;
            }

            for(int i = 0; i < children(); ++i)
            {
                if(child(i)->visible() && Fl::event_inside(child(i)))
                {
                    m_highlighted_widget = child(i);
                    if(!m_dnd)
                        Fl::belowmouse(child(i));
                    return child(i)->handle(FL_ENTER);
                }
            }
            return Fl_Group::handle(msg);
        case FL_PUSH:
            if(m_dnd)
            {
                Fl::belowmouse(m_dnd_widget);
                m_dnd_widget->handle(msg);
                m_dnd = false;
                m_dnd_widget = 0;
            }
            else if(Fl::event_inside(m_vertical_bar))
            {
                Fl::belowmouse(m_vertical_bar);
                return m_vertical_bar->handle(FL_PUSH);
            }            
            else if(Fl::event_inside(m_horizontal_bar))
            {
                Fl::belowmouse(m_horizontal_bar);
                return m_horizontal_bar->handle(FL_PUSH);
            }
            return Fl_Group::handle(msg);
        case FL_DRAG:
            /*
            if(m_dnd)
            {
                for(int i = 0; i < children(); ++i)
                {
                    if(child(i)->visible() && Fl::event_inside(child(i)))
                    {
                        Fl::belowmouse(child(i));
                        child(i)->handle(FL_ENTER);
                    }
                }
                return 1;
            }
            else*/ if(Fl::belowmouse() == m_vertical_bar)
                return m_vertical_bar->handle(FL_DRAG);            
            else if(Fl::belowmouse() == m_horizontal_bar)
                return m_horizontal_bar->handle(FL_DRAG);
            /*
                if(Fl::event_inside(m_vertical_bar))
                {
                    Fl::belowmouse(m_vertical_bar);
                    return m_vertical_bar->handle(FL_DRAG);
                }            
                else if(Fl::event_inside(m_horizontal_bar))
                {
                    Fl::belowmouse(m_horizontal_bar);
                    return m_horizontal_bar->handle(FL_DRAG);
                }                
            */
        default:
            return Fl_Group::handle(msg);
    }
}

void TreeBrowser::scrollbar_callback(Fl_Widget *w, void *d)
{
    w->parent()->redraw();
}

TreeBrowser::TreeNode::~TreeNode() throw()
{
    /*if(m_selected)
      TheUIUtilities::instance().deselect_object(m_id);*/
    m_tree->remove(m_button);
    m_tree->remove(m_widget);
    m_tree->erase_node(m_id);
    Fl::delete_widget(m_button);
    Fl::delete_widget(m_widget);

    // Remember to inform the TreeBrowser that this node's widgets
    // have been erased.
    if(m_tree->get_highlighted() == m_widget || 
       m_tree->get_highlighted() == m_button)
        m_tree->set_highlighted(0);
        
    // If this is the root node, there's no parent to remove it from.
    // Instead, set the root node pointer of the tree to zero.
    if(m_parent == 0)
        m_tree->set_root(0);
    else
        m_parent->remove_child(this);

    while(!m_children.empty())
    {
        // Remember -- the child will remove itself from the parent's
        // map!!!
        delete *m_children.begin();
    }

}

int TreeBrowser::TreeNode::draw_recursive(int x, int y) throw()
{  
    int height = ROW_HEIGHT;
    int indent = m_tree->get_indentation();

    if(m_tree->x()-m_tree->w() < x && x < m_tree->x()+m_tree->w() &&
       m_tree->y()-ROW_HEIGHT < y && y < m_tree->y()+m_tree->h())
    {      
        if(m_selected)
            fl_draw_box(FL_FLAT_BOX, m_tree->x()+4, y,
                        m_tree->w()-8, height-1, fl_rgb_color(210,230,255));
        else
            fl_draw_box(FL_FLAT_BOX, m_tree->x()+4, y,
                        m_tree->w()-8, height-1, FL_WHITE);
        
        m_widget->show();
        m_widget->position(x+indent+4, y);
        if(m_widget->visible())
            m_widget->draw();
        
        fl_line_style(FL_DOT);
        fl_line(x, y + ROW_HEIGHT / 2,
                x+indent+4, y + ROW_HEIGHT / 2);  
        fl_line_style(FL_SOLID);
    }
    else
    {
        m_widget->hide();
    }

    if(!m_children.empty())
    {
        if(m_is_expanded)
        {
            int h = 0, last_h = 0, last_widget_height = 0;
            for(ChildSet::iterator it = m_children.begin();
                it != m_children.end(); ++it)
            {
                h = (*it)->draw_recursive(x + indent,
                                          y + height);
                height += h;
                last_h = h;
                last_widget_height = (*it)->m_widget->h();
            }
            fl_line_style(FL_DOT);
            int starty = y + ROW_HEIGHT/2;
            int endy   = y + height - last_h + last_widget_height/2;
            starty = (starty < MIN_SHORT ? MIN_SHORT : starty);
            starty = (starty > MAX_SHORT ? 0 : starty);
            endy = (endy < MIN_SHORT ? 0 : endy);
            endy = (endy > MAX_SHORT ? MAX_SHORT : endy);
            
            fl_line(x+indent, starty,
                    x+indent, endy);
            fl_line_style(FL_SOLID);
        }
        
        int b_x = x - m_button->w()/2;
        b_x = (b_x < m_tree->x()-m_tree->w() ? 0 : b_x);
        b_x = (b_x > m_tree->x()+m_tree->w() ? m_tree->x()+m_tree->w() : b_x);
        int b_y = (y < m_tree->y()-ROW_HEIGHT ? MIN_SHORT : y);
        b_y     = (y > m_tree->y()+m_tree->h() ? MAX_SHORT : b_y);
        m_button->position(b_x, b_y);
        //m_tree->draw_child(*m_button);
    }   
    return height;
}

int TreeBrowser::TreeNode::draw_buttons(int x, int y) const throw()
{
    int height = 5;

    height = ROW_HEIGHT;
    
    int indent = m_tree->get_indentation();
    
    if(!m_children.empty())
    {
        if(m_tree->x()-m_tree->w() < m_button->x() &&
           m_button->x() < m_tree->x()+m_tree->w() &&
           m_tree->y()-ROW_HEIGHT < m_button->y() &&
           m_button->y() < m_tree->y()+m_tree->h())
        {  
            m_button->show();
            m_button->draw();
        }
        else
            m_button->hide();

        if(m_is_expanded)
        {
            for(ChildSet::iterator it = m_children.begin();
                it != m_children.end(); ++it)
            {
                height += (*it)->draw_buttons(x + indent,
                                              y + height);
            }
        }
    }
    return height;
}

void TreeBrowser::TreeNode::hide_recursive(bool hide) throw()
{
    for(ChildSet::iterator it = m_children.begin();
        it != m_children.end(); ++it)
    {
        if(hide)
        {
            (*it)->m_widget->hide();
            (*it)->m_button->hide();
        }
        else
        {
            (*it)->m_widget->show();
            (*it)->m_button->show();
        }

        if((*it)->m_is_expanded)
            (*it)->hide_recursive(hide);
    }
}

bool TreeBrowser::TreeNode::is_descendant_of(TreeNode *node)
{
    if(m_parent == 0)
        return false;
    else if(m_parent == node || node == this)
        return true;
    else
        return m_parent->is_descendant_of(node);
}

TreeBrowser::NodeWidget::NodeWidget(const char *l, TreeNode *node,
                                    bool selectable)
    : Fl_Group(0, 0, 120, ROW_HEIGHT, l),
      m_tree(node->get_tree()),
      m_node(node),
      m_mouse_over(false),
      m_dnd(false),
      m_selectable(selectable)
{
    copy_label(l);
    align(FL_ALIGN_INSIDE|FL_ALIGN_LEFT);
    end();
}

void TreeBrowser::NodeWidget::draw()
{
    draw_me(x(), y());
}

void TreeBrowser::NodeWidget::draw_overlay()
{
    draw_me(m_dnd_x, m_dnd_y);
}

void TreeBrowser::NodeWidget::draw_me(int x, int y)
{
    int indent = x;
    if(!m_mouse_over)       
        fl_font(FL_HELVETICA, labelsize());
    else
        fl_font(FL_HELVETICA_BOLD, labelsize());
    fl_color(FL_BLACK);
    if(image() != 0)
    {
        indent += image()->w()+1;
        image()->draw(x, y);
    }
    fl_draw(label(), indent, y + h()/2 + labelsize()/2-2);
}

int TreeBrowser::NodeWidget::handle(int msg)
{
    switch(msg)
    {
        case FL_ENTER:
            m_mouse_over = true;
            redraw();
            return 1;
        case FL_LEAVE:
            m_mouse_over = false;
            redraw();
            return 1;
        case FL_PUSH:
            if(m_dnd && m_selectable)
            {
                TheUserInterface::instance().get_main_window()->deregister_overlay(this);
                TheUserInterface::instance().get_main_window()->redraw_overlay();
                m_dnd = false;                
            }
            return 1;
        case FL_DRAG:
            if(m_dnd && m_selectable)
            {                
                m_dnd_x = Fl::event_x();
                m_dnd_y = Fl::event_y();
                TheUserInterface::instance().get_main_window()->redraw_overlay();
                m_tree->handle(FL_MOVE);
                return 1;
            }
            else if(Fl::event_button() == FL_LEFT_MOUSE && m_selectable)
            {
                m_dnd = true;
                m_dnd_x = Fl::event_x();
                m_dnd_y = Fl::event_y();

                TheUserInterface::instance().get_main_window()->register_overlay(
                    this, boost::bind(&TreeBrowser::NodeWidget::draw_overlay, this));
                m_tree->set_dnd(this);
                TheUserInterface::instance().get_main_window()->redraw_overlay();
                return 1;
            }
            else return Fl_Widget::handle(msg);
        case FL_RELEASE:
            if(m_dnd && Fl::event_button() == FL_LEFT_MOUSE && m_selectable)
            {
                TheUserInterface::instance().get_main_window()->deregister_overlay(this);
                TheUserInterface::instance().get_main_window()->redraw_overlay();
                NodeWidget *highlighted = dynamic_cast<NodeWidget*>(
                    m_tree->get_highlighted());
                if(highlighted != 0 && !highlighted->m_node->is_descendant_of(m_node))
                {
                    TheUIUtilities::instance().post_action(
                        new RearrangeObject(
                            PathIdentifier(m_node->get_id()), 
                            PathIdentifier(highlighted->m_node->get_id())));
                }
                m_dnd = false;
                m_tree->stop_dnd();
            }
            else if(Fl::event_button() == FL_LEFT_MOUSE && m_selectable)
            {
                if( (Fl::event_state() & FL_CTRL) != 0)
                {
                    TheUIUtilities::instance().toggle_select(m_node->get_id());
                }
                else if( (Fl::event_state() & FL_SHIFT) != 0)
                {
                    TheUIUtilities::instance().select_object(m_node->get_id());
                }
                else
                {
                    TheUIUtilities::instance().clear_selection();
                    TheUIUtilities::instance().select_object(m_node->get_id());
                }

                parent()->redraw();
            }

            return 1;
        default:
            return Fl_Widget::handle(msg);
    }
}


void TreeBrowser::ExpandButton::draw()
{
    if(m_node->is_expanded())
    {
        if(m_mouse_over)
            fl_draw_box(FL_PLASTIC_DOWN_BOX, x(), y(),
                        w(), h(), (Fl_Color)220);
        else
            fl_draw_box(FL_PLASTIC_DOWN_BOX, x(), y(),
                        w(), h(), FL_WHITE);

        fl_font(FL_COURIER, labelsize());
        fl_color(FL_BLACK);
        fl_draw("-", x(), y(), w(), h(), FL_ALIGN_INSIDE);
        return;
    }
    else
    {
        if(m_mouse_over)
            fl_draw_box(FL_PLASTIC_UP_BOX, x(), y(),
                        w(), h(), (Fl_Color)220);
        else
            fl_draw_box(FL_PLASTIC_UP_BOX, x(), y(),
                        w(), h(), FL_WHITE);

        fl_font(FL_COURIER, labelsize());
        fl_color(FL_BLACK);
        fl_draw("+", x(), y(), w(), h(), FL_ALIGN_INSIDE);
        return;
    }
    

}

int TreeBrowser::ExpandButton::handle(int msg)
{
    switch(msg)
    {
        case FL_ENTER:
            m_mouse_over = true;
            redraw();
            return 1;
        case FL_LEAVE:
            m_mouse_over = false;
            redraw();
            return 1;
        case FL_PUSH:
            if(Fl::event_button() == 1)
            {
                if(m_node->is_expanded())
                {
                    m_node->set_expanded(false);
                    m_node->hide_recursive(true);
                }
                else
                {
                    m_node->set_expanded(true);
                    m_node->hide_recursive(false);
                }

                parent()->redraw();
            }
            return 1;
        default:
            return Fl_Widget::handle(msg);
    }
}
