/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_TREE_BROWSER_HH
#define __PEEKABOT_GUI_TREE_BROWSER_HH

#include <set>
#include <map>
#include <stdexcept>

#include <FL/Fl_Widget.H>
#include <FL/Fl_Group.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Scrollbar.H>
#include <FL/Fl_Image.H>

#include "../Types.hh"



namespace peekabot
{
    namespace gui
    {
        /** \brief Tree widget for browsing the scene tree.
         *
         * This specialized widget draws a graphical representation
         * of the scene tree. It allows the user to select nodes, and
         * to expand or collapse the children of a node.
         */
        class TreeBrowser : public Fl_Group
        {
        public:
            typedef ObjectID IDType;

            class ExpandButton;

            static const int BUTTON_SIZE = 9;
            static const int ROW_HEIGHT = 17;

            class TreeNode
            {
            public:
                typedef std::set<TreeNode*> ChildSet;
                TreeNode(TreeBrowser *tree, IDType id,
                         const char *l, Fl_Image *img,
                         bool selectable = true)
                    : m_tree(tree),
                      m_parent(0),
                      m_id(id),
                      m_button(new ExpandButton(13, this)),
                      m_widget(new NodeWidget(l, this, selectable)),
                      m_is_expanded(false),
                      m_selected(false)
                {
                    tree->add(m_button);
                    m_widget->image(img);
                    tree->add(m_widget);
                }

                ~TreeNode() throw();

                int draw_recursive(int x, int y)
                    throw();

                int draw_buttons(int x, int y)
                    const throw();

                inline void set_parent(TreeNode *p) throw()
                {
                    m_parent = p;
                }

                inline TreeNode* get_parent() throw()
                {
                    return m_parent;
                }

                inline bool add_child(TreeNode *child)
                {
                    return m_children.insert(child).second;
                }

                inline bool remove_child(TreeNode *child)
                {
                    return m_children.erase(child) > 0;
                }
                
                inline void set_widget(Fl_Widget *w) throw()
                {
                    m_widget = w;
                }                

                inline Fl_Widget *get_widget() throw()
                {
                    return m_widget;
                }

                inline const Fl_Widget *get_widget() const throw()
                {
                    return m_widget;
                }


                inline IDType get_id() const throw()
                {
                    return m_id;
                }

                inline bool is_expanded() const throw()
                {
                    return m_is_expanded;
                }

                inline void set_expanded(bool is_expanded) throw()
                {
                    m_is_expanded = is_expanded;
                }
                
                inline void set_selected(bool selected) throw()
                {
                    m_selected = selected;
                }

                inline bool is_selected() throw()
                {
                    return m_selected;
                }

                inline TreeBrowser* get_tree()
                {
                    return m_tree;
                }
                
                bool is_descendant_of(TreeNode *node);

                void hide_recursive(bool hide) throw();

            private:
                TreeBrowser *m_tree;
                TreeNode *m_parent;
                std::set<TreeNode*> m_children;
                IDType m_id;

                ExpandButton *m_button;
                Fl_Widget *m_widget;
                bool m_is_expanded;
                bool m_selected;
            };

            
            class ExpandButton : public Fl_Widget
            {
            public:
                ExpandButton(int size, TreeNode *node)
                    : Fl_Widget(0, 0, size, size),
                      m_size(size),
                      m_node(node),
                      m_mouse_over(false)
                {
                    labelsize(size-1);
                }

                virtual void draw();
                virtual int handle(int msg);
            private:
                int m_size;
                TreeNode *m_node;
                bool m_mouse_over;
            };

            class NodeWidget : public Fl_Group
            {
            public:
                NodeWidget(const char *l,
                           TreeNode *node,
                           bool selectable = true);
                virtual int handle(int msg);
            protected:
                virtual void draw();
                void draw_overlay();
                void draw_me(int x, int y);
            private:
                TreeBrowser *m_tree;
                TreeNode *m_node;
                bool m_mouse_over;
                bool m_dnd;
                bool m_selectable;
                int m_dnd_x;
                int m_dnd_y;
            };

            typedef std::map<IDType, TreeNode*> NodeMap;

            TreeBrowser(int x, int y, int w, int h, const char* l = 0) throw();         
            virtual ~TreeBrowser() throw();

            void add_root(IDType id, const char *l, Fl_Image *img)
                throw(std::runtime_error);
            
            void set_root(TreeNode *root)
            {
                m_root_node = root;
            }

            bool insert_node(IDType id, IDType parent,
                             const char *l, Fl_Image *image);
            bool remove_node(IDType id);
            
            void erase_node(IDType id)
            {
                m_node_map.erase(id);
            }
            virtual void draw() throw();

            void set_indentation(int indent) throw()
            {
                m_indent = indent;
            }

            int get_indentation() const throw()
            {
                return m_indent;
            }

            void set_highlighted(Fl_Widget *w)
            {
                m_highlighted_widget = w;
            }

            Fl_Widget *get_highlighted()
            {
                return m_highlighted_widget;
            }

            void set_selected(IDType id, bool select);
         
            void select_all(bool select) throw();

            void set_label(IDType id, const char* label);
            void set_dnd(NodeWidget *nw);
            void stop_dnd();
            
            virtual int handle(int msg);
            virtual void resize(int x, int y, int w, int h);
        private:
            static void scrollbar_callback(Fl_Widget *w, void *data);
            
            NodeMap m_node_map;
            TreeNode *m_root_node;

            Fl_Scrollbar *m_horizontal_bar;
            Fl_Scrollbar *m_vertical_bar;
            
            int m_indent;

            bool m_dnd;
            NodeWidget *m_dnd_widget;
            Fl_Widget *m_highlighted_widget;
        };
    }
}


#endif // __PEEKABOT_GUI_TREE_BROWSER_HH
