/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_UPDATE_ACTIONS_HH
#define __PEEKABOT_GUI_UPDATE_ACTIONS_HH

#include <string>

#include "../Types.hh"

class Fl_Widget;
class Fl_Input;
class Fl_Valuator;
class Fl_Button;
class Fl_Color_Chooser;


namespace peekabot
{
    namespace gui
    {
        class ViewFrame;

        /* \brief An interface for a GUI update action.
         *
         * Provides the interface for the execute() method, which will
         * be called in the main thread as the event is processed.
         *
         */
        class UpdateAction
        {
        public:
            /** \brief The method called by the main thread to update
             * the GUI.
             *
             * This method is responsible for updating one or several
             * GUI widgets, corresponding to some change in the scene
             * tree.
             */
            virtual void execute() = 0;

            /** \brief Tells the GUIUpdater if it should replace a previous,
             * still pending update with this one, or append it to a list
             * of updates to execute.
             *
             * The default implementation is to return \c true, i.e. to replace
             * any previous updates.
             */
            virtual bool replaces();

            virtual ~UpdateAction() {}
        };

        /** \brief Delete a widget, removing it from it's parent if any.
         */
        class DeleteWidget : public UpdateAction
        {
        public:
            DeleteWidget(Fl_Widget* widget)
                : m_widget(widget) {}
            
            virtual void execute();

        private:
            /** \brief A pointer to the widget to deleted.
             */
            Fl_Widget* m_widget;            
        };

        /** \brief Update action for text field widgets.
         */
        class UpdateInput : public UpdateAction
        {
        public:
            UpdateInput(Fl_Input* widget, std::string str)
                : m_widget(widget), m_string(str) {}
            
            virtual void execute();

        private:
            /** \brief A pointer to the widget to be updated.
             */
            Fl_Input* m_widget;
            
            /** \brief The new value to be inserted in the text
             * field.
             */
            std::string m_string;
        };
           
        /** \brief Update value action for valuators.
         *
         * Changes the current value of the valuator.
         */    
        class UpdateValuator : public UpdateAction
        {
        public:
            UpdateValuator(Fl_Valuator* widget, double value)
                : m_widget(widget), m_value(value) {}            
            
            virtual void execute();

        private:
            /** \brief A pointer to the widget to be updated.
             */
            Fl_Valuator* m_widget;
            
            /** \brief The value to be inserted into the value
             * field.
             */
            double m_value;
        };
        /** \brief Update domain action for valuators.
         *
         * Changes the min/max settings of the valuator.
         */    
        class UpdateValuatorDomain : public UpdateAction
        {
        public:
            UpdateValuatorDomain(Fl_Valuator* widget, double min,
                           double max)
                : m_widget(widget), m_min(min), m_max(max) {}            
            
            virtual void execute();

        private:
            /** \brief A pointer to the widget to be updated.
             */
            Fl_Valuator* m_widget;
            
            /** \brief The new minimum value of the valuator.
             */
            double m_min;

            /** \brief The new minimum value of the valuator.
             */
            double m_max;
        };

        /** \brief Update action for toggle buttons.
         */    
        class UpdateButton : public UpdateAction
        {
        public:
            UpdateButton(Fl_Button* widget, int value)
                : m_widget(widget), m_value(value) {}

            virtual void execute();

        private:
            /** \brief A pointer to the widget to be updated.
             */
            Fl_Button* m_widget;
            
            /** \brief The value to be inserted into the value
             * field.
             */
            int m_value;           
        };

        /** \brief Update action for Fl_Color_Chooser widgets.
         */    
        class UpdateColor : public UpdateAction
        {
        public:
            UpdateColor(Fl_Color_Chooser* widget, double r, double g, double b)
                : m_widget(widget), m_r(r), m_g(g), m_b(b) {}

            virtual void execute();

        private:
            /** \brief A pointer to the widget to be updated.
             */
            Fl_Color_Chooser* m_widget;
            
            /** \brief The color values to be set.
             */
            double m_r, m_g, m_b;
        };

        /** \brief Adds or updates an item in the camera menu of a ViewFrame widget.
         */    
        class UpdateCamera : public UpdateAction
        {
        public:
            UpdateCamera(ViewFrame* widget, ObjectID camera_id,
                          const std::string & camera_name)
                : m_widget(widget),
                  m_camera_id(camera_id),
                  m_camera_name(camera_name) {}
            
            /** \brief Either adds a camera to the camera selection menu, or if
             * a camera with that ID already exists, updates the name of the camera.
             */
            virtual void execute();

            /** \brief Tells the GUIUpdater that it should append UpdateCamera
             * actions to a list instead of replacing them.
             */
            virtual bool replaces();
            
        private:
            /** \brief A pointer to the widget to be updated.
             */
            ViewFrame* m_widget;
            
            /** \brief The ObjectID of the camera to be added.
             */
            ObjectID m_camera_id;
            
            /** \brief The name (name + meta_name) of the camera.
             */
            std::string m_camera_name;
        };

        /** \brief Removes an item from the camera menu of a ViewFrame widget.
         */    
        class RemoveCamera : public UpdateAction
        {
        public:
            RemoveCamera(ViewFrame* widget, ObjectID camera_id)
                : m_widget(widget),
                  m_camera_id(camera_id) {}
            
            /** \brief Removes an item from the camera selection menu.
             */
            virtual void execute();

            /** \brief Tells the GUIUpdater that it should append RemoveCamera
             * actions to a list instead of replacing them.
             */
            virtual bool replaces();

        private:
            /** \brief A pointer to the widget to be updated.
             */
            ViewFrame* m_widget;
            
            /** \brief The ObjectID of the camera to be removed.
             */
            ObjectID m_camera_id;

        };

    }
}

#endif //__PEEKABOT_GUI_UPDATE_ACTIONS_HH
