/*
 * Copyright Anders Boberg 2006-2007.
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <algorithm>
#include <stack>
#include <boost/date_time/posix_time/posix_time.hpp>

#include "CameraTrigger.hh"
#include "UserInterface.hh"
#include "UserInterfaceUtilities.hh"
#include "Workspace.hh"
#include "ViewFrame.hh"
#include "Viewport.hh"
#include "../Types.hh"
#include "MainWindow.hh"
#include "../renderer/Renderer.hh"

#include "../SceneTree.hh"
#include "../SceneTreePtr.hh"
#include "../CameraObject.hh"

using namespace peekabot;
using namespace gui;
using namespace boost::posix_time;
using renderer::TheRenderer;



UserInterface::UserInterface() :
    m_main_window(0),
    m_gl_context(0),
    m_period(milliseconds(50)),
    m_gui_period(milliseconds(33)),
    m_update_period(milliseconds(33)),
    m_action_interval(milliseconds(33)),
    m_last_render_time(microsec_clock::local_time()),
    m_last_update_time(microsec_clock::local_time()),
    m_quit_signal(false)
{
    Fl::visual(FL_DOUBLE | FL_RGB);
}

UserInterface::~UserInterface()
{
    //    delete m_main_window;
}


long UserInterface::time_until_redraw() const
{
    ptime current_time = microsec_clock::local_time();
 
    return (m_last_render_time + m_period -
            current_time).total_milliseconds();
}

long UserInterface::time_until_gui_redraw() const
{
    ptime current_time = microsec_clock::local_time();
 
    return (m_last_render_time + m_gui_period -
            current_time).total_milliseconds();
}

long UserInterface::time_until_update() const
{
    ptime current_time = microsec_clock::local_time();
    
    return (m_last_update_time + m_update_period -
            current_time).total_milliseconds();
}



void UserInterface::redraw_window()
{
    if(m_main_window)
        m_main_window->redraw();
}


void UserInterface::redraw_viewports()
{
    ptime current_time = microsec_clock::local_time();
 
    if(current_time >= m_last_render_time + m_period)
    {
        m_last_render_time = current_time;
        
        TheUIUtilities::instance().redraw_viewports();        
    }
}

void UserInterface::force_redraw()
{
    ptime current_time = microsec_clock::local_time();
    m_last_render_time = current_time;
    
    TheUIUtilities::instance().redraw_viewports();
}

bool UserInterface::set_fps(real32 fps)
{
    if(fps <= 0)
        return true;
    m_period = milliseconds((uint32_t)(1000 / fps));
    return false;
}


real32 UserInterface::get_fps() const throw()
{
    return 1/(m_period.total_milliseconds()/(real32)1000.0);
}


bool UserInterface::set_gui_fps(real32 fps)
{
    if(fps <= 0)
        return true;
    m_gui_period = milliseconds((uint32_t)(1000 / fps));
    return false;
}


real32 UserInterface::get_gui_fps() const throw()
{
    return 1/(m_gui_period.total_milliseconds()/(real32)1000.0);
}


bool UserInterface::set_action_fps(real32 fps)
{
    if(fps <= 0)
        return true;
    m_action_interval = milliseconds((uint32_t)(1000 / fps));;
    return false;
}


real32 UserInterface::get_action_fps() const throw()
{
    return 1/(m_action_interval.total_milliseconds()/(real32)1000.0);
}


real32 UserInterface::get_update_frequency() const throw()
{
    return 1/(m_update_period.total_milliseconds()/(real32)1000.0);
}


time_duration UserInterface::get_action_interval()
{
    return m_action_interval;
}


void UserInterface::update_gui() throw()
{
    m_updater.update_gui();
    
    m_last_update_time = microsec_clock::local_time();
}


void UserInterface::run()
{
    TheUIUtilities::instance().register_handler();

    m_main_window = new MainWindow();
    m_main_window->init();
    m_main_window->show();

    //
    // Update the GUI x or y times per second but at least so and so
    // many times per second.
    //
    while( Fl::check() && !m_quit_signal)
    {
        if( Fl::wait( 
               1/std::max(
                   std::max( get_update_frequency(), 
                             get_gui_fps() ), 
                   get_fps() ) ) >= 0 )
        {
            //
            // Update the GUI if there are pending updates
            //
            if( time_until_redraw() <= 0 && updates_pending() )
                update_gui();

            //
            // If any changes been made to the contents of the 3D views
            // (try to) redraw all the viewports.
            //
            if( get_gl_context() != 0 &&
                renderer::TheRenderer::instance().has_pending_updates() )
            {
                // Process any pending actions (if we don't explicitly process 
                // them they can sometimes hog *all* available memory - we don't
                // want that.
                renderer::TheRenderer::instance().process_updates();

                // Redraw all visible viewports
                TheUIUtilities::instance().redraw_viewports();
            }
        }
    }
    delete m_main_window;
    m_main_window = 0;
    //delete camera_trigger;
}

