/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_USER_INTERFACE_HH
#define __PEEKABOT_GUI_USER_INTERFACE_HH

#include <string>
#include <boost/date_time/posix_time/posix_time.hpp>

#include "../Types.hh"
#include "../Singleton.hh"
#include "GUIUpdater.hh"
#include "MainWindow.hh"

class Fl_Widget;

namespace peekabot
{
    class CameraObject;

namespace gui
{
    class Viewport;
    class UserInterface;
    typedef Singleton<UserInterface> TheUserInterface;

/** \brief Provides an interface to the GUI for outside modules.
 *
 * Whenever outside modules need to access the GUI, they go through this class.
 * Typically, \c DomainExecutionContext objects use \c send_to_console() to sending text messages to the user.
 *
 * The DomainActuator calls \c redraw_viewports() for each frame rendered, and notifies the GUI every time a new \c CameraObject is added to the scene by calling \c add_camera().
 *
 */
class UserInterface
{
public:
    UserInterface();
    ~UserInterface();
    /** \brief Immediately redraws the entire GUI window.
     */
    void redraw_window();

    /** \brief Redraws all viewports in the active workspace if a redraw interval has passed since the last redraw.
     *
     * Accesses the active workspace by way of the \c UserInterfaceUtilities object and calls \c ViewFrame::redraw_viewport() on all ViewFrames in it.
     *
     */
    void redraw_viewports();

    /** \brief Redraws all viewports in the active workspace.
     *
     * Accesses the active workspace by way of the \c UserInterfaceUtilities object and calls \c ViewFrame::redraw_viewport() on all ViewFrames in it.
     *
     */
    void force_redraw();

    /** \brief Prints text in console.
     *
     * Immediately calls \c UserInterfaceUtilities::send_to_console() to forward the text to the console.
     *
     * \param source A SourceID identifying the source of the text.
     * \param text The text to be printed in the console.
     *
     */
    void send_to_console(SourceID source, const std::string &text);

    /** \brief The frame rate at which the viewports are allowed to be redrawn
     * due to non-GUI-triggered actions.
     */
    bool set_fps(real32);

    /**
     * \brief Get the frequency at which the user interface tries to update
     * each individual viewport due to non-GUI triggered actions.
     */
    real32 get_fps() const throw();

    /** \brief The frame rate at which the viewports are allowed to be redrawn
     * due to GUI-triggered actions.
     */
    bool set_gui_fps(real32);

    /**
     * \brief Get the frequency at which the user interface tries to update
     * each individual viewport due to GUI-triggered actions.
     */
    real32 get_gui_fps() const throw();


    /** \brief The frame rate at which the GUI is allowed to send actions.
     */
    bool set_action_fps(real32);

    /**
     * \brief Get the frequency at which the user interface is allowed to
     * post actions.
     */
    real32 get_action_fps() const throw();


    /** \brief Gets the time interval between actions.
     */
    boost::posix_time::time_duration get_action_interval();

    /** \brief Gets the time interval between redraws (non-GUI-triggered
     * actions).
     */
    boost::posix_time::time_duration get_render_interval();

    /** \brief Gets the milliseconds remaining until the next redraw of 
     * the viewports due to non-GUI-triggered actions.
     */
    long time_until_redraw() const;

    /** \brief Gets the milliseconds remaining until the next redraw of
     * the viewports due to GUI-triggered actions.
     */
    long time_until_gui_redraw() const;

    /** \brief Gets the milliseconds remaining until the next time to
     * update GUI widgets.
     */
    long time_until_update() const;

    
    real32 get_update_frequency() const throw();


    /* \brief Wraps the \c GUIUpdater::post_action() method.
     */
    inline void post_action(Fl_Widget* widget, UpdateAction* event)
    {
        m_updater.post_action(widget, event);
    }

    /* \brief Wraps the \c GUIUpdater::update_gui() method.
     */
    void update_gui() throw();

    /* \brief Wraps the \c GUIUpdater::has_events() method.
     */
    inline bool updates_pending()
    {
        return m_updater.has_actions();
    }

    /* \brief Wraps the \c GUIUpdater::remove_updates() method.
     */
    inline void remove_updates(Fl_Widget* widget)
    {
        m_updater.remove_updates(widget);
    }

    /** \brief Initialize the GUI and get FLTK going.
     */
    void run();

    inline const MainWindow* get_main_window() const throw()
    {
        return m_main_window;
    }

    inline MainWindow* get_main_window() throw()
    {
        return m_main_window;
    }

    inline void set_gl_context(void *context) throw()
    {
        m_gl_context = context;
    }

    inline void *get_gl_context() const throw()
    {
        return m_gl_context;
    }


    void quit() throw()
    {
        m_quit_signal = true;
    }


private:
    GUIUpdater m_updater;

    MainWindow *m_main_window;

    
    void *m_gl_context;

    boost::posix_time::time_duration m_period;
    boost::posix_time::time_duration m_gui_period;
    boost::posix_time::time_duration m_update_period;

    boost::posix_time::time_duration m_action_interval;


    boost::posix_time::ptime m_last_render_time;
    boost::posix_time::ptime m_last_update_time;

    bool m_quit_signal;
};

}
}
#endif // __PEEKABOT_GUI_USER_INTERFACE_HH
