/*
 * Copyright Anders Boberg 2006-2007.
 * Copyright Staffan Gimåker 2006-2007, 2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "UserInterfaceUtilities.hh"
#include "UpdateActions.hh"
#include "UserInterface.hh"
#include "../renderer/Renderer.hh"
#include "PropertyFrame.hh"
#include "ViewFrame.hh"
#include "Viewport.hh"
#include "BrowserFrame.hh"
#include "ConsoleFrame.hh"
#include "Workspace.hh"
#include "../Action.hh"
#include "../atl/ActionRouter.hh"
#include "../PrintfFormatter.hh"
#include "../IDTranslationTable.hh"
#include "Manipulator.hh"
#include "CommandRepository.hh"
#include "../Configuration.hh"
#include "../actions/RemoveObject.hh"
#include "../SceneObject.hh"

#include <stack>
#include <cctype>
#include <algorithm>
#include <boost/format.hpp>
#include <boost/bind.hpp>
#include <FL/Fl.H>
#include <FL/fl_ask.H>

using namespace peekabot;
//using namespace boost;
using namespace gui;
using peekabot::renderer::TheRenderer;

UserInterfaceUtilities::UserInterfaceUtilities() :
    m_fullscreen_viewport(0),
    m_active_workspace(0),
    m_manipulator(new Manipulator()),
    m_command_repository(0),
    m_decorator(*this)
{
    TheMessageHub::instance().subscribe(CLIENT_MESSAGE, this);
    TheMessageHub::instance().subscribe(INFO_MESSAGE, this);
    TheMessageHub::instance().subscribe(WARNING_MESSAGE, this);
    TheMessageHub::instance().subscribe(ERROR_MESSAGE, this);

}

UserInterfaceUtilities::~UserInterfaceUtilities()
{
    TheMessageHub::instance().unsubscribe(CLIENT_MESSAGE, this);
    TheMessageHub::instance().unsubscribe(INFO_MESSAGE, this);
    TheMessageHub::instance().unsubscribe(WARNING_MESSAGE, this);
    TheMessageHub::instance().unsubscribe(ERROR_MESSAGE, this);

    delete m_manipulator;
    m_manipulator = 0;
    /*delete m_handler;
      m_handler = 0;*/
    if(m_command_repository != 0)
        delete m_command_repository;
}

void UserInterfaceUtilities::select_object(ObjectID object)
{
    // Second value of pair true if the object was not previously in the selection set
    if(m_selected_objects.insert(object).second)
    {
        // Notify interested parties
        m_on_object_selected(object);

        redraw_viewports();

        // Notify PropertyFrames
        if(m_active_workspace == 0) return;
        std::set<PropertyFrame*>::iterator it =  (
            m_active_workspace->get_property_frames().begin());

        for(; it != m_active_workspace->get_property_frames().end(); ++it)
            (*it)->select_objects(m_selected_objects);

        // Notify BrowserFrames
        if(m_active_workspace == 0) return;
        std::set<BrowserFrame*>::iterator bit =  (
            m_active_workspace->get_browser_frames().begin());

        for(; bit != m_active_workspace->get_browser_frames().end(); ++bit)
            (*bit)->select_object(object);

        // Register the selection trigger...
        {
            // Lock the scene tree to make sure nothing is modified while we build.
            SceneTreePtr ptr = TheSceneTree::acquire();

            SceneObject* object_ptr = ptr->get_object(object);

            if(object_ptr != 0)
            {
                object_ptr->m_on_delete.connect(
                    boost::bind(&UserInterfaceUtilities::on_delete_object, this, _1));
            }
        }
    }
}

void UserInterfaceUtilities::deselect_object(ObjectID object)
{
    ObjectSet::iterator obj_it = m_selected_objects.find(object);

    // If the object isn't in the set, just return.
    if(obj_it == m_selected_objects.end())
        return;

    // If it is, remove it.
    m_selected_objects.erase(obj_it);

    // Notify interested parties
    m_on_object_deselected(object);
    
    redraw_viewports();


    // Notify PropertyFrames
    if(m_active_workspace == 0) return;
    std::set<PropertyFrame*>::iterator it =  m_active_workspace->get_property_frames().begin();
    
    for(; it != m_active_workspace->get_property_frames().end(); ++it)
        (*it)->select_objects(m_selected_objects);    

    // Notify BrowserFrames
    if(m_active_workspace == 0) return;
    std::set<BrowserFrame*>::iterator bit =  (
        m_active_workspace->get_browser_frames().begin());

    for(; bit != m_active_workspace->get_browser_frames().end(); ++bit)
        (*bit)->deselect_object(object);

    // Deregister the selection trigger...
    {
        // Lock the scene tree to make sure nothing is modified while we build.
        SceneTreePtr ptr = TheSceneTree::acquire();
        
        SceneObject* object_ptr = ptr->get_object(object);
        if(object_ptr != 0)
        {
            object_ptr->m_on_delete.disconnect(
                boost::bind(&UserInterfaceUtilities::on_delete_object, this, _1));
        }
    }
}


void UserInterfaceUtilities::toggle_select(ObjectID object)
{
    bool selected = m_selected_objects.find(object) != m_selected_objects.end();

    // If the object is selected, deselect it, if not, select it.
    if( selected )
        deselect_object(object);
    else
        select_object(object);
}

void UserInterfaceUtilities::clear_selection()
{

    for(ObjectSet::iterator it = m_selected_objects.begin();
        it != m_selected_objects.end(); ++it)
    {
        m_on_object_deselected(*it);

        // Deregister the selection trigger...
        {
            // Lock the scene tree to make sure nothing is modified while we build.
            SceneTreePtr ptr = TheSceneTree::acquire();
            
            SceneObject* object_ptr = ptr->get_object(*it);

            if(object_ptr != 0)
            {
                object_ptr->m_on_delete.disconnect(
                    boost::bind(&UserInterfaceUtilities::on_delete_object, this, _1));
            }
        }
    }
 
    m_selected_objects.clear();
    
    redraw_viewports();

    // Notify PropertyFrames
    if(m_active_workspace == 0) return;
    std::set<PropertyFrame*>::iterator it =  m_active_workspace->get_property_frames().begin();
    
    for(; it != m_active_workspace->get_property_frames().end(); ++it)
        (*it)->select_objects(m_selected_objects);

    // Notify BrowserFrames
    if(m_active_workspace == 0) return;
    std::set<BrowserFrame*>::iterator bit =  (
        m_active_workspace->get_browser_frames().begin());
    
    for(; bit != m_active_workspace->get_browser_frames().end(); ++bit)
        (*bit)->clear_selection();    
}


const ObjectSet & UserInterfaceUtilities::get_selection() const
{
    return m_selected_objects;
}

bool UserInterfaceUtilities::is_selected(ObjectID object) const
{
    return m_selected_objects.find(object) != m_selected_objects.end();
}

void UserInterfaceUtilities::activate_workspace(Workspace *ws)
{
    m_active_workspace = ws;
}

peekabot::gui::Workspace*  UserInterfaceUtilities::active_workspace()
{
    return m_active_workspace;
}

Fl_Text_Buffer * UserInterfaceUtilities::get_text_buffer()
{
    return &m_console_buffer;
}


void UserInterfaceUtilities::post_action(Action *action, SourceID dest)
{
    TheActionRouter::instance().post_action(
        action, GUI_SOURCE_ID, dest);
}

boost::recursive_mutex & UserInterfaceUtilities::get_camera_mutex()
{
    return m_camera_mutex;
}

boost::recursive_mutex & UserInterfaceUtilities::get_selection_mutex()
{
    return m_selection_mutex;
}

boost::recursive_mutex & UserInterfaceUtilities::get_console_mutex()
{
    return m_console_mutex;
}


Lock::Lock(boost::recursive_mutex &mutex):
    m_lock(new boost::recursive_mutex::scoped_lock(mutex)) {}

Lock::~Lock()
{
    delete m_lock;
}


void UserInterfaceUtilities::handle_message(const Message &m) throw()
{
    class SendToConsole : public UpdateAction
    {
    public:
        SendToConsole(const std::string &str)
            : m_string(str)
        {
        }

        virtual void execute()
        {
            TheUIUtilities::instance().print_in_console(m_string);
        }

        virtual bool replaces()
        {
            return false;
        }

    private:
        std::string m_string;
    };


    class ErrorMessageBox : public UpdateAction
    {
    public:
        ErrorMessageBox(const std::string &str)
            : m_string(str)
        {
        }

        virtual void execute()
        {
            fl_alert("%s", m_string.c_str());
        }

        virtual bool replaces()
        {
            return false;
        }

    private:
        std::string m_string;
    };


    // Format the string.
    PrintfFormatter formatter("[%t]: %m (%y, %s)\n");
    std::string tagged_string = formatter.format(m);
    
    // Post the printing action on the update queue.
    TheUserInterface::instance().post_action(
        m_active_workspace, new SendToConsole(tagged_string));

    // If we receive a error message from the GUI,
    // display it in an message box
    if( m.get_type() == ERROR_MESSAGE &&
        m.has_source() && m.get_source() == GUI_SOURCE_ID )
    {
        TheUserInterface::instance().post_action(
            m_active_workspace, new ErrorMessageBox(m.get_message()));
    }
}

void UserInterfaceUtilities::print_in_console(const std::string &str)
{
    m_console_buffer.append(str.c_str());

    const std::size_t MAX_LEN = 64*1024;
    if( (std::size_t)m_console_buffer.length() > MAX_LEN )
    {
        m_console_buffer.remove(0, m_console_buffer.length() - MAX_LEN);
        assert( (std::size_t)m_console_buffer.length() == MAX_LEN );
    }

    // Notify ConsoleFrames
    if(m_active_workspace == 0) return;
    std::set<ConsoleFrame*>::iterator it =
    m_active_workspace->get_console_frames().begin();
    
    for(; it != m_active_workspace->get_console_frames().end(); ++it)
        (*it)->update();
}

void UserInterfaceUtilities::delete_selected() throw()
{
    if(m_selected_objects.size() == 1)
    {                
        int choice = fl_choice("Really delete the selected object?",
                               "Cancel", "Delete", 0);
        if(choice == 1)
            post_action(new RemoveObject(*m_selected_objects.begin()));
    }
    else if(m_selected_objects.size() > 1)
        fl_message("Objects may only be deleted one at a time.");

}

int UserInterfaceUtilities::handle(int msg) throw()
{
    if(msg == FL_SHORTCUT)
    {
        KeyTranslationMap & key_translator = TheUIUtilities::instance().m_key_translator;
        CommandRepository * command_repo = TheUIUtilities::instance().m_command_repository;

        KeyTranslationMap::iterator it = key_translator.find(Fl::event_key());
        if(it != key_translator.end())
        {
            command_repo->get_command(it->second)();
            return 1;
        }
        else if(Fl::event_key() == FL_Escape)
            return 1;
    }
    
    return 0;
}

int lower(int c)
{
    return tolower(c);
}

void UserInterfaceUtilities::lowercase(std::string &str)
{
    std::transform(str.begin(), str.end(), str.begin(), lower);
}


void UserInterfaceUtilities::register_handler()
{
    // Map used to map the string representations for keys used in the
    // config file to FLTK Fl::event_key() values.
    std::map<std::string, int> trans;
    trans["esc"] = FL_Escape;
    trans["insert"] = FL_Insert;
    trans["home"] = FL_Home;
    trans["page_up"] = FL_Page_Up;
    trans["delete"] = FL_Delete;
    trans["end"] = FL_End;
    trans["page_down"] = FL_Page_Down;
    trans["backspace"] = FL_BackSpace;

    for(char i = 'a'; i <= 'z'; ++i)
        trans[std::string(1, i)] = i;

    for(char i = '0'; i <= '9'; ++i) 
        trans[std::string(1, i)] = i;

    trans["kp_plus"] = '+' + FL_KP;
    trans["kp_minus"] = '-' + FL_KP;
    trans["kp_times"] = '*' + FL_KP;
    trans["kp_divide"] = '/' + FL_KP;
    
    std::string abort = TheConfiguration::instance().get_option<std::string>(
        "shortcuts.abort", "esc");
    lowercase(abort);
    m_key_translator[trans[abort]] = CMD_ABORT;

    std::string navigate = TheConfiguration::instance().get_option<std::string>(
        "shortcuts.navigate_mode", "n");
    lowercase(navigate);
    m_key_translator[trans[navigate]] = CMD_SET_NAVIGATION_MODE;

    std::string move = TheConfiguration::instance().get_option<std::string>(
        "shortcuts.move_mode", "m");
    lowercase(move);
    m_key_translator[trans[move]] = CMD_SET_MOVE_MODE;

    std::string rotate = TheConfiguration::instance().get_option<std::string>(
        "shortcuts.rotate_mode", "r");
    lowercase(rotate);
    m_key_translator[trans[rotate]] = CMD_SET_ROTATE_MODE;

    std::string camera = TheConfiguration::instance().get_option<std::string>(
        "shortcuts.camera_coordinates", "c");
    lowercase(camera);
    m_key_translator[trans[camera]] = CMD_SET_CAMERA_COORDINATES;
    std::string local = TheConfiguration::instance().get_option<std::string>(
        "shortcuts.local_coordinates", "l");
    lowercase(local);
    m_key_translator[trans[local]] = CMD_SET_LOCAL_COORDINATES;
    std::string world = TheConfiguration::instance().get_option<std::string>(
        "shortcuts.world_coordinates", "w");
    lowercase(world);    
    m_key_translator[trans[world]] = CMD_SET_WORLD_COORDINATES;
    std::string parent = TheConfiguration::instance().get_option<std::string>(
        "shortcuts.parent_coordinates", "p");
    lowercase(parent);
    m_key_translator[trans[parent]] = CMD_SET_PARENT_COORDINATES;

    std::string x = TheConfiguration::instance().get_option<std::string>(
        "shortcuts.x_axis", "x");
    lowercase(x);
    m_key_translator[trans[x]] = CMD_SET_X_AXIS;
    std::string y = TheConfiguration::instance().get_option<std::string>(
        "shortcuts.y_axis", "y");
    lowercase(y);
    m_key_translator[trans[y]] = CMD_SET_Y_AXIS;
    std::string z = TheConfiguration::instance().get_option<std::string>(
        "shortcuts.z_axis", "z");
    lowercase(z);
    m_key_translator[trans[z]] = CMD_SET_Z_AXIS;

    std::string clear = TheConfiguration::instance().get_option<std::string>(
        "shortcuts.clear_selection", "backspace");
    lowercase(clear);
    m_key_translator[trans[clear]] = CMD_CLEAR_SELECTION;

    std::string del = TheConfiguration::instance().get_option<std::string>(
        "shortcuts.delete", "delete");
    lowercase(del);
    m_key_translator[trans[del]] = CMD_DELETE_OBJECT;

    std::string fullscreen = TheConfiguration::instance().get_option<std::string>(
        "shortcuts.fullscreen", "f");
    lowercase(fullscreen);
    m_key_translator[trans[fullscreen]] = CMD_FULLSCREEN;
                           
     


    Fl::add_handler(&UserInterfaceUtilities::handle);
    m_command_repository = new CommandRepository();
}

void UserInterfaceUtilities::do_fullscreen(Viewport *vp) throw()
{
    int x, y, w, h;
    Fl::screen_xywh(x, y, w, h);

    if(m_fullscreen_viewport != 0)
    {
        Fl::delete_widget(m_fullscreen_viewport);
        m_fullscreen_viewport = 0;
    }
    else if(vp == 0)
    {
        if(m_active_workspace != 0)
        {
            ViewFrame *vf = dynamic_cast<ViewFrame*>(
                m_active_workspace->get_selected_frame());
            if(vf != 0)
            {
                m_fullscreen_viewport = new Viewport(vf->get_viewport(), x, y, w, h);
            }
        }
    }
    else
    {
        m_fullscreen_viewport = new Viewport(vp, x, y, w, h);
    }
}


void UserInterfaceUtilities::redraw_viewports()
{
    if(m_fullscreen_viewport != 0)
        m_fullscreen_viewport->redraw();
    else if(m_active_workspace != 0)
    {
        std::set<ViewFrame*> & view_frames =
        m_active_workspace->get_view_frames();
        
        std::set<ViewFrame*>::iterator it;
        for(it = view_frames.begin(); it != view_frames.end(); ++it)
            (*it)->redraw_viewport();
    }
}

void UserInterfaceUtilities::on_delete_object(
    SceneObject *object) throw()
{
    std::stack<SceneObject *> s;
    s.push(object);
    
    while( !s.empty() )
    {
        SceneObject *obj = s.top();
        s.pop();
        
        TheUserInterface::instance().post_action(
            0, new DeselectObject(obj->get_object_id()));
        
        // Push child objects
        SceneObject::ChildContainer &children = obj->get_children();
        for( SceneObject::ChildContainer::const_iterator it = 
             children.begin(); it != children.end(); ++it )
            s.push(*it);
    }
}

UserInterfaceUtilities::DeselectObject::DeselectObject(ObjectID id)
    : m_id(id)
{    
}

void UserInterfaceUtilities::DeselectObject::execute()
{
    TheUIUtilities::instance().deselect_object(m_id);
}

bool UserInterfaceUtilities::DeselectObject::replaces()
{
    return false;
}

