/*
 * Copyright Anders Boberg 2006-2007.
 * Copyright Staffan Gimåker 2006-2007, 2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PEEKABOT_GUI_USER_INTERFACE_UTILITIES_HH_INCLUDED
#define PEEKABOT_GUI_USER_INTERFACE_UTILITIES_HH_INCLUDED


#include <string>
#include <FL/Fl_Text_Buffer.H>
#include <ctime>
#include <cstdio>
#include <set>
#include <boost/thread/recursive_mutex.hpp>
#include <boost/signal.hpp>
#include <boost/utility.hpp>

#include "../SceneObject.hh"
#include "../Singleton.hh"
#include "../MessageHub.hh"
#include "../Subscriber.hh"
#include "GUITypes.hh"
#include "Decorator.hh"
#include "UpdateActions.hh"

//#include "Console.hh"

namespace peekabot
{

class Action;

namespace gui
{
    class MainWindow;
    class Workspace;

    class ViewFrame;
    class Viewport;
    
    class Manipulator;
    class CommandRepository;

    class UserInterfaceUtilities;
    typedef Singleton<UserInterfaceUtilities> TheUIUtilities;

    typedef std::set<ObjectID> ObjectSet;
    
    class Lock
    {
    public:
        Lock(boost::recursive_mutex &mutex);
        ~Lock();
    private:
        boost::recursive_mutex::scoped_lock *m_lock;
    };


/** \brief Provides some general utility functions for GUI classes.
 *
 *  This class keeps track of selected objects, providing functions to select, deselect, and retrieve a list of selected objects.
 *  
 *  The class also keeps track of the currently active workspace, and makes sure all frames in that workspace are notified any time the selection changes.
 *
 * This class also provides the \c post_action() method for any objects that need to modify the scene.
 *
 *  Finally, the class keeps track of all cameras in the scene, and provides a list of them to the \c ViewFrames in the active workspace.
 *
 */
class UserInterfaceUtilities : public Subscriber, 
                               public boost::noncopyable
{
    friend class Lock;
public:
    UserInterfaceUtilities();
    ~UserInterfaceUtilities();

    /** \brief Insert \c object into the selection set.
     *
     * This method inserts the \c ObjectID into the selection set and calls \c PropertyFrame::select_objects() on any \c PropertyFrame object in the active workspace to update it on the new selection. 
     *
     * Typically, this function is called by the \c Viewport::handle() method after performing a pick or by the \c BrowserFrame::browser_callback() function after the the user selects a node in the scene browser. 
     *
     * \param object \c The ObjectID of the SceneObject to be inserted into the list of selected objects.
     * 
     */
    void select_object(ObjectID object);

    /** \brief Removes \c object from the selection set.
     *
     * This method removes the \c ObjectID from the selection set and calls \c PropertyFrame::select_objects() on any \c PropertyFrame object in the active workspace to update it on the new selection. 
     *
     * Typically, this function is called by the \c Viewport::handle() method after performing a pick or by the \c BrowserFrame::browser_callback() function after the the user selects a node in the scene browser. 
     *
     * \param object \c The ObjectID of the SceneObject to be inserted into the list of selected objects.
     * 
     */
    void deselect_object(ObjectID object);


    /** \brief Clears the current selection and sets \c object as selected.
     *
     * This method inserts the \c ObjectID into the selection set. After that, it calls \c PropertyFrame::select_objects() on any \c PropertyFrame object in the active workspace to update it on the new selection. 
     *
     * Typically, this function is called by the \c Viewport::handle() method after performing a pick or by the \c BrowserFrame::browser_callback() function after the the user selects a node in the scene browser. 
     *
     * \param object \c The ObjectID of the SceneObject to be inserted into the list of selected objects.
     * 
     */
    void select(ObjectID object);


    /** \brief Toggles an object between selected and deselected.
     *
     * This function does one of the following
     *
     * - If \c object is not already in the list of selected objects, it adds it.
     *
     * - If \c object is in the list of selected objects, it removes it.
     *
     * After modifying the list of selected objects, it will notify all \c PropertyFrames in the active workspace by calling \c PropertyFrame::select_objects() on on them.
     *
     * Typically, this function is called by the \c Viewport::handle() method after performing a \e pick or by the \c BrowserFrame::node_callback() function after the the user selects a node in the Scene browser. 
     *
     * \param object \c A pointer to the SceneObject to be added or removed from the list of selected objects.
     * 
     */
    void toggle_select(ObjectID object);


    /** \brief Clears the list of selected objects.
     *
     * This function calls clear() on the list of selected objects, and then notifies all \c PropertyFrames in the active workspace by calling \c PropertyFrame::select_objects() on them.
     *
     * Typically, this function is called by the \c handle() method of \c Viewport after performing a pick.
     *
     */
    void clear_selection();


    /** \brief Returns a list of the currently selected objects.
     *
     * This function is typically called by instances of the \c ViewFrame and \c PropertyFrame classes as they are created or go from hidden to shown.
     *
     * \return A constant reference to the list of currently selected objects.
     * 
     */
    const ObjectSet & get_selection() const;

    /**
     * \brief Returns true if \c object is currently selected.
     */ 
    bool is_selected(ObjectID object) const;

    /** \brief Returns a pointer to the active workspace.
     *
     *
     * This function is typically called by objects needing access to particular frames in the currently active workspace.
     *
     * \return A pointer to the currently active workspace.
     *
     */
    Workspace* active_workspace();

    /** \brief Register a new workspace as the currently active one.
     *
     * This function is typically called by the \c Workspace::show() method as the user clicks on a new workspace tab.
     *
     * \param ws A pointer to the workspace to be registered as active.
     *
     *
     */
    void activate_workspace(Workspace* ws);

    /** \brief Prints text in the console.
     *
     * First, the string will be appended with a timestamp (in 24h format) and a source tag (a textual version of the source type).
     *
     * For example, if the current system time is <tt>14:12:03</tt>, \c text is <tt>"Endcountered solid obstacle!"</tt> and the \c source is <tt>{134, CLIENT}</tt>, the string <tt>"[14:12:03] client132: Encountered solid obstacle!"</tt> will be appended to the buffer.
     *
     * Then calls \c Fl_Text_Buffer::append() to put the text in the console and notifies all \c ConsoleFrame objects in the active workspace by calling ConsoleFrame::update().
     *
     * \param source A SourceID identifying the source of the text.
     * \param text The text to be printed in the console.
     *
     */
    void send_to_console(SourceID source, const std::string &text);


    /** \brief Posts an action in the inbound queue.
     *
     * Acquires a write mutex on the inbound queue and posts \c action. 
     *
     *
     * \param action A pointer to the \c Action to be posted.
     */
    void post_action(
        Action* action, 
        SourceID dest = SERVER_SOURCE_ID);

    /** \brief Get the text buffer for console output.
     *
     */
    Fl_Text_Buffer* get_text_buffer();

    /** \brief Get the camera set mutex.
     */
    boost::recursive_mutex &get_camera_mutex();

    /** \brief Get the selection set mutex.
     */
    boost::recursive_mutex &get_selection_mutex();

    /** \brief Handles messages from the message hub by printing
     * them on the console.
     */
    virtual void handle_message(const Message &m) throw();

    /** \brief Get the console set mutex.
     */
    boost::recursive_mutex &get_console_mutex();
    
    /** \brief Non-thread-safe method for printing thins in the
     * console. Only to be used in update actions.
     */
    void print_in_console(const std::string &str);


    Manipulator * get_manipulator()
    {
        return m_manipulator;
    }

    CommandRepository * get_command_repository() throw()
    {
        return m_command_repository;
    }

    static int handle(int msg) throw();

    void register_handler();
    void delete_selected() throw();

    void do_fullscreen(Viewport *vp) throw();
    void redraw_viewports();

    inline boost::signal<void (ObjectID)> &on_object_selected() throw()
    {
        return m_on_object_selected;
    }

    inline boost::signal<void (ObjectID)> &on_object_deselected() throw()
    {
        return m_on_object_deselected;
    }

private:
    /**
     * \brief Takes care of deselecting objects that are deleted.
     */
    void on_delete_object(SceneObject *object) throw();

private:
   // Signals
    boost::signal<void (ObjectID)> m_on_object_selected;

    boost::signal<void (ObjectID)> m_on_object_deselected;



    typedef std::map<int, Command> KeyTranslationMap;

    Viewport *m_fullscreen_viewport;

    /** \brief Recursive mutex for the camera set.
     */
    boost::recursive_mutex m_camera_mutex;
    
    /** \brief Recursive mutex for the selection set.
     */
    boost::recursive_mutex m_selection_mutex;

    /** \brief Recursive mutex for the console.
     */
    boost::recursive_mutex m_console_mutex;

    /** \brief The text buffer shared between all \c ConsoleFrame objects.
     */
    Fl_Text_Buffer m_console_buffer;


    /** \brief A list of pointers to the currently selected objects.
     */
    ObjectSet m_selected_objects;

    /** \brief A pointer to the active workspace.
     */
    Workspace* m_active_workspace;

    Manipulator *m_manipulator;
    KeyTranslationMap m_key_translator;
    
    CommandRepository *m_command_repository;

    void lowercase(std::string &str);

    Decorator m_decorator;

    /**
     * \brief GUI update action for deselecting deleted objects.
     */
    class DeselectObject : public UpdateAction
    {
    public:
        DeselectObject(ObjectID id);
        virtual void execute();
        virtual bool replaces();
    private:
        ObjectID m_id;
    };
};

}
}

#endif // PEEKABOT_GUI_USER_INTERFACE_UTILITIES_HH_INCLUDED
