/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_VIEWFRAME_HH
#define __PEEKABOT_GUI_VIEWFRAME_HH

#include <list>
#include <vector>
#include <string>
#include <FL/Fl_Choice.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Menu_Item.H>
#include <FL/Fl_Group.H>
#include "Frame.hh"
#include "../Types.hh"
#include "../SceneTreePtr.hh"

namespace peekabot
{
    class CameraObject;

namespace gui
{

    class CameraPropertyTrigger;
    class CameraTrigger;
    class Viewport;

    /** \brief Represents view frames.
     *
     * The ViewFrame contains a \c Viewport object for displaying OpenGL graphics, and a simple menu for choosing camera, visible layers, rendering mode and view selection (i.e. orthogonal or perspective). The menu also includes buttons for selecting the mode of operation, which includes moving and rotating objects and cameras.
     *
     * The actual communication with the renderer and viewport mouse control functionality is located in the \c Viewport class.
     *
     * The ViewFrame also inherits the basic frame content selection menu and border drawing functionality from the \c Frame base class.
     *
     */
class ViewFrame : public Frame
{
public:
    /** \brief Constructs a new ViewFrame with the specified size.
     *
     * Will call the \c Frame base class constructor to initialize the frame and draw the frame content selection menu.
     *
     * \param x The x-coordinate of the top left corner of this frame.
     * \param y The y-coordinate of the top left corner of this frame.
     * \param w The width of this frame.
     * \param h The height of this frame.
     * \param l The widget label (for FLTK compatibility)
     */
    ViewFrame(int x, int y, int w, int h, char* l = 0);


    /** \brief Creates a new view frame.
     *
     * The new frame will be created to fit the space occupied by \c frame. The constructor will also take care of registering the frame with the active workspace, and of removing the old frame.
     *
     * \param frame The \c Frame to replace.
     */
    ViewFrame(Frame* frame);


    /** \brief Deletes the \c CameraPropertyTrigger created for this frame.
     */
    virtual ~ViewFrame();

    /** \brief Redraw the viewport in this frame.
     *
     * This method calls the \c redraw() on the viewport in this frame, telling FLTK that it needs its Viewport::draw() method called.
     *
     * This method is typically called by UserInterface::redraw_viewports(), which in turn is called by the DomainActuator when each time a new frame should be drawn.
     *
     */
    void redraw_viewport();
  

    /** \brief Adds a new camera to the camera selection menu.
     *
     * This method adds a new camera item by calling \c Fl_Menu_::add() on the camera selection menu.
     *
     * This method is typically called by \c UserInterfaceUtilities::add_camera().
     *
     * \param camera The ObjectID of the \c CameraObject to be added to the menu.
     *
     */  
    void add_camera(ObjectID camera, const std::string & camera_name);



    /** \brief Removes camera from the camera selection menu.
     *
     * This method removes a item from the menu by calling \c Fl_Menu_::remove() on the camera selection menu. If the currently selected camera is removed from the list, the first camera on the new list will be selected instead.
     *
     * This method is typically called by \c UserInterfaceUtilities::remove_camera().
     *
     * \param camera The ObjectID of the \c CameraObject to be removed from the menu.
     *
     */  
    void remove_camera(ObjectID camera);



    /** \brief Callback function for camera selection.
     *
     * This callback fetches the rendering mode and view type parameters from the new camera, and alter the corresponding button states on the menu to match it.
     *
     * It also calls the \c Viewport::set_camera() method to notify the \c Viewport of this \c ViewFrame of the change.
     *
     * This callback is registered with the camera selection choice button during the construction of the view frame.
     *
     * \param w A pointer to the \c Fl_Choice widget (camera selection button).
     * \param data A pointer to void, as required by FLTK (not used).
     *
     */
    static void camera_selection_callback(Fl_Widget* w, void* data);

    /** \brief Callback for mode of operation buttons.
     *
     * This callback will set the current mode of operation of the \c Viewport in this frame by calling the \c Viewport::set_mode() method.
     *
     * This callback is registered with the buttons during construction of the \c ViewFrame, and is called whenever the user clicks one of the mode of operation buttons.
     *
     * \param w The mode of operation button triggering the callback.
     * \param data A pointer to this ViewFrame.
     */    
    static void mode_of_operation_callback(Fl_Widget* w, void* data);


    /** \brief Callback for visibility buttons. Sets the visibility of a
     * layer.
     *
     * This callback will set a layer as visible or hidden in the viewport depending whether the button is depressed or not. The layer affected by the callback depends on the \c user_data() of the visibility button.
     *
     * This callback is registered with the buttons during the construction of the parent \c ViewFrame, and is called whenever the user clicks one of the visibility buttons.
     *
     * \param w The visibility button triggering the callback.
     * \param data A pointer to the associated view frame.
     */    
    static void layer_button_callback(Fl_Widget* w, void* data);

    static void snapshot_callback(Fl_Widget* w, void* data);

    static void cont_snapshot_callback(Fl_Widget* w, void* data);


    /** \brief Deregisters the \c Frame with its \c Workspace.
     */
    virtual void deregister();

    /** \brief Overrides the \c Fl_Scroll::resize() method to resize the
     * \c Viewport.
     */
    virtual void resize(int xx, int yy, int ww, int hh);

    /** \brief Returns true if the frame can be resized to width \c width.
     *
     * Currently, the minimum width is 50 for view frames.
     */
    virtual bool resizable_width(int width);
    /** \brief Returns true if the frame can be resized to height \c height.
     * 
     * Currently, the minimum height is 50 for view frames.
     */
    virtual bool resizable_height(int height);

      
    Viewport * get_viewport() { return m_viewport;}
protected: 

    void register_trigger(ObjectID, SceneTreePtr&);
    void deregister_trigger(ObjectID, SceneTreePtr&);

    /** \brief Initializes a new view frame to the specified size.
     *
     * It creates the camera selection menu, the layer visibility selection group, the rendering mode, view selection and mode of operation buttons. The \c camera_selection_callback() function will be registered with the camera selection menu.
     *
     * It will also create a \c Viewport showing the default camera view and register its \c callback functions with the visibility and mode of operation buttons.
     *
     * \param x The x-coordinate of the top left corner of this frame.
     * \param y The y-coordinate of the top left corner of this frame.
     * \param w The width of this frame.
     * \param h The height of this frame.
     */
    void init(int x, int y, int w, int h);

    /** \brief Get the type of the frame.
     *
     * \return A \c FrameType of value VIEW_FRAME
     */
    virtual FrameType frame_type() const;

    /** \brief Creates a clone of the frame with the specified screen
     * dimensions.
     */
    virtual Frame* clone(int x, int y, int w, int h);

private:
    /** \brief The callback for the set top-down view button.
     *
     * Sends an action to the server to move the current camera
     * so that it is positioned 10 units above the xy-plane,
     * directed downward.
     *
     */
    static void top_down_callback(Fl_Widget* w, void* data);
    
    static void view_mode_callback(Fl_Widget* w, void* data);

    /** \brief A pointer to the \c Viewport displayed in this \c ViewFrame.
     */
    Viewport* m_viewport;

    /** \brief A pointer to the camera selection widget.
     */
    Fl_Choice* m_camera_selection_menu;

    /** \brief A pointer to the Select mode button.
     */
    Fl_Button* m_select_mode;

    /** \brief A pointer to the Free camera mode button.
     */
    Fl_Button* m_free_camera_mode;

    Fl_Group* m_mode_group;

    Fl_Button *m_layer_buttons[16];

    Fl_Button *m_view_mode_button;
    Fl_Button *m_top_down_button;
    Fl_Button *m_snapshot_button;
    Fl_Button *m_cont_snapshot_button;
    
    CameraPropertyTrigger *m_trigger;

    CameraTrigger* m_camera_trigger;
    
    std::vector<ObjectID*> m_cameras;
};





}
}
#endif //__PEEKABOT_GUI_VIEWFRAME_HH
