/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_VIEWPORT_HH
#define __PEEKABOT_GUI_VIEWPORT_HH

#include <list>
#include <string>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <FL/Fl_Gl_Window.H>

#include "../Types.hh"
#include "../Matrix.hh"

namespace peekabot
{
    
    class CameraObject;
    class SceneTreePtr;
    class FrameCollector;

    

namespace gui
{
    class ViewFrame;
    

    /** \brief A graphical OpenGL area.
     *
     * This class represents a screen area in which 3D graphics will be drawn. This class collaborates with its parent \c ViewFrame, a \c CameraObject associated with the view, and the \c Renderer class, which is responsible for the actual drawing of the graphics.
     * The viewport is responsive to certain mouse events, such as clicking and dragging, the effects of which include changing the selection list, moving and rotating objects and cameras.
     *
     * Other objects, such as the \c PropertyGroups or client actoins, indirectly affect what is drawn in the Viewport, but these classes do not directly interact with the Viewport class.
     *
     */
class Viewport : public Fl_Gl_Window
{
    friend class ViewFrame;
public:
    
    /** \brief Constructs a new \c Viewport with the specified location and size.
     *
     * \param x The x-coordinate of the top left corner of this frame.
     * \param y The y-coordinate of the top left corner of this frame.
     * \param w The width of this frame.
     * \param h The height of this frame.
     * \param l The widget label (for FLTK compatibility)
     */
    Viewport(int x, int y, int w, int h, char* l = 0);

    Viewport(Viewport *vp, int x, int y, int w, int h);

    /** \brief Destructor for Viewport class. Deletes m_camera.
     */
    virtual ~Viewport();

    
    /** \brief Handles mouse clicks within the \c Viewport area.
     *
     * This method will act on \c FL_PUSH, \c FL_RELEASE and \c FL_DRAG events.
     *
     * \b FL_PUSH
     * In the case of an \c FL_PUSH event, it will store mouse coordinates, then check the mode of operation (\c m_operation_mode).
     * - If it was \c VP_ROTATE_CAMERA, \c VP_PAN_CAMERA or \c VP_ZOOM_CAMERA, we don't need to taka any particular measures.
     * - If it was VP_ROTATE_OBJECT, VP_MOVE_OBJECT or VP_SCALE_OBJECT, on the other hand, we will set \c m_selection_pass to \c true,  and call \c redraw() to perform a selection pass (\c redraw() will call \c Renderer::pick() if \c m_selection_pass is \c true). 
     *
     * \b FL_DRAG
     * If an \c FL_DRAG message is received, we set \c m_has_dragged to \c true, 
     *
     * We also check the mode of operation:
     * - If it is VP_ROTATE_CAMERA, VP_PAN_CAMERA or VP_ZOOM_CAMERA, we need take no further action.
     * - If it is VP_ROTATE_OBJECT, VP_MOVE_OBJECT or VP_SCALE_OBJECT, we generate an appropriate transformation or scale action.
     *
     * \b FL_RELEASE
     * When an \c FL_RELEASE is received, we check the mode of operation:
     * - If it is VP_ROTATE_CAMERA, VP_PAN_CAMERA or VP_ZOOM_CAMERA, we need take no further action.
     * - If it is VP_ROTATE_OBJECT, VP_MOVE_OBJECT or VP_SCALE_OBJECT, we first check whether the user clicked a selected object or not. If he did, we take no further action. If he didn't, we check whether the user has dragged the mouse or not:
     *   - If the user has dragged (\c m_has_dragged is \c true), we clear the current selection by calling \c UserInterfaceUtilities::clear_selection().
     *   - If the user has not dragged (\c m_has_dragged is \c false), it will check the stored \c ObjectID.
     *     - If an \c ObjectID of 0 was returned, there was no object under the cursor so we clear the current selection by calling \c UserInterfaceUtilities::clear_selection().
     *     - If the \c ObjectID was non-zero, and \c CTRL was depressed during the click, we will call \c UserInterfaceUtilities::toggle_select(). If \c CTRL was not depressed, we will instead call \c UserInterfaceUtilities::select().
     *
     * \param msg The event ID, passed by FLTK.
     */
    virtual int handle(int msg);





    /** \brief Draws the OpenGL graphics or performs a selection pass.
     *
     * Depending on whether \c m_selection_pass is \c true or \c false, this method will call \c Renderer::pick() or \c Renderer::render() respectively.
     *
     * If we called \c Renderer::pick(), we will store the returned \c ObjectID in \c m_clicked_object. We will also check whether this clicked object was in the list of selected objects or not.
     * - If it was, set \c m_clicked_selected to \c true and calculate the temporary averaged transformation matrix for the list of selected objects.
     * - Otherwise set it to \c false.
     *
     */
    virtual void draw();

    int select_mode(int msg, int select_button = FL_LEFT_MOUSE,
                    int clear_button = FL_RIGHT_MOUSE);
    int free_mode_camera(int msg);
    int camera_pan(int msg, int pan_button = FL_LEFT_MOUSE);
    int camera_rotate(int msg, int rotate_button = FL_LEFT_MOUSE);
    int camera_zoom(int msg, int zoom_button = FL_LEFT_MOUSE);

    int move_object(int msg, int move_button = FL_LEFT_MOUSE);    

    /** \brief Sets the camera for this viewport
     *
     * Sets the pointer to the camera associated with the viewport, this to
     * relieve \c Viewport of the task of polling the camera from the parent
     * window for each call to \c Renderer::render() or \c Renderer::pick().
     *
     * Since a new camera has been selected for this viewport, we need to
     * redraw the OpenGL window, which we do by calling \c redraw().
     *
     * This method is typically called by \c
     * ViewFrame::camera_selection_callback() when the user selects a new
     * camera in the list.
     *
     * \param camera A pointer to the camera object whose \c ObjectID to fetch.
     */
    void set_camera(ObjectID camera);


    /** \brief Gets the camera for this viewport
     *
     * \return The \c ObjectID of the camera shown in the viewport.
     */
    ObjectID get_camera() const {return m_camera;}

    void save_snapshot() throw();

    void toggle_continual_snapshots(bool enable) throw();

    bool *get_visible_layers() throw()
    {
        return m_visible_layers;
    }
private:
    std::string generate_snapshot_filename(
        const std::string &prefix, const std::string &postfix) const throw();

private:
    /** \brief Checks to see that the camera set for this viewport is still
     * valid.
     *
     * \return A pointer to this viewport's camera.
     */
    CameraObject *get_camera(SceneTreePtr &ptr);

    /** \brief The ObjectID of the camera associated with this \c Viewport.*/
    ObjectID m_camera;

    /** \brief This is set to \c true whenever a FL_DRAG message is received by the handle method, to differentiate between clicking and dragging. It is set to \c false when the FL_RELEASE message is received.*/
    bool m_has_dragged;

    /** \brief Specifies which layers are visible in this viewport. <tt>"m_visible_layers[0] == true"</tt> means layer \b 1 is visible. 
     */
    bool m_visible_layers[16];

    /** \brief The time point at which the last action was sent.
     */
    boost::posix_time::ptime m_last_action_time;

    /** \brief The mouse cursor coordinates as they were when the last event
     * occurred.
     */
    int m_old_x;
    int m_old_y;
    

    /** \brief Flag that keeps track of whether the OpenGL context has been 
     * setup. Is false initially and set to true the first time draw() is 
     * called.
     *
     * Note that the GL context needs to be re-setup if lost. Keeping track of
     * whether the context was lost is not handled by this flag, FLTK's valid()
     * method does it for us and is polled before trying to use the context.
     */
    bool m_gl_context_is_setup;

    bool m_is_fullscreen;

    bool m_save_snapshot;

    bool m_continual_snapshots_enabled;

    std::string m_sequence_prefix;
};





}
}
#endif //__PEEKABOT_GUI_VIEWPORT_HH
