/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_VISIBILITY_GROUP_HH
#define __PEEKABOT_GUI_VISIBILITY_GROUP_HH

#include "PropertyVisitor.hh"
#include "PropertyGroup.hh"
#include "../SceneTreePtr.hh"

#include <set>
#include <FL/Fl_Button.H>
#include <FL/Fl_Check_Button.H>
#include <FL/Fl_Slider.H>

#include <boost/date_time/posix_time/posix_time.hpp>

namespace peekabot
{

namespace gui
{
    class VisibilityTrigger;


    /** \brief A group of properties and controls for setting visibility properties.
     *
     * This group contains a \c Fl_Check_Button for hiding/showing an object, a group of \c Fl_Button objects in radio button mode for selecting which layer this object belongs to, two sliders for showing this object's opacity and accumulated opacity, as well as a check button for selecting whether this object should enforce an absolute opacity value.
     *
     * It makes us of the \c VisibilityTrigger class to set and update the
     * values for the controls in the group.
     *
     */
class VisibilityGroup : public PropertyGroup
{
public:
    /** \brief Creates a new \c VisibilityGroup object.
     *
     * Allocates and adds the sub widgets and fills them with properties by
     * creating a \c VisibilityTrigger for the current selection.
     *
     * It also:
     * - Creates a \c Fl_Check_Button widget for hiding or showing this object
     * and register the \c hide_callback() function for this button. 
     * - Creates a group of 16 radio buttons (\c Fl_Button)for selecting which
     * layer the object belongs to, and register the \c set_layer_callback()
     * function for them.
     * - Creates two sliders (\c Fl_Slider), one showing this object's own
     * opacity value, and one showing the opacity accumulated from its
     * ancestors. It also creates a check button (\c Fl_Check_Button) for
     * selecting whether to force an absolute opacity value. It registers the
     * \c set_opacity_callback() for these controls.
     *
     * The radio button group will show the union of layers selected, the
     * hide/show and absolute opacity status of an arbitrarily chosen object,
     * and the averaged opacity value of all objects in the scene. Any
     * alterations to these controls will result in the same setting being
     * applied to all objects in the selection.

     * 
     */
    VisibilityGroup(int x, int y);
    
    /** \brief Deletes the trigger allocated for this group.
     */
    virtual ~VisibilityGroup();


    /** \brief Virtual method for filling this group with visibility information
     * from a set of objects.
     *
     * This method registers a \c VisibilityTrigger for a newly selected object
     * and deregisters triggers for objects that have been deselected.
     *
     * This method is called by the \c PropertyVisitor upon creation, and
     * whenever the selection changes.
     *
     */
    virtual void set_properties(const std::set<ObjectID> &objects);

    /** \brief Callback for hiding or showing objects
     *
     * This is called whenever the user clicks the "Hide" check button.
     *
     * Will generate a \c HideObject action based on the status of the hide
     * check box.
     *
     * \param w A pointer to the \c Fl_Check_Button control triggering the callback.
     * \param data A pointer to \c VisibilityGroup the control is in.
     */
    static void hide_callback(Fl_Widget* w, void* data);


    /** \brief Callback for selecting which layer the object is shown in.
     *
     * This is called whenever the user clicks one of the 16 visibility buttons.
     *
     * Will generate a \c SetLayer action based on the value entered by the user.
     *
     * \param w A pointer to the \c Fl_Button control triggering the callback.
     * \param data A pointer to \c VisibilityGroup the control is in.
     */
    static void set_layer_callback(Fl_Widget* w, void* data);




    /** \brief Callback for setting the opacity value of objects
     *
     * This is called whenever the user drags the upper opacity slider, or when he clicks the "Absolute" check button.
     *
     * Will generate a \c SetOpacity action based on the value entered by the user.
     *
     * \param w A pointer to the \c Fl_Slider or \c Fl_Check_Button control triggering the callback.
     * \param data A pointer to the \c VisibilityGroup the control was in.
     */
    static void set_opacity_callback(Fl_Widget* w, void* data);


    void register_triggers(const std::set<ObjectID> &objects,
                           SceneTreePtr &ptr);

    void deregister_triggers(const std::set<ObjectID> &objects,
                             SceneTreePtr &ptr);

    /** \brief Removes all pending updates for this group.
     */
    void clear_updates();

private:
    Fl_Check_Button *m_absolute_opacity;
    
    Fl_Check_Button *m_hidden;

    Fl_Check_Button *m_apply_recursively;

    Fl_Slider *m_opacity_slider;
    
    Fl_Slider *m_accumulated_opacity_slider;
    
    Fl_Button *m_layer_buttons[16];

    /** \brief A pointer to the trigger this group registers with selected
     * objects.
     */
    VisibilityTrigger* m_trigger;

    std::set<ObjectID> m_objects;

    boost::posix_time::ptime m_last_action_time;
};

}
}
#endif //__PEEKABOT_GUI_VISIBILITY_GROUP_HH
