/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <FL/Fl_Slider.H>
#include <FL/Fl_Check_Button.H>
#include <FL/Fl_Button.H>
#include <cstring>

#include "VisibilityTrigger.hh"
#include "UserInterface.hh"
#include "UpdateActions.hh"
#include "../SceneObject.hh"
#include "../SceneTreePtr.hh"

using namespace peekabot;
using namespace peekabot::gui;


VisibilityTrigger::VisibilityTrigger(
    Fl_Slider* opacity_slider,
    Fl_Slider* accumulated_opacity_slider,
    Fl_Check_Button* absolute_button,
    Fl_Check_Button* hidden_button, Fl_Button** layer_buttons)
    : m_opacity_slider(opacity_slider),
      m_accumulated_opacity_slider(accumulated_opacity_slider),
      m_absolute_button(absolute_button),
      m_hidden_button(hidden_button), m_layer_buttons(layer_buttons)
{
    memset(m_layer_button_values, 0, sizeof(m_layer_button_values));
}


VisibilityTrigger::~VisibilityTrigger()
{
    SceneTreePtr ptr = TheSceneTree::acquire();
    while( !m_objects.empty() )
        detach_from(*m_objects.begin());
}


void VisibilityTrigger::on_opacity_changed(SceneObject *object)
{
    // Set the relative and accumulated opacity sliders to the average of
    // all objects in the activators set.

    real_unit relative_opacity = 0, accumulated_opacity = 0;
    int count = m_objects.size();
    for(std::set<SceneObject *>::const_iterator it = m_objects.begin();
        it != m_objects.end(); ++it)
    {
        relative_opacity    += (*it)->get_opacity().opacity;
        accumulated_opacity += (*it)->get_accumulated_opacity();
    }
    relative_opacity /= count;
    accumulated_opacity /= count;
    
    // Update the opacity sliders
    TheUserInterface::instance().post_action(
        m_opacity_slider, new UpdateValuator(m_opacity_slider, relative_opacity));    

    TheUserInterface::instance().post_action(
        m_accumulated_opacity_slider, new UpdateValuator(m_accumulated_opacity_slider,
                                                         accumulated_opacity));

    // Update the "absolute" check button.
    bool absolute = (*m_objects.begin())->get_opacity().is_absolute;
    TheUserInterface::instance().post_action(
        m_absolute_button, new UpdateButton(m_absolute_button, (absolute) ? 1 : 0)); 

}

void VisibilityTrigger::on_layer_set(SceneObject *object)
{
    int layers[16];

    // Store the old layer button values in a temporary array and zero
    // the old array
    for(int i = 0; i < 16; ++i)
    {
        layers[i] = m_layer_button_values[i];
        m_layer_button_values[i] = 0;
    }
    
    for(std::set<SceneObject *>::const_iterator it = m_objects.begin();
        it != m_objects.end(); ++it)
    {
        // We're displaying the union of layer values for all objects
        m_layer_button_values[(*it)->get_layer()] = 1;
    }
    
    // If any layer button value differs from before, send an update action.
    for(int i = 0; i < 16; ++i)
    {
        if(layers[i] != m_layer_button_values[i])
            TheUserInterface::instance().post_action(
                m_layer_buttons[i], new UpdateButton(m_layer_buttons[i],
                                                     m_layer_button_values[i]));    
    }
}

void VisibilityTrigger::on_hidden_set(SceneObject *object)
{
    bool hidden = object->is_hidden();
    TheUserInterface::instance().post_action(
        m_hidden_button, new UpdateButton(m_hidden_button, (hidden) ? 1 : 0));    
}

void VisibilityTrigger::attach_to(SceneObject *object)
{
    // Add object to the set of registered objects
    if( m_objects.insert(object).second )
    {
        // Register triggers
        object->m_on_layer_set.connect(
            boost::bind(&VisibilityTrigger::on_layer_set, this, _1));
        
        object->m_on_hidden_set.connect(
            boost::bind(&VisibilityTrigger::on_hidden_set, this, _1));
        
        object->m_on_accum_opacity_changed.connect(
            boost::bind(&VisibilityTrigger::on_opacity_changed, this, _1));
        
        object->m_on_subtree_detached.connect(
            boost::bind(&VisibilityTrigger::detach_from, this, _1));
        
        on_opacity_changed(object);
        on_layer_set(object);
        on_hidden_set(object);
    }
}

void VisibilityTrigger::detach_from(SceneObject *object)
{
    // Erase object from set of registered objects
    if( m_objects.erase(object) > 0 )
    {
        // Deregister triggers
        object->m_on_layer_set.disconnect(
            boost::bind(&VisibilityTrigger::on_layer_set, this, _1));
        
        object->m_on_hidden_set.disconnect(
            boost::bind(&VisibilityTrigger::on_hidden_set, this, _1));
        
        object->m_on_accum_opacity_changed.disconnect(
            boost::bind(&VisibilityTrigger::on_opacity_changed, this, _1));
        
        object->m_on_subtree_detached.disconnect(
            boost::bind(&VisibilityTrigger::detach_from, this, _1));

        //on_opacity_changed(object);
        //on_layer_set(object);
        //on_hidden_set(object);
    }
}
