/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_VISIBILITY_TRIGGER_HH
#define __PEEKABOT_GUI_VISIBILITY_TRIGGER_HH

#include <set>

class Fl_Slider;
class Fl_Check_Button;
class Fl_Button;

namespace peekabot
{

class SceneObject;

namespace gui
{


    /** \brief Trigger for updating the visibility fields in a\c VisibilityGroup.
     *
     * This trigger is registered when a \c VisibilityGroup is created in a \c
     * PropertyFrame. It will contain pointers to the visibility fields that it
     * is registered for.
     *
     * It is registered for the ON_OPACITY_CHANGED, ON_LAYER_SET and
     * ON_HIDDEN_SET stimuli.
     *
     * When triggered it generates update events and posts them on the GUI update
     * queue via \c UserInterface::post_event().
     *
     */

class VisibilityTrigger
{

public:
    /** \brief This constructor simply sets the pointers to the opacity and
     * relative opacity sliders, the hidden checkbox and the layer radio group.
     */
    VisibilityTrigger(Fl_Slider* opacity_slider,
                      Fl_Slider* accumulated_opacity_slider,
                      Fl_Check_Button* absolute_button,
                      Fl_Check_Button* hidden_button, Fl_Button** layer_buttons);

    ~VisibilityTrigger();

    void on_opacity_changed(SceneObject *object);
    void on_layer_set(SceneObject *object);
    void on_hidden_set(SceneObject *object);

    void attach_to(SceneObject *object);
    void detach_from(SceneObject *object);

private:

    /** \brief Pointers to the opacity and accumulated opacity sliders of the
     * \c VisibilityGroup.
     */
    Fl_Slider* m_opacity_slider, *m_accumulated_opacity_slider;

    /** \brief A pointer to the "hidden"-checkbox of the group.
     */
    Fl_Check_Button *m_absolute_button, *m_hidden_button;

    /** \brief A pointer to the array of layer buttons.
     */
    Fl_Button** m_layer_buttons;

    int m_layer_button_values[16];

    /** \brief A set storing all objects the trigger is registered with.
     */
    std::set<SceneObject *> m_objects;
};


}
}

#endif //__PEEKABOT_GUI_VISIBILITY_TRIGGER_HH
