/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_GUI_WORKSPACE_HH
#define __PEEKABOT_GUI_WORKSPACE_HH

#include <set>
#include <FL/Fl_Tile.H>

namespace peekabot
{

namespace gui
{
    class Frame;
    class PropertyFrame;
    class BrowserFrame;
    class ConsoleFrame;
    class ViewFrame;

/** \brief Class representing a workspace, specifying a frame layout.
 *
 * This class is meant to provide the user with the ability reshape the layout of frames. It inherits the ability to contain frames (\c Fl_Group objects) and to resize them by the user dragging their borders from \c Fl_Tile. 
 *
 * It is also used by \c UserInterface and \c UserInterfaceUtilities to access frames of a particular type.
 *
 */

class FrameSelector
{
public:
    FrameSelector();

    /** \brief Marks a frame as the currently selected one.
     *
     * This method is typically called by the \c Frame::handle() method when it detects a mouse click within its area.
     *
     * \param frame A pointer to the frame to set as selected.
     *
     */
    void select_frame(Frame* frame);
    Frame* get_selected_frame() const;
protected:
    /** \brief A pointer to the currently selected frame.
     */
    Frame* m_selected_frame;

};

    class Workspace : public Fl_Tile, public FrameSelector
{
public:
    /** \brief Constructs a new Workspace by calling the base \c Fl_Tile
     * constructor.
     */
    Workspace(int x, int y, int w, int h, char* l = 0);

    virtual ~Workspace();


    /** \brief Registers a \c PropertyFrame frame as belonging to this workspace.
     *
     * This method is called whenever a new \c PropertyFrame is created.
     *
     * \param frame A pointer to the \c PropertyFrame to register.
     *
     */    
    void register_property_frame(PropertyFrame* frame);


    /** \brief Registers a \c ViewFrame frame as belonging to this workspace.
     *
     * This method is called whenever a new \c ViewFrame is created.
     *
     * \param frame A pointer to the \c ViewFrame to register.
     *
     */
    void register_view_frame(ViewFrame* frame);


    /** \brief Registers a \c BrowserFrame frame as belonging to this workspace.
     *
     * This method is called whenever a new \c BrowserFrame is created.
     *
     * \param frame A pointer to the \c BrowserFrame to register.
     *
     */
    void register_browser_frame(BrowserFrame* frame);


    /** \brief Registers a \c ConsoleFrame frame as belonging to this workspace.
     *
     * This method is called whenever a new \c ConsoleFrame is created.
     *
     * \param frame A pointer to the \c ConsoleFrame to register.
     *
     */
    void register_console_frame(ConsoleFrame* frame);



    /** \brief Deregisters a \c PropertyFrame from to this workspace.
     *
     * If \c frame is the currently selected frame, the currently selected frame will be set to \c NULL. 
     *
     * This method is called whenever a \c PropertyFrame is destroyed.
     *
     * \param frame A pointer to the \c PropertyFrame to deregister.
     *
     */
    void deregister_property_frame(PropertyFrame* frame);


    /** \brief Deregisters a \c ViewFrame from to this workspace.
     *
     * If \c frame is the currently selected frame, the currently selected frame will be set to \c NULL. 
     *
     * This method is called whenever a \c ViewFrame is destroyed.
     *
     * \param frame A pointer to the \c ViewFrame to deregister.
     *
     */
    void deregister_view_frame(ViewFrame* frame);


    /** \brief Deregisters a \c BrowserFrame from to this workspace.
     *
     * If \c frame is the currently selected frame, the currently selected frame will be set to \c NULL. 
     *
     * This method is called whenever a \c BrowserFrame is destroyed.
     *
     * \param frame A pointer to the \c BrowserFrame to deregister.
     *
     */
    void deregister_browser_frame(BrowserFrame* frame);


    /** \brief Deregisters a \c ConsoleFrame from to this workspace.
     *
     * If \c frame is the currently selected frame, the currently selected frame will be set to \c NULL. 
     *
     * This method is called whenever a \c ConsoleFrame is destroyed.
     *
     * \param frame A pointer to the \c ConsoleFrame to deregister.
     *
     */
    void deregister_console_frame(ConsoleFrame* frame);
    


    /** \brief Returns a set of the \c PropertyFrame objects in this workspace.
     *
     *  This method is usually called by \c UserInterfaceUtilities whenever an object is selected or deselected.
     *
     * \return A constant reference to a set of pointers to the \c PropertyFrame objects in this workspace.
     *
     */
    std::set<PropertyFrame*> & get_property_frames();




    /** \brief Returns a set of the \c ViewFrame objects in this workspace.
     *
     *  This method is usually called from \c UserInterface::redraw_viewports when the viewports are about to be redrawn.
     *
     * \return A constant reference to a set of pointers to the \c ViewFrame objects in this workspace.
     *
     */
    std::set<ViewFrame*> & get_view_frames();



    /** \brief Returns a set of the \c BrowserFrame objects in this workspace.
     *
     *
     * \return A constant reference to a set of pointers to the \c BrowserFrame objects in this workspace.
     *
     */
    std::set<BrowserFrame*> & get_browser_frames();

    /** \brief Returns a set of the \c ConsoleFrames objects in this workspace.
     *
     * Typically called from \c UserInterface::send_to_console() when an Action generates console output.
     *
     * \return A constant reference to a set of pointers to the \c ConsoleFrame objects in this workspace.
     *
     */
    std::set<ConsoleFrame*> & get_console_frames();

private:

    /** \brief A set of pointers to the property frames in this workspace.
     */
    std::set<PropertyFrame*> m_property_frames;

    /** \brief A set of pointers to the view frames in this workspace.
     */
    std::set<ViewFrame*> m_view_frames;

    /** \brief A set of pointers to the browser frames in this workspace.
     */
    std::set<BrowserFrame*> m_browser_frames;

    /** \brief A set of pointers to the console frames in this workspace.
     */
    std::set<ConsoleFrame*> m_console_frames;
};

}
}
#endif //__PEEKABOT_GUI_WORKSPACE_HH
