/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <cmath>
#include "BoundingVolumes.hh"

using namespace peekabot;
using namespace peekabot::renderer;



// ---------------- BoundingSphere implementation ------------------

BoundingSphere::BoundingSphere(const Vector3r &pos, real32 r) throw()
    : m_pos(pos), m_r(r)
{
}

BoundingSphere::BoundingSphere() throw()
{
}


void BoundingSphere::set_radius(real32 r) throw()
{
    m_r = r;
}


void BoundingSphere::set_pos(Vector3r pos) throw()
{
    m_pos = pos;
}


bool BoundingSphere::operator==(const BoundingSphere &bs) const throw()
{
    return ( m_pos == bs.m_pos && m_r == bs.m_r );
}

bool BoundingSphere::operator!=(const BoundingSphere &bs) const throw()
{
    return !(*this == bs);
}


IntersectionTestResult BoundingSphere::contained_in(const BoundingSphere &container) 
    const throw()
{
    // this could probably be replaced by a branch and sq_dist, getting away
    // without the sqrt.
    /*real32 d = m_pos.dist(container.m_pos);

    if( d - m_r > container.m_r )
        return DISJOINT;
    else if( d + m_r <= container.m_r )
        return INSIDE;
    else
        return INTERSECT;*/


    // OPT: optimization possiblity, use this cheap binary test:

	// if the distance between the centers is less than the sum
	// of the radii, then we have an intersection
	// we calculate this using the squared lengths for speed
	if( m_pos.dist_squared(container.m_pos) < 
        ((m_r + container.m_r)*(m_r + container.m_r)))
		return INTERSECT;

    return DISJOINT;
}



IntersectionTestResult BoundingSphere::contained_in(const BoundingCone &container) 
    const throw()
{
    // TODO: Try to optimize this - get rid of the sqrts.

    Vector3r D = m_pos - container.get_vertex() + (m_r/container.get_sin_angle())*container.get_axis();
    
    if( container.get_axis() * D >= D.norm() * container.get_cos_angle() ) 
    { 
        D = m_pos - container.get_vertex();
        real32 D_norm = D.norm();

        if ( -container.get_axis() * D >= D_norm*container.get_sin_angle() ) 
        { 
            return (D_norm <= m_r) ? INTERSECT:DISJOINT; 
        } 
        else 
        { 
            return INTERSECT; 
        } 
    }
    else
    {
        return DISJOINT; 
    } 
}



IntersectionTestResult BoundingSphere::contained_in(const Plane &container) 
    const throw()
{
    real32 dist = container.get_normal() * m_pos + container.get_d();

    if( dist > m_r )
        return DISJOINT;
    else if( dist <= m_r && -dist <= m_r )
        return INTERSECT;
    else
        return INSIDE;
}


IntersectionTestResult BoundingSphere::intersect(const BoundingSphere &bv) const throw()
{
    return DISJOINT;
}


IntersectionTestResult BoundingSphere::intersect(const Plane &bv) const throw()
{
    return DISJOINT;
}






// ---------------- BoundingCone implementation ------------------


BoundingCone::BoundingCone() throw()
{
}

BoundingCone::BoundingCone(Vector3r vertex, Vector3r axis, real32 angle) throw()
    : m_vertex(vertex),
      m_axis(axis),
      m_angle(angle),
      m_sin_angle(sinf(m_angle)),
      m_cos_angle(cosf(m_angle))
{
    m_axis.normalize();
}

const Vector3r &BoundingCone::get_vertex() const throw()
{
    return m_vertex;
}

void BoundingCone::set_vertex(const Vector3r &vertex) throw()
{
    m_vertex = vertex;
}

const Vector3r &BoundingCone::get_axis() const throw()
{
    return m_axis;
}

void BoundingCone::set_axis(const Vector3r &axis) throw()
{
    m_axis = axis;
    m_axis.normalize();
}

void BoundingCone::set_angle(real32 angle) throw()
{
    m_angle = angle;
    m_sin_angle = sinf(angle);
    m_cos_angle = cosf(angle);
}

real32 BoundingCone::get_angle() const throw()
{
    return m_angle;
}

real32 BoundingCone::get_sin_angle() const throw()
{
    return m_sin_angle;
}

real32 BoundingCone::get_cos_angle() const throw()
{
    return m_cos_angle;
}


IntersectionTestResult BoundingCone::intersect(const BoundingSphere &bv) 
    const throw()
{
    return INTERSECT;
}




IntersectionTestResult BoundingCone::intersect(const Plane &bv) const throw()
{
    return DISJOINT;
}





// ---------------- Plane implementation ------------------


Plane::Plane() throw()
{
}

Plane::Plane(Vector3r normal, real32 d) throw()
    : m_normal(normal), 
      m_d(d)
{
}

const Vector3r &Plane::get_normal() const throw()
{
    return m_normal;
}

void Plane::set_normal(const Vector3r &normal) throw()
{
    m_normal = normal;
}

void Plane::set_normal(Vector3r::UnitType x, 
                       Vector3r::UnitType y, 
                       Vector3r::UnitType z) throw()
{
    m_normal(0) = x;
    m_normal(1) = y;
    m_normal(2) = z;
}

inline real32 Plane::get_d() const throw()
{
    return m_d;
}

void Plane::set_d(real32 d) throw()
{
    m_d = d;
}


void Plane::normalize() throw()
{
    m_normal.normalize();
}


IntersectionTestResult Plane::intersect(const BoundingSphere &bv) 
    const throw()
{
    return INTERSECT;
}


IntersectionTestResult Plane::intersect(const Plane &bv) const throw()
{
    return DISJOINT;
}
