/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_RENDERER_BOUNDING_VOLUMES_HH
#define __PEEKABOT_RENDERER_BOUNDING_VOLUMES_HH


#include "../Vector.hh"


namespace peekabot
{
namespace renderer
{
    class BoundingSphere;
    class BoundingCone;
    class Plane;
    class AABB;
    class OBB;

    /**
     * \brief Trinary result for intersection tests.
     */
    enum IntersectionTestResult
    {
        /**
         * \brief The volume is fully contained in the volume it was
         * tested against. I.e., it is a subset of the other volume.
         *
         * When testing against a plane, this enumerator signifies that the
         * object lies on plane's negative normal side, without intersecting
         * the plane.
         */
        INSIDE = 2,
        /**
         * \brief The volume is partially contained in the volume it was
         * tested against. I.e., the volumes intersect.
         */
        INTERSECT = 1,
        /**
         * \brief The two tested volumes are disjoint.
         */
        DISJOINT = 0
    };


    /**
     * \brief Polymorphic bounding volume class.
     *
     * Provides (abstract) interfaces for testing a bounding volume (BV)
     * against a set of common bounding volumes.
     */
    class BoundingVolume
    {
    public:
        virtual ~BoundingVolume() throw() {}

        virtual IntersectionTestResult intersect(const BoundingSphere &bv) const throw() = 0;
        virtual IntersectionTestResult intersect(const Plane &bv) const throw() = 0;
    };

    /*struct BVTests
    {
        static IntersectionTestResult intersect(const Sphere &contains, 
                                           const Vector4 &containee);
        static IntersectionTestResult intersect(const Sphere &contains, 
                                           const Vector4 &containee);
    };*/


    class BoundingSphere : public BoundingVolume
    {
        Vector3r  m_pos;
        real32     m_r;

    public:
        BoundingSphere() throw();

        BoundingSphere(const Vector3r &pos, real32 r) throw();

        inline real32 get_radius() const throw() { return m_r; }

        inline const Vector3r &get_pos() const throw() { return m_pos; }

        void set_radius(real32 r) throw();

        void set_pos(Vector3r pos) throw();

        bool operator==(const BoundingSphere &bs) const throw();

        bool operator!=(const BoundingSphere &bs) const throw();

        IntersectionTestResult contained_in(const BoundingCone &container) const throw();
        IntersectionTestResult contained_in(const BoundingSphere &container) const throw();
        IntersectionTestResult contained_in(const Plane &container) const throw();

        virtual IntersectionTestResult intersect(const BoundingSphere &bv) const throw();
        virtual IntersectionTestResult intersect(const Plane &bv) const throw();
    };


    class BoundingCone : public BoundingVolume
    {
        Vector3r m_vertex;
        Vector3r m_axis; // of unit length
        real32   m_angle; // in rad
        real32   m_sin_angle;
        real32   m_cos_angle;
        
    public:
        BoundingCone() throw();

        BoundingCone(Vector3r vertex, Vector3r axis, real32 angle) throw();

        const Vector3r &get_vertex() const throw();

        void set_vertex(const Vector3r &vertex) throw();

        const Vector3r &get_axis() const throw();

        void set_axis(const Vector3r &vertex) throw();

        void set_angle(real32 angle) throw();

        real32 get_angle() const throw();

        real32 get_sin_angle() const throw();

        real32 get_cos_angle() const throw();



        virtual IntersectionTestResult intersect(const BoundingSphere &bv) const throw();
        virtual IntersectionTestResult intersect(const Plane &bv) const throw();
    };


    class Plane : public BoundingVolume
    {
        Vector3r m_normal;
        real32   m_d;

    public:
        Plane() throw();

        Plane(Vector3r normal, real32 d) throw();

        const Vector3r &get_normal() const throw();

        void set_normal(const Vector3r &normal) throw();

        void set_normal(Vector3r::UnitType x, 
                        Vector3r::UnitType y, 
                        Vector3r::UnitType z) throw();

        real32 get_d() const throw();

        void set_d(real32 d) throw();

        void normalize() throw();

        virtual IntersectionTestResult intersect(const BoundingSphere &bv) const throw();
        virtual IntersectionTestResult intersect(const Plane &bv) const throw();
    };


    class Point : public BoundingVolume
    {
        Vector3r m_pos;
    public:
        
    };

    class AABB : public BoundingVolume
    {
    };

    class OBB : public BoundingVolume
    {
    };


}
}


#endif // __PEEKABOT_RENDERER_BOUNDING_VOLUMES_HH
