/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_RENDERER_CULLABLE_HH
#define __PEEKABOT_RENDERER_CULLABLE_HH


#include <list>
#include <stdexcept>
#include <boost/utility.hpp>
#include <boost/function.hpp>
#include <boost/thread_safe_signal.hpp>

#include "Renderable.hh"
#include "BoundingVolumes.hh"
#include "Cloneable.hh"


namespace peekabot
{
    namespace renderer
    {
        class SceneNode;

        namespace primitives
        {
            class Camera;
        }

        /**
         * \brief An interface that can be implemented in graphical entities 
         * to be able to use the view frustum culling services provided by
         * the Octree.
         *
         * The concept modeled by this interface is, as the overview says, the
         * ability to be culled. Thus, some crucial information must be provided
         * by the implementing class: bounding volume(s) and a way of accessing
         * the renderable entities that the cullable is comprised of.
         *
         * Furthermore, a cullable is usually owned by a SceneNode. The 
         * rendering process uses information from the owning node when 
         * determining what and how to render the cullable's renderables. E.g.,
         * it omits the cullable for rendering if it's parent is hidden or in 
         * a layer not being drawn.
         *
         * \sa Renderable, SceneNode
         */
        class Cullable : public Cloneable, public boost::noncopyable
        {
            friend class SceneNode;

        public:
            typedef std::list<Renderable *> Renderables;

            /**
             * \brief The signature required for the callback functions used to
             * enumerate all of the cullable's renderables.
             */
            typedef void (* EnumerationCallback)(Renderable *);

            typedef std::pair<float, float> OnScreenSize;

            Cullable(OnScreenSize *on_screen_size = 0);

            /**
             * \brief A destroyed cullable automatically removes itself from its
             * parent's list of cullables. Of course this applies only when it's 
             * not been orphaned.
             */
            virtual ~Cullable();

            virtual Cullable *clone() const throw() = 0;


            /**
             * \brief Return the SceneNode that owns this Cullable.
             *
             * \return The owning SceneNode, or null if it has none.
             */
            SceneNode *get_parent() throw();

            /**
             * \copydoc get_parent()
             */
            const SceneNode *get_parent() const throw();

            /**
             * \brief Call the provided callback function for each of the 
             * Cullable's renderables that should be rendered given the 
             * specified camera (can be omitted, see below).
             *
             * Passing the camera along as a parameter allows us to implement 
             * LODs, and do it with much flexibility.
             *
             * \param camera The camera used for rendering. If null is specified
             * all of the cullable's renderables should be "returned".
             * \param callback The callback function to call with the contained
             * renderables.
             */
            virtual void get_renderables(
                primitives::Camera *camera, EnumerationCallback callback) throw() = 0;

            /**
             * \brief Return a list of all the cullables renderables.
             */
            Renderables get_renderables() throw();

            /**
             * \brief Get the cullable's bounding sphere.
             */
            const BoundingSphere &get_bounding_sphere() throw();


            /**
             * \brief Get the cullable's bounding volume.
             *
             * In the stock implementation the bounding sphere is returned.
             * It can be reimplemented to provide a better approximative volume.
             */
            virtual const BoundingVolume &get_bounding_volume() throw();

            
            // TODO:
            //
            // Let this return the template state..
            // That way we can have cullables that aren't templated on the parent node
            // or use, for example, a unscaled template state.
            //virtual const State *get_template_state() const throw();

            //virtual void transform(const Matrix4r &transformation) throw() = 0;


            // force the cullable's bounding volumes to be updated
            virtual void calculate_bvs() throw() = 0;
            
            virtual void transform_geometry(
                const boost::function<Vector3r (const Vector4r &)> &fun);


            bool is_size_invariant() const throw()
            {
                return m_on_screen_size != 0;
            }

            const OnScreenSize &get_on_screen_size() const throw(std::logic_error)
            {
                if( m_on_screen_size == 0 )
                    throw std::logic_error("Cullable is not size-invariant");
                return *m_on_screen_size;
            }


        public:
            //
            // Signals
            //
            boost::signal<void (Cullable *)> on_bv_set;
            


        protected:
            virtual void set_bounding_sphere(const BoundingSphere &bsphere) throw();

        private:
            void set_parent(SceneNode *parent) throw();

        private:
            /**
             * \brief The SceneNode "owning" the Cullable. 
             *
             * Initially set to null, signifying that it has no parent.
             */
            SceneNode *m_parent;

            BoundingSphere m_bsphere;

            OnScreenSize *m_on_screen_size;
        };


    }
}


#endif // __PEEKABOT_RENDERER_CULLABLE_HH
