/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_RENDERER_FRAME_COLLECTOR_HH
#define __PEEKABOT_RENDERER_FRAME_COLLECTOR_HH


#include "../Types.hh"


namespace peekabot
{
    namespace renderer
    {


        /**
         * \brief Provides an interface enabling third parties to retrieve the
         * resulting frame of a render pass.
         *
         * This interface can be used to implement snapshotting and video 
         * recording. 
         *
         * It works like this: when you supply a FrameCollector object when 
         * rendering it will call collect_frame(int, int, const uint8_t *) with 
         * the frame data once the scene has been rendered. The format in which 
         * the data is fed to \c collect_frame can be specified in the 
         * constructor or by calling set_pixel_format(PixelFormat).
         */
        class FrameCollector
        {
        public:
            enum PixelFormat
            {
                RGB,
                BGR
            };

            FrameCollector(PixelFormat format = RGB) : m_pixel_format(format) {}

            virtual ~FrameCollector() {}

            virtual void collect_frame(int w, int h, const uint8_t *pixel_data) throw() = 0;

            PixelFormat get_pixel_format() const throw()
            {
                return m_pixel_format;
            }

        protected:
            void set_pixel_format(PixelFormat format) throw()
            {
                m_pixel_format = format;
            }

        private:
            PixelFormat m_pixel_format;
        };


    }
}


#endif // __PEEKABOT_RENDERER_FRAME_COLLECTOR_HH
