/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_RENDERER_ID_TRANSLATOR_HH
#define __PEEKABOT_RENDERER_ID_TRANSLATOR_HH


#include <map>
#include <set>
#include <stdexcept>
#include "../Types.hh"
#include "../Singleton.hh"
#include "SceneNode.hh"
#include "Cullable.hh"
#include "IDTypes.hh"


namespace peekabot
{
    namespace renderer
    {

        // How do we keep it up to date?? I.e. how do we remove stuff from it...
        //
        // 1. Triggers - NodeTrigger + EntityTrigger. Quite messy, but doable.
        // 2. In the destructors of Node/Entity/Cullable. Not as decoupled, but easy, and less demanding on memory.
        //
        // Entries are added by whomever needs them, typically Renderer.

        /**
         * \brief Translates IDs into SceneNode, SceneEntity and Cullable 
         * pointers.
         */
        class IDTranslator
        {
        public:
            typedef std::set<Cullable *> Primitive;

            void add_alias(NodeID id, SceneNode *node) throw()
            {
                m_id_to_node.insert(std::make_pair(id, node));
                m_node_to_id.insert(std::make_pair(node, id));
            }

            void remove_alias(SceneNode *node) throw()
            {
                NodeToIDMap::iterator it = m_node_to_id.find(node);
                
                if( it != m_node_to_id.end() )
                {
                    // Remove ID -> Node mapping
                    m_id_to_node.erase(it->second);

                    // Remove Node -> ID mapping
                    m_node_to_id.erase(it);
                }
            }

            SceneNode *get_node(NodeID id) throw(std::runtime_error)
            {
                IDToNodeMap::iterator it = m_id_to_node.find(id);
                
                if( it == m_id_to_node.end() )
                    throw std::runtime_error("No node with the given ID found");
                else
                    return it->second;
            }


           
            void add_alias(PrimitiveID id, Cullable *cullable) throw()
            {
                m_id_to_primitive[id].insert(cullable);
                m_cullable_to_id.insert(std::make_pair(cullable, id));
            }

            void remove_alias(Cullable *cullable) throw()
            {
                // Find cullable primitive ID:
                PrimitiveID id = m_cullable_to_id[cullable];

                // Remove Cullable -> ID mapping
                m_cullable_to_id.erase(cullable);

                // Remove cullable from ID -> Primitive mapping
                m_id_to_primitive[id].erase(cullable);

                // Check if all cullables for the primitive has been destroyed,
                // and delete the record of the primitive in that case:
                if( m_id_to_primitive[id].empty() )
                {
                    m_id_to_primitive.erase(id);
                }
            }

            Primitive get_primitive(PrimitiveID id) throw(std::runtime_error)
            {
                IDToPrimitiveMap::iterator it = m_id_to_primitive.find(id);
                
                if( it == m_id_to_primitive.end() )
                    throw std::runtime_error("No primitive with the given ID found");
                else
                    return it->second;
            }


        private:
            typedef std::map<NodeID, SceneNode *> IDToNodeMap;
            typedef std::map<SceneNode *, NodeID> NodeToIDMap;

            IDToNodeMap m_id_to_node;
            NodeToIDMap m_node_to_id;


            typedef std::map<PrimitiveID, Primitive> IDToPrimitiveMap;
            typedef std::map<Cullable *, PrimitiveID> CullableToIDMap;
    
            IDToPrimitiveMap m_id_to_primitive;
            CullableToIDMap m_cullable_to_id;
        };


        typedef Singleton<IDTranslator> TheIDTranslator;

        

    }
}



#endif // __PEEKABOT_RENDERER_ID_TRANSLATOR_HH
