/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <GL/glew.h>
#include <cassert>
#include <cstring>

#include "IndexBuffer.hh"
#include "../Types.hh"
#include "Device.hh"
#include "../Configuration.hh"


using namespace peekabot;
using namespace peekabot::renderer;


IndexBuffer::VBODescriptor IndexBuffer::ms_bound_vbo = 0;


IndexBuffer::IndexBuffer(bool force_no_vbos) throw()
    : m_size(0),
      m_data(0),
      m_is_mapped(false),
      m_usage(STATIC_DRAW),
      m_use_vbos(
          !force_no_vbos &&
          (!TheConfiguration::instance().get_option<bool>(
              "renderer.disable_vbos", false) &&
           Device::has_vbo_support()) )
{
    if( m_use_vbos )
    {
        glGenBuffersARB(1, &m_vbo_name);
        CHECK_GL_ERRORS();
    }
}


IndexBuffer::IndexBuffer(const IndexBuffer &buf) throw()
    : m_size(buf.m_size),
      m_data(0),
      m_is_mapped(false),
      m_usage(buf.m_usage),
      m_use_vbos(buf.m_use_vbos)
{
    if( m_use_vbos )
    {
        glGenBuffersARB(1, &m_vbo_name);
        CHECK_GL_ERRORS();
    }

    /**
     * *** Snip from the VBO extension docs: ***
     *
     * Pointer values returned
     * by MapBuffer may not be passed as parameter values to GL commands.
     * For example, they may not be used to specify array pointers, or to
     * specify or query pixel or texture image data; such actions produce
     * undefined results, although implementations may not check for such
     * behavior for performance reasons.
     *
     * Alas, mapping the buffer and copying it into a our new buffer is illegal
     * and we are forced to make an intermediary copy of the data. Sad but true.
     */

    uint8_t *data = new uint8_t[m_size];
    buf.get_data(data, m_size, 0);
    set_data(data, m_size, m_usage);
    delete[] data;
}


IndexBuffer::~IndexBuffer() throw()
{
    if( m_use_vbos )
    {
        if( is_mapped() )
            unmap();
        
        if( is_bound() )
            unbind();

        glDeleteBuffersARB(1, &m_vbo_name);
        CHECK_GL_ERRORS();
    }
}


void IndexBuffer::set_data(
    const void *data, size_t n_bytes, UsageHint usage)
    throw(std::runtime_error)
{
    m_usage = usage;
    m_size = n_bytes;

    if( m_use_vbos )
    {
        if( !is_bound() )
            bind();

        glBufferDataARB(
            GL_ELEMENT_ARRAY_BUFFER_ARB, (GLsizeiptr)n_bytes, data, usage);
        CHECK_GL_ERRORS();
    }
    else
    {
        if( m_data )
        {
            delete[] m_data;
            m_data = 0;
        }

        m_data = new uint8_t[n_bytes];
        memcpy(m_data, data, n_bytes);
    }
}
        
        
void IndexBuffer::set_data(
    const void *data, size_t n_bytes, size_t offset)
    throw(std::runtime_error)
{
    assert( m_size >= n_bytes+offset );

    if( m_use_vbos )
    {
        if( !is_bound() )
            bind();

        glBufferSubDataARB(
            GL_ELEMENT_ARRAY_BUFFER_ARB, 
            (GLintptr)offset, (GLsizeiptr)n_bytes, data);
        CHECK_GL_ERRORS();
    }
    else
    {
        assert( m_data );
        memcpy(m_data+offset, data, n_bytes);
    }
}


void IndexBuffer::get_data(void *data, size_t n_bytes, size_t offset)
    const throw(std::runtime_error)
{
    if( m_use_vbos )
    {
        if( is_mapped() )
            unmap();
        
        if( !is_bound() )
            bind();

        glGetBufferSubDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, (GLintptr)offset,
                              (GLsizeiptr)n_bytes, data);
        CHECK_GL_ERRORS();
    }
    else
    {
        memcpy(data, m_data, n_bytes);
    }
}

        
void *IndexBuffer::map_vbo(Access access, bool discard) throw()
{
    assert( m_use_vbos );

    m_is_mapped = true;

    if( !is_bound() )
        bind();
    
    /*if( discard )
      glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m_size, 0, m_usage);*/
    
    void *ret = glMapBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, access);
    CHECK_GL_ERRORS();
    return ret;
}


const void *IndexBuffer::map_vbo(bool discard) const throw()
{
    assert( m_use_vbos );

    m_is_mapped = true;

    if( !is_bound() )
        bind();
    
    /*if( discard )
      glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m_size, 0, m_usage);*/
    
    const void *ret = glMapBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, GL_READ_ONLY);
    CHECK_GL_ERRORS();
    return ret;
}
        

bool IndexBuffer::unmap_vbo() const throw()
{
    assert( m_use_vbos );

    if( is_mapped() )
    {
        m_is_mapped = false;

        
        if( !is_bound() )
            bind();
        
        bool ret = !glUnmapBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB);
        CHECK_GL_ERRORS();
        return ret;
    }

    return false;
}


size_t IndexBuffer::get_size() const throw()
{
    return m_size;
}


void IndexBuffer::draw_elements(
    GLenum mode, size_t count, GLenum type, size_t offset) const throw()
{
    if( m_use_vbos )
    {
        if( !is_bound() )
            bind();
        
        if( is_mapped() )
            unmap();

        glDrawElements(mode, count, type, (void *)offset);
    }
    else
    {
        clear_vbo_binding();
        glDrawElements(mode, count, type, (void *)(m_data+offset));
    }
    CHECK_GL_ERRORS();
}


void IndexBuffer::draw_range_elements(
    GLenum mode, size_t start, size_t end,
    size_t count, GLenum type, size_t offset) const throw()
{
    // TODO
    assert(false);
}





void IndexBuffer::bind() const throw()
{
    if( m_use_vbos )
    {
        ms_bound_vbo = m_vbo_name;

        glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m_vbo_name);
        CHECK_GL_ERRORS();
    }
}


void IndexBuffer::unbind() const throw()
{
    if( m_use_vbos )
    {
        ms_bound_vbo = 0;
        glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
        CHECK_GL_ERRORS();
    }
}


void IndexBuffer::clear_vbo_binding() const throw()
{
    if( !Device::has_vbo_support() )
        return;

    if( ms_bound_vbo != 0 )
    {
        ms_bound_vbo = 0;
        glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
        CHECK_GL_ERRORS();
    }
}
