/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef __PEEKABOT_RENDERER_INDEX_BUFFER_HH
#define __PEEKABOT_RENDERER_INDEX_BUFFER_HH


#include <GL/glew.h>
#include <stdexcept>
#include "../Types.hh"


namespace peekabot
{
namespace renderer
{

    // * Only one buffer can be bound at the time
    //
    // * Mapped buffers may become corrupt due to external events such as screen resolution changes
    //
    // * A buffer has to be bound to be mapped
    //
    // * TODO: error handling - how? exceptions or return-value based?
    //
    // The pimpl idiom causes significant function call overhead, for sure... could be 
    // removed later if it inhibits the performance (it shouldn't, the calls themselves
    // should be fairly expensive and "rare"). The interface advantages gained by using
    // the pimpl idiom (built-in factory pattern in the VB-class), and the reduce compile
    // times should be enough to counter the disadvantages..
    

    // Can be used both as an index index buffer and index array (data) buffer
    class IndexBuffer
    {
    public:
        enum UsageHint
        {
            STATIC_DRAW  = GL_STATIC_DRAW_ARB,
            DYNAMIC_DRAW = GL_DYNAMIC_DRAW_ARB,
            STREAM_DRAW  = GL_STREAM_DRAW_ARB
        };

        enum Access
        {
            READ_ONLY  = GL_READ_ONLY_ARB,
            WRITE_ONLY = GL_WRITE_ONLY_ARB,
            READ_WRITE = GL_READ_WRITE_ARB
        };

        
        IndexBuffer(bool force_no_vbos = false) throw();

        IndexBuffer(const IndexBuffer &buf) throw();

        ~IndexBuffer() throw();

        

        // allocate a buffer and set its data
        void set_data(
            const void *data, size_t n_bytes, UsageHint usage) 
            throw(std::runtime_error);
        
        
        // overwrite a subset of the data
        void set_data(
            const void *data, size_t n_bytes, size_t offset)
            throw(std::runtime_error);

        void get_data(void *data, size_t n_bytes, size_t offset)
            const throw(std::runtime_error);
        

        // map the buffer, calls to set_data and set_target are invalid
        // until the buffer is unmapped
        inline void *map(Access access, bool discard = false) throw()
        {
            return m_use_vbos ? map_vbo(access, discard) : m_data;
        }

        // map the buffer, calls to set_data and set_target are invalid
        // until the buffer is unmapped
        //
        // READ-ONLY!
        inline const void *map(bool discard = false) const throw()
        {
            return m_use_vbos ? map_vbo(discard) : m_data;
        }       

        // unmap a mapped buffer
        //
        // \return true if the buffer was corrupted, false otherwise.
        inline bool unmap() const throw()
        {
            return m_use_vbos ? unmap_vbo() : false;
        }
        
        inline bool is_mapped() const throw() { return m_is_mapped; }
        
        
        // get the buffer size
        size_t get_size() const throw();


        void draw_elements(
            GLenum mode, size_t count, GLenum type, size_t offset) const throw();

        void draw_range_elements(
            GLenum mode, size_t start, size_t end,
            size_t count, GLenum type, size_t offset) const throw();


        void bind() const throw();

        void unbind() const throw();

        inline bool is_bound() const throw() { return ms_bound_vbo == m_vbo_name; }
        

    private:
        void clear_vbo_binding() const throw();

        // map the buffer, calls to set_data and set_target are invalid
        // until the buffer is unmapped
        void *map_vbo(Access access, bool discard = false) throw();

        // map the buffer, calls to set_data and set_target are invalid
        // until the buffer is unmapped
        //
        // READ-ONLY!
        const void *map_vbo(bool discard = false) const throw();
        

        // unmap a mapped buffer
        //
        // \return true if the buffer was corrupted, false otherwise.
        bool unmap_vbo() const throw();

    private:
        typedef GLuint VBODescriptor;

        VBODescriptor m_vbo_name;

        size_t m_size;

        uint8_t *m_data;

        mutable bool m_is_mapped;

        UsageHint m_usage;

        bool m_use_vbos;

        static VBODescriptor ms_bound_vbo;
    };

}
}


#endif // __PEEKABOT_RENDERER_INDEX_BUFFER_HH
