/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <GL/glew.h>
#include <boost/scoped_array.hpp>
#include "Mesh.hh"
#include "MeshCullable.hh"
#include "primitives/Camera.hh"
#include "SceneNode.hh"
#include "../GeometryToolbox.hh"


using namespace peekabot;
using namespace peekabot::renderer;
using peekabot::renderer::primitives::Camera;


MeshCullable::MeshCullable(boost::shared_ptr<Mesh> mesh) throw()
    : m_mesh(mesh)
{
}

MeshCullable::~MeshCullable()
{
}

void MeshCullable::get_renderables(
    Camera *camera, EnumerationCallback callback) throw()
{
    callback(m_mesh.get());
}

void MeshCullable::transform_geometry(
    const boost::function<Vector3r (const Vector4r &)> &fun) throw()
{
    assert( false );
}

void MeshCullable::set_vertices(const VertexBased::Vertices &vertices) throw()
{
    boost::shared_ptr<VertexBuffer> buf(new VertexBuffer);

    if( !vertices.empty() )
    {
        boost::scoped_array<float> data(new float[3*vertices.size()]);

        for( size_t i = 0; i < vertices.size(); ++i )
        {
            data[3*i+0] = vertices[i](0);
            data[3*i+1] = vertices[i](1);
            data[3*i+2] = vertices[i](2);
        }

        buf->set_data(
            data.get(),
            3*sizeof(float)*vertices.size(), 
            VertexBuffer::STATIC_DRAW);
    }

    m_mesh->set_vertices(buf);

    if( !m_mesh->has_indices() )
        m_mesh->set_element_count(vertices.size()/m_mesh->vertices_per_element());

    calculate_bvs(vertices);
}

void MeshCullable::add_vertices(const VertexBased::Vertices &vertices) throw()
{
    assert( false ); // not yet implemented!

    /*if( get_vertices.get_size() == 0 )
        set_vertices(vertices);
    else
    {
        boost::scoped_array<uint8_t> data(new uint8_t[get_vertices.get_size()]);
        buf.get_data(&data[0], m_size, 0);
        set_data(data, m_size, m_usage);
        delete[] data;
    }*/
}

void MeshCullable::get_vertices(VertexBased::Vertices &vertices) const throw()
{
    vertices.clear();

    if( !m_mesh->has_vertices() )
        return;

    const boost::shared_ptr<VertexBuffer> verts = m_mesh->get_vertices();
    const float *ptr = (const float *)verts->map();

    size_t N = m_mesh->get_element_count()*m_mesh->vertices_per_element();
    for( size_t i = 0; i < N; ++i )
        vertices.push_back( Vector3f(ptr[3*i+0], ptr[3*i+1], ptr[3*i+2]) );       

    verts->unmap();
}


void MeshCullable::calculate_bvs() throw()
{
    if( get_parent() == 0 )
        return;

    VertexBased::Vertices vertices;
    get_vertices(vertices);
    calculate_bvs(vertices);
}

void MeshCullable::calculate_bvs(const VertexBased::Vertices &vertices) throw()
{
    if( get_parent() == 0 )
        return;

    // TODO: doesn't neccesarily work for indexed primitives!

    //
    // Calculate the midpoint of the vertices
    //
    Vector3f midpoint;
    float r;

    geom::calculate_approximate_bounding_sphere(
        vertices.begin(), vertices.end(),
        midpoint, r,
        get_parent()->get_x_scale(),
        get_parent()->get_y_scale(),
        get_parent()->get_z_scale());

    // Transform the midpoint in to an world-coordinate offset
    // and calculate the BV position
    midpoint = get_parent()->get_mtow() * Vector4f(midpoint, 0);
    
    set_bounding_sphere(
        BoundingSphere(
            get_parent()->get_world_pos() + Vector4f(midpoint, 0), r));
}

boost::shared_ptr<Mesh> MeshCullable::get_mesh() throw()
{
    return m_mesh;
}

const boost::shared_ptr<Mesh> 
MeshCullable::get_mesh() const throw()
{
    return m_mesh;
}


//


PointMeshCullable::PointMeshCullable(boost::shared_ptr<Mesh> mesh) throw()
    : MeshCullable(mesh)
{
}

PointMeshCullable::PointMeshCullable(const PointMeshCullable &x) throw()
    : MeshCullable(boost::shared_ptr<Mesh>(x.get_mesh()->clone()))
{
}

PointMeshCullable::~PointMeshCullable()
{
}

PointMeshCullable *PointMeshCullable::clone() const throw()
{
    return new PointMeshCullable(boost::shared_ptr<Mesh>(get_mesh()->clone()));
}

//

LineMeshCullable::LineMeshCullable(boost::shared_ptr<Mesh> mesh) throw()
    : MeshCullable(mesh)
{
}

LineMeshCullable::LineMeshCullable(const LineMeshCullable &x) throw()
    : MeshCullable(boost::shared_ptr<Mesh>(x.get_mesh()->clone()))
{
}

LineMeshCullable::~LineMeshCullable()
{
}

LineMeshCullable *LineMeshCullable::clone() const throw()
{
    return new LineMeshCullable(boost::shared_ptr<Mesh>(get_mesh()->clone()));
}


void LineMeshCullable::set_line_style(
    float line_width, 
    uint16_t stipple_pattern, 
    int stipple_factor)
{
    m_mesh->get_state().set(
        new statelets::LineParams(
            line_width, stipple_pattern != 0xFFFF,
            stipple_factor, stipple_pattern));
}


//

TriMeshCullable::TriMeshCullable(boost::shared_ptr<Mesh> mesh) throw()
    : MeshCullable(mesh)
{
}

TriMeshCullable::TriMeshCullable(const TriMeshCullable &x) throw()
    : MeshCullable(boost::shared_ptr<Mesh>(x.get_mesh()->clone()))
{
}

TriMeshCullable::~TriMeshCullable()
{
}

TriMeshCullable *TriMeshCullable::clone() const throw()
{
    return new TriMeshCullable(boost::shared_ptr<Mesh>(get_mesh()->clone()));
}

void TriMeshCullable::set_indices(const Indices &indices)
{
    boost::shared_ptr<IndexBuffer> buf(new IndexBuffer);

    buf->set_data(
        &indices[0], 
        sizeof(uint32_t)*indices.size(), 
        IndexBuffer::STATIC_DRAW);

    m_mesh->set_indices(buf);
    m_mesh->set_element_count(indices.size()/m_mesh->vertices_per_element());

    calculate_bvs();
}

void TriMeshCullable::add_indices(const Indices &indices)
{
    assert( false );
}

void TriMeshCullable::get_indices(Indices &indices) const
{
    indices.clear();

    if( !m_mesh->has_indices() )
        return;

    const boost::shared_ptr<IndexBuffer> inds = m_mesh->get_indices();
    const uint32_t *ptr = (const uint32_t *)inds->map();

    size_t N = m_mesh->get_element_count();
    for( size_t i = 0; i < N; ++i )
        indices.push_back( ptr[i] );

    inds->unmap();
}
