/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <cmath>

#include "MeshFactory.hh"
#include "Mesh.hh"

using namespace peekabot;
using namespace peekabot::renderer;


Mesh *MeshFactory::get_disc_mesh(int segments, float r) throw()
{
    // Vertices...
    Vector3r *x = new Vector3r[2*(segments+1)];

    x[0] = x[segments+1] = Vector3r(0,0,0);
    for( int i = 0; i < segments; ++i )
    {
        float alpha = i*(3.14159*2/segments);
        x[i+1] = Vector3r(r*cosf(alpha), r*sinf(alpha), 0);
        x[(segments+1)+i+1] = Vector3r(r*cosf(alpha), r*sinf(alpha), 0);
    }

    boost::shared_ptr<VertexBuffer> vertices(new VertexBuffer);
    vertices->set_data(x, sizeof(Vector3r)*2*(segments+1), 
                       VertexBuffer::STATIC_DRAW);

    // Normals...
    for( int i = 0; i <= segments; ++i )
    {
        x[i] = Vector3r(0, 0, 1);
        x[i+(segments+1)] = Vector3r(0, 0, -1);
    }

    boost::shared_ptr<VertexBuffer> normals(new VertexBuffer);
    normals->set_data(x, sizeof(Vector3r)*2*(segments+1), 
                      VertexBuffer::STATIC_DRAW);

    delete[] x;

    // Indices...
    uint32_t *inds = new uint32_t[2*3*segments];

    inds[0] = 0;
    inds[1] = segments;
    inds[2] = 1;

    inds[3*segments+0] = (segments+1)+0;
    inds[3*segments+1] = (segments+1)+1;
    inds[3*segments+2] = (segments+1)+segments;

    for( int i = 1; i < segments; ++i )
    {
        inds[3*i+0] = 0;
        inds[3*i+1] = i;
        inds[3*i+2] = i+1;

        inds[3*(segments+i)+0] = (segments+1);
        inds[3*(segments+i)+1] = (segments+1)+i+1;
        inds[3*(segments+i)+2] = (segments+1)+i;
    }

    boost::shared_ptr<IndexBuffer> indices(new IndexBuffer);
    indices->set_data(inds, sizeof(Vector3r)*2*segments, 
                      IndexBuffer::STATIC_DRAW);

    delete[] inds;


    return new TriMesh(
        vertices, normals, 
        boost::shared_ptr<VertexBuffer>(), 
        boost::shared_ptr<VertexBuffer>(), 
        indices, 0, 2*segments);
}


Mesh *MeshFactory::get_cylinder_mesh(int segments, float r, float h) throw()
{
    //
    // If you have found a bug in this code, and I am quite confident that 
    // you will, you might as well re-write it from scratch at once, because 
    // it's so hairy and ugly you will most likely die if you try to wrap your
    // head around it.
    //

    // Vertices...
    Vector3r *x = new Vector3r[2*(segments+1) + 2*segments];

    x[0] = Vector3r(0,0,h/2);
    x[segments+1] = Vector3r(0,0,-h/2);
    for( int i = 0; i < segments; ++i )
    {
        // Top/bottom:
        float alpha = i*(3.14159*2/segments);
        x[i+1] = Vector3r(r*cosf(alpha), r*sinf(alpha), h/2);
        x[(segments+1)+i+1] = Vector3r(r*cosf(alpha), r*sinf(alpha), -h/2);

        // Sides:
        x[2*(segments+1)+2*i+0] = Vector3r(r*cosf(alpha), r*sinf(alpha), h/2);
        x[2*(segments+1)+2*i+1] = Vector3r(r*cosf(alpha), r*sinf(alpha), -h/2);
    }

    boost::shared_ptr<VertexBuffer> vertices(new VertexBuffer);
    vertices->set_data(x, sizeof(Vector3r)*(2*(segments+1) + 2*segments), 
                       VertexBuffer::STATIC_DRAW);
    


    // Normals...
    //
    // Top/bottom:
    for( int i = 0; i <= segments; ++i )
    {
        x[i] = Vector3r(0, 0, 1);
        x[i+(segments+1)] = Vector3r(0, 0, -1);
    }
    // Sides:
    for( int i = 0; i < segments; ++i )
    {
        float alpha = i*(3.14159*2/segments);
        x[2*(segments+1)+2*i+0] = Vector3r(cosf(alpha), sinf(alpha), 0);
        x[2*(segments+1)+2*i+1] = Vector3r(cosf(alpha), sinf(alpha), 0);
    }


    boost::shared_ptr<VertexBuffer> normals(new VertexBuffer);
    normals->set_data(x, sizeof(Vector3r)*(2*(segments+1) + 2*segments), 
                      VertexBuffer::STATIC_DRAW);

    delete[] x;

    // Indices...
    uint32_t *inds = new uint32_t[3*(2*segments+2*segments)];

    // Top/bottom:
    inds[0] = 0;
    inds[1] = segments;
    inds[2] = 1;

    inds[3*segments+0] = (segments+1)+0;
    inds[3*segments+1] = (segments+1)+1;
    inds[3*segments+2] = (segments+1)+segments;

    for( int i = 1; i < segments; ++i )
    {
        inds[3*i+0] = 0;
        inds[3*i+1] = i;
        inds[3*i+2] = i+1;

        inds[3*(segments+i)+0] = (segments+1);
        inds[3*(segments+i)+1] = (segments+1)+i+1;
        inds[3*(segments+i)+2] = (segments+1)+i;
    }

    // Sides:
    for( int i = 0; i < segments-1; ++i )
    {
        inds[3*(2*segments+2*i)+0] = 2*(segments+1)+2*i+0;
        inds[3*(2*segments+2*i)+1] = 2*(segments+1)+2*i+1;
        inds[3*(2*segments+2*i)+2] = 2*(segments+1)+2*(i+1)+1;

        inds[3*(2*segments+2*i+1)+0] = 2*(segments+1)+2*i+0;
        inds[3*(2*segments+2*i+1)+1] = 2*(segments+1)+2*(i+1)+1;
        inds[3*(2*segments+2*i+1)+2] = 2*(segments+1)+2*(i+1)+0;
    }

    inds[3*(2*segments+2*(segments-1))+0] = 2*(segments+1)+2*(segments-1)+0;
    inds[3*(2*segments+2*(segments-1))+1] = 2*(segments+1)+2*(segments-1)+1;
    inds[3*(2*segments+2*(segments-1))+2] = 2*(segments+1)+2*0+1;
    
    inds[3*(2*segments+2*(segments-1)+1)+0] = 2*(segments+1)+2*(segments-1)+0;
    inds[3*(2*segments+2*(segments-1)+1)+1] = 2*(segments+1)+2*0+1;
    inds[3*(2*segments+2*(segments-1)+1)+2] = 2*(segments+1)+2*0+0;



    boost::shared_ptr<IndexBuffer> indices(new IndexBuffer);
    indices->set_data(inds, sizeof(uint32_t)*(3*(2*segments+2*segments)), 
                      IndexBuffer::STATIC_DRAW);

    delete[] inds;


    return new TriMesh(
        vertices, normals, 
        boost::shared_ptr<VertexBuffer>(), 
        boost::shared_ptr<VertexBuffer>(), 
        indices, 0, 2*segments+2*segments);
}


Mesh *MeshFactory::get_icosphere_mesh(int subdiv, float r) throw()
{
    const double X = 0.525731112119133606;
    const double Z = 0.850650808352039932;

    static Vector3r vert_data[12] =
    {
        Vector3r(-X, -Z, 0.0), Vector3r(X, -Z, 0.0), 
        Vector3r(-X, Z, 0.0), Vector3r(X, Z, 0.0),
        Vector3r(0.0, -X, Z), Vector3r(0.0, X, Z), 
        Vector3r(0.0, -X, -Z), Vector3r(0.0, X, -Z),
        Vector3r(Z, 0.0, X), Vector3r(-Z, 0.0, X), 
        Vector3r(Z, 0.0, -X), Vector3r(-Z, 0.0, -X)
    };

    static uint32_t ind_data[20][3] =
    {
        {0, 1, 4}, {0, 4, 9}, {9, 4, 5}, {4, 8, 5}, {4, 1, 8},
        {8, 1, 10}, {8, 10, 3}, {5, 8, 3}, {5, 3, 2}, {2, 3, 7},
        {7, 3, 10}, {7, 10, 6}, {7, 6, 11}, {11, 6, 0}, {0, 6, 1},
        {6, 10, 1}, {9, 11, 0}, {9, 2, 11}, {9, 5, 2}, {7, 11, 2}
    };

    std::vector<Vector3r> verts;
    std::vector<uint32_t> inds;

    for( int i = 0; i < 12; ++i )
        verts.push_back(vert_data[i]);


    for( int i = 0; i < 20; ++i )
    {
        subdivide(
            ind_data[i][0], ind_data[i][1], ind_data[i][2], 
            verts, inds, subdiv);
    }

    int n_tris = inds.size()/3;
    int n_verts = verts.size();

    /*boost::shared_ptr<VertexBuffer> normals(new VertexBuffer);
    normals->set_data(&verts[0], sizeof(Vector3r)*n_verts, 
    VertexBuffer::STATIC_DRAW);*/

    for( size_t i = 0; i < verts.size(); ++i )
        verts[i] *= r;

    boost::shared_ptr<VertexBuffer> vertices(new VertexBuffer);
    vertices->set_data(&verts[0], sizeof(Vector3r)*n_verts, 
                       VertexBuffer::STATIC_DRAW);

    boost::shared_ptr<IndexBuffer> indices(new IndexBuffer);
    indices->set_data(&inds[0], 3*sizeof(uint32_t)*n_tris, 
                      IndexBuffer::STATIC_DRAW);   


    return new TriMesh(
        vertices, vertices, //normals, 
        boost::shared_ptr<VertexBuffer>(), 
        boost::shared_ptr<VertexBuffer>(), 
        indices, 0, n_tris);
}


Mesh *MeshFactory::get_circle_mesh(int segments, float r) throw()
{
    std::vector<Vector3r> verts;
    std::vector<uint32_t> inds;


    for( int i = 0; i < segments; ++i )
    {
        verts.push_back(
            Vector3f(r*cosf(2*M_PI*i/segments), 
                     r*sinf(2*M_PI*i/segments), 0));
    }

    for( int i = 0; i < segments; ++i )
    {
        inds.push_back( (i == 0) ? (segments-1):(i-1) );
        inds.push_back(i);
    }

    boost::shared_ptr<VertexBuffer> vertices(new VertexBuffer);
    vertices->set_data(&verts[0], sizeof(Vector3r)*verts.size(), 
                       VertexBuffer::STATIC_DRAW);

    boost::shared_ptr<IndexBuffer> indices(new IndexBuffer);
    indices->set_data(&inds[0], sizeof(uint32_t)*inds.size(), 
                      IndexBuffer::STATIC_DRAW);   


    return new LineMesh(
        vertices, 
        boost::shared_ptr<VertexBuffer>(), 
        boost::shared_ptr<VertexBuffer>(), 
        indices, 0, segments);
}


void MeshFactory::subdivide(
    uint32_t v0_idx, uint32_t v1_idx, uint32_t v2_idx, 
    std::vector<Vector3r> &vertices,
    std::vector<uint32_t> &indices,
    int depth) throw()
{
    if( depth == 0 )
    {
        indices.push_back(v0_idx);
        indices.push_back(v1_idx);
        indices.push_back(v2_idx);
    }
    else
    {
        const Vector3r &v0 = vertices[v0_idx];
        const Vector3r &v1 = vertices[v1_idx];
        const Vector3r &v2 = vertices[v2_idx];
        
        Vector3r v01((v0 + v1) / 2);
        Vector3r v12((v1 + v2) / 2);
        Vector3r v20((v2 + v0) / 2);

        v01.normalize();
        v12.normalize();
        v20.normalize();

        int n = vertices.size();

        vertices.push_back(v01);
        vertices.push_back(v12);
        vertices.push_back(v20);

        subdivide(v0_idx, n+0, n+2, vertices, indices, depth - 1);
        subdivide(v1_idx, n+1, n+0, vertices, indices, depth - 1);
        subdivide(v2_idx, n+2, n+1, vertices, indices, depth - 1);
        subdivide(n+0, n+1, n+2, vertices, indices, depth - 1);
    }
}
