/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "MeshTransformer.hh"
#include "../Vector.hh"


using namespace peekabot;
using namespace renderer;


MeshTransformer::GeometryTransformer::GeometryTransformer(
    const TransformFun &transform, 
    BoundingSphere *bsphere,
    bool copy_on_write) throw()
    : m_transform(transform),
      m_bsphere(bsphere),
      m_copy_on_write(copy_on_write)
{
}


void MeshTransformer::GeometryTransformer::transform_vertices(Mesh &mesh) throw()
{
    if( mesh.has_vertices() )
    {
        // Make a copy?
        if( m_copy_on_write && !mesh.get_vertices().unique() )
        {
            mesh.set_vertices(
                boost::shared_ptr<VertexBuffer>(
                    new VertexBuffer(*mesh.get_vertices())));
        }

        // Transform vertices
        transform_geometry(
            mesh.get_indices(), mesh.get_vertices(), 
            mesh.get_element_count() * mesh.vertices_per_element(), 
            mesh.get_index_offset(), false);
    }
}


void MeshTransformer::GeometryTransformer::visit(TriMesh &mesh) throw()
{
    transform_vertices(mesh);

    if( mesh.has_normals() )
    {
        // Make a copy?
        if( m_copy_on_write && !mesh.get_normals().unique() )
        {
            mesh.set_normals(
                boost::shared_ptr<VertexBuffer>(
                    new VertexBuffer(*mesh.get_normals())));
        }

        // Transform normals
        transform_geometry(
            mesh.get_indices(), mesh.get_normals(), 
            mesh.get_element_count() * mesh.vertices_per_element(), 
            mesh.get_index_offset(), true);
    }
}


void MeshTransformer::GeometryTransformer::visit(LineMesh &mesh) throw()
{
    transform_vertices(mesh);
}


void MeshTransformer::GeometryTransformer::visit(PointMesh &mesh) throw()
{
    transform_vertices(mesh);
}


void MeshTransformer::GeometryTransformer::transform_geometry(
    boost::shared_ptr<IndexBuffer> index_buffer,
    boost::shared_ptr<VertexBuffer> vertex_buffer,
    uint32_t n, uint32_t index_offset, bool normals) throw()
{
    if( index_buffer )
    {
        transform_indexed_geometry(
            index_buffer, vertex_buffer,
            n, index_offset, normals);
    }
    else
    {
        transform_unindexed_geometry(
            vertex_buffer,
            n, normals);
    }
}


void MeshTransformer::GeometryTransformer::transform_indexed_geometry(
    boost::shared_ptr<IndexBuffer> index_buffer,
    boost::shared_ptr<VertexBuffer> vertex_buffer,
    uint32_t n, uint32_t index_offset,
    bool normals) throw()
{
    Vector3r midp(0,0,0);

    assert( vertex_buffer != 0 );
    assert( index_buffer != 0 );

    std::set<uint32_t> transformed;
                                       
    real32 *data = reinterpret_cast<real32 *>(
        vertex_buffer->map(VertexBuffer::READ_WRITE, false));

    uint32_t *indices = reinterpret_cast<uint32_t *>(
        index_buffer->map(IndexBuffer::READ_ONLY, false));

    assert( data != 0 );
    assert( indices != 0 );

    for( uint32_t i = index_offset; i < n+index_offset; ++i )
    {
        // Get indices to vertex...
        uint32_t idx = indices[i];

        if( transformed.insert(idx).second )
        {
            Vector3r r;

            if( normals )
            {
                Vector4r v(data[3*idx+0], data[3*idx+1], data[3*idx+2], 0);
                r = m_transform(v).normalize();
            }
            else
            {
                Vector4r v(data[3*idx+0], data[3*idx+1], data[3*idx+2], 1);
                r = m_transform(v);

                midp += r;
            }
                            
            data[3*idx+0] = r(0);
            data[3*idx+1] = r(1);
            data[3*idx+2] = r(2);
        }
    }


    if( !normals && m_bsphere != 0 )
    {
        //
        // Calculate bounding sphere:
        //
        midp /= transformed.size();

        float sq_r = 0;

        for( uint32_t i = index_offset; i < n+index_offset; ++i )
        {
            uint32_t idx = indices[i];            
            Vector3r v(data[3*idx+0], data[3*idx+1], data[3*idx+2]);

            float sq_d = (midp-v).norm_squared();
            if( sq_r < sq_d )
                sq_r = sq_d;
        }

        m_bsphere->set_radius(sqrtf(sq_r));
        m_bsphere->set_pos(midp);
    }


    vertex_buffer->unmap();
    index_buffer->unmap();
}


void MeshTransformer::GeometryTransformer::transform_unindexed_geometry(
    boost::shared_ptr<VertexBuffer> vertex_buffer,
    uint32_t n, bool normals) throw()
{
    Vector3r midp(0,0,0);

    assert( vertex_buffer != 0 );

    real32 *data = reinterpret_cast<real32 *>(
        vertex_buffer->map(VertexBuffer::READ_WRITE, false));

    assert( data != 0 );

    for( uint32_t i = 0; i < n; ++i )
    {
        Vector3r r;
        
        if( normals )
        {
            Vector4r v(data[3*i+0], data[3*i+1], data[3*i+2], 0);
            r = m_transform(v).normalize();
        }
        else
        {
            Vector4r v(data[3*i+0], data[3*i+1], data[3*i+2], 1);
            r = m_transform(v);
            
            midp += r;
        }
                            
        data[3*i+0] = r(0);
        data[3*i+1] = r(1);
        data[3*i+2] = r(2);
    }


    if( !normals && m_bsphere != 0 )
    {
        //
        // Calculate bounding sphere:
        //
        midp /= n;

        float sq_r = 0;

        for( uint32_t i = 0; i < n; ++i )
        {
            Vector3r v(data[3*i+0], data[3*i+1], data[3*i+2]);

            float sq_d = (midp-v).norm_squared();
            if( sq_r < sq_d )
                sq_r = sq_d;
        }

        m_bsphere->set_radius(sqrtf(sq_r));
        m_bsphere->set_pos(midp);
    }


    vertex_buffer->unmap();
}


void MeshTransformer::transform_geometry(
    Mesh &mesh, const TransformFun &transform, 
    BoundingSphere *bsphere, bool copy_on_write) throw()
{
    GeometryTransformer transformer(transform, bsphere, copy_on_write);
    mesh.accept(transformer);
}

