/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_RENDERER_MESH_TRANSFORMER_HH
#define __PEEKABOT_RENDERER_MESH_TRANSFORMER_HH


#include <boost/function.hpp>

#include "../Matrix.hh"
#include "Mesh.hh"
#include "MeshVisitor.hh"
#include "IndexBuffer.hh"
#include "VertexBuffer.hh"
#include "BoundingVolumes.hh"


namespace peekabot
{
    namespace renderer
    {

        /**
         * \brief An utility that takes a Mesh and transforms its data 
         * (currently geometry only) by the specified matrix.
         */
        class MeshTransformer
        {
            typedef boost::function<Vector3r (const Vector4r &)> TransformFun;

            class GeometryTransformer : public MeshVisitor
            {
                /**
                 * \brief The transformation function by which the mesh data is 
                 * transformed.
                 */
                const TransformFun &m_transform;

                BoundingSphere *m_bsphere;

                bool m_copy_on_write;

            public:
                GeometryTransformer(
                    const TransformFun &transform, 
                    BoundingSphere *bsphere, 
                    bool copy_on_write) throw();

                virtual void visit(TriMesh &mesh) throw();

                virtual void visit(LineMesh &mesh) throw();

                virtual void visit(PointMesh &mesh) throw();

            protected:
                /**
                 * \brief Transforms the geometry in the given vertex buffer 
                 * with \a m_transform.
                 *
                 * If \a normals is true, the geometry is transformed as vectors 
                 * rather than points points and the result is normalized.
                 *
                 * \param index_buffer The index buffer containing in the vertex
                 * indices to be transformed.
                 * \param vertex_buffer The vertex buffer containing the vertex
                 * data.
                 * \param n The number of vertices/normals to be transformed.
                 * \param index_offset The offset into \a index_buffer at which 
                 * the vertex indices to be transformed starts.
                 * \param normals Set to \c false and \c true respectively if 
                 * it's vertices and normals that are to be transformed.
                 */
                void transform_geometry(
                    boost::shared_ptr<IndexBuffer> index_buffer,
                    boost::shared_ptr<VertexBuffer> vertex_buffer,
                    uint32_t n, uint32_t index_offset, bool normals) throw();

                /**
                 * \brief Transforms the geometry in the given vertex buffer 
                 * with \a m_transform.
                 *
                 * If \a normals is true, the geometry is transformed as vectors 
                 * rather than points points and the result is normalized.
                 *
                 * \param index_buffer The index buffer containing in the vertex
                 * indices to be transformed.
                 * \param vertex_buffer The vertex buffer containing the vertex
                 * data.
                 * \param n The number of vertices/normals to be transformed.
                 * \param index_offset The offset into \a index_buffer at which 
                 * the vertex indices to be transformed starts.
                 * \param normals Set to \c false and \c true respectively if 
                 * it's vertices and normals that are to be transformed.
                 */
                void transform_indexed_geometry(
                    boost::shared_ptr<IndexBuffer> index_buffer,
                    boost::shared_ptr<VertexBuffer> vertex_buffer,
                    uint32_t n, uint32_t index_offset, bool normals) throw();

                /**
                 * \brief Transforms the geometry in the given vertex buffer 
                 * with \a m_transform.
                 *
                 * If \a normals is true, the geometry is transformed as vectors 
                 * rather than points points and the result is normalized.
                 *
                 * \param vertex_buffer The vertex buffer containing the vertex
                 * data.
                 * \param n The number of vertices/normals to be transformed.
                 * \param normals Set to \c false and \c true respectively if 
                 * it's vertices and normals that are to be transformed.
                 */
                void transform_unindexed_geometry(
                    boost::shared_ptr<VertexBuffer> vertex_buffer,
                    uint32_t n, bool normals) throw();

                void transform_vertices(Mesh &mesh) throw();
            };

        public:
            /**
             * \brief Transform the geometry (vertices and normals) of the given
             * mesh using the specified transformation matrix.
             *
             * \param mesh The mesh to transform.
             * \param transform The transformation to apply.
             * \param bsphere If non-null, a bounding sphere will be calculated 
             * for the transformed mesh and stored in the variable pointed to 
             * by \a bsphere.
             * \param copy_on_write If enabled, shared vertex buffers will
             * be copied before being manipulated.
             */
            static void transform_geometry(
                Mesh &mesh, const TransformFun &transform, 
                BoundingSphere *bsphere,
                bool copy_on_write = true) throw();
        };

    }
}


#endif // __PEEKABOT_RENDERER_MESH_TRANSFORMER_HH
