/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <vector>
#include <cmath>
#include <cassert>
#include "Octree.hh"
#include "Cullable.hh"
#include "SceneNode.hh"
#include "primitives/Camera.hh"

using namespace peekabot;
using namespace peekabot::renderer;


Octree::Octree(real32 initial_size, real32 min_node_size)
    : m_min_node_size(min_node_size)
{
    m_root = new OctreeNode(0, initial_size, Vector3r(0,0,0), m_min_node_size);
}


Octree::~Octree() throw()
{
    delete m_root; // will cause recursive destruction of all nodes...
}


bool Octree::queue_need_update(Cullable *obj) throw()
{
    return !m_update_set.insert(obj).second;
}


void Octree::process_updates() throw()
{
    // OPT opt. possibilites: implement a "change type" hint 
    // (BV_GREW, BV_SHRANK, MOVED).
    // would reduce the amounts of tests needed..

    while( !m_update_set.empty() )
    {
        Cullable *obj = *m_update_set.begin();
        m_update_set.erase(m_update_set.begin());

        if( !contains(obj) )
        {
            //object has been removed from the octree, skip it
            continue;
        }

        OctreeNode *prev_node = m_objects[obj];

        if( prev_node->encloses(obj) )
        {
            if( obj->get_bounding_sphere().get_radius() < prev_node->m_half_size/2 )
            {
                // it is contained in a child node of prev_node

                // don't update the enclosed count for ancestors,
                // that way, we get away with adding it recursively for the
                // previous node, updating the node count only for its subtree.
                prev_node->remove_object(obj, false);
                OctreeNode *new_node = prev_node->add_object(obj); // will add obj in the right (descendent) node
                
                // Update scene node -> octree node map
                m_objects[obj] = new_node;
            }
            else
            {
                // ok, nothing needs to change - the node remains in the same node.
            }
        }
        else
        {
            // object no longer contained in the same node...
            prev_node->remove_object(obj);

            // Record empty nodes
            if( prev_node->is_empty() )
                m_empty_nodes.insert(prev_node);

            // find ancestor of prev_node that encloses obj
            OctreeNode *node = prev_node->m_parent;
            while( node != 0 && !node->encloses(obj) )
            {
                // Record empty nodes
                if( node->is_empty() )
                    m_empty_nodes.insert(node);

                node = node->m_parent;
            }

                   
            // invariant: node encloses obj, and null if it doesn't
            // node = null => the obj is outside the octree (omg!)

            if( node != 0 )
            {
                // obj should be put in node or a descendent of node...
                OctreeNode *new_node = node->add_object(obj); // recursive add to put it in the right node

                // Fiddle with the enclosed count of ancestors to the new parent (node)
                // to get it right:
                OctreeNode *parents_parent = node->m_parent;
                while( parents_parent != 0 )
                {
                    ++parents_parent->m_enclosed;
                    parents_parent = parents_parent->m_parent;
                }

                // Update scene node -> octree node map
                m_objects[obj] = new_node;
            }
            else
            {
                grow_to_accomodate(obj);
                
                // Add object & update scene node -> octree node map
                m_objects[obj] = m_root->add_object(obj);
            }
        }
    }
}



Octree::OctreeNode *Octree::grow_to_accomodate(Cullable *obj) throw()
{
    // obj outside the octree!
    // deal with special case... by extending the tree outwards,
    // that is, by create a new big node and use the current
    // root node as a child octant in the new node.
    do
    {
        OctreeNode *old_root = m_root;
        m_root = new OctreeNode(0, 
                                old_root->m_half_size*2, 
                                Vector3r(0,0,0), 
                                old_root->m_min_node_size);

        Vector3r configuration[] =
        {
            Vector3r(-1, -1, -1),
            Vector3r( 1, -1,  1),
            Vector3r(-1,  1, -1),
            Vector3r( 1,  1,  1),
            Vector3r(-1, -1, -1),
            Vector3r( 1, -1,  1),
            Vector3r(-1,  1, -1),
            Vector3r( 1,  1,  1)
        };

        real32 min_dist = 0;
        int min_conf = -1;

        for( int i = 0; i < 8; i++ )
        {
            Vector3r confpos = old_root->m_pos + 
            (old_root->m_half_size/2)*configuration[i];
                        
            real32 d = confpos.dist(obj->get_bounding_sphere().get_pos());

            if( min_conf == -1 || d < min_dist )
            {
                min_conf = i;
                min_dist = d;
            }
        }

        // Now we just have to figure out in which directions
        // to extend the octree.
        // This can easily be done by find the configuration p
        // which minimizes the distance dist(obj,p). There
        // are 8 possible configurations.
        //
        // The different possible configurations are the 
        // permutations of <x,y,z>, x,y,z \in {-1,1}. I.e.
        // all the ways the new cube can be offset from the
        // old root node.
        //
        // Having found the best configuration, we implicitly
        // also have the index of the new root node to put the 
        // old one in.
        // index = (x/2+0.5)+2*(y/2+0.5)+4*(z/2+0.5)

        m_root->m_pos = old_root->m_pos + 
            old_root->m_half_size/2*configuration[min_conf];

        /*int index = static_cast<int>(
            1*(configuration[min_conf](0)/2+0.5) +
            2*(configuration[min_conf](1)/2+0.5) +
            4*(configuration[min_conf](2)/2+0.5) );*/

        int index = -1;
        for( int i = 0; i < 8; ++i )
        {
            if( (m_root->octant_offset(i) + m_root->m_pos).equals(
                    old_root->m_pos, m_root->m_half_size * 1e-6) )
            {
                index = i;
                break;
            }
        }

        assert( index != -1 );

        m_root->m_children[index] = old_root;
        old_root->m_parent = m_root;
        m_root->m_enclosed = old_root->m_enclosed;

    } while( !m_root->encloses(obj) );

    return m_root;
}



void Octree::add_object(Cullable *obj) throw()
{
    if( m_objects.find(obj) == m_objects.end() )
    {
        if( !m_root->encloses(obj) )
        {
            grow_to_accomodate(obj);
        }

        OctreeNode *node = m_root->add_object(obj);
        m_objects[obj] = node;

        // Remove re-populated nodes from the empty node set:
        std::set<OctreeNode *>::iterator it = m_empty_nodes.find(node);
        if( it != m_empty_nodes.end() )
            m_empty_nodes.erase(it);
    }
    // else: obj already in octree...
}


void Octree::remove_object(Cullable *obj) throw()
{
    /// \todo This code is sub-optimal in terms of performance, clean up scheduled!
    if( m_objects.find(obj) != m_objects.end() )
    {
        OctreeNode *contained_in = m_objects[obj];
        contained_in->remove_object(obj);
        m_objects.erase(obj);

        // Record an empty node to m_empty_nodes (for later purging):
        if( contained_in->is_empty() )
        {
            // subtree empty, keep track of it!
            m_empty_nodes.insert(contained_in);
        }

        // Remove it from the set of nodes in need of an update (if it's in it):
        m_update_set.erase(obj);
    }
    // else: not found, fail silently
}


bool Octree::contains(Cullable *obj) const throw()
{
    return m_objects.find(obj) != m_objects.end();
}


void Octree::traverse_pvs(primitives::Camera *camera, 
                          OctreeTraverser *traverser) throw()
{
    // Find the order in which to traverse nodes (to achieve rough 
    // front to back traversal)
    //
    // To find the right configuration, sort the sequence based on the y-distance
    // from each octant center to the camera.

    // Find the node the camera is or should be in (it might be hidden)
    OctreeNode *cam_node = m_root;//m_objects.find(camera)->second;

    bool stop = false;
    while( !stop )
    {
        for( int i = 0; i < 8; ++i )
        {
            if( cam_node->enclosed_in_octant(camera, i) )
            {
                if( cam_node->get_octant(i) )
                {
                    cam_node = cam_node->get_octant(i);
                    break;
                }
                else
                {
                    stop = true;
                    break;
                }
            }
            else if( i == 7 )
                stop = true;
        }
    }

    assert( cam_node != 0 );

    Vector3r view_dir = camera->get_view_vector();
    Vector3r view_pos = camera->get_parent()->get_world_pos();

    std::vector<std::pair<real32, int> > tmp(8);

    // Now, for each octant, find the y-distance to the viewing frustum:
    for( int i = 0; i < 8; i++ )
    {
        real32 z_dist = view_dir * (view_pos - cam_node->octant_offset(i));
        tmp[i] = std::make_pair(-z_dist, i);
    }

    std::sort(tmp.begin(), tmp.end());

    int order[8];

    for( int i = 0; i < 8; i++ )
    {
        order[i] = tmp[i].second;
    }
    
    m_root->traverse_pvs(camera, traverser, order, 0x3F);
}


void Octree::remove_subtree_from_empty_nodes(OctreeNode *subtree) throw()
{
    m_empty_nodes.erase(subtree);

    for( int i = 0; i < 8; ++i )
        if( subtree->m_children[i] )
            remove_subtree_from_empty_nodes(subtree->m_children[i]);
}


void Octree::purge() throw()
{
    while( !m_empty_nodes.empty() )
    {
        OctreeNode *node = *m_empty_nodes.begin();

        if( node->is_empty() && 
            node != m_root )
        {
            OctreeNode *parent = node->get_parent();

            // Remove all descendent nodes of node from the empty nodes set
            remove_subtree_from_empty_nodes(node);

            for( int i = 0; i < 8; ++i )
            {
                if( parent->m_children[i] == node )
                {
                    parent->m_children[i] = 0;
                    --parent->m_child_count;
                    break;
                }
            }

            delete node;
        }
        else
        {
            // Non-empty or root node, don't delete it!
            m_empty_nodes.erase(m_empty_nodes.begin());
        }
    }

    assert(m_empty_nodes.empty());
}



unsigned int Octree::get_excess_node_count() const throw()
{
    return m_empty_nodes.size();
}



unsigned int Octree::get_contained_count() const throw()
{
    return m_objects.size();
}
