/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef __PEEKABOT_RENDERER_OCTREE_HH
#define __PEEKABOT_RENDERER_OCTREE_HH


#include <set>
#include <map>
#include "../Vector.hh"


namespace peekabot
{
namespace renderer
{
    class Cullable;
    class OctreeTraverser;

    namespace primitives
    {
        class Camera;
    }

    /**
     * \brief A loose octree implementation.
     *
     * The octree divides the populated space, in its entirety, into an 
     * hierarchy of axis-aligned bounding boxes (cubes), residing in world space
     * coordinates. Each cube is represented by an \c OctreeNode.
     *
     * It is implemented to be fast, but not extravagantly gluttonous for memory. 
     * In practice, it means that the octree is \e not fully expanded. Rather,
     * it is subdivided and modified as needed, using only as much memory as
     * it actually needs.
     *
     * There is presently only one parameter that control subdivision depth. 
     * The \e minimum size of a node is set upon constructing the octree. 
     * It is then guaranteed that the tree  will never produce nodes smaller
     * than set by subdividing itself.
     *
     * \e Size and \e half-size always refers to the bounding cubes' side
     * length (or half side length), never the shortest distance to the cube
     * from it's center point or anything such.
     *
     * \remarks The octree is initially centered around the world origin, but is
     * dynamically extended as needed and best suited.
     * \remarks The <em>loose octree</em> was invented and dubbed by Thatcher 
     * Ulrich, refer to his website for tidbits of good information.
     */
    class Octree
    {   
    private:
        /**
         * \brief A node in the octree.
         *
         * Maintains a set of the objects that are in the node, and
         * references to any octants that further divide up the node 
         * (child nodes).
         *
         * Each node is \e loose in the sense that
         * they are \f$2N\f$ in size, but only objects of size \f$N\f$ in the 
         * smaller \f$N\f$-sized inner volume can be contained in a node. 
         * Thus ensuring that one object
         * is always in one and only one octree node. This avoids the need to
         * split objects or keep an object in multiple nodes.
         */
        struct OctreeNode
        {
            /**
             * \brief Half the size of the octants enclosed volume.
             */
            real32 m_half_size;

            /**
             * \brief The position of node's center, in world coordinates.
             */
            Vector3r m_pos;

            /**
             * \brief The octants the node is divided into.
             *
             * The child nodes of the node that divide itself volume
             * into 8 equally sized volumes. A node can have 0 to 8 child nodes.
             * If m_children[i] is null, the node's <em>i</em>:th octant isn't
             * present.
             * 
             * Octants are numbered according to the following scheme:
             * Let \f$<x,y,z>\f$ be the offset location from the parent's center,
             * \f$x,y,z \in \{-1,1\}\f$, then the index is computed as:
             * \f[ index = \frac{x+1}{2} + 2\frac{y+1}{2} + 4\frac{z+1}{2} \f]
             *
             * \invariant \c m_children[i] \f$= 0 \Leftrightarrow\f$ The object doesn't have
             * an octant with index \e i.
             */
            OctreeNode *m_children[8];

            /**
             * \brief The node's parent node.
             *
             * \invariant Under the condition that the node isn't the root node, 
             *            <tt>m_parent != 0</tt>.
             *
             * \remarks The root node's \c m_parent pointer is the zero pointer.
             */
            OctreeNode *m_parent;

            /**
             * \brief Number of objects enclosed in the node.
             *
             * The \e total number of objects in the octree enclosed in the 
             * node's volume, including objects contained in child nodes.
             *
             * Is updated by \c add_object(Cullable *), 
             * \c remove_object(Cullable *) and \c Octree::process_updates().
             *
             * \sa \c Octree::process_updates()
             */
            int32_t m_enclosed;

            typedef std::set<Cullable *> ContainedSet;

            /**
             * \brief The set of all contained objects.
             *
             * All the objects enclosed by the node but not enclosed by any of
             * it's child nodes.
             */
            ContainedSet m_contained_objects;

            /**
             * \brief A reference to a variable holding the minimum allowed 
             *        size of a node.
             *
             * If this value changes during a node's lifetime, it will obey
             * the new cap when adding new nodes but not update old nodes
             * to comply with it.
             */
            const real32 &m_min_node_size;


            /**
             * \brief The total number of children the node has.
             *
             * The child count is kept in sync by \c Octree::purge() (removes 
             * unnecessary nodes) and \c add_object(Cullable *) (adds nodes).
             *
             * <em>Initial value:</em> 0.
             */
            int m_child_count;

        private:

            void eval_recursive(OctreeTraverser *traverser, 
                                const int order[]) const throw();

        public:
            /**
             * \brief Constructs an octree node, with the given parameters.
             *
             * Initializes members to:
             * - <tt>m_parent = parent</tt>
             * - <tt>m_half_size = half_size</tt>
             * - <tt>m_pos = pos</tt>
             * - <tt>m_enclosed = 0</tt>
             * - <tt>m_children[i] = 0</tt>, \f$ \forall i \f$
             * - <tt>m_contained_objects = </tt>\f$\emptyset\f$
             * - <tt>m_min_node_size = min_node_size</tt>
             * - <tt>m_child_count = 0</tt>
             *
             * \param parent The parent of the object.
             * \param pos The position of the center point of the octant, in
             *        world coordinates.
             * \param min_node_size The minimum size of an octree node.
             */
            OctreeNode(OctreeNode *parent, real32 half_size, 
                       const Vector3r &pos, const real32 &min_node_size);

            /**
             * \brief Destruct the node, and all it's children.
             *
             * Will destroy any resources allocated by the node, and call
             * the destructor of all its child nodes.
             *
             * \pre The invariant documented at \c m_children must hold.
             * \post The resources used by the subtree in its entirety have
             *       been released.
             *
             * \remarks The parent node's child pointers will \e not be
             * automatically updated.
             */
            ~OctreeNode();

            /**
             * \brief Add an object to the node's subtree.
             *
             * May subdivide the node further, if it isn't already.
             * \e Never subdivdes the node into octants of half size smaller 
             * than m_min_node_size.
             *
             * \pre The object must be enclosed in the node, i.e. 
             *      \c encloses(obj) must be \c true.
             * \post The object has been added to the octree.
             *
             * \param obj The object to be added.
             * \return A pointer to the octree node the object was added to 
             *         (is contained in).
             */
            OctreeNode *add_object(Cullable *obj) throw();
            
            /**
             * \brief Remove \c obj from the node.
             *
             * If \c obj is not contained in the node nothing will be done and
             * this "error" silently ignored.
             *
             * \pre The object is contained in the node, i.e. 
             * <tt>contains(obj)</tt> is \c true.
             * \post The object has been removed from the node.
             *
             * \param obj A pointer to the object to be removed.
             * \param update_enclosed_count Whether or not to update the 
             * enclosed count for ancestor nodes. Set to \c false only if 
             * you know what you are doing!
             *
             * \remarks Unlike, \c add_object this method isn't recursive in
             * any way, and thus expects the object in question to be contained
             * in the node the method was called on, \e not any of its child 
             * nodes.
             */
            void remove_object(Cullable *obj,
                               bool update_enclosed_count = true) throw();

            /**
             * \brief Check if an object is enclosed in the octree node.
             *
             * \return \c true if the node encloses <tt>obj</tt>'s entire BV, 
             *         \c false otherwise.
             */
            bool encloses(Cullable *obj) const throw();

            /**
             * \brief Check if an object is enclosed the node's <em>i</em>:th 
             *        octant.
             *
             * \pre \a obj must be enclosed in the node, if it is not this
             * test will (probably) not return the desired result.
             *
             * \param obj The node to check for being enclosed by the octant.
             * \param octant The octant to check against, in the range 0 to 7.
             * \return \c true if the object is enclosed by the node's 
             *         <em>i</em>:th octant, \c false otherwise.
             *
             * \remarks Note that this doesn't check if the object is enclosed
             * by a child node, or even demands that any child nodes exist.
             * \sa \c encloses(const Cullable *), \c m_children (for octant 
             * indexing).
             */
            bool enclosed_in_octant(Cullable *obj, int octant) const throw();

            /**
             * \brief Check if an object is contained in the node.
             *
             * Note that contain and enclose have entirely different meanings.
             * Contain means that the object is \e in this node, and never a 
             * child node.
             *
             * \param obj The object to check for.
             * \return \c true if \c obj is contained in the node, \c false
             *         otherwise.
             */
            bool contains(const Cullable *obj) const throw();
            
            /**
             * \brief Returns whether the node's subtree is empty or not.
             *
             * \return \c true if no node is contained in the node's subtree,
             * \c false otherwise.
             */
            inline bool is_empty() const throw() { return m_enclosed == 0; }

            /**
             * \brief Get a child octants position relative to its parent.
             *
             * \pre \f$ 0 \leq octant < 8 \f$
             *
             * \param octant The octant for which to get the offset from its 
             * parent node.
             * \return A vector describing the offset of node i from its parent.
             *
             * \sa \c m_children (for octant indexing).
             */
            Vector3r octant_offset(int octant) const throw();

            /**
             * \brief Get the node's parent node.
             *
             * \return The node's parent, 0 if none. 
             */
            OctreeNode *get_parent() throw();

            /**
             * \brief Get the node's parent node.
             *
             * \return The node's parent, 0 if none. 
             */
            const OctreeNode *get_parent() const throw();

            /**
             * \brief Get the number of children the node has.
             *
             * \return The node's total number of children, in the range 0-8 
             * inclusive.
             */
            int get_child_count() const throw();

            OctreeNode *get_octant(int i) throw();

            const OctreeNode *get_octant(int i) const throw();


            void traverse_pvs(primitives::Camera *camera, 
                              OctreeTraverser *traverser,
                              const int order[],
                              int plane_mask) const throw();

            static void object_level_cull(
                Cullable *node, 
                primitives::Camera *camera,
                OctreeTraverser *traverser,
                int plane_mask) throw();
        };

        /**
         * \brief The root node of the octree.
         * \invariant Always non-null (except in the constructor).
         * \remarks May change over time (if the tree is enlarged).
         */
        OctreeNode *m_root;

   
        /**
         * \brief The set of all objects having been marked as needing
         *        their octree info to be updated.
         *
         * Objects are added to the set by \c queue_need_update(Cullable *)
         * and processed (and subsequently removed from the set) by 
         * \c process_updates().
         */
        std::set<Cullable *> m_update_set;


        /**
         * \brief Auxillary lookup table used to perform octree updates.
         *
         * Maps a scene object to the octree node it is contained in.
         * This structure must be updated as objects are inserted, moved in
         * and deleted from the octree.
         */
        std::map<Cullable *, OctreeNode *> m_objects;
        

        /**
         * \brief The minimal size of nodes to be constructed in the octree.
         *
         * \remarks Note that while it is guaranteed that no nodes of size 
         * smaller than \c m_min_node_size are constructed, it is not guaranteed
         * that they do not (already) exist (should you change the minimum size
         * after they have been constructed).
         */
        real32 m_min_node_size;


        /**
         * \brief Contains nodes that are \e possibly empty.
         *
         * Nodes are put in this set when they are empty, however, they might
         * remain in here even if they become repopulated before \c purge() 
         * is called.
         * Note that not all empty nodes are guaranteed to be in the set,
         * but it is guaranteed that at least one (that node being a leaf) node 
         * of every empty subtree can be found herein.
         * It is used by \c purge() to remove unnecessary nodes.
         *
         * \remarks \c add_object(Cullable *) removes re-populated
         * nodes from the set.
         */
        std::set<OctreeNode *> m_empty_nodes;


    protected:
        OctreeNode *grow_to_accomodate(Cullable *obj) throw();

        void remove_subtree_from_empty_nodes(OctreeNode *subtree) throw();

        

    public:
        /**
         * \brief Octree constructor.
         *
         * \param initial_size The initial "half" size of the octree. Objects 
         *        who's center point isn't enclosed by this volume (centered in 
         *        the world origin), and nodes larger than this parameter,
         *        cause the octree to grow as needed.
         * \param min_node_size The \e minimum node size allowed. The tree will
         *        never be subdivided in octants with "half size" smaller than
         *        this.
         * \sa \c OctreeNode
         */
        Octree(real32 initial_size, real32 min_node_size);


        /**
         * \brief Octree destructor.
         *
         * Frees all resources allocated by the octree, as well as any 
         * <em>OctreeNode</em>s \e used by it.
         */
        ~Octree() throw();


        /**
         * \brief Notify the octree that a contained object's bounding volume 
         * has been changed or moved.
         *
         * \return \c true if the node was already scheduled for updating, 
         * \c false otherwise.
         *
         * \remarks Does \e not queue children of \c obj, the caller is 
         * responsible for queuing child nodes, need they be updated.
         * \remarks Since rotations never change the bounding volume, the octree
         * doesn't need to be notified of them.
         * Call this method when an objects has been scaled or moved. If you don't,
         * you can expect erroneous behaviour.
         */
        bool queue_need_update(Cullable *obj) throw();


        /**
         * \brief Make needed changes to the octree.
         *
         * Processes the changes indicated by previous calls to 
         * \c queue_need_update(Cullable *). If no calls to 
         * \c queue_need_update(Cullable *) have been made prior to calling
         * the function, or since the last time it was called, it will do
         * nothing and return immediately.
         *
         * <em>Performance:</em>
         * Let \f$m\f$ and \f$n\f$ be the number of scheduled updates and octree
         * nodes respectively. Expected running time (under "normal", 
         * non-degenerate conditions) is then \f$O(m)\f$, worst case is 
         * \f$O(mn)\f$.
         *
         * \remarks This functionality cannot (easily) be put in \c OctreeNode, 
         * since it needs write access to the root node.
         */
        void process_updates() throw();

        /**
         * \brief Add an object to the octree.
         *
         * If the object already exists in the octree this operation will
         * do nothing.
         *
         * \param obj The object to be inserted.
         *
         * \remarks Worst case running time of this operation can be expected
         * to be linear.
         */
        void add_object(Cullable *obj) throw();

        /**
         * \brief Remove an object from the octree.
         *
         * If the object isn't in the octree this operation will do nothing.
         *
         * \param obj The object to be removed.
         */
        void remove_object(Cullable *obj) throw();

        /**
         * \brief Check whether a given object is in the octree or not.
         *
         * \return \c true if \c obj is contained in the octree, \c false
         * otherwise.
         */
        bool contains(Cullable *obj) const throw();

        /**
         * \brief Traverse the octree, in rough front to back order.
         *
         * Depending on the cull tests specified it will also perform
         * frustum culling against both the octree and object bounding volumes.
         *
         * \param frustum The frustum (in world coordinates) used for culling.
         * \param traverser A pointer to the object called for each node in the 
         *        octree, as it's being traversed - its traverse() method is 
         *        called for each cullable in the PVS.
         * \note This interface is \e volatile and subject to change!
         */
        void traverse_pvs(primitives::Camera *camera, 
                          OctreeTraverser *traverser) throw();

        /**
         * \brief Remove unnecessary nodes.
         *
         * Will remove any redundant octree nodes, i.e. empty leaf nodes or
         * paths beginning in empty nodes, and thus necessarily end in empty 
         * nodes.
         *
         * <em>Running time:</em> Linear in the number of excess nodes.
         * Giving a pathological worst case performance of \f$O(n)\f$ in the 
         * total number of contained objects. The number of excess nodes
         * can be retrieved using \c get_excess_node_count().
         *
         * \remarks You may purge the octree periodically to keep memory usage 
         * down. However, refrain from calling it each frame is this could cause
         * a performance hit - especially since some empty nodes are likely to 
         * be re-populated in the next frame.
         * If you have a 
         * very dynamic scene it is recommended that you purge the octree once
         * every few frames. If your particular scene changes little, you 
         * needn't purge it often.
         */
        void purge() throw();


        /**
         * \brief Get the number of excess, unnecessary nodes in the octree.
         *
         * Returns the number of nodes that would be removed by a call 
         * to \c purge().
         *
         * \return The number of unnecessary nodes that can safely be removed.
         */
        unsigned int get_excess_node_count() const throw();


        /**
         * \brief Get the number of objects in the octree.
         */
        unsigned int get_contained_count() const throw();
    };
}
}


#endif // __PEEKABOT_RENDERER_OCTREE_HH
