/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * \brief \c OctreeNode class implementation.
 *
 * \todo Could benefit from clearer separation from \c Octree, i.e. by
 * adding dedicated methods for add/remove child nodes.
 */

#include <GL/glew.h>
#include <iostream>
#include <stack>
#include <cassert>
#include "Octree.hh"
#include "Cullable.hh"
#include "primitives/Camera.hh"
#include "OctreeTraverser.hh"
#include "Frustum.hh"


using namespace peekabot;
using namespace peekabot::renderer;



Octree::OctreeNode::OctreeNode(OctreeNode *parent,
                               real32 half_size,
                               const Vector3r &pos, 
                               const real32 &min_node_size) :
    m_half_size(half_size),
    m_pos(pos), 
    m_parent(parent), 
    m_enclosed(0),
    m_min_node_size(min_node_size),
    m_child_count(0)
{
    for( int i = 0; i < 8; i++ )
        m_children[i] = 0;
}

Octree::OctreeNode::~OctreeNode()
{
    for( int i = 0; i < 8; i++ )
        if( m_children[i] != 0 )
            delete m_children[i];
}




Octree::OctreeNode *Octree::OctreeNode::add_object(Cullable *obj) throw()
{
    // 1. It is enclosed by a child node and the minimum node size has not
    //    been reached.
    //    1.1. That child exists, do a recurisve add.
    //    1.2. It doesn't exist, create it and do a recursive add.
    // 2. Not enclosed by a child, add to this node.

    //assert(encloses(obj));

    ++m_enclosed;

    // Test for 2)
    if( obj->get_bounding_sphere().get_radius() <= m_half_size/2 &&
        m_half_size/2 > m_min_node_size )
    {
        // Add to child node
        for( int i = 0; i < 8; i++ )
        {      
            if( enclosed_in_octant(obj, i) )
            {
                if( m_children[i] == 0 )
                {
                    // invariant: m_half_size/2 > m_min_node_size

                    m_children[i] = new OctreeNode(
                        this, 
                        m_half_size / 2,
                        m_pos + octant_offset(i), 
                        m_min_node_size);

                    ++m_child_count;
                }

                return m_children[i]->add_object(obj);
            }
        }

        TheMessageHub::instance().publish(
            ERROR_MESSAGE,
            "Pending assertion failure! Object not enclosed in a child node, "
            "even though it should meet the requirements to do so. "
            "bounding sphere: pos=(%f, %f, %f), r=%f", 
            obj->get_bounding_sphere().get_pos()(0),
            obj->get_bounding_sphere().get_pos()(1),
            obj->get_bounding_sphere().get_pos()(2),
            obj->get_bounding_sphere().get_radius());
        
        assert(false); // Should never happend
    }
    else
    {
        m_contained_objects.insert(obj);

        return this;
    }
}
            

void Octree::OctreeNode::remove_object(Cullable *obj,
                                       bool update_enclosed_count) throw()
{
    assert(contains(obj));

    m_contained_objects.erase(obj);
    --m_enclosed;

    if( update_enclosed_count )
    {
        // Decrease enclosed count for ancestor nodes
        OctreeNode *parent = m_parent;
        while( parent != 0 )
        {
            --parent->m_enclosed;

            parent = parent->m_parent;
        }
    }
}


bool Octree::OctreeNode::encloses(Cullable *obj) const throw()
{
    if( obj->get_bounding_sphere().get_radius() > m_half_size )
        return false;

    real32 x = obj->get_bounding_sphere().get_pos()(0);
    real32 y = obj->get_bounding_sphere().get_pos()(1);
    real32 z = obj->get_bounding_sphere().get_pos()(2);

    if( x >= m_pos(0) - m_half_size/2 && x < m_pos(0) + m_half_size/2 &&
        y >= m_pos(1) - m_half_size/2 && y < m_pos(1) + m_half_size/2 &&
        z >= m_pos(2) - m_half_size/2 && z < m_pos(2) + m_half_size/2 )
    {
        return true;
    }
    else
        return false;
}


bool Octree::OctreeNode::enclosed_in_octant(Cullable *obj, 
                                            int octant) const throw()
{
    if( obj->get_bounding_sphere().get_radius() > m_half_size/2 )
        return false;

    real32 x = obj->get_bounding_sphere().get_pos()(0);
    real32 y = obj->get_bounding_sphere().get_pos()(1);
    real32 z = obj->get_bounding_sphere().get_pos()(2);

    return ( ( octant&1 ? x >= m_pos(0) : x < m_pos(0) ) && 
             ( octant&2 ? y >= m_pos(1) : y < m_pos(1) ) && 
             ( octant&4 ? z >= m_pos(2) : z < m_pos(2) ) );
}


bool Octree::OctreeNode::contains(const Cullable *obj) const throw()
{
    return m_contained_objects.find(const_cast<Cullable *>(obj)) 
        != m_contained_objects.end();
}



/*bool Octree::OctreeNode::is_empty() const throw()
{
    //assert(m_enclosed >= 0);
    
    return m_enclosed == 0;
}*/



Vector3r Octree::OctreeNode::octant_offset(int octant) const throw()
{
    real32 hs = m_half_size/4;

    return Vector3r( hs*(2*(octant&1)-1),
                     hs*((octant&2)-1),
                     hs*(((octant&4)>>1)-1) );
}


Octree::OctreeNode *Octree::OctreeNode::get_parent() throw()
{
    return m_parent;
}


const Octree::OctreeNode *Octree::OctreeNode::get_parent() const throw()
{
    return m_parent;
}


int Octree::OctreeNode::get_child_count() const throw()
{
    return m_child_count;
}


void Octree::OctreeNode::eval_recursive(
    OctreeTraverser *traverser,
    const int order[]) const throw()
{
    for( ContainedSet::const_iterator it = m_contained_objects.begin();
         it != m_contained_objects.end(); ++it )
    {
        traverser->traverse(*it);
    }

    // Execute fn on recursively on all non-empty child nodes
    for( int i = 0; i < 8; i++ )
    {
        OctreeNode *child = m_children[order[i]];
        if( child && !child->is_empty() )
            child->eval_recursive(traverser, order);
    }
}



void Octree::OctreeNode::traverse_pvs(primitives::Camera *camera, 
                                      OctreeTraverser *traverser,
                                      const int order[],
                                      int plane_mask) const throw()
{
    if( is_empty() )
        return;

    const boost::shared_ptr<Frustum> frustum(camera->get_frustum());

    BoundingSphere bsphere = BoundingSphere(m_pos, m_half_size*M_SQRT2);

    // Cull the node?
    IntersectionTestResult culltest = bsphere.contained_in(
        frustum->get_bounding_sphere());

    //if( culltest != DISJOINT )
    //    culltest = bsphere.contained_in(camera->get_frustum_bounding_cone());

    if( culltest != DISJOINT )
        culltest = frustum->is_in_frustum(bsphere, plane_mask);

    if( culltest == DISJOINT )
    {
        // Cull all child nodes!
        return;
    }
    else if( culltest == INSIDE )
    {
        eval_recursive(traverser, order);
    }
    else // culltest == INTERSECT
    {
        // If partial intersection:
        // 1. Do object level tests for contained objects
        // 2. traverse child nodes

        for( std::set<Cullable *>::iterator it = m_contained_objects.begin();
             it != m_contained_objects.end(); ++it )
        {
            object_level_cull(*it, camera, traverser, plane_mask);
        }


        // Cull subnodes
        for( int i = 0; i < 8; i++ )
        {
            OctreeNode *child = m_children[order[i]];

            if( child ) //&& !child->is_empty() )
            {
                child->traverse_pvs(camera,
                                    traverser,
                                    order,
                                    plane_mask);
            }
        }
    }
}




// Performs sphere-sphere and sphere-frustum tests
void Octree::OctreeNode::object_level_cull(
    Cullable *node, 
    primitives::Camera *camera,
    OctreeTraverser *traverser,
    int plane_mask) throw()
{
    const boost::shared_ptr<Frustum> frustum(camera->get_frustum());

    // Cull the node?
    IntersectionTestResult culltest = node->get_bounding_sphere().contained_in(
        frustum->get_bounding_sphere());

    if( culltest != DISJOINT )
        culltest = frustum->is_in_frustum(
            node->get_bounding_sphere(), plane_mask);

    if( culltest != DISJOINT )
        traverser->traverse(node);

    return;
}


Octree::OctreeNode *Octree::OctreeNode::get_octant(int i) throw()
{
    return m_children[i];
}


const Octree::OctreeNode *Octree::OctreeNode::get_octant(int i) const throw()
{
    return m_children[i];
}


