/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */



#include <iostream>
#include <stdexcept>
#include <cassert>
#include <GL/glew.h>

#include "PointSetNode.hh"

#if 0
using namespace peekabot;
using namespace peekabot::renderer;


static unsigned int BATCH_SIZE = 1024;


// ------------------------- PointBatch implementation -------------------------


static const int8_t INITIAL_CACHE_SCORE  = -10;
static const int8_t MIN_CACHE_SCORE      = -10;
static const int8_t MAX_CACHE_SCORE      =  10;
static const int8_t CACHE_ACQUIRE_BONUS  =  5;
static const int8_t CACHE_RELEASE_BONUS  = -5;
static const int8_t SCORE_DECAY_PENALTY  = -1;
static const int8_t RENDERED_CACHE_BONUS =  1;



PointSetNode::OctreeNode::PointBucket::PointBatch::PointBatch(
    unsigned int size) throw(std::exception *)
    : m_capacity(size),
      m_used(0),
      m_is_cached(false),
      m_cache_score(INITIAL_CACHE_SCORE)
{
    assert(m_capacity > 0);

    if( !(m_colors = new ColorType[3 * m_capacity]) )
        throw new std::runtime_error("Out of memory");

    if( !(m_coords = new CoordType[3 * m_capacity]) )
    {
        delete[] m_colors;
        throw new std::runtime_error("Out of memory");
    }
}


PointSetNode::OctreeNode::PointBucket::PointBatch::~PointBatch() throw()
{
    delete[] m_colors;
    delete[] m_coords;

    if( m_is_cached )
    {
        // Remove batch from the cache
        // TODO
        //ThePointBatchCache::instance().release(*this);
        m_is_cached = false;
        
        assert( false );
    }
}


void PointSetNode::OctreeNode::PointBucket::PointBatch::render() throw()
{
    //m_is_cached = ThePointBatchCache::instance().cache(*this);

    if( m_is_cached )
    {
        // TODO
    }
    else
    {
        // TODO

        glColorPointer(3, GL_UNSIGNED_BYTE, 0, m_colors);
        glVertexPointer(3, GL_FLOAT, 0, m_coords);   
        glDrawArrays(GL_POINTS, 0, m_used);
    }

    award_cache_score(RENDERED_CACHE_BONUS);
}


bool PointSetNode::OctreeNode::PointBucket::PointBatch::is_full() const throw()
{
    return ( m_used == m_capacity );
}


bool PointSetNode::OctreeNode::PointBucket::PointBatch::get_size() const throw()
{
    return m_used;
}


unsigned int PointSetNode::OctreeNode::PointBucket::PointBatch::get_capacity() 
    const throw()
{
    return m_capacity;
}


bool PointSetNode::OctreeNode::PointBucket::PointBatch::add(
    const Vector3r &p, const RGBColor &color) throw()
{
    if( is_full() )
    {
        return true;
    }

    m_coords[3 * m_used + 0] = (CoordType)p(0);
    m_coords[3 * m_used + 1] = (CoordType)p(1);
    m_coords[3 * m_used + 2] = (CoordType)p(2);
    
    // Note: should use round for correctness, but that would lower speed, so...
    m_colors[3 * m_used + 0] = (ColorType)(255*color.r);
    m_colors[3 * m_used + 1] = (ColorType)(255*color.g);
    m_colors[3 * m_used + 2] = (ColorType)(255*color.b);

    if( m_is_cached )
    {
        // Update cache

        // TODO
        assert( false );
    }

    ++m_used;

    return false;
}


Vector3r PointSetNode::OctreeNode::PointBucket::PointBatch::get_coords(
    unsigned int i) const throw()
{
    return Vector3r(m_coords[3*i+0], m_coords[3*i+1], m_coords[3*i+2]);
}


RGBColor PointSetNode::OctreeNode::PointBucket::PointBatch::get_color(
    unsigned int i) const throw()
{
    float f = 1/255.0f;
    return RGBColor(f*m_colors[3*i+0], f*m_colors[3*i+1], f*m_colors[3*i+2]);
}


void PointSetNode::OctreeNode::PointBucket::PointBatch::relocate_points(
    PointSetNode::OctreeNode::OctreeNode *node) throw()
{
    // IMPORTANT: We do NOT update the enclosed count when doing
    // add_object(), as this would ruin it (the enclosed count).

    for( unsigned int i = 0; i < m_used; ++i )
        node->add_object(get_coords(i), get_color(i), false);

    m_used = 0;
}


void PointSetNode::OctreeNode::PointBucket::PointBatch::award_cache_score(
    int8_t bonus) throw()
{
    m_cache_score += bonus;

    if( m_cache_score > MAX_CACHE_SCORE )
        m_cache_score = MAX_CACHE_SCORE;
    else if( m_cache_score < MIN_CACHE_SCORE )
        m_cache_score = MIN_CACHE_SCORE;
}


bool PointSetNode::OctreeNode::PointBucket::PointBatch::is_cached() 
    const throw()
{
    return m_is_cached;
}





// ------------------------- PointBucket implementation ------------------------
            
PointSetNode::OctreeNode::PointBucket::PointBucket()
    : m_size(0)
{
}


PointSetNode::OctreeNode::PointBucket::~PointBucket() throw()
{
    for( ContainerType::iterator it = m_batches.begin();
         it != m_batches.end(); ++it )
    {
        delete *it;
    }

    m_batches.clear();

    m_size = 0;
}


void PointSetNode::OctreeNode::PointBucket::render() throw()
{
    for( ContainerType::iterator it = m_batches.begin();
         it != m_batches.end(); ++it )
    {
        (*it)->render();
    }
}


void PointSetNode::OctreeNode::PointBucket::add(
    const Vector3r &p, const RGBColor &color) throw(std::exception *)
{
    // Is there empty space in an already existing batch?
    if( m_batches.empty() || m_batches.back()->is_full() )
    {
        // No... make room!
        PointBatch *batch = new PointBatch(BATCH_SIZE); // might throw!
        if( !batch )
            throw new std::runtime_error("Out of memory");

        m_batches.push_back(batch);
    }

    PointBatch *batch = m_batches.back();
    batch->add(p, color);
    
    ++m_size;
}


void PointSetNode::OctreeNode::PointBucket::relocate_points(
    PointSetNode::OctreeNode::OctreeNode *node) throw()
{
    for( ContainerType::iterator it = m_batches.begin();
         it != m_batches.end(); ++it )
    {
        (*it)->relocate_points(node);

        delete *it;
    }

    m_batches.clear();

    m_size = 0;
}


unsigned int PointSetNode::OctreeNode::PointBucket::get_size() throw()
{
    return m_size;
}


bool PointSetNode::OctreeNode::PointBucket::is_empty() const throw()
{
    return ( m_size == 0 );
}


void PointSetNode::OctreeNode::PointBucket::penalize_cache_scores() throw()
{
    for( ContainerType::iterator it = m_batches.begin();
         it != m_batches.end(); ++it )
    {
        (*it)->award_cache_score(SCORE_DECAY_PENALTY);
    }
}





// ----------------------- PointBatchCache implementation ----------------------

PointSetNode::OctreeNode::PointBucket::PointBatchCache::PointBatchCache(
    unsigned int block_size, uint32_t initial_size, uint32_t max_size) throw()
    : m_block_size(block_size), m_size(initial_size), m_max_size(max_size)
{
    assert( initial_size <= max_size );

    // TODO
}


PointSetNode::OctreeNode::PointBucket::PointBatchCache::~PointBatchCache() 
    throw()
{
    // TODO
}


std::pair<bool, uint32> 
PointSetNode::OctreeNode::PointBucket::PointBatchCache::cache(
    PointBatch &batch) throw()
{
    // TODO

    return std::make_pair(false, 0);
}


void PointSetNode::OctreeNode::PointBucket::PointBatchCache::release(
    PointBatch &batch) throw()
{
    // TODO

    assert( batch.is_cached() );
}


void PointSetNode::OctreeNode::PointBucket::PointBatchCache::update_cache(
    const PointBatch &batch) throw()
{
    assert( batch.is_cached() );

    // Since we don't (yet?) support removing points it's safe to say
    // a should be added to the cache...
    
    // TODO
}


#endif
