/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * \brief \c PointSetNode class implementation.
 */

#include <GL/glew.h>
#include <iostream>
#include <stack>
#include <cassert>
#include <vector>
#include <cmath>
#include <boost/bind.hpp>
#include <boost/function.hpp>

#include "RenderEngine.hh"
#include "PointSetNode.hh"
#include "BoundingVolumes.hh"

using namespace peekabot;
using namespace peekabot::renderer;


#if 0
static const real32 SQRT_2 = sqrtf(2);


static const real32 INITIAL_OCTREE_SIZE = 10; // in meters


static const real32 MIN_NODE_SIZE = 2; // in meters
static const int    LOD_LEVELS = 8;
static const int    MIN_NODE_OBJECTS = 4096;



// -------------------- PointContainer implementation --------------------

PointSetNode::OctreeNode::PointContainer::PointContainer(
    int bucket_count) throw()
    : m_contained(0), m_bucket_count(bucket_count)
{
    assert( bucket_count > 0 );

    m_buckets = new PointBucket[m_bucket_count];

    // TODO: check for out of memory


    class CacheScorePenalizer : public EngineObserver
    {
        PointContainer &m_pcont;
        int m_refcount;
    public:
        CacheScorePenalizer(PointContainer &pcont)
            : m_pcont(pcont), m_refcount(0)
        {
        }

        virtual void on_registered(EngineEvent event) throw()
        {
            ++m_refcount;
        }

        virtual void on_deregistered(EngineEvent event) throw()
        {
            if( --m_refcount == 0 )
                delete this;
        }

        virtual void notify(EngineEvent event, RenderEngine *subject) throw()
        {
            m_pcont.penalize_cache_scores();
        }
    };

    /*EngineObserver *penalizer = new CacheScorePenalizer(*this);
    RenderEngine::register_observer(ON_PRE_RENDER, penalizer);
    RenderEngine::register_observer(ON_PRE_PICK, penalizer);*/
}


PointSetNode::OctreeNode::PointContainer::~PointContainer() throw()
{
    delete[] m_buckets;
}


void PointSetNode::OctreeNode::PointContainer::relocate_points(
    OctreeNode *node) throw()
{
    for( int i = 0; i < m_bucket_count; ++i )
        m_buckets[i].relocate_points(node);
}


void PointSetNode::OctreeNode::PointContainer::render(int n_buckets) throw()
{
    assert( n_buckets <= m_bucket_count );

    for( int i = 0; i < n_buckets; ++i )
        m_buckets[i].render();
}


void PointSetNode::OctreeNode::PointContainer::add(
    const Vector3r &p, const RGBColor &color) throw()
{
    int active_bucket = (int)roundf((float)rand()/RAND_MAX*(m_bucket_count-1));

    m_buckets[active_bucket].add(p, color);

    // Change active bucket, and keep m_active_bucket mod bucket_count
    //if( ++m_active_bucket >= m_bucket_count )
    //    m_active_bucket = 0;
}



void PointSetNode::OctreeNode::PointContainer::penalize_cache_scores() throw()
{
    for( int i = 0; i < m_bucket_count; ++i )
        m_buckets[i].penalize_cache_scores();
}





// -------------------- OctreeNode implementation --------------------

PointSetNode::OctreeNode::OctreeNode(OctreeNode *parent,
                                     real32 size,
                                     const Vector3r &pos) :
    m_parent(parent), 
    m_size(size),
    m_pos(pos),
    m_enclosed(0),
    m_points(0)
{
    for( int i = 0; i < 8; ++i )
        m_children[i] = 0;
}

PointSetNode::OctreeNode::~OctreeNode()
{
    if( m_points )
        delete m_points;

    for( int i = 0; i < 8; ++i )
        if( m_children[i] != 0 )
            delete m_children[i];
}



const real32 PointSetNode::OctreeNode::get_node_size() const throw()
{
    return m_size;
}


PointSetNode::OctreeNode *PointSetNode::OctreeNode::add_object(
    const Vector3r &p, 
    const RGBColor &color, 
    bool update_enclosed_count) throw()
{
    // 1. It is enclosed by a child node and the minimum node size has not
    //    been reached.
    //    1.1. That child exists, do a recurisve add.
    //    1.2. It doesn't exist, create it and do a recursive add.
    // 2. Not enclosed by a child, add to this node.

    assert(encloses(p));

    if( update_enclosed_count )
        ++m_enclosed;

    if( m_size >= 2 * MIN_NODE_SIZE && 
        m_enclosed >= MIN_NODE_OBJECTS )
    {
        if( m_points )
        {
            // Oops... this node has points, even
            // though there are/can be children! Relocate 'em!

            PointContainer *tmp = m_points;
            m_points = 0;
                       
            tmp->relocate_points(this);
            
            delete tmp;
        }
        

        // Add to child node
        for( int i = 0; i < 8; ++i )
        {
            if( enclosed_in_octant(p, i) )
            {
                if( m_children[i] == 0 )
                {
                    // invariant: m_size/2 >= MIN_NODE_SIZE

                    m_children[i] = new OctreeNode(this, 
                                                   m_size / 2,
                                                   m_pos + octant_offset(i));
                }

                return m_children[i]->add_object(
                    p, color, update_enclosed_count);
            }
        }
    }
    else
    {
        if( !m_points )
        {
            m_points = new PointContainer(LOD_LEVELS);
            // TODO: check for out of memory
        }

        m_points->add(p, color);
        
        return this;
    }
}



bool PointSetNode::OctreeNode::encloses(const Vector3r &p) const throw()
{
    return ( p(0) >= m_pos(0) - m_size/2 && p(0) <= m_pos(0) + m_size/2 &&
             p(1) >= m_pos(1) - m_size/2 && p(1) <= m_pos(1) + m_size/2 &&
             p(2) >= m_pos(2) - m_size/2 && p(2) <= m_pos(2) + m_size/2 );
}


bool PointSetNode::OctreeNode::enclosed_in_octant(
    const Vector3r &p, int octant) const throw()
{
    real32 quarter_size = m_size/4;
    Vector3r pos(m_pos + octant_offset(octant));

    // Note that the z plane tests are least likely to fail (given our typical 
    // scene setup), and are thus  performed last.
    return ( p(0) >= pos(0) - quarter_size && p(0) <= pos(0) + quarter_size &&
             p(1) >= pos(1) - quarter_size && p(1) <= pos(1) + quarter_size &&
             p(2) >= pos(2) - quarter_size && p(2) <= pos(2) + quarter_size );
}



Vector3r PointSetNode::OctreeNode::octant_offset(int octant) const throw()
{
    real32 hs = m_size / 4;

    return Vector3r( hs*(2*(octant&1)-1),
                     hs*((octant&2)-1),
                     hs*(((octant&4)>>1)-1) );
}


PointSetNode::OctreeNode *PointSetNode::OctreeNode::get_parent() throw()
{
    return m_parent;
}


const PointSetNode::OctreeNode *PointSetNode::OctreeNode::get_parent() const throw()
{
    return m_parent;
}



void PointSetNode::OctreeNode::render_points(
    CameraNode *camera, const Vector3r &camera_pos) const throw()
{
    assert( m_points );

    int n_buckets = LOD_LEVELS;


    const float kickin_dist = 20;
    const float min_lod_dist = 120;


    float d = camera_pos.dist(m_pos);


    if( d >= kickin_dist )
    {
        float lod_quotient = 
            1 - (d - kickin_dist) / (min_lod_dist - kickin_dist);

        n_buckets = (int)(lod_quotient * (n_buckets-1)) + 1;
    }

    
    m_points->render(n_buckets);
}



void PointSetNode::OctreeNode::render_recursive(
    CameraNode *camera,
    const int order[],
    const Vector3r &camera_pos) const throw()
{
    if( m_points )
        render_points(camera, camera_pos);

    // Render all non-empty child nodes recursively
    for( int i = 0; i < 8; ++i )
    {
        OctreeNode *child = m_children[order[i]];
        if( child )
            child->render_recursive(camera, order, camera_pos);
    }
}



void PointSetNode::OctreeNode::render(
    CameraNode *camera, 
    const int order[],
    const BoundingSphere &frustum_bsphere,
    const Plane *frustum_planes,
    const Vector3r &camera_pos,
    int plane_mask) const throw()
{
    //if( is_empty() )
    //    return;

    BoundingSphere bsphere = BoundingSphere(m_pos, m_size*SQRT_2);

    // Cull the node?
    IntersectionTestResult culltest = bsphere.contained_in(frustum_bsphere);

    //std::cout << "frust BS: " << frustum_bsphere.get_pos() << ", radius = " << frustum_bsphere.get_radius() << std::endl;
    //std::cout << "node BS: " << bsphere.get_pos() << ", radius = " << bsphere.get_radius() << std::endl;

    if( culltest != DISJOINT )
      culltest = frustum_test(bsphere, frustum_planes, plane_mask);

    if( culltest == DISJOINT )
    {
        // Cull all child nodes - yay!
        return;
    }
    else if( culltest == INSIDE )
    {
        render_recursive(camera, order, camera_pos);
    }
    else // culltest == INTERSECT
    {
        // Partial intersection

        if( m_points )
            render_points(camera, camera_pos);

        // OPT possible optimzaition: put an else here, since if m_points
        // is null, there WILL BE children, if it's non-null, there WILL BE NO
        // children...

        // Cull subnodes
        for( int i = 0; i < 8; ++i )
        {
            OctreeNode *child = m_children[order[i]];

            if( child )
            {
                child->render(camera,
                              order,
                              frustum_bsphere,
                              frustum_planes,
                              camera_pos,
                              plane_mask);
            }
        }
    }
}





PointSetNode::OctreeNode *PointSetNode::grow_to_accomodate(const Vector3r &p) throw()
{
    // obj outside the octree!
    // deal with special case... by extending the tree outwards,
    // that is, by create a new big node and use the current
    // root node as a child octant in the new node.
    do
    {
        OctreeNode *old_root = m_root;
        m_root = new OctreeNode(0, 
                                old_root->get_node_size()*2, 
                                Vector3r(0,0,0));

        Vector3r configuration[] =
        {
            Vector3r(-1, -1, -1),
            Vector3r(-1, -1,  1),
            Vector3r(-1,  1, -1),
            Vector3r(-1,  1,  1),
            Vector3r( 1, -1, -1),
            Vector3r( 1, -1,  1),
            Vector3r( 1,  1, -1),
            Vector3r( 1,  1,  1)
        };

        real32 min_dist;
        int min_conf = -1;

        for( int i = 0; i < 8; i++ )
        {
            Vector3r confpos = old_root->m_pos + 
            (old_root->get_node_size()/2)*configuration[i];
                        
            real32 d = confpos.dist(p);

            if( min_conf == -1 || d < min_dist )
            {
                min_conf = i;
                min_dist = d;
            }
        }

        // Now we just have to figure out in which directions
        // to extend the octree.
        // This can easily be done by find the configuration p
        // which minimizes the distance dist(obj,p). There
        // are 8 possible configurations.
        //
        // The different possible configurations are the 
        // permutations of <x,y,z>, x,y,z \in {-1,1}. I.e.
        // all the ways the new cube can be offset from the
        // old root node.
        //
        // Having found the best configuration, we implicitly
        // also have the index of the new root node to put the 
        // old one in.
        // index = (x/2+0.5)+2*(y/2+0.5)+4*(z/2+0.5)

        m_root->m_pos = old_root->m_pos + 
            old_root->get_node_size()/2*configuration[min_conf];

        /*int index = static_cast<int>(
            1*(configuration[min_conf](0)/2+0.5) +
            2*(configuration[min_conf](1)/2+0.5) +
            4*(configuration[min_conf](2)/2+0.5) );*/

        int index = -1;
        for( int i = 0; i < 8; ++i )
        {
            if( m_root->octant_offset(i) + m_root->m_pos == old_root->m_pos )
            {
                index = i;
                break;
            }
        }

        assert( index != -1 );

        m_root->m_children[index] = old_root;
        old_root->m_parent = m_root;
        m_root->m_enclosed = old_root->m_enclosed;

    } while( !m_root->encloses(p) );

    return m_root;
}



IntersectionTestResult PointSetNode::OctreeNode::frustum_test(
                const BoundingSphere &bsphere,
                const Plane *frustum_planes,
                int &plane_mask) const throw()
{
    register IntersectionTestResult retval = INSIDE;

    for( register int i = 5; i >= 0; --i )
    {
        while( (plane_mask >> (i - 1)) & 1 == 0 )
            if( i > 0 )
                --i;
            else
                return retval;

        IntersectionTestResult test = bsphere.contained_in(frustum_planes[i]);
            
        if( test == DISJOINT )
        {
            plane_mask ^= (1 << i);
            retval = DISJOINT;
        }
        else if( test == INTERSECT && retval != DISJOINT )
            retval = INTERSECT;
    }

    return retval;
}







// -------------------- PointSetNode implementation --------------------

PointSetNode::PointSetNode() throw()
{
    m_max_point_dist = 0;

    m_root = new OctreeNode(0, INITIAL_OCTREE_SIZE, Vector3r(0,0,0));

    /*get_state().set(
        boost::shared_ptr<Statelet>(
            new statelets::Lighting(false)));*/
}


PointSetNode::~PointSetNode() throw()
{
    delete m_root;
}


void PointSetNode::add_object(const Vector3r &p, const RGBColor &color) throw()
{
    if( !m_root->encloses(p) )
    {
        grow_to_accomodate(p);
    }
    
    OctreeNode *node = m_root->add_object(p, color);

    // TODO: verify this code...
    real32 d = p.norm_squared();
    if( d > m_max_point_dist*m_max_point_dist )
    {
        m_max_point_dist = sqrtf(d);

        //invalidate_bvs();

        SceneEntity::trigger(ON_CULLABLE_BV_SET, this, this);
    }
}





void PointSetNode::render(primitives::Camera *camera) throw()
{/*
    //if( !settings.m_no_materials )
    {
        glEnableClientState(GL_COLOR_ARRAY);
    }

   
    glEnableClientState(GL_VERTEX_ARRAY);


    // Translate the camera's frustum bsphere to local space coordinates
    Matrix4r mtow_inv(get_mtow().inverse());

    BoundingSphere frustum_bs = camera->get_frustum_bounding_sphere();
    frustum_bs.set_pos(mtow_inv * Vector4r(frustum_bs.get_pos(), 1));

    // ...and the frustum planes. d values needn't be recalculated, they will
    // remain valid after we've transformed the planes.
    Plane frustum_planes[6];
    camera->get_frustum_planes(frustum_planes);
    for( int i = 0; i < 6; ++i )
    {
        frustum_planes[i].set_normal(
            mtow_inv * Vector4r(frustum_planes[i].get_normal(), 0));
    }



    // Get the camera position in local coordinates
    Vector3r camera_pos(mtow_inv * camera->get_world_pos());





    // Calculate the render order
    Vector3r view_dir(mtow_inv * camera->get_view_vector());
    Vector3r view_pos(mtow_inv * camera->get_world_pos());

    std::vector<std::pair<real32, int> > tmp(8);

    // Now, for each octant, find the y-distance to the viewing frustum:
    for( int i = 0; i < 8; i++ )
    {
        real32 z_dist = view_dir * (view_pos - m_root->octant_offset(i));
        tmp[i] = std::make_pair(-z_dist, i);
    }

    std::sort(tmp.begin(), tmp.end());
    int order[8];
    for( int i = 0; i < 8; i++ )
        order[i] = tmp[i].second;



    // Render!!!
    m_root->render(camera, order, frustum_bs, frustum_planes, camera_pos, 0x3F);
   

    // Clean up
    glDisableClientState(GL_VERTEX_ARRAY);
    
    {
        glDisableClientState(GL_COLOR_ARRAY);
        }*/
}



void PointSetNode::calculate_bvs() throw()
{
    //m_bsphere = BoundingSphere(get_world_pos(), m_max_point_dist);
}

#endif
