/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_RENDERER_POINT_SET_NODE_HH
#define __PEEKABOT_RENDERER_POINT_SET_NODE_HH


#include <cassert>
#include <list>

#include "../Types.hh"
#include "../Singleton.hh"
#include "SceneNode.hh"
#include "Cullable.hh"
#include "BoundingVolumes.hh"
#include "primitives/Camera.hh"


namespace peekabot
{
namespace renderer
{


#if 0
    class PointSetNode : public SceneNode, public Cullable
    {
        // Non-loose, plain vanilla octree implementation
        // Implemented using AABBs in local space, thus the view frustum
        // must be transformed into local space before clipping
        class OctreeNode
        {
            class PointBucket
            {
                class PointBatch
                {
                    typedef real32 CoordType;
                    typedef uint8_t  ColorType;
                    
                    bool m_is_cached;
                    int8_t m_cache_score;
                    uint32_t m_cache_offset;
                    unsigned int m_used;
                    unsigned int m_capacity;
                    CoordType *m_coords;
                    ColorType *m_colors;
                    
                public:
                    PointBatch(unsigned int size) throw(std::exception *);
                    ~PointBatch() throw();
                    void render() throw();
                    bool is_full() const throw();
                    bool get_size() const throw();
                    unsigned int get_capacity() const throw();
                    bool add(const Vector3r &p, const RGBColor &color) throw();
                    void relocate_points(OctreeNode *node) throw();
                    void award_cache_score(int8_t bonus) throw();
                    bool is_cached() const throw();
                    Vector3r get_coords(unsigned int i) const throw();
                    RGBColor get_color(unsigned int i) const throw();
                };



                class PointBatchCache
                {
                    uint32_t m_size;
                    uint32_t m_max_size;
                    unsigned int m_block_size;

                    std::list<uint32> m_free;

                public:
                    PointBatchCache(
                        unsigned int block_size, 
                        uint32_t initial_size, uint32_t max_size) throw();
                    ~PointBatchCache() throw();

                    std::pair<bool, uint32> cache(PointBatch &batch) throw();
                    void release(PointBatch &batch) throw();
                    void update_cache(const PointBatch &batch) throw();
                };

                typedef Singleton<PointBatchCache> ThePointBatchCache;



                typedef std::list<PointBatch *> ContainerType;

                ContainerType m_batches;
                unsigned int m_size;
                
            public:
                void render() throw();
                void add(const Vector3r &p, const RGBColor &color) throw(std::exception *);
                PointBucket();
                ~PointBucket() throw();
                void relocate_points(OctreeNode *node) throw();
                unsigned int get_size() throw();
                bool is_empty() const throw();
                void penalize_cache_scores() throw();
            };



            class PointContainer
            {
                //int m_active_bucket;
                int m_contained;
                PointBucket *m_buckets;
                int m_bucket_count;

            public:
                PointContainer(int bucket_count) throw();
                ~PointContainer() throw();
                void relocate_points(OctreeNode *node) throw();
                void render(int n_buckets) throw();
                void add(const Vector3r &p, const RGBColor &color) throw();
                void penalize_cache_scores() throw();
            };





            int get_buckets_to_use(
                const Vector3r &camera_pos,
                const Vector3r &node_pos);






        public:
            /**
             * \brief The length of the child octant's AABB's side.
             */
            real32 m_size;

            /**
             * \brief The position of node's center, in local space coordinates.
             */
            Vector3r m_pos;

            /**
             * \brief The octants the node is divided into.
             *
             * The child nodes of the node that divide itself volume
             * into 8 equally sized volumes. A node can have 0 to 8 child nodes.
             * If m_children[i] is null, the node does not have the i octant.
             * 
             * Octants are numbered according to the following scheme:
             * Let \f$<x,y,z>\f$ be the offset location from the parent's center,
             * \f$x,y,z \in \{-1,1\}\f$, then the index is computed as:
             * \f[ index = \frac{x+1}{2} + 2\frac{y+1}{2} + 4\frac{z+1}{2} \f]
             *
             * \invariant \c m_children[i] \f$= 0 \Leftrightarrow\f$ The object doesn't have
             * an octant with index \e i.
             */
            OctreeNode *m_children[8];

            /**
             * \brief The node's parent node.
             *
             * \invariant Under the condition that the node isn't the root node, 
             *            <tt>m_parent != 0</tt>.
             *
             * \remarks The root node's \c m_parent pointer is the zero pointer.
             */
            OctreeNode *m_parent;


            unsigned int m_enclosed;

            
            PointContainer *m_points;


        protected:
            void render_recursive(
                CameraNode *camera,
                const int order[],
                const Vector3r &camera_pos) const throw();

            void render_points(
                CameraNode *camera, 
                const Vector3r &camera_pos) const throw();

            IntersectionTestResult frustum_test(
                const BoundingSphere &bsphere,
                const Plane *frustum_planes,
                int &plane_mask) const throw();



        public:

            /**
             * \brief Constructs an octree node, with the given parameters.
             *
             * Initializes members to:
             * - <tt>m_parent = parent</tt>
             * - <tt>m_half_size = half_size</tt>
             * - <tt>m_pos = pos</tt>
             * - <tt>m_enclosed = 0</tt>
             * - <tt>m_children[i] = 0</tt>, \f$ \forall i \f$
             * - <tt>m_contained_objects = </tt>\f$\emptyset\f$
             * - <tt>m_min_node_size = min_node_size</tt>
             * - <tt>m_child_count = 0</tt>
             *
             * \param parent The parent of the object.
             * \param pos The position of the center point of the octant, in
             *        world coordinates.
             * \param min_node_size The minimum size of an octree node.
             */
            OctreeNode(OctreeNode *parent,
                       real32 size,
                       const Vector3r &pos);

            /**
             * \brief Destruct the node, and all it's children.
             *
             * Will destroy any resources allocated by the node, and call
             * the destructor of all its child nodes.
             *
             * \pre The invariant documented at \c m_children must hold.
             * \post The resources used by the subtree in its entirety have
             *       been released.
             *
             * \remarks The parent node's child pointers will \e not be
             * automatically updated.
             */
            ~OctreeNode();

            /**
             * \brief Add an object to the node's subtree.
             *
             * May subdivide the node further, if it isn't already.
             * \e Never subdivdes the node into octants of half size smaller 
             * than m_min_node_size.
             *
             * \pre The object must be enclosed in the node, i.e. 
             *      \c encloses(obj) must be \c true.
             * \post The object has been added to the octree.
             *
             * \param obj The object to be added.
             * \return A pointer to the octree node the object was added to 
             *         (is contained in).
             */
            OctreeNode *add_object(
                const Vector3r &p, const RGBColor &color, 
                bool update_enclosed_count = true) throw();

            /**
             * \brief Check if an object is enclosed in the octree node.
             *
             * \return \c true if the node encloses <tt>obj</tt>'s entire BV, 
             *         \c false otherwise.
             */
            bool encloses(const Vector3r &p) const throw();

            /**
             * \brief Check if an object is enclosed the node's <em>i</em>:th 
             *        octant.
             *
             * \param obj The node to check for being enclosed by the octant.
             * \param octant The octant to check against, in the range 0 to 7.
             * \return \c true if the object is enclosed by the node's 
             *         <em>i</em>:th octant, \c false otherwise.
             *
             * \remarks Note that this doesn't check if the object is enclosed
             * by a child node, or even demands that any child nodes exist.
             * \sa \c encloses(const SceneNode *), \c m_children (for octant 
             * indexing).
             */
            bool enclosed_in_octant(const Vector3r &p, int octant) const throw();
         
            /**
             * \brief Returns whether the node contains any points or not.
             */
            inline bool is_empty() const throw() { return m_enclosed == 0; }

            /**
             * \brief Get a child octants position relative to its parent.
             *
             * \pre \f$ 0 \leq octant < 8 \f$
             *
             * \param octant The octant for which to get the offset from its 
             * parent node.
             * \return A vector describing the offset of node i from its parent.
             *
             * \sa \c m_children (for octant indexing).
             */
            Vector3r octant_offset(int octant) const throw();

            /**
             * \brief Get the node's parent node.
             *
             * \return The node's parent, 0 if none. 
             */
            OctreeNode *get_parent() throw();

            /**
             * \brief Get the node's parent node.
             *
             * \return The node's parent, 0 if none. 
             */
            const OctreeNode *get_parent() const throw();

            const real32 get_node_size() const throw();

            void render(CameraNode *camera, 
                        const int order[],
                        const BoundingSphere &frustum_bsphere,
                        const Plane *frustum_planes,
                        const Vector3r &camera_pos,
                        int plane_mask) const throw();
        };


        OctreeNode *grow_to_accomodate(const Vector3r &p) throw();


        OctreeNode *m_root;

        real32 m_max_point_dist;


    protected:
        virtual void calculate_bvs() throw();

    public:

        PointSetNode() throw();

        virtual ~PointSetNode() throw();

        PointSetNode *clone_cullable() const throw()
        {
            assert( false );
            return 0;
        }


        /**
         * \brief Add an object to the octree.
         *
         * If the object already exists in the octree this operation will
         * do nothing.
         *
         * \param obj The object to be inserted.
         *
         * \remarks Worst case running time of this operation can be expected
         * to be linear.
         */
        void add_object(const Vector3r &p, const RGBColor &color) throw();

        /**
         * \brief Remove an object from the octree.
         *
         * If the object isn't in the octree this operation will do nothing.
         *
         * \param obj The object to be removed.
         */
        //void remove_object(const Vector3r &p, const RGBColor &color);
        
       
        /*int get_size() const throw()
        {
            return m_size;
            }*/

        // PointSetNodes are broken!!! The work done by render() should be done 
        // in get_renderables(), or handeled diffently altogether
        virtual void render(primitives::Camera *camera) throw();

        virtual void render() throw() {}

        virtual void get_renderables(
            primitives::Camera *camera, EnumerationCallback callback) throw() {}
    };
#endif

}
}


#endif // __PEEKABOT_RENDERER_POINT_SET_NODE_HH
