/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "PrimitiveBatch.hh"
#include "VertexBuffer.hh"
#include "Mesh.hh"

#include <iostream>


using namespace peekabot;
using namespace peekabot::renderer;


PrimitiveBatch::Chunk::Chunk(size_t chunk_size)
    : m_avail(chunk_size),
      m_vertices(new VertexBuffer(true)), 
      m_colors(new VertexBuffer(true)) // Don't use VBOs, store in mem instead
{
}



PrimitiveBatch::PrimitiveBatch(size_t chunk_size, RenderContext &context)
    : m_chunk_size(chunk_size),
      m_points(chunk_size),
      m_lines(chunk_size),
      m_unlit_tris(chunk_size),
      m_render_context(context)
{
    get_state().set(new statelets::Lighting(false));
    get_state().set(new statelets::Transformation(Matrix4r::identity()));

    float *tmp = new float[chunk_size*3*3];
    
    try
    {
        m_points.m_vertices->set_data(
            tmp, chunk_size*3*sizeof(float), VertexBuffer::STREAM_DRAW);
        m_points.m_colors->set_data(
            tmp, chunk_size*3*sizeof(float), VertexBuffer::STREAM_DRAW);

        m_lines.m_vertices->set_data(
            tmp, chunk_size*3*2*sizeof(float), VertexBuffer::STREAM_DRAW);
        m_lines.m_colors->set_data(
            tmp, chunk_size*3*2*sizeof(float), VertexBuffer::STREAM_DRAW);

        m_unlit_tris.m_vertices->set_data(
            tmp, chunk_size*3*3*sizeof(float), VertexBuffer::STREAM_DRAW);
        m_unlit_tris.m_colors->set_data(
            tmp, chunk_size*3*3*sizeof(float), VertexBuffer::STREAM_DRAW);
    }
    catch(...)
    {
        delete[] tmp;
        throw;
    }

    delete[] tmp;
}


void PrimitiveBatch::render_point(
    const Vector3f &v, 
    const RGBColor &color) throw()
{
    if( m_points.m_avail < 1 )
        flush_points();

    size_t used = m_chunk_size - m_points.m_avail;

    float *vb = static_cast<float *>(
        m_points.m_vertices->map(VertexBuffer::WRITE_ONLY));
    vb[3*used+0] = v(0);
    vb[3*used+1] = v(1);
    vb[3*used+2] = v(2);
    m_points.m_vertices->unmap();

    float *cb = static_cast<float *>(
        m_points.m_colors->map(VertexBuffer::WRITE_ONLY));
    cb[3*used+0] = color.r;
    cb[3*used+1] = color.g;
    cb[3*used+2] = color.b;
    m_points.m_colors->unmap();

    --m_points.m_avail;
}


void PrimitiveBatch::render_line(
    const Vector3f &v1, 
    const Vector3f &v2,
    const RGBColor &color) throw()
{
    if( m_lines.m_avail < 1 )
        flush_lines();

    size_t used = m_chunk_size - m_lines.m_avail;

    float *vb = static_cast<float *>(
        m_lines.m_vertices->map(VertexBuffer::WRITE_ONLY));
    vb[2*3*used+0] = v1(0);
    vb[2*3*used+1] = v1(1);
    vb[2*3*used+2] = v1(2);
    vb[2*3*used+3] = v2(0);
    vb[2*3*used+4] = v2(1);
    vb[2*3*used+5] = v2(2);
    m_lines.m_vertices->unmap();

    float *cb = static_cast<float *>(
        m_lines.m_colors->map(VertexBuffer::WRITE_ONLY));
    cb[2*3*used+0] = color.r;
    cb[2*3*used+1] = color.g;
    cb[2*3*used+2] = color.b;
    cb[2*3*used+3] = color.r;
    cb[2*3*used+4] = color.g;
    cb[2*3*used+5] = color.b;
    m_lines.m_colors->unmap();

    --m_lines.m_avail;
}


void PrimitiveBatch::render_unlit_tri(
    const Vector3f &v1, 
    const Vector3f &v2,
    const Vector3f &v3,
    const RGBColor &color) throw()
{
    if( m_unlit_tris.m_avail < 1 )
        flush_unlit_tris();

    size_t used = m_chunk_size - m_unlit_tris.m_avail;

    float *vb = static_cast<float *>(
        m_unlit_tris.m_vertices->map(VertexBuffer::WRITE_ONLY));
    vb[3*3*used+0]  = v1(0);
    vb[3*3*used+1]  = v1(1);
    vb[3*3*used+2]  = v1(2);
    vb[3*3*used+3]  = v2(0);
    vb[3*3*used+4]  = v2(1);
    vb[3*3*used+5]  = v2(2);
    vb[3*3*used+6]  = v3(0);
    vb[3*3*used+7]  = v3(1);
    vb[3*3*used+8]  = v3(2);
    m_unlit_tris.m_vertices->unmap();

    float *cb = static_cast<float *>(
        m_unlit_tris.m_colors->map(VertexBuffer::WRITE_ONLY));
    cb[3*3*used+0]  = color.r;
    cb[3*3*used+1]  = color.g;
    cb[3*3*used+2]  = color.b;
    cb[3*3*used+3]  = color.r;
    cb[3*3*used+4]  = color.g;
    cb[3*3*used+5]  = color.b;
    cb[3*3*used+6]  = color.r;
    cb[3*3*used+7]  = color.g;
    cb[3*3*used+8]  = color.b;
    m_unlit_tris.m_colors->unmap();

    --m_unlit_tris.m_avail;
}


void PrimitiveBatch::render_unlit_quad(
    const Vector3f &v1, 
    const Vector3f &v2,
    const Vector3f &v3,
    const Vector3f &v4,
    const RGBColor &color) throw()
{
    if( m_unlit_tris.m_avail < 2 )
        flush_unlit_tris();

    size_t used = m_chunk_size - m_unlit_tris.m_avail;

    float *vb = static_cast<float *>(
        m_unlit_tris.m_vertices->map(VertexBuffer::WRITE_ONLY));

    vb[3*3*used+0]  = v1(0);
    vb[3*3*used+1]  = v1(1);
    vb[3*3*used+2]  = v1(2);
    vb[3*3*used+3]  = v2(0);
    vb[3*3*used+4]  = v2(1);
    vb[3*3*used+5]  = v2(2);
    vb[3*3*used+6]  = v3(0);
    vb[3*3*used+7]  = v3(1);
    vb[3*3*used+8]  = v3(2);

    vb[3*3*(used+1)+0]  = v3(0);
    vb[3*3*(used+1)+1]  = v3(1);
    vb[3*3*(used+1)+2]  = v3(2);
    vb[3*3*(used+1)+3]  = v4(0);
    vb[3*3*(used+1)+4]  = v4(1);
    vb[3*3*(used+1)+5]  = v4(2);
    vb[3*3*(used+1)+6]  = v1(0);
    vb[3*3*(used+1)+7]  = v1(1);
    vb[3*3*(used+1)+8]  = v1(2);

    m_unlit_tris.m_vertices->unmap();

    float *cb = static_cast<float *>(
        m_unlit_tris.m_colors->map(VertexBuffer::WRITE_ONLY));

    cb[3*3*used+0]  = color.r;
    cb[3*3*used+1]  = color.g;
    cb[3*3*used+2]  = color.b;
    cb[3*3*used+3]  = color.r;
    cb[3*3*used+4]  = color.g;
    cb[3*3*used+5]  = color.b;
    cb[3*3*used+6]  = color.r;
    cb[3*3*used+7]  = color.g;
    cb[3*3*used+8]  = color.b;

    cb[3*3*(used+1)+0]  = color.r;
    cb[3*3*(used+1)+1]  = color.g;
    cb[3*3*(used+1)+2]  = color.b;
    cb[3*3*(used+1)+3]  = color.r;
    cb[3*3*(used+1)+4]  = color.g;
    cb[3*3*(used+1)+5]  = color.b;
    cb[3*3*(used+1)+6]  = color.r;
    cb[3*3*(used+1)+7]  = color.g;
    cb[3*3*(used+1)+8]  = color.b;

    m_unlit_tris.m_colors->unmap();

    m_unlit_tris.m_avail -= 2;
}


void PrimitiveBatch::flush() throw()
{
    flush_points();
    flush_lines();
    flush_unlit_tris();
}


void PrimitiveBatch::render(RenderContext &context) throw()
{
    flush();
}


void PrimitiveBatch::flush_points() throw()
{
    if( m_chunk_size - m_points.m_avail > 0 )
    {
        PointMesh m(m_points.m_vertices, m_points.m_colors, 
                    m_chunk_size - m_points.m_avail);
        m.get_state().apply(&get_state());
        m.render(m_render_context);
        
        m_points.m_avail = m_chunk_size;
    }
}


void PrimitiveBatch::flush_lines() throw()
{
    if( m_chunk_size - m_lines.m_avail > 0 )
    {
        LineMesh m(m_lines.m_vertices, m_lines.m_colors, 
                   boost::shared_ptr<VertexBuffer>(),
                   boost::shared_ptr<IndexBuffer>(),
                   0, m_chunk_size - m_lines.m_avail);
        m.get_state().apply(&get_state());
        m.render(m_render_context);
        
        m_lines.m_avail = m_chunk_size;
    }
}


void PrimitiveBatch::flush_unlit_tris() throw()
{
    if( m_chunk_size - m_unlit_tris.m_avail > 0 )
    {
        TriMesh m(m_unlit_tris.m_vertices, 
                  boost::shared_ptr<VertexBuffer>(),
                  m_unlit_tris.m_colors, 
                  boost::shared_ptr<VertexBuffer>(),
                  boost::shared_ptr<IndexBuffer>(),
                  0, m_chunk_size - m_unlit_tris.m_avail);

        // Disable backface culling
        m.get_state().set(new statelets::Culling(false));

        m.get_state().apply(&get_state());
        m.render(m_render_context);
        
        m_unlit_tris.m_avail = m_chunk_size;
    }
}


