/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/scoped_ptr.hpp>
#include "PrimitiveFactory.hh"
#include "MeshFactory.hh"
#include "primitives/MeshBased.hh"
#include "primitives/Circle.hh"


using namespace peekabot;
using namespace peekabot::renderer;


PrimitiveFactory::PrimitiveFactory() throw()
{
    generate_boilerplates();
}

void PrimitiveFactory::generate_boilerplates() throw()
{
    //
    // Disc
    //
    primitives::MeshBased *disc = new primitives::MeshBased(
        4, BoundingSphere(Vector3r(0,0,0), 1.0f), 0.1f);

    boost::scoped_ptr<Mesh> disc1(MeshFactory::get_disc_mesh( 8, 1.0));
    boost::scoped_ptr<Mesh> disc2(MeshFactory::get_disc_mesh(16, 1.0));
    boost::scoped_ptr<Mesh> disc3(MeshFactory::get_disc_mesh(32, 1.0));
    boost::scoped_ptr<Mesh> disc4(MeshFactory::get_disc_mesh(64, 1.0));
    disc->set_lod(0,   0, disc1.get());
    disc->set_lod(1,  15, disc2.get());
    disc->set_lod(2,  75, disc3.get());
    disc->set_lod(3, 400, disc4.get());

    add_boilerplate(DISC, disc);

    //
    // Cylinder
    //
    primitives::MeshBased *cylinder = new primitives::MeshBased(
        4, BoundingSphere(Vector3r(0,0,0), 1.0f), 0.1f);

    boost::scoped_ptr<Mesh> cyl1(MeshFactory::get_cylinder_mesh( 8, 1.0, 1));
    boost::scoped_ptr<Mesh> cyl2(MeshFactory::get_cylinder_mesh(16, 1.0, 1));
    boost::scoped_ptr<Mesh> cyl3(MeshFactory::get_cylinder_mesh(32, 1.0, 1));
    boost::scoped_ptr<Mesh> cyl4(MeshFactory::get_cylinder_mesh(64, 1.0, 1));
    cylinder->set_lod(0,   0, cyl1.get());
    cylinder->set_lod(1,  20, cyl2.get());
    cylinder->set_lod(2, 150, cyl3.get());
    cylinder->set_lod(3, 450, cyl4.get());

    add_boilerplate(CYLINDER, cylinder);

    //
    // Ico-sphere
    //
    primitives::MeshBased *icosphere = new primitives::MeshBased(
        4, BoundingSphere(Vector3r(0,0,0), 1.0f), 0.1f);

    boost::scoped_ptr<Mesh> sphere1(MeshFactory::get_icosphere_mesh(0, 1.0));
    boost::scoped_ptr<Mesh> sphere2(MeshFactory::get_icosphere_mesh(1, 1.0));
    boost::scoped_ptr<Mesh> sphere3(MeshFactory::get_icosphere_mesh(2, 1.0));
    boost::scoped_ptr<Mesh> sphere4(MeshFactory::get_icosphere_mesh(3, 1.0));
    icosphere->set_lod(0,   0, sphere1.get());
    icosphere->set_lod(1,  10, sphere2.get());
    icosphere->set_lod(2,  35, sphere3.get());
    icosphere->set_lod(3, 150, sphere4.get());
    
    add_boilerplate(ICOSPHERE, icosphere);

    //
    // Circle
    //
    primitives::MeshBased *circle = new primitives::Circle(
        4, BoundingSphere(Vector3r(0,0,0), 1.0f), 0.1f);

    boost::scoped_ptr<Mesh> circle1(MeshFactory::get_circle_mesh(8, 1.0));
    boost::scoped_ptr<Mesh> circle2(MeshFactory::get_circle_mesh(16, 1.0));
    boost::scoped_ptr<Mesh> circle3(MeshFactory::get_circle_mesh(32, 1.0));
    boost::scoped_ptr<Mesh> circle4(MeshFactory::get_circle_mesh(64, 1.0));
    circle->set_lod(0,   0, circle1.get());
    circle->set_lod(1,  10, circle2.get());
    circle->set_lod(2,  70, circle3.get());
    circle->set_lod(3, 200, circle4.get());
    
    add_boilerplate(CIRCLE, circle);
}


PrimitiveFactory::~PrimitiveFactory() throw()
{
    for( PrimitiveMap::iterator it = m_boilerplates.begin();
         it != m_boilerplates.end(); ++it )
    {
        delete it->second;
    }

    m_boilerplates.clear();
}


primitives::MeshBased *PrimitiveFactory::get_disc() const throw()
{
    PrimitiveMap::const_iterator it = m_boilerplates.find(DISC);
    return it->second->clone();
}

primitives::MeshBased *PrimitiveFactory::get_cylinder() const throw()
{
    PrimitiveMap::const_iterator it = m_boilerplates.find(CYLINDER);
    return it->second->clone();
}

primitives::MeshBased *PrimitiveFactory::get_icosphere() const throw()
{
    PrimitiveMap::const_iterator it = m_boilerplates.find(ICOSPHERE);
    return it->second->clone();
}

primitives::MeshBased *PrimitiveFactory::get_circle() const throw()
{
    PrimitiveMap::const_iterator it = m_boilerplates.find(CIRCLE);
    return it->second->clone();
}

/*primitives::MeshBased *PrimitiveFactory::get_cube() const throw()
{
    PrimitiveMap::const_iterator it = m_boilerplates.find(CUBE);
    return it->second->clone();
}*/

void PrimitiveFactory::add_boilerplate(
    MeshType type, primitives::MeshBased *boilerplate) throw()
{
    m_boilerplates.insert(std::make_pair(type, boilerplate));
}
