/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <GL/glew.h>
#include <list>
#include <cassert>
#include <boost/bind.hpp>
#include <iostream>

#include "../Configuration.hh"
#include "../MessageHub.hh"
#include "../Bytesex.hh"
#include "RenderEngine.hh"
#include "SceneAction.hh"
#include "FrameCollector.hh"
#include "RenderTraverser.hh"
#include "IDTranslator.hh"

#include "VertexBuffer.hh"
#include "IndexBuffer.hh"
#include "State.hh"
#include "primitives/Camera.hh"
#include "TextureManager.hh"
#include "Texture2D.hh"
#include "PrimitiveBatch.hh"
#include "RenderContext.hh"


#include "PointSetNode.hh"


using namespace peekabot;
using namespace peekabot::renderer;



namespace {
    const RGBColor DEFAULT_BACKGROUND_COLOR(1, 1, 1);
    const RGBColor DEFAULT_SELECTED_COLOR(0, 0.3, 1);
    const RGBColor DEFAULT_PARENT_SELECTED_COLOR(0.5, 0.75, 1);
    
    
    const Axis AXIS_MASK[] = { X_AXIS, Y_AXIS, Z_AXIS };

    const RGBColor AXIS_COLOR[3] = { 
        RGBColor(0.7, 0.2, 0.2), 
        RGBColor(0.2, 0.7, 0.2), 
        RGBColor(0.2, 0.2, 0.7)
    };
}


RenderEngine::RenderEngine(real_unit octree_size,
                           real_unit min_octree_node_size) throw()
    : m_octree(octree_size, min_octree_node_size),
      m_sync_trigger(&m_root_node, m_octree, this),
      m_select_trigger(&m_root_node),
      m_transformation_guides_mode(NO_TRANSFORMATION_GUIDES)
{
    update_configuration();
}


RenderEngine::~RenderEngine() throw()
{
    {
        // It should be safe to access the action queue unlocked now, but hey...
        // ...you know what the three first letters of assumption spell :)
        boost::recursive_mutex::scoped_lock lock(m_action_queue_mutex);
        while( !m_action_queue.empty() )
        {
            delete m_action_queue.front();
            m_action_queue.pop();
        }
    }
}


void RenderEngine::update_configuration() throw()
{
    boost::mutex::scoped_lock lock(m_config_mutex);

    try
    {
        // Get color values
        m_config.m_bg_color = DEFAULT_BACKGROUND_COLOR;

        m_config.m_selected_color = DEFAULT_SELECTED_COLOR;

        m_config.m_parent_selected_color = DEFAULT_PARENT_SELECTED_COLOR;


        m_config.m_enable_multisampling = 
            TheConfiguration::instance().get<bool>(
                "renderer.enable_multisampling", true);

        /*if( has_context )
          Device::toggle_multisampling(m_config.m_enable_multisampling);*/
    }
    catch(std::runtime_error &e)
    {
        TheMessageHub::instance().publish(
            ERROR_MESSAGE,
            "Failed to update RenderEngine configuration: %s (%s:%u)", 
            e.what(), __FILE__, __LINE__);
    }
}


void RenderEngine::init()
{
    boost::recursive_mutex::scoped_lock lock(m_is_active_mutex);

    TheMessageHub::instance().publish(
        DEBUG_MESSAGE, 
        "Initializing GLEW...");
   
    // Setting this magical, undocumented flag will enable GLEW to report
    // that OpenGL >1.2 exists on Macs with crippled hardware (e.g. the
    // Mac Mini, which only supports OpenGL 1.2 hardware accelerated).
    //
    // Note that the extra set of GL functions it reports are OK after setting
    // this flag are software rendered through their LLVM stack!
    // But software rendering is better than nothing I guess?
    glewExperimental = GL_TRUE;

    // Initialize GLEW (Note that it is IMPORTANT that we have a valid
    // OpenGL context before trying to initialize GLEW).
    GLenum err = glewInit();
    if( GLEW_OK != err )
    {
        // Problem: glewInit failed, something is seriously wrong.
        throw std::runtime_error((const char *)glewGetErrorString(err));
    }


    // Log GL version, vendor, renderer and extensions
    TheMessageHub::instance().publish(
        LOG_MESSAGE, 
        "OpenGL vendor: %s", glGetString(GL_VENDOR));

    TheMessageHub::instance().publish(
        LOG_MESSAGE, 
        "OpenGL renderer: %s", glGetString(GL_RENDERER));

    TheMessageHub::instance().publish(
        LOG_MESSAGE, 
        "OpenGL version: %s", glGetString(GL_VERSION));

    TheMessageHub::instance().publish(
        LOG_MESSAGE, 
        "OpenGL extensions: %s", glGetString(GL_EXTENSIONS));


    // We depend on glBlendFunc and glBlendColor - they're not available
    // until OpenGL 1.4, or through the ARB_imaging extension available in 
    // OpenGL 1.2. Make sure we have either one or abort!
    if( !(GLEW_VERSION_1_4 || (GLEW_ARB_imaging && GLEW_VERSION_1_2)) )
    {
        throw std::runtime_error(
            "OpenGL 1.4 or OpenGL 1.2 and the ARB_imaging "
            "extension are required to run peekabot");
    }

    // If we don't have VBO support, print a nice little informational
    // message saying so...
    if( !Device::has_vbo_support() )
    {
        TheMessageHub::instance().publish(
            INFO_MESSAGE, 
            "VBO support not available");
    }

    init_gl_context();
}


void RenderEngine::init_gl_context() throw()
{
    TheMessageHub::instance().publish(
        DEBUG_MESSAGE, 
        "Performing initial configuration of the OpenGL context...");

    // The OpenGL implementation (probably) isn't thread safe...
    boost::recursive_mutex::scoped_lock lock(m_is_active_mutex);

    //
    // Dial in the right settings for a few states that aren't
    // handled by the statelet system (at the moment, at least)
    //
    glShadeModel(GL_SMOOTH);
    glClearStencil(0);
    glEnable(GL_LIGHT0);

    // We need this since we do scaling with matrices.
    glEnable(GL_NORMALIZE);

    Device::toggle_multisampling(m_config.m_enable_multisampling);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_DITHER);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
    //glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);

    glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
    glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);



    State baseline;
    
    // Enable lighting by default
    baseline.set(new statelets::Lighting(true));
    // Disable two-sided lighting by default
    baseline.set(new statelets::LightModel(false));
    // Disable blending by default
    baseline.set(new statelets::BlendMode(GL_ONE, GL_ZERO));
    // Enable backface culling by default
    baseline.set(new statelets::Culling(true));
    // Enable depth testing by default, using the GL_LESS as the depth func
    baseline.set(new statelets::DepthTest(statelets::DepthTest::LESS));
    // Enable depth mask writing by default
    baseline.set(new statelets::DepthMask(true));
    // Enable diffuse reflectance material color tracking
    baseline.set(new statelets::ColorMaterial(true));
    baseline.set(new statelets::PointParams(3, true));
    baseline.set(new statelets::LineParams(1, false, 0, 0));
    baseline.set(new statelets::PolygonMode(
                     statelets::PolygonMode::FRONT_AND_BACK,
                     statelets::PolygonMode::FILL));
    baseline.set(new statelets::StencilTest(false));
    baseline.set(new statelets::StencilFunc(
                     GL_ALWAYS, 1, static_cast<GLuint>(-1),
                     GL_KEEP, GL_KEEP, GL_REPLACE));
    // No texture by default
    baseline.set(new statelets::Texture2D());
    //TheTextureManager::instance().get("coffee_beans.png"));

    // Default to disabling all vertex buffers
    baseline.set(
        new statelets::VertexPointer(boost::shared_ptr<VertexBuffer>()));
    baseline.set(
        new statelets::NormalPointer(boost::shared_ptr<VertexBuffer>()));
    baseline.set(
        new statelets::ColorPointer(boost::shared_ptr<VertexBuffer>()));
    baseline.set(
        new statelets::TexCoordPointer(boost::shared_ptr<VertexBuffer>()));

    // Default material:
    baseline.set(
        new statelets::Material(
            RGBAColor(0.8,0.8,0.8,1),
            RGBAColor(0.2,0.2,0.2,1),
            RGBAColor(0,0,0,1),
            20));

    State::set_baseline(baseline);
}


void RenderEngine::queue_action(SceneAction *action) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_action_queue_mutex);
    
    m_action_queue.push(action);
}


uint32_t RenderEngine::pending_action_count() throw()
{
    boost::recursive_mutex::scoped_lock lock(m_action_queue_mutex);

    return m_action_queue.size();
}


void RenderEngine::process_actions() throw()
{
    on_process_actions(this);
    
    {
        boost::recursive_mutex::scoped_lock lock1(m_action_queue_mutex);
        // Lock render/pick calls out until we're done
        boost::recursive_mutex::scoped_lock lock2(m_is_active_mutex);
        
        while( !m_action_queue.empty() )
        {
            SceneAction *action = m_action_queue.front();
            m_action_queue.pop();
            
            try
            {
                action->execute(this);
            }
            catch(std::exception &e)
            {
                TheMessageHub::instance().publish(
                    ERROR_MESSAGE,
                    "SceneAction threw an exception with message '" + 
                    std::string(e.what()) + "' during execution (%s:%u)",
                    __FILE__, __LINE__);
            }
            catch(...)
            {
                TheMessageHub::instance().publish(
                    ERROR_MESSAGE,
                    "SceneAction threw an unknown exception during execution (%s:%u)",
                    __FILE__, __LINE__);
            }
            
            delete action;
        }
    }

    on_actions_processed(this);
}

primitives::Camera *RenderEngine::get_camera(PrimitiveID id) throw()
{
    try
    {
        IDTranslator::Primitive p = TheIDTranslator::instance().get_primitive(id);
        primitives::Camera *cam = dynamic_cast<primitives::Camera *>(*p.begin());

        if( cam )
            return cam;
    }
    catch(...) {}

    TheMessageHub::instance().publish(
        WARNING_MESSAGE,
        "The requested camera (id=%u) was not found!", id);
    
    return 0;
}


void RenderEngine::ready_scene(const Configuration &config) throw()
{
    // 1. Process changes to the scene
    process_actions();
}


void RenderEngine::set_is_rendering(bool is_rendering) throw()
{
    boost::mutex::scoped_lock lock(m_active_flags_mutex);
    m_is_active = is_rendering;
    m_is_rendering = is_rendering;
}


void RenderEngine::set_is_active(bool is_active) throw()
{
    boost::mutex::scoped_lock lock(m_active_flags_mutex);
    m_is_active = is_active;
    m_is_rendering = is_active;
}


RenderEngine::Configuration RenderEngine::clone_configuration() throw()
{
    boost::mutex::scoped_lock lock(m_config_mutex);

    return m_config;
}



// 1. Ready the scene, i.e.:
//    * Incorporate pending scene modifications
//    * Synchronize the scene and the octree
// 2. Generate the potentially visible set (PVS) of objects through traversing
//    the octree in rough back to front order, culling objects outside the 
//    frustum.
// 3. TODO ...
//
bool RenderEngine::render(int w, int h, PrimitiveID camera_id, 
                          bool *visible_layers,
                          FrameCollector *collector) throw()
{
    if( w < 1 || h < 1 || visible_layers == 0 )
        return true;
    
    boost::recursive_mutex::scoped_lock lock(m_is_active_mutex);

    // Prevent further render/pick calls during our render pass
    set_is_rendering(true);

    // Make a copy of the engine configuration
    Configuration engine_config = clone_configuration();


    ready_scene(engine_config);
    // The scene and octree and now in sync

    primitives::Camera *camera = get_camera(camera_id);
    if( !camera ) // No such camera
        return true;

    on_render(this);


    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);


    // Setup viewport
    camera->set_viewport(0, 0, w, h);



    // Clear color, depth and stencil buffers
    glClearColor(engine_config.m_bg_color.r, 
                 engine_config.m_bg_color.g, 
                 engine_config.m_bg_color.b, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

    // Set up the viewing transform and projection matrix
    use_camera(camera);


    // We're not doing picking, ignore object names
    State::OverrideSet os(State::get_override_set());
    os.insert(statelets::Name::get_type());
    State::push_baseline();
    State::set_baseline(State::get_baseline_state(), os);
    // Unconditionally reapply the baseline state
    State::apply_baseline();


    RenderTraverser traverser(engine_config.m_selected_color, 
                              engine_config.m_parent_selected_color);
    traverser.collect(m_octree, camera, visible_layers);
    traverser.render_opaque();
    if( m_transformation_guides_mode != NO_TRANSFORMATION_GUIDES )
        render_transformation_guides(camera);
    traverser.render_translucent();
    traverser.render_outlines();
    traverser.render_always_on_top();
    traverser.finish();

    State::pop_baseline();

    
    if( collector && w > 0 && h > 0 )
        collect_frame(collector, w, h);


    set_is_rendering(false);

    on_rendered(this);


    return false;
}



uint32_t RenderEngine::pick(int w, int h,
                            real_unit x, real_unit y, 
                            PrimitiveID camera_id, 
                            bool *visible_layers, 
                            real_unit sensitivity) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_is_active_mutex);

    on_pick(this);

    // Prevent further render/pick calls during our render pass
    set_is_active(true);

    // Make a copy of the engine configuration
    Configuration engine_config = clone_configuration();


    ready_scene(engine_config);
    // The scene and octree and now in sync


    primitives::Camera *camera = get_camera(camera_id);
    if( !camera ) // No such camera
        return -1;


    // Setup viewport
    camera->set_viewport(0, 0, w, h);

  
    glMatrixMode(GL_PROJECTION);
    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    glLoadIdentity();
    // OpenGL assumes that the origin is in the lower left 
    // corner of the screen, FLTK does not...
    gluPickMatrix(x, h-y-1, sensitivity, sensitivity, viewport);
    glMatrixMode(GL_MODELVIEW);


    GLuint hit_records[512];
    GLint hits;

    glSelectBuffer(sizeof(hit_records), hit_records);
    glRenderMode(GL_SELECT);
    glInitNames();
    glPushName(0);

    //
    // Render objects in the view frustum
    //

    // Set up the viewing transform and projection matrix
    use_camera(camera);


    // We don't care about colors, materials, textures, ...
    State::OverrideSet os(State::get_override_set());
    os.insert(statelets::Texture2D::get_type());
    os.insert(statelets::Material::get_type());
    os.insert(statelets::ColorMaterial::get_type());
    os.insert(statelets::Color::get_type());
    os.insert(statelets::Alpha::get_type());
    os.insert(statelets::BlendMode::get_type());
    os.insert(statelets::ColorPointer::get_type());
    os.insert(statelets::TexCoordPointer::get_type());
    os.insert(statelets::NormalPointer::get_type());
    os.insert(statelets::Lighting::get_type());
    os.insert(statelets::StencilTest::get_type());
    os.insert(statelets::StencilFunc::get_type());
    State::push_baseline();
    State::set_baseline(State::get_baseline_state(), os);
    // Unconditionally reapply the baseline state
    State::apply_baseline();


    RenderTraverser traverser(engine_config.m_selected_color, 
                              engine_config.m_parent_selected_color);
    traverser.collect(m_octree, camera, visible_layers);
    traverser.render_opaque();
    traverser.render_translucent();
    traverser.render_always_on_top();
    traverser.finish();

    State::pop_baseline();


    //
    // Examine the results
    //
    hits = glRenderMode(GL_RENDER);


    // Extract the picked ID
    uint32_t picked_id = 0;
    GLuint picked_id_x = 0;
    for( int i = 0; i < hits; ++i )
    {
        GLuint name = hit_records[4*i + 3];
        if( name != 0 )
        {
            if( hit_records[4*i+1] < picked_id_x || picked_id == 0 )
            {
                picked_id = name;
                picked_id_x = hit_records[4*i+1];
            }
        }
    }



    set_is_active(false);

    on_picked(this);

    // TODO
    // Clone camera or back up its settings.
    // Change to ortho mode, with a unit cube frustum scaled by sensitivity
    // ...
    return picked_id;
}





void RenderEngine::use_camera(primitives::Camera *camera) throw()
{   
    // Load the projection matrix
    glMatrixMode(GL_PROJECTION);
    camera->apply_projection_matrix();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    Matrix4r camera_matrix = Matrix4r(
            Vector4r( 0,  0, -1,  0),
            Vector4r(-1,  0,  0,  0),
            Vector4r( 0,  1,  0,  0),
            Vector4r( 0,  0,  0,  1)) * camera->get_viewer_mtow().inverse();
    glLoadMatrixf( 
        reinterpret_cast<const GLfloat *>(&camera_matrix(0,0)) ); // TODO: change to something that doesn't break when the unit we use changes

    // Set the light direction in *world* coordinates
    Vector4r pos(1, -1, 2, 0);
    glLightfv(GL_LIGHT0, GL_POSITION, &pos(0));

    Vector4r ambient(0.1, 0.1, 0.1, 1);
    glLightfv(GL_LIGHT0, GL_AMBIENT, &ambient(0));

    //Vector4r specular(0.5, 0.5, 0.5, 1);
    //glLightfv(GL_LIGHT0, GL_SPECULAR, &specular(0));
    //
    //Vector4r diffuse(0.9, 0.9, 0.9, 1);
    //glLightfv(GL_LIGHT0, GL_DIFFUSE, &diffuse(0));
}


void RenderEngine::collect_frame(
    FrameCollector *collector, int w, int h) const throw()
{
    uint8_t pixbuf[w*h*3];
    
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    glReadBuffer(GL_BACK);
    
    if( collector->get_pixel_format() == FrameCollector::RGB )
        glReadPixels(0, 0, w, h, GL_RGB, GL_UNSIGNED_BYTE, pixbuf);
    else if( collector->get_pixel_format() == FrameCollector::BGR )
        glReadPixels(0, 0, w, h, GL_BGR, GL_UNSIGNED_BYTE, pixbuf);
    else
        TheMessageHub::instance().publish(
            WARNING_MESSAGE,
            "Unsupported FrameCollector pixel format used, %s (%s:%u)",
            "the buffer contents will be undefined", __FILE__, __LINE__);
    
    
    collector->collect_frame(w, h, pixbuf);
}


bool RenderEngine::is_active() throw()
{
    if( m_is_active )
        return m_is_active;
    else
    {
        boost::mutex::scoped_lock lock(m_active_flags_mutex);
        return m_is_active;
    }
}


bool RenderEngine::is_rendering() throw()
{
    if( m_is_rendering )
        return m_is_rendering;
    else
    {
        boost::mutex::scoped_lock lock(m_active_flags_mutex);
        return m_is_rendering;
    }
}


SceneNode *RenderEngine::get_root_node() throw()
{
    return &m_root_node;
}


void RenderEngine::render_transformation_guides(
    primitives::Camera *camera) const throw()
{
    if( m_transformation_guides_mode == MOVE_TRANSFORMATION_GUIDES )
        render_move_transformation_guides(camera);
    else if( m_transformation_guides_mode == ROTATE_TRANSFORMATION_GUIDES )
        render_rotate_transformation_guides(camera);
}


void RenderEngine::render_move_transformation_guides(
    primitives::Camera *camera) const throw()
{
    static const Matrix4f eye = Matrix4f::identity();

    const SelectTrigger::SelectedSet &selected = 
        m_select_trigger.get_selected();

    if( selected.empty() )
        return;

    RenderContext context(camera, 1);
    PrimitiveBatch &batch = context.get_batch(0);

    for( SelectTrigger::SelectedSet::const_iterator it = selected.begin();
         it != selected.end(); ++it )
    {
        const Vector4f &pos = (*it)->get_world_pos();

        for( int i = 0; i < 3; ++i )
        {
            if( m_transformation_guides_axes & AXIS_MASK[i] )
            {
                Vector3f v1, v2;
                
                switch( m_transformation_guides_coord_sys )
                {
                    case LOCAL_COORDINATES:
                        v1 = pos - (*it)->get_mtow()(i)*1000;
                        v2 = pos + (*it)->get_mtow()(i)*1000;
                        break;
                        
                    case PARENT_COORDINATES:
                        v1 = pos - (*it)->get_parent_mtow()(i)*1000;
                        v2 = pos + (*it)->get_parent_mtow()(i)*1000;
                        break;

                    case WORLD_COORDINATES:
                        v1 = pos - eye(i)*1000;
                        v2 = pos + eye(i)*1000;
                        break;

                    case CAMERA_COORDINATES:
                        v1 = pos - camera->get_parent()->get_mtow()(i)*1000;
                        v2 = pos + camera->get_parent()->get_mtow()(i)*1000;
                        break;

                    default:
                        assert( false );
                }

                batch.render_line(v1, v2, AXIS_COLOR[i]);
            }
        }
    }

    batch.flush();
}


void RenderEngine::render_rotate_transformation_guides(
    primitives::Camera *camera) const throw()
{
    static const Matrix4f eye = Matrix4f::identity();

    const SelectTrigger::SelectedSet &selected = 
        m_select_trigger.get_selected();

    if( selected.empty() )
        return;

    RenderContext context(camera, 2);
    PrimitiveBatch &solid = context.get_batch(0);
    PrimitiveBatch &dashed = context.get_batch(1);

    dashed.get_state().set(new statelets::LineParams(1, true, 6, 43690));
    solid.get_state().set(new statelets::PointParams(4, true));

    Vector4f avg_pos(0,0,0,0);

    if(  m_transformation_guides_pivot == AVERAGE_CENTER )
    {
        // Calculate the average position of the selected objects
        // Could be done more efficiently using a trigger... then we wouldn't 
        // have to recalculate it every frame
        for( SelectTrigger::SelectedSet::const_iterator it = selected.begin();
             it != selected.end(); ++it )
            avg_pos += (*it)->get_world_pos();

        avg_pos /= selected.size();
    }

    for( SelectTrigger::SelectedSet::const_iterator it = selected.begin();
         it != selected.end(); ++it )
    {
        const Vector4f *ppos = 0;
        if( m_transformation_guides_pivot == LOCAL_CENTER )
            ppos = &(*it)->get_world_pos();
        else if( m_transformation_guides_pivot == PARENT_CENTER )
            ppos = &(*it)->get_parent_mtow()(3);
        else if( m_transformation_guides_pivot == WORLD_CENTER )
            ppos = &m_root_node.get_world_pos();
        else // AVERAGE_CENTER
            ppos = &avg_pos;

        const Vector4f &pivot = *ppos;
        const Vector4f &pos = (*it)->get_world_pos();

        for( int i = 0; i < 3; ++i )
        {
            if( m_transformation_guides_axes & AXIS_MASK[i] )
            {
                Vector4f v1, v2, axis;
                
                if( m_transformation_guides_coord_sys == LOCAL_COORDINATES )
                    axis = (*it)->get_mtow()(i);
                else if( m_transformation_guides_coord_sys == PARENT_COORDINATES )
                    axis = (*it)->get_parent_mtow()(i);
                else if( m_transformation_guides_coord_sys == WORLD_COORDINATES )
                    axis = eye(i);
                else if( m_transformation_guides_coord_sys == CAMERA_COORDINATES )
                    axis = camera->get_parent()->get_mtow()(i);
                else
                    assert( false );

                v1 = pivot - axis*1000;
                v2 = pivot + axis*1000;

                if( m_transformation_guides_coord_sys != CAMERA_COORDINATES 
                    || i != 0 )
                    solid.render_line(v1, v2, AXIS_COLOR[i]);

                if( m_transformation_guides_pivot != LOCAL_CENTER )
                {
                    Vector4f tmp;
                    if( m_transformation_guides_coord_sys != 
                            CAMERA_COORDINATES || i != 0 )
                    {
                        Vector4f proj = axis * (((pivot-pos)*axis) / axis.norm());
                        tmp = pivot - proj;
                    }
                    else
                        tmp = pivot;

                    solid.render_point(tmp, RGBColor(0.2, 0.2, 0.2));
                    dashed.render_line(pos, tmp, RGBColor(0.2, 0.2, 0.2));
                }
            }
        }
    }

    solid.flush();
    dashed.flush();
}


void RenderEngine::set_transformation_guides_mode(
    TransformationGuidesMode mode) throw()
{
    m_transformation_guides_mode = mode;
}

void RenderEngine::set_transformation_guides_coord_sys(
    CoordinateSystem coord_sys) throw()
{
    m_transformation_guides_coord_sys = coord_sys;
}

void RenderEngine::set_transformation_guides_pivot(
    RotationCenterpoint pivot) throw()
{
    m_transformation_guides_pivot = pivot;
}

void RenderEngine::set_transformation_guides_axes(
    Axis axes) throw()
{
    m_transformation_guides_axes = axes;
}
