/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_RENDERER_RENDER_ENGINE_HH
#define __PEEKABOT_RENDERER_RENDER_ENGINE_HH


#include <string>
#include <queue>
#include <set>
#include <map>
#include <boost/signal.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/recursive_mutex.hpp>

#include "RendererTypes.hh"
#include "../Types.hh"
#include "../Configuration.hh"
#include "SceneNode.hh"
#include "Octree.hh"
#include "SyncTrigger.hh"
#include "SelectTrigger.hh"
#include "SceneAction.hh"
#include "IDTypes.hh"
#include "primitives/Camera.hh"


namespace peekabot
{
namespace renderer
{

    class FrameCollector;
    class RenderEngine;


    /**
     * \brief A self-contained scene and rendering environment.
     *
     * The core of the class's functionality is storing a graphical 
     * representation of a scene and providing methods of rendering it as well
     * selecting objects from the rendered image. This functionality is exposed
     * by a few simple methods: queue_action(SceneAction *) (used exclusively 
     * to manipulate the scene), pick(...) and render(...).
     *
     * In addition to these core functions it also provides a trigger system and
     * a method for setting up the OpenGL context. Before anything can be 
     * rendered or picked, the OpenGL must be initialized and set up!
     *
     * \section render_engine_internals Internal structure overview
     *
     * The daily life of the render subsystem typically involves the following
     * "chores":
     * \li Execute incoming SceneActions, incorporating changes from the outside
     *     environment into the subsystems internal representation of the world.
     * \li Update its internal auxillary structures to accomodate the changes
     *     made to its scene. This involves updating a spatial subdivision data 
     *     structure, used to remove hidden objects to speed up rendering.
     * \li Rendering and picking.
     *
     * The scene is stored in a Scene object, to which updates are made. A 
     * trigger registered on it, SyncTrigger, then takes care of updating our
     * spatial subdivision data structure, the Octree.
     *
     * Rendering and picking is carried out in collaboration with the Octree and
     * the RenderTraverser.
     *
     * \sa Renderer (the usually preferred public interface to RenderEngine)
     */
    class RenderEngine
    {
    public:
        /**
         * \brief Create and initialize the render engine.
         *
         * If the VBO subsystem hasn't already been initialized, this method 
         * will check for VBO support and perform the necessary initialization.
         *
         * \param octree_size The initial "half" size of the octree.
         * \param min_octree_node_size The minimum half size of node, 
         * subdivision will stop when this node size is reached.
         * 
         * \sa \c Octree
         */
        RenderEngine(real_unit octree_size = 100.0f,
                     real_unit min_octree_node_size = 1.0f) throw();


        ~RenderEngine() throw();


        void init();

        
        
        /**
         * \brief Render a frame.
         *
         * All changes to the scene queued since the last time \c render was 
         * called will be incorporated into the scene. Subsequently, the scene
         * will be rendered. Also, if \c save_snapshot(const std::string &) has
         * been executed after the previous call to \c render
         * exited, the rendered frame will be saved to hard disk, using the 
         * filename specified to the most recent call to
         * \c save_snapshot(const std::string &).
         *
         * If the camera specified does not exist, the operation will fail.
         *
         * During the duration of this call, \c is_active() will be set to
         * \c true.
         *
         * \pre OpenGL has been initialized and made the current context (by 
         * calling FLTK's \c make_current(), directly or indirectly).
         *
         * \param w The width of the canvas it renders to.
         * \param h The height of the canvas it renders to.
         * \param camera The camera used to render the scene.
         * \param visible_layers An array of \c NUMBER_OF_LAYERS booleans,
         * describing each layers visibility, where \c true means visible and 
         * \c false means it's hidden.
         * \return \c true if the operation fails, \c false otherwise.
         */
        bool render(int w, int h, PrimitiveID camera_id, 
                    bool *visible_layers,
                    FrameCollector *collector) throw();

        /**
         * \brief Return the object under the \f$<x,y>\f$ window coordinate.
         *
         * Will not incorporate pending changes into the scene before running
         * the picking pass, thus, any changes made to the world since the
         * last frame was drawn will not affect the outcome of the pick call.
         *
         * If several objects are drawn at the \f$<x,y>\f$ coordinate, the one
         * closest to the camera will be returned.
         *
         * During the duration of this call, \c is_active() will be set to
         * \c true.
         *
         * \param w The width of the canvas it renders to.
         * \param h The height of the canvas it renders to.
         * \param x The x component of the picking coordinate, in window 
         * coordinates.
         * \param y The y component of the picking coordinate, in window 
         * coordinates.
         * \param camera The camera used to render the scene.
         * \param visible_layers An array of \c NUMBER_OF_LAYERS booleans,
         * \return 0 if no object was under the cursor, otherwise the picked 
         * object's \c ObjectID.
         */
        uint32_t pick(int w, int h,
                    real_unit x, real_unit y, 
                    PrimitiveID camera_id, bool *visible_layers, 
                    real_unit sensitivity) throw();

        /**
         * \brief Queue a change to the scene.
         *
         * Pending scene changes are incorporated into the scene by 
         * calling \c process_actions() (done implicitly through render() and
         * pick(), or explicitly by calling process_actions()).
         */
        void queue_action(SceneAction *action) throw();

        /**
         * \brief Process all changes queued by \c queue_action(SceneAction *) 
         * since the last call to \c process_actions().
         *
         * Takes care of updating the render engine's internal scene 
         * representation (the \c Scene), as well as any supporting data 
         * structures (its octree).
         *
         * \em Important: Note that this method can \em only be when the caller holds
         * a valid OpenGL context.
         *
         * Calls \c SceneAction::execute(Scene *) for each queued action.
         * Each \c SceneAction object itself is responsible for making the 
         * actual changes to the data contained in the renderer, 
         * \c process_actions() merely directs and oversees the process.
         */
        void process_actions() throw();

        /**
         * \brief Returns the number scene actions, submitted through 
         * queue_action that lie waiting to be processed.
         */
        uint32_t pending_action_count() throw();


        /**
         * \brief Returns the state of the render engine, if it is active 
         * performing work or not.
         *
         * \remarks A potentially expensive operation, as it may have to
         * acquire a mutex lock before executing.
         */
        bool is_active() throw();

        /**
         * \brief Returns whether or not the rendering engine is active 
         * rendering.
         *
         * \note If the is_active() is \c true and is_rendering() false,
         * then the render engine is active running a picking pass.
         *
         * \remarks A potentially expensive operation, as it may have to
         * acquire a mutex lock before executing.
         */
        bool is_rendering() throw();

        /**
         * \brief Get the engine's scene's root SceneNode.
         *
         * \return The scene's root node.
         */
        SceneNode *get_root_node() throw();


        void set_transformation_guides_mode(TransformationGuidesMode mode) throw();

        void set_transformation_guides_coord_sys(CoordinateSystem coord_sys) throw();

        void set_transformation_guides_pivot(RotationCenterpoint pivot) throw();

        void set_transformation_guides_axes(Axis axes) throw();


    public:
        //
        // Signals
        //
        boost::signal<void (RenderEngine *)> on_process_actions;

        boost::signal<void (RenderEngine *)> on_actions_processed;

        boost::signal<void (RenderEngine *)> on_render;

        boost::signal<void (RenderEngine *)> on_rendered;

        boost::signal<void (RenderEngine *)> on_pick;

        boost::signal<void (RenderEngine *)> on_picked;



    private:
        struct Configuration
        {
            bool m_occlusion_culling_enabled;

            // maximum number of occluders to use for occlusion culling
            int m_num_occluders;

            // the minimum size for objects to use as occluders
            float m_min_occluder_size;
            
            // skip object if shown fragments < m_min_fragments
            int m_min_fragments;

            RGBColor m_bg_color;

            RGBColor m_selected_color;

            RGBColor m_parent_selected_color;

            RGBColor m_grid_color;

            bool m_enable_grid;

            bool m_enable_multisampling;
        };


        primitives::Camera *get_camera(PrimitiveID id) throw();

        /**
         * \brief Perform initial state configuration on a pristine context.
         *
         * Sets up certain parameters that are the same for all objects - 
         * i.e. use smooth shading, use "local viewer" for lighting 
         * calculations.
         */
        void init_gl_context() throw();

        void update_configuration() throw();


        /**
         * \brief Make the scene and its supporting structures ready to render.
         *
         * After this function has executed, it is guaranteed that:
         * - All pending scene actions have been incorporated into the scene.
         * - The octree is synchronized with the scene, ready to be used for 
         *   rendering.
         */
        void ready_scene(const Configuration &config) throw();


        void use_camera(primitives::Camera *camera) throw();
        
        /**
         * \brief Read the current contents of the frame buffer and call
         * the given FrameCollector with the result.
         *
         * \pre <tt>collector != 0</tt>
         *
         * \param collector The FrameCollector to call with the frame data.
         * \param w The width of the viewport used to generate the current 
         * frame.
         * \param h The height of the viewport used to generate the current 
         * frame.
         */
        void collect_frame(
            FrameCollector *collector, int w, int h) const throw();


        void set_is_rendering(bool is_rendering) throw();

        void set_is_active(bool is_active) throw();

        Configuration clone_configuration() throw();


        void render_grid(const Configuration &config) const throw();

        void render_transformation_guides(primitives::Camera *camera) const throw();

        void render_move_transformation_guides(primitives::Camera *camera) const throw();

        void render_rotate_transformation_guides(primitives::Camera *camera) const throw();


    private:
        SceneNode m_root_node;


        /**
         * \brief A spatial subdivision hierarchy structure, used for frustum
         * culling, and other things.
         *
         * In addition to being used to perform frustum culling, it is Used to 
         * traverse the scene in rough front to back order. This is useful for
         * many purposes, including rendering big occluders for occlusion 
         * culling, ensuring correct rendering of translucent objects and 
         * lessening overdraw.
         *
         * \see \c Octree, \c OctreeNode
         */
        Octree m_octree;

        SyncTrigger m_sync_trigger;

        SelectTrigger m_select_trigger;

        /**
         * \brief A queue of modifications slanted for incorporation in the 
         * scene.
         *
         * Scene changes are queued by \c queue_change(SceneChange *) and
         * processed (and thus remvoed from this queue) by 
         * process_changes().
         */
        std::queue<SceneAction *> m_action_queue;

        /**
         * \brief Access mutex for \c m_change_queue.
         *
         * Must be used for both read and write access to 
         * \c m_change_queue.
         */
        boost::recursive_mutex m_action_queue_mutex;

        /**
         * \brief The current render engine configuration.
         *
         * These are global options that \c may be used
         * by the rendering module to alter its behaviour, accordingly.
         */
        Configuration m_config;

        /**
         * \brief Used to synchronize access to \c m_config.
         */
        boost::mutex m_config_mutex;


        /**
         * \brief Locked when the engine is active performing a render
         * or running a picking pass.
         *
         * If one render pass is in progress and another render is requested,
         * it will stall (waiting on this mutex) until the first render pass
         * is finished.
         *
         * It is locked by the public interface of \c RenderEngine only. I.e.
         * internal methods expect to have \e exclusive access already.
         */
        boost::recursive_mutex m_is_active_mutex;


        boost::mutex m_active_flags_mutex;

        /**
         * \brief Set to true when the render engine is active doing work,
         * i.e. running a render or pick pass.
         */
        volatile bool m_is_active;

        /**
         * \brief Set to true when the render engine is active rendering.
         */
        volatile bool m_is_rendering;

        TransformationGuidesMode m_transformation_guides_mode;

        CoordinateSystem m_transformation_guides_coord_sys;
        
        RotationCenterpoint m_transformation_guides_pivot;

        Axis m_transformation_guides_axes;
    };
}
}


#endif // __PEEKABOT_RENDERER_RENDER_ENGINE_HH
