/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <GL/glew.h>
#include "RenderTraverser.hh"
#include "SceneNode.hh"
#include "Texture2D.hh"
#include "RenderContext.hh"


using namespace peekabot;
using namespace peekabot::renderer;



//
// This is a ugly hack to provide a callback function to enumerate
// all of a cullable's renderables, with using only a regular function.
//
// Using Boost.Function would have been much prettier, however, it's far to slow
// to be a good approach. To overcome that problem this ugly duckling was 
// conceived.
//
namespace
{
    namespace local
    {
        static SceneNode *parent;
        static RenderTraverser *traverser;
        
        static void enum_fun(Renderable *renderable)
        {
            traverser->handle_renderable(renderable, parent);
        };
    }
}



RenderTraverser::RenderTraverser(
    const RGBColor &selected_color, 
    const RGBColor &parent_selected_color) throw()
    : m_selected_color(selected_color),
      m_parent_selected_color(parent_selected_color)
{
}


void RenderTraverser::collect(
    Octree &octree, 
    primitives::Camera *camera, 
    const bool *visible_layers)
{
    m_camera = camera;
    m_visible_layers = visible_layers;
   
    octree.traverse_pvs(camera, this);
}


void RenderTraverser::render_opaque() throw()
{
    render_batch(m_opaque, true);
}


void RenderTraverser::render_translucent() throw()
{
    render_batch(m_translucent, false);
}


void RenderTraverser::render_always_on_top() throw()
{
    // Draw always-on-top renderables
    State::push_baseline();
    State baseline = State::get_baseline_state();
    State::OverrideSet ignore_set;

    baseline.set(new statelets::DepthTest(statelets::DepthTest::ALWAYS));

    ignore_set.insert(statelets::DepthTest::get_type());
    State::set_baseline(baseline, ignore_set);
    
    render_batch(m_opaque_on_top, true);
    render_batch(m_translucent_on_top, false);

    State::pop_baseline();
}


void RenderTraverser::finish() throw()
{
    restore_overridden();
}


void RenderTraverser::traverse(Cullable *cullable)
{
    local::parent    = cullable->get_parent();
    local::traverser = this;

    //
    // Draw cullable if and only if A) it's owner node is not
    // hidden and B) it's in a shown layer.
    //
    if( (cullable->get_parent() &&
         !cullable->get_parent()->is_visible() ||
         !m_visible_layers[cullable->get_parent()->get_layer()]) )
        return;
    
    cullable->get_renderables(m_camera, local::enum_fun);
}


void RenderTraverser::handle_renderable(Renderable *renderable, 
                                        SceneNode *parent)
{
    //
    // Is the renderable translucent? We sort opaque and translucent renderables
    // into different batches, so we need to find out.
    // It is translucent when the alpha used for rendering is less than ~1, 
    // which it is when
    // 1) Its alpha statelet reports an alpha less than 1.
    // 2) Its parent node's template state reports an alpha less than 1 and
    //    the renderable statelet is unset.
    //
    float alpha = 1.0f;
    if( renderable->get_state().is_set<statelets::Alpha>() )
        alpha = renderable->get_state().get<statelets::Alpha>()->get();
    else if( parent && parent->get_template_state().is_set<statelets::Alpha>() )
        alpha = parent->get_template_state().get<statelets::Alpha>()->get();
   
    if( alpha > 0.98f ) // cheat a little :)
    {
        if( renderable->is_always_on_top() )
            m_opaque_on_top.push_back(
                std::make_pair(renderable, &parent->get_template_state()));
        else
            m_opaque.push_back(
                std::make_pair(renderable, &parent->get_template_state()));
    }
    else
    {
        if( renderable->is_always_on_top() )
            m_translucent_on_top.push_back(
                std::make_pair(renderable, &parent->get_template_state()));
        else
            m_translucent.push_back(
                std::make_pair(renderable, &parent->get_template_state()));
    }
    
    if( parent && renderable->selection_effects_enabled() )
    {
        if( parent->is_selected() )
        {
            State s = renderable->get_state();
            s.set(new statelets::StencilTest(true));
            override_state(renderable, s);
            m_selected.push_back(
                std::make_pair(renderable, &parent->get_template_state()));
        }
        else if( parent->has_selected_parent() )
        {
            State s = renderable->get_state();
            s.set(new statelets::StencilTest(true));
            override_state(renderable, s);
            m_parent_selected.push_back(
                std::make_pair(renderable, &parent->get_template_state()));
        }
    }
    
    /*if( m_opaque.size() > 100 )
    {
        render_batch(m_opaque);
        m_opaque.clear();
    }*/
}



void RenderTraverser::render_batch(Batch &batch, bool front_to_back) throw()
{       
    // Get occlusion query results
    // TODO
    //
    // Remove occluded renderables from the batch
    // if( occluded )
    //   m_opaque.erase(...);
    
    // Sort them on state and render them
    //m_opaque.sort(...);

    RenderContext context(m_camera, 1);
    
    if( front_to_back )
    {
        for( Batch::iterator it = batch.begin(); it != batch.end(); ++it )
        {
            it->first->get_state().apply(it->second);
            it->first->render(context);
        }
    }
    else
    {
        for( Batch::reverse_iterator it = batch.rbegin(); it != batch.rend(); ++it )
        {
            it->first->get_state().apply(it->second);
            it->first->render(context);
        }
    }

    context.flush();
}



void RenderTraverser::render_outlines() throw()
{
    State::push_baseline();

    State baseline = State::get_baseline_state();
    State::OverrideSet ignore_set;

    baseline.set(new statelets::ColorMaterial(true));    
    baseline.set(new statelets::Lighting(false));
    baseline.set(new statelets::StencilTest(true));
    baseline.set(new statelets::StencilFunc(
                     GL_NOTEQUAL, 1, static_cast<GLuint>(-1),
                     GL_KEEP, GL_KEEP, GL_REPLACE));
    baseline.set(new statelets::PointParams(4, true));
    baseline.set(new statelets::LineParams(6, false, 0, 0));
    baseline.set(new statelets::PolygonMode(
                     statelets::PolygonMode::FRONT,
                     statelets::PolygonMode::LINE));


    ignore_set.insert(statelets::Lighting::get_type());
    ignore_set.insert(statelets::LightModel::get_type());
    ignore_set.insert(statelets::Material::get_type());
    ignore_set.insert(statelets::Color::get_type());
    ignore_set.insert(statelets::ColorMaterial::get_type());
    ignore_set.insert(statelets::StencilTest::get_type());
    ignore_set.insert(statelets::StencilFunc::get_type());
    ignore_set.insert(statelets::LineParams::get_type());
    ignore_set.insert(statelets::PointParams::get_type());
    ignore_set.insert(statelets::ColorPointer::get_type());
    ignore_set.insert(statelets::NormalPointer::get_type());
    ignore_set.insert(statelets::TexCoordPointer::get_type());
    ignore_set.insert(statelets::Alpha::get_type());
    ignore_set.insert(statelets::BlendMode::get_type());
    ignore_set.insert(statelets::Texture2D::get_type());
    

    baseline.get<statelets::Color>()->set(m_selected_color);
    State::set_baseline(baseline, ignore_set);
    render_batch(m_selected, true);

    baseline.get<statelets::Color>()->set(m_parent_selected_color);
    State::set_baseline(baseline, ignore_set);
    render_batch(m_parent_selected, true);
    
    State::pop_baseline();
}


void RenderTraverser::override_state(Renderable *renderable, const State &new_state) throw()
{
    m_override_stack.push(Override(renderable, renderable->get_state()));
    renderable->set_state(new_state);
}


void RenderTraverser::restore_overridden() throw()
{
    while( !m_override_stack.empty() )
    {
        Override override = m_override_stack.top();
        m_override_stack.pop();
        
        override.first->set_state(override.second);
    }
}
