/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_RENDERER_RENDER_TRAVERSER_HH
#define __PEEKABOT_RENDERER_RENDER_TRAVERSER_HH


#include <vector>
#include <stack>
#include "OctreeTraverser.hh"
#include "Renderable.hh"
#include "primitives/Camera.hh"
#include "Octree.hh"


namespace peekabot
{
    namespace renderer
    {


        //
        // For selection, we need support for disabling many 
        // statelets (translucency, etc. etc.) which are unnecessary
        //
        class RenderTraverser : public OctreeTraverser
        {
            typedef std::vector<std::pair<Renderable *, const State *> > Batch;

            Batch               m_opaque;
            Batch               m_translucent;
            Batch               m_opaque_on_top;
            Batch               m_translucent_on_top;
            Batch               m_selected;
            Batch               m_parent_selected;

            const bool         *m_visible_layers;
            primitives::Camera *m_camera;

            RGBColor            m_selected_color;
            RGBColor            m_parent_selected_color;

            typedef std::pair<Renderable *, State> Override;
            typedef std::stack<Override> OverrideStack;

            OverrideStack  m_override_stack;

        public:
            RenderTraverser(
                const RGBColor &selected_color, 
                const RGBColor &parent_selected_color) throw();

            virtual ~RenderTraverser() throw() {}


            void collect(
                Octree &octree, 
                primitives::Camera *camera, 
                const bool *visible_layers);

            void render_opaque() throw();

            void render_translucent() throw();

            /**
             * \brief Render the "selected effect" outlines of the objects 
             * rendered in a previous call to render_pvs().
             *
             * \pre render_pvs() has been executed, on the same object, with 
             * success prior to calling this method.
             */
            void render_outlines() throw();

            void render_always_on_top() throw();

            void finish() throw();


            virtual void traverse(Cullable *cullable);

            void handle_renderable(Renderable *renderable, SceneNode *parent);

        private:
            void render_batch(Batch &batch, bool front_to_back) throw();

            /**
             * \brief Replace a Renderable's state with the supplied State. The 
             * original state will be restored when restore_overriden() is 
             * called.
             */
            void override_state(Renderable *renderable, const State &new_state) throw();

            /**
             * \brief Restore locally overriden Renderable states.
             *
             * \sa override_state(Renderable *, const State &)
             */
            void restore_overridden() throw();

            /*void fire_occlusion_query(Renderable *renderable) throw()
            {
                // ...
            }*/
        };

    }
}



#endif // __PEEKABOT_RENDERER_RENDER_TRAVERSER_HH
