/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdexcept>
#include <string>
#include <boost/format.hpp>
#include <boost/bind.hpp>

#include "Renderer.hh"
#include "../Configuration.hh"
#include "RenderEngine.hh"
#include "IDTranslator.hh"
#include "SceneAction.hh"
#include "PrimitiveFactory.hh"
#include "SceneNode.hh"
#include "primitives/Camera.hh"
#include "primitives/Cube.hh"
#include "primitives/Polygon.hh"
#include "primitives/MeshBased.hh"
#include "primitives/AxisTripod.hh"
#include "primitives/OccupancyGrid2D.hh"
#include "primitives/Label.hh"
#include "LineBased.hh"
#include "PbmfLoader.hh"

#include "Mesh.hh"
#include "MeshCullable.hh"
#include "VertexBased.hh"


using namespace peekabot;
using namespace peekabot::renderer;


//
// Static members initializaition:
//
boost::shared_ptr<IDAllocator<NodeID> > 
Renderer::ms_node_id_allocator(new DefaultIDAllocator<NodeID>);

boost::shared_ptr<IDAllocator<PrimitiveID> > 
Renderer::ms_primitive_id_allocator(new DefaultIDAllocator<PrimitiveID>);




// -------------------------- Renderer implementation --------------------------




Renderer::Renderer() throw()
    : m_is_initialized(false),
      m_engine(0)
{
    float min_node_size = TheConfiguration::instance().get_option<float>(
        "renderer.octree.min_node_size", 1);
    
    m_engine = new RenderEngine(50, min_node_size);
    
    // Add an entry in the ID translator so that we can 
    // access the root node
    m_root_node_id = ms_node_id_allocator->allocate();
    
    SceneNode *node = m_engine->get_root_node();
    TheIDTranslator::instance().add_alias(m_root_node_id, node);
}


Renderer::~Renderer() throw()
{
    if( m_engine )
        delete m_engine;
}


void Renderer::dispatch_action(SceneAction *action) throw()
{
    m_engine->queue_action(action);
}


NodeID Renderer::allocate_node_id() throw()
{
    return ms_node_id_allocator->allocate();
}

NodeID Renderer::allocate_primitive_id() throw()
{
    return ms_primitive_id_allocator->allocate();
}


void Renderer::initialize()
{
    if( !is_initialized() )
    {
        m_engine->init();
        m_is_initialized = true;
    }
}

bool Renderer::is_initialized() const throw()
{
    return m_is_initialized;
}


bool Renderer::render(int32_t w, int32_t h, 
                      PrimitiveID camera,
                      bool *visible_layers, 
                      FrameCollector *collector)
{
    if( !is_initialized() )
        initialize();

    return m_engine->render(w, h, camera, visible_layers, collector);
}


uint32_t Renderer::pick(int32_t w, int32_t h, int32_t x, int32_t y, 
                      PrimitiveID camera, bool *visible_layers, 
                      real32 sensitivity)
{
    if( !is_initialized() )
        initialize();

    return m_engine->pick(w, h, x, y, camera, visible_layers, sensitivity);
}


bool Renderer::has_pending_updates() const throw()
{
    return (m_engine->pending_action_count() != 0);
}


void Renderer::process_updates()
{
    if( !is_initialized() )
        initialize();

    return m_engine->process_actions();
}


NodeID Renderer::get_root_node_id() const throw()
{
    return m_root_node_id;
}


NodeID Renderer::add_node(NodeID parent) throw()
{
    class AddNode : public SceneAction
    {
        NodeID m_parent_id;
        NodeID m_node_id;
    
    public:
        AddNode(NodeID parent_id, NodeID node_id) throw()
            : m_parent_id(parent_id), m_node_id(node_id) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *parent = TheIDTranslator::instance().get_node(m_parent_id);
            assert( parent );

            SceneNode *node = new SceneNode();
            parent->attach(node);
            
            // Register an alias for the new node
            TheIDTranslator::instance().add_alias(m_node_id, node);
        }
    };

    NodeID node_id = allocate_node_id();
    dispatch_action(new AddNode(parent, node_id));
    return node_id;
}


NodeID Renderer::clone_node(NodeID id, NodeID parent) throw()
{
    class CloneNode : public SceneAction
    {
        NodeID m_source_id;
        NodeID m_parent_id;
        NodeID m_node_id;
    
    public:
        CloneNode(NodeID source_id, NodeID parent_id, NodeID node_id) throw()
            : m_source_id(source_id), 
              m_parent_id(parent_id), 
              m_node_id(node_id) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *source = TheIDTranslator::instance().get_node(m_source_id);
            SceneNode *parent = TheIDTranslator::instance().get_node(m_parent_id);
            assert( parent && source );

            SceneNode *node = source->clone_recursive();
            parent->attach(node);
            
            // Register an alias for the new node
            TheIDTranslator::instance().add_alias(m_node_id, node);
        }
    };

    NodeID node_id = allocate_node_id();
    dispatch_action(new CloneNode(id, parent, node_id));
    return node_id;
}


void Renderer::set_node_transformation(
    NodeID id, const Matrix4ru &m, CoordinateSystem coord_sys) throw()
{
    class SetTransformation : public SceneAction
    {
        NodeID m_node_id;
        Matrix4r m_transformation;
        CoordinateSystem m_coord_sys;
    
    public:
        SetTransformation(NodeID node_id, 
                          const Matrix4r &m, 
                          CoordinateSystem coord_sys) throw()
            : m_node_id(node_id), m_transformation(m), m_coord_sys(coord_sys) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *node = TheIDTranslator::instance().get_node(m_node_id);
            assert( node );

            if( m_coord_sys == LOCAL_COORDINATES )
            {
                node->set_transformation(
                    m_transformation * 
                    node->get_transformation());
            }
            else if( m_coord_sys == PARENT_COORDINATES )
            {
                node->set_transformation(m_transformation);
            }
            else
            {
                node->set_transformation(
                    m_transformation * 
                    node->get_unscaled_mtow().inverse());
            }
        }
    };

    dispatch_action(new SetTransformation(id, m, coord_sys));
}


void Renderer::set_node_color(NodeID id, const RGBColor &color) throw()
{
    class SetNodeColor : public SceneAction
    {
        NodeID m_node_id;
        RGBColor m_color;
    
    public:
        SetNodeColor(NodeID node_id, const RGBColor &color) throw() 
            : m_node_id(node_id), m_color(color) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *node = TheIDTranslator::instance().get_node(m_node_id);
            node->set_color(m_color);
        }
    };

    dispatch_action(new SetNodeColor(id, color));
}

void Renderer::set_node_opacity(NodeID id, const Opacity &opacity) throw()
{
    class SetNodeOpacity : public SceneAction
    {
        NodeID m_node_id;
        Opacity m_opacity;
    
    public:
        SetNodeOpacity(NodeID node_id, const Opacity &opacity) throw() 
            : m_node_id(node_id), m_opacity(opacity) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *node = TheIDTranslator::instance().get_node(m_node_id);
            node->set_opacity(m_opacity);
        }
    };

    dispatch_action(new SetNodeOpacity(id, opacity));
}

void Renderer::set_node_scale(NodeID id, real32 x_scale, real32 y_scale, real32 z_scale) throw()
{
    class SetNodeScale : public SceneAction
    {
        NodeID m_node_id;
        real32 m_x_scale;
        real32 m_y_scale;
        real32 m_z_scale;
    
    public:
        SetNodeScale(NodeID node_id, 
                     real32 x_scale,
                     real32 y_scale, 
                     real32 z_scale) throw() 
            : m_node_id(node_id), 
              m_x_scale(x_scale),
              m_y_scale(y_scale),
              m_z_scale(z_scale) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *node = TheIDTranslator::instance().get_node(m_node_id);
            node->set_scale(m_x_scale, m_y_scale, m_z_scale);
        }
    };

    dispatch_action(new SetNodeScale(id, x_scale, y_scale, z_scale));
}

void Renderer::set_node_visibility(NodeID id, bool hide) throw()
{
    class SetNodeVisibility : public SceneAction
    {
        NodeID m_node_id;
        bool m_hide;
    
    public:
        SetNodeVisibility(NodeID node_id, bool hide) throw() 
            : m_node_id(node_id), m_hide(hide) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *node = TheIDTranslator::instance().get_node(m_node_id);
            if( m_hide )
                node->hide();
            else
                node->show();
        }
    };

    dispatch_action(new SetNodeVisibility(id, hide));
}

void Renderer::set_node_layer(NodeID id, uint32_t layer) throw()
{
    class SetNodeLayer : public SceneAction
    {
        NodeID m_node_id;
        uint32_t m_layer;
    
    public:
        SetNodeLayer(NodeID node_id, uint32_t layer) throw() 
            : m_node_id(node_id), m_layer(layer) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *node = TheIDTranslator::instance().get_node(m_node_id);
            node->set_layer(m_layer);
        }
    };

    dispatch_action(new SetNodeLayer(id, layer));
}

void Renderer::set_node_name(NodeID id, uint32_t name) throw()
{
    class SetNodeName : public SceneAction
    {
        NodeID m_node_id;
        uint32_t m_name;
    
    public:
        SetNodeName(NodeID node_id, uint32_t name) throw() 
            : m_node_id(node_id), m_name(name) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *node = TheIDTranslator::instance().get_node(m_node_id);
            node->set_name(m_name);
        }
    };

    dispatch_action(new SetNodeName(id, name));
}

void Renderer::select_node(NodeID id) throw()
{
    class SelectNode : public SceneAction
    {
        NodeID m_node_id;
    
    public:
        SelectNode(NodeID node_id) throw() : m_node_id(node_id) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *node = TheIDTranslator::instance().get_node(m_node_id);
            node->select();
        }
    };

    dispatch_action(new SelectNode(id));
}


void Renderer::deselect_node(NodeID id) throw()
{
    class DeselectNode : public SceneAction
    {
        NodeID m_node_id;
    
    public:
        DeselectNode(NodeID node_id) throw() : m_node_id(node_id) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *node = TheIDTranslator::instance().get_node(m_node_id);
            node->deselect();
        }
    };

    dispatch_action(new DeselectNode(id));
}

void Renderer::remove_node(NodeID id) throw()
{
    class RemoveNode : public SceneAction
    {
        NodeID m_node_id;
    
    public:
        RemoveNode(NodeID node_id) throw() : m_node_id(node_id) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *node = TheIDTranslator::instance().get_node(m_node_id);
            node->detach();
            delete node;
        }
    };

    dispatch_action(new RemoveNode(id));
}

void Renderer::rearrange_node(NodeID id, NodeID new_parent) throw()
{
    class RearrangeNode : public SceneAction
    {
        NodeID m_node_id;
        NodeID m_new_parent_id;
    
    public:
        RearrangeNode(NodeID node_id, NodeID new_parent_id) throw() 
            : m_node_id(node_id), m_new_parent_id(new_parent_id) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *node = TheIDTranslator::instance().get_node(m_node_id);
            SceneNode *new_parent = TheIDTranslator::instance().get_node(
                m_new_parent_id);
            SceneNode *old_parent = node->get_parent();

            if( old_parent == 0 )
                throw std::runtime_error(
                    "You can't rearrange the root node");

            node->detach();

            if( new_parent->attach(node) )
            {
                // Oops, attaching it to new parent failed. Let's put it back
                // at its original location
                old_parent->attach(node);
                throw std::runtime_error("Failed to rearrange node");
            }
        }
    };

    dispatch_action(new RearrangeNode(id, new_parent));
}


PrimitiveID Renderer::add_camera(NodeID parent, bool anonymous) throw()
{
    class AddCamera : public SceneAction
    {
        NodeID      m_parent_id;
        PrimitiveID m_id;
        bool        m_is_anon;
    
    public:
        AddCamera(NodeID parent_id, PrimitiveID id, bool is_anon) throw()
            : m_parent_id(parent_id), m_id(id), m_is_anon(is_anon) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *parent = TheIDTranslator::instance().get_node(m_parent_id);
            assert( parent );

            primitives::Camera *obj = new primitives::Camera();
            parent->add_cullable(obj);

            if( !m_is_anon )
            {
                // Register an alias for the new primitive
                TheIDTranslator::instance().add_alias(m_id, obj);
            }
        }
    };

    PrimitiveID id = anonymous ? 0 : allocate_primitive_id();
    dispatch_action(new AddCamera(parent, id, anonymous));
    return id;
}

PrimitiveID Renderer::add_sphere(NodeID parent, bool anonymous) throw()
{
    class AddSphere : public SceneAction
    {
        NodeID      m_parent_id;
        PrimitiveID m_id;
        bool        m_is_anon;
    
    public:
        AddSphere(NodeID parent_id, PrimitiveID id, bool is_anon) throw()
            : m_parent_id(parent_id), m_id(id), m_is_anon(is_anon) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *parent = TheIDTranslator::instance().get_node(m_parent_id);
            assert( parent );

            primitives::MeshBased *obj = ThePrimitiveFactory::instance().get_icosphere();
            parent->add_cullable(obj);

            if( !m_is_anon )
            {
                // Register an alias for the new primitive
                TheIDTranslator::instance().add_alias(m_id, obj);
            }
        }
    };

    PrimitiveID id = anonymous ? 0 : allocate_primitive_id();
    dispatch_action(new AddSphere(parent, id, anonymous));
    return id;
}

PrimitiveID Renderer::add_cylinder(NodeID parent, bool anonymous) throw()
{
    class AddCylinder : public SceneAction
    {
        NodeID      m_parent_id;
        PrimitiveID m_id;
        bool        m_is_anon;
    
    public:
        AddCylinder(NodeID parent_id, PrimitiveID id, bool is_anon) throw()
            : m_parent_id(parent_id), m_id(id), m_is_anon(is_anon) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *parent = TheIDTranslator::instance().get_node(m_parent_id);
            assert( parent );

            primitives::MeshBased *obj = ThePrimitiveFactory::instance().get_cylinder();
            parent->add_cullable(obj);

            if( !m_is_anon )
            {
                // Register an alias for the new primitive
                TheIDTranslator::instance().add_alias(m_id, obj);
            }
        }
    };

    PrimitiveID id = anonymous ? 0 : allocate_primitive_id();
    dispatch_action(new AddCylinder(parent, id, anonymous));
    return id;
}

PrimitiveID Renderer::add_cube(NodeID parent, bool anonymous) throw()
{
    class AddCube : public SceneAction
    {
        NodeID      m_parent_id;
        PrimitiveID m_id;
        bool        m_is_anon;
    
    public:
        AddCube(NodeID parent_id, PrimitiveID id, bool is_anon) throw()
            : m_parent_id(parent_id), m_id(id), m_is_anon(is_anon) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *parent = TheIDTranslator::instance().get_node(m_parent_id);
            assert( parent );

            primitives::Cube *obj = new primitives::Cube();
            parent->add_cullable(obj);

            if( !m_is_anon )
            {
                // Register an alias for the new primitive
                TheIDTranslator::instance().add_alias(m_id, obj);
            }
        }
    };

    PrimitiveID id = anonymous ? 0 : allocate_primitive_id();
    dispatch_action(new AddCube(parent, id, anonymous));
    return id;
}


PrimitiveID Renderer::add_model(
    NodeID parent, bool anonymous, const std::string &filename) throw()
{
    class AddModel : public SceneAction
    {
        NodeID      m_parent_id;
        PrimitiveID m_id;
        bool        m_is_anon;
        std::string m_filename;
    
    public:
        AddModel(NodeID parent_id, PrimitiveID id, 
                 bool is_anon, const std::string &filename) throw()
            : m_parent_id(parent_id), m_id(id), 
              m_is_anon(is_anon), m_filename(filename) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *parent = TheIDTranslator::instance().get_node(m_parent_id);
            assert( parent );

            PbmfLoader loader;
            primitives::MeshBased *model = loader.load(m_filename);
            parent->add_cullable(model);

            if( !m_is_anon )
            {
                // Register an alias for the new primitive
                TheIDTranslator::instance().add_alias(m_id, model);
            }
        }
    };

    PrimitiveID id = anonymous ? 0 : allocate_primitive_id();
    dispatch_action(new AddModel(parent, id, anonymous, filename));
    return id;
}


PrimitiveID Renderer::add_circle(
    NodeID parent, bool anonymous) throw()
{
    class AddCircle : public SceneAction
    {
        NodeID      m_parent_id;
        PrimitiveID m_id;
        bool        m_is_anon;
    
    public:
        AddCircle(NodeID parent_id, PrimitiveID id, bool is_anon) throw()
            : m_parent_id(parent_id), m_id(id), m_is_anon(is_anon) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *parent = TheIDTranslator::instance().get_node(m_parent_id);
            assert( parent );

            primitives::MeshBased *obj = ThePrimitiveFactory::instance().get_circle();
            parent->add_cullable(obj);

            if( !m_is_anon )
            {
                // Register an alias for the new primitive
                TheIDTranslator::instance().add_alias(m_id, obj);
            }
        }
    };

    PrimitiveID id = anonymous ? 0 : allocate_primitive_id();
    dispatch_action(new AddCircle(parent, id, anonymous));
    return id;
}


PrimitiveID Renderer::add_disc(NodeID parent, bool anonymous) throw()
{
    class AddDisc : public SceneAction
    {
        NodeID      m_parent_id;
        PrimitiveID m_id;
        bool        m_is_anon;
    
    public:
        AddDisc(NodeID parent_id, PrimitiveID id, bool is_anon) throw()
            : m_parent_id(parent_id), m_id(id), m_is_anon(is_anon) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *parent = TheIDTranslator::instance().get_node(m_parent_id);
            assert( parent );

            primitives::MeshBased *obj = ThePrimitiveFactory::instance().get_disc();
            parent->add_cullable(obj);

            if( !m_is_anon )
            {
                // Register an alias for the new primitive
                TheIDTranslator::instance().add_alias(m_id, obj);
            }
        }
    };

    PrimitiveID id = anonymous ? 0 : allocate_primitive_id();
    dispatch_action(new AddDisc(parent, id, anonymous));
    return id;
}



PrimitiveID Renderer::add_point_blob(
    NodeID parent, bool anonymous, 
    const std::vector<Vector3r> &vertices) throw()
{
    PrimitiveID id = add_point_mesh(parent, false);
    set_primitive_vertices(id, vertices);
    return id;
}


PrimitiveID Renderer::add_line_blob(
    NodeID parent, bool anonymous, 
    const std::vector<Vector3r> &vertices) throw(std::logic_error)
{
    PrimitiveID id = add_line_mesh(parent, false);
    set_primitive_vertices(id, vertices);
    return id;
}


namespace
{
    template<class U, class T>
    class AddMesh : public SceneAction
    {
        NodeID      m_parent_id;
        PrimitiveID m_id;
        bool        m_is_anon;
    
    public:
        AddMesh(NodeID parent_id, PrimitiveID id, bool is_anon) throw()
            : m_parent_id(parent_id), m_id(id), m_is_anon(is_anon) {}
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *parent = TheIDTranslator::instance().get_node(m_parent_id);
            assert( parent );

            boost::shared_ptr<T> mesh(new T);
            MeshCullable *obj = new U(mesh);

            try
            {
                parent->add_cullable(obj);
            }
            catch(...)
            {
                delete obj;
                throw;
            }

            if( !m_is_anon )
            {
                // Register an alias for the new primitive
                TheIDTranslator::instance().add_alias(m_id, obj);
            }
        }
    };
}



PrimitiveID Renderer::add_point_mesh(NodeID parent, bool anonymous) throw()
{
    PrimitiveID id = anonymous ? 0 : allocate_primitive_id();
    dispatch_action(
        new AddMesh<PointMeshCullable, PointMesh>(parent, id, anonymous));
    return id;
}


PrimitiveID Renderer::add_line_mesh(NodeID parent, bool anonymous) throw()
{
    PrimitiveID id = anonymous ? 0 : allocate_primitive_id();
    dispatch_action(
        new AddMesh<LineMeshCullable, LineMesh>(parent, id, anonymous));
    return id;
}


PrimitiveID Renderer::add_tri_mesh(NodeID parent, bool anonymous) throw()
{
    PrimitiveID id = anonymous ? 0 : allocate_primitive_id();
    dispatch_action(
        new AddMesh<TriMeshCullable, TriMesh>(parent, id, anonymous));
    return id;
}



PrimitiveID Renderer::add_polygon(
    NodeID parent, bool anonymous, 
    const std::vector<Vector3r> &vertices) throw()
{
    class AddPolygon : public SceneAction
    {
        NodeID      m_parent_id;
        PrimitiveID m_id;
        bool        m_is_anon;
        std::vector<Vector3r> m_vertices;
    
    public:
        AddPolygon(NodeID parent_id, PrimitiveID id, 
                   bool is_anon, const std::vector<Vector3r> &vertices) throw()
            : m_parent_id(parent_id), 
              m_id(id), 
              m_is_anon(is_anon),
              m_vertices(vertices)
        {
        }
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *parent = TheIDTranslator::instance().get_node(m_parent_id);
            assert( parent );

            primitives::Polygon *obj = new primitives::Polygon();
            obj->set_vertices(m_vertices);
            parent->add_cullable(obj);

            if( !m_is_anon )
            {
                // Register an alias for the new primitive
                TheIDTranslator::instance().add_alias(m_id, obj);
            }
        }
    };

    PrimitiveID id = anonymous ? 0 : allocate_primitive_id();
    dispatch_action(new AddPolygon(parent, id, anonymous, vertices));
    return id;
}


PrimitiveID Renderer::add_label(NodeID parent, bool anonymous) throw()
{
    class AddLabel : public SceneAction
    {
        NodeID      m_parent_id;
        PrimitiveID m_id;
        bool        m_is_anon;
    
    public:
        AddLabel(NodeID parent_id, PrimitiveID id, bool is_anon) throw()
            : m_parent_id(parent_id), 
              m_id(id), 
              m_is_anon(is_anon)
        {
        }
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *parent = TheIDTranslator::instance().get_node(m_parent_id);
            assert( parent );

            primitives::Label *obj = new primitives::Label();
            parent->add_cullable(obj);

            if( !m_is_anon )
            {
                // Register an alias for the new primitive
                TheIDTranslator::instance().add_alias(m_id, obj);
            }
        }
    };

    PrimitiveID id = anonymous ? 0 : allocate_primitive_id();
    dispatch_action(new AddLabel(parent, id, anonymous));
    return id;
}


PrimitiveID Renderer::add_occupancy_grid_2d(
    NodeID parent, bool anonymous, 
    float cell_size,
    const RGBColor &free_color, 
    const RGBColor &occupied_color) throw()
{
    class AddOccupancyGrid2D : public SceneAction
    {
        NodeID      m_parent_id;
        PrimitiveID m_id;
        bool        m_is_anon;
        float       m_cell_size;
        const RGBColor m_free_color;
        const RGBColor m_occupied_color;
    
    public:
        AddOccupancyGrid2D(NodeID parent_id, PrimitiveID id, bool is_anon, 
                           float cell_size,
                           const RGBColor &free_color, 
                           const RGBColor &occupied_color) throw()
            : m_parent_id(parent_id), 
              m_id(id), 
              m_is_anon(is_anon),
              m_cell_size(cell_size),
              m_free_color(free_color),
              m_occupied_color(occupied_color)
        {
        }
    
        virtual void execute(RenderEngine *engine)
        {
            SceneNode *parent = TheIDTranslator::instance().get_node(m_parent_id);
            assert( parent );

            primitives::OccupancyGrid2D *obj = new primitives::OccupancyGrid2D(
                m_cell_size, m_free_color, m_occupied_color, 1024*256);
            parent->add_cullable(obj);

            if( !m_is_anon )
            {
                // Register an alias for the new primitive
                TheIDTranslator::instance().add_alias(m_id, obj);
            }
        }
    };

    PrimitiveID id = anonymous ? 0 : allocate_primitive_id();
    dispatch_action(
        new AddOccupancyGrid2D(
            parent, id, anonymous, cell_size, free_color, occupied_color));
    return id;
}


void Renderer::set_occupancy_grid_2d_cells(
    PrimitiveID id, const std::vector<std::pair<Vector2f, float> > &cells) 
    throw(std::domain_error)
{
    class SetOccupancyGrid2DCells : public SceneAction
    {
        PrimitiveID m_primitive_id;
        std::vector<std::pair<Vector2f, float> > m_cells;
    
    public:
        SetOccupancyGrid2DCells(
            PrimitiveID primitive_id, 
            const std::vector<std::pair<Vector2f, float> > &cells) throw() 
            : m_primitive_id(primitive_id),
              m_cells(cells) {}
    
        virtual void execute(RenderEngine *engine)
        {
            IDTranslator::Primitive primitive = 
                TheIDTranslator::instance().get_primitive(m_primitive_id);

            primitives::OccupancyGrid2D *obj = dynamic_cast<
                primitives::OccupancyGrid2D *>(*primitive.begin());

            if( !obj )
                throw std::runtime_error(
                    "Failed to set occupancy grid (2D) cells, "
                    "wrong primitive type found");

            for( size_t i = 0; i < m_cells.size(); ++i )
            {
               if( m_cells[i].second == -1 )
                    obj->clear_cell(m_cells[i].first);
                else
                    obj->set_belief(m_cells[i].first, m_cells[i].second);
            }
        }
    };

    /*if( belief < 0 || belief > 1 )
      throw std::domain_error("Belief must be in the range [0,1]");*/

    dispatch_action(new SetOccupancyGrid2DCells(id, cells));
}


void Renderer::remove_primitive(PrimitiveID id) throw()
{
    class RemovePrimitive : public SceneAction
    {
        PrimitiveID m_primitive_id;
    
    public:
        RemovePrimitive(PrimitiveID primitive_id) throw() 
            : m_primitive_id(primitive_id) {}
    
        virtual void execute(RenderEngine *engine)
        {
            IDTranslator::Primitive primitive = 
                TheIDTranslator::instance().get_primitive(m_primitive_id);

            for( IDTranslator::Primitive::iterator it = primitive.begin();
                 it != primitive.end(); it++ )
            {
                (*it)->get_parent()->remove_cullable(*it);
            }
        }
    };

    dispatch_action(new RemovePrimitive(id));
}

/*void Renderer::rearrange_primitive(PrimitiveID id, NodeID new_parent) throw()
{
}*/

void Renderer::set_primitive_color(
    PrimitiveID id, const RGBColor &color) throw()
{
    class SetPrimitiveColor : public SceneAction
    {
        PrimitiveID m_primitive_id;
        RGBColor    m_color;
    
    public:
        SetPrimitiveColor(PrimitiveID primitive_id, 
                          const RGBColor &color) throw() 
            : m_primitive_id(primitive_id),
              m_color(color) {}
    
        virtual void execute(RenderEngine *engine)
        {
            IDTranslator::Primitive primitive = 
                TheIDTranslator::instance().get_primitive(m_primitive_id);

            for( IDTranslator::Primitive::iterator it = primitive.begin();
                 it != primitive.end(); it++ )
            {
                Cullable::Renderables r = (*it)->get_renderables();
                for( Cullable::Renderables::iterator jt = r.begin(); 
                     jt != r.end(); jt++ )
                    (*jt)->get_state().get<statelets::Color>()->set(m_color);
            }
        }
    };

    dispatch_action(new SetPrimitiveColor(id, color));
}

void Renderer::set_primitive_alpha(PrimitiveID id, real32 alpha) throw()
{
    class SetPrimitiveAlpha : public SceneAction
    {
        PrimitiveID m_primitive_id;
        real32      m_alpha;
    
    public:
        SetPrimitiveAlpha(PrimitiveID primitive_id, 
                          real32 alpha) throw() 
            : m_primitive_id(primitive_id),
              m_alpha(alpha) {}
    
        virtual void execute(RenderEngine *engine)
        {
            IDTranslator::Primitive primitive = 
                TheIDTranslator::instance().get_primitive(m_primitive_id);

            for( IDTranslator::Primitive::iterator it = primitive.begin();
                 it != primitive.end(); it++ )
            {
                Cullable::Renderables r = (*it)->get_renderables();
                for( Cullable::Renderables::iterator jt = r.begin(); 
                     jt != r.end(); jt++ )
                {
                    Renderable *r = *jt;

                    if( m_alpha < 0.99 )
                    {
                        r->get_state().get<statelets::Alpha>()->set(m_alpha);
                        
                        r->get_state().set(
                            boost::shared_ptr<statelets::BlendMode>(
                                new statelets::BlendMode(
                                    GL_CONSTANT_ALPHA, GL_ONE_MINUS_CONSTANT_ALPHA)));
                        
                        r->get_state().set(
                            boost::shared_ptr<statelets::Culling>(
                                new statelets::Culling(false)));
                        
                        r->get_state().set(
                            boost::shared_ptr<statelets::LightModel>(
                                new statelets::LightModel(true)));
                        
                        r->get_state().set(
                            boost::shared_ptr<statelets::DepthMask>(
                                new statelets::DepthMask(false)));
                    }
                    else
                    {
                        r->get_state().set(
                            boost::shared_ptr<statelets::BlendMode>(
                                new statelets::BlendMode(
                                    GL_ONE, GL_ZERO)));

                        r->get_state().unset<statelets::Alpha>();
                        r->get_state().unset<statelets::Culling>();
                        r->get_state().unset<statelets::LightModel>();
                        r->get_state().unset<statelets::DepthMask>();
                    }
                }
            }
        }
    };

    dispatch_action(new SetPrimitiveAlpha(id, alpha));
}

/*void Renderer::set_primitive_name(PrimitiveID id, uint32_t name) throw()
{
}

void Renderer::set_primitive_point_size(PrimitiveID id, real32 size) throw()
{
}

void Renderer::set_primitive_texture(
    PrimitiveID id, const std::string &filename) throw()
{
}

void Renderer::set_primitive_texture(
    PrimitiveID id, uint32_t w, uint32_t h, 
    const void *tex, bool replace) throw()
{
}

void Renderer::set_primitive_depth_test(PrimitiveID id, bool enable) throw()
{
}*/

void Renderer::transform_primitive_geometry(
    PrimitiveID id, 
    const boost::function<Vector3r (const Vector4r &)> &fun) throw()
{
    class TransformPrimitiveGeometry : public SceneAction
    {
        PrimitiveID m_primitive_id;
        boost::function<Vector3r (const Vector4r &)> m_transform;
    
    public:
        TransformPrimitiveGeometry(
            PrimitiveID primitive_id, 
            const boost::function<Vector3r (const Vector4r &)> &transform) throw() 
            : m_primitive_id(primitive_id),
              m_transform(transform) {}
    
        virtual void execute(RenderEngine *engine)
        {
            IDTranslator::Primitive primitive = 
                TheIDTranslator::instance().get_primitive(m_primitive_id);

            for( IDTranslator::Primitive::iterator it = primitive.begin();
                 it != primitive.end(); it++ )
            {
                (*it)->transform_geometry(m_transform);
            }
        }
    };

    dispatch_action(new TransformPrimitiveGeometry(id, fun));
}


void Renderer::transform_primitive_geometry(
    PrimitiveID id, const Matrix4r &transform) throw()
{
    class TransformPrimitiveGeometry : public SceneAction
    {
        PrimitiveID m_primitive_id;
        Matrix4r    m_transform;
    
    public:
        TransformPrimitiveGeometry(
            PrimitiveID primitive_id, const Matrix4r &transform) throw() 
            : m_primitive_id(primitive_id),
              m_transform(transform) {}

        static Vector3r transform_fun(Matrix4r &transform, const Vector4r &v) throw()
        {
            return transform * v;
        }
   
        virtual void execute(RenderEngine *engine)
        {
            IDTranslator::Primitive primitive = 
                TheIDTranslator::instance().get_primitive(m_primitive_id);

            boost::function<Vector3r (const Vector4r &)> fun(
                boost::bind(&TransformPrimitiveGeometry::transform_fun, m_transform, _1));

            for( IDTranslator::Primitive::iterator it = primitive.begin();
                 it != primitive.end(); it++ )
            {
                (*it)->transform_geometry(fun);
            }
        }
    };

    dispatch_action(new TransformPrimitiveGeometry(id, transform));
}


void Renderer::set_primitive_vertices(
    PrimitiveID id, const std::vector<Vector3r> &vertices) throw()
{
    class SetVertices : public SceneAction
    {
        PrimitiveID m_primitive_id;
        VertexBased::Vertices m_vertices;
    
    public:
        SetVertices(PrimitiveID primitive_id, 
                    const VertexBased::Vertices &vertices) throw() 
            : m_primitive_id(primitive_id), 
              m_vertices(vertices) {}
    
        virtual void execute(RenderEngine *engine)
        {
            IDTranslator::Primitive p = TheIDTranslator::instance().get_primitive(m_primitive_id);

            if( p.empty() )
                throw std::runtime_error(
                    "Failed to set vertices - object not found");

            VertexBased *vertex_based = dynamic_cast<VertexBased *>(*p.begin());
            
            if( !vertex_based )
                throw std::runtime_error(
                    "Failed to set vertices - object is not vertex based");

            vertex_based->set_vertices(m_vertices);
        }
    };

    dispatch_action(new SetVertices(id, vertices));
}


void Renderer::set_primitive_line_style(
    PrimitiveID id, LineStyle stipple_pattern, 
    int stipple_factor, float line_width) throw()
{
    class SetLineStyle : public SceneAction
    {
        PrimitiveID m_primitive_id;
        uint16_t m_stipple_pattern;
        int m_stipple_factor;
        float m_line_width;
    
    public:
        SetLineStyle(PrimitiveID primitive_id, uint16_t stipple_pattern,
                     int stipple_factor, float line_width) throw() 
            : m_primitive_id(primitive_id), 
              m_stipple_pattern(stipple_pattern),
              m_stipple_factor(stipple_factor),
              m_line_width(line_width) {}
    
        virtual void execute(RenderEngine *engine)
        {
            IDTranslator::Primitive p = TheIDTranslator::instance().get_primitive(
                m_primitive_id);

            if( p.empty() )
                throw std::runtime_error(
                    "Failed to set line style - primitive not found");

            int n_processed = 0;

            for( IDTranslator::Primitive::iterator it = p.begin(); 
                 it != p.end(); ++it )
            {
                LineBased *r = dynamic_cast<LineBased *>(*it);
                if( r )
                {
                    r->set_line_style(m_line_width, 
                                      m_stipple_pattern, 
                                      m_stipple_factor);
                    ++n_processed;
                }
            }

            if( n_processed < 1 )
                throw std::runtime_error(
                    "Failed to set line style - no LineBased cullables found");
        }
    };

    dispatch_action(
        new SetLineStyle(id, (uint32_t)stipple_pattern, stipple_factor, line_width));
}


void Renderer::set_camera_fov(PrimitiveID id, real32 fov) throw()
{
    class SetCameraFOV : public SceneAction
    {
        PrimitiveID m_primitive_id;
        real32 m_fov;
    
    public:
        SetCameraFOV(PrimitiveID primitive_id, real32 fov) throw() 
            : m_primitive_id(primitive_id),
              m_fov(fov) {}
    
        virtual void execute(RenderEngine *engine)
        {
            IDTranslator::Primitive p = TheIDTranslator::instance().get_primitive(m_primitive_id);
            primitives::Camera *cam = dynamic_cast<primitives::Camera *>(*p.begin());
            
            if( !cam )
                throw std::runtime_error(
                    "Failed to set camera FOV - camera not found");

            cam->set_fov(m_fov);
        }
    };

    dispatch_action(new SetCameraFOV(id, fov));
}

void Renderer::set_camera_view_mode(PrimitiveID id, bool orthographic) throw()
{
    class SetCameraViewMode : public SceneAction
    {
        PrimitiveID m_primitive_id;
        bool        m_ortho;
    
    public:
        SetCameraViewMode(PrimitiveID primitive_id, bool ortho) throw() 
            : m_primitive_id(primitive_id),
              m_ortho(ortho) {}
    
        virtual void execute(RenderEngine *engine)
        {
            IDTranslator::Primitive p = TheIDTranslator::instance().get_primitive(m_primitive_id);
            primitives::Camera *cam = dynamic_cast<primitives::Camera *>(*p.begin());
            
            if( !cam )
                throw std::runtime_error(
                    "Failed to set camera view mode, camera not found");

            cam->set_orthographic(m_ortho);
        }
    };

    dispatch_action(new SetCameraViewMode(id, orthographic));
}

void Renderer::set_camera_near_plane(PrimitiveID id, real32 _near) throw()
{
    class SetNearPlane : public SceneAction
    {
        PrimitiveID m_primitive_id;
        real32 m_near;
    
    public:
        SetNearPlane(PrimitiveID primitive_id, real32 _near) throw() 
            : m_primitive_id(primitive_id),
              m_near(_near) {}
    
        virtual void execute(RenderEngine *engine)
        {
            IDTranslator::Primitive p = TheIDTranslator::instance().get_primitive(m_primitive_id);
            primitives::Camera *cam = dynamic_cast<primitives::Camera *>(*p.begin());
            
            if( !cam )
                throw std::runtime_error(
                    "Failed to set camera near plane - camera not found");

            cam->set_near_plane(m_near);
        }
    };

    dispatch_action(new SetNearPlane(id, _near));
}

void Renderer::set_camera_far_plane(PrimitiveID id, real32 _far) throw()
{
    class SetFarPlane : public SceneAction
    {
        PrimitiveID m_primitive_id;
        real32 m_far;
    
    public:
        SetFarPlane(PrimitiveID primitive_id, real32 _far) throw() 
            : m_primitive_id(primitive_id),
              m_far(_far) {}
    
        virtual void execute(RenderEngine *engine)
        {
            IDTranslator::Primitive p = TheIDTranslator::instance().get_primitive(m_primitive_id);
            primitives::Camera *cam = dynamic_cast<primitives::Camera *>(*p.begin());
            
            if( !cam )
                throw std::runtime_error(
                    "Failed to set camera far plane - camera not found");

            cam->set_far_plane(m_far);
        }
    };

    dispatch_action(new SetFarPlane(id, _far));
}

void Renderer::set_camera_zoom_distance(PrimitiveID id, float zoom_distance) throw()
{
    class SetCameraZoomDistance : public SceneAction
    {
        PrimitiveID m_primitive_id;
        float m_zoom_distance;
    
    public:
        SetCameraZoomDistance(PrimitiveID primitive_id, float zoom_distance) throw() 
            : m_primitive_id(primitive_id),
              m_zoom_distance(zoom_distance) {}
    
        virtual void execute(RenderEngine *engine)
        {
            IDTranslator::Primitive p = TheIDTranslator::instance().get_primitive(m_primitive_id);
            primitives::Camera *cam = dynamic_cast<primitives::Camera *>(*p.begin());
            
            if( !cam )
                throw std::runtime_error(
                    "Failed to set camera zoom distance - camera not found");

            cam->set_zoom_distance(m_zoom_distance);
        }
    };

    dispatch_action(new SetCameraZoomDistance(id, zoom_distance));
}


void Renderer::set_label_text(PrimitiveID id, const std::string &text) throw()
{
    class SetLabelText : public SceneAction
    {
        PrimitiveID m_primitive_id;
        const std::string &m_text;
    
    public:
        SetLabelText(PrimitiveID primitive_id, const std::string &text) throw() 
            : m_primitive_id(primitive_id),
              m_text(text) {}
    
        virtual void execute(RenderEngine *engine)
        {
            IDTranslator::Primitive p = TheIDTranslator::instance().get_primitive(m_primitive_id);
            primitives::Label *lbl = dynamic_cast<primitives::Label *>(*p.begin());
            
            if( !lbl )
                throw std::runtime_error("Failed to set label text - object "
                                         "not found or of non-label type");

            lbl->set_text(m_text);
        }
    };

    dispatch_action(new SetLabelText(id, text));
}


void Renderer::set_label_alignment(PrimitiveID id, TextAlignment alignment) throw()
{
    class SetLabelAlignment : public SceneAction
    {
        PrimitiveID m_primitive_id;
        TextAlignment m_alignment;
    
    public:
        SetLabelAlignment(PrimitiveID primitive_id, TextAlignment alignment) throw() 
            : m_primitive_id(primitive_id),
              m_alignment(alignment) {}
    
        virtual void execute(RenderEngine *engine)
        {
            IDTranslator::Primitive p = TheIDTranslator::instance().get_primitive(m_primitive_id);
            primitives::Label *lbl = dynamic_cast<primitives::Label *>(*p.begin());
            
            if( !lbl )
                throw std::runtime_error("Failed to set label text - object "
                                         "not found or of non-label type");

            lbl->set_alignment(m_alignment);
        }
    };

    dispatch_action(new SetLabelAlignment(id, alignment));
}


void Renderer::set_transformation_guides_mode(TransformationGuidesMode mode) throw()
{
    class SetTransformationGuidesMode : public SceneAction
    {
        TransformationGuidesMode m_mode;
    
    public:
        SetTransformationGuidesMode(TransformationGuidesMode mode) throw() 
            : m_mode(mode) {}
    
        virtual void execute(RenderEngine *engine)
        {
            engine->set_transformation_guides_mode(m_mode);
        }
    };

    dispatch_action(new SetTransformationGuidesMode(mode));
}

void Renderer::set_transformation_guides_coord_sys(CoordinateSystem coord_sys) throw()
{
    class SetTransformationGuidesCoordSys : public SceneAction
    {
        CoordinateSystem m_coord_sys;
    
    public:
        SetTransformationGuidesCoordSys(CoordinateSystem coord_sys) throw() 
            : m_coord_sys(coord_sys) {}
    
        virtual void execute(RenderEngine *engine)
        {
            engine->set_transformation_guides_coord_sys(m_coord_sys);
        }
    };

    dispatch_action(new SetTransformationGuidesCoordSys(coord_sys));
}


void Renderer::set_transformation_guides_pivot(RotationCenterpoint pivot) throw()
{
    class SetTransformationGuidesPivot : public SceneAction
    {
        RotationCenterpoint m_pivot;
    
    public:
        SetTransformationGuidesPivot(RotationCenterpoint pivot) throw() 
            : m_pivot(pivot) {}
    
        virtual void execute(RenderEngine *engine)
        {
            engine->set_transformation_guides_pivot(m_pivot);
        }
    };

    dispatch_action(new SetTransformationGuidesPivot(pivot));
}


void Renderer::set_transformation_guides_axes(Axis axes) throw()
{
    class SetTransformationGuidesAxes : public SceneAction
    {
        Axis m_axes;
    
    public:
        SetTransformationGuidesAxes(Axis axes) throw() 
            : m_axes(axes) {}
    
        virtual void execute(RenderEngine *engine)
        {
            engine->set_transformation_guides_axes(m_axes);
        }
    };

    dispatch_action(new SetTransformationGuidesAxes(axes));
}
