/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_RENDERER_RENDERER_HH
#define __PEEKABOT_RENDERER_RENDERER_HH


#include <vector>
#include <stdexcept>
#include <boost/shared_ptr.hpp>
#include <boost/function.hpp>

#include "../Types.hh"
#include "../Matrix.hh"
#include "../Vector.hh"
#include "../IDAllocator.hh"
#include "../Singleton.hh"
#include "IDTypes.hh"
#include "RendererTypes.hh"


namespace peekabot
{
    namespace renderer
    {
        class RenderEngine;
        class SceneAction;
        class FrameCollector;


        /**
         * \brief Provides a public interface to an encapsulated rendering 
         * engine.
         *
         * \sa RenderEngine
         */
        class Renderer
        {
        public:
            Renderer() throw();

            virtual ~Renderer() throw();

            // Call after an OpenGL context is set up
            //
            // Must be called with an active and valid OpenGL context
            void initialize();
            
            bool is_initialized() const throw();


            /**
             * \exception std::runtime_error Thrown if the renderer couldn't
             * be initialized.
             */
            bool render(int32_t w, int32_t h, 
                        PrimitiveID camera,
                        bool *visible_layers, 
                        FrameCollector *collector = 0);

            /**
             * \exception std::runtime_error Thrown if the renderer couldn't
             * be initialized.
             */
            uint32_t pick(int32_t w, int32_t h, int32_t x, int32_t y, 
                          PrimitiveID camera, bool *visible_layers, 
                          real32 sensitivity);

            bool has_pending_updates() const throw();

            /**
             * \exception std::runtime_error Thrown if the renderer couldn't
             * be initialized.
             */
            void process_updates();



            
            /**
             * \brief Get the node ID of the Renderer's (RenderEngine's) root 
             * node.
             */
            NodeID get_root_node_id() const throw();

            /**
             * \brief Add a new node in the render engine.
             *
             * \param parent The ID of the new node's parent.
             * \return The ID of the added node.
             */
            NodeID add_node(NodeID parent) throw();

            /**
             * \brief Clone a node along with whatever is attached to it.
             */
            NodeID clone_node(NodeID id, NodeID parent) throw();

            /**
             * \brief Alter the transformation of the specified node.
             *
             * \pre The given \a id is a valid reference to a node.
             *
             * \param id The NodeID of the node to modify.
             * \param m The transformation matrix to use for the node.
             * \param coord_sys Specified whether the given matrix is given in 
             * parent, world or local coordiantes (in which case the 
             * transformation is relative to the current transformation).
             */
            void set_node_transformation(
                NodeID id, const Matrix4ru &m, 
                CoordinateSystem coord_sys = PARENT_COORDINATES) throw();

            /**
             * \brief Set the color of the specified node.
             *
             * \pre The given \a id is a valid reference to a node.
             *
             * \param id The NodeID of the node to modify.
             * \param color The node's new color.
             */
            void set_node_color(NodeID id, const RGBColor &color) throw();

            /**
             * \brief Set the opacity of the specified node.
             *
             * \pre The given \a id is a valid reference to a node.
             *
             * \param id The NodeID of the node to modify.
             * \param opacity The node's new opacity.
             */
            void set_node_opacity(NodeID id, const Opacity &opacity) throw();

            /**
             * \brief Set the scale of the specified node.
             *
             * \pre The given \a id is a valid reference to a node.
             *
             * \param id The NodeID of the node to modify.
             * \param x_scale The node's new scale factor along the X-axis.
             * \param y_scale The node's new scale factor along the Y-axis.
             * \param z_scale The node's new scale factor along the Z-axis.
             */
            void set_node_scale(
                NodeID id, real32 x_scale, 
                real32 y_scale, real32 z_scale) throw();

            /**
             * \brief Hide or show a node in the render engine.
             *
             * \pre The given \a id is a valid reference to a node.
             *
             * \param id The NodeID of the node to modify.
             * \param hide The node will be hidden and shown when this parameter
             * is set to \c true and \c false respectively.
             */
            void set_node_visibility(NodeID id, bool hide) throw();

            /**
             * \brief Change which layer the specified node belongs to.
             *
             * \pre The given \a id is a valid reference to a node.
             *
             * \param id The NodeID of the node to modify.
             * \param layer The node's new layer. Must be zero or larger.
             */
            void set_node_layer(NodeID id, uint32_t layer) throw();

            /**
             * \brief Set the "name" used when rendering the specified node.
             * The set "name" is what's returned by the renderer when the node
             * is picked.
             *
             * \pre The given \a id is a valid reference to a node.
             *
             * \param id The NodeID of the node to modify.
             * \param name The node's new name.
             */
            void set_node_name(NodeID id, uint32_t name) throw();

            /**
             * \brief Mark the specified node as selected.
             *
             * \pre The given \a id is a valid reference to a node.
             *
             * \param id The NodeID of the node to modify.
             */
            void select_node(NodeID id) throw();

            /**
             * \brief Mark the specified node as \e not selected.
             *
             * \pre The given \a id is a valid reference to a node.
             *
             * \param id The NodeID of the node to modify.
             */
            void deselect_node(NodeID id) throw();

            /**
             * \brief Remove specified node and all it's descendents for the 
             * renderer's scene.
             *
             * \pre The given \a id is a valid reference to a node.
             *
             * \param id The NodeID of the node being removed.
             */
            void remove_node(NodeID id) throw();

            /**
             * \brief Move the specified node to a different parent in the 
             * renderer's scene.
             *
             * The absolute (world) position/rotation of the node will not 
             * generally be preserved when a node is rearranged. I.e., the 
             * node's model-to-parent transformation will not be changed to 
             * keep it's model-to-world transformation constant.
             *
             * \pre All given NodeIDs are valid node references.
             *
             * \param id The NodeID of the node being removed.
             * \param new_parent The NodeID of the node's new parent node.
             */
            void rearrange_node(NodeID id, NodeID new_parent) throw();




            // Primitive creation methods:
            // anonymous = no primitive ID allocated, return value undefined
            PrimitiveID add_camera(NodeID parent, bool anonymous) throw();
            PrimitiveID add_sphere(NodeID parent, bool anonymous) throw();
            PrimitiveID add_cylinder(NodeID parent, bool anonymous) throw();
            PrimitiveID add_cube(NodeID parent, bool anonymous) throw();
            PrimitiveID add_model(NodeID parent, bool anonymous, const std::string &filename) throw();
            PrimitiveID add_circle(NodeID parent, bool anonymous) throw();
            PrimitiveID add_disc(NodeID parent, bool anonymous) throw();
            PrimitiveID add_point_blob(NodeID parent, bool anonymous, const std::vector<Vector3r> &vertices) throw();

            PrimitiveID add_point_mesh(NodeID parent, bool anonymous) throw();
            PrimitiveID add_line_mesh(NodeID parent, bool anonymous) throw();
            PrimitiveID add_tri_mesh(NodeID parent, bool anonymous) throw();

            PrimitiveID add_line_blob(
                NodeID parent, 
                bool anonymous, 
                const std::vector<Vector3r> &vertices) throw(std::logic_error);

            PrimitiveID add_polygon(NodeID parent, bool anonymous, const std::vector<Vector3r> &vertices) throw();
            // ... + polyline, polygon, cylinder, circle, disk, plane, (static) point set
            //PrimitiveID add_point_set(NodeID parent) throw();

            PrimitiveID add_label(NodeID parent, bool anonymous) throw();

            PrimitiveID add_occupancy_grid_2d(NodeID parent, bool anonymous, 
                                              float cell_size,
                                              const RGBColor &free_color, 
                                              const RGBColor &occupied_color) throw();

            void set_occupancy_grid_2d_cells(
                PrimitiveID id,
                const std::vector<std::pair<Vector2f, float> > &cells) 
                throw(std::domain_error);

            void remove_primitive(PrimitiveID id) throw();

            void rearrange_primitive(PrimitiveID id, NodeID new_parent) throw();


            void set_primitive_color(PrimitiveID id, const RGBColor &color) throw();
            void set_primitive_alpha(PrimitiveID id, real32 alpha) throw();
            void set_primitive_name(PrimitiveID id, uint32_t name) throw();
            void set_primitive_point_size(PrimitiveID id, real32 size) throw();
            void set_primitive_texture(PrimitiveID id, const std::string &filename) throw();
            void set_primitive_texture(PrimitiveID id, uint32_t w, uint32_t h, const void *tex, bool replace) throw();
            //void set_primitive_tex_coords(PrimitiveID id, size_t n, const void *tex_coords) throw();
            void set_primitive_depth_test(PrimitiveID id, bool enable) throw();
            void set_primitive_vertices(PrimitiveID id, const std::vector<Vector3r> &vertices) throw();
            void set_primitive_line_style(
                PrimitiveID id, LineStyle stipple_pattern, 
                int stipple_factor, float line_width) throw();
            void transform_primitive_geometry(
                PrimitiveID id, 
                const boost::function<Vector3r (const Vector4r &)> &fun) throw();
            void transform_primitive_geometry(PrimitiveID id, const Matrix4r &transform) throw();


            // ...
            void set_point_blob_points(PrimitiveID id, size_t n, const Vector3r *points);

            void set_camera_fov(PrimitiveID id, real32 fov) throw();
            void set_camera_view_mode(PrimitiveID id, bool orthographic) throw();
            void set_camera_near_plane(PrimitiveID id, real32 _near) throw();
            void set_camera_far_plane(PrimitiveID id, real32 _far) throw();
            void set_camera_zoom_distance(PrimitiveID id, float zoom_distance) throw();

            void set_label_text(PrimitiveID id, const std::string &text) throw();
            void set_label_alignment(PrimitiveID id, TextAlignment alignment) throw();

            //void add_point_set_point(PrimitiveID id, const Vector3r &p) throw();

            void set_transformation_guides_mode(TransformationGuidesMode mode) throw();
            void set_transformation_guides_coord_sys(CoordinateSystem coord_sys) throw();
            void set_transformation_guides_axes(Axis axes) throw();
            void set_transformation_guides_pivot(RotationCenterpoint pivot) throw();

        private:
            void dispatch_action(SceneAction *action) throw();

            static NodeID allocate_node_id() throw();

            static NodeID allocate_primitive_id() throw();

        private:
            bool m_is_initialized;

            RenderEngine *m_engine;

            NodeID m_root_node_id;

            static boost::shared_ptr<IDAllocator<NodeID> > ms_node_id_allocator;
            static boost::shared_ptr<IDAllocator<PrimitiveID> > ms_primitive_id_allocator;
        };

        typedef Singleton<Renderer> TheRenderer;

    }
}


#endif // __PEEKABOT_RENDERER_RENDERER_HH
