/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <iostream>
#include <cassert>
#include <exception>
#include <stack>

#include <GL/glew.h>

#include "SceneNode.hh"
#include "BoundingVolumes.hh"
#include "IDTranslator.hh"

#include "primitives/AxisTripod.hh"

using namespace peekabot;
using namespace peekabot::renderer;



static const RGBColor DEFAULT_COLOR(1, 1, 1);



SceneNode::SceneNode() throw()
    : m_parent(0), 
      m_layer(0),
      m_opacity(Opacity(false, 1.0f)),
      m_accum_opacity_is_valid(false),
      m_is_hidden(false),
      m_is_visible_is_valid(false),
      m_mtow_is_valid(false), 
      m_is_selected(false),
      m_has_selected_parent(false),
      m_has_selected_parent_is_valid(false),
      m_x_scale(1),
      m_y_scale(1),
      m_z_scale(1)
{
    m_mtop.load_identity();

    m_template_state.set(new statelets::Transformation(m_scaled_mtow));
}


SceneNode::~SceneNode()
{
    // Remove any aliases registered for this node
    TheIDTranslator::instance().remove_alias(this);
    clear_cullables();

    ChildContainer c = get_children();
    for( ChildContainer::iterator it = c.begin(); it != c.end(); ++it )
    {
        detach(*it);
        delete *it;
    }
}


SceneNode::ChildContainer &SceneNode::get_children() throw()
{
    return m_children;
}


const SceneNode::ChildContainer &SceneNode::get_children() const throw()
{
    return m_children;
}


void SceneNode::set_opacity(Opacity opacity) throw()
{
    // Invalidate the (cached) accumulated opacity for the entire subtree,
    // unless they're absolute!
    invalidate_accum_opacity(true);
    if( m_opacity.is_absolute )
        m_accum_opacity = opacity.opacity;

    m_opacity = opacity;

    on_opacity_set(this);
}



void SceneNode::invalidate_accum_opacity(bool invalidate_absolute) throw()
{
    if( (!m_opacity.is_absolute || invalidate_absolute) && 
        m_accum_opacity_is_valid )
    {
        m_accum_opacity_is_valid = false;

        for( ChildContainer::iterator it = m_children.begin();
             it != m_children.end(); ++it )
        {
            (*it)->invalidate_accum_opacity(false);
        }
    }
}


void SceneNode::invalidate_mtow() throw()
{
    invalidate_mtow_recursive();
}


void SceneNode::invalidate_mtow_recursive() throw()
{
    if( m_mtow_is_valid )
    {
        m_mtow_is_valid = false;

        for( ChildContainer::iterator it = m_children.begin();
             it != m_children.end(); ++it )
        {
            (*it)->invalidate_mtow_recursive();
        }
    }
}


bool SceneNode::has_valid_mtow_cache() const throw()
{
    return m_mtow_is_valid;
}


void SceneNode::invalidate_cache() throw()
{
    invalidate_mtow();
    invalidate_accum_opacity(true);
    invalidate_is_visible();
    invalidate_has_selected_parent();
}


void SceneNode::add_cullable(Cullable *cullable) throw()
{
    // Entities have exclusive ownership, enforce it!
    assert( cullable->get_parent() == 0 );

    m_cullables.insert(cullable);
    cullable->set_parent(this);
    cullable->calculate_bvs();

    on_cullable_added(this, cullable);
}


void SceneNode::remove_cullable(Cullable *cullable) throw()
{
    Cullables::iterator it = m_cullables.find(cullable);
   
    if( it != m_cullables.end() )
    {
        m_cullables.erase(it);
        cullable->set_parent(0);

        on_cullable_removed(this, cullable);

        delete cullable;
    }
}


void SceneNode::clear_cullables() throw()
{
    Cullables cs = m_cullables;
    for( Cullables::iterator it = cs.begin(); it != cs.end(); ++it )
    {
        remove_cullable(*it);
    }
}


const SceneNode::Cullables &SceneNode::get_cullables() const throw()
{
    return m_cullables;
}


Opacity SceneNode::get_opacity() const throw()
{
    return m_opacity;
}


real32 SceneNode::get_accum_opacity() throw()
{
    validate_accum_opacity();
    return m_accum_opacity;
}


void SceneNode::validate_accum_opacity() throw()
{
    if( !m_accum_opacity_is_valid )
    {
        // "Back" track until a node with valid opacity (or an orphaned)
        // node is found, and validate that path.
        //
        // There's a special case - the root.
        // It's accum opacity will be set to m_opacity.opacity.

        /// \todo Profile this code! Naively revalidating the entire subtree
        /// spanned by the first valid ancestor node might be faster in the long
        /// run.

        std::stack<SceneNode *> invalid;
        
        SceneNode *node = this;
        while( node && !node->m_accum_opacity_is_valid )
        {
            invalid.push(node);
            node = node->get_parent();
        }

        while( !invalid.empty() )
        {
            SceneNode *node = invalid.top();
            invalid.pop();

            if( node->get_parent() == 0 )
            {
                // Special case: the root node.
                node->m_accum_opacity = node->m_opacity.opacity;
            }
            else
            {
                // The parent _always_ has a valid accum opacity.. that we know.
                // And the invalid node should, really, never be absolute... but
                // lets be rigorous and check that anyway.
                node->m_accum_opacity = ( node->m_opacity.is_absolute ) ? 
                    node->m_opacity.opacity :
                    node->get_parent()->m_accum_opacity*node->m_opacity.opacity;
            }

            node->m_accum_opacity_is_valid = true;

            if( node->m_accum_opacity < 1.0f )
            {
                node->m_template_state.set(
                    boost::shared_ptr<Statelet>(
                        new statelets::BlendMode(
                            GL_CONSTANT_ALPHA, GL_ONE_MINUS_CONSTANT_ALPHA)));
                
                node->m_template_state.get<statelets::Alpha>()->set(node->m_accum_opacity);
                
                node->m_template_state.set(
                    boost::shared_ptr<Statelet>(
                        new statelets::Culling(false)));
                
                node->m_template_state.set(
                    boost::shared_ptr<Statelet>(
                        new statelets::LightModel(true)));
                
                node->m_template_state.set(
                    boost::shared_ptr<Statelet>(
                        new statelets::DepthMask(false)));
            }
            else
            {
                node->m_template_state.unset<statelets::BlendMode>();
                node->m_template_state.unset<statelets::Alpha>();
                node->m_template_state.unset<statelets::Culling>();
                node->m_template_state.unset<statelets::LightModel>();
                node->m_template_state.unset<statelets::DepthMask>();
            }
        }
    }
}



void SceneNode::set_scale(real32 sx, real32 sy, real32 sz) throw()
{
    m_x_scale = sx;
    m_y_scale = sy;
    m_z_scale = sz;  

    m_scaled_mtow = get_mtow();
    m_scaled_mtow(0) *= m_x_scale;
    m_scaled_mtow(1) *= m_y_scale;
    m_scaled_mtow(2) *= m_z_scale;

    calculate_cullable_bvs();

    on_scale_set(this);
}


real32 SceneNode::get_x_scale() const throw()
{
    return m_x_scale;
}


real32 SceneNode::get_y_scale() const throw()
{
    return m_y_scale;
}


real32 SceneNode::get_z_scale() const throw()
{
    return m_z_scale;
}


SceneNode *SceneNode::get_parent() throw()
{
    return m_parent;
}



const SceneNode *SceneNode::get_parent() const throw()
{
    return m_parent;
}



// Does the following:
// 1. Add to the parent
// 2. Set node->m_parent to this
// 3. Invalidate the nodes caches
// 4. Fires the on_child_attached signal
bool SceneNode::attach(SceneNode *node) throw()
{
    if( node->get_parent() != 0 )
        return true;

    if( m_children.insert(node).second )
    {
        node->m_parent = this;
        node->invalidate_cache();

        on_child_attached(this, node);

        return false;
    }
    else
        return true;
}


// Does the following:
// 1. Remove from parent
// 2. Set node->m_parent to 0
// 3. Invalidate the nodes caches
// 4. Fires the on_detached signal
bool SceneNode::detach(SceneNode *node) throw()
{
    ChildContainer::iterator it = m_children.find(node);

    if( it == m_children.end() )
        return true;
    else
    {
        m_children.erase(it);
        node->m_parent = 0;
        node->invalidate_cache();

        node->on_detached(node);
        
        return false;
    }
}


void SceneNode::detach() throw()
{
    if( get_parent() != 0 )
        get_parent()->detach(this);
    // else: it's the root! cause undefined behaviour! :)
}


uint32_t SceneNode::get_name() throw()
{
    return m_template_state.get<statelets::Name>()->get();
}


void SceneNode::set_name(uint32_t id) throw()
{
    //TODO: trigger(ON_SET_OBJECT_ID);

    m_template_state.get<statelets::Name>()->set(id);

    //TODO: trigger(ON_OBJECT_ID_SET);
}


void SceneNode::validate_mtow() const throw()
{
    // Find a path to the closest ancestor with an valid mtow,
    // then validate that path from top to bottom.
    std::stack<const SceneNode *> invalid;

    const SceneNode *node = this;
    while( node && !node->m_mtow_is_valid )
    {
        invalid.push(node);
        node = node->get_parent();
    }
    
    while( !invalid.empty() )
    {
        const SceneNode *node = invalid.top();
        invalid.pop();
      
        if( node->get_parent() == 0 )
        {
            // Special case: the root node.
            node->m_mtow = node->m_mtop;
        }
        else
        {
            // The parent _always_ has a valid mtow, since we just 
            // calculated it.
            node->m_mtow = node->m_parent->m_mtow * node->m_mtop;
        }

        node->m_mtow_is_valid = true;
        
        node->m_scaled_mtow = node->m_mtow;
        node->m_scaled_mtow(0) *= node->m_x_scale;
        node->m_scaled_mtow(1) *= node->m_y_scale;
        node->m_scaled_mtow(2) *= node->m_z_scale;
    }
}


const Matrix4r &SceneNode::get_mtow() const throw()
{
    if( !m_mtow_is_valid )
        validate_mtow();

    return m_mtow;
}


const Matrix4r &SceneNode::get_parent_mtow() const throw()
{
    return get_parent() ? get_parent()->get_mtow() : get_mtow();
}


Matrix4r SceneNode::get_unscaled_mtow() const throw()
{
    Matrix4r mtow = get_mtow();
    mtow(0).normalize();
    mtow(1).normalize();
    mtow(2).normalize();
    return mtow;
}


const Matrix4r &SceneNode::get_transformation() const throw()
{
    return m_mtop;
}


void SceneNode::set_transformation(const Matrix4r &m) throw()
{
    m_mtop = m;
    invalidate_mtow();

    calculate_cullable_bvs_recursive();

    on_transformation_set(this);
}


const Vector4r &SceneNode::get_pos() const throw()
{
    return m_mtop(3);
}


void SceneNode::set_pos(const Vector3r &pos) throw()
{
    Matrix4r mtop = get_transformation();
    mtop(3) = Vector4r(pos(0), pos(1), pos(2), 1);
    set_transformation(mtop);
}


const Vector4r &SceneNode::get_world_pos() const throw()
{
    return get_mtow()(3);
}


void SceneNode::hide() throw()
{
    if( !m_is_hidden )
    {
        m_is_hidden = true;
        m_is_visible = false;
        invalidate_is_visible();
        m_is_visible_is_valid = true;

        on_visibility_set(this);
    }
    // else: already hidden
}


void SceneNode::show() throw()
{
    if( m_is_hidden )
    {
        m_is_hidden = false;
        invalidate_is_visible();

        on_visibility_set(this);
    }
    // else: already shown
}


bool SceneNode::is_hidden() const throw()
{
    return m_is_hidden;
}


bool SceneNode::is_visible() throw()
{
    if( !m_is_visible_is_valid )
        validate_is_visible();

    return m_is_visible;
}


void SceneNode::invalidate_is_visible() throw()
{
    if( m_is_visible_is_valid )
    {
        m_is_visible_is_valid = false;

        for( ChildContainer::iterator it = m_children.begin();
             it != m_children.end(); ++it )
        {
            (*it)->invalidate_is_visible();
        }
    }
}


void SceneNode::validate_is_visible() throw()
{
    std::stack<SceneNode *> invalid;
        
    SceneNode *node = this;
    while( node && !node->m_is_visible_is_valid )
    {
        invalid.push(node);
        node = node->get_parent();
    }

    while( !invalid.empty() )
    {
        SceneNode *node = invalid.top();
        invalid.pop();

        if( node->get_parent() == 0 )
        {
            // Special case: the root node.
            node->m_is_visible = !m_is_hidden;
        }
        else
        {
            node->m_is_visible = (node->m_is_hidden) ? 
                false : node->get_parent()->m_is_visible;
        }

        node->m_is_visible_is_valid = true;
    }
}


void SceneNode::set_layer(uint32_t layer) throw()
{
    assert( layer < NUMBER_OF_LAYERS );

    m_layer = layer;
}


uint32_t SceneNode::get_layer() const throw()
{
    return m_layer;
}


void SceneNode::select() throw()
{
    if( !is_selected() )
    {
        m_is_selected = true;

        invalidate_has_selected_parent();

        on_selected(this);
    }
}


void SceneNode::deselect() throw()
{
    if( is_selected() )
    {
        m_is_selected = false;

        invalidate_has_selected_parent();

        on_deselected(this);
    }
}


bool SceneNode::is_selected() const throw()
{
    return m_is_selected;
}


void SceneNode::invalidate_has_selected_parent() throw()
{
    invalidate_has_selected_parent_recursive();
}


void SceneNode::invalidate_has_selected_parent_recursive() throw()
{
    if( m_has_selected_parent_is_valid )
    {
        m_has_selected_parent_is_valid = false;

        for( ChildContainer::iterator it = m_children.begin();
             it != m_children.end(); ++it )
        {
            (*it)->invalidate_has_selected_parent_recursive();
        }
    }
}


void SceneNode::validate_has_selected_parent() throw()
{
    std::stack<SceneNode *> invalid;
        
    SceneNode *node = this;
    while( node && !node->m_has_selected_parent_is_valid )
    {
        invalid.push(node);
        node = node->get_parent();
    }

    while( !invalid.empty() )
    {
        SceneNode *node = invalid.top();
        invalid.pop();

        if( node->get_parent() == 0 )
        {
            // Special case: the root node.
            node->m_has_selected_parent = false;
        }
        else
        {
            node->m_has_selected_parent = node->get_parent()->is_selected() || 
                node->get_parent()->has_selected_parent();
        }

        node->m_has_selected_parent_is_valid = true;
    }
}


bool SceneNode::has_selected_parent() throw()
{
    if( !m_has_selected_parent_is_valid )
        validate_has_selected_parent();

    return m_has_selected_parent;
}


void SceneNode::set_color(const RGBColor &color) throw()
{
    m_template_state.get<statelets::Color>()->set(color);
}


/*const RGBColor &SceneNode::get_color() const throw()
{
    return get_state().get<statelets::Color>()->get();
}*/


const State &SceneNode::get_template_state() const throw()
{
    return m_template_state;
}


void SceneNode::calculate_cullable_bvs() throw()
{
    for( Cullables::iterator it = m_cullables.begin();
         it != m_cullables.end(); ++it )
    {
        (*it)->calculate_bvs();
    }
}


void SceneNode::calculate_cullable_bvs_recursive() throw()
{
    calculate_cullable_bvs();

    for( ChildContainer::iterator it = m_children.begin();
         it != m_children.end(); ++it )
    {
        (*it)->calculate_cullable_bvs_recursive();
    }    
}


SceneNode *SceneNode::clone() const throw()
{
    SceneNode *node = new SceneNode;

    // Copy all attributes...
    node->m_mtop = m_mtop;
    node->m_layer = m_layer;
    node->m_opacity = m_opacity;
    node->m_x_scale = m_x_scale;
    node->m_y_scale = m_y_scale;
    node->m_z_scale = m_z_scale;
    node->m_is_hidden = m_is_hidden;
    // note: selected is not inherited

    // Note that we don't copy triggers!

    // *Clone* the node's entire state, i.e. copy ALL the state's statelets
    node->m_template_state = m_template_state;
    // Override all statelets referring to local members...
    node->m_template_state.set(new statelets::Transformation(node->m_scaled_mtow));

    // Clone child entities
    for( Cullables::const_iterator it = m_cullables.begin();
         it != m_cullables.end(); ++it )
    {
        Cullable *cullable = (*it)->clone();
        node->add_cullable(cullable);
    }

    return node;
}

SceneNode *SceneNode::clone_recursive() const throw()
{
    SceneNode *root = clone();

    // Clone children nodes
    for( ChildContainer::const_iterator it = m_children.begin();
         it != m_children.end(); ++it )
    {
        SceneNode *child = (*it)->clone_recursive();
        root->attach(child);
    }

    return root;
}
