/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_RENDERER_SCENE_NODE_HH
#define __PEEKABOT_RENDERER_SCENE_NODE_HH


#include <set>
#include <boost/utility.hpp>
#include <boost/thread_safe_signal.hpp>

#include "../Types.hh"
#include "../Matrix.hh"
#include "BoundingVolumes.hh"
#include "Renderable.hh"
#include "Cullable.hh"
#include "State.hh"


namespace peekabot
{
namespace renderer
{

    namespace primitives { class AxisTripod; }


    /**
     * \brief A node in the scene.
     *
     * Usually has a graphical representation, but really musn't 
     * (groups, for example).
     *
     * The keyword for SceneNode is caching. It caches everything it can, and 
     * does it lazily, validating its cache only when it's really needed.
     */
    class SceneNode : public boost::noncopyable
    {
        friend class Cullable;
           
    public:
        typedef std::set<SceneNode *> ChildContainer;

        typedef std::set<Cullable *> Cullables;


        SceneNode() throw();

        /**
         * \brief Destroy the scene node and any associated resources.
         *
         * Destroys child nodes recursively and deregisters all triggers.
         */
        virtual ~SceneNode();


        void add_cullable(Cullable *cullable) throw();

        void remove_cullable(Cullable *cullable) throw();

        void clear_cullables() throw();

        const Cullables &get_cullables() const throw();



        /**
         * \brief Calculate and cache the model to world space transform.
         *
         * After the model to world space cache has been calculated and stored,
         * the cache is marked as being valid. I.e. \c m_mtow_is_valid is set
         * to true.
         */
        void validate_mtow() const throw();


        void validate_accum_opacity() throw();


        /**
         * \brief Set the node's opacity.
         *
         * \sa \c Opacity
         *
         * \note Invalidates the accumulated opacity cache for all child nodes.
         */
        void set_opacity(Opacity opacity) throw();

        /**
         * \brief Get the node's opacity.
         *
         * \sa \c get_accum_opacity()
         */
        Opacity get_opacity() const throw();

        /**
         * \brief Get the node's accumulated opacity.
         *
         * For a discussion of the meaning of accumulated opacity, refer to the
         * documentation of \c Opacity.
         *
         * \return The node's accumulated (i.e. shown) opacity.
         *
         * \note Will used the cached accumulated opacity if available.
         * If unavailable, it will calculate it and cache it for later use
         * (internally, this is called \e validating the accumulated opacity
         * cache).
         */
        real32 get_accum_opacity() throw();


        /**
         * \brief Perform non-uniform scaling.
         *
         * Scales the mesh along the local coordinate axes.
         *
         * \param sx The scale factor along the local X-axis.
         * \param sy The scale factor along the local Y-axis.
         * \param sz The scale factor along the local Z-axis.
         */
        //!!!virtual void scale(real32 sx, real32 sy, real32 sz)
        //!!!{
        //!!!    if( *m_mesh != 0 )
        //!!!    {
        //!!!        if( m_mesh->get_ref_count() > 1 )
        //!!!            m_mesh = m_mesh->clone();
        //!!!        
        //!!!        m_mesh->scale(sx, sy, sz);
        //!!!    }
        //!!!}

        /**
         * \brief Set the scaling factor for each axis.
         *
         * Will execute <tt>scale(1 / m_x_scale * sx, ...)</tt>.
         * 
         * \param sx The scale factor along the local X-axis.
         * \param sy The scale factor along the local Y-axis.
         * \param sz The scale factor along the local Z-axis.
         */
        void set_scale(real32 sx, real32 sy, real32 sz) throw();

        real32 get_x_scale() const throw();

        real32 get_y_scale() const throw();

        real32 get_z_scale() const throw();



        /**
         * \brief Returns the node's parent.
         *
         * \return 0 if the node is the root node, otherwise the node's parent.
         */
        SceneNode *get_parent() throw();

        /**
         * \brief Returns the node's parent.
         *
         * \return 0 if the node is the root node, otherwise the node's parent.
         */
        const SceneNode *get_parent() const throw();

        /**
         * \brief Get the node's list of children.
         *
         * \return A container with all of the node's child nodes.
         */
        ChildContainer &get_children() throw();

        /**
         * \brief Get the node's list of children.
         *
         * \return A container with all of the node's child nodes.
         */
        const ChildContainer &get_children() const throw();

        /**
         * \brief Add a child node.
         *
         * This operation takes ownership of \c node.
         * If the operation succeedes it will fire all triggers
         * registered with the node, on the \c ON_CHILD_ATTACHED stimuli.
         *
         * \param node The node to add.
         * \return \c true if the node was added, \c false if the node was not
         * added because it is already a child of the node or the node was 
         * already attached to another node.
         *
         * \remarks A node should never be added to several nodes, since adding
         * it means relinquishing the previous owner of his ownership.
         * \remarks Will cause all cached data from the attached node's subtree 
         * to be invalidated.
         */
        bool attach(SceneNode *node) throw();

        /**
         * \brief Remove a child subtree.
         *
         * If the node given is not a child of the node, the operation will 
         * fail.
         *
         * <em>Performance:</em> This operation is guaranteed to not be of
         * worse complexity than linear in the number of child nodes.
         *
         * \param node The subtree to be removed.
         * \return \c true if the operation failed, \c false otherwise.
         *
         * \remarks Will cause all cached data from the detached node's subtree 
         * to be invalidated.
         *
         * \sa \c detach()
         */
        bool detach(SceneNode *node) throw();

        /**
         * \brief Detach the node (and thus its subtree) from its parent.
         *
         * Trying to detach the root node will cause undefined behaviour.
         *
         * \pre The root is \e not the root node.
         * \post The node has been detached.
         *
         * \remarks Will cause all cached data from the node's subtree to be 
         * invalidated.
         *
         * \sa detach(SceneNode *)
         */
        void detach() throw();

        uint32_t get_name() throw();

        void set_name(uint32_t id) throw();


        /**
         * \brief Returns true if the node's model to world transformation is 
         * cached.
         */
        bool has_valid_mtow_cache() const throw();

        /**
         * \brief Get the node's model to world coordinate space 
         * transformation.
         */
        const Matrix4r &get_mtow() const throw();

        /**
         * \brief Get the node's parent's model to world coordinate space 
         * transformation.
         *
         * If the node has no parent, its model-to-world transformation will
         * be returned instead.
         */
        const Matrix4r &get_parent_mtow() const throw();

        Matrix4r get_unscaled_mtow() const throw();

        /**
         * \brief Get the node's model to parent coordinate space 
         * transformation.
         */
        const Matrix4r &get_transformation() const throw();

        /**
         * \brief Set the node's model to parent coordinate space
         * transformation.
         *
         * \param m The new model to parent transform.
         */
        void set_transformation(const Matrix4r &m) throw();

        /**
         * \brief Get the node's position, relative its parent.
         */
        const Vector4r &get_pos() const throw();

        /**
         * \brief Set the node's position, relative its parent.
         */
        void set_pos(const Vector3r &pos) throw();

        /**
         * \brief Get the node's position in world coordinates.
         */
        const Vector4r &get_world_pos() const throw();

        /**
         * \brief Hide the node.
         *
         * If the visiblity state changed (i.e. it was not already hidden) this
         * operation will fire triggers registered on \c ON_HIDDEN.
         */
        void hide() throw();

        /**
         * \brief Unhide the node if it is hidden.
         *
         * If the visiblity state changed (i.e. it was not already shown) this
         * operation will fire triggers registered on \c ON_SHOWN.
         */
        void show() throw();

        /**
         * \brief Check whether or not the node is hidden.
         *
         * \return \c true if the node is hidden, \c false otherwise.
         */
        bool is_hidden() const throw();

        bool is_visible() throw();

        /**
         * \brief Set which layer the node is in.
         *
         * \pre \f$0<layer<NUMBER\_OF\_LAYERS\f$
         *
         * \param layer Which layer to put the node in.
         */
        void set_layer(uint32_t layer) throw();

        /**
         * \brief Get which layer the node is in.
         *
         * \return The node's layer, in the range \f$[0,NUMBER\_OF\_LAYERS)\f$.
         */
        uint32_t get_layer() const throw();

        
        void select() throw();

        void deselect() throw();

        bool is_selected() const throw();

        bool has_selected_parent() throw();
        
        void set_color(const RGBColor &color) throw();

        //const RGBColor &get_color() const throw();

        const State &get_template_state() const throw();

        SceneNode *clone() const throw();

        SceneNode *clone_recursive() const throw();

    public:
        //
        // Signals
        //
        
        boost::signal<void (SceneNode *parent, SceneNode *attached)> on_child_attached;

        boost::signal<void (SceneNode *)> on_detached;

        boost::signal<void (SceneNode *)> on_transformation_set;

        boost::signal<void (SceneNode *)> on_scale_set;

        boost::signal<void (SceneNode *)> on_visibility_set;

        boost::signal<void (SceneNode *)> on_selected;

        boost::signal<void (SceneNode *)> on_deselected;

        boost::signal<void (SceneNode *)> on_opacity_set;

        boost::signal<void (SceneNode *, Cullable *)> on_cullable_added;

        boost::signal<void (SceneNode *, Cullable *)> on_cullable_removed;




    private:
        /**
         * \brief Invalidate the model to world space transform cache 
         * recursively for the node's subtree.
         */
        void invalidate_mtow_recursive() throw();


        void invalidate_is_visible() throw();


        void validate_is_visible() throw();


        void invalidate_has_selected_parent() throw();


        void invalidate_has_selected_parent_recursive() throw();


        void validate_has_selected_parent() throw();

    protected:
        /**
         * \brief Invalidate the accumulated opacity cache for the node
         * and all of its descendents.
         *
         * <em>Running time:</em> Linear in the number of the node's 
         * descendents.
         *
         * \param invalidate_absolute Controls whether or not to invalidate the 
         * accumulated opacity cache for nodes with absolute opacity.
         */
        void invalidate_accum_opacity(bool invalidate_absolute) throw();

        /**
         * \brief Invalidate the model to world space transformation cache
         * for the node and all of its descendents.
         *
         * <em>Running time:</em> Linear in the number of the node's 
         * descendents.
         */
        void invalidate_mtow() throw();

        /**
         * \brief For the node and all of its descendents, invalidate all
         * cached data dependent on the node hierarchy.
         *
         * \remarks In effect, this will invalidate mtow and accumulated
         * opacity caches. However, one can expect that this function 
         * will invalidate a larger set of caches in the future.
         */
        void invalidate_cache() throw();



        void calculate_cullable_bvs() throw();

        void calculate_cullable_bvs_recursive() throw();


    private:
        /**
         * \brief The node's unique ID, used by outside modules to refer
         * to nodes.
         * 
         * Outside (as in other modules) users usually think nodes
         * are the same as objects, through this ID thick outside code can 
         * access an object's counterpart.
         *
         * \invariant <tt>m_id != 0</tt>
         */
        //ObjectID m_id;

        /**
         * \brief Model space to parent space transformation matrix.
         *
         * <em>Initial value:</em> The identity matrix.
         */
        Matrix4r m_mtop;

        /**
         * \brief Model space to world space transformation matrix.
         *
         * \c m_mtow is guaranteed to be valid only when \c m_mtow_is_valid
         * is \c true.
         * \c m_mtow is typically invalidated when the object, or an 
         * ancestor to it, is transformed.
         */
        mutable Matrix4r m_mtow;

        mutable Matrix4r m_scaled_mtow;

         /**
         * \brief References to all of the node's children.
         *
         * The initial value is an empty list.
         *
         * \note The container type used is subject to change.
         */
        ChildContainer m_children;

        /**
         * \brief The node's parent.
         *
         * Is null for the root node, otherwise always non-null. A null value
         * signifies a node with no parent.
         */
        SceneNode *m_parent;

        Cullables m_cullables;

        /**
         * \brief The layer the node is in.
         *
         * \pre \f$0<m\_layer<NUMBER\_OF\_LAYERS\f$
         *
         * <em>Initial value:</em> \c 0.
         *
         * Affects whether the visual representation associated with the node is 
         * rendered.
         * If a node's layer isn't visible when rendering it won't be drawn,
         * however, it doesn't affect the visibility of a child node.
         */
        uint32_t m_layer;

        /**
         * \brief The node's opacity.
         *
         * Affects all child nodes, unless overridden.
         *
         * <em>Initial value:</em> \c <false,1.0>, i.e. it inherits
         * it's parent's full opacity.
         *
         * \sa \c Opacity
         */
        Opacity m_opacity;

        /**
         * \brief The node's accumulated opacity value.
         *
         * The sum the node's opacity and all inherited opacity, unless
         * overridden.
         * This value of this member is only valid when 
         * \c m_accum_opacity_is_valid is \c true, and otherwise undefined.
         *
         * \sa Opacity
         */
        real32 m_accum_opacity;

        /**
         * \brief Denotes the validity of the \c m_accum_opacity member.
         *
         * For correct behaviour, this member must be set to false when a node's
         * accumulated opacity changes, unless it is recalculated to valid 
         * values on the fly. This may typically happen when the opacity
         * is set for a node or any of its ancestor nodes.
         *
         * <em>Initial value:</em> \c false
         */
        bool m_accum_opacity_is_valid:1;

        /**
         * \brief Whether or not the node and its subtree is hidden or shown.
         *
         * If this member is set to \c true, its and any descendent's mesh 
         * won't be drawn.
         *
         * <em>Initial value:</em> \c false.
         *
         * \remarks Setting it to \c true doesn't prevent child nodes from 
         * being shown, though it should be true that if a node has this member
         * set to true the corresponding member for all it's children should 
         * also be true (but this work should be carried out by the function
         * that hides and shows nodes respectively).
         */
        bool m_is_hidden:1;

        /**
         * \brief False if the node is hidden or indirectly hidden through 
         * a hidden ancestor node.
         *
         * Is set when a node is attached to another node (it looks to its 
         * parent to fetch the value) and when hide or show is called on
         * an ancestor node.
         *
         * <em>Initial value:</em> Undefined.
         */
        bool m_is_visible:1;

        bool m_is_visible_is_valid:1;


       /**
         * \brief Denotes whether \c m_mtow is valid or not.
         */
        mutable bool m_mtow_is_valid:1;


        // The mesh associated with the node, if any.
        // initial value = empty handle
        /**
         * \brief The mesh associate with the node.
         *
         * The mesh is usually the only visual element of a node, but depending
         * on the polymorphic behaviour of \c render() this might or might not 
         * be true.
         *
         * <em>Initial value:</em> An empty resource handle, referring to null.
         */
        //!!!Resource<Mesh> m_mesh;

        /**
         * \brief Signifies if the node is selected or not.
         *
         * A selected \e may be rendered differently than its unselected 
         * counterpart.
         */
        bool m_is_selected:1;

        /**
         * \brief Signifies whether any of the node's ancestors are selected.
         *
         * An indirectly selected \e may be rendered differently than its 
         * unselected counterpart.
         */
        bool m_has_selected_parent:1;


        bool m_has_selected_parent_is_valid:1;

        /**
         * \brief Scale factor along the local X-axis.
         *
         * The initial value is 1.
         */
        real32 m_x_scale;

        /**
         * \brief Scale factor along the local Y-axis.
         *
         * The initial value is 1.
         */
        real32 m_y_scale;

        /**
         * \brief Scale factor along the local Z-axis.
         *
         * The initial value is 1.
         */
        real32 m_z_scale;


        State m_template_state;
    };


}
}

#endif // __PEEKABOT_RENDERER_SCENE_NODE_HH
