/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/bind.hpp>
#include "SelectTrigger.hh"
#include "SceneNode.hh"
#include "primitives/AxisTripod.hh"


using namespace peekabot;
using namespace peekabot::renderer;


SelectTrigger::SelectTrigger(SceneNode *root) throw()
    : m_root(root)
{
    add_subtree(m_root);
}


SelectTrigger::~SelectTrigger()
{
    remove_subtree(m_root);
}


const SelectTrigger::SelectedSet &SelectTrigger::get_selected() const throw()
{
    return m_selected;
}


void SelectTrigger::on_child_attached(SceneNode *parent, SceneNode *attached) throw()
{
    add_subtree(attached);
}


void SelectTrigger::on_detached(SceneNode *node) throw()
{
    remove_subtree(node);
}


void SelectTrigger::add_subtree(SceneNode *node) throw()
{
    if( node->is_selected() )
        on_selected(node);

    node->on_child_attached.connect(
        boost::bind(&SelectTrigger::on_child_attached, this, _1, _2));

    node->on_detached.connect(
        boost::bind(&SelectTrigger::on_detached, this, _1));

    node->on_selected.connect(
        boost::bind(&SelectTrigger::on_selected, this, _1));

    node->on_deselected.connect(
        boost::bind(&SelectTrigger::on_deselected, this, _1));
   
    SceneNode::ChildContainer &children = node->get_children();
    for( SceneNode::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        add_subtree(*it);
    }
}


void SelectTrigger::remove_subtree(SceneNode *node) throw()
{
    if( node->is_selected() )
        on_deselected(node);

    node->on_child_attached.disconnect(
        boost::bind(&SelectTrigger::on_child_attached, this, _1, _2));

    node->on_detached.disconnect(
        boost::bind(&SelectTrigger::on_detached, this, _1));

    node->on_selected.disconnect(
        boost::bind(&SelectTrigger::on_selected, this, _1));

    node->on_deselected.disconnect(
        boost::bind(&SelectTrigger::on_deselected, this, _1));

    SceneNode::ChildContainer &children = node->get_children();
    for( SceneNode::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        remove_subtree(*it);
    }
}

void SelectTrigger::on_selected(SceneNode *node) throw()
{
    primitives::AxisTripod *tripod = new primitives::AxisTripod;
    node->add_cullable(tripod);

    m_tripods.insert( std::make_pair(node, tripod) );
    m_selected.insert(node);
}


void SelectTrigger::on_deselected(SceneNode *node) throw()
{
    TripodMap::iterator it = m_tripods.find(node);
    assert( it != m_tripods.end() );

    node->remove_cullable(it->second);

    m_tripods.erase(it);
    m_selected.erase(node);
}

