/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_RENDERER_SELECT_TRIGGER_HH
#define __PEEKABOT_RENDERER_SELECT_TRIGGER_HH


#include <set>
#include <map>
#include <boost/thread_safe_signal.hpp>


namespace peekabot
{
namespace renderer
{

    class SceneNode;
    class Cullable;

   
    class SelectTrigger
    {
    public:
        typedef std::set<SceneNode *> SelectedSet;

        typedef std::map<SceneNode *, Cullable *> TripodMap;

        SelectTrigger(SceneNode *root) throw();

        virtual ~SelectTrigger();


        const SelectedSet &get_selected() const throw();

    private:
        /**
         * \brief For \a node and all of its children, each node will be 
         * added to the octree and the SelectTrigger will register itself on it 
         * and all of its children.
         *
         * \param The subtree to add to the octree and register from the
         * select trigger.
         */
        void on_child_attached(SceneNode *parent, SceneNode *attached) throw();

        /**
         * \brief For \a node and all of its children, each node will be 
         * removed from the octree and deregistered from triggering the 
         * select trigger.
         *
         * Each node in the subtree will be deregistered for \e all the stimuli
         * it's been registered with.
         *
         * \param The subtree to remove from the octree and deregister from the
         * select trigger.
         */
        void on_detached(SceneNode *node) throw();


        void remove_subtree(SceneNode *node) throw();


        void add_subtree(SceneNode *node) throw();

        
        void on_selected(SceneNode *node) throw();


        void on_deselected(SceneNode *node) throw();


    private:
        SceneNode *m_root;

        SelectedSet m_selected;

        TripodMap m_tripods;
    };



}
}


#endif // __PEEKABOT_RENDERER_SELECT_TRIGGER_HH
