/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/bind.hpp>
#include "SyncTrigger.hh"
#include "SceneNode.hh"
#include "Cullable.hh"
#include "RenderEngine.hh"
#include "Configuration.hh"


using namespace peekabot;
using namespace peekabot::renderer;


/**
 * \brief A constant that determines how frequently, and under what premises
 * the octree is purged.
 *
 * Setting this value involves a lot of voodoo. And mind you, there is no 
 * optimal value for all occassions.
 *
 * - Low value: The octree will be purged often. This gives optimal traversal
 * performance, optimal memory usage but can be expensive in highly dynamic
 * scenes (an excess node can be reused when an object moves into the place
 * for that node).
 * - High value: Higher memory waste, potential increase in CPU overhead 
 * during traversal. Can be good if the scene is highly dynamic.
 *
 * The value will have to be set as a compromise. A value that favours 
 * dynamic scenes is probably wise, since the system load will be higher
 * under these circumstances. Thus, a high value may help to compensate 
 * and balance the load.
 */
const int DEFAULT_OCTREE_PURGE_THRESHOLD = 100;


SyncTrigger::SyncTrigger(
    SceneNode *root, 
    Octree &octree, 
    RenderEngine *engine) throw()
    : m_octree(octree),
      m_root(root)
{
    // Notify us when actions have been processed so we can
    // 1) update invalid mtows/accum. opacities
    // 2) update the octree
    engine->on_actions_processed.connect(
        boost::bind(&SyncTrigger::sync, this, _1));

    add_subtree(m_root);
}


SyncTrigger::~SyncTrigger()
{
    remove_subtree(m_root);
}


void SyncTrigger::cullable_added(SceneNode *node, Cullable *cullable) throw()
{
    m_octree.add_object(cullable);

    cullable->on_bv_set.connect(
        boost::bind(&SyncTrigger::queue_update, this, _1));
}


void SyncTrigger::cullable_removed(SceneNode *node, Cullable *cullable) throw()
{
    m_octree.remove_object(cullable);

    cullable->on_bv_set.disconnect(
        boost::bind(&SyncTrigger::queue_update, this, _1));
}


void SyncTrigger::queue_update(Cullable *cullable) throw()
{
    m_octree.queue_need_update(cullable);
}


void SyncTrigger::on_child_attached(SceneNode *parent, SceneNode *attached) throw()
{
    add_subtree(attached);
}



void SyncTrigger::queue_invalid_mtow(SceneNode *node) throw()
{
    m_invalid_mtows.insert(node);

    SceneNode::ChildContainer &children = node->get_children();
    for( SceneNode::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        queue_invalid_mtow(*it);
    }
}



void SyncTrigger::queue_invalid_accum_opacity(SceneNode *node) throw()
{
    m_invalid_accum_opacities.insert(node);

    SceneNode::ChildContainer &children = node->get_children();
    for( SceneNode::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        queue_invalid_accum_opacity(*it);
    }
}


void SyncTrigger::on_detached(SceneNode *node) throw()
{
    remove_subtree(node);
}


void SyncTrigger::visibility_changed(SceneNode *node) throw()
{
    if( node->is_visible() )
        on_shown(node);
    else
        on_hidden(node);
}


void SyncTrigger::on_hidden(SceneNode *node) throw()
{
    SceneNode::Cullables cullables = node->get_cullables();
    for( SceneNode::Cullables::iterator it = cullables.begin();
         it != cullables.end(); ++it )
    {
        (*it)->on_bv_set.disconnect(
            boost::bind(&SyncTrigger::queue_update, this, _1));

        m_octree.remove_object(*it);
    }


    node->on_child_attached.disconnect(
        boost::bind(&SyncTrigger::on_child_attached, this, _1, _2));

    node->on_detached.disconnect(
        boost::bind(&SyncTrigger::on_detached, this, _1));

    node->on_transformation_set.disconnect(
        boost::bind(&SyncTrigger::queue_invalid_mtow, this, _1));

    node->on_opacity_set.disconnect(
        boost::bind(&SyncTrigger::queue_invalid_accum_opacity, this, _1));

    node->on_cullable_added.disconnect(
        boost::bind(&SyncTrigger::cullable_added, this, _1, _2));

    node->on_cullable_removed.disconnect(
        boost::bind(&SyncTrigger::cullable_removed, this, _1, _2));


    // NOTE: ON_SHOWN was NOT deregistered!

    SceneNode::ChildContainer &children = node->get_children();
    for( SceneNode::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        remove_subtree(*it);
    }
}


void SyncTrigger::on_shown(SceneNode *node) throw()
{
    SceneNode::Cullables cullables = node->get_cullables();
    for( SceneNode::Cullables::iterator it = cullables.begin();
         it != cullables.end(); ++it )
    {
        (*it)->on_bv_set.connect(
            boost::bind(&SyncTrigger::queue_update, this, _1));

        m_octree.add_object(*it);
    }


    node->on_child_attached.connect(
        boost::bind(&SyncTrigger::on_child_attached, this, _1, _2));

    node->on_detached.connect(
        boost::bind(&SyncTrigger::on_detached, this, _1));

    node->on_transformation_set.connect(
        boost::bind(&SyncTrigger::queue_invalid_mtow, this, _1));

    node->on_opacity_set.connect(
        boost::bind(&SyncTrigger::queue_invalid_accum_opacity, this, _1));

    node->on_cullable_added.connect(
        boost::bind(&SyncTrigger::cullable_added, this, _1, _2));

    node->on_cullable_removed.connect(
        boost::bind(&SyncTrigger::cullable_removed, this, _1, _2));

    
    SceneNode::ChildContainer &children = node->get_children();
    for( SceneNode::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        add_subtree(*it);
    }
}


void SyncTrigger::add_subtree(SceneNode *node) throw()
{
    SceneNode::Cullables cullables = node->get_cullables();
    for( SceneNode::Cullables::iterator it = cullables.begin();
         it != cullables.end(); ++it )
    {
        (*it)->on_bv_set.connect(
            boost::bind(&SyncTrigger::queue_update, this, _1));

        m_octree.add_object(*it);
    }

    node->on_child_attached.connect(
        boost::bind(&SyncTrigger::on_child_attached, this, _1, _2));

    node->on_detached.connect(
        boost::bind(&SyncTrigger::on_detached, this, _1));

    node->on_transformation_set.connect(
        boost::bind(&SyncTrigger::queue_invalid_mtow, this, _1));

    node->on_visibility_set.connect(
        boost::bind(&SyncTrigger::visibility_changed, this, _1));

    node->on_opacity_set.connect(
        boost::bind(&SyncTrigger::queue_invalid_accum_opacity, this, _1));

    node->on_cullable_added.connect(
        boost::bind(&SyncTrigger::cullable_added, this, _1, _2));

    node->on_cullable_removed.connect(
        boost::bind(&SyncTrigger::cullable_removed, this, _1, _2));


    // Queue for validation just in case the added node has an 
    // invalid mtow/accum opacity
    m_invalid_mtows.insert(node);
    m_invalid_accum_opacities.insert(node);

    
    SceneNode::ChildContainer &children = node->get_children();
    for( SceneNode::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        add_subtree(*it);
    }
}


void SyncTrigger::remove_subtree(SceneNode *node) throw()
{
    SceneNode::Cullables cullables = node->get_cullables();
    for( SceneNode::Cullables::iterator it = cullables.begin();
         it != cullables.end(); ++it )
    {
        (*it)->on_bv_set.disconnect(
            boost::bind(&SyncTrigger::queue_update, this, _1));

        m_octree.remove_object(*it);
    }


    node->on_child_attached.disconnect(
        boost::bind(&SyncTrigger::on_child_attached, this, _1, _2));

    node->on_detached.disconnect(
        boost::bind(&SyncTrigger::on_detached, this, _1));

    node->on_transformation_set.disconnect(
        boost::bind(&SyncTrigger::queue_invalid_mtow, this, _1));

    node->on_visibility_set.disconnect(
        boost::bind(&SyncTrigger::visibility_changed, this, _1));

    node->on_opacity_set.disconnect(
        boost::bind(&SyncTrigger::queue_invalid_accum_opacity, this, _1));

    node->on_cullable_added.disconnect(
        boost::bind(&SyncTrigger::cullable_added, this, _1, _2));

    node->on_cullable_removed.disconnect(
        boost::bind(&SyncTrigger::cullable_removed, this, _1, _2));

    // Remove detached nodes from the invalid mtow/opacity sets
    m_invalid_mtows.erase(node);
    m_invalid_accum_opacities.erase(node);


    SceneNode::ChildContainer &children = node->get_children();
    for( SceneNode::ChildContainer::iterator it = children.begin();
         it != children.end(); ++it )
    {
        remove_subtree(*it);
    }
}


void SyncTrigger::sync(RenderEngine *) throw()
{
    // 1. Validate mtows & accum opacities
    validate_mtows();
    validate_accum_opacities();

    // 2. Sync the octree against the scene
    m_octree.process_updates();

    // 3. Purge the octree (once in a while)
    int purge_threshold = TheConfiguration::instance().get_option<int>(
        "renderer.octree.purge_threshold", DEFAULT_OCTREE_PURGE_THRESHOLD);

    if( m_octree.get_excess_node_count() > (unsigned int)purge_threshold )
        m_octree.purge();
}


void SyncTrigger::validate_mtows() throw()
{
    for( InvalidSet::iterator it = m_invalid_mtows.begin();
         it != m_invalid_mtows.end(); ++it )
    {
        (*it)->validate_mtow();
    }

    m_invalid_mtows.clear();
}


void SyncTrigger::validate_accum_opacities() throw()
{
    for( InvalidSet::iterator it = m_invalid_accum_opacities.begin();
         it != m_invalid_accum_opacities.end(); ++it )
    {
        (*it)->validate_accum_opacity();
    }

    m_invalid_accum_opacities.clear();
}
