/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_RENDERER_SYNC_TRIGGER_HH
#define __PEEKABOT_RENDERER_SYNC_TRIGGER_HH


#include <set>
#include <boost/thread_safe_signal.hpp>
#include "Octree.hh"


namespace peekabot
{
namespace renderer
{

    class SceneNode;
    class RenderEngine;


    /**
     * \brief A trigger responsible for ensuring that all auxillary data 
     * structures are up to date prior to rendering.
     *
     * Automatically registers and deregisters itself to nodes being attached 
     * and detached respectively. It merely needs to be hooked up to the root 
     * node to rock'n'roll!
     *
     * It automatically synchronizes changes from the scene after actions has 
     * been processed by the RenderEngine. No manual intervention is needed.
     *
     * When it detects that scene actions has been processed, it validates
     * all model-to-world transformations, accumulated opacities and updates the
     * Octree.
     */
    class SyncTrigger
    {
    public:
        /**
         * \brief Construct a trigger to keep the given octree and 
         * translation table up to date.
         *
         * \remarks The root node of the scene must be registered manually
         * with the \c SyncTrigger.
         *
         * \param octree The octree to keep current.
         */
        SyncTrigger(SceneNode *root, Octree &octree, RenderEngine *engine) throw();

        virtual ~SyncTrigger();

    private:
        /**
         * \brief For \a node and all of its children, each node will be 
         * added to the octree and the SyncTrigger will register itself on it 
         * and all of its children.
         *
         * \param The subtree to add to the octree and register from the
         * sync trigger.
         */
        void on_child_attached(SceneNode *parent, SceneNode *attached) throw();

        /**
         * \brief For \a node and all of its children, each node will be 
         * removed from the octree and deregistered from triggering the 
         * sync trigger.
         *
         * Each node in the subtree will be deregistered for \e all the stimuli
         * it's been registered with.
         *
         * \param The subtree to remove from the octree and deregister from the
         * sync trigger.
         */
        void on_detached(SceneNode *node) throw();


        void queue_invalid_mtow(SceneNode *node) throw();


        void queue_invalid_accum_opacity(SceneNode *node) throw();


        void visibility_changed(SceneNode *node) throw();


        void on_shown(SceneNode *node) throw();


        void on_hidden(SceneNode *node) throw();


        void remove_subtree(SceneNode *node) throw();


        void add_subtree(SceneNode *node) throw();


        void cullable_added(SceneNode *node, Cullable *cullable) throw();


        void cullable_removed(SceneNode *node, Cullable *cullable) throw();


        void queue_update(Cullable *cullable) throw();


        /**
         * \brief Synchronize auxillary data structures with the recorded
         * changes from the Scene.
         */
        void sync(RenderEngine *) throw();

        /**
         * \brief Validate all invalid model-to-world transformations.
         */
        void validate_mtows() throw();

        /**
         * \brief Validate all invalid accumulated opacitites.
         */
        void validate_accum_opacities() throw();


    private:
        /**
         * \brief The octree which to forward scene changes too.
         */
        Octree &m_octree;

        SceneNode *m_root;

        typedef std::set<SceneNode *> InvalidSet;

        /**
         * \brief The set of SceneNode's whose stored model-to-world 
         * transformation is invalid.
         */
        InvalidSet m_invalid_mtows;

        /**
         * \brief The set of SceneNode's whose accumulated opacity is invalid.
         */
        InvalidSet m_invalid_accum_opacities;
    };



}
}


#endif // __PEEKABOT_RENDERER_SYNC_TRIGGER_HH
