/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cassert>
#include "Tessellator.hh"
#include "../MessageHub.hh"

using namespace peekabot;
using namespace peekabot::renderer;


Tessellator::Tessellator(EmitFaceFunc emit_func)
    : m_tess(gluNewTess()),
      m_emit_func(emit_func)
{
    gluTessProperty(m_tess, GLU_TESS_BOUNDARY_ONLY, GL_FALSE);
    gluTessProperty(m_tess, GLU_TESS_WINDING_RULE, GLU_TESS_WINDING_ODD);
    
    gluTessCallback(
        m_tess, GLU_TESS_BEGIN_DATA, 
        reinterpret_cast<GluTessCallbackType>(
            &Tessellator::begin_callback));

    gluTessCallback(
        m_tess, GLU_TESS_END_DATA, 
        reinterpret_cast<GluTessCallbackType>(
            &Tessellator::end_callback));

    gluTessCallback(
        m_tess, GLU_TESS_VERTEX_DATA, 
        reinterpret_cast<GluTessCallbackType>(
            &Tessellator::tess_callback));

    gluTessCallback(
        m_tess, GLU_TESS_COMBINE_DATA, 
        reinterpret_cast<GluTessCallbackType>(
            &Tessellator::combine_callback));

    gluTessCallback(
        m_tess, GLU_TESS_ERROR, 
        reinterpret_cast<GluTessCallbackType>(
            &Tessellator::error_callback));
}

Tessellator::~Tessellator()
{
    gluDeleteTess(m_tess);

    for( size_t i = 0; i < m_verts.size(); ++i )
        delete m_verts[i];
    m_verts.clear();
}

void Tessellator::begin_polygon()
{
    m_contour_count = 0;
    gluTessBeginPolygon(m_tess, this);
}

void Tessellator::end_polygon()
{
    if( m_contour_count > 0 )
        end_contour();

    gluTessEndPolygon(m_tess);

    for( size_t i = 0; i < m_verts.size(); ++i )
        delete m_verts[i];
    m_verts.clear();
}

void Tessellator::begin_contour()
{
    if( m_contour_count > 0 )
        end_contour();
    ++m_contour_count;
        
    gluTessBeginContour(m_tess);
}

void Tessellator::end_contour()
{
    --m_contour_count;
    gluTessEndContour(m_tess);
}

void Tessellator::vertex(float x, float y, float z)
{
    GLdouble coords[] = { x, y, z };
    gluTessVertex(m_tess, coords, (void *)add_vertex(x,y,z));
}

Tessellator::Vertex *Tessellator::add_vertex(float x, float y, float z)
{
    Vertex *v = new Vertex;
    v->x = x;
    v->y = y;
    v->z = z;
    v->index = m_verts.size();
    m_verts.push_back(v);
    return v;
}

void Tessellator::triangulate_fan(const std::vector<Vertex *> &v)
{
    assert( v.size() > 2 );
    size_t n = v.size()-1;
    for( size_t i = 1; i < n; ++i )
    {
        if( is_ccw(v[0], v[i], v[i+1]) )
        {
            m_emit_func(
                Vector3f(v[0]->x, v[0]->y, v[0]->z),
                Vector3f(v[i]->x, v[i]->y, v[i]->z),
                Vector3f(v[i+1]->x, v[i+1]->y, v[i+1]->z));
        }
        else
        {
            m_emit_func(
                Vector3f(v[0]->x, v[0]->y, v[0]->z),
                Vector3f(v[i+1]->x, v[i+1]->y, v[i+1]->z),
                Vector3f(v[i]->x, v[i]->y, v[i]->z));
        }
    }
}

void Tessellator::triangulate_strip(const std::vector<Vertex *> &v)
{
    assert( v.size() > 2 );
    size_t n = v.size()-1;
    for( size_t i = 1; i < n; ++i )
    {
        if( is_ccw(v[i-1], v[i], v[i+1]) )
        {
            m_emit_func(
                Vector3f(v[i-1]->x, v[i-1]->y, v[i-1]->z),
                Vector3f(v[i]->x, v[i]->y, v[i]->z),
                Vector3f(v[i+1]->x, v[i+1]->y, v[i+1]->z));
        }
        else
        {
            m_emit_func(
                Vector3f(v[i-1]->x, v[i-1]->y, v[i-1]->z),
                Vector3f(v[i+1]->x, v[i+1]->y, v[i+1]->z),
                Vector3f(v[i]->x, v[i]->y, v[i]->z));
        }
    }
}

void Tessellator::triangulate_triangles(const std::vector<Vertex *> &v)
{
    assert( v.size() > 2 );
    assert( v.size() % 3 == 0 );
    for( size_t i = 0; i < v.size(); i += 3 )
    {
        if( is_ccw(v[i+0], v[i+1], v[i+2]) )
        {
            m_emit_func(
                Vector3f(v[i+0]->x, v[i+0]->y, v[i+0]->z),
                Vector3f(v[i+1]->x, v[i+1]->y, v[i+1]->z),
                Vector3f(v[i+2]->x, v[i+2]->y, v[i+2]->z));
        }
        else
        {
            m_emit_func(
                Vector3f(v[i+0]->x, v[i+0]->y, v[i+0]->z),
                Vector3f(v[i+2]->x, v[i+2]->y, v[i+2]->z),
                Vector3f(v[i+1]->x, v[i+1]->y, v[i+1]->z));
        }
    }
}


void Tessellator::tess_callback(Vertex *vertex_data, void *polygon_data)
{
    Tessellator &tess = *(Tessellator *)polygon_data;
    tess.m_q.push_back(vertex_data);
}

void Tessellator::combine_callback(
    GLdouble coords[3],
    Vertex *vertex_data[4],
    GLfloat weight[4],
    void **outdata,
    void *polygon_data)
{
    Tessellator &tess = *(Tessellator *)polygon_data;
    *outdata = (void *)tess.add_vertex(coords[0], coords[1], coords[2]);
}


void Tessellator::begin_callback(GLenum type, void *polygon_data)
{
    Tessellator &tess = *(Tessellator *)polygon_data;
    tess.m_type = type;
}

void Tessellator::end_callback(void *polygon_data)
{
    Tessellator &tess = *(Tessellator *)polygon_data;

    if( tess.m_type == GL_TRIANGLES )
        tess.triangulate_triangles(tess.m_q);
    else if( tess.m_type == GL_TRIANGLE_FAN )
        tess.triangulate_fan(tess.m_q);
    else if( tess.m_type == GL_TRIANGLE_STRIP )
        tess.triangulate_strip(tess.m_q);
    else
        assert( false );

    tess.m_q.clear();
}

void Tessellator::error_callback(GLenum _errno)
{
    TheMessageHub::instance().publish(
        ERROR_MESSAGE,
        "Tessellation error: %s (%s:%u)", 
        gluErrorString(_errno), __FILE__, __LINE__);
}
