/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_RENDERER_TESSELLATOR_HH
#define __PEEKABOT_RENDERER_TESSELLATOR_HH


#include <GL/glew.h>
#include <stdexcept>
#include <vector>
#include <boost/function.hpp>
#include "../Vector.hh"


namespace peekabot
{
    namespace renderer
    {
        class Tessellator
        {
            typedef void (*GluTessCallbackType)();

            struct Vertex
            {
                float x, y, z;
                uint32_t index;
            };

        public:
            typedef boost::function<void (
                const Vector3f &,
                const Vector3f &,
                const Vector3f &)> EmitFaceFunc;

            Tessellator(EmitFaceFunc emit_func);

            ~Tessellator();

            void begin_polygon();

            void end_polygon();

            void begin_contour();

            void vertex(float x, float y, float z);

            Vector3f last_vertex() const
            {
                if( m_verts.empty() )
                    throw std::logic_error("No vertices added");

                Vertex *v = m_verts.back();
                return Vector3f(v->x, v->y, v->z);
            }

            size_t vertices_input() const throw()
            {
                return m_verts.size() + m_q.size();
            }

        private:
            void end_contour();

            Vertex *add_vertex(float x, float y, float z);

            void triangulate_fan(const std::vector<Vertex *> &v);

            void triangulate_strip(const std::vector<Vertex *> &v);

            void triangulate_triangles(const std::vector<Vertex *> &v);

            inline static bool is_ccw(Vertex *v1, Vertex *v2, Vertex *v3) throw()
            {
                return (v2->y-v1->y)*(v3->x-v1->x) < (v3->y-v1->y)*(v2->x-v1->x);
                // else, CW or colinear
            }


        private:
            GLUtesselator *m_tess;
            std::vector<Vertex *> m_verts;

            std::vector<Vertex *> m_q;
            GLenum m_type;

            EmitFaceFunc m_emit_func;

            size_t m_contour_count;

        private:
            static void tess_callback(Vertex *vertex_data, void *polygon_data);

            static void combine_callback(
                GLdouble coords[3],
                Vertex *vertex_data[4],
                GLfloat weight[4],
                void **outdata,
                void *polygon_data);

            static void begin_callback(GLenum type, void *polygon_data);

            static void end_callback(void *polygon_data);

            static void error_callback(GLenum _errno);
        };
    }
}


#endif // __PEEKABOT_RENDERER_TRI_TESSELLATOR_HH
