/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PEEKABOT_RENDERER_PRIMITIVES_CUBE_HH
#define __PEEKABOT_RENDERER_PRIMITIVES_CUBE_HH


#include <GL/glew.h>
#include "../Cullable.hh"
#include "../Renderable.hh"
#include "../SceneNode.hh"


namespace peekabot
{
namespace renderer
{
    namespace primitives
    {

        class Cube : public Cullable, public Renderable
        {
            float m_width;
            float m_height;
            float m_depth;

        protected:
            virtual void calculate_bvs() throw()
            {
                float sw = get_width() * get_parent()->get_x_scale();
                float sh = get_height() * get_parent()->get_z_scale();
                float sd = get_depth() * get_parent()->get_y_scale();

                float s = std::max(sw, std::max(sh, sd));
                set_bounding_sphere(BoundingSphere(get_parent()->get_world_pos(),
                                                   s*sqrtf(3)/2));
            }
            
        public:
            Cube(float width = 1.0f, float height = 1.0f, float depth = 1.0f)
                : m_width(width),
                  m_height(height),
                  m_depth(depth) {}

            virtual Cube *clone() const throw()
            {
                return new Cube(m_width, m_height, m_depth);
            }

            virtual void get_renderables(
                Camera *camera, EnumerationCallback callback) throw()
            {
                callback(this);
            }

            virtual void render(RenderContext &context) throw()
            {
                real32 x = m_width/2;
                real32 y = m_depth/2;
                real32 z = m_height/2;
                
                glBegin(GL_QUADS);
                
                
                glNormal3f(0, 1, 0);
                glVertex3f( x, y,-z);			// Top Right Of The Quad (Top)
                glVertex3f(-x, y,-z);			// Top Left Of The Quad (Top)
                glVertex3f(-x, y, z);			// Bottom Left Of The Quad (Top)
                glVertex3f( x, y, z);			// Bottom Right Of The Quad (Top)
                
                
                glNormal3f(0, -1, 0);
                glVertex3f( x,-y, z);			// Top Right Of The Quad (Bottom)
                glVertex3f(-x,-y, z);			// Top Left Of The Quad (Bottom)
                glVertex3f(-x,-y,-z);			// Bottom Left Of The Quad (Bottom)
                glVertex3f( x,-y,-z);			// Bottom Right Of The Quad (Bottom)
            
		
                glNormal3f(0, 0, 1);
                glVertex3f( x, y, z);			// Top Right Of The Quad (Front)
                glVertex3f(-x, y, z);			// Top Left Of The Quad (Front)
                glVertex3f(-x,-y, z);			// Bottom Left Of The Quad (Front)
                glVertex3f( x,-y, z);			// Bottom Right Of The Quad (Front)


                glNormal3f(0, 0, -1);
                glVertex3f( x,-y,-z);			// Bottom Left Of The Quad (Back)
                glVertex3f(-x,-y,-z);			// Bottom Right Of The Quad (Back)
                glVertex3f(-x, y,-z);			// Top Right Of The Quad (Back)
                glVertex3f( x, y,-z);			// Top Left Of The Quad (Back)


                glNormal3f(-1, 0, 0);
                glVertex3f(-x, y, z);			// Top Right Of The Quad (Left)
                glVertex3f(-x, y,-z);			// Top Left Of The Quad (Left)
                glVertex3f(-x,-y,-z);			// Bottom Left Of The Quad (Left)
                glVertex3f(-x,-y, z);			// Bottom Right Of The Quad (Left)

                glNormal3f(1, 0, 0);
                glVertex3f( x, y,-z);			// Top Right Of The Quad (Right)
                glVertex3f( x, y, z);			// Top Left Of The Quad (Right)
                glVertex3f( x,-y, z);			// Bottom Left Of The Quad (Right)
                glVertex3f( x,-y,-z);			// Bottom Right Of The Quad (Right)

                glEnd();
            }

            float get_width() const throw() { return m_width; }

            float get_height() const throw() { return m_height; }

            float get_depth() const throw() { return m_depth; }

            void set_width(float x) throw()
            {
                m_width = x;
                calculate_bvs();
            }

            void set_height(float x) throw()
            {
                m_height = x;
                calculate_bvs();
            }

            void set_depth(float x) throw()
            {
                m_depth = x;
                calculate_bvs();
            }
        };


    }
}
}


#endif // __PEEKABOT_RENDERER_PRIMITIVES_CUBE_HH
