/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <cassert>
#include <cmath>
#include <iostream>

#include "MeshBased.hh"
#include "Camera.hh"
#include "../SceneNode.hh"
#include "../MeshTransformer.hh"


using namespace peekabot;
using namespace peekabot::renderer;
using namespace peekabot::renderer::primitives;


MeshBased::MeshBased(uint8_t lod_count, 
                     const BoundingSphere &bsphere, 
                     real32 hysteresis) throw()
    : m_base_bsphere(bsphere),
      m_n_lods(lod_count),
      m_curr_lod(0),
      m_hysteresis((uint8_t)roundf(100-hysteresis*100)),
      m_lods(new LOD[m_n_lods])
{
}

MeshBased::~MeshBased()
{
    delete[] m_lods;
}

MeshBased::MeshBased(const MeshBased &x) throw()
    : m_base_bsphere(x.m_base_bsphere),
      m_n_lods(x.m_n_lods),
      m_curr_lod(0),
      m_hysteresis(x.m_hysteresis),
      m_lods(new LOD[m_n_lods])
{
    for( int i = 0; i < m_n_lods; ++i )
        set_lod(i, x.m_lods[i].m_threshold, x.m_lods[i].m_mesh);
}

MeshBased *MeshBased::clone() const throw()
{
    return new MeshBased(*this);
}

void MeshBased::set_lod(uint8_t lod_no, uint16_t coverage, const Mesh *mesh) throw()
{
    assert( lod_no < m_n_lods );
    
    if( m_lods[lod_no].m_mesh != 0 )
        delete m_lods[lod_no].m_mesh;
    
    m_lods[lod_no].m_threshold = coverage;
    m_lods[lod_no].m_mesh = mesh->clone();
}

void MeshBased::get_renderables(
    Camera *camera, EnumerationCallback callback) throw()
{  
    // Vertical screen coverage, in pixels
    float cov;
    float r = get_bounding_sphere().get_radius();
    
    if( camera->is_orthographic() )
    {
        // In this case, the coverage depends only on zoom distance and the
        // object's size - and the viewport height, of course.
        //
        // Note that it *does not* depend on viewer-object distance!
        
        real32 d = camera->get_zoom_distance();
        real32 half_h = d*tanf(camera->get_fov_rad()/2);

        cov = camera->get_viewport_height() * (r / half_h);
    }
    else
    {
        // Calculate pp the vector between the viewer and the object position,
        // projected on the view vectorx
        Vector4r pp(Vector4r(get_bounding_sphere().get_pos(),1) -
                    camera->get_viewer_world_pos());
        pp = pp*camera->get_view_vector() * camera->get_view_vector();
        
        // The x-distance between the camera view point and the object
        float dist = pp.norm();

        cov = camera->get_viewport_height() * (
            r * camera->get_near_plane()/dist / 
            (camera->get_near_plane()*tanf(camera->get_fov()/2/180*M_PI)));
    }
    
    if( cov < m_lods[m_curr_lod].m_threshold*m_hysteresis/100 )
        while( cov < m_lods[m_curr_lod].m_threshold*m_hysteresis/100 && m_curr_lod > 0 )
            --m_curr_lod;
    else if( m_curr_lod < m_n_lods-1 && cov >= m_lods[m_curr_lod+1].m_threshold )
        while( m_curr_lod < m_n_lods-1 && cov >= m_lods[m_curr_lod+1].m_threshold )
            ++m_curr_lod;
    
    assert( m_curr_lod < m_n_lods );
    assert( m_lods[m_curr_lod].m_mesh != 0 );
    
    callback(m_lods[m_curr_lod].m_mesh);
}

void MeshBased::calculate_bvs() throw()
{
    float s = std::max(
        get_parent()->get_x_scale(), 
        std::max(get_parent()->get_y_scale(), 
                 get_parent()->get_z_scale()));

    Vector3r pos(Vector4r(m_base_bsphere.get_pos()*s, 0) + 
                 get_parent()->get_world_pos());
    
    set_bounding_sphere(
        BoundingSphere(
            pos,
            s*m_base_bsphere.get_radius()));
}


void MeshBased::transform_geometry(
    const boost::function<Vector3r (const Vector4r &)> &fun) throw()
{
    for( uint8_t i = 0; i < m_n_lods; ++i )
    {
        if( m_lods[i].m_mesh )
            MeshTransformer::transform_geometry(
                *m_lods[i].m_mesh, fun,
                &m_base_bsphere, true);
    }

    if( get_parent() )
        calculate_bvs();
}
