/*
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <GL/glew.h>
#include <boost/scoped_array.hpp>
#include <boost/bind.hpp>
#include <boost/ref.hpp>

#include "Polygon.hh"
#include "../State.hh"
#include "../SceneNode.hh"
#include "../../GeometryToolbox.hh"
#include "../Tessellator.hh"


using namespace peekabot;
using namespace peekabot::renderer;
using namespace peekabot::renderer::primitives;


namespace
{
    void tess_callback(
        std::vector<float> &vertex_data,
        std::vector<float> &norms,
        const Vector3f &v1,
        const Vector3f &v2,
        const Vector3f &v3)
    {
        Vector3f n;

        // CCW
        vertex_data.push_back(v1(0));
        vertex_data.push_back(v1(1));
        vertex_data.push_back(v1(2));

        vertex_data.push_back(v2(0));
        vertex_data.push_back(v2(1));
        vertex_data.push_back(v2(2));

        vertex_data.push_back(v3(0));
        vertex_data.push_back(v3(1));
        vertex_data.push_back(v3(2));

        n = (v2-v1).cross_product(v3-v1);
        n.normalize();
        for( int i = 0; i < 3; ++i )
            for( int j = 0; j < 3; ++j )
                norms.push_back(n(j));

        // CW
        vertex_data.push_back(v1(0));
        vertex_data.push_back(v1(1));
        vertex_data.push_back(v1(2));

        vertex_data.push_back(v3(0));
        vertex_data.push_back(v3(1));
        vertex_data.push_back(v3(2));

        vertex_data.push_back(v2(0));
        vertex_data.push_back(v2(1));
        vertex_data.push_back(v2(2));

        n = (v3-v1).cross_product(v2-v1);
        n.normalize();
        for( int i = 0; i < 3; ++i )
            for( int j = 0; j < 3; ++j )
                norms.push_back(n(j));
    }
}


namespace
{
    static inline boost::arg<10> _10() { return boost::arg<10>(); }
}


Polygon::Polygon()
{
}


Polygon::Polygon(const Polygon &x)
    : VertexBased(x),
      m_mesh(x.m_mesh)
{
}


Polygon *Polygon::clone() const throw()
{
    return new Polygon(*this);
}

void Polygon::get_renderables(
    Camera *camera, EnumerationCallback callback) throw()
{
    callback(&m_mesh);
}

void Polygon::set_vertices(const VertexBased::Vertices &vertices) throw()
{
    if( vertices.size() <= 3 )
    {
        // TODO
    }


    std::vector<float> vertex_data, norms;

    Tessellator tess(
        boost::bind(&tess_callback, 
                    boost::ref(vertex_data),
                    boost::ref(norms),
                    _1, _2, _3));

    tess.begin_polygon();
    tess.begin_contour();
    for( size_t i = 0; i < vertices.size(); ++i )
        tess.vertex( vertices[i](0), vertices[i](1), vertices[i](2) );
    tess.end_polygon();

    assert( norms.size() == vertex_data.size() );


    boost::shared_ptr<VertexBuffer> verts(new VertexBuffer);
    boost::shared_ptr<VertexBuffer> normals(new VertexBuffer);

    verts->set_data(
        &vertex_data[0],
        sizeof(float)*vertex_data.size(), 
        VertexBuffer::STATIC_DRAW);

    normals->set_data(
        &norms[0], 
        sizeof(float)*norms.size(), 
        VertexBuffer::STATIC_DRAW);

    m_mesh.set_vertices(verts);
    m_mesh.set_normals(normals);

    m_mesh.set_element_count(
        vertex_data.size()/3/m_mesh.vertices_per_element());

    calculate_bvs(vertices);
}

void Polygon::add_vertices(const VertexBased::Vertices &vertices) throw()
{
    assert( false ); // not yet implemented!

    /*if( get_vertices.get_size() == 0 )
        set_vertices(vertices);
    else
    {
        boost::scoped_array<uint8_t> data(new uint8_t[get_vertices.get_size()]);
        buf.get_data(&data[0], m_size, 0);
        set_data(data, m_size, m_usage);
        delete[] data;
    }*/
}

void Polygon::get_vertices(VertexBased::Vertices &vertices) const throw()
{
    vertices.clear();

    if( !m_mesh.has_vertices() )
        return;

    const boost::shared_ptr<VertexBuffer> verts = m_mesh.get_vertices();
    const float *ptr = (const float *)verts->map();

    size_t N = m_mesh.get_element_count()*m_mesh.vertices_per_element();
    for( size_t i = 0; i < N; ++i )
        vertices.push_back( Vector3f(ptr[3*i+0], ptr[3*i+1], ptr[3*i+2]) );       

    verts->unmap();
}


void Polygon::calculate_bvs() throw()
{
    if( get_parent() == 0 )
        return;

    VertexBased::Vertices vertices;
    get_vertices(vertices);
    calculate_bvs(vertices);
}

void Polygon::calculate_bvs(const VertexBased::Vertices &vertices) throw()
{
    if( get_parent() == 0 )
        return;

    // TODO: doesn't neccesarily work for indexed primitives!

    //
    // Calculate the midpoint of the vertices
    //
    Vector3f midpoint;
    float r;

    geom::calculate_approximate_bounding_sphere(
        vertices.begin(), vertices.end(),
        midpoint, r,
        get_parent()->get_x_scale(),
        get_parent()->get_y_scale(),
        get_parent()->get_z_scale());

    // Transform the midpoint in to an world-coordinate offset
    // and calculate the BV position
    midpoint = get_parent()->get_mtow() * Vector4f(midpoint, 0);
    
    set_bounding_sphere(
        BoundingSphere(
            get_parent()->get_world_pos() + Vector4f(midpoint, 0), r));
}
