/*
 * Copyright Staffan Gimåker 2008.
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */

#ifndef __PEEKABOT_SERIALIZATION_CONTAINERS_HH
#define __PEEKABOT_SERIALIZATION_CONTAINERS_HH


#include <map>
#include <set>
#include <vector>
#include <list>

#include "SerializationInterface.hh"
#include "DeserializationInterface.hh"


namespace peekabot
{
    // ------------- std::vector -----------

    template<class T>
    SerializationInterface &operator<<(
        SerializationInterface &ar, const std::vector<T> &x)
    {
        ar << (uint32_t)x.size();

        for( size_t i = 0; i < x.size(); ++i )
            ar << x[i];

        return ar;
    }

    template<class T>
    DeserializationInterface &operator>>(
        DeserializationInterface &ar, std::vector<T> &x)
    {
        uint32_t len;
        ar >> len;
        x = std::vector<T>(len);

        for( size_t i = 0; i < x.size(); ++i )
            ar >> x[i];

        return ar;
    }

    // ----------- std::list ----------

    template<class T>
    SerializationInterface &operator<<(
        SerializationInterface &ar, const std::list<T> &x)
    {
        ar << (uint32_t)x.size();

        for( typename std::list<T>::const_iterator it = x.begin(); 
             it != x.end(); ++it )
            ar << *it;

        return ar;
    }

    template<class T>
    DeserializationInterface &operator>>(
        DeserializationInterface &ar, std::list<T> &x)
    {
        uint32_t len;
        ar >> len;
        x.clear();

        for( uint32_t i = 0; i < len; ++i )
        {
            T tmp;
            ar >> tmp;
            x.push_back(tmp);
        }

        return ar;
    }

    // -------------- std::set --------------
    
    template<class T>
    SerializationInterface &operator<<(
        SerializationInterface &ar, const std::set<T> &x)
    {
        ar << (uint32_t)x.size();

        for( typename std::set<T>::const_iterator it = x.begin(); 
             it != x.end(); ++it )
            ar << *it;

        return ar;
    }

    template<class T>
    DeserializationInterface &operator>>(
        DeserializationInterface &ar, std::set<T> &x)
    {
        uint32_t len;
        ar >> len;
        x.clear();

        for( uint32_t i = 0; i < len; ++i )
        {
            T tmp;
            ar >> tmp;
            x.insert(tmp);
        }

        return ar;
    }

    // ------------- std::map -------------

    template<class K, class V>
    SerializationInterface &operator<<(
        SerializationInterface &ar, const std::map<K, V> &x)
    {
        ar << (uint32_t)x.size();

        for( typename std::map<K, V>::const_iterator it = x.begin(); 
             it != x.end(); ++it )
            ar << it->first << it->second;

        return ar;
    }

    template<class K, class V>
    DeserializationInterface &operator>>(
        DeserializationInterface &ar, std::map<K, V> &x)
    {
        uint32_t len;
        ar >> len;
        x.clear();

        for( uint32_t i = 0; i < len; ++i )
        {
            K key; V val;
            ar >> key >> val;
            x.insert(std::make_pair(key, val));
        }

        return ar;
    }
}


#endif // __PEEKABOT_SERIALIZATION_CONTAINERS_HH
