/*
 * Copyright Staffan Gimåker 2008-2009.
 *
 * ---
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */

#ifndef PEEKABOT_SERIALIZATION_DESERIALIZATION_INTERFACE_HH_INCLUDED
#define PEEKABOT_SERIALIZATION_DESERIALIZATION_INTERFACE_HH_INCLUDED


#include <stdexcept>
#include <boost/type_traits.hpp>
#include <boost/mpl/and.hpp>
#include <boost/mpl/not.hpp>
#include <boost/mpl/or.hpp>
#include <boost/utility/enable_if.hpp>

#include "../Bytesex.hh"
#include "Serializable.hh"
#include "SerializableFactory.hh"
#include "DeserializationBuffer.hh"


namespace peekabot
{
    /**
     * \brief An interface for reading binary data independent of their
     * byte order. It's independent of the underlying data source.
     *
     * \sa SerializationInterface
     */
    class DeserializationInterface : public detail::SerializableFriend
    {
    public:
        DeserializationInterface(
            DeserializationBuffer &buf,
            bool buffer_is_big_endian) throw()
            : m_buf(buf),
              m_buffer_is_big_endian(buffer_is_big_endian)
        {
        }

        inline bool is_big_endian() const throw()
        {
            return m_buffer_is_big_endian;
        }

        inline void load_binary(void *x, size_t n)
        {
            m_buf.read(x, n);
        }

        // Non-pointer, non-POD
        /*template<class T> inline
        typename boost::disable_if_c<
            boost::mpl::or_<
                boost::is_pointer<T>,
                boost::is_pod<T>
                >::value, 
            DeserializationInterface &>::type 
        operator>>(T &t)
        {
            load(t, *this);
            return *this;
            }*/
        
        // Pointer to a Serializable-derivative, which
        // we expect to be registered!
        template<class T> inline
        typename boost::enable_if_c<
            boost::is_base_and_derived<Serializable, typename boost::remove_pointer<T>::type>::value, 
            DeserializationInterface &>::type 
        operator>>(T* &t)
        {
            SerializableFactory::IdType id;
            *this >> id; // Load class ID
            t = dynamic_cast<T *>(TheSerializableFactory::instance().create(id));
            load(*t, *this);
            return *this;
        }

        // Enum
        template<class T> inline
        typename boost::enable_if_c<
            boost::is_enum<T>::value, 
            DeserializationInterface &>::type 
        operator>>(T &t)
        {
            int32_t tmp;
            *this >> tmp;
            t = (T)tmp;
            return *this;
        }

        // bool
        template<class T> inline
        typename boost::enable_if_c<
            boost::is_same<bool, T>::value,
            DeserializationInterface &>::type 
        operator>>(T &t)
        {
            uint8_t tmp;
            *this >> tmp;
            t = (T)tmp;
            return *this;
        }

        // POD that is not
        //   1) a pointer or (=> not an array),
        //   2) an enum
        template<class T> inline
        typename boost::enable_if_c<
            boost::mpl::and_<
                boost::is_arithmetic<T>,
                boost::mpl::not_< boost::is_pointer<T> >,
                boost::mpl::not_< boost::is_same<T, bool> >
                >::value,
            DeserializationInterface &>::type 
        operator>>(T &t)
        { 
            load_binary(&t, sizeof(T));
            if( sizeof(T) > 1 && 
                m_buffer_is_big_endian != (PEEKABOT_BYTE_ORDER == PEEKABOT_BIG_ENDIAN) )
                switch_byte_order(&t, 1);
            return *this;
        }

        // Array of PODs that are not
        //   1) pointers, or
        //   2) enums
        template<class T, size_t n> inline
        typename boost::enable_if_c<
            boost::mpl::and_<
                boost::is_arithmetic<T>, 
                boost::mpl::not_< boost::is_pointer<T> >,
                boost::mpl::not_< boost::is_same<T, bool> >
                >::value,
            DeserializationInterface &>::type 
        operator>>(T (&t)[n])
        {
            // An array of PODs is a POD, and thus its memory representation
            // is contiguous => we can do an efficient load!
            load_binary(t, sizeof(T)*n);
            if( sizeof(T) > 1 && 
                m_buffer_is_big_endian != (PEEKABOT_BYTE_ORDER == PEEKABOT_BIG_ENDIAN) )
                switch_byte_order(t, n);
            return *this;
        }

        // Array of non-PODs, pointers or enums
        template<class T, size_t n> inline
        typename boost::enable_if_c<
            boost::mpl::or_<
                boost::mpl::not_< boost::is_arithmetic<T> >,
                boost::is_pointer<T>,
                boost::is_same<T, bool>
                >::value, 
            DeserializationInterface &>::type 
        operator>>(T (&t)[n])
        {
            for( size_t i = 0; i < n; ++i )
                *this >> t[i];
            return *this;
        }

    private:
        DeserializationBuffer &m_buf;

        bool m_buffer_is_big_endian;
    };


    inline DeserializationInterface &operator>>(DeserializationInterface &ar, std::string &str)
    {
        uint32_t len;
        ar.load_binary(&len, sizeof(uint32_t));
        if( len > 0 )
        {
            char *buf = new char[len+1];
            ar.load_binary(buf, len);
            buf[len] = 0;
            str = buf;
            delete[] buf;
        }
        else
            str = "";

        return ar;
    }
}


#endif // PEEKABOT_SERIALIZATION_DESERIALIZATION_INTERFACE_HH_INCLUDED
