/*
 * Copyright Staffan Gimåker 2008-2009.
 *
 * ---
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */

#ifndef PEEKABOT_SERIALIZATION_SERIALIZABLE_FACTORY_HH_INCLUDED
#define PEEKABOT_SERIALIZATION_SERIALIZABLE_FACTORY_HH_INCLUDED


#include <map>
#include <string>
#include <stdexcept>
#include <boost/function.hpp>
#include <boost/format.hpp>
#include <boost/cstdint.hpp>
#include "../Singleton.hh"
#include "Serializable.hh"


namespace peekabot
{
    class SerializableFactory
    {
        friend class SerializationInterface;
        friend class DeserializationInterface;

        typedef boost::function<Serializable *(void)> CreateFun;

    public:
        struct UnregisteredClass : public std::runtime_error
        {
            UnregisteredClass(const std::string &msg) 
                : std::runtime_error(msg) {}
        };

        typedef boost::uint16_t IdType;

        template<class T>
        void register_class(IdType id) throw(std::runtime_error)
        {
            m_type_to_id.insert(std::make_pair(typeid(T).name(), id));
            if( !m_id_to_create_fun.insert(
                    std::make_pair(id, CreateFun(&create<T>))).second )
            {
                throw std::runtime_error(
                    (boost::format("Id already registered, id=%1%")%id).str());
            }
        }

        void deregister_all() throw()
        {
            m_type_to_id.clear();
            m_id_to_create_fun.clear();
        }

    private:
        typedef std::map<const char *, IdType> TypeToId;
        typedef std::map<IdType, CreateFun> IdToCreateFun;


        Serializable *create(IdType id)
        {
            IdToCreateFun::const_iterator it;
            it = m_id_to_create_fun.find(id);

            if( it == m_id_to_create_fun.end() )
                throw UnregisteredClass(
                    (boost::format("Class not registered, id=%1%")%id).str());
        
            Serializable *prod = it->second();
            assert( prod != 0 );
            return prod;
        }

        IdType get_id(const Serializable *x) const throw(std::runtime_error)
        {
            std::map<const char *, IdType>::const_iterator it = m_type_to_id.find(
                typeid(*x).name());

            if( it == m_type_to_id.end() )
                throw UnregisteredClass("Class not registered");

            return it->second;
        }

    private:
        template<class T>
        static Serializable *create()
        {
            return new T();
        }

        TypeToId m_type_to_id;
        IdToCreateFun m_id_to_create_fun;
    };

    typedef Singleton<SerializableFactory> TheSerializableFactory;
}

#endif // PEEKABOT_SERIALIZATION_SERIALIZABLE_FACTORY_HH_INCLUDED
