/*
 * Copyright Staffan Gimåker 2008-2009.
 *
 * ---
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or copy at
 * http://www.boost.org/LICENSE_1_0.txt)
 */

#ifndef PEEKABOT_SERIALIZATION_SERIALIZATION_INTERFACE_HH_INCLUDED
#define PEEKABOT_SERIALIZATION_SERIALIZATION_INTERFACE_HH_INCLUDED


#include <stdexcept>
#include <boost/type_traits.hpp>
#include <boost/mpl/and.hpp>
#include <boost/mpl/not.hpp>
#include <boost/mpl/or.hpp>
#include <boost/utility/enable_if.hpp>

#include "Serializable.hh"
#include "SerializableFactory.hh"
#include "SerializationBuffer.hh"


namespace peekabot
{
    /**
     * \brief An interface for writing binary data in a format compatible with
     * DeserializationBuffer. It's independent of the underlying data source.
     *
     * \sa DeserializationInterface
     */
    class SerializationInterface : public detail::SerializableFriend
    {
    public:
        SerializationInterface(SerializationBuffer &buf) throw()
            : m_buf(buf)
        {
        }

        inline void save_binary(const void *x, size_t n)
        {
            m_buf.write(x, n);
        }

        // Non-pointer, non-POD
        /*template<class T> inline
        typename boost::disable_if_c<
            boost::mpl::or_<
                boost::is_pointer<T>,
                boost::is_pod<T>
                >::value, 
            SerializationInterface &>::type 
        operator<<(const T &t)
        {
            save(t, *this);
            return *this;
        }*/
        
        // Pointer to a Serializable-derivative, which
        // we expect to be registered!
        template<class T> inline
        typename boost::enable_if_c<
            boost::is_base_and_derived<Serializable, T>::value, 
            SerializationInterface &>::type 
        operator<<(const T *t)
        { 
            *this << TheSerializableFactory::instance().get_id(t); // Save class ID
            save(*t, *this);
            return *this;
        }

        // Enum
        template<class T> inline
        typename boost::enable_if_c<
            boost::is_enum<T>::value, 
            SerializationInterface &>::type 
        operator<<(const T t)
        {
            int32_t tmp = (int32_t)t;
            *this << tmp;
            return *this;
        }

        // bool
        template<class T> inline
        typename boost::enable_if_c<
            boost::is_same<bool, T>::value,
            SerializationInterface &>::type 
        operator<<(const T t)
        {
            uint8_t tmp = (uint8_t)t;
            *this << tmp;
            return *this;
        }

        // POD that is not
        //   1) a pointer or (=> not an array),
        //   2) an enum
        template<class T> inline
        typename boost::enable_if_c<
            boost::mpl::and_<
                boost::is_arithmetic<T>,
                boost::mpl::not_< boost::is_pointer<T> >,
                boost::mpl::not_< boost::is_same<T, bool> >
                >::value,
            SerializationInterface &>::type 
        operator<<(const T &t)
        {
            save_binary(&t, sizeof(T));
            return *this;
        }

        // Array of PODs that are not
        //   1) pointers, or
        //   2) enums
        template<class T, size_t n> inline
        typename boost::enable_if_c<
            boost::mpl::and_<
                boost::is_arithmetic<T>, 
                boost::mpl::not_< boost::is_pointer<T> >,
                boost::mpl::not_< boost::is_same<T, bool> >
                >::value,
            SerializationInterface &>::type 
        operator<<(const T (&t)[n])
        {
            // An array of PODs is a POD, and thus its memory representation
            // is contiguous => we can do an efficient save!
            save_binary(t, sizeof(T)*n);
            return *this;
        }

        // Array of non-PODs, pointers or enums
        template<class T, size_t n> inline
        typename boost::enable_if_c<
            boost::mpl::or_<
                boost::mpl::not_< boost::is_arithmetic<T> >,
                boost::is_pointer<T>,
                boost::is_same<T, bool>
                >::value, 
            SerializationInterface &>::type 
        operator<<(const T (&t)[n])
        {
            for( size_t i = 0; i < n; ++i )
                *this << t[i];
            return *this;
        }

    private:
        SerializationBuffer &m_buf;
    };


    inline SerializationInterface &operator<<(SerializationInterface &ar, const std::string &str)
    {
        uint32_t len = (uint32_t)str.length();
        ar.save_binary(&len, sizeof(uint32_t));
        ar.save_binary(str.c_str(), len);
        return ar;
    }
}


#endif // PEEKABOT_SERIALIZATION_SERIALIZATION_INTERFACE_HH_INCLUDED
